/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <endian.h>
#include <stdlib.h>
#include <string.h>

#include "cons_buf.h"
#include "pctextmode.h"
#include "term_dyn.h"

/*

offset | length | type    | description
-------+--------+---------+--------------
     0 |     22 | pstring | Identifier "TheDraw Save File",0x0d,0x0a,0x1a.
    22 |    ??? | pstring | Version
    27 |      2 |  uint16 | Cursor X (from 1)
    29 |      2 |  uint16 | Cursor Y (from 1)
    31 |      2 |  uint16 | Selected page (from 1)
    33 |      2 |  uint16 | ? = 2 or 8
    35 |      2 |  uint16 | ? = 1
    37 |      2 |  uint16 | ? = 1
    39 |      2 |  uint16 | Width (=80)
    41 |      2 |  uint16 | Height (>=50)
    43 |      2 |  uint16 | ? = 1

    56 |      1 |   uint8 | Pages
    57 |      1 |   uint8 | 0 or 1
    58 |      1 |   uint8 | 0 or 1
    59 |      1 |   uint8 | 0 or 1
    60 |      1 |   uint8 | 0 or 1
    61 |      1 |   uint8 | 0 or 1
    62 |      1 |   uint8 | 0 or 1

   162 |      1 |   uint8 | Selected character palette

                                       00 00 00
00 60 0e 00 00 00 00 02 -- 00 00 00 00 00 00 50
c4 c4 c4 c4 c4 c4 c4 c4 c4 fe c4 c4 c4 c4 c4 c4

@144:

24 0f 00 00 00 00 00 00 00 d4 00 00 00 00 50 00
32 00 01 00 50

@144:

37 0f 00 00 00 00 00 00 00 d4 00 00 00 00 91 ab
92 ab 01 00 50

@336:

00 00 00 00 50 00 32 00 01 00 01 00 00 20 7e 54
7e 68 65 44 72 61 77 20 7e 4c 7e 69 62 72 00

@336:

00 00 00 00 50 00 32 00 01 00 01 00 00 00 00 00
00 00 00 00 00 00 00 00 00 44 3a 5c 55 54 00

   367 |      2 |  uint16 | Length of compressed data
   369 |        |         | Compressed data

*/

typedef struct{
    cairo_surface_t *surface[2];
    int frame_count;
}page_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    page_t *page;
};


static const char **
_td_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-thedraw",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_td_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->page = NULL;
    h->info->threadsafe = 1;
    return h;
}


static void
_td_free(abydos_plugin_handle_t *h)
{
    if(h->page){
        int i;
        for(i=0;i<h->info->page_count;++i){
            int j;
            for(j=0;j<h->page[i].frame_count;++j)
                if(h->page[i].surface[j])
                    cairo_surface_destroy(h->page[i].surface[j]);
        }
        free(h->page);
    }
    free(h);
}


#define get16(a) le16toh(*(uint16_t*)(a))


static int
_uncompress(Terminal *term, const uint8_t *src, size_t len)
{
    const uint8_t *end = src+len;
    for(;;){
        int b;
        if(src >= end)
            return 0;
        b = *src++;
        if(b <= 15){
            /* change foreground color */
            term_dyn_foreground(term,b);
        }else if(b <= 23){
            /* change background color (exluding the blink bit) */
            term_dyn_background(term,(term_dyn_get_background(term)&0x8)|(b-16));
        }else if(24==b){
            /* next line */
            term_dyn_write(term,'\r');
        }else if(25==b){
            /* Multi space output */
            if(src >= end)
                return -1;
            b = *src++ + 1;
            while(b-->0)
                term_dyn_write(term,' ');
        }else if(26==b){
            /* Multi output */
            int ch;
            if(src+2 > end)
                return -1;
            b = *src++ + 1;
            ch = *src++;
            while(b-->0)
                term_dyn_write(term,ch);
        }else if(27==b){
            /* toggle blink bit */
            term_dyn_background(term,term_dyn_get_background(term)^0x8);
        }else if(b <= 31){
            /* invalid */
            return -1;
        }else{
            /* write character */
            term_dyn_write(term,b);
        }
    }
}


static int
_td_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    int res;
    Console *cons;
    Terminal *term;
    int width;
    int height;
    int i;

    if(len < 370){
        h->info->error = "Data too short";
        return -1;
    }
    if(memcmp(data+1,"TheDraw Save File",17)){
        h->info->error = "Invalid file";
        return -1;
    }

    h->info->pixel_ratio = 20.0/27;
    width = get16(data+39);
    height = get16(data+41);
    h->info->width = width*9;
    h->info->height = height*16;
    h->info->default_page = get16(data+31)-1;
    h->info->page_count = data[56];
    DEBUG_p("dimensions: %dx%d",width,height);
    data += 367;
    len += 367;

    h->page = (page_t*)malloc(sizeof(page_t)*h->info->page_count);
    cons = cons_buf_new();
    cons_buf_set_size(cons,width,height,0x07);
    term = term_dyn_new(&cons_buf_if,cons);
    for(i=0;i<h->info->page_count;++i){
        int size = get16(data);
        term_dyn_gotoxy(term,0,0);
        term_dyn_set_attribute(term,0x07);
        term_dyn_clrscr(term);
        if(size > len){
            DEBUG_p("data too short");
            res = -1;
        }else{
            res = _uncompress(term, (uint8_t*)data+2, size);
            size += 2;
            data += size;
            len -= size;
        }
        if(res >= 0){
            h->page[i].surface[0] = abydos_image_surface_from_console(cons, S_ON, 9);
            if(abydos_console_has_blink(cons)){
                h->page[i].frame_count = 2;
                h->page[i].surface[1] = abydos_image_surface_from_console(cons, S_OFF, 9);
            }else{
                h->page[i].frame_count = 1;
            }
        }else{
            h->page[i].frame_count = 1;
            h->page[i].surface[0] = NULL;
        }
    }
    term_dyn_free(term);
    cons_buf_free(cons);
    return res;
}


static cairo_surface_t *
_td_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    DEBUG_p("page: %d, frame: %d",page,frame);
    return cairo_surface_reference(h->page[page].surface[frame]);
}


static void
_td_get_page_info(abydos_plugin_handle_t *h, int page, abydos_plugin_page_info_t *info)
{
    info->frame_count = h->page[page].frame_count;
}


static void
_td_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->page[i].frame_count;++i)
        duration[i] = BLINK_DURATION;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _td_get_mime_types,
        .new = _td_new,
        .free = _td_free,
        .create_from_data = _td_create_from_data,
        .get_image_surface = _td_get_image_surface,
        .get_page_info = _td_get_page_info,
        .get_duration = _td_get_duration
    };
    return &p;
}
