/*
 * PI1 image loader
 *
 * Copyright (C) 2004 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */

#include <abydos-plugin.h>
#include <debug.h>
#include <endian.h>
#include <nil-atarist.h>
#include <nil-cycle.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>                     /* malloc() */
#include <string.h>                     /* strcasecmp() */

#include "atarist.h"


#define ENABLE_PC1

#define DEGAS_ANIM_LEFT 0
#define DEGAS_ANIM_OFF 1
#define DEGAS_ANIM_RIGHT 2

typedef struct{
    uint16_t left_color[4];
    uint16_t right_color[4];
    uint16_t direction[4];
    uint16_t delay[4];
}degas_animation_t;

typedef struct{
    uint16_t resolution;
    uint16_t palette[16];
    uint8_t packed[];
}pc1_t;

typedef struct{
    uint16_t resolution;
    uint16_t palette[16];
    uint16_t image[16000];
    degas_animation_t animation;
}pi1_t;

typedef cairo_surface_t*(*decode_image_func_t)(uint16_t*, atarist_cairo_color_t*);

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    enum{
        T_PI1,
        T_PC1
    }type;
    int resolution;
    decode_image_func_t decode_image;
    atarist_cairo_color_t cairo_palette[16];
    uint16_t atari_image[16000];
    nil_cycle_t cycle[4];
    cairo_surface_t *cached_surface;
};


static const char **
_degas_get_mime_types()
{
    static const char *mime_types[] = {
#ifdef ENABLE_PC1
        "image/x-pc1",
        "image/x-pc2",
        "image/x-pc3",
#endif
        "image/x-pi1",
        "image/x-pi2",
        "image/x-pi3",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_degas_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
#ifdef ENABLE_PC1
    if(!strcasecmp(mime_type,"image/x-pc1")||
       !strcasecmp(mime_type,"image/x-pc2")||
       !strcasecmp(mime_type,"image/x-pc3"))
    {
        h->type = T_PC1;
    }else
#endif
    {
        h->type = T_PI1;
    }
    h->info = info;
    info->threadsafe = 1;
    h->cached_surface = NULL;
    return h;
}


static void
_degas_free(abydos_plugin_handle_t *h)
{
    if(h->cached_surface)
        cairo_surface_destroy(h->cached_surface);
    free(h);
}


#ifdef ENABLE_PC1
static int
_packbits_unpack(uint8_t *src, int len, uint16_t *atari_image, int resolution)
{
    uint8_t linebuffer[160];
    int x;
    int y;
    for(y=0;y<200;++y){
        for(x=0;x<160;){
            int b;
            if(len<2){
                DEBUG_p("out of data");
                return -1;
            }
            b = *src++;
            --len;
            if(b>=128){
                b = 257-b;
                if(x+b>160){
                    DEBUG_p("invalid runlength");
                    return -1;
                }
                memset(linebuffer+x,*src,b);
                ++src;
                --len;
            }else{
                ++b;
                if(x+b>160){
                    DEBUG_p("invalid data");
                    return -1;
                }else if(b>len){
                    DEBUG_p("out of data");
                    return -1;
                }
                memcpy(linebuffer+x,src,b);
                src += b;
                len -= b;
            }
            x += b;
        }
        switch(resolution){
        case 0:
            for(x=0;x<20;++x){
                atari_image[y*80+x*4+0] = ((uint16_t*)linebuffer)[x+0];
                atari_image[y*80+x*4+1] = ((uint16_t*)linebuffer)[x+20];
                atari_image[y*80+x*4+2] = ((uint16_t*)linebuffer)[x+40];
                atari_image[y*80+x*4+3] = ((uint16_t*)linebuffer)[x+60];
            }
            break;
        case 1:
            for(x=0;x<40;++x){
                atari_image[y*80+x*2+0] = ((uint16_t*)linebuffer)[x+0];
                atari_image[y*80+x*2+1] = ((uint16_t*)linebuffer)[x+40];
            }
            break;
        case 2:
            memcpy(atari_image+y*80,linebuffer,160);
            break;
        }
    }
    return len;
}
#endif


static int
_setup_resolution(abydos_plugin_handle_t *h)
{
    switch(h->resolution){
    case 3:
        /* Sometimes set to 3 for low resolution. */
        h->resolution = 0;
        /* fall through */
    case 0:
        h->info->width = 320;
        h->info->height = 200;
        h->info->pixel_ratio = 5.0/6;
        h->decode_image = atarist_decode_image_0;
        break;
    case 1:
        h->info->width = 640;
        h->info->height = 200;
        h->info->pixel_ratio = 5.0/12;
        h->decode_image = atarist_decode_image_1;
        break;
    case 2:
        h->info->width = 640;
        h->info->height = 400;
        h->info->pixel_ratio = 5.0/6;
        h->decode_image = atarist_decode_image_2;
        break;
    default:
        return -1;
    }
    return 0;
}


static void
_setup_animation(abydos_plugin_handle_t *h, degas_animation_t *a)
{
    int max_colors;
    int i;
    switch(h->resolution){
    case 0:
        max_colors = 16;
        break;
    case 1:
        max_colors = 4;
        break;
    default:
        /* For high reolution no color cycling should occur. */
        max_colors = 0;
        break;
    }
    for(i=0;i<4;++i){
        int ticks;
        h->cycle[i].first = be16toh(a->left_color[i]);
        h->cycle[i].last = be16toh(a->right_color[i]);
        if(h->cycle[i].first >= h->cycle[i].last || h->cycle[i].last >= max_colors){
            /* Invalid range specified, disable this cycle. */
            DEBUG_p("invalid color range for cycle %d, disabling",i);
            h->cycle[i].direction = DEGAS_ANIM_OFF;
            continue;
        }
        switch(be16toh(a->direction[i])){
        case DEGAS_ANIM_LEFT:
            h->cycle[i].direction = NIL_CYCLE_LEFT;
            break;
        case DEGAS_ANIM_RIGHT:
            h->cycle[i].direction = NIL_CYCLE_RIGHT;
            break;
        case DEGAS_ANIM_OFF:
        default:
            h->cycle[i].direction = NIL_CYCLE_OFF;
            break;
        }
        //DEBUG_p("delay value is %d",(int)ntohs(a->delay[i]));
        ticks = 128 - be16toh(a->delay[i]);
        //DEBUG_p("delay is %d",h->anim[i].delay);
        h->cycle[i].duration = (double)ticks / 60;
        //DEBUG_p("duration is %f seconds",h->anim[i].frame_duration);
    }
    /* FIXME!! disable cycles that overlap, calculate least common multiple for all steps */
    h->info->frame_count = nil_cycle_total_frames(h->cycle,1);
    DEBUG_p("frame count: %d",h->info->frame_count);
}


static int
_degas_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    switch(h->type){
    case T_PI1:
        if(len < offsetof(pi1_t,animation)){
            h->info->error = "data is too short";
            return -1;
        }
        {
            pi1_t *pi1 = (pi1_t*)data;
            h->resolution = be16toh(pi1->resolution)&3;
            if(_setup_resolution(h)<0){
                h->info->error = "invalid resolution specified";
                return -1;
            }
            if(len >= sizeof(pi1_t)){
                /* animation is included */
                DEBUG_p("animation is included");
                _setup_animation(h, &pi1->animation);
            }
            atarist_decode_palette(pi1->palette, h->cairo_palette);
            if(h->info->frame_count > 1){
                memcpy(h->atari_image,pi1->image,16000*sizeof(uint16_t));
            }else{
                h->cached_surface = h->decode_image(pi1->image, h->cairo_palette);
            }
        }
        break;
#ifdef ENABLE_PC1
    case T_PC1:
        {
            pc1_t *pc1 = (pc1_t*)data;
            int res;
            h->resolution = be16toh(pc1->resolution)&3;
            _setup_resolution(h);
            res = _packbits_unpack(pc1->packed,len-sizeof(pc1_t),h->atari_image,h->resolution);
            if(res<0){
                h->info->error = "invalid data";
                return res;
            }
            if(res >= sizeof(degas_animation_t))
                _setup_animation(h, (degas_animation_t*)(data+len-sizeof(degas_animation_t)));
            atarist_decode_palette(pc1->palette, h->cairo_palette);
            if(h->info->frame_count <= 1){
                h->cached_surface = h->decode_image(h->atari_image, h->cairo_palette);
            }
        }
        break;
#endif
    }
    return 0;
}


static cairo_surface_t *
_degas_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant,double time,int frame)
{
    if(h->info->frame_count > 1){
        atarist_cairo_color_t cairo_palette[16];
        memcpy(cairo_palette,h->cairo_palette,16*sizeof(atarist_cairo_color_t));
        nil_cycle_apply_time(&h->cycle[0], time, cairo_palette, sizeof(atarist_cairo_color_t));
        return h->decode_image(h->atari_image, cairo_palette);
    }else{
        return cairo_surface_reference(h->cached_surface);
    }
}


static void
_degas_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->cycle[0].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _degas_get_mime_types,
        .new = _degas_new,
        .free = _degas_free,
        .create_from_data = _degas_create_from_data,
        .get_image_surface = _degas_get_image_surface,
        .get_duration = _degas_get_duration
    };
    return &p;
}
