/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <ctype.h>                      /* tolower() */
#include <debug.h>
#include <string.h>

#include "cgmread.h"
#include "cgmread-text.h"


static const uint8_t *
_get_token(const uint8_t *data, const uint8_t *end, const char **r_token, ssize_t *r_len)
{
    enum{
        S_BEFORE = 0,
        S_IN_VALUE,
        S_IN_STRING,
        S_AFTER
    }state = 0;
    int in_comment = 0;
    *r_token = NULL;
    *r_len = -1;

    for(;;){
        if(data >= end){
            if(*r_token && *r_len < 0)
                *r_len = (char*)data - *r_token;
            return data;
        }
        if(in_comment){
            if('%'==*data)
                in_comment = 0;
        }else{
            switch(state){
            case S_BEFORE:
                switch(*data){
                case ' ':
                case '\n':
                case '\r':
                case '\t':
                case ',':
                    break;
                case '%':
                    in_comment = 1;
                    break;
                case '"':
                    *r_token = (char*)data+1;
                    state = S_IN_STRING;
                    break;
                case ';':
                    /* shouldn't happen */
                    return data;
                default:
                    *r_token = (char*)data;
                    state = S_IN_VALUE;
                    break;
                }
                break;
            case S_IN_VALUE:
                switch(*data){
                case ' ':
                case '\n':
                case '\r':
                case '\t':
                case ',':
                case '(':
                case ')':
                    *r_len = (char*)data - *r_token;
                    state = S_AFTER;
                    break;
                case ';':
                    *r_len = (char*)data - *r_token;
                    return data;
                }
                break;
            case S_IN_STRING:
                switch(*data){
                case '"':
                    *r_len = (char*)data - *r_token;
                    state = S_AFTER;
                    break;
                }
                break;
            case S_AFTER:
                switch(*data){
                case ' ':
                case '\n':
                case '\r':
                case '\t':
                case ',':
                    break;
                case '%':
                    in_comment = 1;
                    break;
                default:
                    return data;
                }
                break;
            }
        }
        ++data;
    }
}


static int
_atoin(const char* s,size_t len)
{
    const char *end = s+len;
    int neg;
    int res;
    while(s<end && *s!='-' && (*s<'0' || *s>'9'))
        ++s;
    if(s>=end)
        return 0;
    if('-'==*s){
        neg = 1;
        ++s;
    }else{
        neg = 0;
    }
    res = 0;
    while(s<end && *s>='0' && *s<='9'){
        res *= 10;
        res += *s-'0';
        ++s;
    }
    return neg ? -res : res;
}


static double
_atofn(const char *s,size_t len){
    const char *end = s+len;
    int neg;
    double res;
    int frac;
    int in_frac;

    while(s<end && *s!='-' && *s!='.' && (*s<'0' || *s>'9'))
        ++s;
    if(s>=end)
        return 0;
    if('-'==*s){
        neg = 1;
        ++s;
    }else{
        neg = 0;
    }
    res=0;
    frac = 0;
    in_frac = 0;
    while(s<end){
        if('.'==*s){
             if (in_frac)
                return 0;
            in_frac=1;
        }else{
            res *= 10;
            res += *s-'0';
            if(in_frac)
                --frac;
        }
        ++s;
    }
    res *= pow(10,frac);
    return neg ? -res : res;
}


static int
_text_eoe(cgm_read_t *r){
    return r->data >= r->end || ';'==*r->data;
}


static int
_text_get_int(cgm_read_t *r)
{
    const char *token;
    ssize_t token_len;
    DEBUG_p("begin");
    r->data = _get_token(r->data, r->end, &token, &token_len);
    if(token_len <= 0)
        return 0;
    return _atoin((char*)token, token_len);
}


static unsigned int
_text_get_uint(cgm_read_t *r)
{
    const char *token;
    ssize_t token_len;
    DEBUG_p("begin");
    r->data = _get_token(r->data, r->end, &token, &token_len);
    if(token_len <= 0)
        return 0;
    return _atoin((char*)token, token_len);
}


static double
_text_get_real(cgm_read_t *r)
{
    const char *token;
    ssize_t token_len;
    DEBUG_p("begin");
    r->data = _get_token(r->data, r->end, &token, &token_len);
    if(token_len <= 0)
        return 0;
    return _atofn((char*)token, token_len);
}


static const char *
_text_get_string(cgm_read_t *r, int *r_len)
{
    const char *token;
    ssize_t token_len;
    DEBUG_p("begin");
    r->data = _get_token(r->data, r->end, &token, &token_len);
    if(token_len <= 0)
        return NULL;
    *r_len = token_len;
    return (char*)token;
}


static double
_text_get_vdc(cgm_read_t *r)
{
    DEBUG_p("begin");
    switch(r->vdc_type){
    case VDC_INT:
    default:
        return _text_get_int(r);
    case VDC_REAL:
        return _text_get_real(r);
    }
}


static int
_text_get_int_precision_bytes(cgm_read_t *r)
{
    int range;
    DEBUG_p("begin");
    range = _text_get_int(r);
    if(range < 0)
        range = _text_get_int(r) - range;
    if(range <= UINT8_MAX)
        return 1;
    if(range <= UINT16_MAX)
        return 2;
    return 4;
}


vdc_type_t
_text_get_vdc_type(cgm_read_t *r)
{
    const char *token;
    ssize_t token_len;
    DEBUG_p("begin");
    r->data = _get_token(r->data, r->end, &token, &token_len);
    return token_len > 0 && !strncasecmp(token, "REAL", token_len) ? VDC_REAL : VDC_INT;
}


static ScalingMode
_text_get_scaling_mode(cgm_read_t *r)
{
    const char *token;
    ssize_t token_len;
    DEBUG_p("begin");
    r->data = _get_token(r->data, r->end, &token, &token_len);
    return token_len > 0 && !strncasecmp(token, "MATRIC", token_len) ? SM_METRIC : SM_ABSTRACT;
}


static ColorSelectionMode
_text_get_color_selection_mode(cgm_read_t *r)
{
    const char *token;
    ssize_t token_len;
    DEBUG_p("begin");
    r->data = _get_token(r->data, r->end, &token, &token_len);
    return token_len > 0 && !strncasecmp(token, "DIRECT", token_len) ? CSM_DIRECT : CSM_INDEXED;
}


static close_type_t
_text_get_close_type(cgm_read_t *r)
{
    const char *token;
    ssize_t token_len;
    DEBUG_p("begin");
    r->data = _get_token(r->data, r->end, &token, &token_len);
    return token_len > 0 && !strncasecmp(token, "CHORD", token_len) ? CT_CHORD : CT_PIE;
}


static TextPrecision
_text_get_text_precision(cgm_read_t *r)
{
    const char *token;
    ssize_t token_len;
    DEBUG_p("begin");
    r->data = _get_token(r->data, r->end, &token, &token_len);
    if(token_len > 0){
        if(!strncasecmp(token, "STRING", token_len))
            return TPR_STRING;
        if(!strncasecmp(token, "CHARACTER", token_len))
            return TPR_CHARACTER;
        if(!strncasecmp(token, "STROKE", token_len))
            return TPR_STROKE;
    }
    return TPR_UNDEFINED;
}


static TextPath
_text_get_text_path(cgm_read_t *r)
{
    const char *token;
    ssize_t token_len;
    DEBUG_p("begin");
    r->data = _get_token(r->data, r->end, &token, &token_len);
    if(token_len > 0){
        if(!strncasecmp(token, "LEFT", token_len))
            return TPR_LEFT;
        if(!strncasecmp(token, "UP", token_len))
            return TPR_UP;
        if(!strncasecmp(token, "DOWN", token_len))
            return TPR_DOWN;
    }
    return TPR_RIGHT;
}


static FinalFlag
_text_get_final_flag(cgm_read_t *r)
{
    const char *token;
    ssize_t token_len;
    DEBUG_p("begin");
    r->data = _get_token(r->data, r->end, &token, &token_len);
    return token_len > 0 && !strncasecmp(token, "FINAL", token_len) ? FF_FINAL : FF_NOT_FINAL;
}


static SpecMode
_text_get_spec_mode(cgm_read_t *r)
{
    const char *token;
    ssize_t token_len;
    DEBUG_p("begin");
    r->data = _get_token(r->data, r->end, &token, &token_len);
    return token_len > 0 && !strncasecmp(token, "SCALED", token_len) ? SM_SCALED : SM_ABSOLUTE;
}


static EdgeVisibility
_text_get_edge_visibility(cgm_read_t *r)
{
    const char *token;
    ssize_t token_len;
    DEBUG_p("begin");
    r->data = _get_token(r->data, r->end, &token, &token_len);
    return token_len > 0 && !strncasecmp(token, "ON", token_len) ? EV_ON : EV_OFF;
}


static edge_flag_t
_text_get_edge_flag(cgm_read_t *r)
{
    const char *token;
    ssize_t token_len;
    DEBUG_p("begin");
    r->data = _get_token(r->data, r->end, &token, &token_len);
    if(token_len > 0){
        if(!strncasecmp(token, "VIS", token_len))
            return EF_VISIBLE;
        if(!strncasecmp(token, "CLOSEVIS", token_len))
            return EF_CLOSE_VISIBLE;
    }
    return EF_INVISIBLE;
}


static MarkerType
_text_get_marker_type(cgm_read_t *r)
{
    DEBUG_p("not implement");
}


static int
_xstrncasecmp(const char *a, size_t len, const char *b){
    for(;;){
        if(len>0 && *b){
            if(tolower(*a) != tolower(*b))
                return tolower(*a) - tolower(*b);
        }else if(len>0){
            return 1;
        }else if(*b){
            return -1;
        }else{
            return 0;
        }
        ++a;
        --len;
        ++b;
    }
}


static cgm_element_t cgm_element[] = {
    {"BegMF", 0, 0x01},
    {"EndMF", 0, 0x02},
    {"BegPic", 0, 0x03},
    {"BegPicBody", 0, 0x04},
    {"EndPic", 0, 0x05},

    {"MFVersion", 1, 0x01},
    {"MFDesc", 1, 0x02},
    {"VDCType", 1, 0x03},
    {"IntegerPrec", 1, 0x04},
    {"ColrPrec", 1, 0x07},
    {"ColrIndexPrec", 1, 0x08},
    {"MaxColrIndex", 1, 0x09},
    {"MFElemList", 1, 0x0b},
    {"FontList", 1, 0x0d},

    {"ColrMode", 2, 0x02}, /* ? */
    {"LineWidthMode", 2, 0x03},
    {"MarkerSizeMode", 2, 0x04},
    {"EdgeWidthMode", 2, 0x05},
    {"VDCExt", 2, 0x06},
    {"BackColr", 2, 0x07},
    {"IntStyle", 2, 0x10}, /* ? */

    {"Line", 4, 0x01},
    {"DisjtLine", 4, 0x02},
    {"Marker", 4, 0x03}, /* ? */
    {"Text", 4, 0x04},
    {"RestrText", 4, 0x05},
    {"ApndText", 4, 0x06},
    {"Polygon", 4, 0x07},
    {"PolygonSet", 4, 0x08},
    {"CellArray", 4, 0x09},
    {"Rect", 4, 0x0b},
    {"Circle", 4, 0x0c},
    {"Arc3Pt", 4, 0x0d},
    {"Arc3PtClose", 4, 0x0e},
    {"ArcCtr", 4, 0x0f},
    {"ArcCtrClose", 4, 0x10},
    {"Ellipse", 4, 0x11},
    {"EllipArc", 4, 0x12},
    {"EllipArcClose", 4, 0x13},

    {"LineType", 5, 0x02},
    {"LineWidth", 5, 0x03},
    {"LineColr", 5, 0x04},
    {"MarkerType", 5, 0x06},
    {"MarkerSize", 5, 0x07},
    {"MarkerColr", 5, 0x08},
    {"TextFontIndex", 5, 0x0a},
    {"TextPrec", 5, 0x0b},
    {"CharExpan", 5, 0x0c},
    {"CharSpace", 5, 0x0d},
    {"TextColr", 5, 0x0e},
    {"CharHeight", 5, 0x0f},
    {"CharOri", 5, 0x10},
    {"TextPath", 5, 0x11},
    {"FillColr", 5, 0x17},
    {"HatchIndex", 5, 0x18},
    {"EdgeWidth", 5, 0x1c},
    {"EdgeColr", 5, 0x1d},
    {"EdgeVis", 5, 0x1e},
    {"ColrTable", 5, 0x22},

    {"ApplData", 7, 0x02},

    {NULL}
};


int
cgm_read_text(const uint8_t *data, size_t len, cgm_handler_t userfunc, void *userdata)
{
    static const cgm_read_interface_t read_text_vt = {
        .eoe = _text_eoe,

        .get_I = _text_get_int,
        .get_IX = _text_get_uint,
        .get_R = _text_get_real,
        .get_S = _text_get_string,

        .get_color_channel = _text_get_uint,
        .get_color_index = _text_get_uint,
        .get_vdc = _text_get_vdc,
        .get_int_precision_bytes = _text_get_int_precision_bytes,

        .get_vdc_type = _text_get_vdc_type,
        .get_scaling_mode = _text_get_scaling_mode,
        .get_color_selection_mode = _text_get_color_selection_mode,
        .get_close_type = _text_get_close_type,
        .get_text_precision = _text_get_text_precision,
        .get_text_path = _text_get_text_path,
        .get_final_flag = _text_get_final_flag,
        .get_spec_mode = _text_get_spec_mode,
        .get_edge_visibility = _text_get_edge_visibility,
        .get_edge_flag = _text_get_edge_flag,
        .get_marker_type = _text_get_marker_type
    };
    const uint8_t *end = data + len;
    cgm_read_t r;
    cgm_read_init(&r);
    r.vt = &read_text_vt;
    while(data < end){
        const char *token;
        ssize_t token_len;
        int i;
        data = _get_token(data, end, &token, &token_len);
        if(!token){
            DEBUG_p("no more elements");
            break;
        }
        for(i=0;;++i){
            if(!cgm_element[i].string){
                DEBUG_p("element '%.*s' is unknown", (int)token_len, token);
                break;
            }
            if(!_xstrncasecmp(token, token_len, cgm_element[i].string)){
                DEBUG_p("class: %d, id: 0x%02x, %s", cgm_element[i].klass, cgm_element[i].id, cgm_element[i].string);
                cgm_read_set(&r, data, end-data);
                if(userfunc(userdata, &cgm_element[i], &r) < 0)
                    return -1;
                data = r.data;
                break;
            }
        }
        while(*data != ';' && data<end)
            ++data;
        ++data;
    }
    DEBUG_p("successfully done");
    return 0;
}
