/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <abydos-plugin.h>
#include <libgxps/gxps.h>
#include <math.h>                       /* ceil() */
#include <stdlib.h>
#include <string.h>


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    GXPSPage **page;
};


static const char **
_gxps_get_mime_types()
{
    static const char *mime_types[] = {
        "application/oxps",
        "application/vnd.ms-xpsdocument",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_gxps_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->threadsafe = 1;
    h->page = NULL;
    return h;
}


static void
_gxps_free(abydos_plugin_handle_t *h)
{
    if(h->page){
        int i;
        for(i=0;i<h->info->page_count;++i)
            if(h->page[i])
                g_object_unref(h->page[i]);
        free(h->page);
    }
    free(h);
}


static int
_gxps_create_from_file(abydos_plugin_handle_t *h,const char *filename)
{
    GFile *file = g_file_new_for_path(filename);
    GXPSFile *f = gxps_file_new(file, NULL);
    GXPSDocument **doc;
    int doc_count;
    int pagenr;
    int res = -1;
    int i;
    g_object_unref(file);
    if(!f){
        return -1;
    }
    h->info->page_count = 0;
    doc_count = gxps_file_get_n_documents(f);
    if(doc_count <= 0)
        return -1;
    doc = calloc(doc_count, sizeof(GXPSDocument*));
    for(i=0;i<doc_count;++i){
        doc[i] = gxps_file_get_document(f, i, NULL);
        if(!doc[i])
            continue;
        h->info->page_count += gxps_document_get_n_pages(doc[i]);
    }
    if(h->info->page_count <= 0)
        goto out;

    h->page = calloc(h->info->page_count, sizeof(GXPSDocument*));
    pagenr = 0;
    for(i=0;i<doc_count;++i){
        int page_count;
        int j;
        if(!doc[i])
            continue;
        page_count = gxps_document_get_n_pages(doc[i]);
        for(j=0;j<page_count;++j)
            h->page[pagenr++] = gxps_document_get_page(doc[i], j, NULL);
    }
    res = 0;
out:
    for(i=0;i<doc_count;++i)
        if(doc[i])
            g_object_unref(doc[i]);
    free(doc);
    return res;
}


static void
_gxps_render(abydos_plugin_handle_t *h,
             cairo_t *cr,
             int page,
             cairo_bool_t *layer_visibility,
             int variant,
             double time,
             int frame)
{
    GError *err = NULL;
    gxps_page_render(h->page[page], cr, &err);
}


const void
_gxps_get_page_info(abydos_plugin_handle_t *h, int page, abydos_plugin_page_info_t *info)
{
    double width;
    double height;
    gxps_page_get_size(h->page[page], &width, &height);
    info->width = ceil(width);
    info->height = ceil(height);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _gxps_get_mime_types,
        .new = _gxps_new,
        .free = _gxps_free,
        .create_from_file = _gxps_create_from_file,
        .render = _gxps_render,
        .get_page_info = _gxps_get_page_info
    };
    return &p;
}
