/*
 * Copyright (C) 2019 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdio.h>                      /* FILE, needed by jpeglib.h */
#include <jpeglib.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_jpeg_get_mime_types()
{
    static const char *mime_types[] = {
        "image/jpeg",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_jpeg_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    h->surface = NULL;
    return h;
}


static void
_jpeg_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static cairo_surface_t *
_surface_from_jpeg(struct jpeg_decompress_struct *dec)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,dec->output_width,dec->output_height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - dec->output_width;
    unsigned char *row = malloc(dec->output_width * dec->output_components);
    int y;
    for(y=0;y<dec->output_height;++y){
        unsigned char *src;
        int x;
        jpeg_read_scanlines(dec, &row, 1);
        src = row;
        if(1 == dec->output_components){
            for(x=0;x<dec->output_width;++x){
                int c = *src++;
                *dst++ = c<<16 | c<< 8 | c<< 0;
            }
        }else{
            for(x=0;x<dec->output_width;++x){
                int r = *src++;
                int g = *src++;
                int b = *src++;
                *dst++ = r<<16 | g<< 8 | b<< 0;
            }
        }
        dst += skip;
    }
    free(row);
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_jpeg_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    struct jpeg_decompress_struct dec;
    struct jpeg_error_mgr jerr;

    dec.err = jpeg_std_error(&jerr);
    jpeg_create_decompress(&dec);
    jpeg_mem_src(&dec, (const unsigned char*)data, len);

    if(jpeg_read_header(&dec, TRUE) != 1){
        jpeg_destroy_decompress(&dec);
        return -1;
    }
    if(1 == dec.num_components || 3 == dec.num_components){
        (void)jpeg_start_decompress(&dec);

        h->info->width = dec.output_width;
        h->info->height = dec.output_height;
        h->surface = _surface_from_jpeg(&dec);
    }

    (void)jpeg_finish_decompress(&dec);
    jpeg_destroy_decompress(&dec);
    return h->surface ? 0 : -1;
}


static cairo_surface_t *
_jpeg_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _jpeg_get_mime_types,
        .new = _jpeg_new,
        .free = _jpeg_free,
        .create_from_data = _jpeg_create_from_data,
        .get_image_surface = _jpeg_get_image_surface
    };
    return &p;
}
