/*
 * Copyright (C) 2020 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <endian.h>
#include <float.h>
#include <loa-buffer.h>
#include <loa-cairo.h>
#include <math.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>


#if __BYTE_ORDER == __BIG_ENDIAN
# define NCIF (('n'<<24)|('c'<<16)|('i'<<8)|('f'<<0))
#else
# define NCIF (('n'<<0)|('c'<<8)|('i'<<16)|('f'<<24))
#endif

#define STYLE_RGBA     1
#define STYLE_GRADIENT 2
#define STYLE_RGB      3
#define STYLE_GRAYA    4
#define STYLE_GRAY     5

#define SHAPE_PATHS 10

#define EFFECT_AFFINE      20
#define EFFECT_CONTOUR     21
#define EFFECT_PERSPECTIVE 22
#define EFFECT_STROKE      23

#define GRADIENT_LINEAR   0
#define GRADIENT_CIRCULAR 1
#define GRADIENT_DIAMOND  2
#define GRADIENT_CONIC    3
#define GRADIENT_XY       4
#define GRADIENT_SQRT_XY  5

#define GRADIENT_FLAG_TRANSFORM     (1<<1)
#define GRADIENT_FLAG_NO_ALPHA      (1<<2)
#define GRADIENT_FLAG_16_BIT_COLORS (1<<3)
#define GRADIENT_FLAG_GRAYS         (1<<4)

#define PATH_CLOSED   (1<<1)
#define PATH_COMMANDS (1<<2)
#define PATH_LINES    (1<<3)

#define SHAPE_TRANSFORM    (1<<1)
#define SHAPE_HINTING      (1<<2)
#define SHAPE_LOD_SCALE    (1<<3)
#define SHAPE_EFFECTS      (1<<4)
#define SHAPE_TRANSLATION  (1<<5)

#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif

#define LOD(a) ((a)<=0 ? -DBL_MAX : ((a)>=255 ? DBL_MAX : 4.0*(a)/255))

typedef struct{
    double x;
}transform_t;

typedef struct{
    double r;
    double g;
    double b;
    double a;
}color_t;

typedef struct{
    int gradient;
    color_t c;
    cairo_pattern_t *pattern;
}style_t;

typedef enum{
    C_NONE = -1,
    C_HLINE = 0,
    C_VLINE = 1,
    C_LINE = 2,
    C_CURVE = 3
}command_t;

typedef struct{
    command_t command;
    loa_point_t point[3];
}node_t;

typedef struct{
    int type;
    union{
        struct{
            int width;
            int join;
            int miter;
        }contour;
        struct{
            int width;
            int options;
            int miter;
        }stroke;
    };
}effect_t;

struct _loa_object_t{
    const style_t *style;
    /* offset in the shape_path array */
    int path;
    int path_count;
    double lod_min;
    double lod_max;
    effect_t effect[256];
    int effect_count;
};

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    style_t style[256];
    int style_count;
    loa_buffer_t vb;
    loa_array_t shape_path_array;
    loa_path_t path[256];
    int path_count;
    loa_object_t shape[256];
    int shape_count;
};


#define _get_shape_path(h,i) ((int*)loa_array_get(&(h)->shape_path_array,i))
#define _shape_path_count(h) loa_array_length(&(h)->shape_path_array)
#define _add_shape_paths(h,c) (int*)loa_array_append(&(h)->shape_path_array, c)


static const char **
_hvif_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-hvif",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_hvif_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->width = 64;
    h->info->height = 64;
    h->info->scalable_size = 1;
    h->info->threadsafe = 1;
    loa_buffer_init(&h->vb, sizeof(loa_object_t), 0, 0, 0);
    loa_array_init(&h->shape_path_array, sizeof(int));
    return h;
}


static void
_hvif_free(abydos_plugin_handle_t *h)
{
    loa_buffer_done(&h->vb);
    loa_array_done(&h->shape_path_array);
    free(h);
}


static int
_get_value(const uint8_t **p, const uint8_t *end, double *rvalue)
{
    int val;
    if(*p >= end){
        DEBUG_p("tried to read value beyond end");
        return -1;
    }
    val = *(*p)++;

    if(val & 128){
        /* high bit set, the next byte is part of the value */
        int val2;
        if(*p >= end)
            return -1;
        val2 = *(*p)++ | (val&127) << 8;
        *rvalue = (double)val2 / 102 - 128;
    }else{
        *rvalue = (double)val - 32;
    }
    return 0;
}


static int
_get_points(const uint8_t **p, const uint8_t *end, loa_point_t *point, int count){
    while(count-->0){
        if(_get_value(p,end,&point->x)<0)
            return -1;
        if(_get_value(p,end,&point->y)<0)
            return -1;
        DEBUG_p("got %f x %f",point->x,point->y);
        ++point;
    }
    return 0;
}


#if 0
static int
_get_short(const uint8_t **p, const uint8_t *end, int *rvalue)
{
    if(*p+2 > end){
        DEBUG_p("truncated data");
        return -1;
    }
    *rvalue = le16toh(*(uint16_t*)*p);
    *p += 2;
    return 0;
}
#endif


static int
_get_float(const uint8_t **p, const uint8_t *end, double *rvalue)
{
    int v;
    if(*p+3 > end){
        DEBUG_p("truncated data");
        return -1;
    }

    v = (*p)[0] << 16 | (*p)[1] << 8 | (*p)[2];
    *p += 3;

    if(0 == v){
        *rvalue = 0;
    }else{
        int sign = (v & 0x800000) >> 23;
        int exponent = ((v & 0x7e0000) >> 17) - 32;
        int mantissa = (v & 0x01ffff) << 6;
        uint32_t tmp = (sign << 31) | ((exponent + 127) << 23) | mantissa;
        *rvalue = *(float*)&tmp;
    }
    return 0;
}


static int
_get_transform(const uint8_t **p, const uint8_t *end, transform_t *t)
{
    double value[6];
    int i;
    if(*p+6*3 > end){
        DEBUG_p("truncated data");
        return -1;
    }
    for(i=0;i<6;++i)
        if(_get_float(p, end, &value[i])<0)
            return -1;
    /* FIXME!! setup transform */
    return 0;
}


static int
_get_translate(const uint8_t **p, const uint8_t *end, transform_t *t)
{
    double x;
    double y;
    if(_get_value(p, end, &x)<0)
        return 0;
    if(_get_value(p, end, &y)<0)
        return 0;
    /* FIXME!! setup transform */
    return 0;
}


static int
_handle_effect(const uint8_t **p, const uint8_t *end, effect_t *e)
{
    if(*p >= end)
        return -1;
    e->type = *(*p)++;
    switch(e->type){
    case EFFECT_AFFINE:
        {
            double matrix[6];
            int i;
            DEBUG_p("affine");
            for(i=0;i<6;++i)
                if(_get_float(p, end, &matrix[i])<0)
                    return -1;
        }
    case EFFECT_CONTOUR:
        {
            DEBUG_p("contour");
            if(*p+3 > end)
                return -1;
            e->contour.width = *(*p)++ - 128;
            e->contour.join = *(*p)++;
            e->contour.miter = *(*p)++;
        }
        break;
    case EFFECT_PERSPECTIVE:
        {
            DEBUG_p("perspective");
            return -1;
        }
    case EFFECT_STROKE:
        {
            DEBUG_p("stroke");
            if(*p+3 > end)
                return -1;
            e->stroke.width = *(*p)++ - 128;
            e->stroke.options = *(*p)++;
            e->stroke.miter = *(*p)++;
        }
        break;
    default:
        DEBUG_p("invalied data");
        return -1;
    }
    return 0;
}


static int
_handle_lines(loa_buffer_t *vb, loa_path_t *path, const uint8_t **p, const uint8_t *end)
{
    int seg_count;
    int i;

    if(*p >= end)
        goto error_truncated;
    seg_count = *(*p)++;
    DEBUG_p("part count: %d",seg_count);

    path->segment = loa_buffer_segment_count(vb);
    for(i=0;i<seg_count;++i){
        loa_point_t *point;
        if(0==i){
            point = loa_buffer_push_segment(vb,path,LOA_GOTO);
        }else{
            point = loa_buffer_push_segment(vb,path,LOA_LINE);
        }
        if(_get_points(p,end,point,1)<0)
            goto error_truncated;
    }
    return 0;
error_truncated:
    DEBUG_p("truncated data");
    return -1;
}


static int
_handle_curves(loa_buffer_t *vb, loa_path_t *path, const uint8_t **p, const uint8_t *end)
{
    int seg_count;
    node_t first;
    loa_point_t last_point;
    node_t node;
    loa_point_t *point;
    int i;

    if(*p >= end)
        goto error_truncated;

    /* just to shut up the compiler */
    last_point.x = last_point.y = 0;

    seg_count = *(*p)++;
    DEBUG_p("part count: %d",seg_count);
    path->segment = loa_buffer_segment_count(vb);
    for(i=0;i<seg_count;++i){
        if(_get_points(p,end,node.point,3)<0)
            goto error_truncated;
        if(0==i){
            point = loa_buffer_push_segment(vb,path,LOA_GOTO);
            point[0] = node.point[0];
            first.point[0] = node.point[0];
            first.point[1] = node.point[1];
        }else{
            point = loa_buffer_push_segment(vb,path,LOA_CURVE);
            point[0] = last_point;
            point[1] = node.point[1];
            point[2] = node.point[0];
        }
        last_point = node.point[2];
    }

    /* close the path */
    point = loa_buffer_push_segment(vb,path,LOA_CURVE);
    point[0] = last_point;
    point[1] = first.point[1];
    point[2] = first.point[0];

    return 0;
error_truncated:
    DEBUG_p("truncated data");
    return -1;
}


static int
_handle_commands(loa_buffer_t *vb, loa_path_t *path, const uint8_t **p, const uint8_t *end)
{
    int seg_count;
    int command_bytes;
    const uint8_t *command;
    node_t first;
    node_t last;
    node_t node;
    loa_point_t *point;
    int i;

    if(*p >= end)
        goto error_truncated;
    seg_count = *(*p)++;
    command_bytes = (seg_count+3)/4;
    DEBUG_p("part count: %d (%d bytes)",seg_count,command_bytes);
    if(*p+command_bytes > end)
        goto error_truncated;

    /* just to shut up the compiler */
    last.point[2].x = last.point[2].y = 0;

    command = *p;
    *p += command_bytes;

    path->segment = loa_buffer_segment_count(vb);
    last.command = C_NONE;

    for(i=0;i<seg_count;++i){
        node.command = (command[i/4] >> 2*(i%4)) & 3;
        switch(node.command){
        case C_HLINE:
            if(C_NONE==last.command)
                goto error_invalid;
            node.point[0].y = last.point[0].y;
            if(_get_value(p,end,&node.point[0].x)<0)
                goto error_truncated;
            break;
        case C_VLINE:
            if(C_NONE==last.command)
                goto error_invalid;
            node.point[0].x = last.point[0].x;
            if(_get_value(p,end,&node.point[0].y)<0)
                goto error_truncated;
            break;
        case C_LINE:
            if(_get_points(p,end,node.point,1)<0)
                goto error_truncated;
            break;
        case C_CURVE:
            if(_get_points(p,end,node.point,3)<0)
                goto error_truncated;
            break;
        case C_NONE:
            goto error_internal;
        }
        switch(last.command){
        case C_NONE:
            point = loa_buffer_push_segment(vb,path,LOA_GOTO);
            point[0] = node.point[0];
            /* keep the first points in order to loop the path */
            first.command = node.command;
            first.point[0] = node.point[0];
            if(C_CURVE==node.command)
                first.point[1] = node.point[1];
            break;
        case C_HLINE:
        case C_VLINE:
        case C_LINE:
            if(C_CURVE==node.command){
                point = loa_buffer_push_segment(vb,path,LOA_CURVE);
                point[0] = last.point[0];
                point[1] = node.point[1];
                point[2] = node.point[0];
            }else{
                point = loa_buffer_push_segment(vb,path,LOA_LINE);
                point[0] = node.point[0];
            }
            break;
        case C_CURVE:
            point = loa_buffer_push_segment(vb,path,LOA_CURVE);
            if(C_CURVE==node.command){
                point[0] = last.point[2];
                point[1] = node.point[1];
                point[2] = node.point[0];
            }else{
                point[0] = last.point[2];
                point[1] = node.point[0];
                point[2] = node.point[0];
            }
            break;
        }
        last.command = node.command;
        last.point[0] = node.point[0];
        if(C_CURVE==node.command)
            last.point[2] = node.point[2];
    }

    /* close the path */
    switch(last.command){
    case C_HLINE:
    case C_VLINE:
    case C_LINE:
        if(C_CURVE==first.command){
            point = loa_buffer_push_segment(vb,path,LOA_CURVE);
            point[0] = last.point[0];
            point[1] = first.point[1];
            point[2] = first.point[0];
        }else{
            /* Just closing the path will have the same effect */
            /*
            point = loa_buffer_push_part(vb,path,LOA_LINE,point,1);
            point[0] = first.point[0];
            */
        }
        break;
    case C_CURVE:
        point = loa_buffer_push_segment(vb,path,LOA_CURVE);
        if(C_CURVE==first.command){
            point[0] = last.point[2];
            point[1] = first.point[1];
            point[2] = first.point[0];
        }else{
            point[0] = last.point[2];
            point[1] = first.point[0];
            point[2] = first.point[0];
        }
        break;
    case C_NONE:
        goto error_internal;
    }

    return 0;
error_truncated:
    DEBUG_p("truncated data");
    return -1;
error_invalid:
    DEBUG_p("invalid data");
    return -1;
error_internal:
    DEBUG_p("internal error");
    return -1;
}


static int
_handle_color(color_t *c, int encoding, const uint8_t **p, const uint8_t *end)
{
    switch(encoding){
    case STYLE_RGBA:
        if(*p+4 > end)
            goto error_truncated;
        c->r = (double)(*p)[0]/255;
        c->g = (double)(*p)[1]/255;
        c->b = (double)(*p)[2]/255;
        c->a = (double)(*p)[3]/255;
        *p += 4;
        break;
    case STYLE_RGB:
        if(*p+3 > end)
            goto error_truncated;
        c->r = (double)(*p)[0]/255;
        c->g = (double)(*p)[1]/255;
        c->b = (double)(*p)[2]/255;
        c->a = 1.0;
        *p += 3;
        break;
    case STYLE_GRAYA:
        if(*p+2 > end)
            goto error_truncated;
        c->r = c->g = c->b = (double)(*p)[0]/255;
        c->a = (double)(*p)[1]/255;
        *p += 2;
        break;
    case STYLE_GRAY:
        if(*p >= end)
            goto error_truncated;
        c->r = c->g = c->b = (double)(*p)[0]/255;
        c->a = 1.0;
        *p += 1;
        break;
    default:
        goto error_invalid;
    }
    return 0;
error_truncated:
    DEBUG_p("truncated data");
    return -1;
error_invalid:
    DEBUG_p("invalid data");
    return -1;
}


static int
_hvif_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    const uint8_t *p = (uint8_t*)data;
    const uint8_t *end = p+len;
    int i;
    int j;
    if(len < 4+3)
        goto error_truncated;
    if(NCIF != *(uint32_t*)data)
        goto error_invalid;
    p += 4;

    /* Read styles */
    h->style_count = *p++;
    DEBUG_p("style count: %d",h->style_count);
    for(i=0;i<h->style_count;++i){
        int style;
        if(p >= end)
            goto error_truncated;
        style = *p++;
        switch(style){
        case STYLE_RGBA:
        case STYLE_RGB:
        case STYLE_GRAYA:
        case STYLE_GRAY:
            if(_handle_color(&h->style[i].c, style, &p, end)<0)
                goto error;
            h->style[i].gradient = -1;
            h->style[i].pattern = cairo_pattern_create_rgba(h->style[i].c.r,h->style[i].c.g,h->style[i].c.b,h->style[i].c.a);
            break;
        case STYLE_GRADIENT:
            DEBUG_p("gradient");
            if(p+3 > end)
                goto error_truncated;
            h->style[i].gradient = *p++;
            {
                int flags = *p++;
                int count;
                int j;
                if(flags & ~30){
                    DEBUG_p("incorrect gradient flags");
                    goto error_parse;
                }
                count = *p++;
                DEBUG_p("color count: %d",count);
                if(flags & GRADIENT_FLAG_TRANSFORM){
                    if(_get_transform(&p, end, NULL)<0)
                        goto error;
                }
                if(flags & GRADIENT_FLAG_NO_ALPHA){
                    if(flags & GRADIENT_FLAG_GRAYS){
                        style = STYLE_GRAY;
                    }else{
                        style = STYLE_RGB;
                    }
                }else{
                    if(flags & GRADIENT_FLAG_GRAYS){
                        style = STYLE_GRAYA;
                    }else{
                        style = STYLE_RGBA;
                    }
                }
                if(p+2 > end)
                    goto error_truncated;
                DEBUG_p("gradient type: %d",h->style[i].gradient);
                //h->style[i].pattern = cairo_pattern_create_linear(0,-32,0,12);
                h->style[i].pattern = cairo_pattern_create_linear(0,64,0,0);
                for(j=0;j<count;++j){
                    int stop;
                    color_t c;
                    if(p+2 > end)
                        goto error;
                    stop = *p++;
                    if(_handle_color(&c, style, &p, end)<0)
                        goto error;
                    DEBUG_p("stop: %d",stop);
                    cairo_pattern_add_color_stop_rgba(h->style[i].pattern, (double)stop/(255), c.r, c.g, c.b, c.a);
                }
            }
            break;
#if 0
        case STYLE_RGBA:
            if(p+4 > end)
                goto error_truncated;
            h->style[i].c.r = (double)p[0]/255;
            h->style[i].c.g = (double)p[1]/255;
            h->style[i].c.b = (double)p[2]/255;
            h->style[i].c.a = (double)p[3]/255;
            p += 4;
            break;
        case STYLE_RGB:
            if(p+3 > end)
                goto error_truncated;
            h->style[i].c.r = (double)p[0]/255;
            h->style[i].c.g = (double)p[1]/255;
            h->style[i].c.b = (double)p[2]/255;
            h->style[i].c.a = 1.0;
            p += 3;
            break;
        case STYLE_GRAYA:
            if(p+2 > end)
                goto error_truncated;
            h->style[i].c.r = h->style[i].c.g = h->style[i].c.b = (double)p[0]/255;
            h->style[i].c.a = (double)p[1]/255;
            p += 2;
            break;
        case STYLE_GRAY:
            if(p >= end)
                goto error_truncated;
            h->style[i].c.r = h->style[i].c.g = h->style[i].c.b = (double)p[0]/255;
            h->style[i].c.a = 1.0;
            ++p;
            break;
#endif
        default:
            DEBUG_p("invalid style: %d",style);
            goto error_invalid;
        }
    }

    /* Read paths */
    if(p+2 > end)
        goto error_truncated;
    h->path_count = *p++;
    DEBUG_p("path count: %d",h->path_count);
    for(i=0;i<h->path_count;++i){
        int flags;
        if(p >= end)
            goto error_truncated;
        flags = *p++;
        DEBUG_p("flags: %x",flags);
        if(flags & ~14){
            DEBUG_p("incorrect path flags");
            goto error_invalid;
        }
        if(flags & PATH_LINES){
            DEBUG_p("---lines");
            if(_handle_lines(&h->vb, &h->path[i], &p, end)<0)
                goto error_truncated;
        }else if(flags & PATH_COMMANDS){
            DEBUG_p("---commands");
            if(_handle_commands(&h->vb, &h->path[i], &p, end)<0)
                goto error;
        }else{
            /* only curves */
            DEBUG_p("---curves");
            if(_handle_curves(&h->vb, &h->path[i], &p, end)<0)
                goto error;
        }
        if(flags & PATH_CLOSED)
            loa_buffer_push_segment(&h->vb,&h->path[i],LOA_CLOSE);
    }

    /* Read shapes */
    if(p >= end)
        goto error_truncated;
    h->shape_count = *p++;
    DEBUG_p("shape count: %d",h->shape_count);
    for(i=0;i<h->shape_count;++i){
        int type;
        int *path_index;
        int flags;
        if(p+4 > end)
            goto error_truncated;
        type = *p++;
        DEBUG_p("shape type %d",type);
        if(type != SHAPE_PATHS){
#if 1
            DEBUG_p("invalid shape type");
            goto error_invalid;
#else
            int len;
            if(_get_short(&p, end, &len)<0)
                goto error;
            DEBUG_p("unknown shape tag, size: %d",len);
            if(p+len > end)
                goto error_truncated;
            p += len;
            continue;
#endif
        }
        h->shape[i].style = h->style + *p++;
        h->shape[i].path_count = *p++;
        if(p+h->shape[i].path_count+1 > end)
            goto error_truncated;
        DEBUG_p("shape %d has %d paths (style: %d)",i,h->shape[i].path_count,(int)(h->shape[i].style-h->style));
        h->shape[i].path = _shape_path_count(h);
        path_index = _add_shape_paths(h,h->shape[i].path_count);
        for(j=0;j<h->shape[i].path_count;++j)
            path_index[j] = *p++;

        flags = *p++;
        DEBUG_p("flags: %x",flags);
        if(flags & ~62){
            DEBUG_p("incorrect shape flags");
            goto error_invalid;
        }

        if(flags & SHAPE_TRANSFORM){
            if(_get_transform(&p, end, NULL)<0)
                goto error;
        }
        if(flags & SHAPE_HINTING){
            DEBUG_p("enable shape hinting");
            /* FIXME!! what should be done with this information? */
        }
        if(flags & SHAPE_LOD_SCALE){
            DEBUG_p("there are lod scale bytes");
            if(p+2 > end)
                goto error_truncated;
            h->shape[i].lod_min = LOD(p[0]);
            h->shape[i].lod_max = LOD(p[1]);
            p += 2;
        }else{
            h->shape[i].lod_min = -DBL_MAX;
            h->shape[i].lod_max = DBL_MAX;
        }
        if(flags & SHAPE_EFFECTS){
            int j;
            if(p >= end)
                goto error_truncated;
            h->shape[i].effect_count = *p++;
            DEBUG_p("shape transform count: %d",h->shape[i].effect_count);
            for(j=0;j<h->shape[i].effect_count;++j){
                if(_handle_effect(&p, end, &h->shape[i].effect[j])<0)
                    goto error;
            }
        }
        if(flags & SHAPE_TRANSLATION){
            if(_get_translate(&p, end, NULL)<0)
                goto error;
        }
    }

    DEBUG_p("bytes left: %d",(int)(end-p));

    return 0;
error_truncated:
    DEBUG_p("truncated data");
    return -1;
error_invalid:
    DEBUG_p("invalid data");
    return -1;
error_parse:
    DEBUG_p("parsing error");
error:
    return -1;
}


static void
_hvif_render(abydos_plugin_handle_t *h,
             cairo_t *cr,
             int page,
             cairo_bool_t *layer_visibility,
             int variant,
             double time,
             int frame)
{
    cairo_matrix_t matrix;
    double scale;
    int i;

    cairo_get_matrix(cr,&matrix);
    scale = MAX(sqrt(matrix.xx*matrix.xx+matrix.yx*matrix.yx),
                sqrt(matrix.xy*matrix.xy+matrix.yy*matrix.yy));

    cairo_save(cr);

    for(i=0;i<h->shape_count;++i){
        loa_object_t *shape = &h->shape[i];
        const color_t *c;
        int *path_index;
        int j;
        if(scale < shape->lod_min || scale > shape->lod_max)
            continue;
        path_index = _get_shape_path(h,h->shape[i].path);
        for(j=0;j<h->shape[i].path_count;++j){
            const loa_path_t *path = &h->path[path_index[j]];
            loa_cairo_draw_path(cr, path, loa_buffer_get_segment(&h->vb, 0), loa_buffer_get_point(&h->vb, 0));
        }
        c = &h->shape[i].style->c;
        cairo_set_operator(cr, CAIRO_OPERATOR_OVER);
        for(j=0;j<shape->effect_count;++j){
            effect_t *e = &shape->effect[j];
            switch(e->type){
            case EFFECT_CONTOUR:
                cairo_save(cr);
                break;
            case EFFECT_STROKE:
                cairo_set_source_rgba(cr,0,0,0,1);
                cairo_set_line_width(cr,e->stroke.width);
                cairo_stroke_preserve(cr);
                break;
            }
        }
        cairo_set_source(cr, shape->style->pattern);
        cairo_fill(cr);
        for(j=0;j<shape->effect_count;++j){
            effect_t *e = &shape->effect[j];
            switch(e->type){
            case EFFECT_CONTOUR:
                cairo_restore(cr);
                cairo_set_source_rgba(cr,0,0,0,c->a);
                cairo_set_line_width(cr,e->contour.width);
                cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
                cairo_stroke(cr);
                break;
            }
        }
    }
    cairo_restore(cr);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _hvif_get_mime_types,
        .new = _hvif_new,
        .free = _hvif_free,
        .create_from_data = _hvif_create_from_data,
        .render = _hvif_render
    };
    return &p;
}
