/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos.h>
#include <abydos-plugin.h>
#include <debug.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>

#include "loa-abstract-abydos.h"
#include "ni-mng.h"


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    ni_mng_t *mng;
};


static const char **
_mng_get_mime_types()
{
    static const char *mime_types[] = {
        "video/x-mng", "image/x-mng",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_mng_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    info->threadsafe = 1;
    h->info = info;
    return h;
}


static void
_mng_free(abydos_plugin_handle_t *h)
{
    if(h->mng)
        ni_mng_free(h->mng);
    free(h);
}


#if 0
static int
_mng_examine(abydos_plugin_handle_t *h, mng_info *info, const char *data, const char *end)
{
    int frame_count = 0;
    ni_mng_defi_t defi;

    //h->object_count = 0;
    defi.hide = 0;

    while(end - data >= 12){
        uint32_t chunk_len = get32(data);
        uint32_t chunk_id = *(uint32_t*)(data+4);
        DEBUG_p("chunk id: %.4s, length: %d", data+4, chunk_len);
        data += 8;
        switch(chunk_id){
        case MNG_MHDR:
            /* min 28 bytes */
            DEBUG_p("width: %d", get32(data));
            DEBUG_p("height: %d", get32(data+4));
            DEBUG_p("ticks per second: %d", get32(data+8));
            DEBUG_p("layers count: %d", get32(data+12));
            DEBUG_p("frame count: %d", get32(data+16));
            DEBUG_p("play time: %d", get32(data+20));
            DEBUG_p("simplicity profile: 0x%04x", get32(data+24));
            h->info->width = get32(data);
            h->info->height = get32(data+4);
            info->ticks_per_second = get32(data+8);
            h->info->frame_count = get32(data+12);
            break;
        case MNG_IHDR:
            if(defi.hide){
                defi.hide = 0;
            }else{
                ++frame_count;
            }
            break;
        case MNG_DEFI:
            if(ni_mng_handle_defi(&defi, data, chunk_len))
                return -1;
            //if(defi.object_id >= h->object_count)
            //    h->object_count = defi.object_id + 1;
            break;
        case MNG_SHOW:
            ++frame_count;
        }
        data += chunk_len + 4; /* also skip past checksum */
    }
    DEBUG_p("reported frame count: %d, measured frame count: %d", h->info->frame_count, frame_count);
    if(0 == h->info->frame_count && frame_count > 0)
        h->info->frame_count = frame_count;
    return 0;
}
#endif


typedef struct{
    abydos_t *ar;
}read_data_t;


static int
_on_png_begin(void *userdata)
{
    read_data_t *d = userdata;
    d->ar = abydos_create("image/png");
    abydos_load_begin(d->ar, NULL, NULL, NULL);
    return 0;
}


static int
_on_jng_begin(void *userdata)
{
    read_data_t *d = userdata;
    d->ar = abydos_create("image/x-jng");
    abydos_load_begin(d->ar, NULL, NULL, NULL);
    return 0;
}


static int
_on_feed(void *userdata, const char *data, size_t len)
{
    read_data_t *d = userdata;
    return abydos_load_feed(d->ar, data, len);
}


static loa_abstract_t *
_on_end(void *userdata)
{
    read_data_t *d = userdata;
    if(0 == abydos_load_end(d->ar)){
        return loa_abstract_abydos_from_abydos(d->ar);
    }else{
        abydos_destroy(d->ar);
        return loa_abstract_abydos_create(userdata, 1, 1);
    }
}


static int
_mng_create_from_data(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    static const ni_mng_callbacks_t vt = {
        .abstract = {
            .create = loa_abstract_abydos_create,
            .free = loa_abstract_abydos_free,
            .get_size = loa_abstract_abydos_get_size,
            .copy = loa_abstract_abydos_copy,
            .fill = loa_abstract_abydos_fill,
            .paste = loa_abstract_abydos_paste
        },
        .png_begin = _on_png_begin,
        .png_feed = _on_feed,
        .png_end = _on_end,
        .jng_begin = _on_jng_begin,
        .jng_feed = _on_feed,
        .jng_end = _on_end
    };
    read_data_t rd;
    h->mng = ni_mng_decode(data, len, &vt, &rd);
    if(!h->mng)
        return -1;
    h->info->width = ni_mng_get_width(h->mng);
    h->info->height = ni_mng_get_height(h->mng);
    h->info->scalable_time = ni_mng_get_scalable_time(h->mng);
    h->info->frame_count = ni_mng_get_frame_count(h->mng, 0);
    h->info->page_count = ni_mng_get_page_count(h->mng);
    return 0;
}


static int
_on_draw(void *userdata, double x, double y, double xscale, double yscale, int flags, loa_abstract_t *abst, nil_clip_t *clip)
{
    cairo_t *cr = userdata;
#if 1
    loa_abstract_abydos_render(x, y, xscale, yscale, flags, abst, clip, cr);
#else
    switch(abst->type){
    case ABST_EMPTY:
        break;
    case ABST_ABYDOS:
        cairo_save(cr);
        cairo_translate(cr, x, y);
        abydos_render(abst->ar, cr);
        cairo_restore(cr);
        break;
    case ABST_SURFACE:
        cairo_set_source_surface(cr, abst->surface, x, y);
        cairo_paint(cr);
        break;
    }
#endif
    return 0;
}


static void
_mng_render(abydos_plugin_handle_t *h,
            cairo_t *cr,
            int page,
            cairo_bool_t *layer_visibility,
            int variant,
            double time,
            int frame)
{
    //page = 1;
    //time = 2;
    cairo_save(cr);
    cairo_rectangle(cr, 0, 0, h->info->width, h->info->height);
    cairo_clip(cr);
    ni_mng_draw(h->mng, page, layer_visibility, time, _on_draw, cr);
    cairo_restore(cr);
}


static void
_mng_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int frame_count = ni_mng_get_frame_count(h->mng, page);
    int i;
    for(i=0;i<frame_count;++i)
        duration[i] = ni_mng_get_frame_duration(h->mng, page, i);
}


static void
_mng_get_page_info(abydos_plugin_handle_t *h, int page, abydos_plugin_page_info_t *info)
{
    info->frame_count = ni_mng_get_frame_count(h->mng, page);
    info->layer_count = ni_mng_get_layer_count(h->mng, page);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _mng_get_mime_types,
        .new = _mng_new,
        .free = _mng_free,
        .create_from_data = _mng_create_from_data,
        .render = _mng_render,
        .get_duration = _mng_get_duration,
        .get_page_info = _mng_get_page_info
    };
    return &p;
}
