/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <debug.h>
#include <stdint.h>
#include <string.h>                     /* strnlen() */

#include "ni-mng.h"

#define get32s(a) (int32_t)be32toh(*(uint32_t*)(a))
#define get32u(a) be32toh(*(uint32_t*)(a))
#define get16(a) be16toh(*(uint16_t*)(a))
#define get8(a) (*(uint8_t*)(a))


int
ni_mng_parse_mhdr(ni_mng_mhdr_t *d, const char *data, size_t len)
{
    if(len<28)
        return -1;
    d->width = get32u(data);
    d->height = get32u(data+4);
    d->ticks_per_second = get32u(data+8);
    d->layer_count = get32u(data+12);
    d->frame_count = get32u(data+16);
    d->play_time = get32u(data+20);
    d->simplicity_profile = get32u(data+24);
    return 0;
}


int
ni_mng_parse_loop(ni_mng_loop_t *d, const char *data, size_t len)
{
    if(len<5)
        return -1;
    d->nest_level = get8(data);
    d->iteration_count = get32u(data+1);
    d->termination_condition = len >= 6 ? get8(data+5) : 0;
    if((3 == d->termination_condition || 7 == d->termination_condition) && len<18)
        return -1;
    d->iteration_min = len >= 10 ? get32u(data+6) : 1;
    d->iteration_max = len >= 14 ? get32u(data+10) : -1;
    d->signal_number = len >= 18 ? get32u(data+14) : -1;
    return 0;
}


int
ni_mng_parse_endl(ni_mng_endl_t *d, const char *data, size_t len)
{
    if(len<1)
        return -1;
    d->nest_level = get8(data);
    return 0;
}


int
ni_mng_parse_defi(ni_mng_defi_t *d, const char *data, size_t len)
{
    if(len<2)
        return -1;
    d->object_id = get16(data);
    d->hide = len >= 3 ? get8(data+2) : 0;
    d->concrete = len >= 4 ? get8(data+3) : 0;
    d->posx = len >= 12 ? get32s(data+4) : 0;
    d->posy = len >= 12 ? get32s(data+8) : 0;
    if((d->clip_available = len >= 28)){
        d->clipx1 = get32s(data+12);
        d->clipy1 = get32s(data+16);
        d->clipx2 = get32s(data+20);
        d->clipy2 = get32s(data+24);
    }
    return 0;
}


int ni_mng_parse_basi(ni_mng_basi_t *d, const char *data, size_t len)
{
    if(len<13)
        return -1;
    d->width = get32u(data);
    d->height = get32u(data+4);
    d->sample_depth = get8(data+8);
    d->color_type = get8(data+9);
#ifdef USE_DEBUG
    d->compression_method = get8(data+10);
    d->filter_method = get8(data+11);
    d->interlace_method = get8(data+12);
#endif
    if(len>=21){
        int mask = 0xffff >> (16-d->sample_depth);
        int max = (1 << d->sample_depth) - 1;
        d->red = (double)(get16(data+13)&mask) / max;
        if(0 == d->color_type || 4 == d->color_type){
            d->green = d->red;
            d->blue = d->red;
        }else{
            d->green = (double)(get16(data+15)&mask) / max;
            d->blue = (double)(get16(data+17)&mask) / max;
        }
        d->alpha = (double)(get16(data+19)&mask) / max;
        if((0 == d->color_type || 2 == d->color_type) && d->alpha > 0)
            d->alpha = 1;
        d->viewable = get8(data+20);
    }else{
        d->red = 0;
        d->green = 0;
        d->blue = 0;
        d->alpha = 0;
        d->viewable = 0;
    }
    return 0;
}


int ni_mng_parse_clon(ni_mng_clon_t *d, const char *data, size_t len)
{
    if(len<4)
        return -1;
    d->src_id = get16(data);
    d->dst_id = get16(data+2);
    d->clone_type = len >= 5 ? get8(data+4) : 0;
    d->hide = len >= 6 ? get8(data+5) : -1;
    d->concrete = len >= 7 ? get8(data+6) : -1;
    if(len >= 16){
        d->loca_delta_type = get8(data+7);
        d->x_location = get32s(data+8);
        d->y_location = get32s(data+12);
    }else{
        d->loca_delta_type = -1;
    }
    return 0;
}


int ni_mng_parse_dhdr(ni_mng_dhdr_t *d, const char *data, size_t len)
{
    if(len<4)
        return -1;
    d->object_id = get16(data);
    d->image_type = get8(data+2);
    d->delta_type = get8(data+3);
    if(7 == d->delta_type){
        d->block_width = -1;
        d->block_height = -1;
    }else{
        if(len<12)
            return -1;
        d->block_width = get32u(data+4);
        d->block_height = get32u(data+8);
    }
    if(0 == d->delta_type || 7 == d->delta_type){
        d->x_location = 0;
        d->y_location = 0;
    }else{
        if(len<20)
            return -1;
        d->x_location = get32u(data+12);
        d->y_location = get32u(data+16);
    }
    return 0;
}


int ni_mng_parse_past(ni_mng_past_t *d, const char *data, size_t len)
{
    if(len<11)
        return -1;
    d->destination_id = get16(data);
    d->target_delta_type = get8(data+2);
    d->target_x = get32s(data+3);
    d->target_y = get32s(data+7);
    return 0;
}


int ni_mng_parse_past_entry(ni_mng_past_entry_t *d, const char *data, size_t len)
{
    if(len<30)
        return -1;
    d->source_id = get16(data);
    d->composition_mode = get8(data+2);
    d->orientation = get8(data+3);
    d->offset_origin = get8(data+4);
    d->x_offset = get32s(data+5);
    d->y_offset = get32s(data+9);
    d->boundary_origin = get8(data+13);
    d->left_past_cb = get32s(data+14);
    d->right_past_cb = get32s(data+18);
    d->top_past_cb = get32s(data+22);
    d->bottom_past_cb = get32s(data+26);
    return 0;
}


/*
    __first_____    __last______    _xm_    __mx________    __my
    0x00	0x01	0x00	0x01	0x02	0x00	0x02	0x00
    ____    __ml________    __mr________    __mt________    __mb
    0x78	0x00	0x02	0x00	0x02	0x00	0x78	0x00
    ____    _ym_
    0x78	0x01

    __first_____    __last______    _?__    _m__    __xscale_____
    0x00	0x01	0x00	0x01	0x00	0x01	0x00	0x06
    __yscale____
    0x00	0x03

*/
int ni_mng_parse_magn(ni_mng_magn_t *d, const char *data, size_t len)
{
    d->first_id = len >= 2 ? get16(data) : -1;
    d->last_id = len >= 4 ? get16(data+2) : d->first_id;
    d->x_method = len >= 5 ? get8(data+4) : 0;
    /* This is a hack to handle files that seem to have a padding zero
     * inserted here.
     */
    if(0 == d->x_method && len > 5){
        d->x_method = get8(data+5);
        ++data;
        --len;
    }
    d->mx = d->x_method != 0 && len >= 7 ? get16(data+5) : 1;
    d->my = len >= 9 ? get16(data+7) : d->mx;
    d->ml = len >= 11 ? get16(data+9) : d->mx;
    d->mr = len >= 13 ? get16(data+11) : d->mx;
    d->mt = len >= 15 ? get16(data+13) : d->my;
    d->mb = len >= 17 ? get16(data+15) : d->my;
    d->y_method = len >= 18 ? get8(data+17) : d->x_method;
    return 0;
}


int ni_mng_parse_term(ni_mng_term_t *d, const char *data, size_t len)
{
    if(len<1)
        return -1;
    d->termination_action = get8(data);
    if(len >= 10){
        d->action_after_iterations = get8(data+1);
        d->delay = get32u(data+2);
        d->iteration_max = get32u(data+6);
    }else{
        d->action_after_iterations = 0;
        d->delay = 0;
        d->iteration_max = 0x7fffffff;
    }
    return 0;
}


int
ni_mng_parse_back(ni_mng_back_t *d, const char *data, size_t len)
{
    if(len<6)
        return -1;
    d->red   = (double)get16(data) / UINT16_MAX;
    d->green = (double)get16(data+2) / UINT16_MAX;
    d->blue  = (double)get16(data+4) / UINT16_MAX;
    d->mandatory = len>= 7 ? get8(data+6) : 0;
    d->object_id = len >= 9 ? get16(data+7) : -1;
    d->tiling = len >= 10 ? get8(data+9) : 0;
    return 0;
}


int
ni_mng_parse_fram(ni_mng_fram_t *d, const char *data, size_t len)
{
    d->framing_mode = len >= 1 ? get8(data) : 0;
    d->interframe_delay.change = 0;
    d->timeout_and_termination.change = 0;
    d->clipping_boundaries.change = 0;
    d->sync_id_list.change = 0;
    if(len >= 2){
        const char *end = data + len;
        d->name = ++data;
        d->name_len = strnlen(d->name, end - d->name);
        if(data + d->name_len >= end)
            return -1;
        data += d->name_len + 1;
        if(data + 4 <= end){
            d->interframe_delay.change = get8(data);
            d->timeout_and_termination.change = get8(data+1);
            d->clipping_boundaries.change = get8(data+2);
            d->sync_id_list.change = get8(data+3);
            data += 4;

            if(d->interframe_delay.change){
                if(data + 4 > end)
                    return -1;
                d->interframe_delay.ticks = get32u(data);
                data += 4;
            }
            if(d->timeout_and_termination.change){
                if(data + 4 > end)
                    return -1;
                d->timeout_and_termination.ticks = get32u(data);
                data += 4;
            }
            if(d->clipping_boundaries.change){
                if(data + 17 > end)
                    return -1;
                d->clipping_boundaries.type = get8(data);
                d->clipping_boundaries.left = get32s(data+1);
                d->clipping_boundaries.right = get32s(data+5);
                d->clipping_boundaries.top = get32s(data+9);
                d->clipping_boundaries.bottom = get32s(data+13);
                data += 16;
            }
            if(d->sync_id_list.change){
                if(data + 4 > end)
                    return -1;
                d->sync_id_list.id = get32u(data);
                //data += 4;
            }
        }
    }else{
        d->name = "";
        d->name_len = 0;
    }
    return 0;
}


int ni_mng_parse_move(ni_mng_move_t *d, const char *data, size_t len)
{
    if(len < 13)
        return -1;
    d->first = get16(data);
    d->last = get16(data+2);
    d->type = get8(data+4);
    d->x = get32s(data+5);
    d->y = get32s(data+9);
    return 0;
}


int
ni_mng_parse_clip(ni_mng_clip_t *d, const char *data, size_t len)
{
    if(len < 21)
        return -1;
    d->first = get16(data);
    d->last = get16(data+2);
    d->type = get8(data+4);
    d->left_cb = get32s(data+5);
    d->right_cb = get32s(data+9);
    d->top_cb = get32s(data+13);
    d->bottom_cb = get32s(data+17);
    return 0;
}


int
ni_mng_parse_show(ni_mng_show_t *d, const char *data, size_t len)
{
    if(len < 2)
        return -1;
    d->first = get16(data);
    d->last = len >= 4 ? get16(data+2) : d->first;
    d->mode = len >= 5 ? get8(data+4) : 0;
    return 0;
}


int
ni_mng_parse_prom(ni_mng_prom_t *d, const char *data, size_t len)
{
    if(len < 3)
        return -1;
    d->new_color_type = get8(data);
    d->new_sample_depth = get8(data+1);
    d->fill_method = get8(data+2);
    return 0;
}
