/*
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This file is public domain.
 */
#ifndef ni_mng_h
#define ni_mng_h

#include <endian.h>
#include <stddef.h>
#include <stdint.h>

#include "loa-abstract.h"
#include "ni-xng.h"

#define MNG_MHDR XNG_ID('M','H','D','R')
#define MNG_MEND XNG_ID('M','E','N','D')
#define MNG_LOOP XNG_ID('L','O','O','P')
#define MNG_ENDL XNG_ID('E','N','D','L')

#define MNG_DEFI XNG_ID('D','E','F','I')
#define MNG_PLTE XNG_ID('P','L','T','E')
#define MNG_tRNS XNG_ID('t','R','N','S')
#define MNG_IHDR XNG_ID('I','H','D','R')
#define MNG_JHDR XNG_ID('J','H','D','R')
#define MNG_BASI XNG_ID('B','A','S','I')
#define MNG_IDAT XNG_ID('I','D','A','T')
#define MNG_IEND XNG_ID('I','E','N','D')
#define MNG_CLON XNG_ID('C','L','O','N')
#define MNG_DHDR XNG_ID('D','H','D','R')
#define MNG_PAST XNG_ID('P','A','S','T')
#define MNG_MAGN XNG_ID('M','A','G','N')
/* DISC */
#define MNG_TERM XNG_ID('T','E','R','M')

#define MNG_BACK XNG_ID('B','A','C','K')
#define MNG_FRAM XNG_ID('F','R','A','M')
#define MNG_MOVE XNG_ID('M','O','V','E')
#define MNG_CLIP XNG_ID('C','L','I','P')
#define MNG_SHOW XNG_ID('S','H','O','W')

#define MNG_SAVE XNG_ID('S','A','V','E')
#define MNG_SEEK XNG_ID('S','E','E','K')

/* eXPI */
/* fPRI */
/* nEED */
/* pHYg */

#define MNG_tIME XNG_ID('t','I','M','E')

#define MNG_PROM XNG_ID('P','R','O','M')

typedef struct{
    int width;
    int height;
    int ticks_per_second;
    int layer_count;
    int frame_count;
    int play_time;
    uint32_t simplicity_profile;
}ni_mng_mhdr_t;

typedef struct{
    int nest_level;
    int iteration_count;
    int termination_condition;
    int iteration_min;
    int iteration_max;
    int signal_number;
}ni_mng_loop_t;

typedef struct{
    int nest_level;
}ni_mng_endl_t;

typedef struct{
    int object_id;
    int hide;
    int concrete;
    int posx;
    int posy;
    int clip_available;
    int clipx1;
    int clipy1;
    int clipx2;
    int clipy2;
}ni_mng_defi_t;

typedef struct{
    int width;
    int height;
    int sample_depth;
    int color_type;
    int compression_method;
    int filter_method;
    int interlace_method;
    double red;
    double green;
    double blue;
    double alpha;
    int viewable;
}ni_mng_basi_t;

typedef struct{
    int src_id;
    int dst_id;
    int clone_type;
    int hide;
    int concrete;
    int loca_delta_type;
    int x_location;
    int y_location;
}ni_mng_clon_t;

typedef struct{
    int object_id;
    int image_type;
    int delta_type;
    int block_width;
    int block_height;
    int x_location;
    int y_location;
}ni_mng_dhdr_t;

typedef struct{
    int destination_id;
    int target_delta_type;
    int target_x;
    int target_y;
}ni_mng_past_t;

typedef struct{
    int source_id;
    int composition_mode;
    int orientation;
    int offset_origin;
    int x_offset;
    int y_offset;
    int boundary_origin;
    int left_past_cb;
    int right_past_cb;
    int top_past_cb;
    int bottom_past_cb;
}ni_mng_past_entry_t;

typedef struct{
    int first_id;
    int last_id;
    int x_method;
    int mx;
    int my;
    int ml;
    int mr;
    int mt;
    int mb;
    int y_method;
}ni_mng_magn_t;

typedef struct{
    int termination_action;
    int action_after_iterations;
    int delay;
    int iteration_max;
}ni_mng_term_t;

typedef struct{
    double red;
    double green;
    double blue;
    int mandatory;
    int object_id;
    int tiling;
}ni_mng_back_t;

typedef struct{
    int framing_mode;
    const char *name;
    int name_len;
    struct{
        int change;
        int ticks;
    }interframe_delay;
    struct{
        int change;
        int ticks;
    }timeout_and_termination;
    struct{
        int change;
        int type;
        int left;
        int right;
        int top;
        int bottom;
    }clipping_boundaries;
    struct{
        int change;
        int id;
    }sync_id_list;
}ni_mng_fram_t;

typedef struct{
    int first;
    int last;
    enum{
        NI_MNG_MOVE_ABSOLUTE = 0,
        NI_MNG_MOVE_RELATIVE = 1
    }type;
    int x;
    int y;
}ni_mng_move_t;

typedef struct{
    int first;
    int last;
    enum{
        NI_MNG_CLIP_ABSOLUTE = 0,
        NI_MNG_CLIP_RELATIVE = 1
    }type;
    int left_cb;
    int right_cb;
    int top_cb;
    int bottom_cb;
}ni_mng_clip_t;

typedef struct{
    int first;
    int last;
    int mode;
}ni_mng_show_t;

typedef struct{
    int new_color_type;
    int new_sample_depth;
    int fill_method;
}ni_mng_prom_t;

typedef struct _ni_mng_t ni_mng_t;

typedef struct{
    loa_abstract_interface_t abstract;
    int(*png_begin)(void *userdata);
    int(*png_feed)(void *userdata, const char*, size_t);
    loa_abstract_t*(*png_end)(void *userdata);
    int(*jng_begin)(void *userdata);
    int(*jng_feed)(void *userdata, const char*, size_t);
    loa_abstract_t*(*jng_end)(void *userdata);
}ni_mng_callbacks_t;

typedef int(*ni_mng_draw_func_t)(void *userdata, double x, double y, double xscale, double yscale, int flags, loa_abstract_t*, nil_clip_t*);

int ni_mng_parse_mhdr(ni_mng_mhdr_t*, const char*, size_t);
int ni_mng_parse_loop(ni_mng_loop_t*, const char*, size_t);
int ni_mng_parse_endl(ni_mng_endl_t*, const char*, size_t);
int ni_mng_parse_defi(ni_mng_defi_t*, const char*, size_t);
int ni_mng_parse_basi(ni_mng_basi_t*, const char*, size_t);
int ni_mng_parse_clon(ni_mng_clon_t*, const char*, size_t);
int ni_mng_parse_dhdr(ni_mng_dhdr_t*, const char*, size_t);
int ni_mng_parse_past(ni_mng_past_t*, const char*, size_t);
int ni_mng_parse_past_entry(ni_mng_past_entry_t*, const char*, size_t);
int ni_mng_parse_magn(ni_mng_magn_t*, const char*, size_t);
int ni_mng_parse_term(ni_mng_term_t*, const char*, size_t);
int ni_mng_parse_back(ni_mng_back_t*, const char*, size_t);
int ni_mng_parse_fram(ni_mng_fram_t*, const char*, size_t);
int ni_mng_parse_move(ni_mng_move_t*, const char*, size_t);
int ni_mng_parse_clip(ni_mng_clip_t*, const char*, size_t);
int ni_mng_parse_show(ni_mng_show_t*, const char*, size_t);
int ni_mng_parse_prom(ni_mng_prom_t*, const char*, size_t);

ni_mng_t *ni_mng_decode(const char *data, size_t, const ni_mng_callbacks_t*, void *userdata);
void ni_mng_free(ni_mng_t*);
int ni_mng_get_width(ni_mng_t*);
int ni_mng_get_height(ni_mng_t*);
int ni_mng_get_scalable_time(ni_mng_t*);
int ni_mng_get_page_count(ni_mng_t*);
const char *ni_mng_get_page_name(ni_mng_t*, int page);
int ni_mng_get_layer_count(ni_mng_t*, int page);
int ni_mng_get_frame_count(ni_mng_t*, int page);
double ni_mng_get_frame_duration(ni_mng_t*, int page, int frame);
int ni_mng_draw(const ni_mng_t*, int page, int *layer_visibility, double time, ni_mng_draw_func_t, void *userdata);

#endif /* ni_mng_h */
