/*
 * Copyright (C) 2018 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <abydos-plugin.h>
#include <netpbm/pam.h>

#include "debug.h"


struct _abydos_plugin_handle_t{
    cairo_surface_t *surface;
    abydos_plugin_info_t *info;
};


static const char **
_netpbm_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-portable-anymap",
        "image/x-portable-arbitrarymap",
        "image/x-portable-bitmap",
        "image/x-portable-graymap",
        "image/x-portable-greymap",
        "image/x-portable-pixmap",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_netpbm_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_netpbm_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static cairo_surface_t *
_surface_from_pam(abydos_plugin_handle_t *h, struct pam *pam)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,h->info->width,h->info->height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-h->info->width;
    tuple *row = pnm_allocpamrow(pam);
    int x;
    int y;
    if(!strcmp(pam->tuple_type,PAM_PBM_TUPLETYPE)|
       !strcmp(pam->tuple_type,PAM_PGM_TUPLETYPE))
    {
        for(y=0;y<h->info->height;++y){
            pnm_readpamrow(pam,row);
            for(x=0;x<h->info->width;++x){
                int val = 255*row[x][0]/pam->maxval;
                *dst++ = (val<<16)|
                         (val<< 8)|
                         (val<< 0)|
                         (255<<24);
            }
            dst += rowskip;
        }
    }else if(!strcmp(pam->tuple_type,PAM_PPM_TUPLETYPE)){
        for(y=0;y<h->info->height;++y){
            pnm_readpamrow(pam,row);
            for(x=0;x<h->info->width;++x){
                *dst++ = ((255*row[x][0]/pam->maxval)<<16)|
                         ((255*row[x][1]/pam->maxval)<< 8)|
                         ((255*row[x][2]/pam->maxval)<< 0)|
                         (255<<24);
            }
            dst += rowskip;
        }
    }else if(!strcmp(pam->tuple_type,PAM_PBM_ALPHA_TUPLETYPE)|
             !strcmp(pam->tuple_type,PAM_PGM_ALPHA_TUPLETYPE))
    {
        int alphamax = pam->maxval*pam->maxval;
        for(y=0;y<h->info->height;++y){
            pnm_readpamrow(pam,row);
            for(x=0;x<h->info->width;++x){
                int alpha = row[x][3];
                int val = 255*alpha*row[x][0]/alphamax;
                *dst++ = (val<<16)|
                         (val<< 8)|
                         (val<< 0)|
                         ((255*alpha/pam->maxval)<<24);
            }
            dst += rowskip;
        }
    }else if(!strcmp(pam->tuple_type,PAM_PPM_ALPHA_TUPLETYPE)){
        int alphamax = pam->maxval*pam->maxval;
        for(y=0;y<h->info->height;++y){
            pnm_readpamrow(pam,row);
            for(x=0;x<h->info->width;++x){
                int alpha = row[x][3];
                *dst++ = ((255*alpha*row[x][0]/alphamax)<<16)|
                         ((255*alpha*row[x][1]/alphamax)<< 8)|
                         ((255*alpha*row[x][2]/alphamax)<< 0)|
                         ((255*alpha/pam->maxval)<<24);
            }
            dst += rowskip;
        }
    }else{
        cairo_surface_destroy(surface);
        surface = NULL;
        h->info->error = "Unknown pixel type";
    }
    pnm_freepamrow(row);
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_netpbm_create_from_stream(abydos_plugin_handle_t *h,FILE *f)
{
    struct pam pam;
    pnm_readpaminit(f, &pam, PAM_STRUCT_SIZE(tuple_type));
    h->info->width = pam.width;
    h->info->height = pam.height;
    h->surface = _surface_from_pam(h,&pam);
    return h->surface ? 0 : -1;
}


static int
_netpbm_create_from_data(abydos_plugin_handle_t *h, const char *bytes, size_t len)
{
    FILE *f;
    int res;
    f = fmemopen((void*)bytes,len,"r");
    if(!f)
        return -1;
    res = _netpbm_create_from_stream(h,f);
    fclose(f);
    return res;
}


static int
_netpbm_create_from_file(abydos_plugin_handle_t *h, const char *filename)
{
    FILE *f;
    int res;
    f = fopen(filename,"r");
    if(!f)
        return -1;
    res = _netpbm_create_from_stream(h,f);
    fclose(f);
    return res;
}


static cairo_surface_t *
_netpbm_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant, double time, int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0(){
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _netpbm_get_mime_types,
        .new = _netpbm_new,
        .free = _netpbm_free,
        .create_from_data = _netpbm_create_from_data,
        .create_from_file = _netpbm_create_from_file,
        .get_image_surface = _netpbm_get_image_surface
    };
    pm_init(PACKAGE,0);
    return &p;
}
