/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
extern "C"{
#define new new_
#include <abydos-plugin.h>
#undef new
#include <debug.h>
}
#include <OpenImageIO/imageio.h>

//#define ENABLE_CINEON /* not working */
#define ENABLE_DDS
#define ENABLE_DPX
//#define ENABLE_FITS /* segfaulting */
#define ENABLE_OPENEXR

/*
  BMP
* Cineon
  DDS
  DPX
  Field3D
  FITS
  GIF
  HDR/RGBE
  ICO
  IFF
  JPEG
  JPEG-2000
  Movie formats (using ffmpeg)
  OpenEXR
  PNG
  PNM / Netpbm
  PSD
  Ptex
  RLA
  SGI
  Softimage PIC
  RAW digital camera files
  Targa
  TIFF
  Webp
  Zfile
*/

OIIO_NAMESPACE_USING


typedef struct{
    cairo_surface_t *surface;
}layer_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    layer_t *layer;
    cairo_surface_t *surface;
};


static cairo_surface_t *
_surface_from_any(unsigned char *src, int width, int height, int channels)
{
    cairo_surface_t *surface = cairo_image_surface_create(channels>=4?CAIRO_FORMAT_ARGB32:CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-width;
    switch(channels){
    case 1:
        for(int y=0;y<height;++y){
            for(int x=0;x<width;++x){
                int v = *src++;
                *dst++ = v<<16|
                         v<< 8|
                         v<< 0;
            }
            dst += rowskip;
        }
        break;
    case 4:
        for(int y=0;y<height;++y){
            for(int x=0;x<width;++x){
                int r = *src++;
                int g = *src++;
                int b = *src++;
                int alpha = *src++;
                *dst++ = (alpha*r/255)<<16|
                         (alpha*g/255)<< 8|
                         (alpha*b/255)<< 0|
                         alpha<<24;
            }
            dst += rowskip;
        }
        break;
    case 12:
        for(int y=0;y<height;++y){
            for(int x=0;x<width;++x){
                int r = src[0];
                int g = src[1];
                int b = src[2];
                int alpha = src[3];
                *dst++ = (alpha*r/255)<<16|
                         (alpha*g/255)<< 8|
                         (alpha*b/255)<< 0|
                         alpha<<24;
                src += channels;
            }
            dst += rowskip;
        }
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


extern "C" const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = [](){
            static const char *mime_types[] = {
#ifdef ENABLE_CINEON
                "image/x-cineon",
#endif
#ifdef ENABLE_DPX
                "image/x-digital-picture-exchange", "image/x-dpx",
#endif
#ifdef ENABLE_FITS
                "application/fits", "image/fits",
#endif
#ifdef ENABLE_OPENEXR
                "image/x-exr",
#endif
                NULL
            };
            return mime_types;
        },
        .new_ = [](const char *mime_type, abydos_plugin_info_t *info){
            auto *h = new abydos_plugin_handle_t;
            h->info = info;
            h->info->threadsafe = 1;
            h->surface = NULL;
            return h;
        },
        .free = [](abydos_plugin_handle_t *h){
            if(h->surface)
                cairo_surface_destroy(h->surface);
            delete h;
        },
        .create_from_file = [](abydos_plugin_handle_t *h,const char *filename){
            auto in = ImageInput::open(filename);
            if(!in)
                return -1;
            const ImageSpec &spec = in->spec();
            h->info->width = spec.width;
            h->info->height = spec.height;
            for (int i = 0; i < spec.nchannels; ++i){
                DEBUG_p("Channel %d is %s",i,spec.channelnames[i].c_str());
            }
            std::vector<uint8_t> pixels(h->info->width*h->info->height*spec.nchannels);
            in->read_image (TypeDesc::UINT8, &pixels[0]);
            in->close();
            h->surface = _surface_from_any(&pixels[0], h->info->width, h->info->height, spec.nchannels);
            return h->surface ? 0 : -1;
        },
        .get_image_surface = [](abydos_plugin_handle_t *h,
                                int page,
                                cairo_bool_t *layer_visibility,
                                int variant,
                                double time,
                                int frame)
        {
            return cairo_surface_reference(h->surface);
        },
    };
    return &p;
}
