/*
 * Copyright (C) 2020 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#define PY_SSIZE_T_CLEAN
#include <abydos-plugin.h>
#include <debug.h>
#include <Imaging/Imaging.h>
#include <nil-cairo.h>
#include <Python.h>
#include <stdlib.h>
#include <string.h>

/*

Type code   | write |
------------+-------<
  BLP       | no    |
* BMP       | yes   |
  CUR       | no    |
  DCX       | no    |
  DDS       | no    |
  DIB       | yes   |
* EPS       | yes   |
  FLI, FLC  | no    |
* FPX       | no    |
  FTEX      | no    |
  GBR       | no    |
  GD        | no    |
* GIF       | yes   |
  ICNS      | yes   |
  ICO       | yes   |
* IM        | yes   |
  IMT       | no    |
  IPTC/NAA  | no    |
* JPEG      | yes   |
* JPEG 2000 | yes   |
  MCIDAS    | no    |
  MIC       | no    |
  MPO       | no    |
  MSP       | yes   |
  PCD       | no    |
* PCX       | yes   |
  PIXAR     | no    |
* PNG       | yes   |
  PPM       | yes   |
  PSD       | no    |
  SGI       | yes   |
* SPIDER    | yes   |
* TGA       | yes   |
* TIFF      | yes   |
  WAL       | no    |
  WebP      | yes   |
  WMF       | no    |
  XBM       | yes   |
  XPM       | no    |
------------^-------'

*/

typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};

typedef struct{
    PyObject_HEAD
    Imaging image;
}image_object_t;


static const char **
_pil_get_mime_types()
{
    static const char *mime_types[] = {
        "image/bmp",
        "image/x-eps", "application/eps", "application/x-eps", "image/eps",
        "image/vnd.fpx", "image/x-flash-pix",
        "image/gif",
        "image/jpeg",
        "image/jp2", "image/jpeg2000", "image/jpx",
        "image/x-labeye-image",
        "image/vnd.zbrush.pcx",
        "image/png",
        "image/x-spider",
        "image/x-tga",
        "image/tiff",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_pil_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe_render = 1;
    h->frame = NULL;
    return h;
}


static void
_pil_free(abydos_plugin_handle_t *h)
{
    free(h);
}


static cairo_surface_t *
_surface_from_image(Imaging img)
{
    cairo_surface_t *surface;
    int is_indexed = 0;
    nile_format_t dst_format;
    nile_format_t src_format;
    int channel_width = img->pixelsize / img->bands;
    nile_blit_t blit;
    uint8_t *dst;
    int dst_stride;
    int y;
    if(!strcmp(img->mode, "F")){
        dst_format = NIL_CAIRO_FORMAT_RGB24;
        src_format = NILE_FORMAT_g32;
    }else if(!strcmp(img->mode, "P")){
        is_indexed = 1;
        dst_format = NIL_CAIRO_FORMAT_ARGB32;
        src_format = NILE_FORMAT_r8g8b8a8;
    }else if(!strcmp(img->mode, "RGB")){
        switch(channel_width){
        case 1:
            src_format = NILE_FORMAT_r8g8b8;
            dst_format = NIL_CAIRO_FORMAT_RGB24;
            break;
        case 2:
            src_format = NILE_FORMAT_r16g16b16;
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
            dst_format = NIL_CAIRO_FORMAT_RGB96F;
#else
            dst_format = NIL_CAIRO_FORMAT_RGB24;
#endif
            break;
        case 4:
            src_format = NILE_FORMAT_r32g32b32;
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
            dst_format = NIL_CAIRO_FORMAT_RGB96F;
#else
            dst_format = NIL_CAIRO_FORMAT_RGB24;
#endif
            break;
        default:
            return NULL;
        }
    }else if(!strcmp(img->mode, "RGBA")){
        switch(channel_width){
        case 1:
            src_format = NILE_FORMAT_r8g8b8a8;
            dst_format = NIL_CAIRO_FORMAT_ARGB32;
            break;
        case 2:
            src_format = NILE_FORMAT_r16g16b16a16;
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
            dst_format = NIL_CAIRO_FORMAT_RGBA128F;
#else
            dst_format = NIL_CAIRO_FORMAT_ARGB24;
#endif
            break;
        case 4:
            src_format = NILE_FORMAT_r32g32b32a32;
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
            dst_format = NIL_CAIRO_FORMAT_RGBA128F;
#else
            dst_format = NIL_CAIRO_FORMAT_ARGB24;
#endif
            break;
        default:
            return NULL;
        }
    }else{
        return NULL;
    }
    channel_width = img->pixelsize / img->bands;
    surface = nil_cairo_surface_create_for_format(dst_format, img->xsize, img->ysize);
    dst = cairo_image_surface_get_data(surface);
    dst_stride = cairo_image_surface_get_stride(surface);
    if(is_indexed){
        nile_blit_init_indexed_rgb(&blit, 4, dst_format, img->palette->palette, 256, 4, src_format);
    }else{
        nile_blit_init_direct_rgb(&blit, 4, dst_format, 4, src_format);
    }
    for(y=0;y<img->ysize;++y){
        nile_blit_line(&blit, dst, (uint8_t*)img->image[y], img->xsize);
        dst += dst_stride;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_pil_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    int ret = -1;
    PyObject *bytes = NULL;
    PyObject *io = NULL;
    PyObject *img = NULL;
    PyObject *im = NULL;
    PyObject *mod;
    PyObject *pixels;
    Imaging image;

    Py_Initialize();

    bytes = PyBytes_FromStringAndSize(data, len);

    mod = PyImport_ImportModule("io");
    if(!mod){
        DEBUG_p("io module not found");
        PyErr_Print();
        goto out;
    }
    io = PyObject_CallMethod(mod, "BytesIO", "O", bytes);

    mod = PyImport_ImportModule("PIL.Image");
    if(!mod){
        DEBUG_p("PIL.Image not found");
        PyErr_Print();
    }
    img = PyObject_CallMethod(mod, "open", "O", io);
    if(!img){
        DEBUG_p("image could not be decoded with PIL");
        goto out;
    }
    pixels = PyObject_CallMethod(img, "load", "");
    Py_DECREF(pixels);
    im = PyObject_GetAttrString(img, "im");
    image = ((image_object_t*)im)->image;

    h->info->width = image->xsize;
    h->info->height = image->ysize;
    printf("width: %d, height: %d\n", h->info->width, h->info->height);

    h->frame = (frame_t*)malloc(h->info->frame_count * sizeof(frame_t));
    h->frame[0].surface = _surface_from_image(image);
    ret = 0;

out:
    if(im)
        Py_DECREF(im);
    if(img)
        Py_DECREF(img);
    if(io)
        Py_DECREF(io);
    if(bytes)
        Py_DECREF(bytes);
    Py_Finalize();
    return ret;
}


static cairo_surface_t *
_pil_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_pil_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _pil_get_mime_types,
        .new = _pil_new,
        .free = _pil_free,
        .create_from_data = _pil_create_from_data,
        .get_image_surface = _pil_get_image_surface,
        .get_duration = _pil_get_duration
    };
    return &p;
}
