/*
 * Copyright (C) 2019 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos.h>
#include <abydos-plugin.h>
#include <debug.h>
#include <endian.h>
#include <nil-cairo.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strncasecmp() */

#ifdef ENABLE_FFMPEG
# include "ffmpegdec.h"
#endif

/*

  idsc atom:

offset | size     | description
-------+----------+---------------------------------------------------------
     0 | 32 bits  | Atom size (=94)
     4 |  4 bytes | Atom type (='idsc')
     8 | 32 bits  | Image description size
    12 |  4 bytes | Compression (see below)
    16 | 16 bits  | Reserved 1
    18 | 16 bits  | Reserved 2
    20 | 32 bits  | Reservet 3
    24 | 32 bits  | Data version
    28 |  4 bytes | Vendor code (='appl' for Apple)
    32 | 32 bits  | Temporal quality (0 = no temporal compression)
    36 | 32 bits  | Spatial quality
    40 | 16 bits  | Width
    42 | 16 bits  | Height
    44 | 32 bits  | Horizontal resolution
    48 | 32 bits  | Vertical resolution
    52 | 32 bits  | Data size
    56 | 16 bits  | Frame count
    58 | 32 bytes | Compressor name (Pascal string)
    90 | 16 bits  | Bit depth
    92 | 16 bits  | Color lookup table
-------^----------^---------------------------------------------------------

  Compression:

    'rle ' : Animation
    'pxlt' : Apple Pixlet Video
    'h263' : Apple VC H.263
    'dvcp' : DV - PAL
    'dvc ' : DV/DVCPRO - NTSC
    'dvpp' : DVCPRO - PAL
    'dv5n' : DVCPRO50 - NTSC
    'dv5p' : DVCPRO50 - PAL
    'jpeg' : Foto - JPEG
    'avc1' : H.264
    'raw ' : None
    'mjp2' : JPEG 2000
    'mp4v' : MPEG-4 Video
    'png ' : PNG

    'cvid' : Cinepak
    'smc ' : Graphics
    'rpza' : Apple video
    'kpcd' : Kodak Photo CD
    'mjpa' : Motion-JPEG (format A)
    'mjpb' : Motion-JPEG (format B)
    'SVQ1' : Sorenson video, version 1
    'SVQ3' : Sorenson video 3
    'gif ' : CompuServe Graphics Interchange Format
    'tiff' : Tagged Image File Format
    '2vuY' : Uncompressed Y´CbCr, 8-bit-per-component 4:2:2
    'yuv2' : Uncompressed Y´CbCr, 8-bit-per-component 4:2:2
    'v308' : Uncompressed Y´CbCr, 8-bit-per-component 4:4:4
    'v408' : Uncompressed Y´CbCr, 8-bit-per-component 4:4:4:4
    'v216' : Uncompressed Y´CbCr, 10, 12, 14, or 16-bit-per-component 4:2:2
    'v410' : Uncompressed Y´CbCr, 10-bit-per-component 4:4:4
    'v210' : Uncompressed Y´CbCr, 10-bit-per-component 4:2:2

*/

#if __BYTE_ORDER == __BIG_ENDIAN
# define ID(a,b,c,d) ((a<<24)|(b<<16)|(c<<8)|(d<<0))
#else
# define ID(a,b,c,d) ((a<<0)|(b<<8)|(c<<16)|(d<<24))
#endif

typedef enum{
  SUBTYPE_ANIMATION    = ID('r','l','e',' '),
  SUBTYPE_PIXLET       = ID('p','x','l','t'),
  SUBTYPE_H263         = ID('h','2','6','3'),
  SUBTYPE_DV_PAL       = ID('d','v','c','p'),
  SUBTYPE_DV_HTSC      = ID('d','v','c',' '),
  SUBTYPE_DVPRO_PAL    = ID('d','v','p','p'),
  SUBTYPE_DVPRO50_NTSC = ID('d','v','5','n'),
  SUBTYPE_DVPRO50_PAL  = ID('d','v','5','p'),
  SUBTYPE_JPEG         = ID('j','p','e','g'),
  SUBTYPE_H264         = ID('a','v','c','1'),
  SUBTYPE_NONE         = ID('r','a','w',' '),
  SUBTYPE_JPEG2000     = ID('m','j','p','2'),
  SUBTYPE_MPEG4        = ID('m','p','4','v'),
  SUBTYPE_PNG          = ID('p','n','g',' ')
}subtype_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
    abydos_t *ar;
};


static const char **
_qtif_get_mime_types()
{
    static const char *mime_types[] = {
        "image/qtif",
        "image/x-quicktime",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_qtif_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    h->ar = NULL;
    return h;
}


static void
_qtif_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    if(h->ar)
        abydos_destroy(h->ar);
    free(h);
}


#define _get16(p) be16toh(*(uint16_t*)(p))
#define _get32(p) be32toh(*(uint32_t*)(p))


static const char *
_find_atom(uint32_t id,const char *p,size_t len){
    while(len>8){
        size_t atomsize = _get32(p);
        if(atomsize>len)
            return NULL;
        if(*(uint32_t*)(p+4)==id)
            return p;
        p += atomsize;
        len -= atomsize;
    }
    return NULL;
}


static inline uint32_t
convert_pixelRGBA(uint32_t pixel)
{
#if __BYTE_ORDER == __BIG_ENDIAN
    int b = pixel & 0xff;
    int g = (pixel>>8) & 0xff;
    int r = (pixel>>16) & 0xff;
    int alpha = pixel >> 24;
#else
    int alpha = pixel & 0xff;
    int r = (pixel>>8) & 0xff;
    int g = (pixel>>16) & 0xff;
    int b = pixel >> 24;
#endif
    return r<<16|g<<8|b|alpha<<24;
}


/*

  14 byte feader (?):

    0x40     =     64 = channels << 8 ?
    0x03d520 = 251168 = data length
    0x0008   =      8 = additional bytes
    0x0000   =      0 = starting line?
    0x0000   =      0
    0x0258   =      600 = height?
    0x0000   =      0


*/
static int
_rle_unpack(const uint8_t *src, size_t src_len, cairo_surface_t *surface, int channels)
{
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int height = cairo_image_surface_get_height(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t); // channels * width;
    const uint8_t *src_end = src + src_len;
    uint32_t *dst_end;
    enum{
        S_LINE,
        S_SKIP,
        S_RLE
    }state = S_LINE;
    int header_length;

    if(src_len<6)
        return -1;

    header_length = _get16(src+4);
    DEBUG_p("header length: %d",header_length);
    if(src_len<6+header_length)
        return -1;
    if(header_length>=2)
        dst += stride*_get16(src+6);
    if(header_length>=6)
        height = _get16(src+10);

    src += 6 + header_length;

    dst_end = dst + stride * height;
    while(dst<dst_end){
        uint32_t *line_end;
        int b;
        if(S_LINE==state){
            line_end = dst + stride;
            state = S_SKIP;
        }else if(S_SKIP==state){
            if(src>=src_end){
                /* premature end of data */
                DEBUG_p("premature end of data");
                return -1;
            }
            b = *src++;
            if(0==b){
                /* end of image */
                DEBUG_p("end of image");
                return 0;
            }
            //b = channels*(b-1);
            --b;
            if(dst+b>dst_end){
                /* image overflow */
                DEBUG_p("image overflow");
                return -1;
            }
            dst += b;
            state = S_RLE;
        }else /* if(S_RLE==state) */ {
            if(src>=src_end){
                /* premature end of data */
                DEBUG_p("premature end of data");
                return -1;
            }
            b = *(int8_t*)src++;
            if(0==b){
                state = S_SKIP;
            }else if(-1==b){
                /* end of line */
                dst = line_end;
                state = S_LINE;
            }else if(b>0){
                if(src+b*channels>src_end){
                    /* premature end of data */
                    DEBUG_p("premature end of data");
                    return -1;
                }
                if(dst+b>dst_end){
                    /* image overflow */
                    DEBUG_p("image overflow");
                    return -1;
                }
                while(b-->0){
                    *dst++ = convert_pixelRGBA(*(uint32_t*)src);
                    src += channels;
                }
            }else if(3==channels){
                /* FIXME!! */
                return -1;
            }else if(4==channels){
                uint32_t pixel;
                if(src+4 > src_end){
                    /* premature end of data */
                    DEBUG_p("premature end of data");
                    return -1;
                }
                if(dst-b > dst_end){
                    /* image overflow */
                    DEBUG_p("image overflow");
                    return -1;
                }
                pixel = convert_pixelRGBA(*(uint32_t*)src);
                src += 4;
                while(b++<0)
                    *dst++ = pixel;
            }
        }
    }
    DEBUG_p("src left %d",(int)(src_end-src));
    DEBUG_p("dst left %d",(int)(dst_end-dst));
    cairo_surface_mark_dirty(surface);
    return 0;
}


static int
_qtif_create_from_data(abydos_plugin_handle_t *h,const char *bytes, size_t len)
{
    const char* p;
    size_t atomsize;
    size_t datasize;
    int is_rle = 0;
    nile_format_t format;
    int channels = 0;
#ifdef ENABLE_FFMPEG
    enum AVCodecID ffmpeg_id = 0;
#endif
    const char* mime_type = NULL;
    int ret;

    /* A QTIF file cannot possibly be smaller than 25 bytes */
    if(len<25)
        return -1;
    p = _find_atom(*(uint32_t*)"idsc",bytes,len);
    if(!p)
        return -1;
    atomsize = _get32(p+0);
    if(atomsize<16)
        return -1;
    DEBUG_p("content type: '%4s'",p+12);
    switch(*(uint32_t*)(p+12)){
    case SUBTYPE_PIXLET:
#ifdef _ENABLE_FFMPEG
        /* not working */
        ffmpeg_id = AV_CODEC_ID_PIXLET;
        break;
#else
        return -1;
#endif
    case SUBTYPE_H263:
        ffmpeg_id = AV_CODEC_ID_H263;
        break;
    case SUBTYPE_DV_PAL:
    case SUBTYPE_DV_HTSC:
    case SUBTYPE_DVPRO_PAL:
    case SUBTYPE_DVPRO50_NTSC:
    case SUBTYPE_DVPRO50_PAL:
        ffmpeg_id = AV_CODEC_ID_DVVIDEO;
        break;
    case SUBTYPE_JPEG:
#ifdef _ENABLE_FFMPEG
        /* not working */
        ffmpeg_id = AV_CODEC_ID_MJPEG;
#else
        mime_type = "image/jpeg";
#endif
        break;
    case SUBTYPE_H264:
#ifdef _ENABLE_FFMPEG
        /* not working */
        ffmpeg_id = AV_CODEC_ID_H264;
        break;
#else
        return -1;
#endif
    case SUBTYPE_ANIMATION:
        is_rle = 1;
        /* fall through */
    case SUBTYPE_NONE:
        /* for the raw subformat we need the full idsc atom */
        if(atomsize<94)
            return -1;
        switch(_get16(p+90)){
        case 24:
            format = NILE_FORMAT_r8g8b8;
            channels = 3;
            break;
        case 32:
            format = NILE_FORMAT_a8r8g8b8;
            channels = 4;
            break;
        default:
            DEBUG_p("unhandled raw bit depth %d",_get16(p+90));
            return -1;
        }
        break;
    case SUBTYPE_JPEG2000:
#ifdef ENABLE_FFMPEG
        ffmpeg_id = AV_CODEC_ID_JPEG2000;
#else
        mime_type = "image/jp2";
#endif
        break;
    case SUBTYPE_MPEG4:
#ifdef _ENABLE_FFMPEG
        /* not working */
        ffmpeg_id = AV_CODEC_ID_MPEG4;
        break;
#else
        return -1;
#endif
    case SUBTYPE_PNG:
#ifdef ENABLE_FFMPEG
        ffmpeg_id = AV_CODEC_ID_PNG;
#else
        mime_type = "image/png";
#endif
        break;
    default:
        /* unknown data type */
        return -1;
    }
    if(atomsize>=44){
        h->info->width = _get16(p+40);
        h->info->height = _get16(p+42);
        if(atomsize>=52){
            double h_dpi = (double)_get32(p+44)/65536;
            double v_dpi = (double)_get32(p+48)/65536;
            DEBUG_p("DPI: %fx%f",h_dpi,v_dpi);
            if(h_dpi > 0 && v_dpi > 0){
                h->info->pixel_ratio = h_dpi/v_dpi;
            }
        }
    }
    datasize = atomsize>=56 ? _get32(p+52) : 0;

    p = _find_atom(*(uint32_t*)"idat",bytes,len);
    if(!p)
        return -1;
    if(datasize > 0){
        atomsize = _get32(p+0);
        if(datasize != atomsize-8)
            return -1;
    }
    datasize = atomsize-8;

    if(channels>0){
        size_t realsize = channels * h->info->width * h->info->height;
        if(is_rle){
            h->surface = cairo_image_surface_create(channels >=4 ? CAIRO_FORMAT_ARGB32 : CAIRO_FORMAT_RGB24, h->info->width, h->info->height);
            return _rle_unpack((uint8_t*)p+8, datasize, h->surface, channels);
        }else{
            if(realsize > datasize)
                return -1;
            h->surface = nil_cairo_surface_create_from_direct((uint8_t*)p+8,h->info->width,h->info->height,-1,channels,format,0);
            return 0;
        }
    }

#ifdef ENABLE_FFMPEG
    if(ffmpeg_id){
        h->surface = ffmpeg_decode(p+8,datasize,ffmpeg_id);
        if(!h->surface)
            return -1;
        h->info->width = cairo_image_surface_get_width(h->surface);
        h->info->height = cairo_image_surface_get_height(h->surface);
        return 0;
    }
#endif

    h->ar = abydos_create(mime_type);
    if(!h->ar)
        return -1;
    ret = abydos_from_data(h->ar,p+8,datasize);
    if(ret)
        return ret;
    abydos_get_size(h->ar,&h->info->width,&h->info->height);
    return 0;
}


static void
_qtif_render(abydos_plugin_handle_t *h,
             cairo_t *cr,
             int page,
             cairo_bool_t *visible_layers,
             int variant,
             double time,
             int frame)
{
    abydos_render(h->ar, cr);
}


static cairo_surface_t *
_qtif_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return h->surface ? cairo_surface_reference(h->surface) : abydos_get_image_surface(h->ar);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _qtif_get_mime_types,
        .new = _qtif_new,
        .free = _qtif_free,
        .create_from_data = _qtif_create_from_data,
        .render = _qtif_render,
        .get_image_surface = _qtif_get_image_surface
    };
    return &p;
}
