/*
 * Copyright (C) 2019 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <limits.h>                     /* INT_MAX */
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */

#include "recoil.h"

#define MAX_LAYERS 4

#define ENABLE_ILBM
/* #define ENABLE_PC1 */
#define ENABLE_SPECTRUM512

/*

Platform      | extension(s)  | description                         | frames
--------------+---------------+-------------------------------------+-------
 Amiga        | ACBM          | Amiga Continuous Bitmap             |   1
 Amiga DCTV   | DCT, DCTV     | DCTV                                |   1
 Amiga        | DEEP          | TVPaint                             |   1
 Amiga        | HAM6          | Hold-And-Modify 6                   |   1
 Amiga        | HAM8          | Hold-And-Modify 8                   |   1
*Amiga        | IFF           | Interchange File Format             |   1
 Amiga HAM-E  | IFF           | HAM-E                               |   1
 Amiga        | INFO          | Icon                                |   1
*Amiga        | LBM, ILBM     | Interleaved Bitmap                  |   1
 Amiga        | DHR, DR, MP, BEAM | Multi-palette                   |   1
 Amiga        | RGB8          | Turbo Silver                        |   1
 Amiga        | RGBN          | Turbo Silver                        |   1
 Amiga        | SHAM          | Sliced HAM 4096 colors              |   1
 Amstrad CPC  | CM5+GFX       | Mode 5                              |   1
 Amstrad CPC  | FNT           | 8x8 font                            |   1
 Amstrad CPC  | HGB           | HinterGrundBild                     |   1
 Amstrad CPC  | PPH+ODD+EVE   | Perfect Pix                         |   2
 Amstrad CPC  | SCR+PAL       | Advanced OCP Art Studio             |   1
 Amstrad CPC  | SGX           | SymbOS graphic                      |   1
 Amstrad CPC  | WIN+PAL       | Advanced OCP Art Studio             |   1
 Apple II     | HGR           | High Resolution                     |   1
 Apple II     | SPR           | Sprites mono                        |   1
 Apple IIe    | DHGR, DHR	Double High-Resolution	560x192	mono	1
 Apple IIGS   | SH3		320x200	3200 colors	1
 Apple IIGS   | SHR		320x200	256 colors	1
*Apple Macintosh | MAC, PNT, PNTG | MacPaint, 576x720 mono          |   1
 Atari 8-bit  | 256, AP2      | 80x96 256 colors                    |   1
 Atari 8-bit  | 4MI           | AtariTools-800 4 missiles           |   1
 Atari 8-bit  | 4PL	AtariTools-800 4 players			1
 Atari 8-bit  | 4PM	AtariTools-800 4 players and 4 missiles			1
 Atari 8-bit  | A4R           | Anime 4ever                         |   1
 Atari 8-bit  | ACS	AtariTools-800 font		4 colors	1
 Atari 8-bit  | AGP	AtariTools-800 graphic			1
 Atari 8-bit  | ALL	Graph	160x192	5 colors	1
 Atari 8-bit  | AP3, APV, DGI, DGP, ESC, ILC, PZM		80x192	256 colors	2
 Atari 8-bit  | APC, PLM		80x96	256 colors	1
 Atari 8-bit  | APL	Atari Player Editor		4 colors	1
 Atari 8-bit  | APP		80x192	256 colors	2
 Atari 8-bit  | ART	Ascii-Art Editor	up to 64x24 characters	mono	1
 Atari 8-bit  | ART (arbitrary)	Artist by David Eaton	160x80	4 colors	1
 Atari 8-bit  | BG9, G09		160x192	16-level grayscale	1
 Atari 8-bit  | BGP	Bugbiter APAC239i	80x239	256 colors	2
 Atari 8-bit  | BKG	Movie Maker background	160x96	4 colors	1
 Atari 8-bit  | CCI	Champions' Interlace (compressed)	160x192		2
 Atari 8-bit  | CHR	Blazing Paddles font		mono	1
 Atari 8-bit  | CIN	Champions' Interlace	up to 160x200		2
 Atari 8-bit  | CPI	Marco Pixel Editor	160x192	4 colors	1
 Atari 8-bit  | CPR	Trzmiel (compressed)	320x192	mono	1
 Atari 8-bit  | CUT	Cut Creator	96x99	mono	1
 Atari 8-bit VBXE | DAP     | SlideShow for VBXE	320x240	256 colors	1
 Atari 8-bit  | DIN		320x192	10 colors	2
 Atari 8-bit  | DIT (arbitrary)	DrawIt	160x96	4 colors	1
 Atari 8-bit  | DLM	Dir Logo Maker	11x16 characters	mono	1
*Atari 8-bit  | DRG           | Atari CAD                           |   1
 Atari 8-bit  | F80	The Last Word font		mono	1
 Atari 8-bit  | FGE	Floor Designer	64x40	16-level grayscale	1
 Atari 8-bit  | FN2	Atari FontMaker		mono	1
 Atari 8-bit	FNT	8x8 font		mono	1
 Atari 8-bit	FWA	Fun with Art	160x192	128 colors	1
 Atari 8-bit	G10	Graphics 10	up to 80x240	9 colors	1
 Atari 8-bit	G11	Graphics 11	up to 80x240	16 colors	1
 Atari 8-bit	G2F	Graph2Font	up to 336x240	256 colors	1
 Atari 8-bit	GED	GED	160x200	128 colors	1
 Atari 8-bit	GHG	Gephard Hires Graphics	up to 320x200	mono	1
 Atari 8-bit	GR1	Mad Studio	20x24 characters	4 colors	1
 Atari 8-bit	GR2	Mad Studio	20x12 characters	4 colors	1
 Atari 8-bit	GR3	Mad Studio	40x24	4 colors	1
 Atari 8-bit	GR7	Graphics 7	up to 160x120	4 colors	1
 Atari 8-bit	GR8	Graphics 8	up to 320x240	mono	1
 Atari 8-bit	GR9	Graphics 9	up to 80x240	16-level grayscale	1
 Atari 8-bit	HCI, HR2		320x200	5 colors	2
 Atari 8-bit	HCM	Hard Color Map	128x192	9 colors	1
 Atari 8-bit	HIP	Hard Interlace Picture	160x200	30-level grayscale	2
 Atari 8-bit	HPM	Grass' Slideshow	160x192	4 colors	1
 Atari 8-bit	HR		256x239	3 colors	2
 Atari 8-bit	ICE	Interlace Character Editor font			2
 Atari 8-bit	ICN	ICE CIN	160x192	80 colors	2
 Atari 8-bit	IGE	Interlace Graphics Editor	128x96	16 colors	2
 Atari 8-bit	ILD	Interlace Logo Designer	128x128	7 colors	2
 Atari 8-bit	IMN	ICE MIN	160x192	80 colors	2
 Atari 8-bit	ING	ING 15	160x200	7 colors	2
 Atari 8-bit	INP		160x200	7 colors	2
 Atari 8-bit	INT	INT95a	up to 160x239	16 colors	2
 Atari 8-bit	IP2	ICE PCIN+	160x192	45 colors	2
 Atari 8-bit	IPC	ICE PCIN	160x192	35 colors	2
 Atari 8-bit	IR2	Super IRG 2	160x192	25 colors	2
 Atari 8-bit	IRG	Super IRG	160x192	15 colors	2
 Atari 8-bit	IST	Interlace Studio	160x200		2
 Atari 8-bit	JGP	Jet Graphics Planner		4 colors	1
 Atari 8-bit	KPR	Kompresor do Animatora		4 colors	1
 Atari 8-bit	KSS (arbitrary)	KSS-Paint	160x160	4 colors	1
 Atari 8-bit	LDM	Ludek Maker		4 colors	1
 Atari 8-bit	LEO	Larka Edytor Obiektów trybu $4+		5 colors	1
 Atari 8-bit	LUM+COL	Technicolor Dream	80x119	256 colors	1
 Atari 8-bit	KFX		56x60	mono	1
 Atari 8-bit	MAP	Envision			1
 Atari 8-bit	MAP	EnvisionPC			1
 Atari 8-bit	MAX	XL-Paint MAX	160x192		2
 Atari 8-bit	MBG	Mad Designer	512x256	mono	1
 Atari 8-bit	MCH	Graph2Font	up to 336x240	256 colors	1
 Atari 8-bit	MCP	McPainter	160x200	16 colors	2
 Atari 8-bit	MCPP	Paradox	160x100	16 colors	1
 Atari 8-bit	MCS		160x192	9 colors	1
 Atari 8-bit	MGP (arbitrary)	Magic Painter	160x96		1
 Atari 8-bit	MIC	Micro Illustrator / Graphics 15	up to 160x240	4 colors	1
 Atari 8-bit	MIC (arbitrary)	AtariGraphics	160x192	4 colors	1
 Atari 8-bit	MIC+COL	Graph2Font	160x240	128 colors	1
 Atari 8-bit	MIS	AtariTools-800 missile	2x240	mono	1
 Atari 8-bit	NLQ	Daisy-Dot font		mono	1
 Atari 8-bit	PGR	PowerGraphics	up to 336x240	256 colors	1
 Atari 8-bit	PIC	Graphic Arts Department	160x96	128 colors	1
 Atari 8-bit	PIC	Koala MicroIllustrator	160x192	4 colors	1
 Atari 8-bit	PLA	AtariTools-800 player	8x240	mono	1
 Atari 8-bit	PMD	PMG Designer by Henryk Karpowicz			1
 Atari 8-bit	RAP	Vidig Paint	80x192	16 colors	1
 Atari 8-bit	RAW	XL-Paint MAX	160x192	16 colors	2
 Atari 8-bit	RGB	ColorViewSquash	up to 160x192		3
 Atari 8-bit	RIP	Rocky Interlace Picture	up to 320x239		1 or 2
 Atari 8-bit	RM0	Rambrandt	160x96	99 colors	1
 Atari 8-bit	RM1	Rambrandt	80x192	256 colors	1
 Atari 8-bit	RM2	Rambrandt	80x192	104 colors	1
 Atari 8-bit	RM3	Rambrandt	80x192	128 colors	1
 Atari 8-bit	RM4	Rambrandt	160x192	99 colors	1
 Atari 8-bit	RYS	Mamut	160x96	4 colors	1
 Atari 8-bit	SG3	Standard Graphics 3	40x24	4 colors	1
 Atari 8-bit	SGE	Semi-Graphic logos Editor	40x24 characters	mono	1
 Atari 8-bit	SHC	SAMAR Hi-res Interlace with Map of Colours	320x192		2
 Atari 8-bit	SHP	Blazing Paddles shape table		mono	1
 Atari 8-bit	SHP	Movie Maker shapes	160x96	4 colors	1
 Atari 8-bit	SIF	Super-IRG Font		4 colors	2
 Atari 8-bit	SKP	Sketch-PadDles	160x192	mono	1
 Atari 8-bit	SPC	The Graphics Magician Picture Painter	160x192	128 colors	1
 Atari 8-bit	SXS	16x16 font		mono	1
 Atari 8-bit	TIP	Taquart Interlace Picture	up to 160x119		2
 Atari 8-bit	TX0	Texture Maker0 16x16x16	16x16	16 colors	1
 Atari 8-bit	TXE	Texture Editor by Mikey	80x96	16-level grayscale	1
 Atari 8-bit	TXS		16x16	16-level grayscale	1
 Atari 8-bit	VZI	VertiZontal Interlacing	160x200	31-level grayscale	2
 Atari 8-bit	WND	Blazing Paddles window	up to 160x192	4 colors	1
 Atari 8-bit	XLP	XL-Paint	up to 160x200	7 colors	2
 Atari Portfolio	PGC		240x64	mono	1
 Atari Portfolio	PGF		240x64	mono	1
 Atari ST/STE | ART           | Art Director                        |   1
*Atari ST/STE | ART           | GFA Artist                          |   1
 Atari ST/STE	ART	Palette Master	320x200	512 colors	1
 Atari ST/STE	BL1	DEGAS Elite block		16 colors	1
 Atari ST/STE	BL2	DEGAS Elite block		4 colors	1
 Atari ST/STE	BL3	DEGAS Elite block		mono	1
 Atari ST/STE	BLD	MegaPaint		mono	1
 Atari ST/STE	BRU	DEGAS Elite brush	8x8	mono	1
 Atari ST/STE	CA1	CrackArt	320x200	16 colors	1
 Atari ST/STE	CA2	CrackArt	640x200	4 colors	1
 Atari ST/STE	CA3	CrackArt	640x400	mono	1
 Atari ST/STE	CE1	ComputerEyes	320x200	262144 colors	1
 Atari ST/STE	CE2	ComputerEyes	640x200	32768 colors	1
 Atari ST/STE	CE3	ComputerEyes	640x400	192-level grayscale	1
 Atari ST/STE	CEL	Cyber Paint Cell	up to 320x200	16 colors	1
 Atari ST/STE	CMP	Public Painter	up to 640x800	mono	1
 Atari ST/STE	CP3	Picworks	640x400	mono	1
 Atari ST/STE	CPT	Canvas (compressed)			1
 Atari ST/STE	CPT+HBL	Canvas (raster)	up to 640x200		1
 Atari ST/STE	CRG	Calamus Raster Graphic		mono	1
 Atari ST/STE	DA4	PaintShop	640x800	mono	1
 Atari ST/STE	DOO	Doodle	640x400	mono	1
 Atari ST/STE	DU1, DUO		416x273	136 colors	2
 Atari ST/STE	DU2		832x273	10 colors	2
 Atari ST/STE	EZA	EZ-Art Professional	320x200	16 colors	1
 Atari ST/STE	FNT	8x16 font		mono	1
 Atari ST/STE	FNT	GDOS font		mono	1
 Atari ST/STE	FUL	Canvas (raster)	up to 640x200		1
 Atari ST/STE	GFB	DeskPic			1
 Atari ST/STE	HPK	Dali (compressed)	640x400	mono	1
 Atari ST/STE	HRM	HighresMedium	640x200		2
 Atari ST/STE	IC1	Imagic	320x200	16 colors	1
 Atari ST/STE	IC2	Imagic	640x200	4 colors	1
 Atari ST/STE	IC3	Imagic	640x400	mono	1
 Atari ST/STE	ICN	DEGAS Elite icon		mono	1
*Atari ST/STE | IFF           | Interchange File Format             |   1
*Atari ST/STE | IMG           | GEM Bit Image                       |   1
 Atari ST/STE	KID	Fullscreen Construction Kit	448x272	16 colors	1
 Atari ST/STE	LPK	Dali (compressed)	320x200	16 colors	1
 Atari ST/STE	MPK	Dali (compressed)	640x200	4 colors	1
*Atari ST/STE | MPP           | Multi Palette Picture               | 1 - 2
 Atari ST/STE	MUR+PAL	C.O.L.R. Object Editor	320x200	16 colors	1
*Atari ST/STE | NEO           | NEOchrome                           |  1
 Atari ST/STE | NEO           | NEOchrome Master 512 colors         |  1
 Atari ST/STE | NEO+RST	NEOchrome Master 320x200 512 colors         |  1
 Atari ST/STE	P3C	D-GRAPH (compressed)	320x200	136 colors	2
 Atari ST/STE	PA3	Pablo Paint 2.5	640x400	mono	1
*Atari ST/STE | PAC           | STAD                                |   1
*Atari ST/STE | PBX           | QuantumPaint                        |   1
*Atari ST/STE | PC1           | DEGAS Elite	320x200	16 colors       |   1
*Atari ST/STE | PC2           | DEGAS Elite	640x200	4 colors        |   1
*Atari ST/STE | PC3           | DEGAS Elite	640x400	mono            |   1
*Atari ST/STE | PCI           | Tobias Richter Fullscreen Slideshow |   2
 Atari ST/STE	PCS	PhotoChrome	320x199		1 or 2
 Atari ST/STE	PG0	Paintworks	320x400	16 colors	1
 Atari ST/STE	PG1	Graphics Processor	320x200	16 colors	1
 Atari ST/STE	PG1	Paintworks	640x400	4 colors	1
 Atari ST/STE	PG2	Graphics Processor	640x200	4 colors	1
 Atari ST/STE	PG2	Paintworks	640x800	mono	1
 Atari ST/STE	PG3	Graphics Processor	640x400	mono	1
 Atari ST/STE | PI1           | DEGAS	up to 416x560	16 colors   |   1
 Atari ST/STE | PI2           | DEGAS	640x200	4 colors            |   1
 Atari ST/STE | PI3, SUH      | DEGAS	640x400	mono                |   1
 Atari ST/STE	PIC	PaintPro ST			1
 Atari ST/STE	PL4		320x200	256 colors	2
*Atari ST/STE | PPP           | Pablo Paint 1.1                     |   1
 Atari ST/STE	PSC	PaintShop (compressed)	640x400	mono	1
*Atari ST/STE | RGB           | RGB Intermediate                    |   3
 Atari ST/STE	RGH	ZZ_ROUGH	320x200	16 colors	1
 Atari ST/STE	CL0, SC0	Paintworks	320x200	16 colors	1
 Atari ST/STE	CL1, SC1	Paintworks	640x200	4 colors	1
 Atari ST/STE	CL2, SC2	Paintworks	640x400	mono	1
 Atari ST/STE	SD0	Dali	320x200	16 colors	1
 Atari ST/STE	SD1	Dali	640x200	4 colors	1
 Atari ST/STE	SD2	Dali	640x400	mono	1
*Atari ST/STE | SPC           | Spectrum 512 (compressed) 320x199 4096 colors | 1
*Atari ST/STE | SPS           | Spectrum 512 (smooshed) 320x199	512 colors | 1
*Atari ST/STE | SPU           | Spectrum 512	320x199	4096 colors	1
 Atari ST/STE | SPU           | Spectrum 512 (enhanced)	320x199	29791 colors	2
 Atari ST/STE | SPX           | Spectrum 512 (extended)	320x199	4096 colors	1
 Atari ST/STE	SRT	Synthetic Arts	640x200	4 colors	1
 Atari ST/STE	SSB	Sinbad Slideshow	320x200	16 colors	1
 Atari ST/STE	TN1, TN4	Tiny Stuff	320x200	16 colors	1
 Atari ST/STE	TN2, TN5	Tiny Stuff	640x200	4 colors	1
 Atari ST/STE	TN3, TN6	Tiny Stuff	640x400	mono	1
 Atari ST/STE	TNY	Tiny Stuff			1
 Atari ST/STE	XIMG	Extended GEM Bit Image			1
 Atari TT     | PI4           | 320x480 256 colors                  |   1
 Atari TT     | PI5           | 640x480 16 colors                   |   1
 Atari TT     | PI6           | 1280x960 mono                       |   1
*Atari Falcon | B&W, B_W      | ImageLab                            |   1
 Atari Falcon	DC1	DuneGraph (compressed)	320x200	256 colors	1
 Atari Falcon	DEL	DelmPaint	320x240	256 colors	1
 Atari Falcon	DG1	DuneGraph	320x200	256 colors	1
 Atari Falcon	DPH	DelmPaint	640x480	256 colors	1
 Atari Falcon	ESM	TmS Cranach Paint/Studio			1
 Atari Falcon	FTC	Falcon True Color	384x240	65536 colors	1
 Atari Falcon	FUN	Funny Paint			1
 Atari Falcon	GOD	GodPaint		65536 colors	1
 Atari Falcon	HIR	Print-Technik		128-level grayscale	1
 Atari Falcon	IB3	ICDRAW icon	32x32	16 colors	1
 Atari Falcon	IBI	ICDRAW icon	32x32	16 colors	1
 Atari Falcon	IIM	InShape			1
 Atari Falcon	PI4, PI7, PI9	Fuckpaint	up to 640x480	256 colors	1
 Atari Falcon	PIX	PixArt			1
*Atari Falcon | PNT, TPI      | Prism Paint                         |   1
 Atari Falcon	RAG	RAG-D			1
 Atari Falcon	RAW, RWH, RWL	IMG Scan	up to 640x400	256-level grayscale	1
 Atari Falcon	TCP	Rembrandt		65536 colors	1
 Atari Falcon	TG1	COKE		65536 colors	1
 Atari Falcon	TIMG			32768 colors	1
 Atari Falcon	TRE	Spooky Sprites (compressed)		65536 colors	1
 Atari Falcon	TRP	EggPaint		65536 colors	1
 Atari Falcon	TRP	Spooky Sprites		65536 colors	1
 Atari Falcon	TRU	IndyPaint		65536 colors	1
*Atari Falcon | XGA           |                                     |   1
 BBC Micro    | BB0	Mode 0	640x256	mono	1
 BBC Micro    | BB1	Mode 1	320x256	4 colors	1
 BBC Micro    | BB2	Mode 2	160x256	8 colors	1
 BBC Micro    | BB4	Mode 4	320x256	mono	1
 BBC Micro    | BB5	Mode 5	160x256	4 colors	1
 BBC Micro    | BBG	LdPic	up to 640x256		1
 Commodore 16 | P4I           |                                     |   1
 Commodore 64 | 4BT	GoDot	320x200	16 colors	1
 Commodore 64 | 64C	8x8 font		mono	1
 Commodore 64 | A             | SEUCK sprites                       |   1
*Commodore 64 | A64, WIG      | Wigmore Artist 64                   |   1
 Commodore 64 | AAS, ART      | Art Studio                          |   1
 Commodore 64 | AFL           | AFLI-editor v2.0 by Topaz Beerlin   |   1
 Commodore 64 | AMI           | Amica Paint                         |   1
 Commodore 64 | BFLI          | Big FLI (148x400)                   |   1
 Commodore 64 | BML, FLI      | FLI Graph 2.2 by Blackmail          |   1
 Commodore 64 | CDU           | CDU-Paint                           |   1
*Commodore 64 | CHE           | Cheese                              |   1
 Commodore 64 | CWG           | Create with Garfield                |   1
 Commodore 64 | DD            | Doodle!                             |   1
 Commodore 64 | DOL           | Dolphin Ed                          |   1
 Commodore 64 | DRL, DLP      | Drazlace                            |   2
*Commodore 64 | DRZ, DRP      | Drazpaint                           |   1
 Commodore 64 | ECI           | ECI Graphic Editor                  |   2
 Commodore 64 | ECP           | ECI Graphic Editor (compressed)     |   2
 Commodore 64 | FBI           | Flip                                |   1
*Commodore 64 | FCP, FPT      | Face Painter                        |   1
 Commodore 64 | FD2           | FLI Designer                        |   1
 Commodore 64 | FFLI          |                                     |   2
 Commodore 64 | FLI           | FLI                                 |   1
 Commodore 64 | FUN, FP2      | Funpaint                            |   2
 Commodore 64 | FPR           | FLI Profi                           |   1
 Commodore 64 | G             | SEUCK font                          |   1
 Commodore 64 | GCD, MON      |                                     |   1
 Commodore 64 | GG            | Koala Painter (compressed)          |   1
 Commodore 64 | GUN, IFL      | Gunpaint                            |   2
 Commodore 64 | HBM, HIR, HPI, GIH, FGS | Hires-Bitmap              |   1
*Commodore 64 | HED           | Hi-Eddi                             |   1
 Commodore 64 | HFC           | Hires FLI by Crest                  |   1
 Commodore 64 | HIM           | Hires Manager by Cosmos             |   1
 Commodore 64 | HLF           | Hires Interlace by Feniks           |   2
 Commodore 64 | HPC           | Hi-Pic Creator                      |   1
 Commodore 64 | IHE           | Interlace Hires Editor              |   2
 Commodore 64 | IPH           | Interpaint                          |   1
 Commodore 64 | IPT           | Interpaint                          |   1
 Commodore 64 | ISM           | Image System                        |   1
 Commodore 64 | JJ            | Doodle! (compressed)                |   1
*Commodore 64 | GIG, KLA, KOA | Koala Painter                       |   1
 Commodore 64 | LP3           | Logo Painter 3/3+                   |   1
 Commodore 64 | MCI           | True Paint                          |   2
 Commodore 64 | MIL           | Micro Illustrator                   |   1
 Commodore 64 | MLE           | Multi-Lace Editor                   |   2
 Commodore 64 | OCP           | Advanced Art Studio                 |   1
*Commodore 64 | P64           | Picasso 64                          |   1
 Commodore 64 | PET           | PETSCII Editor                      |   1
*Commodore 64 | PI            | Blazing Paddles                     |   1
 Commodore 64 | PMG           | Paint Magic                         |   1
 Commodore 64 | PP            | Pixel Perfect                       |   2
 Commodore 64 | PPP           | Pixel Perfect (compressed)          |   2
 Commodore 64 | RP            | Rainbow Painter                     |   1
 Commodore 64 | RPM           | Run Paint                           |   1
*Commodore 64 | SAR           | Saracen Paint                       |   1
 Commodore 64 | SCR+COL       | PETSCII Editor                      |   1
 Commodore 64 | SPD           | SpritePad                           |   1
 Commodore 64 | VIC           |                                     |   1
*Commodore 64 | VID           | Vidcom 64Paint                      |   1
 Commodore 128| BM, VBM       | VDC BitMap                          |   1
 Commodore 128| BRUS, PICT    | BASIC 8                             |   1
 Commodore 128| IP            | I Paint                             |   1
 MSX	MIG	MIG	up to 512x424		1 or 2
 MSX	SC2, GRP	Screen 2	256x192	15 colors	1
 MSX	SC3	Screen 3	64x48	15 colors	1
 MSX2	FNT	Dynamic Publisher font	512x160	mono	1
 MSX2	GL5+PL5			16 colors	1
 MSX2	GL6+PL6			4 colors	1
 MSX2	GL7+PL7			16 colors	1
 MSX2	GL8			256 colors	1
 MSX2         | MAG, MAX      | Maki-chan Graphics                  |   1
 MSX2         | PCT           | Dynamic Publisher screen            |   1
 MSX2         | SC4           | Screen 4 (256x192 16 colors)        |   1
 MSX2         | SC5, GE5      | Screen 5 (256x212 16 colors)        |   1
 MSX2	SC5+S15	Screen 5 interlaced	256x424	16 colors	2
 MSX2	SC6	Screen 6	512x212	4 colors	1
 MSX2	SC6+S16	Screen 6 interlaced	512x424	4 colors	2
 MSX2	SC7, GE7	Screen 7	512x212	16 colors	1
 MSX2	SC7+S17	Screen 7 interlaced	512x424	16 colors	2
 MSX2	SC8, GE8, PIC	Screen 8	256x212	256 colors	1
 MSX2	SC8+S18	Screen 8 interlaced	256x424	256 colors	2
 MSX2	SR5+PL5	Graph Saurus	256x212	16 colors	1
 MSX2	SR6+PL6	Graph Saurus	512x212	4 colors	1
 MSX2	SR7+PL7	Graph Saurus	512x212	16 colors	1
 MSX2	SR8	Graph Saurus	256x212	256 colors	1
 MSX2	SRI+PL7	Graph Saurus	512x424	16 colors	2
 MSX2         | STP           | Dynamic Publisher stamp             |   1
 MSX2+ V9990 VDP | G9B        | GFX9k (32768 colors)                |   1
 MSX2+        | GLA+PLA       | 12515 colors                        |   1
 MSX2+        | GLC, GLS      | 19268 colors                        |   1
 MSX2+        | SCA           | Screen 10 (256x212 12515 colors)    |   1
 MSX2+	SCA+S1A	Screen 10 interlaced	256x424	12515 colors	2
 MSX2+	SCC, SRS, YJK	Screen 12	up to 256x212	19268 colors	1
 MSX2+	SCC+S1C	Screen 12 interlaced	up to 256x424	19268 colors	2
 NEC PC-88    | MAG           | Maki-chan Graphics (8 colors)       |   1
 NEC PC-98    | MAG, MKI      | Maki-chan Graphics (256 colors)     |   1
 NEC PC-98    | PI            | 256 colors                          |   1
 NEC PC-98    | Q4            | XLD4                                |   1
 Oric         | CHS           | 6x8 font                            |   1
 Oric         | HIR, HRS      | 240x200 8 colors                    |   1
 PC           | EPA           | Award BIOS logo                     |   1
 PC           | HS2           | Handy Scanner 2000 POSTERING        |   1
 PC           | FNT           | Image 72 font                       |   1
 PC           | MSP           | Microsoft Paint version 1 or 2 mono |   1
 PlayStation  | TIM           | TIM (32768 colors)                  |   1
 Psion Series3| PIC, ICN      |                                     |   1
 SAM Coupe    | SCS4, SS4     | Mode 4 256x192 16 colors            |   1
 Sharp X68000 | MAG           | Maki-chan Graphics (16 colors)      |   1
 Sharp X68000 | PIC           |                                     |   1
 Timex 2048   | HRG           | Hi-res Gigascreen                   |   2
 Timex 2048   | SCR           | Hi-color                            |   1
 Timex 2048   | SCR           | Hi-res                              |   1
 TRS-80       | HR            |                                     |   1
 TRS-80       | RLE           | CompuServe RLE                      |   1
 TRS-80       | SHR           | MagicDraw                           |   1
 TRS-80 CC    | CLP           |                                     |   1
 TRS-80 CC    | GRF, MAX, P41, PIX |                                |   1
 TRS-80 CC    | P11           | 128x96 4 colors                     |   1
 ZX81         | P             | Sinclair BASIC	32x24 characters	mono	1
 ZX81         | RAW           | ZXPaintyONE v2.0	32x24 characters	mono	1
 ZX81         | ZP1 (arbitrary)	ZXpaintyONE	32x24 characters	mono	1
 ZX Spectrum  | 3             | 256x192 8 colors                    |   3
 ZX Spectrum  | ATR           | Attributes	32x24 characters	15 colors	1
 ZX Spectrum  | BMC4          | Border Screen Multicolor 8x4	384x304	15 colors	1
 ZX Spectrum  | BSC           | Border Screen	384x304	15 colors	1
 ZX Spectrum  | CH$           | CHR$                                | 1 - 2
 ZX Spectrum  | CH4           | 4x8 font		mono	1
 ZX Spectrum  | CH6           | 6x8 font		mono	1
 ZX Spectrum  | CH8           | 8x8 font		mono	1
 ZX Spectrum  | HLR           | Attributes Gigascreen               |   2
 ZX Spectrum  | IFL           | Multicolor 8x2 256x192 15 colors    |   1
 ZX Spectrum  | IMG           | Gigascreen                          |   2
 ZX Spectrum  | MC            | Multicolor 8x1 256x192 15 colors    |   1
 ZX Spectrum  | MG1           | MultiArtist 8x1 256x192             |   2
 ZX Spectrum  | MG2           | MultiArtist 8x2 256x192             |   2
 ZX Spectrum  | MG4           | MultiArtist 8x4 256x192             |   2
 ZX Spectrum  | MG8           | MultiArtist 8x8 256x192             |   2
 ZX Spectrum  | MLT           | 256x192 15 colors                   |   1
 ZX Spectrum  | RGB           | 256x192 8 colors                    |   3
 ZX Spectrum  | SCR           | 256x192 15 colors                   |   1
 ZX Spectrum ULAplus |SCR     | ULAplus (256x192 64 colors)         |   1
 ZX Spectrum Evolution |SXG   | Speccy eXtended Graphics (256 c.)   |   1
 ZX Spectrum  | ZXP           | ZX-Paintbrush (256x192 15 colors)   |   1

*/

#ifndef ABS
# define ABS(a) ((a)<0?-(a):(a))
#endif
#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif

#define ATARI_ST_LOW_PIXEL_RATIO (5.0/6)
#define ATARI_ST_MED_PIXEL_RATIO (5.0/12)
#define ATARI_ST_HIGH_PIXEL_RATIO (5.0/6)
#if 0
#define C64_LOW_PIXEL_RATIO (5.0/3)
#define C64_HIGH_PIXEL_RATIO (5.0/6)
#endif
#define FALCON_PIXEL_RATIO (5.0/6)

typedef bool(*decode_func_t)(RECOIL *, uint8_t const *, int);

typedef struct{
    cairo_surface_t *surface;
}layer_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    decode_func_t decode_func;
    /* This is used for a hack to get individual "layers" out of recoil */
    int supports_layers;
    layer_t layer[MAX_LAYERS];
};


static const char **
_recoil_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-artist-64",
        "image/x-atari-cad",
        "image/x-atari-duo",
        "image/x-atari-rgb-intermediate",
        "image/x-blazing-paddles",
        "image/x-cheese",
        "image/x-draz-paint",
        "image/x-face-painter",
        "image/x-gem",
        "image/x-gfa-artist",
        "image/x-hi-eddi",
#ifdef ENABLE_ILBM
        "image/x-ilbm",
#endif
        "image/x-imagelab",
        "image/x-macpaint", "application/mac-paint",
        "image/x-multi-palette-picture",
        "image/x-neochrome","image/x-neo",
        "image/x-pablo-packed-picture",
#ifdef ENABLE_PC1
        "image/x-pc1",
        "image/x-pc2",
        "image/x-pc3",
#endif
        "image/x-picasso-64",
        "image/x-photochrome-screen",
        "image/x-prism-paint",
        "image/x-quantum-paint",
        "image/x-saracen-paint",
#ifdef ENABLE_SPECTRUM512
        "image/x-spectrum512-compressed",
        "image/x-spectrum512-smooshed",
        "image/x-spectrum512-uncompressed",
#endif
        "image/x-stad",
        "image/x-tobias-richter-fullscreen-slideshow",
        "image/x-tricolor-rgb",
        "image/x-vidcom-64",
        "image/x-xga",
        "image/x-zx-spectrum-standard-screen",
        NULL
    };
    return mime_types;
}


#include "recoil.c"


static bool
_decode_a64_wig(RECOIL *r, uint8_t const *data, int len)
{
    if(len != 10242)
        return false;
    return RECOIL_DecodeC64Multicolor(r, data, 2, 0x2002, 0x2402, 0x2801);
}


static bool
_decode_blazing_paddles(RECOIL *r, uint8_t const *data, int len)
{
    if(len != 10242)
        return false;
    return RECOIL_DecodeC64Multicolor(r, data, 2, 0x2002, 0x2402, 0x1f82);
}


static bool
_decode_che(RECOIL *r, uint8_t const *data, int len)
{
    if(len != 20482)
        return false;
    return RECOIL_DecodeC64Multicolor(r, data, 2, 0x4202, 0x4802, 0x4fff);
}


static bool
_decode_fcp_ftp(RECOIL *r, uint8_t const *data, int len)
{
    if(len != 10004)
        return false;
    return RECOIL_DecodeC64Multicolor(r, data, 2, 0x1f42, 0x232a, 0x2712);
}


#ifdef ENABLE_ILBM
static bool
_decode_ilbm(RECOIL *r, uint8_t const *data, int len)
{
    return RECOIL_DecodeIff(r, data, len, RECOILResolution_AMIGA1X1);
}
#endif


static bool
_decode_neo(RECOIL *r, uint8_t const *data, int len)
{
    return RECOIL_DecodeNeo(r, ".neo", data, len);
}


static bool
_decode_p64(RECOIL *r, uint8_t const *data, int len)
{
    if(len != 10050)
        return false;
    return RECOIL_DecodeC64Multicolor(r, data, 0x802, 0x402, 2, 0x801);
}


static bool
_decode_sar(RECOIL *r, uint8_t const *data, int len)
{
    if(len != 10219)
        return false;
    return RECOIL_DecodeC64Multicolor(r, data, 0x402, 2, 0x2402, 0x3f2);
}


static bool
_decode_zx_scr(RECOIL *r, uint8_t const *data, int len)
{
    int attroffs;
    switch(len){
    case 6144:
        attroffs = -1;
        break;
    case 6912:
    case 6913:
        attroffs = 0x1800;
        break;
    default:
        return false;
    }
    RECOIL_SetZx(r, RECOIL_GetResolutionSpectrum1x1(r), 1);
    RECOIL_DecodeZx(r, data, 0, attroffs, RECOIL_ZX_ATTRIBUTES8X8, 0);
    return true;
}


static abydos_plugin_handle_t *
_recoil_new(const char *mime_type,abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->supports_layers = 0;
    if(!strcasecmp(mime_type,"image/x-artist-64")){
        h->decode_func = _decode_a64_wig;
    }else if(!strcasecmp(mime_type,"image/x-atari-cad")){
        h->info->pixel_ratio = 2.0/3;
        h->decode_func = RECOIL_DecodeDrg;
    }else if(!strcasecmp(mime_type,"image/x-atari-duo")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->decode_func = RECOIL_DecodeDuo;
        h->supports_layers = 1;
    }else if(!strcasecmp(mime_type,"image/x-atari-rgb-intermediate")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->decode_func = RECOIL_DecodeStRgb;
    }else if(!strcasecmp(mime_type,"image/x-blazing-paddles")){
        h->decode_func = _decode_blazing_paddles;
    }else if(!strcasecmp(mime_type,"image/x-cheese")){
        h->decode_func = _decode_che;
    }else if(!strcasecmp(mime_type,"image/x-draz-paint")){
        h->decode_func = RECOIL_DecodeDrz;
    }else if(!strcasecmp(mime_type,"image/x-face-painter")){
        h->decode_func = _decode_fcp_ftp;
    }else if(!strcasecmp(mime_type,"image/x-gem")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->decode_func = RECOIL_DecodeStImg;
    }else if(!strcasecmp(mime_type,"image/x-gfa-artist")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->decode_func = RECOIL_DecodeGfaArtist;
    }else if(!strcasecmp(mime_type,"image/x-hi-eddi")){
        h->decode_func = RECOIL_DecodeIsh;
    }else
#ifdef ENABLE_ILBM
    if(!strcasecmp(mime_type,"image/x-ilbm")){
        h->decode_func = _decode_ilbm;
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-imagelab")){
        h->decode_func = RECOIL_DecodeBw;
    }else if(!strcasecmp(mime_type,"image/x-macpaint") ||
             !strcasecmp(mime_type,"application/mac-paint"))
    {
        h->decode_func = RECOIL_DecodeMac;
    }else if(!strcasecmp(mime_type,"image/x-multi-palette-picture")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->decode_func = RECOIL_DecodeMpp;
        /* not tested */
        h->supports_layers = 1;
    }else if(!strcasecmp(mime_type,"image/x-neo") ||
             !strcasecmp(mime_type,"image/x-neochrome"))
    {
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->decode_func = _decode_neo;
    }else if(!strcasecmp(mime_type,"image/x-pablo-packed-picture")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->decode_func = RECOIL_DecodeStPpp;
    }else
#ifdef ENABLE_PC1
    if(!strcasecmp(mime_type,"image/x-pc1")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->recoil_type = ".pc1";
    }else if(!strcasecmp(mime_type,"image/x-pc2")){
        h->info->pixel_ratio = ATARI_ST_MED_PIXEL_RATIO;
        h->recoil_type = ".pc2";
    }else if(!strcasecmp(mime_type,"image/x-pc3")){
        h->info->pixel_ratio = ATARI_ST_HIGH_PIXEL_RATIO;
        h->recoil_type = ".pc3";
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-picasso-64")){
        h->decode_func = _decode_p64;
    }else if(!strcasecmp(mime_type,"image/x-photochrome-screen")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->decode_func = RECOIL_DecodePcs;
    }else if(!strcasecmp(mime_type,"image/x-prism-paint")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->decode_func = RECOIL_DecodeFalconPnt;
    }else if(!strcasecmp(mime_type,"image/x-quantum-paint")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->decode_func = RECOIL_DecodePbx;
    }else if(!strcasecmp(mime_type,"image/x-saracen-paint")){
        h->decode_func = _decode_sar;
    }else
#ifdef ENABLE_SPECTRUM512
    if(!strcasecmp(mime_type,"image/x-spectrum512-compressed")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->decode_func = RECOIL_DecodeStSpc;
    }else if(!strcasecmp(mime_type,"image/x-spectrum512-smooshed")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->decode_func = RECOIL_DecodeSps;
    }else if(!strcasecmp(mime_type,"image/x-spectrum512-uncompressed")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->decode_func = RECOIL_DecodeSpu;
    }else
#endif
    if(!strcasecmp(mime_type,"image/x-stad")){
        h->info->pixel_ratio = ATARI_ST_HIGH_PIXEL_RATIO;
        h->decode_func = RECOIL_DecodePac;
    }else if(!strcasecmp(mime_type,"image/x-tobias-richter-fullscreen-slideshow")){
        h->info->pixel_ratio = ATARI_ST_LOW_PIXEL_RATIO;
        h->decode_func = RECOIL_DecodePci;
        h->supports_layers = 1;
    }else if(!strcasecmp(mime_type,"image/x-tricolor-rgb")){
        h->decode_func = RECOIL_Decode3;
        h->supports_layers = 1;
    }else if(!strcasecmp(mime_type,"image/x-vidcom-64")){
        h->decode_func = RECOIL_DecodeDol;
    }else if(!strcasecmp(mime_type,"image/x-xga")){
        h->info->pixel_ratio = FALCON_PIXEL_RATIO;
        h->decode_func = RECOIL_DecodeXga;
    }else if(!strcasecmp(mime_type,"image/x-zx-spectrum-standard-screen")){
        h->decode_func = _decode_zx_scr;
    }else{
        /* not reached */
        free(h);
        return NULL;
    }
    info->threadsafe = 1;
    h->layer[0].surface = NULL;
    return h;
}


static void
_recoil_free(abydos_plugin_handle_t *h){
    if(h->layer[0].surface){
        int i;
        for(i=0;i<h->info->layer_count;++i)
            cairo_surface_destroy(h->layer[i].surface);
    }
    free(h);
}


static cairo_surface_t *
_surface_from_image(abydos_plugin_handle_t *h, const int *src, int xscale, int yscale)
{
    int height = h->info->height;
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,h->info->width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - h->info->width;
    while(height-->0){
        int x;
        for(x=0;x<h->info->width;++x){
            *dst++ = *src;
            src += xscale;
        }
        src += h->info->width * (yscale-1);
        dst += rowskip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


#if 0
static cairo_surface_t *
_surface_from_image_with_palette(abydos_plugin_handle_t *h, const unsigned char *src, const uint32_t *palette)
{
    int height = h->info->height;
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,h->info->width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - h->info->width;
    while(height-->0){
        int x;
        for(x=0;x<h->info->width;++x){
            *dst++ = palette[*src];
            src += h->xscale;
        }
        src += h->info->width * (h->yscale-1);
        dst += rowskip;
    }
    return surface;
}
#endif


#if 0
static void
_sort_palette(uint32_t *dst, const int *reference)
{
    int i;
    for(i=0;i<16;++i){
        int best_index = 0;
        int best_val = INT_MAX;
        int j;
        for(j=0;j<16;++j){
            int val = ABS(reference[i] - RECOIL_C64_PALETTE[j]);
            if(val<best_val){
                best_index = j;
                best_val = val;
            }
        }
        DEBUG_p("mapping %02d -> %02d",i,best_index);
        dst[i] = c64_color[best_index];
    }
}
#endif


static int
_recoil_create_from_data(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    int res = -1;
    RECOIL *recoil = RECOIL_New();
    if(h->decode_func(recoil,(const unsigned char*)data,len)){
        int xscale;
        int yscale;
        int xppm;
        int yppm;
        int i;
        DEBUG_p("success");
        h->info->width = RECOIL_GetOriginalWidth(recoil);
        h->info->height = RECOIL_GetOriginalHeight(recoil);
        xscale = RECOIL_GetWidth(recoil) / h->info->width;
        yscale = RECOIL_GetHeight(recoil) / h->info->height;
        xppm = RECOIL_GetXPixelsPerMeter(recoil);
        yppm = RECOIL_GetYPixelsPerMeter(recoil);
        if(xppm > 0 && yppm > 0){
            h->info->pixel_ratio = (double)(yppm * xscale) / (xppm * yscale);
        }else{
            DEBUG_p("pixel ratio information not available");
        }
        DEBUG_p("pixel ratio %f", h->info->pixel_ratio);
        if(h->supports_layers)
            h->info->layer_count = MIN(RECOIL_GetFrames(recoil), MAX_LAYERS);
        DEBUG_p("layers: %d",h->info->layer_count);
        for(i=0;i<h->info->layer_count;++i)
            h->layer[i].surface = _surface_from_image(h,RECOIL_GetPixels(recoil)+i*h->info->width*h->info->height, xscale, yscale);
        res = 0;
    }else{
        DEBUG_p("failure");
    }
    RECOIL_Delete(recoil);
    return res;
}


static cairo_surface_t *
_recoil_get_image_surface(abydos_plugin_handle_t *h,
                          int page,
                          cairo_bool_t *visible_layers,
                          int prefered_variant,
                          double time,
                          int frame)
{
    cairo_surface_t *surface;
    cairo_t *cr;
    int visible_count = 0;
    int i;
    if(1==h->info->layer_count)
        return cairo_surface_reference(h->layer[0].surface);
    for(i=0;i<h->info->layer_count;++i)
        if(visible_layers[i])
            ++visible_count;
    DEBUG_p("number of visible layers: %d",visible_count);
    switch(visible_count){
    case 0:
        return cairo_image_surface_create(CAIRO_FORMAT_ARGB32,h->info->width,h->info->height);
    case 1:
        for(i=0;i<h->info->layer_count;++i)
            if(visible_layers[i])
                return cairo_surface_reference(h->layer[i].surface);
        /* not reached */
    default:
        /* blend two or more layers */
        surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,h->info->width,h->info->height);
        cr = cairo_create(surface);
        visible_count = 0;
        for(i=0;i<h->info->layer_count;++i){
            if(!visible_layers[i])
                continue;
            ++visible_count;
            cairo_set_source_surface(cr,h->layer[i].surface,0,0);
            cairo_paint_with_alpha(cr,1.0/visible_count);
        }
        cairo_destroy(cr);
        return surface;
    }
}

const abydos_plugin_interface_t *
Abydos0(){
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _recoil_get_mime_types,
        .new = _recoil_new,
        .free = _recoil_free,
        .create_from_data = _recoil_create_from_data,
        .get_image_surface = _recoil_get_image_surface
    };
    return &p;
}
