/*
 * Copyright (C) 2019 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <glib.h>
#include <libspectre/spectre.h>
#include <math.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */

//#define ENABLE_OFFSET_HACK


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    SpectreDocument *document;
};


static const char **
_spectre_get_mime_types()
{
    static const char *mime_types[] = {
        "application/postscript",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_spectre_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->threadsafe = 1;
    h->document = NULL;
    return h;
}


static void
_spectre_free(abydos_plugin_handle_t *h)
{
    if(h->document)
        spectre_document_free(h->document);
    free(h);
}


static int
_spectre_create_from_file(abydos_plugin_handle_t *h, const char *filename)
{
    h->document = spectre_document_new();
    spectre_document_load(h->document,filename);
    /* FIXME!! check for errors somehow */
    h->info->page_count = spectre_document_get_n_pages(h->document);
    return 0;
}


static void
_spectre_render(abydos_plugin_handle_t *h, cairo_t *cr, int pagenr, int *visible_layers, int variant, double time, int frame)
{
    SpectrePage *page = spectre_document_get_page(h->document, pagenr);
    int page_width;
    int page_height;
    double x1;
    double y1;
    double x2;
    double y2;
    double sx;
    double sy;
    unsigned char* data;
    int stride;

    spectre_page_get_size(page, &page_width, &page_height);
    DEBUG_p("dimensions: %dx%d", page_width, page_height);

    cairo_clip_extents(cr,&x1,&y1,&x2,&y2);
    DEBUG_p("cairo clip extents: (%f)x(%f) - (%f)x(%f)", x1, y1, x2, y2);
    if(x1>0){
        cairo_translate(cr,x1,0);
    }else{
        x1 = 0;
    }
    if(y1>0){
        cairo_translate(cr,0,y1);
    }else{
        y1 = 0;
    }
    x2 = MIN(x2, page_width);
    y2 = MIN(y2, page_height);

    {
        cairo_matrix_t matrix;
        cairo_get_matrix(cr,&matrix);
        sx = MAX(1,sqrt(matrix.xx*matrix.xx+matrix.yx*matrix.yx));
        sy = MAX(1,sqrt(matrix.xy*matrix.xy+matrix.yy*matrix.yy));
        DEBUG_p("scale: %fx%f", sx, sy);
        cairo_scale(cr, 1/sx, 1/sy);
    }

    DEBUG_p("area: (%f)x(%f)-(%f)x(%f)", x1, y1, x2, y2);

    {
        SpectreRenderContext *rc = spectre_render_context_new();
        int unscaled_width = ceil(x2-x1);
        int unscaled_height = ceil(y2-y1);
        int unscaled_yoffs = page_height - unscaled_height - y1;
        spectre_render_context_set_scale(rc, sx, sy);
        DEBUG_p("render region: %dx%d %dx%d", (int)x1, (int)y1, unscaled_width, unscaled_height);

#ifdef ENABLE_OFFSET_HACK
        spectre_page_render_slice(page,rc,0,0,unscaled_width+x1,unscaled_height+unscaled_yoffs,&data,&stride);
#else
        spectre_page_render_slice(page,rc,x1,unscaled_yoffs,unscaled_width,unscaled_height,&data,&stride);
#endif
        spectre_page_free(page);
        spectre_render_context_free(rc);
    }
    {
        int scaled_width = ceil((x2-x1)*sx);
        int scaled_height = ceil((y2-y1)*sy);
#ifdef ENABLE_OFFSET_HACK
        int scaled_x = x1*sx;
        int scaled_y = (h->info->height-scaled_height-y1)*sy;
#endif
        cairo_surface_t *surface;
        DEBUG_p("width: %d (%d), stride: %d",scaled_width,scaled_width*4,stride);
#ifdef ENABLE_OFFSET_HACK
        surface = cairo_image_surface_create_for_data(data+scaled_x*4,CAIRO_FORMAT_RGB24,scaled_width,scaled_height,stride);
#else
        surface = cairo_image_surface_create_for_data(data,CAIRO_FORMAT_RGB24,scaled_width,scaled_height,stride);
#endif
        cairo_set_source_surface(cr,surface,0,0);
        cairo_paint(cr);
        cairo_surface_destroy(surface);
    }
    free(data);
}


static void
_spectre_get_page_info(abydos_plugin_handle_t *h, int page, abydos_plugin_page_info_t *info)
{
    int i;
    for(i=0;i<h->info->page_count;++i){
        SpectrePage *page = spectre_document_get_page(h->document, i);
        spectre_page_get_size(page, &info->width, &info->height);
        spectre_page_free(page);
    }
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _spectre_get_mime_types,
        .new = _spectre_new,
        .free = _spectre_free,
        .create_from_file = _spectre_create_from_file,
        .render = _spectre_render,
        .get_page_info = _spectre_get_page_info
    };
    return &p;
}
