/*
 * Copyright (C) 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-cairo.h>
#include <stdlib.h>
#include <string.h>

#define STB_IMAGE_IMPLEMENTATION
#define STBI_NO_STDIO
#include "stb_image.h"


typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    enum{
        T_8,
        T_16,
        T_FLOAT,
        T_GIF
    }type;
    frame_t *frame;
};


static const char **
_stb_get_mime_types()
{
    static const char *mime_types[] = {
        "image/bmp",
        "image/gif",
        "image/jpeg",
        "image/png",
        "image/vnd.adobe.photoshop",
        "image/vnd.radiance",
        "image/x-portable-anymap",
        "image/x-softimage-pic",
        "image/x-tga",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_stb_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    if(!strcasecmp(mime_type, "image/bmp")){
        h->type = T_16;
    }else if(!strcasecmp(mime_type, "image/gif")){
        h->type = T_GIF;
    }else if(!strcasecmp(mime_type, "image/png")){
        h->type = T_16;
    }else if(!strcasecmp(mime_type, "image/vnd.adobe.photoshop")){
        h->type = T_FLOAT;
    }else if(!strcasecmp(mime_type, "image/vnd.radiance")){
        h->type = T_FLOAT;
    }else{
        h->type = T_8;
    }
    h->frame = NULL;

    return h;
}


static void
_stb_free(abydos_plugin_handle_t *h)
{
    if(h->frame){
        int i;
        for(i=0;i<h->info->frame_count;++i)
            cairo_surface_destroy(h->frame[i].surface);
        free(h->frame);
    }
    free(h);
}


static int
_stb_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    uint8_t *img = NULL;
    int *delays = NULL;
    int channels;
    nile_format_t format = NILE_FORMAT_UNKNOWN;
    int stride;
    int bpc;
    int framesize;

    switch(h->type){
    case T_8:
        img = stbi_load_from_memory((stbi_uc*)data, len, &h->info->width, &h->info->height, &channels, 4);
        if(!img)
            return -1;
        switch(channels){
        case 3:
            format = NILE_FORMAT_r8g8b8;
            break;
        case 4:
            format = NILE_FORMAT_r8g8b8a8;
            break;
        }
        bpc = 4;
        stride = h->info->width * bpc;
        break;
    case T_16:
        img = (uint8_t*)stbi_load_16_from_memory((stbi_uc*)data, len, &h->info->width, &h->info->height, &channels, 4);
        if(!img)
            return -1;
        switch(channels){
        case 3:
            format = NILE_FORMAT_r16g16b16;
            break;
        case 4:
            format = NILE_FORMAT_r16g16b16a16;
            break;
        }
        bpc = 8;
        stride = h->info->width * bpc;
        break;
    case T_FLOAT:
        img = (uint8_t*)stbi_loadf_from_memory((stbi_uc*)data, len, &h->info->width, &h->info->height, &channels, 4);
        if(!img)
            return -1;
        switch(channels){
        case 3:
            format = NILE_FORMAT_r32g32b32;
            break;
        case 4:
            format = NILE_FORMAT_r32g32b32a32;
            break;
        }
        bpc = 16;
        stride = h->info->width * bpc;
        break;
    case T_GIF:
        img = stbi_load_gif_from_memory((stbi_uc*)data, len, &delays, &h->info->width, &h->info->height, &h->info->frame_count, &channels, 4);
        if(!img)
            return -1;

        switch(channels){
        case 3:
            format = NILE_FORMAT_r8g8b8;
            break;
        case 4:
            format = NILE_FORMAT_r8g8b8a8;
            break;
        }
        bpc = 4;
        stride = h->info->width * bpc;
        framesize = stride * h->info->height;
        break;
    }
    if(img){
        if(format){
            int i;
            DEBUG_p("bytes per channel: %d", NILE_GET_BYTES_PER_CHANNEL(format));
            h->frame = malloc(h->info->frame_count * sizeof(frame_t));
            for(i=0;i<h->info->frame_count;++i)
                h->frame[i].surface = nil_cairo_surface_create_from_direct(img + i * framesize,
                                                                           h->info->width, h->info->height,
                                                                           stride, bpc, format, 0);
            if(delays){
                for(i=0;i<h->info->frame_count;++i)
                    h->frame[i].duration = (double)delays[i] / 100;
                free(delays);
            }
        }
        free(img);
    }
    return h->frame ? 0 : -1;
}


static cairo_surface_t *
_stb_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_stb_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _stb_get_mime_types,
        .new = _stb_new,
        .free = _stb_free,
        .create_from_data = _stb_create_from_data,
        .get_image_surface = _stb_get_image_surface,
        .get_duration = _stb_get_duration
    };
    return &p;
}
