/*
 * Copyright (C) 2018 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <librsvg/rsvg.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    RsvgHandle *handle;
    char *error_msg;
    abydos_plugin_info_func_t info_func;
    abydos_plugin_progress_func_t progress_func;
    void *userdata;
};


static const char **
_svg_get_mime_types()
{
    static const char *mime_types[] = {
        "image/svg+xml",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_svg_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    info->scalable_size = 1;
    info->threadsafe = 1;
    h->info = info;
    h->handle = NULL;
    h->error_msg = NULL;
    return h;
}


static void
_svg_free(abydos_plugin_handle_t *h)
{
    if(h->handle)
        g_object_unref(h->handle);
    if(h->error_msg)
        free(h->error_msg);
    free(h);
}



static void
_svg_progressive_begin(abydos_plugin_handle_t *h,
                       abydos_plugin_info_func_t info_func,
                       abydos_plugin_progress_func_t progress_func,
                       void *userdata)
{
    h->handle = rsvg_handle_new_with_flags(RSVG_HANDLE_FLAG_KEEP_IMAGE_DATA);
    h->info_func = info_func;
    h->progress_func = progress_func;
    h->userdata = userdata;
}


static int
_svg_progressive_feed(abydos_plugin_handle_t *h,const char* data, size_t len)
{
    GError *err = NULL;
    if(!rsvg_handle_write(h->handle, (const unsigned char*)data, len, &err)){
        h->info->error = h->error_msg = strdup(err->message);
        g_error_free(err);
        return -1;
    }
    return 0;
}


static int
_svg_progressive_end(abydos_plugin_handle_t *h)
{
    RsvgDimensionData dim;
    GError *err = NULL;
    if(!rsvg_handle_close(h->handle, &err)){
        h->info->error = h->error_msg = strdup(err->message);
        g_error_free(err);
        return -1;
    }
    rsvg_handle_get_dimensions(h->handle,&dim);
    h->info->width = dim.width;
    h->info->height = dim.height;
    if(h->info_func)
        h->info_func(h->userdata);
    if(h->progress_func){
        cairo_rectangle_int_t rect;
        rect.x = 0;
        rect.y = 0;
        rect.width = dim.width;
        rect.height = dim.height;
        h->progress_func(h->userdata,&rect);
    }
    return 0;
}



static void
_svg_render(abydos_plugin_handle_t *h, cairo_t *cr,int page,cairo_bool_t *visible_layers,int prefered_variant,double time,int frame)
{
    rsvg_handle_render_cairo(h->handle, cr);
}



const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _svg_get_mime_types,
        .new = _svg_new,
        .free = _svg_free,
        .progressive_begin = _svg_progressive_begin,
        .progressive_feed = _svg_progressive_feed,
        .progressive_end = _svg_progressive_end,
        .render = _svg_render
    };
    return &p;
}
