/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <debug.h>

#include "teletext_cons.h"
#include "teletext_dec.h"


static
#include "teletext-12x10.c"


const teletext_cairo_color_t abydos_teletext_palette[8] = {
#ifdef TT_CAIRO_BITS16
    0x0000,     /* black */
    0xf800,     /* red */
    0x07e0,     /* green */
    0xffe0,     /* yellow */
    0x001f,     /* blue */
    0xf81f,     /* magenta */
    0x07ff,     /* cyan */
    0xffff      /* white */
#else
    0x00000000,     /* black */
    0x00ff0000,     /* red */
    0x0000ff00,     /* green */
    0x00ffff00,     /* yellow */
    0x000000ff,     /* blue */
    0x00ff00ff,     /* magenta */
    0x0000ffff,     /* cyan */
    0x00ffffff      /* white */
#endif
};


static void
_draw_rect(teletext_cairo_color_t *dst,int stride,int width,int height,teletext_cairo_color_t col)
{
    while(height-->0){
        int x;
        for(x=0;x<width;++x)
            dst[x] = col;
        dst += stride;
    }
}


static void
_draw_graphic(teletext_cairo_color_t *dst,int stride,int ch,teletext_cairo_color_t bg,teletext_cairo_color_t fg)
{
    _draw_rect(dst           ,stride,6,3,ch&0x01?fg:bg);
    _draw_rect(dst+6         ,stride,6,3,ch&0x02?fg:bg);
    _draw_rect(dst  +stride*3,stride,6,4,ch&0x04?fg:bg);
    _draw_rect(dst+6+stride*3,stride,6,4,ch&0x08?fg:bg);
    _draw_rect(dst  +stride*7,stride,6,3,ch&0x10?fg:bg);
    _draw_rect(dst+6+stride*7,stride,6,3,ch&0x40?fg:bg);
}


static void
_draw_character(teletext_cairo_color_t *dst,int stride,int ch,teletext_cairo_color_t bg,teletext_cairo_color_t fg)
{
    uint16_t *fontchar = &teletext_12x10[(ch-' ')*TT_CHAR_HEIGHT];
    int x;
    int y;
    for(y=0;y<TT_CHAR_HEIGHT;++y){
        int pixels = *fontchar++;
        for(x=0;x<TT_CHAR_WIDTH;++x){
            pixels <<= 1;
            dst[x] = pixels & (1<<TT_CHAR_WIDTH) ? fg : bg;
        }
        dst += stride;
    }
}


cairo_bool_t
teletext_console_has_flash(Console *cons)
{
    int size = cons_buf_width(cons) * cons_buf_height(cons);
    const ConsoleCell *p = cons_buf_buffer(cons);
    while(size-->0){
        union{
            TeletextState state;
            int attr:CONSOLE_BITS_ATTR;
        }s;
        s.attr = p->attr;
        if(s.state.flash){
            DEBUG_p("has flash");
            return 1;
        }
        ++p;
    }
    DEBUG_p("no flash");
    return 0;
}


cairo_surface_t *
abydos_image_surface_create_from_teletext_console(Console *cons, TeletextConsoleState state)
{
    const ConsoleCell *src = cons_buf_buffer(cons);
    int width = cons_buf_width(cons);
    int height = cons_buf_height(cons);
#ifdef TT_CAIRO_BITS16
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB16_565,width*TT_CHAR_WIDTH,height*TT_CHAR_HEIGHT);
#else
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width*TT_CHAR_WIDTH,height*TT_CHAR_HEIGHT);
#endif
    teletext_cairo_color_t *dst = (teletext_cairo_color_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(teletext_cairo_color_t);
    int x;
    int y;
    for(y=0;y<height;++y){
        for(x=0;x<width;++x){
            union{
                TeletextState state;
                int attr:CONSOLE_BITS_ATTR;
            }s;
            ConsoleCell cell = *src++;
            teletext_cairo_color_t bg;
            teletext_cairo_color_t fg;
            s.attr = cell.attr;
            //DEBUG_p("attr: 0x%x, bg: %d, fg: %d",s.attr,s.state.bg,s.state.fg);
            bg = abydos_teletext_palette[s.state.bg];
            fg = abydos_teletext_palette[s.state.fg];
            if(S_ON == state || !s.state.flash){
                if(s.state.graphic && (cell.ch&0x60) != 0x40){
                    _draw_graphic(dst+x*TT_CHAR_WIDTH,stride,cell.ch,bg,fg);
                }else{
                    _draw_character(dst+x*TT_CHAR_WIDTH,stride,cell.ch,bg,fg);
                }
            }else{
                _draw_character(dst+x*TT_CHAR_WIDTH,stride,' ',bg,fg);
            }
        }
        dst += stride*TT_CHAR_HEIGHT;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}
