/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */
#include <tiffio.h>


typedef struct{
    cairo_surface_t *surface;
    int width;
    int height;
}page_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    page_t *page;
};


static const char **
_tiff_get_mime_types()
{
    static const char *mime_types[] = {
        "image/tiff",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_tiff_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    info->threadsafe = 1;
    h->page = NULL;
    return h;
}


static void
_tiff_free(abydos_plugin_handle_t *h)
{
    if(h->page){
        int i;
        for(i=0;i<h->info->page_count;++i)
            cairo_surface_destroy(h->page[i].surface);
        free(h->page);
    }
    free(h);
}


static cairo_surface_t *
_surface_from_rgba(uint32_t *src, int width, int height)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int x;
    int y;
    dst += stride * height;
    for(y=0;y<height;++y){
        dst -= stride;
        for(x=0;x<width;++x){
            uint32_t pix = *src++;
            dst[x] = (pix&0x000000ff)<<16|
                     (pix&0x00ff0000)>>16|
                     (pix&0xff00ff00);
        }
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_tiff_create_from_file(abydos_plugin_handle_t *h,const char *filename)
{
    TIFF *tiff = TIFFOpen(filename,"r");
    int i;
    if(!tiff)
        goto error;
    h->info->page_count = 1;
    while(TIFFReadDirectory(tiff))
        ++h->info->page_count;
    h->page = (page_t*)malloc(h->info->page_count * sizeof(page_t));
    for(i=0;i<h->info->page_count;++i){
        uint32_t width;
        uint32_t height;
        uint32_t *pixels;
        TIFFSetDirectory(tiff,i);
        TIFFGetField(tiff,TIFFTAG_IMAGEWIDTH,&width);
        TIFFGetField(tiff,TIFFTAG_IMAGELENGTH,&height);

        pixels = (uint32_t*)malloc(width*height*sizeof(uint32_t));
        /* FIXME!! check for errors */
        TIFFReadRGBAImage(tiff,width,height,pixels,0);
        h->page[i].surface = _surface_from_rgba(pixels,width,height);
        free(pixels);

        h->page[i].width = width;
        h->page[i].height = height;
        if(width > h->info->width)
            h->info->width = width;
        if(height > h->info->height)
            h->info->height = height;
    }
    TIFFClose(tiff);
    return 0;

error:
    h->info->error = "failed to decode the image";
    return -1;
}


static cairo_surface_t *
_tiff_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->page[page].surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _tiff_get_mime_types,
        .new = _tiff_new,
        .free = _tiff_free,
        .create_from_file = _tiff_create_from_file,
        .get_image_surface = _tiff_get_image_surface,
    };
    return &p;
}
