/*
 * Copyright (C) 2017  Teemu Suutari
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 */
#include <cstdlib>
#include <cstring>
#include <memory>

#include "buffer.h"

/* --- Buffer ----------------------------------------------------------- */

uint8_t &Buffer::operator[](size_t i)
{
    if (i>=size())
        throw OutOfBoundsError();
    return data()[i];
}

const uint8_t &Buffer::operator[](size_t i) const
{
    if (i>=size())
        throw OutOfBoundsError();
    return data()[i];
}

uint32_t Buffer::readBE32(size_t offset) const
{
    if (offset+4>size())
        throw OutOfBoundsError();
    const uint8_t *ptr=data()+offset;
    return (uint32_t(ptr[0])<<24)|(uint32_t(ptr[1])<<16)|(uint32_t(ptr[2])<<8)|uint32_t(ptr[3]);
}

uint16_t Buffer::readBE16(size_t offset) const
{
    if (offset+2>size())
        throw OutOfBoundsError();
    const uint8_t *ptr=data()+offset;
    return (uint16_t(ptr[0])<<8)|uint16_t(ptr[1]);
}

uint64_t Buffer::readLE64(size_t offset) const
{
    if (offset+8>size())
        throw OutOfBoundsError();
    const uint8_t *ptr=data()+offset;
    return (uint64_t(ptr[7])<<56)|(uint64_t(ptr[6])<<48)|(uint64_t(ptr[5])<<40)|(uint64_t(ptr[4])<<32)|
        (uint64_t(ptr[3])<<24)|(uint64_t(ptr[2])<<16)|(uint64_t(ptr[1])<<8)|uint64_t(ptr[0]);
}

uint32_t Buffer::readLE32(size_t offset) const
{
    if (offset+4>size())
        throw OutOfBoundsError();
    const uint8_t *ptr=data()+offset;
    return (uint32_t(ptr[3])<<24)|(uint32_t(ptr[2])<<16)|(uint32_t(ptr[1])<<8)|uint32_t(ptr[0]);
}

uint16_t Buffer::readLE16(size_t offset) const
{
    if (offset+2>size())
        throw OutOfBoundsError();
    const uint8_t *ptr=data()+offset;
    return (uint16_t(ptr[1])<<8)|uint16_t(ptr[0]);
}

uint8_t Buffer::read8(size_t offset) const
{
    if (offset>=size())
        throw OutOfBoundsError();
    const uint8_t *ptr=reinterpret_cast<const uint8_t*>(data())+offset;
    return ptr[0];
}

/* --- MemoryBuffer ----------------------------------------------------- */

MemoryBuffer::MemoryBuffer(size_t size) :
    Buffer(size),
    _owned(true),
    _data(reinterpret_cast<uint8_t*>(std::malloc(size)))
{
    if (!_data)
        throw OutOfMemoryError();
}

MemoryBuffer::MemoryBuffer(uint8_t *data, size_t size):
    Buffer(size),
    _owned(false),
    _data(data)
{}

MemoryBuffer::MemoryBuffer(const Buffer &src,size_t offset,size_t size) :
    MemoryBuffer(size)
{
    if(offset+size>src.size())
        throw InvalidOperationError();
    std::memcpy(_data,src.data()+offset,size);
}

MemoryBuffer::~MemoryBuffer()
{
    if(_owned)
        std::free(_data);
}

void MemoryBuffer::resize(size_t newSize)
{
    if(!isResizable())
        throw InvalidOperationError();
    _data=reinterpret_cast<uint8_t*>(std::realloc(_data,newSize));
    _size=newSize;
    if (!_data){
        _size=0;
        throw OutOfMemoryError();
    }
}

/* --- ConstStaticBuffer ------------------------------------------------ */

ConstStaticBuffer::ConstStaticBuffer(const uint8_t *data,size_t size) :
    Buffer(size),
    _data(data)
{}

ConstStaticBuffer::ConstStaticBuffer(const Buffer &src,size_t offset,size_t size) :
    ConstStaticBuffer(src.data()+offset,size)
{
    if(offset+size>src.size())
        throw InvalidOperationError();
}
