/*
 * Copyright (C) 2017  Teemu Suutari
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 */
#ifndef BUFFER_HPP
#define BUFFER_HPP

#include <cstddef>
#include <cstdint>
#include <exception>


class Buffer {
protected:
    Buffer(size_t size) noexcept : _size(size) {};

    size_t _size;

public:
    class Error : public std::exception {};
    class OutOfBoundsError : public Error {};
    class InvalidOperationError : public Error {};
    class OutOfMemoryError : public Error {};

    Buffer(const Buffer&)=delete;
    Buffer& operator=(const Buffer&)=delete;

    virtual ~Buffer(){};

    virtual const uint8_t *data() const noexcept=0;
    virtual uint8_t *data()=0;
    virtual size_t size() const noexcept {return _size;}

    template<typename T> const T *cast() const noexcept {return reinterpret_cast<const T*>(data());}
    template<typename T> T *cast() {return reinterpret_cast<T*>(data());}

    virtual bool isResizable() const noexcept=0;
    virtual void resize(size_t newSize){throw InvalidOperationError();};

    uint8_t &operator[](size_t i);
    const uint8_t &operator[](size_t i) const;

    uint32_t readBE32(size_t offset) const;
    uint16_t readBE16(size_t offset) const;

    uint64_t readLE64(size_t offset) const;
    uint32_t readLE32(size_t offset) const;
    uint16_t readLE16(size_t offset) const;

    uint8_t read8(size_t offset) const;
};

class MemoryBuffer : public Buffer {
public:
    MemoryBuffer(size_t size);
    MemoryBuffer(uint8_t *data, size_t size);
    MemoryBuffer(const Buffer &src,size_t offset,size_t size);
    virtual ~MemoryBuffer() override final;

    virtual const uint8_t *data() const noexcept override final {return _data;};
    virtual uint8_t *data() override final {return _data;};

    virtual bool isResizable() const noexcept override final {return _owned;};
    virtual void resize(size_t newSize) override final;

private:
    bool _owned;
    uint8_t* _data;
};

class ConstStaticBuffer : public Buffer {
public:
    ConstStaticBuffer(const ConstStaticBuffer&)=delete;
    ConstStaticBuffer& operator=(const ConstStaticBuffer&)=delete;

    ConstStaticBuffer(const uint8_t *data,size_t length);
    ConstStaticBuffer(const Buffer &src, size_t offset, size_t size);
    virtual ~ConstStaticBuffer(){};

    virtual const uint8_t *data() const noexcept override {return _data;};
    virtual uint8_t *data() override {throw InvalidOperationError();};

    virtual bool isResizable() const noexcept override {return false;};

private:
    const uint8_t *_data;
};

#endif
