/*
 * Copyright (C) 2017  Teemu Suutari
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 */
#include "dec_hfmn.h"
#include "dec_huffman.h"
#include "streami.h"
#include "streamo.h"


HFMNDecompressor::HFMNDecompressor(const Buffer &packedData) :
	XPKDecompressor(packedData)
{
	if (packedData.size()<4)
		throw Decompressor::InvalidFormatError();
	uint16_t tmp=packedData.readBE16(0);
	if (tmp&3) throw Decompressor::InvalidFormatError();	// header is being written in 4 byte chunks
	_headerSize=size_t(tmp&0x1ff);				// the top 7 bits are flags. No definition what they are and they are ignored in decoder...
	if (size_t(_headerSize)+4>packedData.size()) throw Decompressor::InvalidFormatError();
	tmp=packedData.readBE16(_headerSize+2);
	_rawSize=size_t(tmp);
	if (!_rawSize) throw Decompressor::InvalidFormatError();
	_headerSize+=4;
}

void HFMNDecompressor::decompressImpl(Buffer &rawData,const Buffer &previousData,bool verify)
{
	if (rawData.size()!=_rawSize) throw Decompressor::DecompressionError();
	ForwardInputStream inputStream(_packedData,2,_headerSize);
	MSBBitReader<ForwardInputStream> bitReader(inputStream);
	auto readBit=[&]()->uint32_t
	{
		return bitReader.readBits8(1);
	};

	ForwardOutputStream outputStream(rawData,0,rawData.size());

	HuffmanDecoder<uint32_t> decoder;
	uint32_t code=1;
	uint32_t codeBits=1;
	for (;;)
	{
		if (!readBit())
		{
			uint32_t lit=0;
			for (uint32_t i=0;i<8;i++) lit|=readBit()<<i;
			decoder.insert(HuffmanCode<uint32_t>{codeBits,code,lit});
			while (!(code&1) && codeBits)
			{
				codeBits--;
				code>>=1;
			}
			if (!codeBits) break;
			code--;
		} else {
			code=(code<<1)+1;
			codeBits++;
		}
	}
	inputStream=ForwardInputStream(_packedData,_headerSize,_packedData.size());
	bitReader.reset();

	while (!outputStream.eof())
		outputStream.writeByte(decoder.decode(readBit));
}
