/*
 * Copyright (C) 2019 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#ifndef _GNU_SOURCE
# define _GNU_SOURCE                    /* vasprintf */
#endif
#include <ctype.h>                      /* isspace() */
#include <dirent.h>
#include <dlfcn.h>                      /* dlopen() */
#include <errno.h>
#include <fcntl.h>                      /* open() */
#include <pthread.h>
#include <pwd.h>                        /* getpwuid() */
#include <stdarg.h>
#include <stdio.h>                      /* fopen() */
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>                   /* mmap() */
#include <sys/stat.h>                   /* mkdir() */
#include <unistd.h>                     /* close() */

#include "abydos-internal.h"
#include "abydos-plugin.h"
#include "debug.h"
#include "priv.h"

#define ALIAS_CONFIG "alias.conf"
#define PLUGINS_CACHE "plugins.cache"
#define PLUGINS_CONFIG "plugins.conf"


static HashMap *type_map;               /**< Maps MIME types to plugin. */
static HashMap *alias_map;              /**< Maps MIME type aliases to proper types. */


typedef struct{
    const char *mime_type;
    size_t mime_type_len;
    const char *filename;
    size_t filename_len;
    int priority;
} config_item_t;

typedef struct{
    time_t timestamp;
    int fd;
    const void *data;
    size_t data_len;
    HashMap *item_map;
} config_t;

typedef struct{
    plugin_t *plugin;
    List *mime_type_list;
    int confirmed;                      /* The plugin is present, not just in the cache. */
} plugin_info_t;

typedef struct{
    config_t config;
    time_t timestamp;
    HashMap *plugin_info_map;           /* Maps filenames to plugin info. */
    int changed;
} cache_t;


static int
_atoin(const char* s,size_t len)
{
    const char *end = s+len;
    int neg;
    int res;
    while(s<end && isspace(*s))
        ++s;
    if(s>=end)
        return 0;
    if('-'==*s){
        neg = 1;
        ++s;
    }else{
        neg = 0;
    }
    res = 0;
    while(s<end && *s>='0' && *s<='9'){
        res *= 10;
        res += *s-'0';
        ++s;
    }
    return neg ? -res : res;
}


static const char *
_strnrchr(const char *s,size_t len,int c)
{
    const char *p = s+len-1;
    for(;;--p)
        if(p<s){
            return NULL;
        }else if(*p==c){
            return p;
        }
}


static const char *
_strnchrnul(const char *s,size_t len,int c)
{
    const char *end = s+len;
    for(;s<end && *s!=c;++s);
    return s;
}


static char *
_strdupf(const char *format,...)
{
    char *s;
    va_list ap;
    va_start(ap,format);
    (void)(1+vasprintf(&s,format,ap));
    va_end(ap);
    return s;
}


static void
_trim(const char** r_str,size_t *r_len)
{
    const char *str = *r_str;
    size_t len = *r_len;
    while(len>0 && isspace(str[len-1]))
        --len;
    while(len>0 && isspace(str[0])){
        ++str;
        --len;
    }
    *r_str = str;
    *r_len = len;
}


static int
_split(const char *s,ssize_t len,const char *sep,const char **args,int max)
{
    int count = 0;
    const char *end = len>=0 ? s+len : strchr(s,'\0');
    while(s<end){
        args[2*count++] = s;
        if(count>=max){
            args[2*count-1] = end;
            break;
        }
        while(s<end && !strchr(sep,*s))
            ++s;
        args[2*count-1] = s;
        ++s;
    }
    return count;
}


static int
_mmap_open(const char *filename,size_t len,int *r_fd,const void **r_data)
{
    *r_fd = open(filename,O_RDONLY);
    if(*r_fd<0)
        return -1;
    *r_data = mmap(NULL,len,PROT_READ,MAP_SHARED|MAP_POPULATE,*r_fd,0);
    if(MAP_FAILED==*r_data){
        close(*r_fd);
        return -1;
    }
    return 0;
}


static void
_mmap_close(int fd,const void *data,size_t len)
{
    munmap((void*)data,len);
    close(fd);
}


static int
_mkdir_recursive(char *path)
{
    char *p = strrchr(path,'/');
    if(p){
        *p = '\0';
        if(_mkdir_recursive(path))
            return -1;
        DEBUG_p("creating: %s",path);
        if(*path && mkdir(path,0777) && errno != EEXIST)
            return -1;
        *p = '/';
    }
    return 0;
}


static const char *
_get_home_dir()
{
    const char *home_dir;
    home_dir = getenv("HOME");
    if(!home_dir)
        home_dir = getpwuid(getuid())->pw_dir;
    return home_dir;
}


static char *
_get_config_filename(const char *filename)
{
    const char *var;
    char *path;
    struct stat statbuf;

    var = getenv("XDG_CONFIG_HOME");
    if(var && *var){
        path = _strdupf("%s/" PACKAGE "/%s",var,filename);
    }else{
        path = _strdupf("%s/.config/" PACKAGE "/%s",_get_home_dir(),filename);
    }
    if(!stat(path,&statbuf))
        return path;
    free(path);

#if defined(SYSCONFDIR) && !defined(WIN32)
    /* try the install directory if specified */
    path = _strdupf(SYSCONFDIR "/" PACKAGE "/%s",filename);
    if(!stat(path,&statbuf))
        return path;
    free(path);
#endif

    var = getenv("XDG_CONFIG_DIRS");
    if(var){
        const char *end = strchr(var,'\0');
        while(var<=end){
            const char *p = _strnchrnul(var,end-var,':');
            size_t len = p-var;
            _trim(&var,&len);
            if(len>0){
                path = _strdupf("%.*s/" PACKAGE "/%s",(int)len,var,filename);
                if(!stat(path,&statbuf))
                    return path;
                free(path);
            }
            var = p+1;
        }
    }
    return NULL;
}


static char *
_get_cache_filename(const char *filename,int for_writing)
{
    const char *cache_dir = getenv("XDG_CACHE_HOME");
    char *path;
    struct stat info;
    if(cache_dir && *cache_dir){
        path = _strdupf("%s/" PACKAGE "/%s",cache_dir,filename);
    }else{
        path = _strdupf("%s/.cache/" PACKAGE "/%s",_get_home_dir(),filename);
    }
    if(for_writing)
        return path;
    if(!stat(path,&info))
        return path;
    free(path);
    return NULL;
}



static plugin_t *
_create_plugin(char *filename)
{
    plugin_t *plugin = (plugin_t*)malloc(sizeof(plugin_t));
    plugin->refcount = 1;
    plugin->filename = filename;
    plugin->handle = NULL;
    pthread_mutex_init(&plugin->lock,NULL);
    return plugin;
}


static plugin_info_t *
_create_plugin_info(char *filename)
{
    plugin_info_t *info = (plugin_info_t*)malloc(sizeof(plugin_info_t));
    info->plugin = _create_plugin(filename);
    info->mime_type_list = NULL;
    info->confirmed = 0;
    return info;
}


static void
_free_plugin_info(plugin_info_t *info)
{
    if(info->plugin)
        _abydos_unref_plugin(info->plugin);
    list_free_full(info->mime_type_list,free);
    free(info);
}


static int
_strn_equal(const char *str1,ssize_t len1,const char *str2,ssize_t len2)
{
    if(len1>=0 && len2>=2){
        if(len1 != len2)
            return 0;
        return !strncmp(str1,str2,len1);
    }else if(len1>=0){
        return !strncmp(str1,str2,len1);
    }else if(len2>=0){
        return !strncmp(str1,str2,len2);
    }else{
        return !strcmp(str1,str2);
    }
}


static int
_strncase_equal(const char *str1,ssize_t len1,const char *str2,ssize_t len2)
{
    if(len1>=0 && len2>=2){
        if(len1 != len2)
            return 0;
        return !strncasecmp(str1,str2,len1);
    }else if(len1>=0){
        return !strncasecmp(str1,str2,len1);
    }else if(len2>=0){
        return !strncasecmp(str1,str2,len2);
    }else{
        return !strcasecmp(str1,str2);
    }
}


static void
_plugin_basename(const char **r_str,ssize_t *r_len)
{
    const char *str = *r_str;
    const char *p;
    ssize_t len = *r_len;
    if(len<0)
        len = strlen(str);
    p = _strnrchr(str,len,'/');
    if(p){
        ++p;
        len -= p-str;
        str = p;
    }
    p = _strnrchr(str,len,'.');
    if(p)
        len = p-str;
    *r_str = str;
    *r_len = len;
}


static int
_plugin_equal(const char *str1,ssize_t len1,const char *str2,ssize_t len2)
{
    _plugin_basename(&str1,&len1);
    _plugin_basename(&str2,&len2);
    return _strn_equal(str1,len1,str2,len2);
}


static int
_config_item_equal(const void *a,const void *b)
{
    config_item_t *item1 = (config_item_t*)a;
    config_item_t *item2 = (config_item_t*)b;
    if(!_strncase_equal(item1->mime_type,item1->mime_type_len,item2->mime_type,item2->mime_type_len))
        return 0;
    return _plugin_equal(item1->filename,item1->filename_len,item2->filename,item2->filename_len);
}


static unsigned int
_strcase_hash_update(unsigned int h,const char *s,ssize_t len)
{
    const signed char *p = (const signed char*)s;
    const signed char *end = len>=0 ? p+len : (void*)-1;
    for(;*p && p<end;++p)
        h = (h<<5)-h+toupper(*p);
    return h;
}


static unsigned int
_config_item_hash(const void *a){
    config_item_t *item = (config_item_t*)a;
    unsigned int h = _strcase_hash_update(0,item->mime_type,item->mime_type_len);
    const char *str = item->filename;
    ssize_t len = item->filename_len;
    _plugin_basename(&str,&len);
    return str_hash_update(h,str,len);
}


static void
_config_init(config_t *c)
{
    char *path = _get_config_filename(PLUGINS_CONFIG);
    struct stat statbuf;
    const char *s;
    const char *end;
    if(!path)
        goto nope;
    if(stat(path,&statbuf))
        goto nope;
    if(_mmap_open(path,statbuf.st_size,&c->fd,&c->data))
        goto nope;
    free(path);
    c->data_len = statbuf.st_size;
    c->timestamp = statbuf.st_mtime;
    c->item_map = hash_map_new(_config_item_hash,_config_item_equal,NULL,free);
    end = c->data+c->data_len;
    for(s=c->data;s<end;){
        const char *p = _strnchrnul(s,end-s,'\n');
        size_t len = p-s;
        _trim(&s,&len);
        if(len>0 && *s!='#'){
            const char *args[6];
            if(_split(s,len,":",args,3)>=3){
                config_item_t *item = (config_item_t*)malloc(sizeof(config_item_t));
                len = args[1]-args[0];
                _trim(&args[0],&len);
                item->mime_type = args[0];
                item->mime_type_len = len;
                len = args[3]-args[2];
                _trim(&args[2],&len);
                item->filename = args[2];
                item->filename_len = len;
                len = args[5]-args[4];
                //_trim(&args[4],&len);
                item->priority = _atoin(args[4],len);
                /* remove previous entry, if there is one */
                hash_map_remove(c->item_map,item);
                hash_map_add(c->item_map,item,item);
            }
        }
        s = p+1;
    }
    return;
nope:
    if(path)
        free(path);
    c->timestamp = 0;
    c->item_map = NULL;
}


static void
_config_done(config_t *c)
{
    if(c->item_map){
        hash_map_free(c->item_map);
        _mmap_close(c->fd,c->data,c->data_len);
    }
}


static int
_config_priority(config_t *c, const char *mime_type, const char *filename)
{
    if(c->item_map){
        config_item_t search;
        config_item_t *result;
        search.mime_type = mime_type;
        search.mime_type_len = -1;
        search.filename = filename;
        search.filename_len = -1;
        result = (config_item_t*)hash_map_get(c->item_map,&search);
        if(result){
            DEBUG_p("found entry: %d",result->priority);
            return result->priority;
        }
    }
    return 1;
}


static int
_cache_init(cache_t *c)
{
    char *path;
    struct stat statbuf;
    int fd;
    const void *data;
    const char *s;
    const char *end;
    plugin_info_t *info = NULL;

    _config_init(&c->config);
    c->plugin_info_map = hash_map_new(str_hash,str_equal,NULL,(HashMapNotifyFunc)_free_plugin_info);
    c->changed = 0;
    path = _get_cache_filename(PLUGINS_CACHE,0);
    if(!path)
        goto nope;
    if(stat(path,&statbuf))
        goto nope;
    if(statbuf.st_mtime <= c->config.timestamp)
        goto nope;
    if(_mmap_open(path,statbuf.st_size,&fd,&data))
        goto nope;
    free(path);
    c->timestamp = statbuf.st_mtime;
    end = data+statbuf.st_size;
    for(s=data;s<end;){
        const char *p = _strnchrnul(s,end-s,'\n');
        size_t len = p-s;
        _trim(&s,&len);
        if(len>0 && *s!='#'){
            if(':'==*s){
                if(!info){
                    /* Found a type before any plugin, just ignore it */
                }else{
                    ++s;
                    --len;
                    info->mime_type_list = list_prepend(info->mime_type_list,strndup(s,len));
                }
            }else{
                info = _create_plugin_info(strndup(s,len));
                hash_map_add(c->plugin_info_map,info->plugin->filename,info);
            }
        }
        s = p+1;
    }
    return 0;
nope:
    if(path)
        free(path);
    c->timestamp = 0;
    return -1;
}


static int
_cache_add(const char *mime_type, plugin_info_t *info, FILE *f)
{
    List *iter;
    fprintf(f,"%s\n",info->plugin->filename);
    for(iter=info->mime_type_list;iter;iter=iter->next)
        fprintf(f,":%s\n",(const char*)iter->data);
    return 0;
}


static void
_cache_done(cache_t *c)
{
    if(c->changed){
        char *path;
        FILE* f;
        path = _get_cache_filename(PLUGINS_CACHE,1);
        if(_mkdir_recursive(path)){
            free(path);
            DEBUG_p("ensure_path() failed, but that's OK");
            goto error;
        }
        f = fopen(path,"w");
        free(path);
        if(!f){
            DEBUG_p("fopen() failed, but that's OK");
            goto error;
        }
        hash_map_foreach(c->plugin_info_map,(HashMapForeachFunc)_cache_add,f);
        fclose(f);
    }
error:
    hash_map_free(c->plugin_info_map);
    _config_done(&c->config);
}


static int
_register_plugin(const char *filename, plugin_info_t *info, cache_t *c)
{
    List *iter;
    if(!info->confirmed){
        /* The plugin was in the cache but wasn't found */
        c->changed = 1;
        return 0;
    }

    DEBUG_p("--> %s",info->plugin->filename);
    for(iter=info->mime_type_list;iter;iter=iter->next){
        const char *mime_type = iter->data;
        plugin_t *old_plugin;
        int old_prio;
        int new_prio;

        new_prio = _config_priority(&c->config, mime_type, filename);
        if(new_prio<=0){
            DEBUG_p("black listed entry for %s",mime_type);
            continue;
        }

        old_plugin = (plugin_t*)hash_map_get(type_map,mime_type);
        if(old_plugin){
            old_prio = _config_priority(&c->config, mime_type, old_plugin->filename);
            if(old_prio >= new_prio){
                if(old_prio == new_prio)
                    DEBUG_p("WARNING: ambiguous priority for %s",mime_type);
                DEBUG_p("already have plugin with equal or higher priority for %s",mime_type);
                continue;
            }

        }
        DEBUG_p("regestering for %s",mime_type);

        hash_map_add(type_map,strdup(mime_type),_abydos_ref_plugin(info->plugin));
    }
    return 0;
}


static int
_load_plugin(plugin_t *plugin)
{
    abydos_plugin_func_t func;
    DEBUG_p("loading %s",plugin->filename);
    /* From python 3.8 the math module isn't linked to libpython, which
     * makes the PIL plugin require RTLF_GLOBAL to work.
     */
    plugin->handle = dlopen(plugin->filename,RTLD_NOW|RTLD_GLOBAL);
    if(!plugin->handle){
        DEBUG_p("dlopen failed: %s",dlerror());
        return -1;
    }
    func = dlsym(plugin->handle,"Abydos0");
    if(!func){
        dlclose(plugin->handle);
        plugin->handle = NULL;
        return -1;
    }
    plugin->vt = func();
    /* check if all required functions are implemented */
    if(  !plugin->vt->get_mime_types ||
         !plugin->vt->free ||
       ((!plugin->vt->progressive_begin ||
         !plugin->vt->progressive_feed ||
         !plugin->vt->progressive_end) &&
         !plugin->vt->create_from_data &&
         !plugin->vt->create_from_file) ||
        (!plugin->vt->render &&
         !plugin->vt->get_image_surface))
    {
        DEBUG_p("not all required functions are provided");
        dlclose(plugin->handle);
        plugin->handle = NULL;
        return -1;
    }
    return 0;
}


static void
_setup_aliases()
{
    char *path = _get_config_filename(ALIAS_CONFIG);
    struct stat statbuf;
    int fd;
    const void *data;
    const char *s;
    const char *end;
    if(!path)
        goto nope;
    if(stat(path,&statbuf))
        goto nope;
    if(_mmap_open(path,statbuf.st_size,&fd,&data))
        goto nope;
    free(path);
    end = data+statbuf.st_size;
    for(s=data;s<end;){
        const char *p = _strnchrnul(s,end-s,'\n');
        size_t len = p-s;
        _trim(&s,&len);
        if(len>0 && *s!='#'){
            const char *args[4];
            /* FIXME!! allow multiple aliases per line */
            if(_split(s,len,":",args,2)>=2){
                char *tmp;
                const char *proper_type;
                len = args[1]-args[0];
                _trim(&args[0],&len);
                tmp = strndup(args[0],len);
                proper_type = hash_map_get_key(type_map,tmp);
                free(tmp);
                if(proper_type){
                    char *alias_type;
                    len = args[3]-args[2];
                    _trim(&args[2],&len);
                    alias_type = strndup(args[2],len);
                    DEBUG_p("'%s' is an alias of '%s'",alias_type,proper_type);
                    hash_map_add(alias_map,alias_type,(void*)proper_type);
                }else{
                    DEBUG_p("no plugin handles %s, no need to register alias",proper_type);
                }
            }
        }
        s = p+1;
    }
    _mmap_close(fd,data,statbuf.st_size);
    return;
nope:
    if(path)
        free(path);
}


static void
_abydos_load_plugins()
{
    static pthread_mutex_t init_lock = PTHREAD_MUTEX_INITIALIZER;
    cache_t cache;
    DIR *d;
    struct dirent *e;
    size_t len1;

    pthread_mutex_lock(&init_lock);
    if(type_map){
        pthread_mutex_unlock(&init_lock);
        return;
    }

    type_map = hash_map_new(strcase_hash,strcase_equal,free,(HashMapNotifyFunc)_abydos_unref_plugin);
    alias_map = hash_map_new(strcase_hash,strcase_equal,free,NULL);

    _cache_init(&cache);

    DEBUG_p("searching for plugins in %s",PLUGINDIR);
    d = opendir(PLUGINDIR);
    if(!d)
        return;
    len1 = strlen(PLUGINDIR);
    while ((e = readdir(d))) {
        size_t len2;
        char *path;
        struct stat statbuf;
        plugin_info_t *info;
        if('.' == e->d_name[0])
            continue;
        len2 = strlen(e->d_name);
        path = (char*)malloc(len1 + 1 + len2 + 1);
        memcpy(path,PLUGINDIR,len1);
        path[len1] = '/';
        memcpy(path+len1+1,e->d_name,len2);
        path[len1+1+len2] = '\0';

        /* if plugin date > cache date */
        if(stat(path,&statbuf)){
            DEBUG_p("failed to load %s",path);
            free(path);
            continue;
        }

        info = hash_map_get(cache.plugin_info_map,path);
        DEBUG_p("modified: %ld (current: %ld), newer: %s",statbuf.st_mtime,cache.timestamp,statbuf.st_mtime>cache.timestamp?"yes":"no");
        if(info && statbuf.st_mtime < cache.timestamp){
            info->confirmed = 1;
            free(path);
        }else{
            if(info){
                /* drop old MIME types, in case the plugin now has a different set */
                list_free_full(info->mime_type_list,free);
                info->mime_type_list = NULL;
            }else{
                info = _create_plugin_info(path);
                hash_map_add(cache.plugin_info_map,info->plugin->filename,info);
            }
            info->confirmed = 1;
            if(_load_plugin(info->plugin)<0){
                /* Cache defunct plugins as plugins without any MIME types */
            }else{
                const char **mime_types = info->plugin->vt->get_mime_types();
                int i;
                for(i=0;mime_types[i];++i)
                    info->mime_type_list = list_prepend(info->mime_type_list,strdup(mime_types[i]));
            }
            cache.changed = 1;
        }
    }
    closedir(d);

    hash_map_foreach(cache.plugin_info_map,(HashMapForeachFunc)_register_plugin,&cache);

    _cache_done(&cache);

    _setup_aliases();

    pthread_mutex_unlock(&init_lock);
}


const char *
_abydos_unalias(const char *mime_type)
{
    const char *proper_type;
    if(!alias_map)
        _abydos_load_plugins();
    proper_type = hash_map_get(alias_map, mime_type);
    return proper_type ? proper_type : mime_type;
}


plugin_t *
_abydos_get_plugin(const char *mime_type)
{
    plugin_t *plugin;
    if(!type_map)
        _abydos_load_plugins();

    plugin = (plugin_t*)hash_map_get(type_map,mime_type);
    if(!plugin)
        return NULL;
    if(!plugin->handle)
        if(_load_plugin(plugin))
            return NULL;
    return _abydos_ref_plugin(plugin);
}


plugin_t *
_abydos_ref_plugin(plugin_t *p)
{
    ++p->refcount;
    //DEBUG_p("%s refcount: %d",p->filename,p->refcount);
    return p;
}


void
_abydos_unref_plugin(plugin_t *p)
{
    /* Static plugins always have a refcount of 0. */
    if(p->refcount <= 0)
        return;
    //DEBUG_p("%s refcount: %d",p->filename,p->refcount-1);
    if(--p->refcount > 0)
        return;
    free(p->filename);
    if(p->handle){
        if(p->vt->done)
            p->vt->done();
        dlclose(p->handle);
    }
    free(p);
}


typedef struct{
    const char **arr;
    int index;
}mime_types_data_t;

static int
_mime_types_cb1(void *key,void *value,void *userdata)
{
    mime_types_data_t *d = (mime_types_data_t*)userdata;
    d->arr[d->index++] = key;
    return 0;
}

static int
_mime_types_cb2(void *key,void *value,void *userdata)
{
    /* skip MIME types if they were already found in the type map */
    if(hash_map_get(type_map,key))
        return 0;
    return _mime_types_cb1(key, value, userdata);
}

static int
_mime_types_cb3(const void *p1, const void *p2)
{
    return strcasecmp(*(char**)p1,*(char**)p2);
}

/**
 * abydos_mime_types
 *
 * Returns an array with all MIME types supported with the currently
 * installed plugins.
 *
 * returns: An array of strings. You should free the array but not the
 *          strings themselves.
 */
const char **
abydos_mime_types()
{
    mime_types_data_t d;
    if(!type_map)
        _abydos_load_plugins();
    d.arr = (const char**)malloc(sizeof(const char*)*(hash_map_size(type_map)+hash_map_size(alias_map)+1));
    d.index = 0;
    hash_map_foreach(type_map,_mime_types_cb1,&d);
    hash_map_foreach(alias_map,_mime_types_cb2,&d);
    d.arr[d.index] = NULL;
    qsort(d.arr,d.index,sizeof(const char*),_mime_types_cb3);
    return d.arr;
}


/**
 * abydos_unload_plugins
 *
 * Unloads all plugins and frees up all global memory used by abydos.
 * This isn't something you normally need to do. But It might be useful for
 * memory debugging. And it can also be used to force abydos to pick up
 * changes in the configuration file or plugins that have been updated or
 * added. Plugins that are in use will not be affected but kept around as
 * long as they remain in use. Keep in mind that this function isn't
 * completely thread safe. It must not be called at the same time as
 * abydos_create() or abydos_mime_types().
 */
void abydos_unload_plugins()
{
    if(type_map){
        hash_map_free(type_map);
        type_map = NULL;
    }
}
