/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <endian.h>
#include <stdio.h>
#include <webp/encode.h>
#include <webp/mux.h>

#include "abydos.h"
#include "debug.h"


/* Minumum duration of a subframe. */
#define MIN_DURATION (0.05)

typedef union{
    uint32_t bits;
    struct{
        uint8_t b;
        uint8_t g;
        uint8_t r;
        uint8_t a;
    };
}pixel_t;


static int
_encode_frame(void *src, int w, int h, int stride, WebPData *dst){
    uint32_t *p = src;
    int rowskip = stride/sizeof(uint32_t) - w;
    int y = h;
    DEBUG_p("row skip: %d",rowskip);
    /* Do byte swapping and unprescaling. */
    while(y-->0){
        int x = w;
        while(x-->0){
            pixel_t px;
            /* Byte swap on big endian. */
            px.bits = le32toh(*p);
            if(px.a>0 && px.a<255){
                px.r = 255*px.r/px.a;
                px.g = 255*px.g/px.a;
                px.b = 255*px.b/px.a;
            }
            *p++ = px.bits;
        }
        p += rowskip;
    }
    dst->size = WebPEncodeLosslessBGRA(src, w, h, stride, (uint8_t**)&dst->bytes);
    return 0;
}


static int
_encode_animation(abydos_t *ar, int w, int h, WebPData *dst){
    WebPMux* mux;
    WebPMuxError err;
    abydos_info_t info;
    cairo_surface_t *surface = NULL;
    void *src;
    int stride;
    cairo_t *cr = NULL;
    WebPMuxFrameInfo frame;
    int i;

    mux = WebPMuxNew();
    if(!mux)
        return -1;

    abydos_get_info(ar, &info);

    surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,w,h);
    src = cairo_image_surface_get_data(surface);
    stride = cairo_image_surface_get_stride(surface);
    cr = cairo_create(surface);
    cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);

    frame.id = WEBP_CHUNK_ANMF;
    frame.x_offset = frame.y_offset = 0;
    frame.dispose_method = WEBP_MUX_DISPOSE_BACKGROUND;
    frame.blend_method = WEBP_MUX_NO_BLEND;

    for(i=0;i<abydos_get_frame_count(ar);++i){
        double frame_duration;
        int subframe_count;
        double subframe_duration;
        int j;

        frame_duration = abydos_get_frame_duration(ar,i);
        if(info.features & ABYDOS_FEATURE_SCALABLE_TIME){
            subframe_count = frame_duration / MIN_DURATION;
            subframe_duration = frame_duration / subframe_count;
        }else{
            subframe_count = 1;
            subframe_duration = frame_duration;
        }

        abydos_set_frame(ar,i);
        for(j=0;j<subframe_count;++j){
            frame.duration = subframe_duration*1000;
            cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
            cairo_set_source_rgba(cr, 0, 0, 0, 0);
            cairo_paint(cr);
            cairo_set_operator(cr, CAIRO_OPERATOR_OVER);
            abydos_render(ar, cr);
            cairo_surface_flush(surface);
            DEBUG_p("encoding frame %d:%d (%dms)",i,j,frame.duration);
            _encode_frame(src, w, h, stride, &frame.bitstream);
            err = WebPMuxPushFrame(mux, &frame, 1);
            WebPDataClear(&frame.bitstream);
            if(err != WEBP_MUX_OK){
                DEBUG_p("could not add frame to animation");
                goto out;
            }
            abydos_add_time(ar, subframe_duration);
        }
    }
    {
        WebPMuxAnimParams anim = {0, 0};
        err = WebPMuxSetAnimationParams(mux, &anim);
        if (err != WEBP_MUX_OK) {
            DEBUG_p("could not set animation parameters");
            goto out;
        }
    }
    err = WebPMuxAssemble(mux, dst);

out:
    WebPMuxDelete(mux);
    if(cr)
        cairo_destroy(cr);
    if(surface)
        cairo_surface_destroy(surface);

    return WEBP_MUX_OK==err ? 0 : -1;
}


/**
 * abydos_write_to_webp
 * @ar: an #abydos_t
 * @filename: file to write to
 *
 * Creates a WEBP file with the current state of @ar. Since the WEBP format
 * supports animation (as separate frames) the animation will be perserved
 * but only one variant of the current page. Also any scalable element will
 * be converted to pixels. You may want to call abydos_set_page_default()
 * before calling this if you don't have a particular wish to save a certain
 * page. This function is only available if WEBP support was enabled at
 * compile time.
 */
cairo_status_t
abydos_write_to_webp(abydos_t *ar, const char *filename)
{
    abydos_info_t info;
    WebPData dst;
    int res;

    abydos_get_info(ar,&info);
    if(abydos_can_animate(ar)){
        res = _encode_animation(ar, info.width, info.height, &dst);
    }else{
        cairo_surface_t *surface = abydos_get_image_surface(ar);
        void *src = cairo_image_surface_get_data(surface);
        int stride = cairo_image_surface_get_stride(surface);
        res = _encode_frame(src,info.width,info.height,stride,&dst);
        cairo_surface_destroy(surface);
    }
    if(0==res){
        FILE *f = fopen(filename,"w");
        fwrite(dst.bytes, dst.size, 1, f);
        fclose(f);
    }
    WebPDataClear(&dst);
    return res;
}
