/*
 * Copyright (C) 2021 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <stddef.h>                     /* NULL */
#include <stdlib.h>

#include "nil-cairo.h"


cairo_format_t
nil_cairo_format_from_nile(nile_format_t format)
{
    if(NILE_FORMAT_r5g6b5 == format){
        return CAIRO_FORMAT_RGB16_565;
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
    }else if(NILE_GET_BYTES_PER_CHANNEL(format) > 1){
        if(NILE_HAS_ALPHA(format)){
            return CAIRO_FORMAT_RGBA128F;
        }else{
            return CAIRO_FORMAT_RGB96F;
        }
#endif
    }else if(NILE_HAS_ALPHA(format)){
        return CAIRO_FORMAT_ARGB32;
    }else{
        return CAIRO_FORMAT_RGB24;
    }
}


nile_format_t
nil_cairo_format_to_nile(cairo_format_t format)
{
    switch(format){
    case CAIRO_FORMAT_INVALID:
    case CAIRO_FORMAT_A8: /* Not supported */
    case CAIRO_FORMAT_A1: /* Not supported */
    case CAIRO_FORMAT_RGB30: /* Not supported (yet) */
        break;
    case CAIRO_FORMAT_ARGB32:
        return NIL_CAIRO_FORMAT_ARGB32;
    case CAIRO_FORMAT_RGB24:
        return NIL_CAIRO_FORMAT_RGB24;
    case CAIRO_FORMAT_RGB16_565:
        return NIL_CAIRO_FORMAT_RGB16_565;
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
    case CAIRO_FORMAT_RGB96F:
        return NIL_CAIRO_FORMAT_RGB96F;
    case CAIRO_FORMAT_RGBA128F:
        return NIL_CAIRO_FORMAT_RGBA128F;
#endif
    }
    return NILE_FORMAT_UNKNOWN;
}


int
nil_cairo_format_to_bpc(cairo_format_t format)
{
    switch(format){
    case CAIRO_FORMAT_INVALID:
    case CAIRO_FORMAT_A8: /* Not supported */
    case CAIRO_FORMAT_A1: /* Not supported */
    case CAIRO_FORMAT_RGB30: /* Not supported (yet) */
        break;
    case CAIRO_FORMAT_ARGB32:
    case CAIRO_FORMAT_RGB24:
        return 4;
    case CAIRO_FORMAT_RGB16_565:
        return 2;
#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1,17,2)
    case CAIRO_FORMAT_RGB96F:
        return 12;
    case CAIRO_FORMAT_RGBA128F:
        return 16;
#endif
    }
    return 0;
}



int
nil_cairo_convert_palette(nile_t *ni)
{
    /*
    int cairo_bpc;
    cairo_format_t cairo_format;
    nile_format_t nile_format;
    */
    nile_t tmp;
    if(!ni->palette)
        return -1;
    nile_init(&tmp);
    if(NILE_FORMAT_r5g6b5 == ni->format){
        tmp.bpc = 2;
        //cairo_format = CAIRO_FORMAT_RGB565;
        tmp.format = NIL_CAIRO_FORMAT_RGB16_565;
    }else if(NILE_HAS_ALPHA(ni->format)){
        tmp.bpc = 4;
        //cairo_format = CAIRO_FORMAT_ARGB32;
        tmp.format = NIL_CAIRO_FORMAT_ARGB32;
    }else{
        tmp.bpc = 4;
        //cairo_format = CAIRO_FORMAT_RGB24;
        tmp.format = NIL_CAIRO_FORMAT_RGB24;
    }
    nile_ensure_palette(&tmp, ni->colors, tmp.bpc);
    nile_copy_colors_rgb(tmp.palette, tmp.bpc, tmp.format, ni->palette, ni->bpc, ni->format, ni->colors);
    nile_move_palette(ni, &tmp);
    nile_done(&tmp);
    return 0;
}


cairo_surface_t *
nil_cairo_surface_create_for_format(nile_format_t format, int width, int height)
{
    return cairo_image_surface_create(nil_cairo_format_from_nile(format), width, height);
}


nile_t *
nil_cairo_from_surface(nile_t *ni, cairo_surface_t *surface)
{
    nile_drop_pixels(ni);
    ni->width = cairo_image_surface_get_width(surface);
    ni->height = cairo_image_surface_get_height(surface);
    ni->type = NILE_TYPE_DIRECT;
    ni->pixels = cairo_image_surface_get_data(surface);
    ni->stride = cairo_image_surface_get_stride(surface);
    ni->bpc = 4;
    ni->format = nil_cairo_format_to_nile(cairo_image_surface_get_format(surface));
    return ni;
}


cairo_surface_t *
nil_cairo_surface_create_from_bitplanes(uint8_t **src,
                                        int width,
                                        int height,
                                        int stride,
                                        int depth,
                                        const uint8_t *palette,
                                        int colors,
                                        int bpc,
                                        nile_format_t format,
                                        int flags)
{
    cairo_format_t cairo_format = nil_cairo_format_from_nile(format);
    cairo_surface_t *surface = cairo_image_surface_create(cairo_format, width, height);
    uint8_t *dst = cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface);
    if(stride < 0)
        stride = (width-1)/8+1;
    nile_blit_direct_from_bitplanes_rgb(dst, dst_stride,
                                        nil_cairo_format_to_bpc(cairo_format),
                                        nil_cairo_format_to_nile(cairo_format),
                                        src, stride, depth,
                                        palette, colors,
                                        bpc, format,
                                        width, height,
                                        flags);
    cairo_surface_mark_dirty(surface);
    return surface;
}


cairo_surface_t *
nil_cairo_surface_create_from_indexed(const uint8_t *src,
                                      int width,
                                      int height,
                                      int stride,
                                      const uint8_t *palette,
                                      int colors,
                                      int bpc,
                                      nile_format_t format,
                                      int flags)
{
    cairo_format_t cairo_format = nil_cairo_format_from_nile(format);
    cairo_surface_t *surface = cairo_image_surface_create(cairo_format, width, height);
    uint8_t *dst = cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface);
    if(stride < 0)
        stride = width;
    nile_blit_direct_from_indexed_rgb(dst, dst_stride,
                                      nil_cairo_format_to_bpc(cairo_format),
                                      nil_cairo_format_to_nile(cairo_format),
                                      src, stride,
                                      palette, colors,
                                      bpc, format,
                                      width, height,
                                      flags);
    cairo_surface_mark_dirty(surface);
    return surface;
}


cairo_surface_t *
nil_cairo_surface_create_from_direct(const uint8_t *src,int width,int height,int stride, int bpc, nile_format_t format, int flags)
{
    cairo_format_t cairo_format = nil_cairo_format_from_nile(format);
    cairo_surface_t *surface = cairo_image_surface_create(cairo_format, width, height);
    uint8_t *dst = cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface);
    if(stride < 0)
        stride = width*bpc;
    nile_blit_direct_rgb(dst, dst_stride,
                         nil_cairo_format_to_bpc(cairo_format),
                         nil_cairo_format_to_nile(cairo_format),
                         src, stride,
                         bpc, format,
                         width, height,
                         flags);
    cairo_surface_mark_dirty(surface);
    return surface;
}


cairo_surface_t *
nil_cairo_surface_create(nile_t *ni, int flags)
{
#if 0
    if(ni->bpc > 4)
        return NULL;
#endif
    switch(ni->type){
    case NILE_TYPE_NONE:
        return NULL;
    case NILE_TYPE_BITPLANES:
        return nil_cairo_surface_create_from_bitplanes(ni->bitplanes, ni->width, ni->height, nile_get_stride(ni),
                                                       ni->depth, ni->palette, ni->colors, ni->bpc, ni->format, flags);
    case NILE_TYPE_INDEXED:
        return nil_cairo_surface_create_from_indexed(ni->pixels, ni->width, ni->height, nile_get_stride(ni),
                                                     ni->palette, ni->colors, ni->bpc, ni->format, flags);
    case NILE_TYPE_DIRECT:
        return nil_cairo_surface_create_from_direct(ni->pixels, ni->width, ni->height, nile_get_stride(ni),
                                                    ni->bpc, ni->format, flags);
    }
    return NULL;
}
