/*
 * Copyright (C) 2023  Magnus Bergman <magnus@snisurset.net>
 * Copyright (C) 2022  Nigel Tao.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * QOIR is a fast, simple image file format.
 *
 * This file also contains a stand-alone implementation of LZ4 block
 * compression, a general format that is not limited to compressing images. The
 * qoir_lz4_block_decode and qoir_lz4_block_encode functions also read from and
 * write to a contiguous block of memory.
 */
#include <abydos-plugin.h>
#include <cairo.h>
#include <endian.h>
#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#if defined(QOIR_CONFIG__USE_OFFICIAL_LZ4_LIBRARY)
# include <limits.h>
# include <lz4.h>
#endif
#ifndef QOIR_CONFIG__DISABLE_SIMD
# if (defined(__GNUC__) && defined(__x86_64__)) || (defined(_MSC_VER) && defined(_M_X64))
#  define QOIR_USE_SIMD_SSE2
#  include <emmintrin.h>
# endif
#endif

//#define ENABLE_METADATA

#if __BYTE_ORDER == __BIG_ENDIAN
# define ID(a,b,c,d) ((a<<24)|(b<<16)|(c<<8)|(d<<0))
#else
# define ID(a,b,c,d) ((a<<0)|(b<<8)|(c<<16)|(d<<24))
#endif

#define ID_QOIR ID('Q','O','I','R')
#define ID_QEND ID('Q','E','N','D')
#define ID_QPIX ID('Q','P','I','X')
#define ID_CICP ID('C','I','C','P')
#define ID_ICCP ID('I','C','C','P')
#define ID_EXIF ID('E','X','I','F')
#define ID_XMP  ID('X','M','P',' ')

// -------- Compile-time Configuration

// The compile-time configuration macros are:
//  - QOIR_CONFIG__DISABLE_LARGE_LOOK_UP_TABLES
//  - QOIR_CONFIG__DISABLE_SIMD
//  - QOIR_CONFIG__STATIC_FUNCTIONS
//  - QOIR_CONFIG__USE_OFFICIAL_LZ4_LIBRARY

// ----

// If using e.g. "CFLAGS='-DQOIR_CONFIG__USE_OFFICIAL_LZ4_LIBRARY -O3'") then
// you probably also want "LDFLAGS=-llz4", otherwise you'll get "undefined
// reference to `LZ4_decompress_safe'".

// -------- Pixel Buffers

// A pixel format combines an alpha transparency choice, a color model choice
// and other configuration (such as pixel byte order).
//
// Values less than 0x10 are directly representable by the file format (and by
// this implementation's API), using the same bit pattern.
//
// Values greater than or equal to 0x10 are representable by the API but not by
// the file format:
//  - the 0x10 bit means 3 (not 4) bytes per (fully opaque) pixel.
//  - the 0x20 bit means RGBA (not BGRA) byte order.

typedef enum{
    PIXEL_ALPHA_TRANSPARENCY__OPAQUE                 = 0x01,
    PIXEL_ALPHA_TRANSPARENCY__NONPREMULTIPLIED_ALPHA = 0x02,
    PIXEL_ALPHA_TRANSPARENCY__PREMULTIPLIED_ALPHA    = 0x03
}pixel_alpha_t;

#define QOIR_PIXEL_FORMAT__MASK_FOR_ALPHA_TRANSPARENCY  0x03
#define QOIR_PIXEL_FORMAT__MASK_FOR_COLOR_MODEL         0x0C

typedef enum{
    PIXEL_FORMAT__INVALID        = 0x00,
    PIXEL_FORMAT__BGRX           = 0x01,
    PIXEL_FORMAT__BGRA_NONPREMUL = 0x02,
    PIXEL_FORMAT__BGRA_PREMUL    = 0x03,
    PIXEL_FORMAT__BGR            = 0x11,
    PIXEL_FORMAT__RGBX           = 0x21,
    PIXEL_FORMAT__RGBA_NONPREMUL = 0x22,
    PIXEL_FORMAT__RGBA_PREMUL    = 0x23,
    PIXEL_FORMAT__RGB            = 0x31
}pixel_format_t;

#if __BYTE_ORDER == __BIG_ENDIAN
# define PIXEL_FORMAT_CAIRO_ARGB32 PIXEL_FORMAT__ARGB_PREMUL
# define PIXEL_FORMAT_CAIRO_RGB24  PIXEL_FORMAT__XRGB
#else
# define PIXEL_FORMAT_CAIRO_ARGB32 PIXEL_FORMAT__BGRA_PREMUL
# define PIXEL_FORMAT_CAIRO_RGB24  PIXEL_FORMAT__BGRX
#endif

typedef struct qoir_pixel_buffer_struct {
    uint32_t width;
    uint32_t height;
    pixel_format_t format;
    uint8_t *data;
    size_t stride;
}pixel_buffer_t;

// All int32_t points (x, y) such that ((x0 <= x) && (x < x1) && (y0 <= y) &&
// (y < y1)). The low bounds are inclusive. The high bounds are exclusive.
typedef struct{
    int32_t x0;
    int32_t y0;
    int32_t x1;
    int32_t y1;
}rectangle_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static uint32_t
_pixel_format_get_bpc(pixel_format_t format)
{
    return (format & 0x10) ? 3 : 4;
}


static int
_pixel_format_has_alpha(pixel_format_t format)
{
    switch((pixel_alpha_t)(format & QOIR_PIXEL_FORMAT__MASK_FOR_ALPHA_TRANSPARENCY)){
    case PIXEL_ALPHA_TRANSPARENCY__NONPREMULTIPLIED_ALPHA:
    case PIXEL_ALPHA_TRANSPARENCY__PREMULTIPLIED_ALPHA:
        return 1;
    default:
        return 0;
    }
}


#if 0
static inline bool
pixel_buffer__is_zero(pixel_buffer_t pixbuf)
{
  return (pixbuf.format == 0) &&
         (pixbuf.width == 0) &&
         (pixbuf.height == 0) &&
         (pixbuf.data == NULL) &&
         (pixbuf.stride == 0);
}
#endif


static rectangle_t
make_rectangle(int32_t x0, int32_t y0, int32_t x1, int32_t y1)
{
    rectangle_t r;
    r.x0 = x0;
    r.y0 = y0;
    r.x1 = x1;
    r.y1 = y1;
    return r;
}


static rectangle_t
rectangle__intersect(rectangle_t r, rectangle_t s)
{
    rectangle_t ret;
    ret.x0 = (r.x0 > s.x0) ? r.x0 : s.x0;
    ret.y0 = (r.y0 > s.y0) ? r.y0 : s.y0;
    ret.x1 = (r.x1 < s.x1) ? r.x1 : s.x1;
    ret.y1 = (r.y1 < s.y1) ? r.y1 : s.y1;
    return ret;
}


static int
rectangle__is_empty(rectangle_t r)
{
    return (r.x1 <= r.x0) || (r.y1 <= r.y0);
}


static uint32_t
rectangle__width(rectangle_t r)
{
    return (r.x1 > r.x0) ? ((uint32_t)r.x1 - (uint32_t)r.x0) : 0;
}


static uint32_t
rectangle__height(rectangle_t r)
{
    return (r.y1 > r.y0) ? ((uint32_t)r.y1 - (uint32_t)r.y0) : 0;
}

/* --- Tiling --- */

#define QOIR_TILE_MASK 0x3F
#define QOIR_TILE_SIZE 0x40
#define QOIR_TILE_SHIFT 6

// QOIR_LITERALS_PRE_PADDING is large enough to hold the previous pixel, at 4
// bytes per pixel.
#define QOIR_LITERALS_PRE_PADDING 4

// QOIR_TS2 is the maximum (inclusive) number of pixels in a tile.
#define QOIR_TS2 (QOIR_TILE_SIZE * QOIR_TILE_SIZE)


static uint32_t
calculate_number_of_tiles_1d(uint32_t number_of_pixels)
{
  uint64_t rounded_up = (uint64_t)number_of_pixels + QOIR_TILE_MASK;
  return (uint32_t)(rounded_up >> QOIR_TILE_SHIFT);
}


#if 0
static uint64_t
calculate_number_of_tiles_2d(uint32_t width_in_pixels, uint32_t height_in_pixels)
{
  uint64_t w = ((uint64_t)width_in_pixels + QOIR_TILE_MASK) >> QOIR_TILE_SHIFT;
  uint64_t h = ((uint64_t)height_in_pixels + QOIR_TILE_MASK) >> QOIR_TILE_SHIFT;
  return w * h;
}
#endif

/* --- LZ4 Decode --- */

// QOIR_LZ4_BLOCK_DECODE_MAX_INCL_SRC_LEN is the maximum (inclusive) supported
// input length for this file's LZ4 decode functions. The LZ4 block format can
// generally support longer inputs, but this implementation specifically is
// more limited, to simplify overflow checking.
//
// With sufficiently large input, qoir_lz4_block_encode (note that that's
// encode, not decode) may very well produce output that is longer than this.
// That output is valid (in terms of the LZ4 file format) but isn't decodable
// by qoir_lz4_block_decode.
//
// 0x00FFFFFF = 16777215, which is over 16 million bytes.
#define QOIR_LZ4_BLOCK_DECODE_MAX_INCL_SRC_LEN 0x00FFFFFF

// qoir_lz4_block_decode writes to dst the LZ4 block decompressed form of src,
// returning the number of bytes written.
//
// It fails with qoir_lz4_status_message__error_dst_is_too_short if dst_len is
// not long enough to hold the decompressed form.
static ssize_t
qoir_lz4_block_decode(uint8_t *dst_ptr, size_t dst_len, const uint8_t *src_ptr, size_t src_len);

/* --- QOIR Decode --- */

typedef struct{
    struct {
        // ops has to be before literals, so that (worst case) we can read (and
        // ignore) 8 bytes past the end of the ops array. See §
        uint8_t ops[4 * QOIR_TS2];
        uint8_t literals[QOIR_LITERALS_PRE_PADDING + (4 * QOIR_TS2)];
    }private_impl;
}decode_buffer_t;

#if 0
typedef struct{
    const char *status_message;
    void *owned_memory;
    pixel_buffer_t dst_pixbuf;
}decode_result_t;
#endif

#if 0
typedef struct{
#if 0
    // Custom malloc/free implementations. NULL etc_func pointers means to use
    // the standard malloc and free functions. Non-NULL etc_func pointers will be
    // passed the memory_func_context.
    void* (*contextual_malloc_func)(void* memory_func_context, size_t len);
    void (*contextual_free_func)(void* memory_func_context, void* ptr);
    void* memory_func_context;
#endif

    // Pre-allocated 'scratch space' used during decoding. Its contents need not
    // be initialized when passed to qoir_decode.
    //
    // If NULL, the 'scratch space' will be dynamically allocated and freed.
    decode_buffer_t *decbuf;

    // Pre-allocated pixel buffer to decode into.
    //
    // If zero, it will be dynamically allocated and memory ownership (i.e. the
    // responsibility to call free or equivalent) will be returned as the
    // qoir_decode_result owned_memory field.
    pixel_buffer_t pixbuf;

    // If non-zero, this is the pixel format to use when dynamically allocating
    // the pixel buffer to decode into (if the pixbuf field is zero). This pixfmt
    // field has no effect if the pixbuf field is non-zero.
    //
    // A zero-valued pixfmt field is equivalent to the default pixel format:
    // QOIR_PIXEL_FORMAT__RGBA_NONPREMUL.
    pixel_format_t format;

    // Clipping rectangles, in the destination or source (or both) coordinate
    // spaces. The clips (the qoir_rectangle typed fields) have no effect unless
    // the corresponding boolean typed field is true.
    rectangle_t dst_clip_rectangle;
    rectangle_t src_clip_rectangle;
    bool use_dst_clip_rectangle;
    bool use_src_clip_rectangle;

    // The position (in destination coordinate space) to place the top-left
    // corner of the decoded source image. The Y axis grows down.
    int32_t offset_x;
    int32_t offset_y;
}decode_options_t;
#endif

// Decodes a pixel buffer from the QOIR format.
//
// A NULL options is valid and is equivalent to a non-NULL pointer to a
// zero-valued struct (where all fields are zero / NULL / false).
//static decode_result_t
//qoir_decode(const uint8_t* src, size_t len, const decode_options_t*);

// -------- Look-up Tables

#if 0
// qoir_private_table_noise is the "LDR_LLL1_42.png" 16x16 blue noise texture
// from the zip file linked to from http://momentsingraphics.de/BlueNoise.html
// and whose LICENSE.txt says "To the extent possible under law, Christoph
// Peters has waived all copyright and related or neighboring rights".
static uint8_t qoir_private_table_noise[16][16] = {
  {0xAE,0xE7,0x89,0xF2,0x26,0x1A,0x8C,0xF7,0xC1,0x69,0xA4,0x30,0x91,0xCC,0x64,0x52},
  {0x71,0x1C,0x5E,0x7B,0xA9,0xD5,0x62,0x08,0xCE,0x49,0x0E,0x55,0x7A,0x18,0xA6,0x05},
  {0xD8,0xBF,0x99,0x01,0x56,0xE4,0x72,0xB2,0x36,0x97,0xD9,0xED,0xBC,0x25,0xF8,0x96},
  {0x2B,0x3B,0xFE,0xB6,0x41,0x2F,0x9F,0xE9,0x1D,0x85,0xAC,0x6D,0x3E,0xD1,0x47,0x80},
  {0x66,0x4F,0x13,0xD2,0x83,0xC2,0x0F,0x51,0xFB,0x60,0x2A,0x00,0x9D,0x5B,0xB5,0xE2},
  {0xC8,0xAA,0x90,0xEB,0x6B,0x23,0x94,0x79,0x40,0xCA,0xDE,0x76,0xF5,0x16,0x8B,0x0B},
  {0xF3,0x20,0x77,0x33,0x5C,0xCB,0xE1,0xA8,0xBB,0x14,0x8E,0xB0,0x4D,0xC3,0x34,0x70},
  {0x58,0x45,0xDB,0xA5,0x0D,0x4A,0xF4,0x04,0x68,0x57,0x31,0xE5,0x1E,0x82,0xE8,0xA2},
  {0x95,0x03,0xBD,0xFA,0x8A,0xB3,0x28,0x3A,0x86,0xA1,0xD3,0x43,0x9A,0x63,0xCD,0x27},
  {0xD7,0x65,0x81,0x1B,0x54,0x6E,0x98,0xDA,0xC5,0xEE,0x0C,0x6F,0xFF,0x06,0x3D,0xB9},
  {0xF1,0xAD,0x2E,0x3F,0xCF,0xEC,0x7C,0x19,0x5F,0x22,0x7F,0xAF,0xBE,0x53,0x7D,0x17},
  {0x48,0x73,0xE6,0xC4,0xA0,0x07,0x46,0xB7,0xF9,0x4E,0x37,0x92,0x2C,0xDC,0xA7,0x8D},
  {0x5A,0x09,0x93,0x24,0x61,0xD6,0x32,0x8F,0x6C,0xA3,0xD0,0xEA,0x15,0x67,0xF6,0x35},
  {0xD4,0xB4,0xFC,0x50,0x84,0xAB,0x11,0xC6,0xE3,0x02,0x42,0x75,0xC7,0x9E,0x21,0xC0},
  {0x10,0x9B,0x6A,0x12,0xDF,0x74,0xF0,0x59,0x29,0xB1,0x88,0x5D,0x0A,0x4C,0x87,0x78},
  {0x2D,0x44,0xC9,0x38,0xBA,0x4B,0x9C,0x3C,0x7E,0x1F,0xE0,0xFD,0xB8,0x39,0xEF,0xDD},
};
#endif

// The table was generated by script/gen_table_unlossify.go
static uint8_t qoir_private_table_unlossify[7][256] = {{
  0x00,0x02,0x04,0x06,0x08,0x0A,0x0C,0x0E,0x10,0x12,0x14,0x16,0x18,0x1A,0x1C,0x1E,
  0x20,0x22,0x24,0x26,0x28,0x2A,0x2C,0x2E,0x30,0x32,0x34,0x36,0x38,0x3A,0x3C,0x3E,
  0x40,0x42,0x44,0x46,0x48,0x4A,0x4C,0x4E,0x50,0x52,0x54,0x56,0x58,0x5A,0x5C,0x5E,
  0x60,0x62,0x64,0x66,0x68,0x6A,0x6C,0x6E,0x70,0x72,0x74,0x76,0x78,0x7A,0x7C,0x7E,
  0x81,0x83,0x85,0x87,0x89,0x8B,0x8D,0x8F,0x91,0x93,0x95,0x97,0x99,0x9B,0x9D,0x9F,
  0xA1,0xA3,0xA5,0xA7,0xA9,0xAB,0xAD,0xAF,0xB1,0xB3,0xB5,0xB7,0xB9,0xBB,0xBD,0xBF,
  0xC1,0xC3,0xC5,0xC7,0xC9,0xCB,0xCD,0xCF,0xD1,0xD3,0xD5,0xD7,0xD9,0xDB,0xDD,0xDF,
  0xE1,0xE3,0xE5,0xE7,0xE9,0xEB,0xED,0xEF,0xF1,0xF3,0xF5,0xF7,0xF9,0xFB,0xFD,0xFF,
  0x00,0x02,0x04,0x06,0x08,0x0A,0x0C,0x0E,0x10,0x12,0x14,0x16,0x18,0x1A,0x1C,0x1E,
  0x20,0x22,0x24,0x26,0x28,0x2A,0x2C,0x2E,0x30,0x32,0x34,0x36,0x38,0x3A,0x3C,0x3E,
  0x40,0x42,0x44,0x46,0x48,0x4A,0x4C,0x4E,0x50,0x52,0x54,0x56,0x58,0x5A,0x5C,0x5E,
  0x60,0x62,0x64,0x66,0x68,0x6A,0x6C,0x6E,0x70,0x72,0x74,0x76,0x78,0x7A,0x7C,0x7E,
  0x81,0x83,0x85,0x87,0x89,0x8B,0x8D,0x8F,0x91,0x93,0x95,0x97,0x99,0x9B,0x9D,0x9F,
  0xA1,0xA3,0xA5,0xA7,0xA9,0xAB,0xAD,0xAF,0xB1,0xB3,0xB5,0xB7,0xB9,0xBB,0xBD,0xBF,
  0xC1,0xC3,0xC5,0xC7,0xC9,0xCB,0xCD,0xCF,0xD1,0xD3,0xD5,0xD7,0xD9,0xDB,0xDD,0xDF,
  0xE1,0xE3,0xE5,0xE7,0xE9,0xEB,0xED,0xEF,0xF1,0xF3,0xF5,0xF7,0xF9,0xFB,0xFD,0xFF,
}, {
  0x00,0x04,0x08,0x0C,0x10,0x14,0x18,0x1C,0x20,0x24,0x28,0x2C,0x30,0x34,0x38,0x3C,
  0x41,0x45,0x49,0x4D,0x51,0x55,0x59,0x5D,0x61,0x65,0x69,0x6D,0x71,0x75,0x79,0x7D,
  0x82,0x86,0x8A,0x8E,0x92,0x96,0x9A,0x9E,0xA2,0xA6,0xAA,0xAE,0xB2,0xB6,0xBA,0xBE,
  0xC3,0xC7,0xCB,0xCF,0xD3,0xD7,0xDB,0xDF,0xE3,0xE7,0xEB,0xEF,0xF3,0xF7,0xFB,0xFF,
  0x00,0x04,0x08,0x0C,0x10,0x14,0x18,0x1C,0x20,0x24,0x28,0x2C,0x30,0x34,0x38,0x3C,
  0x41,0x45,0x49,0x4D,0x51,0x55,0x59,0x5D,0x61,0x65,0x69,0x6D,0x71,0x75,0x79,0x7D,
  0x82,0x86,0x8A,0x8E,0x92,0x96,0x9A,0x9E,0xA2,0xA6,0xAA,0xAE,0xB2,0xB6,0xBA,0xBE,
  0xC3,0xC7,0xCB,0xCF,0xD3,0xD7,0xDB,0xDF,0xE3,0xE7,0xEB,0xEF,0xF3,0xF7,0xFB,0xFF,
  0x00,0x04,0x08,0x0C,0x10,0x14,0x18,0x1C,0x20,0x24,0x28,0x2C,0x30,0x34,0x38,0x3C,
  0x41,0x45,0x49,0x4D,0x51,0x55,0x59,0x5D,0x61,0x65,0x69,0x6D,0x71,0x75,0x79,0x7D,
  0x82,0x86,0x8A,0x8E,0x92,0x96,0x9A,0x9E,0xA2,0xA6,0xAA,0xAE,0xB2,0xB6,0xBA,0xBE,
  0xC3,0xC7,0xCB,0xCF,0xD3,0xD7,0xDB,0xDF,0xE3,0xE7,0xEB,0xEF,0xF3,0xF7,0xFB,0xFF,
  0x00,0x04,0x08,0x0C,0x10,0x14,0x18,0x1C,0x20,0x24,0x28,0x2C,0x30,0x34,0x38,0x3C,
  0x41,0x45,0x49,0x4D,0x51,0x55,0x59,0x5D,0x61,0x65,0x69,0x6D,0x71,0x75,0x79,0x7D,
  0x82,0x86,0x8A,0x8E,0x92,0x96,0x9A,0x9E,0xA2,0xA6,0xAA,0xAE,0xB2,0xB6,0xBA,0xBE,
  0xC3,0xC7,0xCB,0xCF,0xD3,0xD7,0xDB,0xDF,0xE3,0xE7,0xEB,0xEF,0xF3,0xF7,0xFB,0xFF,
}, {
  0x00,0x08,0x10,0x18,0x21,0x29,0x31,0x39,0x42,0x4A,0x52,0x5A,0x63,0x6B,0x73,0x7B,
  0x84,0x8C,0x94,0x9C,0xA5,0xAD,0xB5,0xBD,0xC6,0xCE,0xD6,0xDE,0xE7,0xEF,0xF7,0xFF,
  0x00,0x08,0x10,0x18,0x21,0x29,0x31,0x39,0x42,0x4A,0x52,0x5A,0x63,0x6B,0x73,0x7B,
  0x84,0x8C,0x94,0x9C,0xA5,0xAD,0xB5,0xBD,0xC6,0xCE,0xD6,0xDE,0xE7,0xEF,0xF7,0xFF,
  0x00,0x08,0x10,0x18,0x21,0x29,0x31,0x39,0x42,0x4A,0x52,0x5A,0x63,0x6B,0x73,0x7B,
  0x84,0x8C,0x94,0x9C,0xA5,0xAD,0xB5,0xBD,0xC6,0xCE,0xD6,0xDE,0xE7,0xEF,0xF7,0xFF,
  0x00,0x08,0x10,0x18,0x21,0x29,0x31,0x39,0x42,0x4A,0x52,0x5A,0x63,0x6B,0x73,0x7B,
  0x84,0x8C,0x94,0x9C,0xA5,0xAD,0xB5,0xBD,0xC6,0xCE,0xD6,0xDE,0xE7,0xEF,0xF7,0xFF,
  0x00,0x08,0x10,0x18,0x21,0x29,0x31,0x39,0x42,0x4A,0x52,0x5A,0x63,0x6B,0x73,0x7B,
  0x84,0x8C,0x94,0x9C,0xA5,0xAD,0xB5,0xBD,0xC6,0xCE,0xD6,0xDE,0xE7,0xEF,0xF7,0xFF,
  0x00,0x08,0x10,0x18,0x21,0x29,0x31,0x39,0x42,0x4A,0x52,0x5A,0x63,0x6B,0x73,0x7B,
  0x84,0x8C,0x94,0x9C,0xA5,0xAD,0xB5,0xBD,0xC6,0xCE,0xD6,0xDE,0xE7,0xEF,0xF7,0xFF,
  0x00,0x08,0x10,0x18,0x21,0x29,0x31,0x39,0x42,0x4A,0x52,0x5A,0x63,0x6B,0x73,0x7B,
  0x84,0x8C,0x94,0x9C,0xA5,0xAD,0xB5,0xBD,0xC6,0xCE,0xD6,0xDE,0xE7,0xEF,0xF7,0xFF,
  0x00,0x08,0x10,0x18,0x21,0x29,0x31,0x39,0x42,0x4A,0x52,0x5A,0x63,0x6B,0x73,0x7B,
  0x84,0x8C,0x94,0x9C,0xA5,0xAD,0xB5,0xBD,0xC6,0xCE,0xD6,0xDE,0xE7,0xEF,0xF7,0xFF,
}, {
  0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,0x88,0x99,0xAA,0xBB,0xCC,0xDD,0xEE,0xFF,
  0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,0x88,0x99,0xAA,0xBB,0xCC,0xDD,0xEE,0xFF,
  0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,0x88,0x99,0xAA,0xBB,0xCC,0xDD,0xEE,0xFF,
  0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,0x88,0x99,0xAA,0xBB,0xCC,0xDD,0xEE,0xFF,
  0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,0x88,0x99,0xAA,0xBB,0xCC,0xDD,0xEE,0xFF,
  0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,0x88,0x99,0xAA,0xBB,0xCC,0xDD,0xEE,0xFF,
  0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,0x88,0x99,0xAA,0xBB,0xCC,0xDD,0xEE,0xFF,
  0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,0x88,0x99,0xAA,0xBB,0xCC,0xDD,0xEE,0xFF,
  0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,0x88,0x99,0xAA,0xBB,0xCC,0xDD,0xEE,0xFF,
  0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,0x88,0x99,0xAA,0xBB,0xCC,0xDD,0xEE,0xFF,
  0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,0x88,0x99,0xAA,0xBB,0xCC,0xDD,0xEE,0xFF,
  0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,0x88,0x99,0xAA,0xBB,0xCC,0xDD,0xEE,0xFF,
  0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,0x88,0x99,0xAA,0xBB,0xCC,0xDD,0xEE,0xFF,
  0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,0x88,0x99,0xAA,0xBB,0xCC,0xDD,0xEE,0xFF,
  0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,0x88,0x99,0xAA,0xBB,0xCC,0xDD,0xEE,0xFF,
  0x00,0x11,0x22,0x33,0x44,0x55,0x66,0x77,0x88,0x99,0xAA,0xBB,0xCC,0xDD,0xEE,0xFF,
}, {
  0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,
  0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,
  0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,
  0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,
  0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,
  0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,
  0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,
  0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,
  0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,
  0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,
  0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,
  0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,
  0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,
  0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,
  0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,
  0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,0x00,0x24,0x49,0x6D,0x92,0xB6,0xDB,0xFF,
}, {
  0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,
  0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,
  0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,
  0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,
  0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,
  0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,
  0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,
  0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,
  0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,
  0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,
  0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,
  0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,
  0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,
  0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,
  0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,
  0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,0x00,0x55,0xAA,0xFF,
}, {
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
  0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,
}};

#if !defined(QOIR_CONFIG__DISABLE_LARGE_LOOK_UP_TABLES)
// The table was generated by script/gen_table_luma.go
static uint8_t qoir_private_table_luma[65536] = {
    0xD8,0xE0,0xD8,0,0xD9,0xE1,0xD9,0,0xDA,0xE2,0xDA,0,0xDB,0xE3,0xDB,0,
    0xDC,0xE4,0xDC,0,0xDD,0xE5,0xDD,0,0xDE,0xE6,0xDE,0,0xDF,0xE7,0xDF,0,
    0xE0,0xE8,0xE0,0,0xE1,0xE9,0xE1,0,0xE2,0xEA,0xE2,0,0xE3,0xEB,0xE3,0,
    0xE4,0xEC,0xE4,0,0xE5,0xED,0xE5,0,0xE6,0xEE,0xE6,0,0xE7,0xEF,0xE7,0,
    0xE8,0xF0,0xE8,0,0xE9,0xF1,0xE9,0,0xEA,0xF2,0xEA,0,0xEB,0xF3,0xEB,0,
    0xEC,0xF4,0xEC,0,0xED,0xF5,0xED,0,0xEE,0xF6,0xEE,0,0xEF,0xF7,0xEF,0,
    0xF0,0xF8,0xF0,0,0xF1,0xF9,0xF1,0,0xF2,0xFA,0xF2,0,0xF3,0xFB,0xF3,0,
    0xF4,0xFC,0xF4,0,0xF5,0xFD,0xF5,0,0xF6,0xFE,0xF6,0,0xF7,0xFF,0xF7,0,
    0xF8,0x00,0xF8,0,0xF9,0x01,0xF9,0,0xFA,0x02,0xFA,0,0xFB,0x03,0xFB,0,
    0xFC,0x04,0xFC,0,0xFD,0x05,0xFD,0,0xFE,0x06,0xFE,0,0xFF,0x07,0xFF,0,
    0x00,0x08,0x00,0,0x01,0x09,0x01,0,0x02,0x0A,0x02,0,0x03,0x0B,0x03,0,
    0x04,0x0C,0x04,0,0x05,0x0D,0x05,0,0x06,0x0E,0x06,0,0x07,0x0F,0x07,0,
    0x08,0x10,0x08,0,0x09,0x11,0x09,0,0x0A,0x12,0x0A,0,0x0B,0x13,0x0B,0,
    0x0C,0x14,0x0C,0,0x0D,0x15,0x0D,0,0x0E,0x16,0x0E,0,0x0F,0x17,0x0F,0,
    0x10,0x18,0x10,0,0x11,0x19,0x11,0,0x12,0x1A,0x12,0,0x13,0x1B,0x13,0,
    0x14,0x1C,0x14,0,0x15,0x1D,0x15,0,0x16,0x1E,0x16,0,0x17,0x1F,0x17,0,
    0xD9,0xE0,0xD8,0,0xDA,0xE1,0xD9,0,0xDB,0xE2,0xDA,0,0xDC,0xE3,0xDB,0,
    0xDD,0xE4,0xDC,0,0xDE,0xE5,0xDD,0,0xDF,0xE6,0xDE,0,0xE0,0xE7,0xDF,0,
    0xE1,0xE8,0xE0,0,0xE2,0xE9,0xE1,0,0xE3,0xEA,0xE2,0,0xE4,0xEB,0xE3,0,
    0xE5,0xEC,0xE4,0,0xE6,0xED,0xE5,0,0xE7,0xEE,0xE6,0,0xE8,0xEF,0xE7,0,
    0xE9,0xF0,0xE8,0,0xEA,0xF1,0xE9,0,0xEB,0xF2,0xEA,0,0xEC,0xF3,0xEB,0,
    0xED,0xF4,0xEC,0,0xEE,0xF5,0xED,0,0xEF,0xF6,0xEE,0,0xF0,0xF7,0xEF,0,
    0xF1,0xF8,0xF0,0,0xF2,0xF9,0xF1,0,0xF3,0xFA,0xF2,0,0xF4,0xFB,0xF3,0,
    0xF5,0xFC,0xF4,0,0xF6,0xFD,0xF5,0,0xF7,0xFE,0xF6,0,0xF8,0xFF,0xF7,0,
    0xF9,0x00,0xF8,0,0xFA,0x01,0xF9,0,0xFB,0x02,0xFA,0,0xFC,0x03,0xFB,0,
    0xFD,0x04,0xFC,0,0xFE,0x05,0xFD,0,0xFF,0x06,0xFE,0,0x00,0x07,0xFF,0,
    0x01,0x08,0x00,0,0x02,0x09,0x01,0,0x03,0x0A,0x02,0,0x04,0x0B,0x03,0,
    0x05,0x0C,0x04,0,0x06,0x0D,0x05,0,0x07,0x0E,0x06,0,0x08,0x0F,0x07,0,
    0x09,0x10,0x08,0,0x0A,0x11,0x09,0,0x0B,0x12,0x0A,0,0x0C,0x13,0x0B,0,
    0x0D,0x14,0x0C,0,0x0E,0x15,0x0D,0,0x0F,0x16,0x0E,0,0x10,0x17,0x0F,0,
    0x11,0x18,0x10,0,0x12,0x19,0x11,0,0x13,0x1A,0x12,0,0x14,0x1B,0x13,0,
    0x15,0x1C,0x14,0,0x16,0x1D,0x15,0,0x17,0x1E,0x16,0,0x18,0x1F,0x17,0,
    0xDA,0xE0,0xD8,0,0xDB,0xE1,0xD9,0,0xDC,0xE2,0xDA,0,0xDD,0xE3,0xDB,0,
    0xDE,0xE4,0xDC,0,0xDF,0xE5,0xDD,0,0xE0,0xE6,0xDE,0,0xE1,0xE7,0xDF,0,
    0xE2,0xE8,0xE0,0,0xE3,0xE9,0xE1,0,0xE4,0xEA,0xE2,0,0xE5,0xEB,0xE3,0,
    0xE6,0xEC,0xE4,0,0xE7,0xED,0xE5,0,0xE8,0xEE,0xE6,0,0xE9,0xEF,0xE7,0,
    0xEA,0xF0,0xE8,0,0xEB,0xF1,0xE9,0,0xEC,0xF2,0xEA,0,0xED,0xF3,0xEB,0,
    0xEE,0xF4,0xEC,0,0xEF,0xF5,0xED,0,0xF0,0xF6,0xEE,0,0xF1,0xF7,0xEF,0,
    0xF2,0xF8,0xF0,0,0xF3,0xF9,0xF1,0,0xF4,0xFA,0xF2,0,0xF5,0xFB,0xF3,0,
    0xF6,0xFC,0xF4,0,0xF7,0xFD,0xF5,0,0xF8,0xFE,0xF6,0,0xF9,0xFF,0xF7,0,
    0xFA,0x00,0xF8,0,0xFB,0x01,0xF9,0,0xFC,0x02,0xFA,0,0xFD,0x03,0xFB,0,
    0xFE,0x04,0xFC,0,0xFF,0x05,0xFD,0,0x00,0x06,0xFE,0,0x01,0x07,0xFF,0,
    0x02,0x08,0x00,0,0x03,0x09,0x01,0,0x04,0x0A,0x02,0,0x05,0x0B,0x03,0,
    0x06,0x0C,0x04,0,0x07,0x0D,0x05,0,0x08,0x0E,0x06,0,0x09,0x0F,0x07,0,
    0x0A,0x10,0x08,0,0x0B,0x11,0x09,0,0x0C,0x12,0x0A,0,0x0D,0x13,0x0B,0,
    0x0E,0x14,0x0C,0,0x0F,0x15,0x0D,0,0x10,0x16,0x0E,0,0x11,0x17,0x0F,0,
    0x12,0x18,0x10,0,0x13,0x19,0x11,0,0x14,0x1A,0x12,0,0x15,0x1B,0x13,0,
    0x16,0x1C,0x14,0,0x17,0x1D,0x15,0,0x18,0x1E,0x16,0,0x19,0x1F,0x17,0,
    0xDB,0xE0,0xD8,0,0xDC,0xE1,0xD9,0,0xDD,0xE2,0xDA,0,0xDE,0xE3,0xDB,0,
    0xDF,0xE4,0xDC,0,0xE0,0xE5,0xDD,0,0xE1,0xE6,0xDE,0,0xE2,0xE7,0xDF,0,
    0xE3,0xE8,0xE0,0,0xE4,0xE9,0xE1,0,0xE5,0xEA,0xE2,0,0xE6,0xEB,0xE3,0,
    0xE7,0xEC,0xE4,0,0xE8,0xED,0xE5,0,0xE9,0xEE,0xE6,0,0xEA,0xEF,0xE7,0,
    0xEB,0xF0,0xE8,0,0xEC,0xF1,0xE9,0,0xED,0xF2,0xEA,0,0xEE,0xF3,0xEB,0,
    0xEF,0xF4,0xEC,0,0xF0,0xF5,0xED,0,0xF1,0xF6,0xEE,0,0xF2,0xF7,0xEF,0,
    0xF3,0xF8,0xF0,0,0xF4,0xF9,0xF1,0,0xF5,0xFA,0xF2,0,0xF6,0xFB,0xF3,0,
    0xF7,0xFC,0xF4,0,0xF8,0xFD,0xF5,0,0xF9,0xFE,0xF6,0,0xFA,0xFF,0xF7,0,
    0xFB,0x00,0xF8,0,0xFC,0x01,0xF9,0,0xFD,0x02,0xFA,0,0xFE,0x03,0xFB,0,
    0xFF,0x04,0xFC,0,0x00,0x05,0xFD,0,0x01,0x06,0xFE,0,0x02,0x07,0xFF,0,
    0x03,0x08,0x00,0,0x04,0x09,0x01,0,0x05,0x0A,0x02,0,0x06,0x0B,0x03,0,
    0x07,0x0C,0x04,0,0x08,0x0D,0x05,0,0x09,0x0E,0x06,0,0x0A,0x0F,0x07,0,
    0x0B,0x10,0x08,0,0x0C,0x11,0x09,0,0x0D,0x12,0x0A,0,0x0E,0x13,0x0B,0,
    0x0F,0x14,0x0C,0,0x10,0x15,0x0D,0,0x11,0x16,0x0E,0,0x12,0x17,0x0F,0,
    0x13,0x18,0x10,0,0x14,0x19,0x11,0,0x15,0x1A,0x12,0,0x16,0x1B,0x13,0,
    0x17,0x1C,0x14,0,0x18,0x1D,0x15,0,0x19,0x1E,0x16,0,0x1A,0x1F,0x17,0,
    0xDC,0xE0,0xD8,0,0xDD,0xE1,0xD9,0,0xDE,0xE2,0xDA,0,0xDF,0xE3,0xDB,0,
    0xE0,0xE4,0xDC,0,0xE1,0xE5,0xDD,0,0xE2,0xE6,0xDE,0,0xE3,0xE7,0xDF,0,
    0xE4,0xE8,0xE0,0,0xE5,0xE9,0xE1,0,0xE6,0xEA,0xE2,0,0xE7,0xEB,0xE3,0,
    0xE8,0xEC,0xE4,0,0xE9,0xED,0xE5,0,0xEA,0xEE,0xE6,0,0xEB,0xEF,0xE7,0,
    0xEC,0xF0,0xE8,0,0xED,0xF1,0xE9,0,0xEE,0xF2,0xEA,0,0xEF,0xF3,0xEB,0,
    0xF0,0xF4,0xEC,0,0xF1,0xF5,0xED,0,0xF2,0xF6,0xEE,0,0xF3,0xF7,0xEF,0,
    0xF4,0xF8,0xF0,0,0xF5,0xF9,0xF1,0,0xF6,0xFA,0xF2,0,0xF7,0xFB,0xF3,0,
    0xF8,0xFC,0xF4,0,0xF9,0xFD,0xF5,0,0xFA,0xFE,0xF6,0,0xFB,0xFF,0xF7,0,
    0xFC,0x00,0xF8,0,0xFD,0x01,0xF9,0,0xFE,0x02,0xFA,0,0xFF,0x03,0xFB,0,
    0x00,0x04,0xFC,0,0x01,0x05,0xFD,0,0x02,0x06,0xFE,0,0x03,0x07,0xFF,0,
    0x04,0x08,0x00,0,0x05,0x09,0x01,0,0x06,0x0A,0x02,0,0x07,0x0B,0x03,0,
    0x08,0x0C,0x04,0,0x09,0x0D,0x05,0,0x0A,0x0E,0x06,0,0x0B,0x0F,0x07,0,
    0x0C,0x10,0x08,0,0x0D,0x11,0x09,0,0x0E,0x12,0x0A,0,0x0F,0x13,0x0B,0,
    0x10,0x14,0x0C,0,0x11,0x15,0x0D,0,0x12,0x16,0x0E,0,0x13,0x17,0x0F,0,
    0x14,0x18,0x10,0,0x15,0x19,0x11,0,0x16,0x1A,0x12,0,0x17,0x1B,0x13,0,
    0x18,0x1C,0x14,0,0x19,0x1D,0x15,0,0x1A,0x1E,0x16,0,0x1B,0x1F,0x17,0,
    0xDD,0xE0,0xD8,0,0xDE,0xE1,0xD9,0,0xDF,0xE2,0xDA,0,0xE0,0xE3,0xDB,0,
    0xE1,0xE4,0xDC,0,0xE2,0xE5,0xDD,0,0xE3,0xE6,0xDE,0,0xE4,0xE7,0xDF,0,
    0xE5,0xE8,0xE0,0,0xE6,0xE9,0xE1,0,0xE7,0xEA,0xE2,0,0xE8,0xEB,0xE3,0,
    0xE9,0xEC,0xE4,0,0xEA,0xED,0xE5,0,0xEB,0xEE,0xE6,0,0xEC,0xEF,0xE7,0,
    0xED,0xF0,0xE8,0,0xEE,0xF1,0xE9,0,0xEF,0xF2,0xEA,0,0xF0,0xF3,0xEB,0,
    0xF1,0xF4,0xEC,0,0xF2,0xF5,0xED,0,0xF3,0xF6,0xEE,0,0xF4,0xF7,0xEF,0,
    0xF5,0xF8,0xF0,0,0xF6,0xF9,0xF1,0,0xF7,0xFA,0xF2,0,0xF8,0xFB,0xF3,0,
    0xF9,0xFC,0xF4,0,0xFA,0xFD,0xF5,0,0xFB,0xFE,0xF6,0,0xFC,0xFF,0xF7,0,
    0xFD,0x00,0xF8,0,0xFE,0x01,0xF9,0,0xFF,0x02,0xFA,0,0x00,0x03,0xFB,0,
    0x01,0x04,0xFC,0,0x02,0x05,0xFD,0,0x03,0x06,0xFE,0,0x04,0x07,0xFF,0,
    0x05,0x08,0x00,0,0x06,0x09,0x01,0,0x07,0x0A,0x02,0,0x08,0x0B,0x03,0,
    0x09,0x0C,0x04,0,0x0A,0x0D,0x05,0,0x0B,0x0E,0x06,0,0x0C,0x0F,0x07,0,
    0x0D,0x10,0x08,0,0x0E,0x11,0x09,0,0x0F,0x12,0x0A,0,0x10,0x13,0x0B,0,
    0x11,0x14,0x0C,0,0x12,0x15,0x0D,0,0x13,0x16,0x0E,0,0x14,0x17,0x0F,0,
    0x15,0x18,0x10,0,0x16,0x19,0x11,0,0x17,0x1A,0x12,0,0x18,0x1B,0x13,0,
    0x19,0x1C,0x14,0,0x1A,0x1D,0x15,0,0x1B,0x1E,0x16,0,0x1C,0x1F,0x17,0,
    0xDE,0xE0,0xD8,0,0xDF,0xE1,0xD9,0,0xE0,0xE2,0xDA,0,0xE1,0xE3,0xDB,0,
    0xE2,0xE4,0xDC,0,0xE3,0xE5,0xDD,0,0xE4,0xE6,0xDE,0,0xE5,0xE7,0xDF,0,
    0xE6,0xE8,0xE0,0,0xE7,0xE9,0xE1,0,0xE8,0xEA,0xE2,0,0xE9,0xEB,0xE3,0,
    0xEA,0xEC,0xE4,0,0xEB,0xED,0xE5,0,0xEC,0xEE,0xE6,0,0xED,0xEF,0xE7,0,
    0xEE,0xF0,0xE8,0,0xEF,0xF1,0xE9,0,0xF0,0xF2,0xEA,0,0xF1,0xF3,0xEB,0,
    0xF2,0xF4,0xEC,0,0xF3,0xF5,0xED,0,0xF4,0xF6,0xEE,0,0xF5,0xF7,0xEF,0,
    0xF6,0xF8,0xF0,0,0xF7,0xF9,0xF1,0,0xF8,0xFA,0xF2,0,0xF9,0xFB,0xF3,0,
    0xFA,0xFC,0xF4,0,0xFB,0xFD,0xF5,0,0xFC,0xFE,0xF6,0,0xFD,0xFF,0xF7,0,
    0xFE,0x00,0xF8,0,0xFF,0x01,0xF9,0,0x00,0x02,0xFA,0,0x01,0x03,0xFB,0,
    0x02,0x04,0xFC,0,0x03,0x05,0xFD,0,0x04,0x06,0xFE,0,0x05,0x07,0xFF,0,
    0x06,0x08,0x00,0,0x07,0x09,0x01,0,0x08,0x0A,0x02,0,0x09,0x0B,0x03,0,
    0x0A,0x0C,0x04,0,0x0B,0x0D,0x05,0,0x0C,0x0E,0x06,0,0x0D,0x0F,0x07,0,
    0x0E,0x10,0x08,0,0x0F,0x11,0x09,0,0x10,0x12,0x0A,0,0x11,0x13,0x0B,0,
    0x12,0x14,0x0C,0,0x13,0x15,0x0D,0,0x14,0x16,0x0E,0,0x15,0x17,0x0F,0,
    0x16,0x18,0x10,0,0x17,0x19,0x11,0,0x18,0x1A,0x12,0,0x19,0x1B,0x13,0,
    0x1A,0x1C,0x14,0,0x1B,0x1D,0x15,0,0x1C,0x1E,0x16,0,0x1D,0x1F,0x17,0,
    0xDF,0xE0,0xD8,0,0xE0,0xE1,0xD9,0,0xE1,0xE2,0xDA,0,0xE2,0xE3,0xDB,0,
    0xE3,0xE4,0xDC,0,0xE4,0xE5,0xDD,0,0xE5,0xE6,0xDE,0,0xE6,0xE7,0xDF,0,
    0xE7,0xE8,0xE0,0,0xE8,0xE9,0xE1,0,0xE9,0xEA,0xE2,0,0xEA,0xEB,0xE3,0,
    0xEB,0xEC,0xE4,0,0xEC,0xED,0xE5,0,0xED,0xEE,0xE6,0,0xEE,0xEF,0xE7,0,
    0xEF,0xF0,0xE8,0,0xF0,0xF1,0xE9,0,0xF1,0xF2,0xEA,0,0xF2,0xF3,0xEB,0,
    0xF3,0xF4,0xEC,0,0xF4,0xF5,0xED,0,0xF5,0xF6,0xEE,0,0xF6,0xF7,0xEF,0,
    0xF7,0xF8,0xF0,0,0xF8,0xF9,0xF1,0,0xF9,0xFA,0xF2,0,0xFA,0xFB,0xF3,0,
    0xFB,0xFC,0xF4,0,0xFC,0xFD,0xF5,0,0xFD,0xFE,0xF6,0,0xFE,0xFF,0xF7,0,
    0xFF,0x00,0xF8,0,0x00,0x01,0xF9,0,0x01,0x02,0xFA,0,0x02,0x03,0xFB,0,
    0x03,0x04,0xFC,0,0x04,0x05,0xFD,0,0x05,0x06,0xFE,0,0x06,0x07,0xFF,0,
    0x07,0x08,0x00,0,0x08,0x09,0x01,0,0x09,0x0A,0x02,0,0x0A,0x0B,0x03,0,
    0x0B,0x0C,0x04,0,0x0C,0x0D,0x05,0,0x0D,0x0E,0x06,0,0x0E,0x0F,0x07,0,
    0x0F,0x10,0x08,0,0x10,0x11,0x09,0,0x11,0x12,0x0A,0,0x12,0x13,0x0B,0,
    0x13,0x14,0x0C,0,0x14,0x15,0x0D,0,0x15,0x16,0x0E,0,0x16,0x17,0x0F,0,
    0x17,0x18,0x10,0,0x18,0x19,0x11,0,0x19,0x1A,0x12,0,0x1A,0x1B,0x13,0,
    0x1B,0x1C,0x14,0,0x1C,0x1D,0x15,0,0x1D,0x1E,0x16,0,0x1E,0x1F,0x17,0,
    0xE0,0xE0,0xD8,0,0xE1,0xE1,0xD9,0,0xE2,0xE2,0xDA,0,0xE3,0xE3,0xDB,0,
    0xE4,0xE4,0xDC,0,0xE5,0xE5,0xDD,0,0xE6,0xE6,0xDE,0,0xE7,0xE7,0xDF,0,
    0xE8,0xE8,0xE0,0,0xE9,0xE9,0xE1,0,0xEA,0xEA,0xE2,0,0xEB,0xEB,0xE3,0,
    0xEC,0xEC,0xE4,0,0xED,0xED,0xE5,0,0xEE,0xEE,0xE6,0,0xEF,0xEF,0xE7,0,
    0xF0,0xF0,0xE8,0,0xF1,0xF1,0xE9,0,0xF2,0xF2,0xEA,0,0xF3,0xF3,0xEB,0,
    0xF4,0xF4,0xEC,0,0xF5,0xF5,0xED,0,0xF6,0xF6,0xEE,0,0xF7,0xF7,0xEF,0,
    0xF8,0xF8,0xF0,0,0xF9,0xF9,0xF1,0,0xFA,0xFA,0xF2,0,0xFB,0xFB,0xF3,0,
    0xFC,0xFC,0xF4,0,0xFD,0xFD,0xF5,0,0xFE,0xFE,0xF6,0,0xFF,0xFF,0xF7,0,
    0x00,0x00,0xF8,0,0x01,0x01,0xF9,0,0x02,0x02,0xFA,0,0x03,0x03,0xFB,0,
    0x04,0x04,0xFC,0,0x05,0x05,0xFD,0,0x06,0x06,0xFE,0,0x07,0x07,0xFF,0,
    0x08,0x08,0x00,0,0x09,0x09,0x01,0,0x0A,0x0A,0x02,0,0x0B,0x0B,0x03,0,
    0x0C,0x0C,0x04,0,0x0D,0x0D,0x05,0,0x0E,0x0E,0x06,0,0x0F,0x0F,0x07,0,
    0x10,0x10,0x08,0,0x11,0x11,0x09,0,0x12,0x12,0x0A,0,0x13,0x13,0x0B,0,
    0x14,0x14,0x0C,0,0x15,0x15,0x0D,0,0x16,0x16,0x0E,0,0x17,0x17,0x0F,0,
    0x18,0x18,0x10,0,0x19,0x19,0x11,0,0x1A,0x1A,0x12,0,0x1B,0x1B,0x13,0,
    0x1C,0x1C,0x14,0,0x1D,0x1D,0x15,0,0x1E,0x1E,0x16,0,0x1F,0x1F,0x17,0,
    0xE1,0xE0,0xD8,0,0xE2,0xE1,0xD9,0,0xE3,0xE2,0xDA,0,0xE4,0xE3,0xDB,0,
    0xE5,0xE4,0xDC,0,0xE6,0xE5,0xDD,0,0xE7,0xE6,0xDE,0,0xE8,0xE7,0xDF,0,
    0xE9,0xE8,0xE0,0,0xEA,0xE9,0xE1,0,0xEB,0xEA,0xE2,0,0xEC,0xEB,0xE3,0,
    0xED,0xEC,0xE4,0,0xEE,0xED,0xE5,0,0xEF,0xEE,0xE6,0,0xF0,0xEF,0xE7,0,
    0xF1,0xF0,0xE8,0,0xF2,0xF1,0xE9,0,0xF3,0xF2,0xEA,0,0xF4,0xF3,0xEB,0,
    0xF5,0xF4,0xEC,0,0xF6,0xF5,0xED,0,0xF7,0xF6,0xEE,0,0xF8,0xF7,0xEF,0,
    0xF9,0xF8,0xF0,0,0xFA,0xF9,0xF1,0,0xFB,0xFA,0xF2,0,0xFC,0xFB,0xF3,0,
    0xFD,0xFC,0xF4,0,0xFE,0xFD,0xF5,0,0xFF,0xFE,0xF6,0,0x00,0xFF,0xF7,0,
    0x01,0x00,0xF8,0,0x02,0x01,0xF9,0,0x03,0x02,0xFA,0,0x04,0x03,0xFB,0,
    0x05,0x04,0xFC,0,0x06,0x05,0xFD,0,0x07,0x06,0xFE,0,0x08,0x07,0xFF,0,
    0x09,0x08,0x00,0,0x0A,0x09,0x01,0,0x0B,0x0A,0x02,0,0x0C,0x0B,0x03,0,
    0x0D,0x0C,0x04,0,0x0E,0x0D,0x05,0,0x0F,0x0E,0x06,0,0x10,0x0F,0x07,0,
    0x11,0x10,0x08,0,0x12,0x11,0x09,0,0x13,0x12,0x0A,0,0x14,0x13,0x0B,0,
    0x15,0x14,0x0C,0,0x16,0x15,0x0D,0,0x17,0x16,0x0E,0,0x18,0x17,0x0F,0,
    0x19,0x18,0x10,0,0x1A,0x19,0x11,0,0x1B,0x1A,0x12,0,0x1C,0x1B,0x13,0,
    0x1D,0x1C,0x14,0,0x1E,0x1D,0x15,0,0x1F,0x1E,0x16,0,0x20,0x1F,0x17,0,
    0xE2,0xE0,0xD8,0,0xE3,0xE1,0xD9,0,0xE4,0xE2,0xDA,0,0xE5,0xE3,0xDB,0,
    0xE6,0xE4,0xDC,0,0xE7,0xE5,0xDD,0,0xE8,0xE6,0xDE,0,0xE9,0xE7,0xDF,0,
    0xEA,0xE8,0xE0,0,0xEB,0xE9,0xE1,0,0xEC,0xEA,0xE2,0,0xED,0xEB,0xE3,0,
    0xEE,0xEC,0xE4,0,0xEF,0xED,0xE5,0,0xF0,0xEE,0xE6,0,0xF1,0xEF,0xE7,0,
    0xF2,0xF0,0xE8,0,0xF3,0xF1,0xE9,0,0xF4,0xF2,0xEA,0,0xF5,0xF3,0xEB,0,
    0xF6,0xF4,0xEC,0,0xF7,0xF5,0xED,0,0xF8,0xF6,0xEE,0,0xF9,0xF7,0xEF,0,
    0xFA,0xF8,0xF0,0,0xFB,0xF9,0xF1,0,0xFC,0xFA,0xF2,0,0xFD,0xFB,0xF3,0,
    0xFE,0xFC,0xF4,0,0xFF,0xFD,0xF5,0,0x00,0xFE,0xF6,0,0x01,0xFF,0xF7,0,
    0x02,0x00,0xF8,0,0x03,0x01,0xF9,0,0x04,0x02,0xFA,0,0x05,0x03,0xFB,0,
    0x06,0x04,0xFC,0,0x07,0x05,0xFD,0,0x08,0x06,0xFE,0,0x09,0x07,0xFF,0,
    0x0A,0x08,0x00,0,0x0B,0x09,0x01,0,0x0C,0x0A,0x02,0,0x0D,0x0B,0x03,0,
    0x0E,0x0C,0x04,0,0x0F,0x0D,0x05,0,0x10,0x0E,0x06,0,0x11,0x0F,0x07,0,
    0x12,0x10,0x08,0,0x13,0x11,0x09,0,0x14,0x12,0x0A,0,0x15,0x13,0x0B,0,
    0x16,0x14,0x0C,0,0x17,0x15,0x0D,0,0x18,0x16,0x0E,0,0x19,0x17,0x0F,0,
    0x1A,0x18,0x10,0,0x1B,0x19,0x11,0,0x1C,0x1A,0x12,0,0x1D,0x1B,0x13,0,
    0x1E,0x1C,0x14,0,0x1F,0x1D,0x15,0,0x20,0x1E,0x16,0,0x21,0x1F,0x17,0,
    0xE3,0xE0,0xD8,0,0xE4,0xE1,0xD9,0,0xE5,0xE2,0xDA,0,0xE6,0xE3,0xDB,0,
    0xE7,0xE4,0xDC,0,0xE8,0xE5,0xDD,0,0xE9,0xE6,0xDE,0,0xEA,0xE7,0xDF,0,
    0xEB,0xE8,0xE0,0,0xEC,0xE9,0xE1,0,0xED,0xEA,0xE2,0,0xEE,0xEB,0xE3,0,
    0xEF,0xEC,0xE4,0,0xF0,0xED,0xE5,0,0xF1,0xEE,0xE6,0,0xF2,0xEF,0xE7,0,
    0xF3,0xF0,0xE8,0,0xF4,0xF1,0xE9,0,0xF5,0xF2,0xEA,0,0xF6,0xF3,0xEB,0,
    0xF7,0xF4,0xEC,0,0xF8,0xF5,0xED,0,0xF9,0xF6,0xEE,0,0xFA,0xF7,0xEF,0,
    0xFB,0xF8,0xF0,0,0xFC,0xF9,0xF1,0,0xFD,0xFA,0xF2,0,0xFE,0xFB,0xF3,0,
    0xFF,0xFC,0xF4,0,0x00,0xFD,0xF5,0,0x01,0xFE,0xF6,0,0x02,0xFF,0xF7,0,
    0x03,0x00,0xF8,0,0x04,0x01,0xF9,0,0x05,0x02,0xFA,0,0x06,0x03,0xFB,0,
    0x07,0x04,0xFC,0,0x08,0x05,0xFD,0,0x09,0x06,0xFE,0,0x0A,0x07,0xFF,0,
    0x0B,0x08,0x00,0,0x0C,0x09,0x01,0,0x0D,0x0A,0x02,0,0x0E,0x0B,0x03,0,
    0x0F,0x0C,0x04,0,0x10,0x0D,0x05,0,0x11,0x0E,0x06,0,0x12,0x0F,0x07,0,
    0x13,0x10,0x08,0,0x14,0x11,0x09,0,0x15,0x12,0x0A,0,0x16,0x13,0x0B,0,
    0x17,0x14,0x0C,0,0x18,0x15,0x0D,0,0x19,0x16,0x0E,0,0x1A,0x17,0x0F,0,
    0x1B,0x18,0x10,0,0x1C,0x19,0x11,0,0x1D,0x1A,0x12,0,0x1E,0x1B,0x13,0,
    0x1F,0x1C,0x14,0,0x20,0x1D,0x15,0,0x21,0x1E,0x16,0,0x22,0x1F,0x17,0,
    0xE4,0xE0,0xD8,0,0xE5,0xE1,0xD9,0,0xE6,0xE2,0xDA,0,0xE7,0xE3,0xDB,0,
    0xE8,0xE4,0xDC,0,0xE9,0xE5,0xDD,0,0xEA,0xE6,0xDE,0,0xEB,0xE7,0xDF,0,
    0xEC,0xE8,0xE0,0,0xED,0xE9,0xE1,0,0xEE,0xEA,0xE2,0,0xEF,0xEB,0xE3,0,
    0xF0,0xEC,0xE4,0,0xF1,0xED,0xE5,0,0xF2,0xEE,0xE6,0,0xF3,0xEF,0xE7,0,
    0xF4,0xF0,0xE8,0,0xF5,0xF1,0xE9,0,0xF6,0xF2,0xEA,0,0xF7,0xF3,0xEB,0,
    0xF8,0xF4,0xEC,0,0xF9,0xF5,0xED,0,0xFA,0xF6,0xEE,0,0xFB,0xF7,0xEF,0,
    0xFC,0xF8,0xF0,0,0xFD,0xF9,0xF1,0,0xFE,0xFA,0xF2,0,0xFF,0xFB,0xF3,0,
    0x00,0xFC,0xF4,0,0x01,0xFD,0xF5,0,0x02,0xFE,0xF6,0,0x03,0xFF,0xF7,0,
    0x04,0x00,0xF8,0,0x05,0x01,0xF9,0,0x06,0x02,0xFA,0,0x07,0x03,0xFB,0,
    0x08,0x04,0xFC,0,0x09,0x05,0xFD,0,0x0A,0x06,0xFE,0,0x0B,0x07,0xFF,0,
    0x0C,0x08,0x00,0,0x0D,0x09,0x01,0,0x0E,0x0A,0x02,0,0x0F,0x0B,0x03,0,
    0x10,0x0C,0x04,0,0x11,0x0D,0x05,0,0x12,0x0E,0x06,0,0x13,0x0F,0x07,0,
    0x14,0x10,0x08,0,0x15,0x11,0x09,0,0x16,0x12,0x0A,0,0x17,0x13,0x0B,0,
    0x18,0x14,0x0C,0,0x19,0x15,0x0D,0,0x1A,0x16,0x0E,0,0x1B,0x17,0x0F,0,
    0x1C,0x18,0x10,0,0x1D,0x19,0x11,0,0x1E,0x1A,0x12,0,0x1F,0x1B,0x13,0,
    0x20,0x1C,0x14,0,0x21,0x1D,0x15,0,0x22,0x1E,0x16,0,0x23,0x1F,0x17,0,
    0xE5,0xE0,0xD8,0,0xE6,0xE1,0xD9,0,0xE7,0xE2,0xDA,0,0xE8,0xE3,0xDB,0,
    0xE9,0xE4,0xDC,0,0xEA,0xE5,0xDD,0,0xEB,0xE6,0xDE,0,0xEC,0xE7,0xDF,0,
    0xED,0xE8,0xE0,0,0xEE,0xE9,0xE1,0,0xEF,0xEA,0xE2,0,0xF0,0xEB,0xE3,0,
    0xF1,0xEC,0xE4,0,0xF2,0xED,0xE5,0,0xF3,0xEE,0xE6,0,0xF4,0xEF,0xE7,0,
    0xF5,0xF0,0xE8,0,0xF6,0xF1,0xE9,0,0xF7,0xF2,0xEA,0,0xF8,0xF3,0xEB,0,
    0xF9,0xF4,0xEC,0,0xFA,0xF5,0xED,0,0xFB,0xF6,0xEE,0,0xFC,0xF7,0xEF,0,
    0xFD,0xF8,0xF0,0,0xFE,0xF9,0xF1,0,0xFF,0xFA,0xF2,0,0x00,0xFB,0xF3,0,
    0x01,0xFC,0xF4,0,0x02,0xFD,0xF5,0,0x03,0xFE,0xF6,0,0x04,0xFF,0xF7,0,
    0x05,0x00,0xF8,0,0x06,0x01,0xF9,0,0x07,0x02,0xFA,0,0x08,0x03,0xFB,0,
    0x09,0x04,0xFC,0,0x0A,0x05,0xFD,0,0x0B,0x06,0xFE,0,0x0C,0x07,0xFF,0,
    0x0D,0x08,0x00,0,0x0E,0x09,0x01,0,0x0F,0x0A,0x02,0,0x10,0x0B,0x03,0,
    0x11,0x0C,0x04,0,0x12,0x0D,0x05,0,0x13,0x0E,0x06,0,0x14,0x0F,0x07,0,
    0x15,0x10,0x08,0,0x16,0x11,0x09,0,0x17,0x12,0x0A,0,0x18,0x13,0x0B,0,
    0x19,0x14,0x0C,0,0x1A,0x15,0x0D,0,0x1B,0x16,0x0E,0,0x1C,0x17,0x0F,0,
    0x1D,0x18,0x10,0,0x1E,0x19,0x11,0,0x1F,0x1A,0x12,0,0x20,0x1B,0x13,0,
    0x21,0x1C,0x14,0,0x22,0x1D,0x15,0,0x23,0x1E,0x16,0,0x24,0x1F,0x17,0,
    0xE6,0xE0,0xD8,0,0xE7,0xE1,0xD9,0,0xE8,0xE2,0xDA,0,0xE9,0xE3,0xDB,0,
    0xEA,0xE4,0xDC,0,0xEB,0xE5,0xDD,0,0xEC,0xE6,0xDE,0,0xED,0xE7,0xDF,0,
    0xEE,0xE8,0xE0,0,0xEF,0xE9,0xE1,0,0xF0,0xEA,0xE2,0,0xF1,0xEB,0xE3,0,
    0xF2,0xEC,0xE4,0,0xF3,0xED,0xE5,0,0xF4,0xEE,0xE6,0,0xF5,0xEF,0xE7,0,
    0xF6,0xF0,0xE8,0,0xF7,0xF1,0xE9,0,0xF8,0xF2,0xEA,0,0xF9,0xF3,0xEB,0,
    0xFA,0xF4,0xEC,0,0xFB,0xF5,0xED,0,0xFC,0xF6,0xEE,0,0xFD,0xF7,0xEF,0,
    0xFE,0xF8,0xF0,0,0xFF,0xF9,0xF1,0,0x00,0xFA,0xF2,0,0x01,0xFB,0xF3,0,
    0x02,0xFC,0xF4,0,0x03,0xFD,0xF5,0,0x04,0xFE,0xF6,0,0x05,0xFF,0xF7,0,
    0x06,0x00,0xF8,0,0x07,0x01,0xF9,0,0x08,0x02,0xFA,0,0x09,0x03,0xFB,0,
    0x0A,0x04,0xFC,0,0x0B,0x05,0xFD,0,0x0C,0x06,0xFE,0,0x0D,0x07,0xFF,0,
    0x0E,0x08,0x00,0,0x0F,0x09,0x01,0,0x10,0x0A,0x02,0,0x11,0x0B,0x03,0,
    0x12,0x0C,0x04,0,0x13,0x0D,0x05,0,0x14,0x0E,0x06,0,0x15,0x0F,0x07,0,
    0x16,0x10,0x08,0,0x17,0x11,0x09,0,0x18,0x12,0x0A,0,0x19,0x13,0x0B,0,
    0x1A,0x14,0x0C,0,0x1B,0x15,0x0D,0,0x1C,0x16,0x0E,0,0x1D,0x17,0x0F,0,
    0x1E,0x18,0x10,0,0x1F,0x19,0x11,0,0x20,0x1A,0x12,0,0x21,0x1B,0x13,0,
    0x22,0x1C,0x14,0,0x23,0x1D,0x15,0,0x24,0x1E,0x16,0,0x25,0x1F,0x17,0,
    0xE7,0xE0,0xD8,0,0xE8,0xE1,0xD9,0,0xE9,0xE2,0xDA,0,0xEA,0xE3,0xDB,0,
    0xEB,0xE4,0xDC,0,0xEC,0xE5,0xDD,0,0xED,0xE6,0xDE,0,0xEE,0xE7,0xDF,0,
    0xEF,0xE8,0xE0,0,0xF0,0xE9,0xE1,0,0xF1,0xEA,0xE2,0,0xF2,0xEB,0xE3,0,
    0xF3,0xEC,0xE4,0,0xF4,0xED,0xE5,0,0xF5,0xEE,0xE6,0,0xF6,0xEF,0xE7,0,
    0xF7,0xF0,0xE8,0,0xF8,0xF1,0xE9,0,0xF9,0xF2,0xEA,0,0xFA,0xF3,0xEB,0,
    0xFB,0xF4,0xEC,0,0xFC,0xF5,0xED,0,0xFD,0xF6,0xEE,0,0xFE,0xF7,0xEF,0,
    0xFF,0xF8,0xF0,0,0x00,0xF9,0xF1,0,0x01,0xFA,0xF2,0,0x02,0xFB,0xF3,0,
    0x03,0xFC,0xF4,0,0x04,0xFD,0xF5,0,0x05,0xFE,0xF6,0,0x06,0xFF,0xF7,0,
    0x07,0x00,0xF8,0,0x08,0x01,0xF9,0,0x09,0x02,0xFA,0,0x0A,0x03,0xFB,0,
    0x0B,0x04,0xFC,0,0x0C,0x05,0xFD,0,0x0D,0x06,0xFE,0,0x0E,0x07,0xFF,0,
    0x0F,0x08,0x00,0,0x10,0x09,0x01,0,0x11,0x0A,0x02,0,0x12,0x0B,0x03,0,
    0x13,0x0C,0x04,0,0x14,0x0D,0x05,0,0x15,0x0E,0x06,0,0x16,0x0F,0x07,0,
    0x17,0x10,0x08,0,0x18,0x11,0x09,0,0x19,0x12,0x0A,0,0x1A,0x13,0x0B,0,
    0x1B,0x14,0x0C,0,0x1C,0x15,0x0D,0,0x1D,0x16,0x0E,0,0x1E,0x17,0x0F,0,
    0x1F,0x18,0x10,0,0x20,0x19,0x11,0,0x21,0x1A,0x12,0,0x22,0x1B,0x13,0,
    0x23,0x1C,0x14,0,0x24,0x1D,0x15,0,0x25,0x1E,0x16,0,0x26,0x1F,0x17,0,
    0xD8,0xE0,0xD9,0,0xD9,0xE1,0xDA,0,0xDA,0xE2,0xDB,0,0xDB,0xE3,0xDC,0,
    0xDC,0xE4,0xDD,0,0xDD,0xE5,0xDE,0,0xDE,0xE6,0xDF,0,0xDF,0xE7,0xE0,0,
    0xE0,0xE8,0xE1,0,0xE1,0xE9,0xE2,0,0xE2,0xEA,0xE3,0,0xE3,0xEB,0xE4,0,
    0xE4,0xEC,0xE5,0,0xE5,0xED,0xE6,0,0xE6,0xEE,0xE7,0,0xE7,0xEF,0xE8,0,
    0xE8,0xF0,0xE9,0,0xE9,0xF1,0xEA,0,0xEA,0xF2,0xEB,0,0xEB,0xF3,0xEC,0,
    0xEC,0xF4,0xED,0,0xED,0xF5,0xEE,0,0xEE,0xF6,0xEF,0,0xEF,0xF7,0xF0,0,
    0xF0,0xF8,0xF1,0,0xF1,0xF9,0xF2,0,0xF2,0xFA,0xF3,0,0xF3,0xFB,0xF4,0,
    0xF4,0xFC,0xF5,0,0xF5,0xFD,0xF6,0,0xF6,0xFE,0xF7,0,0xF7,0xFF,0xF8,0,
    0xF8,0x00,0xF9,0,0xF9,0x01,0xFA,0,0xFA,0x02,0xFB,0,0xFB,0x03,0xFC,0,
    0xFC,0x04,0xFD,0,0xFD,0x05,0xFE,0,0xFE,0x06,0xFF,0,0xFF,0x07,0x00,0,
    0x00,0x08,0x01,0,0x01,0x09,0x02,0,0x02,0x0A,0x03,0,0x03,0x0B,0x04,0,
    0x04,0x0C,0x05,0,0x05,0x0D,0x06,0,0x06,0x0E,0x07,0,0x07,0x0F,0x08,0,
    0x08,0x10,0x09,0,0x09,0x11,0x0A,0,0x0A,0x12,0x0B,0,0x0B,0x13,0x0C,0,
    0x0C,0x14,0x0D,0,0x0D,0x15,0x0E,0,0x0E,0x16,0x0F,0,0x0F,0x17,0x10,0,
    0x10,0x18,0x11,0,0x11,0x19,0x12,0,0x12,0x1A,0x13,0,0x13,0x1B,0x14,0,
    0x14,0x1C,0x15,0,0x15,0x1D,0x16,0,0x16,0x1E,0x17,0,0x17,0x1F,0x18,0,
    0xD9,0xE0,0xD9,0,0xDA,0xE1,0xDA,0,0xDB,0xE2,0xDB,0,0xDC,0xE3,0xDC,0,
    0xDD,0xE4,0xDD,0,0xDE,0xE5,0xDE,0,0xDF,0xE6,0xDF,0,0xE0,0xE7,0xE0,0,
    0xE1,0xE8,0xE1,0,0xE2,0xE9,0xE2,0,0xE3,0xEA,0xE3,0,0xE4,0xEB,0xE4,0,
    0xE5,0xEC,0xE5,0,0xE6,0xED,0xE6,0,0xE7,0xEE,0xE7,0,0xE8,0xEF,0xE8,0,
    0xE9,0xF0,0xE9,0,0xEA,0xF1,0xEA,0,0xEB,0xF2,0xEB,0,0xEC,0xF3,0xEC,0,
    0xED,0xF4,0xED,0,0xEE,0xF5,0xEE,0,0xEF,0xF6,0xEF,0,0xF0,0xF7,0xF0,0,
    0xF1,0xF8,0xF1,0,0xF2,0xF9,0xF2,0,0xF3,0xFA,0xF3,0,0xF4,0xFB,0xF4,0,
    0xF5,0xFC,0xF5,0,0xF6,0xFD,0xF6,0,0xF7,0xFE,0xF7,0,0xF8,0xFF,0xF8,0,
    0xF9,0x00,0xF9,0,0xFA,0x01,0xFA,0,0xFB,0x02,0xFB,0,0xFC,0x03,0xFC,0,
    0xFD,0x04,0xFD,0,0xFE,0x05,0xFE,0,0xFF,0x06,0xFF,0,0x00,0x07,0x00,0,
    0x01,0x08,0x01,0,0x02,0x09,0x02,0,0x03,0x0A,0x03,0,0x04,0x0B,0x04,0,
    0x05,0x0C,0x05,0,0x06,0x0D,0x06,0,0x07,0x0E,0x07,0,0x08,0x0F,0x08,0,
    0x09,0x10,0x09,0,0x0A,0x11,0x0A,0,0x0B,0x12,0x0B,0,0x0C,0x13,0x0C,0,
    0x0D,0x14,0x0D,0,0x0E,0x15,0x0E,0,0x0F,0x16,0x0F,0,0x10,0x17,0x10,0,
    0x11,0x18,0x11,0,0x12,0x19,0x12,0,0x13,0x1A,0x13,0,0x14,0x1B,0x14,0,
    0x15,0x1C,0x15,0,0x16,0x1D,0x16,0,0x17,0x1E,0x17,0,0x18,0x1F,0x18,0,
    0xDA,0xE0,0xD9,0,0xDB,0xE1,0xDA,0,0xDC,0xE2,0xDB,0,0xDD,0xE3,0xDC,0,
    0xDE,0xE4,0xDD,0,0xDF,0xE5,0xDE,0,0xE0,0xE6,0xDF,0,0xE1,0xE7,0xE0,0,
    0xE2,0xE8,0xE1,0,0xE3,0xE9,0xE2,0,0xE4,0xEA,0xE3,0,0xE5,0xEB,0xE4,0,
    0xE6,0xEC,0xE5,0,0xE7,0xED,0xE6,0,0xE8,0xEE,0xE7,0,0xE9,0xEF,0xE8,0,
    0xEA,0xF0,0xE9,0,0xEB,0xF1,0xEA,0,0xEC,0xF2,0xEB,0,0xED,0xF3,0xEC,0,
    0xEE,0xF4,0xED,0,0xEF,0xF5,0xEE,0,0xF0,0xF6,0xEF,0,0xF1,0xF7,0xF0,0,
    0xF2,0xF8,0xF1,0,0xF3,0xF9,0xF2,0,0xF4,0xFA,0xF3,0,0xF5,0xFB,0xF4,0,
    0xF6,0xFC,0xF5,0,0xF7,0xFD,0xF6,0,0xF8,0xFE,0xF7,0,0xF9,0xFF,0xF8,0,
    0xFA,0x00,0xF9,0,0xFB,0x01,0xFA,0,0xFC,0x02,0xFB,0,0xFD,0x03,0xFC,0,
    0xFE,0x04,0xFD,0,0xFF,0x05,0xFE,0,0x00,0x06,0xFF,0,0x01,0x07,0x00,0,
    0x02,0x08,0x01,0,0x03,0x09,0x02,0,0x04,0x0A,0x03,0,0x05,0x0B,0x04,0,
    0x06,0x0C,0x05,0,0x07,0x0D,0x06,0,0x08,0x0E,0x07,0,0x09,0x0F,0x08,0,
    0x0A,0x10,0x09,0,0x0B,0x11,0x0A,0,0x0C,0x12,0x0B,0,0x0D,0x13,0x0C,0,
    0x0E,0x14,0x0D,0,0x0F,0x15,0x0E,0,0x10,0x16,0x0F,0,0x11,0x17,0x10,0,
    0x12,0x18,0x11,0,0x13,0x19,0x12,0,0x14,0x1A,0x13,0,0x15,0x1B,0x14,0,
    0x16,0x1C,0x15,0,0x17,0x1D,0x16,0,0x18,0x1E,0x17,0,0x19,0x1F,0x18,0,
    0xDB,0xE0,0xD9,0,0xDC,0xE1,0xDA,0,0xDD,0xE2,0xDB,0,0xDE,0xE3,0xDC,0,
    0xDF,0xE4,0xDD,0,0xE0,0xE5,0xDE,0,0xE1,0xE6,0xDF,0,0xE2,0xE7,0xE0,0,
    0xE3,0xE8,0xE1,0,0xE4,0xE9,0xE2,0,0xE5,0xEA,0xE3,0,0xE6,0xEB,0xE4,0,
    0xE7,0xEC,0xE5,0,0xE8,0xED,0xE6,0,0xE9,0xEE,0xE7,0,0xEA,0xEF,0xE8,0,
    0xEB,0xF0,0xE9,0,0xEC,0xF1,0xEA,0,0xED,0xF2,0xEB,0,0xEE,0xF3,0xEC,0,
    0xEF,0xF4,0xED,0,0xF0,0xF5,0xEE,0,0xF1,0xF6,0xEF,0,0xF2,0xF7,0xF0,0,
    0xF3,0xF8,0xF1,0,0xF4,0xF9,0xF2,0,0xF5,0xFA,0xF3,0,0xF6,0xFB,0xF4,0,
    0xF7,0xFC,0xF5,0,0xF8,0xFD,0xF6,0,0xF9,0xFE,0xF7,0,0xFA,0xFF,0xF8,0,
    0xFB,0x00,0xF9,0,0xFC,0x01,0xFA,0,0xFD,0x02,0xFB,0,0xFE,0x03,0xFC,0,
    0xFF,0x04,0xFD,0,0x00,0x05,0xFE,0,0x01,0x06,0xFF,0,0x02,0x07,0x00,0,
    0x03,0x08,0x01,0,0x04,0x09,0x02,0,0x05,0x0A,0x03,0,0x06,0x0B,0x04,0,
    0x07,0x0C,0x05,0,0x08,0x0D,0x06,0,0x09,0x0E,0x07,0,0x0A,0x0F,0x08,0,
    0x0B,0x10,0x09,0,0x0C,0x11,0x0A,0,0x0D,0x12,0x0B,0,0x0E,0x13,0x0C,0,
    0x0F,0x14,0x0D,0,0x10,0x15,0x0E,0,0x11,0x16,0x0F,0,0x12,0x17,0x10,0,
    0x13,0x18,0x11,0,0x14,0x19,0x12,0,0x15,0x1A,0x13,0,0x16,0x1B,0x14,0,
    0x17,0x1C,0x15,0,0x18,0x1D,0x16,0,0x19,0x1E,0x17,0,0x1A,0x1F,0x18,0,
    0xDC,0xE0,0xD9,0,0xDD,0xE1,0xDA,0,0xDE,0xE2,0xDB,0,0xDF,0xE3,0xDC,0,
    0xE0,0xE4,0xDD,0,0xE1,0xE5,0xDE,0,0xE2,0xE6,0xDF,0,0xE3,0xE7,0xE0,0,
    0xE4,0xE8,0xE1,0,0xE5,0xE9,0xE2,0,0xE6,0xEA,0xE3,0,0xE7,0xEB,0xE4,0,
    0xE8,0xEC,0xE5,0,0xE9,0xED,0xE6,0,0xEA,0xEE,0xE7,0,0xEB,0xEF,0xE8,0,
    0xEC,0xF0,0xE9,0,0xED,0xF1,0xEA,0,0xEE,0xF2,0xEB,0,0xEF,0xF3,0xEC,0,
    0xF0,0xF4,0xED,0,0xF1,0xF5,0xEE,0,0xF2,0xF6,0xEF,0,0xF3,0xF7,0xF0,0,
    0xF4,0xF8,0xF1,0,0xF5,0xF9,0xF2,0,0xF6,0xFA,0xF3,0,0xF7,0xFB,0xF4,0,
    0xF8,0xFC,0xF5,0,0xF9,0xFD,0xF6,0,0xFA,0xFE,0xF7,0,0xFB,0xFF,0xF8,0,
    0xFC,0x00,0xF9,0,0xFD,0x01,0xFA,0,0xFE,0x02,0xFB,0,0xFF,0x03,0xFC,0,
    0x00,0x04,0xFD,0,0x01,0x05,0xFE,0,0x02,0x06,0xFF,0,0x03,0x07,0x00,0,
    0x04,0x08,0x01,0,0x05,0x09,0x02,0,0x06,0x0A,0x03,0,0x07,0x0B,0x04,0,
    0x08,0x0C,0x05,0,0x09,0x0D,0x06,0,0x0A,0x0E,0x07,0,0x0B,0x0F,0x08,0,
    0x0C,0x10,0x09,0,0x0D,0x11,0x0A,0,0x0E,0x12,0x0B,0,0x0F,0x13,0x0C,0,
    0x10,0x14,0x0D,0,0x11,0x15,0x0E,0,0x12,0x16,0x0F,0,0x13,0x17,0x10,0,
    0x14,0x18,0x11,0,0x15,0x19,0x12,0,0x16,0x1A,0x13,0,0x17,0x1B,0x14,0,
    0x18,0x1C,0x15,0,0x19,0x1D,0x16,0,0x1A,0x1E,0x17,0,0x1B,0x1F,0x18,0,
    0xDD,0xE0,0xD9,0,0xDE,0xE1,0xDA,0,0xDF,0xE2,0xDB,0,0xE0,0xE3,0xDC,0,
    0xE1,0xE4,0xDD,0,0xE2,0xE5,0xDE,0,0xE3,0xE6,0xDF,0,0xE4,0xE7,0xE0,0,
    0xE5,0xE8,0xE1,0,0xE6,0xE9,0xE2,0,0xE7,0xEA,0xE3,0,0xE8,0xEB,0xE4,0,
    0xE9,0xEC,0xE5,0,0xEA,0xED,0xE6,0,0xEB,0xEE,0xE7,0,0xEC,0xEF,0xE8,0,
    0xED,0xF0,0xE9,0,0xEE,0xF1,0xEA,0,0xEF,0xF2,0xEB,0,0xF0,0xF3,0xEC,0,
    0xF1,0xF4,0xED,0,0xF2,0xF5,0xEE,0,0xF3,0xF6,0xEF,0,0xF4,0xF7,0xF0,0,
    0xF5,0xF8,0xF1,0,0xF6,0xF9,0xF2,0,0xF7,0xFA,0xF3,0,0xF8,0xFB,0xF4,0,
    0xF9,0xFC,0xF5,0,0xFA,0xFD,0xF6,0,0xFB,0xFE,0xF7,0,0xFC,0xFF,0xF8,0,
    0xFD,0x00,0xF9,0,0xFE,0x01,0xFA,0,0xFF,0x02,0xFB,0,0x00,0x03,0xFC,0,
    0x01,0x04,0xFD,0,0x02,0x05,0xFE,0,0x03,0x06,0xFF,0,0x04,0x07,0x00,0,
    0x05,0x08,0x01,0,0x06,0x09,0x02,0,0x07,0x0A,0x03,0,0x08,0x0B,0x04,0,
    0x09,0x0C,0x05,0,0x0A,0x0D,0x06,0,0x0B,0x0E,0x07,0,0x0C,0x0F,0x08,0,
    0x0D,0x10,0x09,0,0x0E,0x11,0x0A,0,0x0F,0x12,0x0B,0,0x10,0x13,0x0C,0,
    0x11,0x14,0x0D,0,0x12,0x15,0x0E,0,0x13,0x16,0x0F,0,0x14,0x17,0x10,0,
    0x15,0x18,0x11,0,0x16,0x19,0x12,0,0x17,0x1A,0x13,0,0x18,0x1B,0x14,0,
    0x19,0x1C,0x15,0,0x1A,0x1D,0x16,0,0x1B,0x1E,0x17,0,0x1C,0x1F,0x18,0,
    0xDE,0xE0,0xD9,0,0xDF,0xE1,0xDA,0,0xE0,0xE2,0xDB,0,0xE1,0xE3,0xDC,0,
    0xE2,0xE4,0xDD,0,0xE3,0xE5,0xDE,0,0xE4,0xE6,0xDF,0,0xE5,0xE7,0xE0,0,
    0xE6,0xE8,0xE1,0,0xE7,0xE9,0xE2,0,0xE8,0xEA,0xE3,0,0xE9,0xEB,0xE4,0,
    0xEA,0xEC,0xE5,0,0xEB,0xED,0xE6,0,0xEC,0xEE,0xE7,0,0xED,0xEF,0xE8,0,
    0xEE,0xF0,0xE9,0,0xEF,0xF1,0xEA,0,0xF0,0xF2,0xEB,0,0xF1,0xF3,0xEC,0,
    0xF2,0xF4,0xED,0,0xF3,0xF5,0xEE,0,0xF4,0xF6,0xEF,0,0xF5,0xF7,0xF0,0,
    0xF6,0xF8,0xF1,0,0xF7,0xF9,0xF2,0,0xF8,0xFA,0xF3,0,0xF9,0xFB,0xF4,0,
    0xFA,0xFC,0xF5,0,0xFB,0xFD,0xF6,0,0xFC,0xFE,0xF7,0,0xFD,0xFF,0xF8,0,
    0xFE,0x00,0xF9,0,0xFF,0x01,0xFA,0,0x00,0x02,0xFB,0,0x01,0x03,0xFC,0,
    0x02,0x04,0xFD,0,0x03,0x05,0xFE,0,0x04,0x06,0xFF,0,0x05,0x07,0x00,0,
    0x06,0x08,0x01,0,0x07,0x09,0x02,0,0x08,0x0A,0x03,0,0x09,0x0B,0x04,0,
    0x0A,0x0C,0x05,0,0x0B,0x0D,0x06,0,0x0C,0x0E,0x07,0,0x0D,0x0F,0x08,0,
    0x0E,0x10,0x09,0,0x0F,0x11,0x0A,0,0x10,0x12,0x0B,0,0x11,0x13,0x0C,0,
    0x12,0x14,0x0D,0,0x13,0x15,0x0E,0,0x14,0x16,0x0F,0,0x15,0x17,0x10,0,
    0x16,0x18,0x11,0,0x17,0x19,0x12,0,0x18,0x1A,0x13,0,0x19,0x1B,0x14,0,
    0x1A,0x1C,0x15,0,0x1B,0x1D,0x16,0,0x1C,0x1E,0x17,0,0x1D,0x1F,0x18,0,
    0xDF,0xE0,0xD9,0,0xE0,0xE1,0xDA,0,0xE1,0xE2,0xDB,0,0xE2,0xE3,0xDC,0,
    0xE3,0xE4,0xDD,0,0xE4,0xE5,0xDE,0,0xE5,0xE6,0xDF,0,0xE6,0xE7,0xE0,0,
    0xE7,0xE8,0xE1,0,0xE8,0xE9,0xE2,0,0xE9,0xEA,0xE3,0,0xEA,0xEB,0xE4,0,
    0xEB,0xEC,0xE5,0,0xEC,0xED,0xE6,0,0xED,0xEE,0xE7,0,0xEE,0xEF,0xE8,0,
    0xEF,0xF0,0xE9,0,0xF0,0xF1,0xEA,0,0xF1,0xF2,0xEB,0,0xF2,0xF3,0xEC,0,
    0xF3,0xF4,0xED,0,0xF4,0xF5,0xEE,0,0xF5,0xF6,0xEF,0,0xF6,0xF7,0xF0,0,
    0xF7,0xF8,0xF1,0,0xF8,0xF9,0xF2,0,0xF9,0xFA,0xF3,0,0xFA,0xFB,0xF4,0,
    0xFB,0xFC,0xF5,0,0xFC,0xFD,0xF6,0,0xFD,0xFE,0xF7,0,0xFE,0xFF,0xF8,0,
    0xFF,0x00,0xF9,0,0x00,0x01,0xFA,0,0x01,0x02,0xFB,0,0x02,0x03,0xFC,0,
    0x03,0x04,0xFD,0,0x04,0x05,0xFE,0,0x05,0x06,0xFF,0,0x06,0x07,0x00,0,
    0x07,0x08,0x01,0,0x08,0x09,0x02,0,0x09,0x0A,0x03,0,0x0A,0x0B,0x04,0,
    0x0B,0x0C,0x05,0,0x0C,0x0D,0x06,0,0x0D,0x0E,0x07,0,0x0E,0x0F,0x08,0,
    0x0F,0x10,0x09,0,0x10,0x11,0x0A,0,0x11,0x12,0x0B,0,0x12,0x13,0x0C,0,
    0x13,0x14,0x0D,0,0x14,0x15,0x0E,0,0x15,0x16,0x0F,0,0x16,0x17,0x10,0,
    0x17,0x18,0x11,0,0x18,0x19,0x12,0,0x19,0x1A,0x13,0,0x1A,0x1B,0x14,0,
    0x1B,0x1C,0x15,0,0x1C,0x1D,0x16,0,0x1D,0x1E,0x17,0,0x1E,0x1F,0x18,0,
    0xE0,0xE0,0xD9,0,0xE1,0xE1,0xDA,0,0xE2,0xE2,0xDB,0,0xE3,0xE3,0xDC,0,
    0xE4,0xE4,0xDD,0,0xE5,0xE5,0xDE,0,0xE6,0xE6,0xDF,0,0xE7,0xE7,0xE0,0,
    0xE8,0xE8,0xE1,0,0xE9,0xE9,0xE2,0,0xEA,0xEA,0xE3,0,0xEB,0xEB,0xE4,0,
    0xEC,0xEC,0xE5,0,0xED,0xED,0xE6,0,0xEE,0xEE,0xE7,0,0xEF,0xEF,0xE8,0,
    0xF0,0xF0,0xE9,0,0xF1,0xF1,0xEA,0,0xF2,0xF2,0xEB,0,0xF3,0xF3,0xEC,0,
    0xF4,0xF4,0xED,0,0xF5,0xF5,0xEE,0,0xF6,0xF6,0xEF,0,0xF7,0xF7,0xF0,0,
    0xF8,0xF8,0xF1,0,0xF9,0xF9,0xF2,0,0xFA,0xFA,0xF3,0,0xFB,0xFB,0xF4,0,
    0xFC,0xFC,0xF5,0,0xFD,0xFD,0xF6,0,0xFE,0xFE,0xF7,0,0xFF,0xFF,0xF8,0,
    0x00,0x00,0xF9,0,0x01,0x01,0xFA,0,0x02,0x02,0xFB,0,0x03,0x03,0xFC,0,
    0x04,0x04,0xFD,0,0x05,0x05,0xFE,0,0x06,0x06,0xFF,0,0x07,0x07,0x00,0,
    0x08,0x08,0x01,0,0x09,0x09,0x02,0,0x0A,0x0A,0x03,0,0x0B,0x0B,0x04,0,
    0x0C,0x0C,0x05,0,0x0D,0x0D,0x06,0,0x0E,0x0E,0x07,0,0x0F,0x0F,0x08,0,
    0x10,0x10,0x09,0,0x11,0x11,0x0A,0,0x12,0x12,0x0B,0,0x13,0x13,0x0C,0,
    0x14,0x14,0x0D,0,0x15,0x15,0x0E,0,0x16,0x16,0x0F,0,0x17,0x17,0x10,0,
    0x18,0x18,0x11,0,0x19,0x19,0x12,0,0x1A,0x1A,0x13,0,0x1B,0x1B,0x14,0,
    0x1C,0x1C,0x15,0,0x1D,0x1D,0x16,0,0x1E,0x1E,0x17,0,0x1F,0x1F,0x18,0,
    0xE1,0xE0,0xD9,0,0xE2,0xE1,0xDA,0,0xE3,0xE2,0xDB,0,0xE4,0xE3,0xDC,0,
    0xE5,0xE4,0xDD,0,0xE6,0xE5,0xDE,0,0xE7,0xE6,0xDF,0,0xE8,0xE7,0xE0,0,
    0xE9,0xE8,0xE1,0,0xEA,0xE9,0xE2,0,0xEB,0xEA,0xE3,0,0xEC,0xEB,0xE4,0,
    0xED,0xEC,0xE5,0,0xEE,0xED,0xE6,0,0xEF,0xEE,0xE7,0,0xF0,0xEF,0xE8,0,
    0xF1,0xF0,0xE9,0,0xF2,0xF1,0xEA,0,0xF3,0xF2,0xEB,0,0xF4,0xF3,0xEC,0,
    0xF5,0xF4,0xED,0,0xF6,0xF5,0xEE,0,0xF7,0xF6,0xEF,0,0xF8,0xF7,0xF0,0,
    0xF9,0xF8,0xF1,0,0xFA,0xF9,0xF2,0,0xFB,0xFA,0xF3,0,0xFC,0xFB,0xF4,0,
    0xFD,0xFC,0xF5,0,0xFE,0xFD,0xF6,0,0xFF,0xFE,0xF7,0,0x00,0xFF,0xF8,0,
    0x01,0x00,0xF9,0,0x02,0x01,0xFA,0,0x03,0x02,0xFB,0,0x04,0x03,0xFC,0,
    0x05,0x04,0xFD,0,0x06,0x05,0xFE,0,0x07,0x06,0xFF,0,0x08,0x07,0x00,0,
    0x09,0x08,0x01,0,0x0A,0x09,0x02,0,0x0B,0x0A,0x03,0,0x0C,0x0B,0x04,0,
    0x0D,0x0C,0x05,0,0x0E,0x0D,0x06,0,0x0F,0x0E,0x07,0,0x10,0x0F,0x08,0,
    0x11,0x10,0x09,0,0x12,0x11,0x0A,0,0x13,0x12,0x0B,0,0x14,0x13,0x0C,0,
    0x15,0x14,0x0D,0,0x16,0x15,0x0E,0,0x17,0x16,0x0F,0,0x18,0x17,0x10,0,
    0x19,0x18,0x11,0,0x1A,0x19,0x12,0,0x1B,0x1A,0x13,0,0x1C,0x1B,0x14,0,
    0x1D,0x1C,0x15,0,0x1E,0x1D,0x16,0,0x1F,0x1E,0x17,0,0x20,0x1F,0x18,0,
    0xE2,0xE0,0xD9,0,0xE3,0xE1,0xDA,0,0xE4,0xE2,0xDB,0,0xE5,0xE3,0xDC,0,
    0xE6,0xE4,0xDD,0,0xE7,0xE5,0xDE,0,0xE8,0xE6,0xDF,0,0xE9,0xE7,0xE0,0,
    0xEA,0xE8,0xE1,0,0xEB,0xE9,0xE2,0,0xEC,0xEA,0xE3,0,0xED,0xEB,0xE4,0,
    0xEE,0xEC,0xE5,0,0xEF,0xED,0xE6,0,0xF0,0xEE,0xE7,0,0xF1,0xEF,0xE8,0,
    0xF2,0xF0,0xE9,0,0xF3,0xF1,0xEA,0,0xF4,0xF2,0xEB,0,0xF5,0xF3,0xEC,0,
    0xF6,0xF4,0xED,0,0xF7,0xF5,0xEE,0,0xF8,0xF6,0xEF,0,0xF9,0xF7,0xF0,0,
    0xFA,0xF8,0xF1,0,0xFB,0xF9,0xF2,0,0xFC,0xFA,0xF3,0,0xFD,0xFB,0xF4,0,
    0xFE,0xFC,0xF5,0,0xFF,0xFD,0xF6,0,0x00,0xFE,0xF7,0,0x01,0xFF,0xF8,0,
    0x02,0x00,0xF9,0,0x03,0x01,0xFA,0,0x04,0x02,0xFB,0,0x05,0x03,0xFC,0,
    0x06,0x04,0xFD,0,0x07,0x05,0xFE,0,0x08,0x06,0xFF,0,0x09,0x07,0x00,0,
    0x0A,0x08,0x01,0,0x0B,0x09,0x02,0,0x0C,0x0A,0x03,0,0x0D,0x0B,0x04,0,
    0x0E,0x0C,0x05,0,0x0F,0x0D,0x06,0,0x10,0x0E,0x07,0,0x11,0x0F,0x08,0,
    0x12,0x10,0x09,0,0x13,0x11,0x0A,0,0x14,0x12,0x0B,0,0x15,0x13,0x0C,0,
    0x16,0x14,0x0D,0,0x17,0x15,0x0E,0,0x18,0x16,0x0F,0,0x19,0x17,0x10,0,
    0x1A,0x18,0x11,0,0x1B,0x19,0x12,0,0x1C,0x1A,0x13,0,0x1D,0x1B,0x14,0,
    0x1E,0x1C,0x15,0,0x1F,0x1D,0x16,0,0x20,0x1E,0x17,0,0x21,0x1F,0x18,0,
    0xE3,0xE0,0xD9,0,0xE4,0xE1,0xDA,0,0xE5,0xE2,0xDB,0,0xE6,0xE3,0xDC,0,
    0xE7,0xE4,0xDD,0,0xE8,0xE5,0xDE,0,0xE9,0xE6,0xDF,0,0xEA,0xE7,0xE0,0,
    0xEB,0xE8,0xE1,0,0xEC,0xE9,0xE2,0,0xED,0xEA,0xE3,0,0xEE,0xEB,0xE4,0,
    0xEF,0xEC,0xE5,0,0xF0,0xED,0xE6,0,0xF1,0xEE,0xE7,0,0xF2,0xEF,0xE8,0,
    0xF3,0xF0,0xE9,0,0xF4,0xF1,0xEA,0,0xF5,0xF2,0xEB,0,0xF6,0xF3,0xEC,0,
    0xF7,0xF4,0xED,0,0xF8,0xF5,0xEE,0,0xF9,0xF6,0xEF,0,0xFA,0xF7,0xF0,0,
    0xFB,0xF8,0xF1,0,0xFC,0xF9,0xF2,0,0xFD,0xFA,0xF3,0,0xFE,0xFB,0xF4,0,
    0xFF,0xFC,0xF5,0,0x00,0xFD,0xF6,0,0x01,0xFE,0xF7,0,0x02,0xFF,0xF8,0,
    0x03,0x00,0xF9,0,0x04,0x01,0xFA,0,0x05,0x02,0xFB,0,0x06,0x03,0xFC,0,
    0x07,0x04,0xFD,0,0x08,0x05,0xFE,0,0x09,0x06,0xFF,0,0x0A,0x07,0x00,0,
    0x0B,0x08,0x01,0,0x0C,0x09,0x02,0,0x0D,0x0A,0x03,0,0x0E,0x0B,0x04,0,
    0x0F,0x0C,0x05,0,0x10,0x0D,0x06,0,0x11,0x0E,0x07,0,0x12,0x0F,0x08,0,
    0x13,0x10,0x09,0,0x14,0x11,0x0A,0,0x15,0x12,0x0B,0,0x16,0x13,0x0C,0,
    0x17,0x14,0x0D,0,0x18,0x15,0x0E,0,0x19,0x16,0x0F,0,0x1A,0x17,0x10,0,
    0x1B,0x18,0x11,0,0x1C,0x19,0x12,0,0x1D,0x1A,0x13,0,0x1E,0x1B,0x14,0,
    0x1F,0x1C,0x15,0,0x20,0x1D,0x16,0,0x21,0x1E,0x17,0,0x22,0x1F,0x18,0,
    0xE4,0xE0,0xD9,0,0xE5,0xE1,0xDA,0,0xE6,0xE2,0xDB,0,0xE7,0xE3,0xDC,0,
    0xE8,0xE4,0xDD,0,0xE9,0xE5,0xDE,0,0xEA,0xE6,0xDF,0,0xEB,0xE7,0xE0,0,
    0xEC,0xE8,0xE1,0,0xED,0xE9,0xE2,0,0xEE,0xEA,0xE3,0,0xEF,0xEB,0xE4,0,
    0xF0,0xEC,0xE5,0,0xF1,0xED,0xE6,0,0xF2,0xEE,0xE7,0,0xF3,0xEF,0xE8,0,
    0xF4,0xF0,0xE9,0,0xF5,0xF1,0xEA,0,0xF6,0xF2,0xEB,0,0xF7,0xF3,0xEC,0,
    0xF8,0xF4,0xED,0,0xF9,0xF5,0xEE,0,0xFA,0xF6,0xEF,0,0xFB,0xF7,0xF0,0,
    0xFC,0xF8,0xF1,0,0xFD,0xF9,0xF2,0,0xFE,0xFA,0xF3,0,0xFF,0xFB,0xF4,0,
    0x00,0xFC,0xF5,0,0x01,0xFD,0xF6,0,0x02,0xFE,0xF7,0,0x03,0xFF,0xF8,0,
    0x04,0x00,0xF9,0,0x05,0x01,0xFA,0,0x06,0x02,0xFB,0,0x07,0x03,0xFC,0,
    0x08,0x04,0xFD,0,0x09,0x05,0xFE,0,0x0A,0x06,0xFF,0,0x0B,0x07,0x00,0,
    0x0C,0x08,0x01,0,0x0D,0x09,0x02,0,0x0E,0x0A,0x03,0,0x0F,0x0B,0x04,0,
    0x10,0x0C,0x05,0,0x11,0x0D,0x06,0,0x12,0x0E,0x07,0,0x13,0x0F,0x08,0,
    0x14,0x10,0x09,0,0x15,0x11,0x0A,0,0x16,0x12,0x0B,0,0x17,0x13,0x0C,0,
    0x18,0x14,0x0D,0,0x19,0x15,0x0E,0,0x1A,0x16,0x0F,0,0x1B,0x17,0x10,0,
    0x1C,0x18,0x11,0,0x1D,0x19,0x12,0,0x1E,0x1A,0x13,0,0x1F,0x1B,0x14,0,
    0x20,0x1C,0x15,0,0x21,0x1D,0x16,0,0x22,0x1E,0x17,0,0x23,0x1F,0x18,0,
    0xE5,0xE0,0xD9,0,0xE6,0xE1,0xDA,0,0xE7,0xE2,0xDB,0,0xE8,0xE3,0xDC,0,
    0xE9,0xE4,0xDD,0,0xEA,0xE5,0xDE,0,0xEB,0xE6,0xDF,0,0xEC,0xE7,0xE0,0,
    0xED,0xE8,0xE1,0,0xEE,0xE9,0xE2,0,0xEF,0xEA,0xE3,0,0xF0,0xEB,0xE4,0,
    0xF1,0xEC,0xE5,0,0xF2,0xED,0xE6,0,0xF3,0xEE,0xE7,0,0xF4,0xEF,0xE8,0,
    0xF5,0xF0,0xE9,0,0xF6,0xF1,0xEA,0,0xF7,0xF2,0xEB,0,0xF8,0xF3,0xEC,0,
    0xF9,0xF4,0xED,0,0xFA,0xF5,0xEE,0,0xFB,0xF6,0xEF,0,0xFC,0xF7,0xF0,0,
    0xFD,0xF8,0xF1,0,0xFE,0xF9,0xF2,0,0xFF,0xFA,0xF3,0,0x00,0xFB,0xF4,0,
    0x01,0xFC,0xF5,0,0x02,0xFD,0xF6,0,0x03,0xFE,0xF7,0,0x04,0xFF,0xF8,0,
    0x05,0x00,0xF9,0,0x06,0x01,0xFA,0,0x07,0x02,0xFB,0,0x08,0x03,0xFC,0,
    0x09,0x04,0xFD,0,0x0A,0x05,0xFE,0,0x0B,0x06,0xFF,0,0x0C,0x07,0x00,0,
    0x0D,0x08,0x01,0,0x0E,0x09,0x02,0,0x0F,0x0A,0x03,0,0x10,0x0B,0x04,0,
    0x11,0x0C,0x05,0,0x12,0x0D,0x06,0,0x13,0x0E,0x07,0,0x14,0x0F,0x08,0,
    0x15,0x10,0x09,0,0x16,0x11,0x0A,0,0x17,0x12,0x0B,0,0x18,0x13,0x0C,0,
    0x19,0x14,0x0D,0,0x1A,0x15,0x0E,0,0x1B,0x16,0x0F,0,0x1C,0x17,0x10,0,
    0x1D,0x18,0x11,0,0x1E,0x19,0x12,0,0x1F,0x1A,0x13,0,0x20,0x1B,0x14,0,
    0x21,0x1C,0x15,0,0x22,0x1D,0x16,0,0x23,0x1E,0x17,0,0x24,0x1F,0x18,0,
    0xE6,0xE0,0xD9,0,0xE7,0xE1,0xDA,0,0xE8,0xE2,0xDB,0,0xE9,0xE3,0xDC,0,
    0xEA,0xE4,0xDD,0,0xEB,0xE5,0xDE,0,0xEC,0xE6,0xDF,0,0xED,0xE7,0xE0,0,
    0xEE,0xE8,0xE1,0,0xEF,0xE9,0xE2,0,0xF0,0xEA,0xE3,0,0xF1,0xEB,0xE4,0,
    0xF2,0xEC,0xE5,0,0xF3,0xED,0xE6,0,0xF4,0xEE,0xE7,0,0xF5,0xEF,0xE8,0,
    0xF6,0xF0,0xE9,0,0xF7,0xF1,0xEA,0,0xF8,0xF2,0xEB,0,0xF9,0xF3,0xEC,0,
    0xFA,0xF4,0xED,0,0xFB,0xF5,0xEE,0,0xFC,0xF6,0xEF,0,0xFD,0xF7,0xF0,0,
    0xFE,0xF8,0xF1,0,0xFF,0xF9,0xF2,0,0x00,0xFA,0xF3,0,0x01,0xFB,0xF4,0,
    0x02,0xFC,0xF5,0,0x03,0xFD,0xF6,0,0x04,0xFE,0xF7,0,0x05,0xFF,0xF8,0,
    0x06,0x00,0xF9,0,0x07,0x01,0xFA,0,0x08,0x02,0xFB,0,0x09,0x03,0xFC,0,
    0x0A,0x04,0xFD,0,0x0B,0x05,0xFE,0,0x0C,0x06,0xFF,0,0x0D,0x07,0x00,0,
    0x0E,0x08,0x01,0,0x0F,0x09,0x02,0,0x10,0x0A,0x03,0,0x11,0x0B,0x04,0,
    0x12,0x0C,0x05,0,0x13,0x0D,0x06,0,0x14,0x0E,0x07,0,0x15,0x0F,0x08,0,
    0x16,0x10,0x09,0,0x17,0x11,0x0A,0,0x18,0x12,0x0B,0,0x19,0x13,0x0C,0,
    0x1A,0x14,0x0D,0,0x1B,0x15,0x0E,0,0x1C,0x16,0x0F,0,0x1D,0x17,0x10,0,
    0x1E,0x18,0x11,0,0x1F,0x19,0x12,0,0x20,0x1A,0x13,0,0x21,0x1B,0x14,0,
    0x22,0x1C,0x15,0,0x23,0x1D,0x16,0,0x24,0x1E,0x17,0,0x25,0x1F,0x18,0,
    0xE7,0xE0,0xD9,0,0xE8,0xE1,0xDA,0,0xE9,0xE2,0xDB,0,0xEA,0xE3,0xDC,0,
    0xEB,0xE4,0xDD,0,0xEC,0xE5,0xDE,0,0xED,0xE6,0xDF,0,0xEE,0xE7,0xE0,0,
    0xEF,0xE8,0xE1,0,0xF0,0xE9,0xE2,0,0xF1,0xEA,0xE3,0,0xF2,0xEB,0xE4,0,
    0xF3,0xEC,0xE5,0,0xF4,0xED,0xE6,0,0xF5,0xEE,0xE7,0,0xF6,0xEF,0xE8,0,
    0xF7,0xF0,0xE9,0,0xF8,0xF1,0xEA,0,0xF9,0xF2,0xEB,0,0xFA,0xF3,0xEC,0,
    0xFB,0xF4,0xED,0,0xFC,0xF5,0xEE,0,0xFD,0xF6,0xEF,0,0xFE,0xF7,0xF0,0,
    0xFF,0xF8,0xF1,0,0x00,0xF9,0xF2,0,0x01,0xFA,0xF3,0,0x02,0xFB,0xF4,0,
    0x03,0xFC,0xF5,0,0x04,0xFD,0xF6,0,0x05,0xFE,0xF7,0,0x06,0xFF,0xF8,0,
    0x07,0x00,0xF9,0,0x08,0x01,0xFA,0,0x09,0x02,0xFB,0,0x0A,0x03,0xFC,0,
    0x0B,0x04,0xFD,0,0x0C,0x05,0xFE,0,0x0D,0x06,0xFF,0,0x0E,0x07,0x00,0,
    0x0F,0x08,0x01,0,0x10,0x09,0x02,0,0x11,0x0A,0x03,0,0x12,0x0B,0x04,0,
    0x13,0x0C,0x05,0,0x14,0x0D,0x06,0,0x15,0x0E,0x07,0,0x16,0x0F,0x08,0,
    0x17,0x10,0x09,0,0x18,0x11,0x0A,0,0x19,0x12,0x0B,0,0x1A,0x13,0x0C,0,
    0x1B,0x14,0x0D,0,0x1C,0x15,0x0E,0,0x1D,0x16,0x0F,0,0x1E,0x17,0x10,0,
    0x1F,0x18,0x11,0,0x20,0x19,0x12,0,0x21,0x1A,0x13,0,0x22,0x1B,0x14,0,
    0x23,0x1C,0x15,0,0x24,0x1D,0x16,0,0x25,0x1E,0x17,0,0x26,0x1F,0x18,0,
    0xD8,0xE0,0xDA,0,0xD9,0xE1,0xDB,0,0xDA,0xE2,0xDC,0,0xDB,0xE3,0xDD,0,
    0xDC,0xE4,0xDE,0,0xDD,0xE5,0xDF,0,0xDE,0xE6,0xE0,0,0xDF,0xE7,0xE1,0,
    0xE0,0xE8,0xE2,0,0xE1,0xE9,0xE3,0,0xE2,0xEA,0xE4,0,0xE3,0xEB,0xE5,0,
    0xE4,0xEC,0xE6,0,0xE5,0xED,0xE7,0,0xE6,0xEE,0xE8,0,0xE7,0xEF,0xE9,0,
    0xE8,0xF0,0xEA,0,0xE9,0xF1,0xEB,0,0xEA,0xF2,0xEC,0,0xEB,0xF3,0xED,0,
    0xEC,0xF4,0xEE,0,0xED,0xF5,0xEF,0,0xEE,0xF6,0xF0,0,0xEF,0xF7,0xF1,0,
    0xF0,0xF8,0xF2,0,0xF1,0xF9,0xF3,0,0xF2,0xFA,0xF4,0,0xF3,0xFB,0xF5,0,
    0xF4,0xFC,0xF6,0,0xF5,0xFD,0xF7,0,0xF6,0xFE,0xF8,0,0xF7,0xFF,0xF9,0,
    0xF8,0x00,0xFA,0,0xF9,0x01,0xFB,0,0xFA,0x02,0xFC,0,0xFB,0x03,0xFD,0,
    0xFC,0x04,0xFE,0,0xFD,0x05,0xFF,0,0xFE,0x06,0x00,0,0xFF,0x07,0x01,0,
    0x00,0x08,0x02,0,0x01,0x09,0x03,0,0x02,0x0A,0x04,0,0x03,0x0B,0x05,0,
    0x04,0x0C,0x06,0,0x05,0x0D,0x07,0,0x06,0x0E,0x08,0,0x07,0x0F,0x09,0,
    0x08,0x10,0x0A,0,0x09,0x11,0x0B,0,0x0A,0x12,0x0C,0,0x0B,0x13,0x0D,0,
    0x0C,0x14,0x0E,0,0x0D,0x15,0x0F,0,0x0E,0x16,0x10,0,0x0F,0x17,0x11,0,
    0x10,0x18,0x12,0,0x11,0x19,0x13,0,0x12,0x1A,0x14,0,0x13,0x1B,0x15,0,
    0x14,0x1C,0x16,0,0x15,0x1D,0x17,0,0x16,0x1E,0x18,0,0x17,0x1F,0x19,0,
    0xD9,0xE0,0xDA,0,0xDA,0xE1,0xDB,0,0xDB,0xE2,0xDC,0,0xDC,0xE3,0xDD,0,
    0xDD,0xE4,0xDE,0,0xDE,0xE5,0xDF,0,0xDF,0xE6,0xE0,0,0xE0,0xE7,0xE1,0,
    0xE1,0xE8,0xE2,0,0xE2,0xE9,0xE3,0,0xE3,0xEA,0xE4,0,0xE4,0xEB,0xE5,0,
    0xE5,0xEC,0xE6,0,0xE6,0xED,0xE7,0,0xE7,0xEE,0xE8,0,0xE8,0xEF,0xE9,0,
    0xE9,0xF0,0xEA,0,0xEA,0xF1,0xEB,0,0xEB,0xF2,0xEC,0,0xEC,0xF3,0xED,0,
    0xED,0xF4,0xEE,0,0xEE,0xF5,0xEF,0,0xEF,0xF6,0xF0,0,0xF0,0xF7,0xF1,0,
    0xF1,0xF8,0xF2,0,0xF2,0xF9,0xF3,0,0xF3,0xFA,0xF4,0,0xF4,0xFB,0xF5,0,
    0xF5,0xFC,0xF6,0,0xF6,0xFD,0xF7,0,0xF7,0xFE,0xF8,0,0xF8,0xFF,0xF9,0,
    0xF9,0x00,0xFA,0,0xFA,0x01,0xFB,0,0xFB,0x02,0xFC,0,0xFC,0x03,0xFD,0,
    0xFD,0x04,0xFE,0,0xFE,0x05,0xFF,0,0xFF,0x06,0x00,0,0x00,0x07,0x01,0,
    0x01,0x08,0x02,0,0x02,0x09,0x03,0,0x03,0x0A,0x04,0,0x04,0x0B,0x05,0,
    0x05,0x0C,0x06,0,0x06,0x0D,0x07,0,0x07,0x0E,0x08,0,0x08,0x0F,0x09,0,
    0x09,0x10,0x0A,0,0x0A,0x11,0x0B,0,0x0B,0x12,0x0C,0,0x0C,0x13,0x0D,0,
    0x0D,0x14,0x0E,0,0x0E,0x15,0x0F,0,0x0F,0x16,0x10,0,0x10,0x17,0x11,0,
    0x11,0x18,0x12,0,0x12,0x19,0x13,0,0x13,0x1A,0x14,0,0x14,0x1B,0x15,0,
    0x15,0x1C,0x16,0,0x16,0x1D,0x17,0,0x17,0x1E,0x18,0,0x18,0x1F,0x19,0,
    0xDA,0xE0,0xDA,0,0xDB,0xE1,0xDB,0,0xDC,0xE2,0xDC,0,0xDD,0xE3,0xDD,0,
    0xDE,0xE4,0xDE,0,0xDF,0xE5,0xDF,0,0xE0,0xE6,0xE0,0,0xE1,0xE7,0xE1,0,
    0xE2,0xE8,0xE2,0,0xE3,0xE9,0xE3,0,0xE4,0xEA,0xE4,0,0xE5,0xEB,0xE5,0,
    0xE6,0xEC,0xE6,0,0xE7,0xED,0xE7,0,0xE8,0xEE,0xE8,0,0xE9,0xEF,0xE9,0,
    0xEA,0xF0,0xEA,0,0xEB,0xF1,0xEB,0,0xEC,0xF2,0xEC,0,0xED,0xF3,0xED,0,
    0xEE,0xF4,0xEE,0,0xEF,0xF5,0xEF,0,0xF0,0xF6,0xF0,0,0xF1,0xF7,0xF1,0,
    0xF2,0xF8,0xF2,0,0xF3,0xF9,0xF3,0,0xF4,0xFA,0xF4,0,0xF5,0xFB,0xF5,0,
    0xF6,0xFC,0xF6,0,0xF7,0xFD,0xF7,0,0xF8,0xFE,0xF8,0,0xF9,0xFF,0xF9,0,
    0xFA,0x00,0xFA,0,0xFB,0x01,0xFB,0,0xFC,0x02,0xFC,0,0xFD,0x03,0xFD,0,
    0xFE,0x04,0xFE,0,0xFF,0x05,0xFF,0,0x00,0x06,0x00,0,0x01,0x07,0x01,0,
    0x02,0x08,0x02,0,0x03,0x09,0x03,0,0x04,0x0A,0x04,0,0x05,0x0B,0x05,0,
    0x06,0x0C,0x06,0,0x07,0x0D,0x07,0,0x08,0x0E,0x08,0,0x09,0x0F,0x09,0,
    0x0A,0x10,0x0A,0,0x0B,0x11,0x0B,0,0x0C,0x12,0x0C,0,0x0D,0x13,0x0D,0,
    0x0E,0x14,0x0E,0,0x0F,0x15,0x0F,0,0x10,0x16,0x10,0,0x11,0x17,0x11,0,
    0x12,0x18,0x12,0,0x13,0x19,0x13,0,0x14,0x1A,0x14,0,0x15,0x1B,0x15,0,
    0x16,0x1C,0x16,0,0x17,0x1D,0x17,0,0x18,0x1E,0x18,0,0x19,0x1F,0x19,0,
    0xDB,0xE0,0xDA,0,0xDC,0xE1,0xDB,0,0xDD,0xE2,0xDC,0,0xDE,0xE3,0xDD,0,
    0xDF,0xE4,0xDE,0,0xE0,0xE5,0xDF,0,0xE1,0xE6,0xE0,0,0xE2,0xE7,0xE1,0,
    0xE3,0xE8,0xE2,0,0xE4,0xE9,0xE3,0,0xE5,0xEA,0xE4,0,0xE6,0xEB,0xE5,0,
    0xE7,0xEC,0xE6,0,0xE8,0xED,0xE7,0,0xE9,0xEE,0xE8,0,0xEA,0xEF,0xE9,0,
    0xEB,0xF0,0xEA,0,0xEC,0xF1,0xEB,0,0xED,0xF2,0xEC,0,0xEE,0xF3,0xED,0,
    0xEF,0xF4,0xEE,0,0xF0,0xF5,0xEF,0,0xF1,0xF6,0xF0,0,0xF2,0xF7,0xF1,0,
    0xF3,0xF8,0xF2,0,0xF4,0xF9,0xF3,0,0xF5,0xFA,0xF4,0,0xF6,0xFB,0xF5,0,
    0xF7,0xFC,0xF6,0,0xF8,0xFD,0xF7,0,0xF9,0xFE,0xF8,0,0xFA,0xFF,0xF9,0,
    0xFB,0x00,0xFA,0,0xFC,0x01,0xFB,0,0xFD,0x02,0xFC,0,0xFE,0x03,0xFD,0,
    0xFF,0x04,0xFE,0,0x00,0x05,0xFF,0,0x01,0x06,0x00,0,0x02,0x07,0x01,0,
    0x03,0x08,0x02,0,0x04,0x09,0x03,0,0x05,0x0A,0x04,0,0x06,0x0B,0x05,0,
    0x07,0x0C,0x06,0,0x08,0x0D,0x07,0,0x09,0x0E,0x08,0,0x0A,0x0F,0x09,0,
    0x0B,0x10,0x0A,0,0x0C,0x11,0x0B,0,0x0D,0x12,0x0C,0,0x0E,0x13,0x0D,0,
    0x0F,0x14,0x0E,0,0x10,0x15,0x0F,0,0x11,0x16,0x10,0,0x12,0x17,0x11,0,
    0x13,0x18,0x12,0,0x14,0x19,0x13,0,0x15,0x1A,0x14,0,0x16,0x1B,0x15,0,
    0x17,0x1C,0x16,0,0x18,0x1D,0x17,0,0x19,0x1E,0x18,0,0x1A,0x1F,0x19,0,
    0xDC,0xE0,0xDA,0,0xDD,0xE1,0xDB,0,0xDE,0xE2,0xDC,0,0xDF,0xE3,0xDD,0,
    0xE0,0xE4,0xDE,0,0xE1,0xE5,0xDF,0,0xE2,0xE6,0xE0,0,0xE3,0xE7,0xE1,0,
    0xE4,0xE8,0xE2,0,0xE5,0xE9,0xE3,0,0xE6,0xEA,0xE4,0,0xE7,0xEB,0xE5,0,
    0xE8,0xEC,0xE6,0,0xE9,0xED,0xE7,0,0xEA,0xEE,0xE8,0,0xEB,0xEF,0xE9,0,
    0xEC,0xF0,0xEA,0,0xED,0xF1,0xEB,0,0xEE,0xF2,0xEC,0,0xEF,0xF3,0xED,0,
    0xF0,0xF4,0xEE,0,0xF1,0xF5,0xEF,0,0xF2,0xF6,0xF0,0,0xF3,0xF7,0xF1,0,
    0xF4,0xF8,0xF2,0,0xF5,0xF9,0xF3,0,0xF6,0xFA,0xF4,0,0xF7,0xFB,0xF5,0,
    0xF8,0xFC,0xF6,0,0xF9,0xFD,0xF7,0,0xFA,0xFE,0xF8,0,0xFB,0xFF,0xF9,0,
    0xFC,0x00,0xFA,0,0xFD,0x01,0xFB,0,0xFE,0x02,0xFC,0,0xFF,0x03,0xFD,0,
    0x00,0x04,0xFE,0,0x01,0x05,0xFF,0,0x02,0x06,0x00,0,0x03,0x07,0x01,0,
    0x04,0x08,0x02,0,0x05,0x09,0x03,0,0x06,0x0A,0x04,0,0x07,0x0B,0x05,0,
    0x08,0x0C,0x06,0,0x09,0x0D,0x07,0,0x0A,0x0E,0x08,0,0x0B,0x0F,0x09,0,
    0x0C,0x10,0x0A,0,0x0D,0x11,0x0B,0,0x0E,0x12,0x0C,0,0x0F,0x13,0x0D,0,
    0x10,0x14,0x0E,0,0x11,0x15,0x0F,0,0x12,0x16,0x10,0,0x13,0x17,0x11,0,
    0x14,0x18,0x12,0,0x15,0x19,0x13,0,0x16,0x1A,0x14,0,0x17,0x1B,0x15,0,
    0x18,0x1C,0x16,0,0x19,0x1D,0x17,0,0x1A,0x1E,0x18,0,0x1B,0x1F,0x19,0,
    0xDD,0xE0,0xDA,0,0xDE,0xE1,0xDB,0,0xDF,0xE2,0xDC,0,0xE0,0xE3,0xDD,0,
    0xE1,0xE4,0xDE,0,0xE2,0xE5,0xDF,0,0xE3,0xE6,0xE0,0,0xE4,0xE7,0xE1,0,
    0xE5,0xE8,0xE2,0,0xE6,0xE9,0xE3,0,0xE7,0xEA,0xE4,0,0xE8,0xEB,0xE5,0,
    0xE9,0xEC,0xE6,0,0xEA,0xED,0xE7,0,0xEB,0xEE,0xE8,0,0xEC,0xEF,0xE9,0,
    0xED,0xF0,0xEA,0,0xEE,0xF1,0xEB,0,0xEF,0xF2,0xEC,0,0xF0,0xF3,0xED,0,
    0xF1,0xF4,0xEE,0,0xF2,0xF5,0xEF,0,0xF3,0xF6,0xF0,0,0xF4,0xF7,0xF1,0,
    0xF5,0xF8,0xF2,0,0xF6,0xF9,0xF3,0,0xF7,0xFA,0xF4,0,0xF8,0xFB,0xF5,0,
    0xF9,0xFC,0xF6,0,0xFA,0xFD,0xF7,0,0xFB,0xFE,0xF8,0,0xFC,0xFF,0xF9,0,
    0xFD,0x00,0xFA,0,0xFE,0x01,0xFB,0,0xFF,0x02,0xFC,0,0x00,0x03,0xFD,0,
    0x01,0x04,0xFE,0,0x02,0x05,0xFF,0,0x03,0x06,0x00,0,0x04,0x07,0x01,0,
    0x05,0x08,0x02,0,0x06,0x09,0x03,0,0x07,0x0A,0x04,0,0x08,0x0B,0x05,0,
    0x09,0x0C,0x06,0,0x0A,0x0D,0x07,0,0x0B,0x0E,0x08,0,0x0C,0x0F,0x09,0,
    0x0D,0x10,0x0A,0,0x0E,0x11,0x0B,0,0x0F,0x12,0x0C,0,0x10,0x13,0x0D,0,
    0x11,0x14,0x0E,0,0x12,0x15,0x0F,0,0x13,0x16,0x10,0,0x14,0x17,0x11,0,
    0x15,0x18,0x12,0,0x16,0x19,0x13,0,0x17,0x1A,0x14,0,0x18,0x1B,0x15,0,
    0x19,0x1C,0x16,0,0x1A,0x1D,0x17,0,0x1B,0x1E,0x18,0,0x1C,0x1F,0x19,0,
    0xDE,0xE0,0xDA,0,0xDF,0xE1,0xDB,0,0xE0,0xE2,0xDC,0,0xE1,0xE3,0xDD,0,
    0xE2,0xE4,0xDE,0,0xE3,0xE5,0xDF,0,0xE4,0xE6,0xE0,0,0xE5,0xE7,0xE1,0,
    0xE6,0xE8,0xE2,0,0xE7,0xE9,0xE3,0,0xE8,0xEA,0xE4,0,0xE9,0xEB,0xE5,0,
    0xEA,0xEC,0xE6,0,0xEB,0xED,0xE7,0,0xEC,0xEE,0xE8,0,0xED,0xEF,0xE9,0,
    0xEE,0xF0,0xEA,0,0xEF,0xF1,0xEB,0,0xF0,0xF2,0xEC,0,0xF1,0xF3,0xED,0,
    0xF2,0xF4,0xEE,0,0xF3,0xF5,0xEF,0,0xF4,0xF6,0xF0,0,0xF5,0xF7,0xF1,0,
    0xF6,0xF8,0xF2,0,0xF7,0xF9,0xF3,0,0xF8,0xFA,0xF4,0,0xF9,0xFB,0xF5,0,
    0xFA,0xFC,0xF6,0,0xFB,0xFD,0xF7,0,0xFC,0xFE,0xF8,0,0xFD,0xFF,0xF9,0,
    0xFE,0x00,0xFA,0,0xFF,0x01,0xFB,0,0x00,0x02,0xFC,0,0x01,0x03,0xFD,0,
    0x02,0x04,0xFE,0,0x03,0x05,0xFF,0,0x04,0x06,0x00,0,0x05,0x07,0x01,0,
    0x06,0x08,0x02,0,0x07,0x09,0x03,0,0x08,0x0A,0x04,0,0x09,0x0B,0x05,0,
    0x0A,0x0C,0x06,0,0x0B,0x0D,0x07,0,0x0C,0x0E,0x08,0,0x0D,0x0F,0x09,0,
    0x0E,0x10,0x0A,0,0x0F,0x11,0x0B,0,0x10,0x12,0x0C,0,0x11,0x13,0x0D,0,
    0x12,0x14,0x0E,0,0x13,0x15,0x0F,0,0x14,0x16,0x10,0,0x15,0x17,0x11,0,
    0x16,0x18,0x12,0,0x17,0x19,0x13,0,0x18,0x1A,0x14,0,0x19,0x1B,0x15,0,
    0x1A,0x1C,0x16,0,0x1B,0x1D,0x17,0,0x1C,0x1E,0x18,0,0x1D,0x1F,0x19,0,
    0xDF,0xE0,0xDA,0,0xE0,0xE1,0xDB,0,0xE1,0xE2,0xDC,0,0xE2,0xE3,0xDD,0,
    0xE3,0xE4,0xDE,0,0xE4,0xE5,0xDF,0,0xE5,0xE6,0xE0,0,0xE6,0xE7,0xE1,0,
    0xE7,0xE8,0xE2,0,0xE8,0xE9,0xE3,0,0xE9,0xEA,0xE4,0,0xEA,0xEB,0xE5,0,
    0xEB,0xEC,0xE6,0,0xEC,0xED,0xE7,0,0xED,0xEE,0xE8,0,0xEE,0xEF,0xE9,0,
    0xEF,0xF0,0xEA,0,0xF0,0xF1,0xEB,0,0xF1,0xF2,0xEC,0,0xF2,0xF3,0xED,0,
    0xF3,0xF4,0xEE,0,0xF4,0xF5,0xEF,0,0xF5,0xF6,0xF0,0,0xF6,0xF7,0xF1,0,
    0xF7,0xF8,0xF2,0,0xF8,0xF9,0xF3,0,0xF9,0xFA,0xF4,0,0xFA,0xFB,0xF5,0,
    0xFB,0xFC,0xF6,0,0xFC,0xFD,0xF7,0,0xFD,0xFE,0xF8,0,0xFE,0xFF,0xF9,0,
    0xFF,0x00,0xFA,0,0x00,0x01,0xFB,0,0x01,0x02,0xFC,0,0x02,0x03,0xFD,0,
    0x03,0x04,0xFE,0,0x04,0x05,0xFF,0,0x05,0x06,0x00,0,0x06,0x07,0x01,0,
    0x07,0x08,0x02,0,0x08,0x09,0x03,0,0x09,0x0A,0x04,0,0x0A,0x0B,0x05,0,
    0x0B,0x0C,0x06,0,0x0C,0x0D,0x07,0,0x0D,0x0E,0x08,0,0x0E,0x0F,0x09,0,
    0x0F,0x10,0x0A,0,0x10,0x11,0x0B,0,0x11,0x12,0x0C,0,0x12,0x13,0x0D,0,
    0x13,0x14,0x0E,0,0x14,0x15,0x0F,0,0x15,0x16,0x10,0,0x16,0x17,0x11,0,
    0x17,0x18,0x12,0,0x18,0x19,0x13,0,0x19,0x1A,0x14,0,0x1A,0x1B,0x15,0,
    0x1B,0x1C,0x16,0,0x1C,0x1D,0x17,0,0x1D,0x1E,0x18,0,0x1E,0x1F,0x19,0,
    0xE0,0xE0,0xDA,0,0xE1,0xE1,0xDB,0,0xE2,0xE2,0xDC,0,0xE3,0xE3,0xDD,0,
    0xE4,0xE4,0xDE,0,0xE5,0xE5,0xDF,0,0xE6,0xE6,0xE0,0,0xE7,0xE7,0xE1,0,
    0xE8,0xE8,0xE2,0,0xE9,0xE9,0xE3,0,0xEA,0xEA,0xE4,0,0xEB,0xEB,0xE5,0,
    0xEC,0xEC,0xE6,0,0xED,0xED,0xE7,0,0xEE,0xEE,0xE8,0,0xEF,0xEF,0xE9,0,
    0xF0,0xF0,0xEA,0,0xF1,0xF1,0xEB,0,0xF2,0xF2,0xEC,0,0xF3,0xF3,0xED,0,
    0xF4,0xF4,0xEE,0,0xF5,0xF5,0xEF,0,0xF6,0xF6,0xF0,0,0xF7,0xF7,0xF1,0,
    0xF8,0xF8,0xF2,0,0xF9,0xF9,0xF3,0,0xFA,0xFA,0xF4,0,0xFB,0xFB,0xF5,0,
    0xFC,0xFC,0xF6,0,0xFD,0xFD,0xF7,0,0xFE,0xFE,0xF8,0,0xFF,0xFF,0xF9,0,
    0x00,0x00,0xFA,0,0x01,0x01,0xFB,0,0x02,0x02,0xFC,0,0x03,0x03,0xFD,0,
    0x04,0x04,0xFE,0,0x05,0x05,0xFF,0,0x06,0x06,0x00,0,0x07,0x07,0x01,0,
    0x08,0x08,0x02,0,0x09,0x09,0x03,0,0x0A,0x0A,0x04,0,0x0B,0x0B,0x05,0,
    0x0C,0x0C,0x06,0,0x0D,0x0D,0x07,0,0x0E,0x0E,0x08,0,0x0F,0x0F,0x09,0,
    0x10,0x10,0x0A,0,0x11,0x11,0x0B,0,0x12,0x12,0x0C,0,0x13,0x13,0x0D,0,
    0x14,0x14,0x0E,0,0x15,0x15,0x0F,0,0x16,0x16,0x10,0,0x17,0x17,0x11,0,
    0x18,0x18,0x12,0,0x19,0x19,0x13,0,0x1A,0x1A,0x14,0,0x1B,0x1B,0x15,0,
    0x1C,0x1C,0x16,0,0x1D,0x1D,0x17,0,0x1E,0x1E,0x18,0,0x1F,0x1F,0x19,0,
    0xE1,0xE0,0xDA,0,0xE2,0xE1,0xDB,0,0xE3,0xE2,0xDC,0,0xE4,0xE3,0xDD,0,
    0xE5,0xE4,0xDE,0,0xE6,0xE5,0xDF,0,0xE7,0xE6,0xE0,0,0xE8,0xE7,0xE1,0,
    0xE9,0xE8,0xE2,0,0xEA,0xE9,0xE3,0,0xEB,0xEA,0xE4,0,0xEC,0xEB,0xE5,0,
    0xED,0xEC,0xE6,0,0xEE,0xED,0xE7,0,0xEF,0xEE,0xE8,0,0xF0,0xEF,0xE9,0,
    0xF1,0xF0,0xEA,0,0xF2,0xF1,0xEB,0,0xF3,0xF2,0xEC,0,0xF4,0xF3,0xED,0,
    0xF5,0xF4,0xEE,0,0xF6,0xF5,0xEF,0,0xF7,0xF6,0xF0,0,0xF8,0xF7,0xF1,0,
    0xF9,0xF8,0xF2,0,0xFA,0xF9,0xF3,0,0xFB,0xFA,0xF4,0,0xFC,0xFB,0xF5,0,
    0xFD,0xFC,0xF6,0,0xFE,0xFD,0xF7,0,0xFF,0xFE,0xF8,0,0x00,0xFF,0xF9,0,
    0x01,0x00,0xFA,0,0x02,0x01,0xFB,0,0x03,0x02,0xFC,0,0x04,0x03,0xFD,0,
    0x05,0x04,0xFE,0,0x06,0x05,0xFF,0,0x07,0x06,0x00,0,0x08,0x07,0x01,0,
    0x09,0x08,0x02,0,0x0A,0x09,0x03,0,0x0B,0x0A,0x04,0,0x0C,0x0B,0x05,0,
    0x0D,0x0C,0x06,0,0x0E,0x0D,0x07,0,0x0F,0x0E,0x08,0,0x10,0x0F,0x09,0,
    0x11,0x10,0x0A,0,0x12,0x11,0x0B,0,0x13,0x12,0x0C,0,0x14,0x13,0x0D,0,
    0x15,0x14,0x0E,0,0x16,0x15,0x0F,0,0x17,0x16,0x10,0,0x18,0x17,0x11,0,
    0x19,0x18,0x12,0,0x1A,0x19,0x13,0,0x1B,0x1A,0x14,0,0x1C,0x1B,0x15,0,
    0x1D,0x1C,0x16,0,0x1E,0x1D,0x17,0,0x1F,0x1E,0x18,0,0x20,0x1F,0x19,0,
    0xE2,0xE0,0xDA,0,0xE3,0xE1,0xDB,0,0xE4,0xE2,0xDC,0,0xE5,0xE3,0xDD,0,
    0xE6,0xE4,0xDE,0,0xE7,0xE5,0xDF,0,0xE8,0xE6,0xE0,0,0xE9,0xE7,0xE1,0,
    0xEA,0xE8,0xE2,0,0xEB,0xE9,0xE3,0,0xEC,0xEA,0xE4,0,0xED,0xEB,0xE5,0,
    0xEE,0xEC,0xE6,0,0xEF,0xED,0xE7,0,0xF0,0xEE,0xE8,0,0xF1,0xEF,0xE9,0,
    0xF2,0xF0,0xEA,0,0xF3,0xF1,0xEB,0,0xF4,0xF2,0xEC,0,0xF5,0xF3,0xED,0,
    0xF6,0xF4,0xEE,0,0xF7,0xF5,0xEF,0,0xF8,0xF6,0xF0,0,0xF9,0xF7,0xF1,0,
    0xFA,0xF8,0xF2,0,0xFB,0xF9,0xF3,0,0xFC,0xFA,0xF4,0,0xFD,0xFB,0xF5,0,
    0xFE,0xFC,0xF6,0,0xFF,0xFD,0xF7,0,0x00,0xFE,0xF8,0,0x01,0xFF,0xF9,0,
    0x02,0x00,0xFA,0,0x03,0x01,0xFB,0,0x04,0x02,0xFC,0,0x05,0x03,0xFD,0,
    0x06,0x04,0xFE,0,0x07,0x05,0xFF,0,0x08,0x06,0x00,0,0x09,0x07,0x01,0,
    0x0A,0x08,0x02,0,0x0B,0x09,0x03,0,0x0C,0x0A,0x04,0,0x0D,0x0B,0x05,0,
    0x0E,0x0C,0x06,0,0x0F,0x0D,0x07,0,0x10,0x0E,0x08,0,0x11,0x0F,0x09,0,
    0x12,0x10,0x0A,0,0x13,0x11,0x0B,0,0x14,0x12,0x0C,0,0x15,0x13,0x0D,0,
    0x16,0x14,0x0E,0,0x17,0x15,0x0F,0,0x18,0x16,0x10,0,0x19,0x17,0x11,0,
    0x1A,0x18,0x12,0,0x1B,0x19,0x13,0,0x1C,0x1A,0x14,0,0x1D,0x1B,0x15,0,
    0x1E,0x1C,0x16,0,0x1F,0x1D,0x17,0,0x20,0x1E,0x18,0,0x21,0x1F,0x19,0,
    0xE3,0xE0,0xDA,0,0xE4,0xE1,0xDB,0,0xE5,0xE2,0xDC,0,0xE6,0xE3,0xDD,0,
    0xE7,0xE4,0xDE,0,0xE8,0xE5,0xDF,0,0xE9,0xE6,0xE0,0,0xEA,0xE7,0xE1,0,
    0xEB,0xE8,0xE2,0,0xEC,0xE9,0xE3,0,0xED,0xEA,0xE4,0,0xEE,0xEB,0xE5,0,
    0xEF,0xEC,0xE6,0,0xF0,0xED,0xE7,0,0xF1,0xEE,0xE8,0,0xF2,0xEF,0xE9,0,
    0xF3,0xF0,0xEA,0,0xF4,0xF1,0xEB,0,0xF5,0xF2,0xEC,0,0xF6,0xF3,0xED,0,
    0xF7,0xF4,0xEE,0,0xF8,0xF5,0xEF,0,0xF9,0xF6,0xF0,0,0xFA,0xF7,0xF1,0,
    0xFB,0xF8,0xF2,0,0xFC,0xF9,0xF3,0,0xFD,0xFA,0xF4,0,0xFE,0xFB,0xF5,0,
    0xFF,0xFC,0xF6,0,0x00,0xFD,0xF7,0,0x01,0xFE,0xF8,0,0x02,0xFF,0xF9,0,
    0x03,0x00,0xFA,0,0x04,0x01,0xFB,0,0x05,0x02,0xFC,0,0x06,0x03,0xFD,0,
    0x07,0x04,0xFE,0,0x08,0x05,0xFF,0,0x09,0x06,0x00,0,0x0A,0x07,0x01,0,
    0x0B,0x08,0x02,0,0x0C,0x09,0x03,0,0x0D,0x0A,0x04,0,0x0E,0x0B,0x05,0,
    0x0F,0x0C,0x06,0,0x10,0x0D,0x07,0,0x11,0x0E,0x08,0,0x12,0x0F,0x09,0,
    0x13,0x10,0x0A,0,0x14,0x11,0x0B,0,0x15,0x12,0x0C,0,0x16,0x13,0x0D,0,
    0x17,0x14,0x0E,0,0x18,0x15,0x0F,0,0x19,0x16,0x10,0,0x1A,0x17,0x11,0,
    0x1B,0x18,0x12,0,0x1C,0x19,0x13,0,0x1D,0x1A,0x14,0,0x1E,0x1B,0x15,0,
    0x1F,0x1C,0x16,0,0x20,0x1D,0x17,0,0x21,0x1E,0x18,0,0x22,0x1F,0x19,0,
    0xE4,0xE0,0xDA,0,0xE5,0xE1,0xDB,0,0xE6,0xE2,0xDC,0,0xE7,0xE3,0xDD,0,
    0xE8,0xE4,0xDE,0,0xE9,0xE5,0xDF,0,0xEA,0xE6,0xE0,0,0xEB,0xE7,0xE1,0,
    0xEC,0xE8,0xE2,0,0xED,0xE9,0xE3,0,0xEE,0xEA,0xE4,0,0xEF,0xEB,0xE5,0,
    0xF0,0xEC,0xE6,0,0xF1,0xED,0xE7,0,0xF2,0xEE,0xE8,0,0xF3,0xEF,0xE9,0,
    0xF4,0xF0,0xEA,0,0xF5,0xF1,0xEB,0,0xF6,0xF2,0xEC,0,0xF7,0xF3,0xED,0,
    0xF8,0xF4,0xEE,0,0xF9,0xF5,0xEF,0,0xFA,0xF6,0xF0,0,0xFB,0xF7,0xF1,0,
    0xFC,0xF8,0xF2,0,0xFD,0xF9,0xF3,0,0xFE,0xFA,0xF4,0,0xFF,0xFB,0xF5,0,
    0x00,0xFC,0xF6,0,0x01,0xFD,0xF7,0,0x02,0xFE,0xF8,0,0x03,0xFF,0xF9,0,
    0x04,0x00,0xFA,0,0x05,0x01,0xFB,0,0x06,0x02,0xFC,0,0x07,0x03,0xFD,0,
    0x08,0x04,0xFE,0,0x09,0x05,0xFF,0,0x0A,0x06,0x00,0,0x0B,0x07,0x01,0,
    0x0C,0x08,0x02,0,0x0D,0x09,0x03,0,0x0E,0x0A,0x04,0,0x0F,0x0B,0x05,0,
    0x10,0x0C,0x06,0,0x11,0x0D,0x07,0,0x12,0x0E,0x08,0,0x13,0x0F,0x09,0,
    0x14,0x10,0x0A,0,0x15,0x11,0x0B,0,0x16,0x12,0x0C,0,0x17,0x13,0x0D,0,
    0x18,0x14,0x0E,0,0x19,0x15,0x0F,0,0x1A,0x16,0x10,0,0x1B,0x17,0x11,0,
    0x1C,0x18,0x12,0,0x1D,0x19,0x13,0,0x1E,0x1A,0x14,0,0x1F,0x1B,0x15,0,
    0x20,0x1C,0x16,0,0x21,0x1D,0x17,0,0x22,0x1E,0x18,0,0x23,0x1F,0x19,0,
    0xE5,0xE0,0xDA,0,0xE6,0xE1,0xDB,0,0xE7,0xE2,0xDC,0,0xE8,0xE3,0xDD,0,
    0xE9,0xE4,0xDE,0,0xEA,0xE5,0xDF,0,0xEB,0xE6,0xE0,0,0xEC,0xE7,0xE1,0,
    0xED,0xE8,0xE2,0,0xEE,0xE9,0xE3,0,0xEF,0xEA,0xE4,0,0xF0,0xEB,0xE5,0,
    0xF1,0xEC,0xE6,0,0xF2,0xED,0xE7,0,0xF3,0xEE,0xE8,0,0xF4,0xEF,0xE9,0,
    0xF5,0xF0,0xEA,0,0xF6,0xF1,0xEB,0,0xF7,0xF2,0xEC,0,0xF8,0xF3,0xED,0,
    0xF9,0xF4,0xEE,0,0xFA,0xF5,0xEF,0,0xFB,0xF6,0xF0,0,0xFC,0xF7,0xF1,0,
    0xFD,0xF8,0xF2,0,0xFE,0xF9,0xF3,0,0xFF,0xFA,0xF4,0,0x00,0xFB,0xF5,0,
    0x01,0xFC,0xF6,0,0x02,0xFD,0xF7,0,0x03,0xFE,0xF8,0,0x04,0xFF,0xF9,0,
    0x05,0x00,0xFA,0,0x06,0x01,0xFB,0,0x07,0x02,0xFC,0,0x08,0x03,0xFD,0,
    0x09,0x04,0xFE,0,0x0A,0x05,0xFF,0,0x0B,0x06,0x00,0,0x0C,0x07,0x01,0,
    0x0D,0x08,0x02,0,0x0E,0x09,0x03,0,0x0F,0x0A,0x04,0,0x10,0x0B,0x05,0,
    0x11,0x0C,0x06,0,0x12,0x0D,0x07,0,0x13,0x0E,0x08,0,0x14,0x0F,0x09,0,
    0x15,0x10,0x0A,0,0x16,0x11,0x0B,0,0x17,0x12,0x0C,0,0x18,0x13,0x0D,0,
    0x19,0x14,0x0E,0,0x1A,0x15,0x0F,0,0x1B,0x16,0x10,0,0x1C,0x17,0x11,0,
    0x1D,0x18,0x12,0,0x1E,0x19,0x13,0,0x1F,0x1A,0x14,0,0x20,0x1B,0x15,0,
    0x21,0x1C,0x16,0,0x22,0x1D,0x17,0,0x23,0x1E,0x18,0,0x24,0x1F,0x19,0,
    0xE6,0xE0,0xDA,0,0xE7,0xE1,0xDB,0,0xE8,0xE2,0xDC,0,0xE9,0xE3,0xDD,0,
    0xEA,0xE4,0xDE,0,0xEB,0xE5,0xDF,0,0xEC,0xE6,0xE0,0,0xED,0xE7,0xE1,0,
    0xEE,0xE8,0xE2,0,0xEF,0xE9,0xE3,0,0xF0,0xEA,0xE4,0,0xF1,0xEB,0xE5,0,
    0xF2,0xEC,0xE6,0,0xF3,0xED,0xE7,0,0xF4,0xEE,0xE8,0,0xF5,0xEF,0xE9,0,
    0xF6,0xF0,0xEA,0,0xF7,0xF1,0xEB,0,0xF8,0xF2,0xEC,0,0xF9,0xF3,0xED,0,
    0xFA,0xF4,0xEE,0,0xFB,0xF5,0xEF,0,0xFC,0xF6,0xF0,0,0xFD,0xF7,0xF1,0,
    0xFE,0xF8,0xF2,0,0xFF,0xF9,0xF3,0,0x00,0xFA,0xF4,0,0x01,0xFB,0xF5,0,
    0x02,0xFC,0xF6,0,0x03,0xFD,0xF7,0,0x04,0xFE,0xF8,0,0x05,0xFF,0xF9,0,
    0x06,0x00,0xFA,0,0x07,0x01,0xFB,0,0x08,0x02,0xFC,0,0x09,0x03,0xFD,0,
    0x0A,0x04,0xFE,0,0x0B,0x05,0xFF,0,0x0C,0x06,0x00,0,0x0D,0x07,0x01,0,
    0x0E,0x08,0x02,0,0x0F,0x09,0x03,0,0x10,0x0A,0x04,0,0x11,0x0B,0x05,0,
    0x12,0x0C,0x06,0,0x13,0x0D,0x07,0,0x14,0x0E,0x08,0,0x15,0x0F,0x09,0,
    0x16,0x10,0x0A,0,0x17,0x11,0x0B,0,0x18,0x12,0x0C,0,0x19,0x13,0x0D,0,
    0x1A,0x14,0x0E,0,0x1B,0x15,0x0F,0,0x1C,0x16,0x10,0,0x1D,0x17,0x11,0,
    0x1E,0x18,0x12,0,0x1F,0x19,0x13,0,0x20,0x1A,0x14,0,0x21,0x1B,0x15,0,
    0x22,0x1C,0x16,0,0x23,0x1D,0x17,0,0x24,0x1E,0x18,0,0x25,0x1F,0x19,0,
    0xE7,0xE0,0xDA,0,0xE8,0xE1,0xDB,0,0xE9,0xE2,0xDC,0,0xEA,0xE3,0xDD,0,
    0xEB,0xE4,0xDE,0,0xEC,0xE5,0xDF,0,0xED,0xE6,0xE0,0,0xEE,0xE7,0xE1,0,
    0xEF,0xE8,0xE2,0,0xF0,0xE9,0xE3,0,0xF1,0xEA,0xE4,0,0xF2,0xEB,0xE5,0,
    0xF3,0xEC,0xE6,0,0xF4,0xED,0xE7,0,0xF5,0xEE,0xE8,0,0xF6,0xEF,0xE9,0,
    0xF7,0xF0,0xEA,0,0xF8,0xF1,0xEB,0,0xF9,0xF2,0xEC,0,0xFA,0xF3,0xED,0,
    0xFB,0xF4,0xEE,0,0xFC,0xF5,0xEF,0,0xFD,0xF6,0xF0,0,0xFE,0xF7,0xF1,0,
    0xFF,0xF8,0xF2,0,0x00,0xF9,0xF3,0,0x01,0xFA,0xF4,0,0x02,0xFB,0xF5,0,
    0x03,0xFC,0xF6,0,0x04,0xFD,0xF7,0,0x05,0xFE,0xF8,0,0x06,0xFF,0xF9,0,
    0x07,0x00,0xFA,0,0x08,0x01,0xFB,0,0x09,0x02,0xFC,0,0x0A,0x03,0xFD,0,
    0x0B,0x04,0xFE,0,0x0C,0x05,0xFF,0,0x0D,0x06,0x00,0,0x0E,0x07,0x01,0,
    0x0F,0x08,0x02,0,0x10,0x09,0x03,0,0x11,0x0A,0x04,0,0x12,0x0B,0x05,0,
    0x13,0x0C,0x06,0,0x14,0x0D,0x07,0,0x15,0x0E,0x08,0,0x16,0x0F,0x09,0,
    0x17,0x10,0x0A,0,0x18,0x11,0x0B,0,0x19,0x12,0x0C,0,0x1A,0x13,0x0D,0,
    0x1B,0x14,0x0E,0,0x1C,0x15,0x0F,0,0x1D,0x16,0x10,0,0x1E,0x17,0x11,0,
    0x1F,0x18,0x12,0,0x20,0x19,0x13,0,0x21,0x1A,0x14,0,0x22,0x1B,0x15,0,
    0x23,0x1C,0x16,0,0x24,0x1D,0x17,0,0x25,0x1E,0x18,0,0x26,0x1F,0x19,0,
    0xD8,0xE0,0xDB,0,0xD9,0xE1,0xDC,0,0xDA,0xE2,0xDD,0,0xDB,0xE3,0xDE,0,
    0xDC,0xE4,0xDF,0,0xDD,0xE5,0xE0,0,0xDE,0xE6,0xE1,0,0xDF,0xE7,0xE2,0,
    0xE0,0xE8,0xE3,0,0xE1,0xE9,0xE4,0,0xE2,0xEA,0xE5,0,0xE3,0xEB,0xE6,0,
    0xE4,0xEC,0xE7,0,0xE5,0xED,0xE8,0,0xE6,0xEE,0xE9,0,0xE7,0xEF,0xEA,0,
    0xE8,0xF0,0xEB,0,0xE9,0xF1,0xEC,0,0xEA,0xF2,0xED,0,0xEB,0xF3,0xEE,0,
    0xEC,0xF4,0xEF,0,0xED,0xF5,0xF0,0,0xEE,0xF6,0xF1,0,0xEF,0xF7,0xF2,0,
    0xF0,0xF8,0xF3,0,0xF1,0xF9,0xF4,0,0xF2,0xFA,0xF5,0,0xF3,0xFB,0xF6,0,
    0xF4,0xFC,0xF7,0,0xF5,0xFD,0xF8,0,0xF6,0xFE,0xF9,0,0xF7,0xFF,0xFA,0,
    0xF8,0x00,0xFB,0,0xF9,0x01,0xFC,0,0xFA,0x02,0xFD,0,0xFB,0x03,0xFE,0,
    0xFC,0x04,0xFF,0,0xFD,0x05,0x00,0,0xFE,0x06,0x01,0,0xFF,0x07,0x02,0,
    0x00,0x08,0x03,0,0x01,0x09,0x04,0,0x02,0x0A,0x05,0,0x03,0x0B,0x06,0,
    0x04,0x0C,0x07,0,0x05,0x0D,0x08,0,0x06,0x0E,0x09,0,0x07,0x0F,0x0A,0,
    0x08,0x10,0x0B,0,0x09,0x11,0x0C,0,0x0A,0x12,0x0D,0,0x0B,0x13,0x0E,0,
    0x0C,0x14,0x0F,0,0x0D,0x15,0x10,0,0x0E,0x16,0x11,0,0x0F,0x17,0x12,0,
    0x10,0x18,0x13,0,0x11,0x19,0x14,0,0x12,0x1A,0x15,0,0x13,0x1B,0x16,0,
    0x14,0x1C,0x17,0,0x15,0x1D,0x18,0,0x16,0x1E,0x19,0,0x17,0x1F,0x1A,0,
    0xD9,0xE0,0xDB,0,0xDA,0xE1,0xDC,0,0xDB,0xE2,0xDD,0,0xDC,0xE3,0xDE,0,
    0xDD,0xE4,0xDF,0,0xDE,0xE5,0xE0,0,0xDF,0xE6,0xE1,0,0xE0,0xE7,0xE2,0,
    0xE1,0xE8,0xE3,0,0xE2,0xE9,0xE4,0,0xE3,0xEA,0xE5,0,0xE4,0xEB,0xE6,0,
    0xE5,0xEC,0xE7,0,0xE6,0xED,0xE8,0,0xE7,0xEE,0xE9,0,0xE8,0xEF,0xEA,0,
    0xE9,0xF0,0xEB,0,0xEA,0xF1,0xEC,0,0xEB,0xF2,0xED,0,0xEC,0xF3,0xEE,0,
    0xED,0xF4,0xEF,0,0xEE,0xF5,0xF0,0,0xEF,0xF6,0xF1,0,0xF0,0xF7,0xF2,0,
    0xF1,0xF8,0xF3,0,0xF2,0xF9,0xF4,0,0xF3,0xFA,0xF5,0,0xF4,0xFB,0xF6,0,
    0xF5,0xFC,0xF7,0,0xF6,0xFD,0xF8,0,0xF7,0xFE,0xF9,0,0xF8,0xFF,0xFA,0,
    0xF9,0x00,0xFB,0,0xFA,0x01,0xFC,0,0xFB,0x02,0xFD,0,0xFC,0x03,0xFE,0,
    0xFD,0x04,0xFF,0,0xFE,0x05,0x00,0,0xFF,0x06,0x01,0,0x00,0x07,0x02,0,
    0x01,0x08,0x03,0,0x02,0x09,0x04,0,0x03,0x0A,0x05,0,0x04,0x0B,0x06,0,
    0x05,0x0C,0x07,0,0x06,0x0D,0x08,0,0x07,0x0E,0x09,0,0x08,0x0F,0x0A,0,
    0x09,0x10,0x0B,0,0x0A,0x11,0x0C,0,0x0B,0x12,0x0D,0,0x0C,0x13,0x0E,0,
    0x0D,0x14,0x0F,0,0x0E,0x15,0x10,0,0x0F,0x16,0x11,0,0x10,0x17,0x12,0,
    0x11,0x18,0x13,0,0x12,0x19,0x14,0,0x13,0x1A,0x15,0,0x14,0x1B,0x16,0,
    0x15,0x1C,0x17,0,0x16,0x1D,0x18,0,0x17,0x1E,0x19,0,0x18,0x1F,0x1A,0,
    0xDA,0xE0,0xDB,0,0xDB,0xE1,0xDC,0,0xDC,0xE2,0xDD,0,0xDD,0xE3,0xDE,0,
    0xDE,0xE4,0xDF,0,0xDF,0xE5,0xE0,0,0xE0,0xE6,0xE1,0,0xE1,0xE7,0xE2,0,
    0xE2,0xE8,0xE3,0,0xE3,0xE9,0xE4,0,0xE4,0xEA,0xE5,0,0xE5,0xEB,0xE6,0,
    0xE6,0xEC,0xE7,0,0xE7,0xED,0xE8,0,0xE8,0xEE,0xE9,0,0xE9,0xEF,0xEA,0,
    0xEA,0xF0,0xEB,0,0xEB,0xF1,0xEC,0,0xEC,0xF2,0xED,0,0xED,0xF3,0xEE,0,
    0xEE,0xF4,0xEF,0,0xEF,0xF5,0xF0,0,0xF0,0xF6,0xF1,0,0xF1,0xF7,0xF2,0,
    0xF2,0xF8,0xF3,0,0xF3,0xF9,0xF4,0,0xF4,0xFA,0xF5,0,0xF5,0xFB,0xF6,0,
    0xF6,0xFC,0xF7,0,0xF7,0xFD,0xF8,0,0xF8,0xFE,0xF9,0,0xF9,0xFF,0xFA,0,
    0xFA,0x00,0xFB,0,0xFB,0x01,0xFC,0,0xFC,0x02,0xFD,0,0xFD,0x03,0xFE,0,
    0xFE,0x04,0xFF,0,0xFF,0x05,0x00,0,0x00,0x06,0x01,0,0x01,0x07,0x02,0,
    0x02,0x08,0x03,0,0x03,0x09,0x04,0,0x04,0x0A,0x05,0,0x05,0x0B,0x06,0,
    0x06,0x0C,0x07,0,0x07,0x0D,0x08,0,0x08,0x0E,0x09,0,0x09,0x0F,0x0A,0,
    0x0A,0x10,0x0B,0,0x0B,0x11,0x0C,0,0x0C,0x12,0x0D,0,0x0D,0x13,0x0E,0,
    0x0E,0x14,0x0F,0,0x0F,0x15,0x10,0,0x10,0x16,0x11,0,0x11,0x17,0x12,0,
    0x12,0x18,0x13,0,0x13,0x19,0x14,0,0x14,0x1A,0x15,0,0x15,0x1B,0x16,0,
    0x16,0x1C,0x17,0,0x17,0x1D,0x18,0,0x18,0x1E,0x19,0,0x19,0x1F,0x1A,0,
    0xDB,0xE0,0xDB,0,0xDC,0xE1,0xDC,0,0xDD,0xE2,0xDD,0,0xDE,0xE3,0xDE,0,
    0xDF,0xE4,0xDF,0,0xE0,0xE5,0xE0,0,0xE1,0xE6,0xE1,0,0xE2,0xE7,0xE2,0,
    0xE3,0xE8,0xE3,0,0xE4,0xE9,0xE4,0,0xE5,0xEA,0xE5,0,0xE6,0xEB,0xE6,0,
    0xE7,0xEC,0xE7,0,0xE8,0xED,0xE8,0,0xE9,0xEE,0xE9,0,0xEA,0xEF,0xEA,0,
    0xEB,0xF0,0xEB,0,0xEC,0xF1,0xEC,0,0xED,0xF2,0xED,0,0xEE,0xF3,0xEE,0,
    0xEF,0xF4,0xEF,0,0xF0,0xF5,0xF0,0,0xF1,0xF6,0xF1,0,0xF2,0xF7,0xF2,0,
    0xF3,0xF8,0xF3,0,0xF4,0xF9,0xF4,0,0xF5,0xFA,0xF5,0,0xF6,0xFB,0xF6,0,
    0xF7,0xFC,0xF7,0,0xF8,0xFD,0xF8,0,0xF9,0xFE,0xF9,0,0xFA,0xFF,0xFA,0,
    0xFB,0x00,0xFB,0,0xFC,0x01,0xFC,0,0xFD,0x02,0xFD,0,0xFE,0x03,0xFE,0,
    0xFF,0x04,0xFF,0,0x00,0x05,0x00,0,0x01,0x06,0x01,0,0x02,0x07,0x02,0,
    0x03,0x08,0x03,0,0x04,0x09,0x04,0,0x05,0x0A,0x05,0,0x06,0x0B,0x06,0,
    0x07,0x0C,0x07,0,0x08,0x0D,0x08,0,0x09,0x0E,0x09,0,0x0A,0x0F,0x0A,0,
    0x0B,0x10,0x0B,0,0x0C,0x11,0x0C,0,0x0D,0x12,0x0D,0,0x0E,0x13,0x0E,0,
    0x0F,0x14,0x0F,0,0x10,0x15,0x10,0,0x11,0x16,0x11,0,0x12,0x17,0x12,0,
    0x13,0x18,0x13,0,0x14,0x19,0x14,0,0x15,0x1A,0x15,0,0x16,0x1B,0x16,0,
    0x17,0x1C,0x17,0,0x18,0x1D,0x18,0,0x19,0x1E,0x19,0,0x1A,0x1F,0x1A,0,
    0xDC,0xE0,0xDB,0,0xDD,0xE1,0xDC,0,0xDE,0xE2,0xDD,0,0xDF,0xE3,0xDE,0,
    0xE0,0xE4,0xDF,0,0xE1,0xE5,0xE0,0,0xE2,0xE6,0xE1,0,0xE3,0xE7,0xE2,0,
    0xE4,0xE8,0xE3,0,0xE5,0xE9,0xE4,0,0xE6,0xEA,0xE5,0,0xE7,0xEB,0xE6,0,
    0xE8,0xEC,0xE7,0,0xE9,0xED,0xE8,0,0xEA,0xEE,0xE9,0,0xEB,0xEF,0xEA,0,
    0xEC,0xF0,0xEB,0,0xED,0xF1,0xEC,0,0xEE,0xF2,0xED,0,0xEF,0xF3,0xEE,0,
    0xF0,0xF4,0xEF,0,0xF1,0xF5,0xF0,0,0xF2,0xF6,0xF1,0,0xF3,0xF7,0xF2,0,
    0xF4,0xF8,0xF3,0,0xF5,0xF9,0xF4,0,0xF6,0xFA,0xF5,0,0xF7,0xFB,0xF6,0,
    0xF8,0xFC,0xF7,0,0xF9,0xFD,0xF8,0,0xFA,0xFE,0xF9,0,0xFB,0xFF,0xFA,0,
    0xFC,0x00,0xFB,0,0xFD,0x01,0xFC,0,0xFE,0x02,0xFD,0,0xFF,0x03,0xFE,0,
    0x00,0x04,0xFF,0,0x01,0x05,0x00,0,0x02,0x06,0x01,0,0x03,0x07,0x02,0,
    0x04,0x08,0x03,0,0x05,0x09,0x04,0,0x06,0x0A,0x05,0,0x07,0x0B,0x06,0,
    0x08,0x0C,0x07,0,0x09,0x0D,0x08,0,0x0A,0x0E,0x09,0,0x0B,0x0F,0x0A,0,
    0x0C,0x10,0x0B,0,0x0D,0x11,0x0C,0,0x0E,0x12,0x0D,0,0x0F,0x13,0x0E,0,
    0x10,0x14,0x0F,0,0x11,0x15,0x10,0,0x12,0x16,0x11,0,0x13,0x17,0x12,0,
    0x14,0x18,0x13,0,0x15,0x19,0x14,0,0x16,0x1A,0x15,0,0x17,0x1B,0x16,0,
    0x18,0x1C,0x17,0,0x19,0x1D,0x18,0,0x1A,0x1E,0x19,0,0x1B,0x1F,0x1A,0,
    0xDD,0xE0,0xDB,0,0xDE,0xE1,0xDC,0,0xDF,0xE2,0xDD,0,0xE0,0xE3,0xDE,0,
    0xE1,0xE4,0xDF,0,0xE2,0xE5,0xE0,0,0xE3,0xE6,0xE1,0,0xE4,0xE7,0xE2,0,
    0xE5,0xE8,0xE3,0,0xE6,0xE9,0xE4,0,0xE7,0xEA,0xE5,0,0xE8,0xEB,0xE6,0,
    0xE9,0xEC,0xE7,0,0xEA,0xED,0xE8,0,0xEB,0xEE,0xE9,0,0xEC,0xEF,0xEA,0,
    0xED,0xF0,0xEB,0,0xEE,0xF1,0xEC,0,0xEF,0xF2,0xED,0,0xF0,0xF3,0xEE,0,
    0xF1,0xF4,0xEF,0,0xF2,0xF5,0xF0,0,0xF3,0xF6,0xF1,0,0xF4,0xF7,0xF2,0,
    0xF5,0xF8,0xF3,0,0xF6,0xF9,0xF4,0,0xF7,0xFA,0xF5,0,0xF8,0xFB,0xF6,0,
    0xF9,0xFC,0xF7,0,0xFA,0xFD,0xF8,0,0xFB,0xFE,0xF9,0,0xFC,0xFF,0xFA,0,
    0xFD,0x00,0xFB,0,0xFE,0x01,0xFC,0,0xFF,0x02,0xFD,0,0x00,0x03,0xFE,0,
    0x01,0x04,0xFF,0,0x02,0x05,0x00,0,0x03,0x06,0x01,0,0x04,0x07,0x02,0,
    0x05,0x08,0x03,0,0x06,0x09,0x04,0,0x07,0x0A,0x05,0,0x08,0x0B,0x06,0,
    0x09,0x0C,0x07,0,0x0A,0x0D,0x08,0,0x0B,0x0E,0x09,0,0x0C,0x0F,0x0A,0,
    0x0D,0x10,0x0B,0,0x0E,0x11,0x0C,0,0x0F,0x12,0x0D,0,0x10,0x13,0x0E,0,
    0x11,0x14,0x0F,0,0x12,0x15,0x10,0,0x13,0x16,0x11,0,0x14,0x17,0x12,0,
    0x15,0x18,0x13,0,0x16,0x19,0x14,0,0x17,0x1A,0x15,0,0x18,0x1B,0x16,0,
    0x19,0x1C,0x17,0,0x1A,0x1D,0x18,0,0x1B,0x1E,0x19,0,0x1C,0x1F,0x1A,0,
    0xDE,0xE0,0xDB,0,0xDF,0xE1,0xDC,0,0xE0,0xE2,0xDD,0,0xE1,0xE3,0xDE,0,
    0xE2,0xE4,0xDF,0,0xE3,0xE5,0xE0,0,0xE4,0xE6,0xE1,0,0xE5,0xE7,0xE2,0,
    0xE6,0xE8,0xE3,0,0xE7,0xE9,0xE4,0,0xE8,0xEA,0xE5,0,0xE9,0xEB,0xE6,0,
    0xEA,0xEC,0xE7,0,0xEB,0xED,0xE8,0,0xEC,0xEE,0xE9,0,0xED,0xEF,0xEA,0,
    0xEE,0xF0,0xEB,0,0xEF,0xF1,0xEC,0,0xF0,0xF2,0xED,0,0xF1,0xF3,0xEE,0,
    0xF2,0xF4,0xEF,0,0xF3,0xF5,0xF0,0,0xF4,0xF6,0xF1,0,0xF5,0xF7,0xF2,0,
    0xF6,0xF8,0xF3,0,0xF7,0xF9,0xF4,0,0xF8,0xFA,0xF5,0,0xF9,0xFB,0xF6,0,
    0xFA,0xFC,0xF7,0,0xFB,0xFD,0xF8,0,0xFC,0xFE,0xF9,0,0xFD,0xFF,0xFA,0,
    0xFE,0x00,0xFB,0,0xFF,0x01,0xFC,0,0x00,0x02,0xFD,0,0x01,0x03,0xFE,0,
    0x02,0x04,0xFF,0,0x03,0x05,0x00,0,0x04,0x06,0x01,0,0x05,0x07,0x02,0,
    0x06,0x08,0x03,0,0x07,0x09,0x04,0,0x08,0x0A,0x05,0,0x09,0x0B,0x06,0,
    0x0A,0x0C,0x07,0,0x0B,0x0D,0x08,0,0x0C,0x0E,0x09,0,0x0D,0x0F,0x0A,0,
    0x0E,0x10,0x0B,0,0x0F,0x11,0x0C,0,0x10,0x12,0x0D,0,0x11,0x13,0x0E,0,
    0x12,0x14,0x0F,0,0x13,0x15,0x10,0,0x14,0x16,0x11,0,0x15,0x17,0x12,0,
    0x16,0x18,0x13,0,0x17,0x19,0x14,0,0x18,0x1A,0x15,0,0x19,0x1B,0x16,0,
    0x1A,0x1C,0x17,0,0x1B,0x1D,0x18,0,0x1C,0x1E,0x19,0,0x1D,0x1F,0x1A,0,
    0xDF,0xE0,0xDB,0,0xE0,0xE1,0xDC,0,0xE1,0xE2,0xDD,0,0xE2,0xE3,0xDE,0,
    0xE3,0xE4,0xDF,0,0xE4,0xE5,0xE0,0,0xE5,0xE6,0xE1,0,0xE6,0xE7,0xE2,0,
    0xE7,0xE8,0xE3,0,0xE8,0xE9,0xE4,0,0xE9,0xEA,0xE5,0,0xEA,0xEB,0xE6,0,
    0xEB,0xEC,0xE7,0,0xEC,0xED,0xE8,0,0xED,0xEE,0xE9,0,0xEE,0xEF,0xEA,0,
    0xEF,0xF0,0xEB,0,0xF0,0xF1,0xEC,0,0xF1,0xF2,0xED,0,0xF2,0xF3,0xEE,0,
    0xF3,0xF4,0xEF,0,0xF4,0xF5,0xF0,0,0xF5,0xF6,0xF1,0,0xF6,0xF7,0xF2,0,
    0xF7,0xF8,0xF3,0,0xF8,0xF9,0xF4,0,0xF9,0xFA,0xF5,0,0xFA,0xFB,0xF6,0,
    0xFB,0xFC,0xF7,0,0xFC,0xFD,0xF8,0,0xFD,0xFE,0xF9,0,0xFE,0xFF,0xFA,0,
    0xFF,0x00,0xFB,0,0x00,0x01,0xFC,0,0x01,0x02,0xFD,0,0x02,0x03,0xFE,0,
    0x03,0x04,0xFF,0,0x04,0x05,0x00,0,0x05,0x06,0x01,0,0x06,0x07,0x02,0,
    0x07,0x08,0x03,0,0x08,0x09,0x04,0,0x09,0x0A,0x05,0,0x0A,0x0B,0x06,0,
    0x0B,0x0C,0x07,0,0x0C,0x0D,0x08,0,0x0D,0x0E,0x09,0,0x0E,0x0F,0x0A,0,
    0x0F,0x10,0x0B,0,0x10,0x11,0x0C,0,0x11,0x12,0x0D,0,0x12,0x13,0x0E,0,
    0x13,0x14,0x0F,0,0x14,0x15,0x10,0,0x15,0x16,0x11,0,0x16,0x17,0x12,0,
    0x17,0x18,0x13,0,0x18,0x19,0x14,0,0x19,0x1A,0x15,0,0x1A,0x1B,0x16,0,
    0x1B,0x1C,0x17,0,0x1C,0x1D,0x18,0,0x1D,0x1E,0x19,0,0x1E,0x1F,0x1A,0,
    0xE0,0xE0,0xDB,0,0xE1,0xE1,0xDC,0,0xE2,0xE2,0xDD,0,0xE3,0xE3,0xDE,0,
    0xE4,0xE4,0xDF,0,0xE5,0xE5,0xE0,0,0xE6,0xE6,0xE1,0,0xE7,0xE7,0xE2,0,
    0xE8,0xE8,0xE3,0,0xE9,0xE9,0xE4,0,0xEA,0xEA,0xE5,0,0xEB,0xEB,0xE6,0,
    0xEC,0xEC,0xE7,0,0xED,0xED,0xE8,0,0xEE,0xEE,0xE9,0,0xEF,0xEF,0xEA,0,
    0xF0,0xF0,0xEB,0,0xF1,0xF1,0xEC,0,0xF2,0xF2,0xED,0,0xF3,0xF3,0xEE,0,
    0xF4,0xF4,0xEF,0,0xF5,0xF5,0xF0,0,0xF6,0xF6,0xF1,0,0xF7,0xF7,0xF2,0,
    0xF8,0xF8,0xF3,0,0xF9,0xF9,0xF4,0,0xFA,0xFA,0xF5,0,0xFB,0xFB,0xF6,0,
    0xFC,0xFC,0xF7,0,0xFD,0xFD,0xF8,0,0xFE,0xFE,0xF9,0,0xFF,0xFF,0xFA,0,
    0x00,0x00,0xFB,0,0x01,0x01,0xFC,0,0x02,0x02,0xFD,0,0x03,0x03,0xFE,0,
    0x04,0x04,0xFF,0,0x05,0x05,0x00,0,0x06,0x06,0x01,0,0x07,0x07,0x02,0,
    0x08,0x08,0x03,0,0x09,0x09,0x04,0,0x0A,0x0A,0x05,0,0x0B,0x0B,0x06,0,
    0x0C,0x0C,0x07,0,0x0D,0x0D,0x08,0,0x0E,0x0E,0x09,0,0x0F,0x0F,0x0A,0,
    0x10,0x10,0x0B,0,0x11,0x11,0x0C,0,0x12,0x12,0x0D,0,0x13,0x13,0x0E,0,
    0x14,0x14,0x0F,0,0x15,0x15,0x10,0,0x16,0x16,0x11,0,0x17,0x17,0x12,0,
    0x18,0x18,0x13,0,0x19,0x19,0x14,0,0x1A,0x1A,0x15,0,0x1B,0x1B,0x16,0,
    0x1C,0x1C,0x17,0,0x1D,0x1D,0x18,0,0x1E,0x1E,0x19,0,0x1F,0x1F,0x1A,0,
    0xE1,0xE0,0xDB,0,0xE2,0xE1,0xDC,0,0xE3,0xE2,0xDD,0,0xE4,0xE3,0xDE,0,
    0xE5,0xE4,0xDF,0,0xE6,0xE5,0xE0,0,0xE7,0xE6,0xE1,0,0xE8,0xE7,0xE2,0,
    0xE9,0xE8,0xE3,0,0xEA,0xE9,0xE4,0,0xEB,0xEA,0xE5,0,0xEC,0xEB,0xE6,0,
    0xED,0xEC,0xE7,0,0xEE,0xED,0xE8,0,0xEF,0xEE,0xE9,0,0xF0,0xEF,0xEA,0,
    0xF1,0xF0,0xEB,0,0xF2,0xF1,0xEC,0,0xF3,0xF2,0xED,0,0xF4,0xF3,0xEE,0,
    0xF5,0xF4,0xEF,0,0xF6,0xF5,0xF0,0,0xF7,0xF6,0xF1,0,0xF8,0xF7,0xF2,0,
    0xF9,0xF8,0xF3,0,0xFA,0xF9,0xF4,0,0xFB,0xFA,0xF5,0,0xFC,0xFB,0xF6,0,
    0xFD,0xFC,0xF7,0,0xFE,0xFD,0xF8,0,0xFF,0xFE,0xF9,0,0x00,0xFF,0xFA,0,
    0x01,0x00,0xFB,0,0x02,0x01,0xFC,0,0x03,0x02,0xFD,0,0x04,0x03,0xFE,0,
    0x05,0x04,0xFF,0,0x06,0x05,0x00,0,0x07,0x06,0x01,0,0x08,0x07,0x02,0,
    0x09,0x08,0x03,0,0x0A,0x09,0x04,0,0x0B,0x0A,0x05,0,0x0C,0x0B,0x06,0,
    0x0D,0x0C,0x07,0,0x0E,0x0D,0x08,0,0x0F,0x0E,0x09,0,0x10,0x0F,0x0A,0,
    0x11,0x10,0x0B,0,0x12,0x11,0x0C,0,0x13,0x12,0x0D,0,0x14,0x13,0x0E,0,
    0x15,0x14,0x0F,0,0x16,0x15,0x10,0,0x17,0x16,0x11,0,0x18,0x17,0x12,0,
    0x19,0x18,0x13,0,0x1A,0x19,0x14,0,0x1B,0x1A,0x15,0,0x1C,0x1B,0x16,0,
    0x1D,0x1C,0x17,0,0x1E,0x1D,0x18,0,0x1F,0x1E,0x19,0,0x20,0x1F,0x1A,0,
    0xE2,0xE0,0xDB,0,0xE3,0xE1,0xDC,0,0xE4,0xE2,0xDD,0,0xE5,0xE3,0xDE,0,
    0xE6,0xE4,0xDF,0,0xE7,0xE5,0xE0,0,0xE8,0xE6,0xE1,0,0xE9,0xE7,0xE2,0,
    0xEA,0xE8,0xE3,0,0xEB,0xE9,0xE4,0,0xEC,0xEA,0xE5,0,0xED,0xEB,0xE6,0,
    0xEE,0xEC,0xE7,0,0xEF,0xED,0xE8,0,0xF0,0xEE,0xE9,0,0xF1,0xEF,0xEA,0,
    0xF2,0xF0,0xEB,0,0xF3,0xF1,0xEC,0,0xF4,0xF2,0xED,0,0xF5,0xF3,0xEE,0,
    0xF6,0xF4,0xEF,0,0xF7,0xF5,0xF0,0,0xF8,0xF6,0xF1,0,0xF9,0xF7,0xF2,0,
    0xFA,0xF8,0xF3,0,0xFB,0xF9,0xF4,0,0xFC,0xFA,0xF5,0,0xFD,0xFB,0xF6,0,
    0xFE,0xFC,0xF7,0,0xFF,0xFD,0xF8,0,0x00,0xFE,0xF9,0,0x01,0xFF,0xFA,0,
    0x02,0x00,0xFB,0,0x03,0x01,0xFC,0,0x04,0x02,0xFD,0,0x05,0x03,0xFE,0,
    0x06,0x04,0xFF,0,0x07,0x05,0x00,0,0x08,0x06,0x01,0,0x09,0x07,0x02,0,
    0x0A,0x08,0x03,0,0x0B,0x09,0x04,0,0x0C,0x0A,0x05,0,0x0D,0x0B,0x06,0,
    0x0E,0x0C,0x07,0,0x0F,0x0D,0x08,0,0x10,0x0E,0x09,0,0x11,0x0F,0x0A,0,
    0x12,0x10,0x0B,0,0x13,0x11,0x0C,0,0x14,0x12,0x0D,0,0x15,0x13,0x0E,0,
    0x16,0x14,0x0F,0,0x17,0x15,0x10,0,0x18,0x16,0x11,0,0x19,0x17,0x12,0,
    0x1A,0x18,0x13,0,0x1B,0x19,0x14,0,0x1C,0x1A,0x15,0,0x1D,0x1B,0x16,0,
    0x1E,0x1C,0x17,0,0x1F,0x1D,0x18,0,0x20,0x1E,0x19,0,0x21,0x1F,0x1A,0,
    0xE3,0xE0,0xDB,0,0xE4,0xE1,0xDC,0,0xE5,0xE2,0xDD,0,0xE6,0xE3,0xDE,0,
    0xE7,0xE4,0xDF,0,0xE8,0xE5,0xE0,0,0xE9,0xE6,0xE1,0,0xEA,0xE7,0xE2,0,
    0xEB,0xE8,0xE3,0,0xEC,0xE9,0xE4,0,0xED,0xEA,0xE5,0,0xEE,0xEB,0xE6,0,
    0xEF,0xEC,0xE7,0,0xF0,0xED,0xE8,0,0xF1,0xEE,0xE9,0,0xF2,0xEF,0xEA,0,
    0xF3,0xF0,0xEB,0,0xF4,0xF1,0xEC,0,0xF5,0xF2,0xED,0,0xF6,0xF3,0xEE,0,
    0xF7,0xF4,0xEF,0,0xF8,0xF5,0xF0,0,0xF9,0xF6,0xF1,0,0xFA,0xF7,0xF2,0,
    0xFB,0xF8,0xF3,0,0xFC,0xF9,0xF4,0,0xFD,0xFA,0xF5,0,0xFE,0xFB,0xF6,0,
    0xFF,0xFC,0xF7,0,0x00,0xFD,0xF8,0,0x01,0xFE,0xF9,0,0x02,0xFF,0xFA,0,
    0x03,0x00,0xFB,0,0x04,0x01,0xFC,0,0x05,0x02,0xFD,0,0x06,0x03,0xFE,0,
    0x07,0x04,0xFF,0,0x08,0x05,0x00,0,0x09,0x06,0x01,0,0x0A,0x07,0x02,0,
    0x0B,0x08,0x03,0,0x0C,0x09,0x04,0,0x0D,0x0A,0x05,0,0x0E,0x0B,0x06,0,
    0x0F,0x0C,0x07,0,0x10,0x0D,0x08,0,0x11,0x0E,0x09,0,0x12,0x0F,0x0A,0,
    0x13,0x10,0x0B,0,0x14,0x11,0x0C,0,0x15,0x12,0x0D,0,0x16,0x13,0x0E,0,
    0x17,0x14,0x0F,0,0x18,0x15,0x10,0,0x19,0x16,0x11,0,0x1A,0x17,0x12,0,
    0x1B,0x18,0x13,0,0x1C,0x19,0x14,0,0x1D,0x1A,0x15,0,0x1E,0x1B,0x16,0,
    0x1F,0x1C,0x17,0,0x20,0x1D,0x18,0,0x21,0x1E,0x19,0,0x22,0x1F,0x1A,0,
    0xE4,0xE0,0xDB,0,0xE5,0xE1,0xDC,0,0xE6,0xE2,0xDD,0,0xE7,0xE3,0xDE,0,
    0xE8,0xE4,0xDF,0,0xE9,0xE5,0xE0,0,0xEA,0xE6,0xE1,0,0xEB,0xE7,0xE2,0,
    0xEC,0xE8,0xE3,0,0xED,0xE9,0xE4,0,0xEE,0xEA,0xE5,0,0xEF,0xEB,0xE6,0,
    0xF0,0xEC,0xE7,0,0xF1,0xED,0xE8,0,0xF2,0xEE,0xE9,0,0xF3,0xEF,0xEA,0,
    0xF4,0xF0,0xEB,0,0xF5,0xF1,0xEC,0,0xF6,0xF2,0xED,0,0xF7,0xF3,0xEE,0,
    0xF8,0xF4,0xEF,0,0xF9,0xF5,0xF0,0,0xFA,0xF6,0xF1,0,0xFB,0xF7,0xF2,0,
    0xFC,0xF8,0xF3,0,0xFD,0xF9,0xF4,0,0xFE,0xFA,0xF5,0,0xFF,0xFB,0xF6,0,
    0x00,0xFC,0xF7,0,0x01,0xFD,0xF8,0,0x02,0xFE,0xF9,0,0x03,0xFF,0xFA,0,
    0x04,0x00,0xFB,0,0x05,0x01,0xFC,0,0x06,0x02,0xFD,0,0x07,0x03,0xFE,0,
    0x08,0x04,0xFF,0,0x09,0x05,0x00,0,0x0A,0x06,0x01,0,0x0B,0x07,0x02,0,
    0x0C,0x08,0x03,0,0x0D,0x09,0x04,0,0x0E,0x0A,0x05,0,0x0F,0x0B,0x06,0,
    0x10,0x0C,0x07,0,0x11,0x0D,0x08,0,0x12,0x0E,0x09,0,0x13,0x0F,0x0A,0,
    0x14,0x10,0x0B,0,0x15,0x11,0x0C,0,0x16,0x12,0x0D,0,0x17,0x13,0x0E,0,
    0x18,0x14,0x0F,0,0x19,0x15,0x10,0,0x1A,0x16,0x11,0,0x1B,0x17,0x12,0,
    0x1C,0x18,0x13,0,0x1D,0x19,0x14,0,0x1E,0x1A,0x15,0,0x1F,0x1B,0x16,0,
    0x20,0x1C,0x17,0,0x21,0x1D,0x18,0,0x22,0x1E,0x19,0,0x23,0x1F,0x1A,0,
    0xE5,0xE0,0xDB,0,0xE6,0xE1,0xDC,0,0xE7,0xE2,0xDD,0,0xE8,0xE3,0xDE,0,
    0xE9,0xE4,0xDF,0,0xEA,0xE5,0xE0,0,0xEB,0xE6,0xE1,0,0xEC,0xE7,0xE2,0,
    0xED,0xE8,0xE3,0,0xEE,0xE9,0xE4,0,0xEF,0xEA,0xE5,0,0xF0,0xEB,0xE6,0,
    0xF1,0xEC,0xE7,0,0xF2,0xED,0xE8,0,0xF3,0xEE,0xE9,0,0xF4,0xEF,0xEA,0,
    0xF5,0xF0,0xEB,0,0xF6,0xF1,0xEC,0,0xF7,0xF2,0xED,0,0xF8,0xF3,0xEE,0,
    0xF9,0xF4,0xEF,0,0xFA,0xF5,0xF0,0,0xFB,0xF6,0xF1,0,0xFC,0xF7,0xF2,0,
    0xFD,0xF8,0xF3,0,0xFE,0xF9,0xF4,0,0xFF,0xFA,0xF5,0,0x00,0xFB,0xF6,0,
    0x01,0xFC,0xF7,0,0x02,0xFD,0xF8,0,0x03,0xFE,0xF9,0,0x04,0xFF,0xFA,0,
    0x05,0x00,0xFB,0,0x06,0x01,0xFC,0,0x07,0x02,0xFD,0,0x08,0x03,0xFE,0,
    0x09,0x04,0xFF,0,0x0A,0x05,0x00,0,0x0B,0x06,0x01,0,0x0C,0x07,0x02,0,
    0x0D,0x08,0x03,0,0x0E,0x09,0x04,0,0x0F,0x0A,0x05,0,0x10,0x0B,0x06,0,
    0x11,0x0C,0x07,0,0x12,0x0D,0x08,0,0x13,0x0E,0x09,0,0x14,0x0F,0x0A,0,
    0x15,0x10,0x0B,0,0x16,0x11,0x0C,0,0x17,0x12,0x0D,0,0x18,0x13,0x0E,0,
    0x19,0x14,0x0F,0,0x1A,0x15,0x10,0,0x1B,0x16,0x11,0,0x1C,0x17,0x12,0,
    0x1D,0x18,0x13,0,0x1E,0x19,0x14,0,0x1F,0x1A,0x15,0,0x20,0x1B,0x16,0,
    0x21,0x1C,0x17,0,0x22,0x1D,0x18,0,0x23,0x1E,0x19,0,0x24,0x1F,0x1A,0,
    0xE6,0xE0,0xDB,0,0xE7,0xE1,0xDC,0,0xE8,0xE2,0xDD,0,0xE9,0xE3,0xDE,0,
    0xEA,0xE4,0xDF,0,0xEB,0xE5,0xE0,0,0xEC,0xE6,0xE1,0,0xED,0xE7,0xE2,0,
    0xEE,0xE8,0xE3,0,0xEF,0xE9,0xE4,0,0xF0,0xEA,0xE5,0,0xF1,0xEB,0xE6,0,
    0xF2,0xEC,0xE7,0,0xF3,0xED,0xE8,0,0xF4,0xEE,0xE9,0,0xF5,0xEF,0xEA,0,
    0xF6,0xF0,0xEB,0,0xF7,0xF1,0xEC,0,0xF8,0xF2,0xED,0,0xF9,0xF3,0xEE,0,
    0xFA,0xF4,0xEF,0,0xFB,0xF5,0xF0,0,0xFC,0xF6,0xF1,0,0xFD,0xF7,0xF2,0,
    0xFE,0xF8,0xF3,0,0xFF,0xF9,0xF4,0,0x00,0xFA,0xF5,0,0x01,0xFB,0xF6,0,
    0x02,0xFC,0xF7,0,0x03,0xFD,0xF8,0,0x04,0xFE,0xF9,0,0x05,0xFF,0xFA,0,
    0x06,0x00,0xFB,0,0x07,0x01,0xFC,0,0x08,0x02,0xFD,0,0x09,0x03,0xFE,0,
    0x0A,0x04,0xFF,0,0x0B,0x05,0x00,0,0x0C,0x06,0x01,0,0x0D,0x07,0x02,0,
    0x0E,0x08,0x03,0,0x0F,0x09,0x04,0,0x10,0x0A,0x05,0,0x11,0x0B,0x06,0,
    0x12,0x0C,0x07,0,0x13,0x0D,0x08,0,0x14,0x0E,0x09,0,0x15,0x0F,0x0A,0,
    0x16,0x10,0x0B,0,0x17,0x11,0x0C,0,0x18,0x12,0x0D,0,0x19,0x13,0x0E,0,
    0x1A,0x14,0x0F,0,0x1B,0x15,0x10,0,0x1C,0x16,0x11,0,0x1D,0x17,0x12,0,
    0x1E,0x18,0x13,0,0x1F,0x19,0x14,0,0x20,0x1A,0x15,0,0x21,0x1B,0x16,0,
    0x22,0x1C,0x17,0,0x23,0x1D,0x18,0,0x24,0x1E,0x19,0,0x25,0x1F,0x1A,0,
    0xE7,0xE0,0xDB,0,0xE8,0xE1,0xDC,0,0xE9,0xE2,0xDD,0,0xEA,0xE3,0xDE,0,
    0xEB,0xE4,0xDF,0,0xEC,0xE5,0xE0,0,0xED,0xE6,0xE1,0,0xEE,0xE7,0xE2,0,
    0xEF,0xE8,0xE3,0,0xF0,0xE9,0xE4,0,0xF1,0xEA,0xE5,0,0xF2,0xEB,0xE6,0,
    0xF3,0xEC,0xE7,0,0xF4,0xED,0xE8,0,0xF5,0xEE,0xE9,0,0xF6,0xEF,0xEA,0,
    0xF7,0xF0,0xEB,0,0xF8,0xF1,0xEC,0,0xF9,0xF2,0xED,0,0xFA,0xF3,0xEE,0,
    0xFB,0xF4,0xEF,0,0xFC,0xF5,0xF0,0,0xFD,0xF6,0xF1,0,0xFE,0xF7,0xF2,0,
    0xFF,0xF8,0xF3,0,0x00,0xF9,0xF4,0,0x01,0xFA,0xF5,0,0x02,0xFB,0xF6,0,
    0x03,0xFC,0xF7,0,0x04,0xFD,0xF8,0,0x05,0xFE,0xF9,0,0x06,0xFF,0xFA,0,
    0x07,0x00,0xFB,0,0x08,0x01,0xFC,0,0x09,0x02,0xFD,0,0x0A,0x03,0xFE,0,
    0x0B,0x04,0xFF,0,0x0C,0x05,0x00,0,0x0D,0x06,0x01,0,0x0E,0x07,0x02,0,
    0x0F,0x08,0x03,0,0x10,0x09,0x04,0,0x11,0x0A,0x05,0,0x12,0x0B,0x06,0,
    0x13,0x0C,0x07,0,0x14,0x0D,0x08,0,0x15,0x0E,0x09,0,0x16,0x0F,0x0A,0,
    0x17,0x10,0x0B,0,0x18,0x11,0x0C,0,0x19,0x12,0x0D,0,0x1A,0x13,0x0E,0,
    0x1B,0x14,0x0F,0,0x1C,0x15,0x10,0,0x1D,0x16,0x11,0,0x1E,0x17,0x12,0,
    0x1F,0x18,0x13,0,0x20,0x19,0x14,0,0x21,0x1A,0x15,0,0x22,0x1B,0x16,0,
    0x23,0x1C,0x17,0,0x24,0x1D,0x18,0,0x25,0x1E,0x19,0,0x26,0x1F,0x1A,0,
    0xD8,0xE0,0xDC,0,0xD9,0xE1,0xDD,0,0xDA,0xE2,0xDE,0,0xDB,0xE3,0xDF,0,
    0xDC,0xE4,0xE0,0,0xDD,0xE5,0xE1,0,0xDE,0xE6,0xE2,0,0xDF,0xE7,0xE3,0,
    0xE0,0xE8,0xE4,0,0xE1,0xE9,0xE5,0,0xE2,0xEA,0xE6,0,0xE3,0xEB,0xE7,0,
    0xE4,0xEC,0xE8,0,0xE5,0xED,0xE9,0,0xE6,0xEE,0xEA,0,0xE7,0xEF,0xEB,0,
    0xE8,0xF0,0xEC,0,0xE9,0xF1,0xED,0,0xEA,0xF2,0xEE,0,0xEB,0xF3,0xEF,0,
    0xEC,0xF4,0xF0,0,0xED,0xF5,0xF1,0,0xEE,0xF6,0xF2,0,0xEF,0xF7,0xF3,0,
    0xF0,0xF8,0xF4,0,0xF1,0xF9,0xF5,0,0xF2,0xFA,0xF6,0,0xF3,0xFB,0xF7,0,
    0xF4,0xFC,0xF8,0,0xF5,0xFD,0xF9,0,0xF6,0xFE,0xFA,0,0xF7,0xFF,0xFB,0,
    0xF8,0x00,0xFC,0,0xF9,0x01,0xFD,0,0xFA,0x02,0xFE,0,0xFB,0x03,0xFF,0,
    0xFC,0x04,0x00,0,0xFD,0x05,0x01,0,0xFE,0x06,0x02,0,0xFF,0x07,0x03,0,
    0x00,0x08,0x04,0,0x01,0x09,0x05,0,0x02,0x0A,0x06,0,0x03,0x0B,0x07,0,
    0x04,0x0C,0x08,0,0x05,0x0D,0x09,0,0x06,0x0E,0x0A,0,0x07,0x0F,0x0B,0,
    0x08,0x10,0x0C,0,0x09,0x11,0x0D,0,0x0A,0x12,0x0E,0,0x0B,0x13,0x0F,0,
    0x0C,0x14,0x10,0,0x0D,0x15,0x11,0,0x0E,0x16,0x12,0,0x0F,0x17,0x13,0,
    0x10,0x18,0x14,0,0x11,0x19,0x15,0,0x12,0x1A,0x16,0,0x13,0x1B,0x17,0,
    0x14,0x1C,0x18,0,0x15,0x1D,0x19,0,0x16,0x1E,0x1A,0,0x17,0x1F,0x1B,0,
    0xD9,0xE0,0xDC,0,0xDA,0xE1,0xDD,0,0xDB,0xE2,0xDE,0,0xDC,0xE3,0xDF,0,
    0xDD,0xE4,0xE0,0,0xDE,0xE5,0xE1,0,0xDF,0xE6,0xE2,0,0xE0,0xE7,0xE3,0,
    0xE1,0xE8,0xE4,0,0xE2,0xE9,0xE5,0,0xE3,0xEA,0xE6,0,0xE4,0xEB,0xE7,0,
    0xE5,0xEC,0xE8,0,0xE6,0xED,0xE9,0,0xE7,0xEE,0xEA,0,0xE8,0xEF,0xEB,0,
    0xE9,0xF0,0xEC,0,0xEA,0xF1,0xED,0,0xEB,0xF2,0xEE,0,0xEC,0xF3,0xEF,0,
    0xED,0xF4,0xF0,0,0xEE,0xF5,0xF1,0,0xEF,0xF6,0xF2,0,0xF0,0xF7,0xF3,0,
    0xF1,0xF8,0xF4,0,0xF2,0xF9,0xF5,0,0xF3,0xFA,0xF6,0,0xF4,0xFB,0xF7,0,
    0xF5,0xFC,0xF8,0,0xF6,0xFD,0xF9,0,0xF7,0xFE,0xFA,0,0xF8,0xFF,0xFB,0,
    0xF9,0x00,0xFC,0,0xFA,0x01,0xFD,0,0xFB,0x02,0xFE,0,0xFC,0x03,0xFF,0,
    0xFD,0x04,0x00,0,0xFE,0x05,0x01,0,0xFF,0x06,0x02,0,0x00,0x07,0x03,0,
    0x01,0x08,0x04,0,0x02,0x09,0x05,0,0x03,0x0A,0x06,0,0x04,0x0B,0x07,0,
    0x05,0x0C,0x08,0,0x06,0x0D,0x09,0,0x07,0x0E,0x0A,0,0x08,0x0F,0x0B,0,
    0x09,0x10,0x0C,0,0x0A,0x11,0x0D,0,0x0B,0x12,0x0E,0,0x0C,0x13,0x0F,0,
    0x0D,0x14,0x10,0,0x0E,0x15,0x11,0,0x0F,0x16,0x12,0,0x10,0x17,0x13,0,
    0x11,0x18,0x14,0,0x12,0x19,0x15,0,0x13,0x1A,0x16,0,0x14,0x1B,0x17,0,
    0x15,0x1C,0x18,0,0x16,0x1D,0x19,0,0x17,0x1E,0x1A,0,0x18,0x1F,0x1B,0,
    0xDA,0xE0,0xDC,0,0xDB,0xE1,0xDD,0,0xDC,0xE2,0xDE,0,0xDD,0xE3,0xDF,0,
    0xDE,0xE4,0xE0,0,0xDF,0xE5,0xE1,0,0xE0,0xE6,0xE2,0,0xE1,0xE7,0xE3,0,
    0xE2,0xE8,0xE4,0,0xE3,0xE9,0xE5,0,0xE4,0xEA,0xE6,0,0xE5,0xEB,0xE7,0,
    0xE6,0xEC,0xE8,0,0xE7,0xED,0xE9,0,0xE8,0xEE,0xEA,0,0xE9,0xEF,0xEB,0,
    0xEA,0xF0,0xEC,0,0xEB,0xF1,0xED,0,0xEC,0xF2,0xEE,0,0xED,0xF3,0xEF,0,
    0xEE,0xF4,0xF0,0,0xEF,0xF5,0xF1,0,0xF0,0xF6,0xF2,0,0xF1,0xF7,0xF3,0,
    0xF2,0xF8,0xF4,0,0xF3,0xF9,0xF5,0,0xF4,0xFA,0xF6,0,0xF5,0xFB,0xF7,0,
    0xF6,0xFC,0xF8,0,0xF7,0xFD,0xF9,0,0xF8,0xFE,0xFA,0,0xF9,0xFF,0xFB,0,
    0xFA,0x00,0xFC,0,0xFB,0x01,0xFD,0,0xFC,0x02,0xFE,0,0xFD,0x03,0xFF,0,
    0xFE,0x04,0x00,0,0xFF,0x05,0x01,0,0x00,0x06,0x02,0,0x01,0x07,0x03,0,
    0x02,0x08,0x04,0,0x03,0x09,0x05,0,0x04,0x0A,0x06,0,0x05,0x0B,0x07,0,
    0x06,0x0C,0x08,0,0x07,0x0D,0x09,0,0x08,0x0E,0x0A,0,0x09,0x0F,0x0B,0,
    0x0A,0x10,0x0C,0,0x0B,0x11,0x0D,0,0x0C,0x12,0x0E,0,0x0D,0x13,0x0F,0,
    0x0E,0x14,0x10,0,0x0F,0x15,0x11,0,0x10,0x16,0x12,0,0x11,0x17,0x13,0,
    0x12,0x18,0x14,0,0x13,0x19,0x15,0,0x14,0x1A,0x16,0,0x15,0x1B,0x17,0,
    0x16,0x1C,0x18,0,0x17,0x1D,0x19,0,0x18,0x1E,0x1A,0,0x19,0x1F,0x1B,0,
    0xDB,0xE0,0xDC,0,0xDC,0xE1,0xDD,0,0xDD,0xE2,0xDE,0,0xDE,0xE3,0xDF,0,
    0xDF,0xE4,0xE0,0,0xE0,0xE5,0xE1,0,0xE1,0xE6,0xE2,0,0xE2,0xE7,0xE3,0,
    0xE3,0xE8,0xE4,0,0xE4,0xE9,0xE5,0,0xE5,0xEA,0xE6,0,0xE6,0xEB,0xE7,0,
    0xE7,0xEC,0xE8,0,0xE8,0xED,0xE9,0,0xE9,0xEE,0xEA,0,0xEA,0xEF,0xEB,0,
    0xEB,0xF0,0xEC,0,0xEC,0xF1,0xED,0,0xED,0xF2,0xEE,0,0xEE,0xF3,0xEF,0,
    0xEF,0xF4,0xF0,0,0xF0,0xF5,0xF1,0,0xF1,0xF6,0xF2,0,0xF2,0xF7,0xF3,0,
    0xF3,0xF8,0xF4,0,0xF4,0xF9,0xF5,0,0xF5,0xFA,0xF6,0,0xF6,0xFB,0xF7,0,
    0xF7,0xFC,0xF8,0,0xF8,0xFD,0xF9,0,0xF9,0xFE,0xFA,0,0xFA,0xFF,0xFB,0,
    0xFB,0x00,0xFC,0,0xFC,0x01,0xFD,0,0xFD,0x02,0xFE,0,0xFE,0x03,0xFF,0,
    0xFF,0x04,0x00,0,0x00,0x05,0x01,0,0x01,0x06,0x02,0,0x02,0x07,0x03,0,
    0x03,0x08,0x04,0,0x04,0x09,0x05,0,0x05,0x0A,0x06,0,0x06,0x0B,0x07,0,
    0x07,0x0C,0x08,0,0x08,0x0D,0x09,0,0x09,0x0E,0x0A,0,0x0A,0x0F,0x0B,0,
    0x0B,0x10,0x0C,0,0x0C,0x11,0x0D,0,0x0D,0x12,0x0E,0,0x0E,0x13,0x0F,0,
    0x0F,0x14,0x10,0,0x10,0x15,0x11,0,0x11,0x16,0x12,0,0x12,0x17,0x13,0,
    0x13,0x18,0x14,0,0x14,0x19,0x15,0,0x15,0x1A,0x16,0,0x16,0x1B,0x17,0,
    0x17,0x1C,0x18,0,0x18,0x1D,0x19,0,0x19,0x1E,0x1A,0,0x1A,0x1F,0x1B,0,
    0xDC,0xE0,0xDC,0,0xDD,0xE1,0xDD,0,0xDE,0xE2,0xDE,0,0xDF,0xE3,0xDF,0,
    0xE0,0xE4,0xE0,0,0xE1,0xE5,0xE1,0,0xE2,0xE6,0xE2,0,0xE3,0xE7,0xE3,0,
    0xE4,0xE8,0xE4,0,0xE5,0xE9,0xE5,0,0xE6,0xEA,0xE6,0,0xE7,0xEB,0xE7,0,
    0xE8,0xEC,0xE8,0,0xE9,0xED,0xE9,0,0xEA,0xEE,0xEA,0,0xEB,0xEF,0xEB,0,
    0xEC,0xF0,0xEC,0,0xED,0xF1,0xED,0,0xEE,0xF2,0xEE,0,0xEF,0xF3,0xEF,0,
    0xF0,0xF4,0xF0,0,0xF1,0xF5,0xF1,0,0xF2,0xF6,0xF2,0,0xF3,0xF7,0xF3,0,
    0xF4,0xF8,0xF4,0,0xF5,0xF9,0xF5,0,0xF6,0xFA,0xF6,0,0xF7,0xFB,0xF7,0,
    0xF8,0xFC,0xF8,0,0xF9,0xFD,0xF9,0,0xFA,0xFE,0xFA,0,0xFB,0xFF,0xFB,0,
    0xFC,0x00,0xFC,0,0xFD,0x01,0xFD,0,0xFE,0x02,0xFE,0,0xFF,0x03,0xFF,0,
    0x00,0x04,0x00,0,0x01,0x05,0x01,0,0x02,0x06,0x02,0,0x03,0x07,0x03,0,
    0x04,0x08,0x04,0,0x05,0x09,0x05,0,0x06,0x0A,0x06,0,0x07,0x0B,0x07,0,
    0x08,0x0C,0x08,0,0x09,0x0D,0x09,0,0x0A,0x0E,0x0A,0,0x0B,0x0F,0x0B,0,
    0x0C,0x10,0x0C,0,0x0D,0x11,0x0D,0,0x0E,0x12,0x0E,0,0x0F,0x13,0x0F,0,
    0x10,0x14,0x10,0,0x11,0x15,0x11,0,0x12,0x16,0x12,0,0x13,0x17,0x13,0,
    0x14,0x18,0x14,0,0x15,0x19,0x15,0,0x16,0x1A,0x16,0,0x17,0x1B,0x17,0,
    0x18,0x1C,0x18,0,0x19,0x1D,0x19,0,0x1A,0x1E,0x1A,0,0x1B,0x1F,0x1B,0,
    0xDD,0xE0,0xDC,0,0xDE,0xE1,0xDD,0,0xDF,0xE2,0xDE,0,0xE0,0xE3,0xDF,0,
    0xE1,0xE4,0xE0,0,0xE2,0xE5,0xE1,0,0xE3,0xE6,0xE2,0,0xE4,0xE7,0xE3,0,
    0xE5,0xE8,0xE4,0,0xE6,0xE9,0xE5,0,0xE7,0xEA,0xE6,0,0xE8,0xEB,0xE7,0,
    0xE9,0xEC,0xE8,0,0xEA,0xED,0xE9,0,0xEB,0xEE,0xEA,0,0xEC,0xEF,0xEB,0,
    0xED,0xF0,0xEC,0,0xEE,0xF1,0xED,0,0xEF,0xF2,0xEE,0,0xF0,0xF3,0xEF,0,
    0xF1,0xF4,0xF0,0,0xF2,0xF5,0xF1,0,0xF3,0xF6,0xF2,0,0xF4,0xF7,0xF3,0,
    0xF5,0xF8,0xF4,0,0xF6,0xF9,0xF5,0,0xF7,0xFA,0xF6,0,0xF8,0xFB,0xF7,0,
    0xF9,0xFC,0xF8,0,0xFA,0xFD,0xF9,0,0xFB,0xFE,0xFA,0,0xFC,0xFF,0xFB,0,
    0xFD,0x00,0xFC,0,0xFE,0x01,0xFD,0,0xFF,0x02,0xFE,0,0x00,0x03,0xFF,0,
    0x01,0x04,0x00,0,0x02,0x05,0x01,0,0x03,0x06,0x02,0,0x04,0x07,0x03,0,
    0x05,0x08,0x04,0,0x06,0x09,0x05,0,0x07,0x0A,0x06,0,0x08,0x0B,0x07,0,
    0x09,0x0C,0x08,0,0x0A,0x0D,0x09,0,0x0B,0x0E,0x0A,0,0x0C,0x0F,0x0B,0,
    0x0D,0x10,0x0C,0,0x0E,0x11,0x0D,0,0x0F,0x12,0x0E,0,0x10,0x13,0x0F,0,
    0x11,0x14,0x10,0,0x12,0x15,0x11,0,0x13,0x16,0x12,0,0x14,0x17,0x13,0,
    0x15,0x18,0x14,0,0x16,0x19,0x15,0,0x17,0x1A,0x16,0,0x18,0x1B,0x17,0,
    0x19,0x1C,0x18,0,0x1A,0x1D,0x19,0,0x1B,0x1E,0x1A,0,0x1C,0x1F,0x1B,0,
    0xDE,0xE0,0xDC,0,0xDF,0xE1,0xDD,0,0xE0,0xE2,0xDE,0,0xE1,0xE3,0xDF,0,
    0xE2,0xE4,0xE0,0,0xE3,0xE5,0xE1,0,0xE4,0xE6,0xE2,0,0xE5,0xE7,0xE3,0,
    0xE6,0xE8,0xE4,0,0xE7,0xE9,0xE5,0,0xE8,0xEA,0xE6,0,0xE9,0xEB,0xE7,0,
    0xEA,0xEC,0xE8,0,0xEB,0xED,0xE9,0,0xEC,0xEE,0xEA,0,0xED,0xEF,0xEB,0,
    0xEE,0xF0,0xEC,0,0xEF,0xF1,0xED,0,0xF0,0xF2,0xEE,0,0xF1,0xF3,0xEF,0,
    0xF2,0xF4,0xF0,0,0xF3,0xF5,0xF1,0,0xF4,0xF6,0xF2,0,0xF5,0xF7,0xF3,0,
    0xF6,0xF8,0xF4,0,0xF7,0xF9,0xF5,0,0xF8,0xFA,0xF6,0,0xF9,0xFB,0xF7,0,
    0xFA,0xFC,0xF8,0,0xFB,0xFD,0xF9,0,0xFC,0xFE,0xFA,0,0xFD,0xFF,0xFB,0,
    0xFE,0x00,0xFC,0,0xFF,0x01,0xFD,0,0x00,0x02,0xFE,0,0x01,0x03,0xFF,0,
    0x02,0x04,0x00,0,0x03,0x05,0x01,0,0x04,0x06,0x02,0,0x05,0x07,0x03,0,
    0x06,0x08,0x04,0,0x07,0x09,0x05,0,0x08,0x0A,0x06,0,0x09,0x0B,0x07,0,
    0x0A,0x0C,0x08,0,0x0B,0x0D,0x09,0,0x0C,0x0E,0x0A,0,0x0D,0x0F,0x0B,0,
    0x0E,0x10,0x0C,0,0x0F,0x11,0x0D,0,0x10,0x12,0x0E,0,0x11,0x13,0x0F,0,
    0x12,0x14,0x10,0,0x13,0x15,0x11,0,0x14,0x16,0x12,0,0x15,0x17,0x13,0,
    0x16,0x18,0x14,0,0x17,0x19,0x15,0,0x18,0x1A,0x16,0,0x19,0x1B,0x17,0,
    0x1A,0x1C,0x18,0,0x1B,0x1D,0x19,0,0x1C,0x1E,0x1A,0,0x1D,0x1F,0x1B,0,
    0xDF,0xE0,0xDC,0,0xE0,0xE1,0xDD,0,0xE1,0xE2,0xDE,0,0xE2,0xE3,0xDF,0,
    0xE3,0xE4,0xE0,0,0xE4,0xE5,0xE1,0,0xE5,0xE6,0xE2,0,0xE6,0xE7,0xE3,0,
    0xE7,0xE8,0xE4,0,0xE8,0xE9,0xE5,0,0xE9,0xEA,0xE6,0,0xEA,0xEB,0xE7,0,
    0xEB,0xEC,0xE8,0,0xEC,0xED,0xE9,0,0xED,0xEE,0xEA,0,0xEE,0xEF,0xEB,0,
    0xEF,0xF0,0xEC,0,0xF0,0xF1,0xED,0,0xF1,0xF2,0xEE,0,0xF2,0xF3,0xEF,0,
    0xF3,0xF4,0xF0,0,0xF4,0xF5,0xF1,0,0xF5,0xF6,0xF2,0,0xF6,0xF7,0xF3,0,
    0xF7,0xF8,0xF4,0,0xF8,0xF9,0xF5,0,0xF9,0xFA,0xF6,0,0xFA,0xFB,0xF7,0,
    0xFB,0xFC,0xF8,0,0xFC,0xFD,0xF9,0,0xFD,0xFE,0xFA,0,0xFE,0xFF,0xFB,0,
    0xFF,0x00,0xFC,0,0x00,0x01,0xFD,0,0x01,0x02,0xFE,0,0x02,0x03,0xFF,0,
    0x03,0x04,0x00,0,0x04,0x05,0x01,0,0x05,0x06,0x02,0,0x06,0x07,0x03,0,
    0x07,0x08,0x04,0,0x08,0x09,0x05,0,0x09,0x0A,0x06,0,0x0A,0x0B,0x07,0,
    0x0B,0x0C,0x08,0,0x0C,0x0D,0x09,0,0x0D,0x0E,0x0A,0,0x0E,0x0F,0x0B,0,
    0x0F,0x10,0x0C,0,0x10,0x11,0x0D,0,0x11,0x12,0x0E,0,0x12,0x13,0x0F,0,
    0x13,0x14,0x10,0,0x14,0x15,0x11,0,0x15,0x16,0x12,0,0x16,0x17,0x13,0,
    0x17,0x18,0x14,0,0x18,0x19,0x15,0,0x19,0x1A,0x16,0,0x1A,0x1B,0x17,0,
    0x1B,0x1C,0x18,0,0x1C,0x1D,0x19,0,0x1D,0x1E,0x1A,0,0x1E,0x1F,0x1B,0,
    0xE0,0xE0,0xDC,0,0xE1,0xE1,0xDD,0,0xE2,0xE2,0xDE,0,0xE3,0xE3,0xDF,0,
    0xE4,0xE4,0xE0,0,0xE5,0xE5,0xE1,0,0xE6,0xE6,0xE2,0,0xE7,0xE7,0xE3,0,
    0xE8,0xE8,0xE4,0,0xE9,0xE9,0xE5,0,0xEA,0xEA,0xE6,0,0xEB,0xEB,0xE7,0,
    0xEC,0xEC,0xE8,0,0xED,0xED,0xE9,0,0xEE,0xEE,0xEA,0,0xEF,0xEF,0xEB,0,
    0xF0,0xF0,0xEC,0,0xF1,0xF1,0xED,0,0xF2,0xF2,0xEE,0,0xF3,0xF3,0xEF,0,
    0xF4,0xF4,0xF0,0,0xF5,0xF5,0xF1,0,0xF6,0xF6,0xF2,0,0xF7,0xF7,0xF3,0,
    0xF8,0xF8,0xF4,0,0xF9,0xF9,0xF5,0,0xFA,0xFA,0xF6,0,0xFB,0xFB,0xF7,0,
    0xFC,0xFC,0xF8,0,0xFD,0xFD,0xF9,0,0xFE,0xFE,0xFA,0,0xFF,0xFF,0xFB,0,
    0x00,0x00,0xFC,0,0x01,0x01,0xFD,0,0x02,0x02,0xFE,0,0x03,0x03,0xFF,0,
    0x04,0x04,0x00,0,0x05,0x05,0x01,0,0x06,0x06,0x02,0,0x07,0x07,0x03,0,
    0x08,0x08,0x04,0,0x09,0x09,0x05,0,0x0A,0x0A,0x06,0,0x0B,0x0B,0x07,0,
    0x0C,0x0C,0x08,0,0x0D,0x0D,0x09,0,0x0E,0x0E,0x0A,0,0x0F,0x0F,0x0B,0,
    0x10,0x10,0x0C,0,0x11,0x11,0x0D,0,0x12,0x12,0x0E,0,0x13,0x13,0x0F,0,
    0x14,0x14,0x10,0,0x15,0x15,0x11,0,0x16,0x16,0x12,0,0x17,0x17,0x13,0,
    0x18,0x18,0x14,0,0x19,0x19,0x15,0,0x1A,0x1A,0x16,0,0x1B,0x1B,0x17,0,
    0x1C,0x1C,0x18,0,0x1D,0x1D,0x19,0,0x1E,0x1E,0x1A,0,0x1F,0x1F,0x1B,0,
    0xE1,0xE0,0xDC,0,0xE2,0xE1,0xDD,0,0xE3,0xE2,0xDE,0,0xE4,0xE3,0xDF,0,
    0xE5,0xE4,0xE0,0,0xE6,0xE5,0xE1,0,0xE7,0xE6,0xE2,0,0xE8,0xE7,0xE3,0,
    0xE9,0xE8,0xE4,0,0xEA,0xE9,0xE5,0,0xEB,0xEA,0xE6,0,0xEC,0xEB,0xE7,0,
    0xED,0xEC,0xE8,0,0xEE,0xED,0xE9,0,0xEF,0xEE,0xEA,0,0xF0,0xEF,0xEB,0,
    0xF1,0xF0,0xEC,0,0xF2,0xF1,0xED,0,0xF3,0xF2,0xEE,0,0xF4,0xF3,0xEF,0,
    0xF5,0xF4,0xF0,0,0xF6,0xF5,0xF1,0,0xF7,0xF6,0xF2,0,0xF8,0xF7,0xF3,0,
    0xF9,0xF8,0xF4,0,0xFA,0xF9,0xF5,0,0xFB,0xFA,0xF6,0,0xFC,0xFB,0xF7,0,
    0xFD,0xFC,0xF8,0,0xFE,0xFD,0xF9,0,0xFF,0xFE,0xFA,0,0x00,0xFF,0xFB,0,
    0x01,0x00,0xFC,0,0x02,0x01,0xFD,0,0x03,0x02,0xFE,0,0x04,0x03,0xFF,0,
    0x05,0x04,0x00,0,0x06,0x05,0x01,0,0x07,0x06,0x02,0,0x08,0x07,0x03,0,
    0x09,0x08,0x04,0,0x0A,0x09,0x05,0,0x0B,0x0A,0x06,0,0x0C,0x0B,0x07,0,
    0x0D,0x0C,0x08,0,0x0E,0x0D,0x09,0,0x0F,0x0E,0x0A,0,0x10,0x0F,0x0B,0,
    0x11,0x10,0x0C,0,0x12,0x11,0x0D,0,0x13,0x12,0x0E,0,0x14,0x13,0x0F,0,
    0x15,0x14,0x10,0,0x16,0x15,0x11,0,0x17,0x16,0x12,0,0x18,0x17,0x13,0,
    0x19,0x18,0x14,0,0x1A,0x19,0x15,0,0x1B,0x1A,0x16,0,0x1C,0x1B,0x17,0,
    0x1D,0x1C,0x18,0,0x1E,0x1D,0x19,0,0x1F,0x1E,0x1A,0,0x20,0x1F,0x1B,0,
    0xE2,0xE0,0xDC,0,0xE3,0xE1,0xDD,0,0xE4,0xE2,0xDE,0,0xE5,0xE3,0xDF,0,
    0xE6,0xE4,0xE0,0,0xE7,0xE5,0xE1,0,0xE8,0xE6,0xE2,0,0xE9,0xE7,0xE3,0,
    0xEA,0xE8,0xE4,0,0xEB,0xE9,0xE5,0,0xEC,0xEA,0xE6,0,0xED,0xEB,0xE7,0,
    0xEE,0xEC,0xE8,0,0xEF,0xED,0xE9,0,0xF0,0xEE,0xEA,0,0xF1,0xEF,0xEB,0,
    0xF2,0xF0,0xEC,0,0xF3,0xF1,0xED,0,0xF4,0xF2,0xEE,0,0xF5,0xF3,0xEF,0,
    0xF6,0xF4,0xF0,0,0xF7,0xF5,0xF1,0,0xF8,0xF6,0xF2,0,0xF9,0xF7,0xF3,0,
    0xFA,0xF8,0xF4,0,0xFB,0xF9,0xF5,0,0xFC,0xFA,0xF6,0,0xFD,0xFB,0xF7,0,
    0xFE,0xFC,0xF8,0,0xFF,0xFD,0xF9,0,0x00,0xFE,0xFA,0,0x01,0xFF,0xFB,0,
    0x02,0x00,0xFC,0,0x03,0x01,0xFD,0,0x04,0x02,0xFE,0,0x05,0x03,0xFF,0,
    0x06,0x04,0x00,0,0x07,0x05,0x01,0,0x08,0x06,0x02,0,0x09,0x07,0x03,0,
    0x0A,0x08,0x04,0,0x0B,0x09,0x05,0,0x0C,0x0A,0x06,0,0x0D,0x0B,0x07,0,
    0x0E,0x0C,0x08,0,0x0F,0x0D,0x09,0,0x10,0x0E,0x0A,0,0x11,0x0F,0x0B,0,
    0x12,0x10,0x0C,0,0x13,0x11,0x0D,0,0x14,0x12,0x0E,0,0x15,0x13,0x0F,0,
    0x16,0x14,0x10,0,0x17,0x15,0x11,0,0x18,0x16,0x12,0,0x19,0x17,0x13,0,
    0x1A,0x18,0x14,0,0x1B,0x19,0x15,0,0x1C,0x1A,0x16,0,0x1D,0x1B,0x17,0,
    0x1E,0x1C,0x18,0,0x1F,0x1D,0x19,0,0x20,0x1E,0x1A,0,0x21,0x1F,0x1B,0,
    0xE3,0xE0,0xDC,0,0xE4,0xE1,0xDD,0,0xE5,0xE2,0xDE,0,0xE6,0xE3,0xDF,0,
    0xE7,0xE4,0xE0,0,0xE8,0xE5,0xE1,0,0xE9,0xE6,0xE2,0,0xEA,0xE7,0xE3,0,
    0xEB,0xE8,0xE4,0,0xEC,0xE9,0xE5,0,0xED,0xEA,0xE6,0,0xEE,0xEB,0xE7,0,
    0xEF,0xEC,0xE8,0,0xF0,0xED,0xE9,0,0xF1,0xEE,0xEA,0,0xF2,0xEF,0xEB,0,
    0xF3,0xF0,0xEC,0,0xF4,0xF1,0xED,0,0xF5,0xF2,0xEE,0,0xF6,0xF3,0xEF,0,
    0xF7,0xF4,0xF0,0,0xF8,0xF5,0xF1,0,0xF9,0xF6,0xF2,0,0xFA,0xF7,0xF3,0,
    0xFB,0xF8,0xF4,0,0xFC,0xF9,0xF5,0,0xFD,0xFA,0xF6,0,0xFE,0xFB,0xF7,0,
    0xFF,0xFC,0xF8,0,0x00,0xFD,0xF9,0,0x01,0xFE,0xFA,0,0x02,0xFF,0xFB,0,
    0x03,0x00,0xFC,0,0x04,0x01,0xFD,0,0x05,0x02,0xFE,0,0x06,0x03,0xFF,0,
    0x07,0x04,0x00,0,0x08,0x05,0x01,0,0x09,0x06,0x02,0,0x0A,0x07,0x03,0,
    0x0B,0x08,0x04,0,0x0C,0x09,0x05,0,0x0D,0x0A,0x06,0,0x0E,0x0B,0x07,0,
    0x0F,0x0C,0x08,0,0x10,0x0D,0x09,0,0x11,0x0E,0x0A,0,0x12,0x0F,0x0B,0,
    0x13,0x10,0x0C,0,0x14,0x11,0x0D,0,0x15,0x12,0x0E,0,0x16,0x13,0x0F,0,
    0x17,0x14,0x10,0,0x18,0x15,0x11,0,0x19,0x16,0x12,0,0x1A,0x17,0x13,0,
    0x1B,0x18,0x14,0,0x1C,0x19,0x15,0,0x1D,0x1A,0x16,0,0x1E,0x1B,0x17,0,
    0x1F,0x1C,0x18,0,0x20,0x1D,0x19,0,0x21,0x1E,0x1A,0,0x22,0x1F,0x1B,0,
    0xE4,0xE0,0xDC,0,0xE5,0xE1,0xDD,0,0xE6,0xE2,0xDE,0,0xE7,0xE3,0xDF,0,
    0xE8,0xE4,0xE0,0,0xE9,0xE5,0xE1,0,0xEA,0xE6,0xE2,0,0xEB,0xE7,0xE3,0,
    0xEC,0xE8,0xE4,0,0xED,0xE9,0xE5,0,0xEE,0xEA,0xE6,0,0xEF,0xEB,0xE7,0,
    0xF0,0xEC,0xE8,0,0xF1,0xED,0xE9,0,0xF2,0xEE,0xEA,0,0xF3,0xEF,0xEB,0,
    0xF4,0xF0,0xEC,0,0xF5,0xF1,0xED,0,0xF6,0xF2,0xEE,0,0xF7,0xF3,0xEF,0,
    0xF8,0xF4,0xF0,0,0xF9,0xF5,0xF1,0,0xFA,0xF6,0xF2,0,0xFB,0xF7,0xF3,0,
    0xFC,0xF8,0xF4,0,0xFD,0xF9,0xF5,0,0xFE,0xFA,0xF6,0,0xFF,0xFB,0xF7,0,
    0x00,0xFC,0xF8,0,0x01,0xFD,0xF9,0,0x02,0xFE,0xFA,0,0x03,0xFF,0xFB,0,
    0x04,0x00,0xFC,0,0x05,0x01,0xFD,0,0x06,0x02,0xFE,0,0x07,0x03,0xFF,0,
    0x08,0x04,0x00,0,0x09,0x05,0x01,0,0x0A,0x06,0x02,0,0x0B,0x07,0x03,0,
    0x0C,0x08,0x04,0,0x0D,0x09,0x05,0,0x0E,0x0A,0x06,0,0x0F,0x0B,0x07,0,
    0x10,0x0C,0x08,0,0x11,0x0D,0x09,0,0x12,0x0E,0x0A,0,0x13,0x0F,0x0B,0,
    0x14,0x10,0x0C,0,0x15,0x11,0x0D,0,0x16,0x12,0x0E,0,0x17,0x13,0x0F,0,
    0x18,0x14,0x10,0,0x19,0x15,0x11,0,0x1A,0x16,0x12,0,0x1B,0x17,0x13,0,
    0x1C,0x18,0x14,0,0x1D,0x19,0x15,0,0x1E,0x1A,0x16,0,0x1F,0x1B,0x17,0,
    0x20,0x1C,0x18,0,0x21,0x1D,0x19,0,0x22,0x1E,0x1A,0,0x23,0x1F,0x1B,0,
    0xE5,0xE0,0xDC,0,0xE6,0xE1,0xDD,0,0xE7,0xE2,0xDE,0,0xE8,0xE3,0xDF,0,
    0xE9,0xE4,0xE0,0,0xEA,0xE5,0xE1,0,0xEB,0xE6,0xE2,0,0xEC,0xE7,0xE3,0,
    0xED,0xE8,0xE4,0,0xEE,0xE9,0xE5,0,0xEF,0xEA,0xE6,0,0xF0,0xEB,0xE7,0,
    0xF1,0xEC,0xE8,0,0xF2,0xED,0xE9,0,0xF3,0xEE,0xEA,0,0xF4,0xEF,0xEB,0,
    0xF5,0xF0,0xEC,0,0xF6,0xF1,0xED,0,0xF7,0xF2,0xEE,0,0xF8,0xF3,0xEF,0,
    0xF9,0xF4,0xF0,0,0xFA,0xF5,0xF1,0,0xFB,0xF6,0xF2,0,0xFC,0xF7,0xF3,0,
    0xFD,0xF8,0xF4,0,0xFE,0xF9,0xF5,0,0xFF,0xFA,0xF6,0,0x00,0xFB,0xF7,0,
    0x01,0xFC,0xF8,0,0x02,0xFD,0xF9,0,0x03,0xFE,0xFA,0,0x04,0xFF,0xFB,0,
    0x05,0x00,0xFC,0,0x06,0x01,0xFD,0,0x07,0x02,0xFE,0,0x08,0x03,0xFF,0,
    0x09,0x04,0x00,0,0x0A,0x05,0x01,0,0x0B,0x06,0x02,0,0x0C,0x07,0x03,0,
    0x0D,0x08,0x04,0,0x0E,0x09,0x05,0,0x0F,0x0A,0x06,0,0x10,0x0B,0x07,0,
    0x11,0x0C,0x08,0,0x12,0x0D,0x09,0,0x13,0x0E,0x0A,0,0x14,0x0F,0x0B,0,
    0x15,0x10,0x0C,0,0x16,0x11,0x0D,0,0x17,0x12,0x0E,0,0x18,0x13,0x0F,0,
    0x19,0x14,0x10,0,0x1A,0x15,0x11,0,0x1B,0x16,0x12,0,0x1C,0x17,0x13,0,
    0x1D,0x18,0x14,0,0x1E,0x19,0x15,0,0x1F,0x1A,0x16,0,0x20,0x1B,0x17,0,
    0x21,0x1C,0x18,0,0x22,0x1D,0x19,0,0x23,0x1E,0x1A,0,0x24,0x1F,0x1B,0,
    0xE6,0xE0,0xDC,0,0xE7,0xE1,0xDD,0,0xE8,0xE2,0xDE,0,0xE9,0xE3,0xDF,0,
    0xEA,0xE4,0xE0,0,0xEB,0xE5,0xE1,0,0xEC,0xE6,0xE2,0,0xED,0xE7,0xE3,0,
    0xEE,0xE8,0xE4,0,0xEF,0xE9,0xE5,0,0xF0,0xEA,0xE6,0,0xF1,0xEB,0xE7,0,
    0xF2,0xEC,0xE8,0,0xF3,0xED,0xE9,0,0xF4,0xEE,0xEA,0,0xF5,0xEF,0xEB,0,
    0xF6,0xF0,0xEC,0,0xF7,0xF1,0xED,0,0xF8,0xF2,0xEE,0,0xF9,0xF3,0xEF,0,
    0xFA,0xF4,0xF0,0,0xFB,0xF5,0xF1,0,0xFC,0xF6,0xF2,0,0xFD,0xF7,0xF3,0,
    0xFE,0xF8,0xF4,0,0xFF,0xF9,0xF5,0,0x00,0xFA,0xF6,0,0x01,0xFB,0xF7,0,
    0x02,0xFC,0xF8,0,0x03,0xFD,0xF9,0,0x04,0xFE,0xFA,0,0x05,0xFF,0xFB,0,
    0x06,0x00,0xFC,0,0x07,0x01,0xFD,0,0x08,0x02,0xFE,0,0x09,0x03,0xFF,0,
    0x0A,0x04,0x00,0,0x0B,0x05,0x01,0,0x0C,0x06,0x02,0,0x0D,0x07,0x03,0,
    0x0E,0x08,0x04,0,0x0F,0x09,0x05,0,0x10,0x0A,0x06,0,0x11,0x0B,0x07,0,
    0x12,0x0C,0x08,0,0x13,0x0D,0x09,0,0x14,0x0E,0x0A,0,0x15,0x0F,0x0B,0,
    0x16,0x10,0x0C,0,0x17,0x11,0x0D,0,0x18,0x12,0x0E,0,0x19,0x13,0x0F,0,
    0x1A,0x14,0x10,0,0x1B,0x15,0x11,0,0x1C,0x16,0x12,0,0x1D,0x17,0x13,0,
    0x1E,0x18,0x14,0,0x1F,0x19,0x15,0,0x20,0x1A,0x16,0,0x21,0x1B,0x17,0,
    0x22,0x1C,0x18,0,0x23,0x1D,0x19,0,0x24,0x1E,0x1A,0,0x25,0x1F,0x1B,0,
    0xE7,0xE0,0xDC,0,0xE8,0xE1,0xDD,0,0xE9,0xE2,0xDE,0,0xEA,0xE3,0xDF,0,
    0xEB,0xE4,0xE0,0,0xEC,0xE5,0xE1,0,0xED,0xE6,0xE2,0,0xEE,0xE7,0xE3,0,
    0xEF,0xE8,0xE4,0,0xF0,0xE9,0xE5,0,0xF1,0xEA,0xE6,0,0xF2,0xEB,0xE7,0,
    0xF3,0xEC,0xE8,0,0xF4,0xED,0xE9,0,0xF5,0xEE,0xEA,0,0xF6,0xEF,0xEB,0,
    0xF7,0xF0,0xEC,0,0xF8,0xF1,0xED,0,0xF9,0xF2,0xEE,0,0xFA,0xF3,0xEF,0,
    0xFB,0xF4,0xF0,0,0xFC,0xF5,0xF1,0,0xFD,0xF6,0xF2,0,0xFE,0xF7,0xF3,0,
    0xFF,0xF8,0xF4,0,0x00,0xF9,0xF5,0,0x01,0xFA,0xF6,0,0x02,0xFB,0xF7,0,
    0x03,0xFC,0xF8,0,0x04,0xFD,0xF9,0,0x05,0xFE,0xFA,0,0x06,0xFF,0xFB,0,
    0x07,0x00,0xFC,0,0x08,0x01,0xFD,0,0x09,0x02,0xFE,0,0x0A,0x03,0xFF,0,
    0x0B,0x04,0x00,0,0x0C,0x05,0x01,0,0x0D,0x06,0x02,0,0x0E,0x07,0x03,0,
    0x0F,0x08,0x04,0,0x10,0x09,0x05,0,0x11,0x0A,0x06,0,0x12,0x0B,0x07,0,
    0x13,0x0C,0x08,0,0x14,0x0D,0x09,0,0x15,0x0E,0x0A,0,0x16,0x0F,0x0B,0,
    0x17,0x10,0x0C,0,0x18,0x11,0x0D,0,0x19,0x12,0x0E,0,0x1A,0x13,0x0F,0,
    0x1B,0x14,0x10,0,0x1C,0x15,0x11,0,0x1D,0x16,0x12,0,0x1E,0x17,0x13,0,
    0x1F,0x18,0x14,0,0x20,0x19,0x15,0,0x21,0x1A,0x16,0,0x22,0x1B,0x17,0,
    0x23,0x1C,0x18,0,0x24,0x1D,0x19,0,0x25,0x1E,0x1A,0,0x26,0x1F,0x1B,0,
    0xD8,0xE0,0xDD,0,0xD9,0xE1,0xDE,0,0xDA,0xE2,0xDF,0,0xDB,0xE3,0xE0,0,
    0xDC,0xE4,0xE1,0,0xDD,0xE5,0xE2,0,0xDE,0xE6,0xE3,0,0xDF,0xE7,0xE4,0,
    0xE0,0xE8,0xE5,0,0xE1,0xE9,0xE6,0,0xE2,0xEA,0xE7,0,0xE3,0xEB,0xE8,0,
    0xE4,0xEC,0xE9,0,0xE5,0xED,0xEA,0,0xE6,0xEE,0xEB,0,0xE7,0xEF,0xEC,0,
    0xE8,0xF0,0xED,0,0xE9,0xF1,0xEE,0,0xEA,0xF2,0xEF,0,0xEB,0xF3,0xF0,0,
    0xEC,0xF4,0xF1,0,0xED,0xF5,0xF2,0,0xEE,0xF6,0xF3,0,0xEF,0xF7,0xF4,0,
    0xF0,0xF8,0xF5,0,0xF1,0xF9,0xF6,0,0xF2,0xFA,0xF7,0,0xF3,0xFB,0xF8,0,
    0xF4,0xFC,0xF9,0,0xF5,0xFD,0xFA,0,0xF6,0xFE,0xFB,0,0xF7,0xFF,0xFC,0,
    0xF8,0x00,0xFD,0,0xF9,0x01,0xFE,0,0xFA,0x02,0xFF,0,0xFB,0x03,0x00,0,
    0xFC,0x04,0x01,0,0xFD,0x05,0x02,0,0xFE,0x06,0x03,0,0xFF,0x07,0x04,0,
    0x00,0x08,0x05,0,0x01,0x09,0x06,0,0x02,0x0A,0x07,0,0x03,0x0B,0x08,0,
    0x04,0x0C,0x09,0,0x05,0x0D,0x0A,0,0x06,0x0E,0x0B,0,0x07,0x0F,0x0C,0,
    0x08,0x10,0x0D,0,0x09,0x11,0x0E,0,0x0A,0x12,0x0F,0,0x0B,0x13,0x10,0,
    0x0C,0x14,0x11,0,0x0D,0x15,0x12,0,0x0E,0x16,0x13,0,0x0F,0x17,0x14,0,
    0x10,0x18,0x15,0,0x11,0x19,0x16,0,0x12,0x1A,0x17,0,0x13,0x1B,0x18,0,
    0x14,0x1C,0x19,0,0x15,0x1D,0x1A,0,0x16,0x1E,0x1B,0,0x17,0x1F,0x1C,0,
    0xD9,0xE0,0xDD,0,0xDA,0xE1,0xDE,0,0xDB,0xE2,0xDF,0,0xDC,0xE3,0xE0,0,
    0xDD,0xE4,0xE1,0,0xDE,0xE5,0xE2,0,0xDF,0xE6,0xE3,0,0xE0,0xE7,0xE4,0,
    0xE1,0xE8,0xE5,0,0xE2,0xE9,0xE6,0,0xE3,0xEA,0xE7,0,0xE4,0xEB,0xE8,0,
    0xE5,0xEC,0xE9,0,0xE6,0xED,0xEA,0,0xE7,0xEE,0xEB,0,0xE8,0xEF,0xEC,0,
    0xE9,0xF0,0xED,0,0xEA,0xF1,0xEE,0,0xEB,0xF2,0xEF,0,0xEC,0xF3,0xF0,0,
    0xED,0xF4,0xF1,0,0xEE,0xF5,0xF2,0,0xEF,0xF6,0xF3,0,0xF0,0xF7,0xF4,0,
    0xF1,0xF8,0xF5,0,0xF2,0xF9,0xF6,0,0xF3,0xFA,0xF7,0,0xF4,0xFB,0xF8,0,
    0xF5,0xFC,0xF9,0,0xF6,0xFD,0xFA,0,0xF7,0xFE,0xFB,0,0xF8,0xFF,0xFC,0,
    0xF9,0x00,0xFD,0,0xFA,0x01,0xFE,0,0xFB,0x02,0xFF,0,0xFC,0x03,0x00,0,
    0xFD,0x04,0x01,0,0xFE,0x05,0x02,0,0xFF,0x06,0x03,0,0x00,0x07,0x04,0,
    0x01,0x08,0x05,0,0x02,0x09,0x06,0,0x03,0x0A,0x07,0,0x04,0x0B,0x08,0,
    0x05,0x0C,0x09,0,0x06,0x0D,0x0A,0,0x07,0x0E,0x0B,0,0x08,0x0F,0x0C,0,
    0x09,0x10,0x0D,0,0x0A,0x11,0x0E,0,0x0B,0x12,0x0F,0,0x0C,0x13,0x10,0,
    0x0D,0x14,0x11,0,0x0E,0x15,0x12,0,0x0F,0x16,0x13,0,0x10,0x17,0x14,0,
    0x11,0x18,0x15,0,0x12,0x19,0x16,0,0x13,0x1A,0x17,0,0x14,0x1B,0x18,0,
    0x15,0x1C,0x19,0,0x16,0x1D,0x1A,0,0x17,0x1E,0x1B,0,0x18,0x1F,0x1C,0,
    0xDA,0xE0,0xDD,0,0xDB,0xE1,0xDE,0,0xDC,0xE2,0xDF,0,0xDD,0xE3,0xE0,0,
    0xDE,0xE4,0xE1,0,0xDF,0xE5,0xE2,0,0xE0,0xE6,0xE3,0,0xE1,0xE7,0xE4,0,
    0xE2,0xE8,0xE5,0,0xE3,0xE9,0xE6,0,0xE4,0xEA,0xE7,0,0xE5,0xEB,0xE8,0,
    0xE6,0xEC,0xE9,0,0xE7,0xED,0xEA,0,0xE8,0xEE,0xEB,0,0xE9,0xEF,0xEC,0,
    0xEA,0xF0,0xED,0,0xEB,0xF1,0xEE,0,0xEC,0xF2,0xEF,0,0xED,0xF3,0xF0,0,
    0xEE,0xF4,0xF1,0,0xEF,0xF5,0xF2,0,0xF0,0xF6,0xF3,0,0xF1,0xF7,0xF4,0,
    0xF2,0xF8,0xF5,0,0xF3,0xF9,0xF6,0,0xF4,0xFA,0xF7,0,0xF5,0xFB,0xF8,0,
    0xF6,0xFC,0xF9,0,0xF7,0xFD,0xFA,0,0xF8,0xFE,0xFB,0,0xF9,0xFF,0xFC,0,
    0xFA,0x00,0xFD,0,0xFB,0x01,0xFE,0,0xFC,0x02,0xFF,0,0xFD,0x03,0x00,0,
    0xFE,0x04,0x01,0,0xFF,0x05,0x02,0,0x00,0x06,0x03,0,0x01,0x07,0x04,0,
    0x02,0x08,0x05,0,0x03,0x09,0x06,0,0x04,0x0A,0x07,0,0x05,0x0B,0x08,0,
    0x06,0x0C,0x09,0,0x07,0x0D,0x0A,0,0x08,0x0E,0x0B,0,0x09,0x0F,0x0C,0,
    0x0A,0x10,0x0D,0,0x0B,0x11,0x0E,0,0x0C,0x12,0x0F,0,0x0D,0x13,0x10,0,
    0x0E,0x14,0x11,0,0x0F,0x15,0x12,0,0x10,0x16,0x13,0,0x11,0x17,0x14,0,
    0x12,0x18,0x15,0,0x13,0x19,0x16,0,0x14,0x1A,0x17,0,0x15,0x1B,0x18,0,
    0x16,0x1C,0x19,0,0x17,0x1D,0x1A,0,0x18,0x1E,0x1B,0,0x19,0x1F,0x1C,0,
    0xDB,0xE0,0xDD,0,0xDC,0xE1,0xDE,0,0xDD,0xE2,0xDF,0,0xDE,0xE3,0xE0,0,
    0xDF,0xE4,0xE1,0,0xE0,0xE5,0xE2,0,0xE1,0xE6,0xE3,0,0xE2,0xE7,0xE4,0,
    0xE3,0xE8,0xE5,0,0xE4,0xE9,0xE6,0,0xE5,0xEA,0xE7,0,0xE6,0xEB,0xE8,0,
    0xE7,0xEC,0xE9,0,0xE8,0xED,0xEA,0,0xE9,0xEE,0xEB,0,0xEA,0xEF,0xEC,0,
    0xEB,0xF0,0xED,0,0xEC,0xF1,0xEE,0,0xED,0xF2,0xEF,0,0xEE,0xF3,0xF0,0,
    0xEF,0xF4,0xF1,0,0xF0,0xF5,0xF2,0,0xF1,0xF6,0xF3,0,0xF2,0xF7,0xF4,0,
    0xF3,0xF8,0xF5,0,0xF4,0xF9,0xF6,0,0xF5,0xFA,0xF7,0,0xF6,0xFB,0xF8,0,
    0xF7,0xFC,0xF9,0,0xF8,0xFD,0xFA,0,0xF9,0xFE,0xFB,0,0xFA,0xFF,0xFC,0,
    0xFB,0x00,0xFD,0,0xFC,0x01,0xFE,0,0xFD,0x02,0xFF,0,0xFE,0x03,0x00,0,
    0xFF,0x04,0x01,0,0x00,0x05,0x02,0,0x01,0x06,0x03,0,0x02,0x07,0x04,0,
    0x03,0x08,0x05,0,0x04,0x09,0x06,0,0x05,0x0A,0x07,0,0x06,0x0B,0x08,0,
    0x07,0x0C,0x09,0,0x08,0x0D,0x0A,0,0x09,0x0E,0x0B,0,0x0A,0x0F,0x0C,0,
    0x0B,0x10,0x0D,0,0x0C,0x11,0x0E,0,0x0D,0x12,0x0F,0,0x0E,0x13,0x10,0,
    0x0F,0x14,0x11,0,0x10,0x15,0x12,0,0x11,0x16,0x13,0,0x12,0x17,0x14,0,
    0x13,0x18,0x15,0,0x14,0x19,0x16,0,0x15,0x1A,0x17,0,0x16,0x1B,0x18,0,
    0x17,0x1C,0x19,0,0x18,0x1D,0x1A,0,0x19,0x1E,0x1B,0,0x1A,0x1F,0x1C,0,
    0xDC,0xE0,0xDD,0,0xDD,0xE1,0xDE,0,0xDE,0xE2,0xDF,0,0xDF,0xE3,0xE0,0,
    0xE0,0xE4,0xE1,0,0xE1,0xE5,0xE2,0,0xE2,0xE6,0xE3,0,0xE3,0xE7,0xE4,0,
    0xE4,0xE8,0xE5,0,0xE5,0xE9,0xE6,0,0xE6,0xEA,0xE7,0,0xE7,0xEB,0xE8,0,
    0xE8,0xEC,0xE9,0,0xE9,0xED,0xEA,0,0xEA,0xEE,0xEB,0,0xEB,0xEF,0xEC,0,
    0xEC,0xF0,0xED,0,0xED,0xF1,0xEE,0,0xEE,0xF2,0xEF,0,0xEF,0xF3,0xF0,0,
    0xF0,0xF4,0xF1,0,0xF1,0xF5,0xF2,0,0xF2,0xF6,0xF3,0,0xF3,0xF7,0xF4,0,
    0xF4,0xF8,0xF5,0,0xF5,0xF9,0xF6,0,0xF6,0xFA,0xF7,0,0xF7,0xFB,0xF8,0,
    0xF8,0xFC,0xF9,0,0xF9,0xFD,0xFA,0,0xFA,0xFE,0xFB,0,0xFB,0xFF,0xFC,0,
    0xFC,0x00,0xFD,0,0xFD,0x01,0xFE,0,0xFE,0x02,0xFF,0,0xFF,0x03,0x00,0,
    0x00,0x04,0x01,0,0x01,0x05,0x02,0,0x02,0x06,0x03,0,0x03,0x07,0x04,0,
    0x04,0x08,0x05,0,0x05,0x09,0x06,0,0x06,0x0A,0x07,0,0x07,0x0B,0x08,0,
    0x08,0x0C,0x09,0,0x09,0x0D,0x0A,0,0x0A,0x0E,0x0B,0,0x0B,0x0F,0x0C,0,
    0x0C,0x10,0x0D,0,0x0D,0x11,0x0E,0,0x0E,0x12,0x0F,0,0x0F,0x13,0x10,0,
    0x10,0x14,0x11,0,0x11,0x15,0x12,0,0x12,0x16,0x13,0,0x13,0x17,0x14,0,
    0x14,0x18,0x15,0,0x15,0x19,0x16,0,0x16,0x1A,0x17,0,0x17,0x1B,0x18,0,
    0x18,0x1C,0x19,0,0x19,0x1D,0x1A,0,0x1A,0x1E,0x1B,0,0x1B,0x1F,0x1C,0,
    0xDD,0xE0,0xDD,0,0xDE,0xE1,0xDE,0,0xDF,0xE2,0xDF,0,0xE0,0xE3,0xE0,0,
    0xE1,0xE4,0xE1,0,0xE2,0xE5,0xE2,0,0xE3,0xE6,0xE3,0,0xE4,0xE7,0xE4,0,
    0xE5,0xE8,0xE5,0,0xE6,0xE9,0xE6,0,0xE7,0xEA,0xE7,0,0xE8,0xEB,0xE8,0,
    0xE9,0xEC,0xE9,0,0xEA,0xED,0xEA,0,0xEB,0xEE,0xEB,0,0xEC,0xEF,0xEC,0,
    0xED,0xF0,0xED,0,0xEE,0xF1,0xEE,0,0xEF,0xF2,0xEF,0,0xF0,0xF3,0xF0,0,
    0xF1,0xF4,0xF1,0,0xF2,0xF5,0xF2,0,0xF3,0xF6,0xF3,0,0xF4,0xF7,0xF4,0,
    0xF5,0xF8,0xF5,0,0xF6,0xF9,0xF6,0,0xF7,0xFA,0xF7,0,0xF8,0xFB,0xF8,0,
    0xF9,0xFC,0xF9,0,0xFA,0xFD,0xFA,0,0xFB,0xFE,0xFB,0,0xFC,0xFF,0xFC,0,
    0xFD,0x00,0xFD,0,0xFE,0x01,0xFE,0,0xFF,0x02,0xFF,0,0x00,0x03,0x00,0,
    0x01,0x04,0x01,0,0x02,0x05,0x02,0,0x03,0x06,0x03,0,0x04,0x07,0x04,0,
    0x05,0x08,0x05,0,0x06,0x09,0x06,0,0x07,0x0A,0x07,0,0x08,0x0B,0x08,0,
    0x09,0x0C,0x09,0,0x0A,0x0D,0x0A,0,0x0B,0x0E,0x0B,0,0x0C,0x0F,0x0C,0,
    0x0D,0x10,0x0D,0,0x0E,0x11,0x0E,0,0x0F,0x12,0x0F,0,0x10,0x13,0x10,0,
    0x11,0x14,0x11,0,0x12,0x15,0x12,0,0x13,0x16,0x13,0,0x14,0x17,0x14,0,
    0x15,0x18,0x15,0,0x16,0x19,0x16,0,0x17,0x1A,0x17,0,0x18,0x1B,0x18,0,
    0x19,0x1C,0x19,0,0x1A,0x1D,0x1A,0,0x1B,0x1E,0x1B,0,0x1C,0x1F,0x1C,0,
    0xDE,0xE0,0xDD,0,0xDF,0xE1,0xDE,0,0xE0,0xE2,0xDF,0,0xE1,0xE3,0xE0,0,
    0xE2,0xE4,0xE1,0,0xE3,0xE5,0xE2,0,0xE4,0xE6,0xE3,0,0xE5,0xE7,0xE4,0,
    0xE6,0xE8,0xE5,0,0xE7,0xE9,0xE6,0,0xE8,0xEA,0xE7,0,0xE9,0xEB,0xE8,0,
    0xEA,0xEC,0xE9,0,0xEB,0xED,0xEA,0,0xEC,0xEE,0xEB,0,0xED,0xEF,0xEC,0,
    0xEE,0xF0,0xED,0,0xEF,0xF1,0xEE,0,0xF0,0xF2,0xEF,0,0xF1,0xF3,0xF0,0,
    0xF2,0xF4,0xF1,0,0xF3,0xF5,0xF2,0,0xF4,0xF6,0xF3,0,0xF5,0xF7,0xF4,0,
    0xF6,0xF8,0xF5,0,0xF7,0xF9,0xF6,0,0xF8,0xFA,0xF7,0,0xF9,0xFB,0xF8,0,
    0xFA,0xFC,0xF9,0,0xFB,0xFD,0xFA,0,0xFC,0xFE,0xFB,0,0xFD,0xFF,0xFC,0,
    0xFE,0x00,0xFD,0,0xFF,0x01,0xFE,0,0x00,0x02,0xFF,0,0x01,0x03,0x00,0,
    0x02,0x04,0x01,0,0x03,0x05,0x02,0,0x04,0x06,0x03,0,0x05,0x07,0x04,0,
    0x06,0x08,0x05,0,0x07,0x09,0x06,0,0x08,0x0A,0x07,0,0x09,0x0B,0x08,0,
    0x0A,0x0C,0x09,0,0x0B,0x0D,0x0A,0,0x0C,0x0E,0x0B,0,0x0D,0x0F,0x0C,0,
    0x0E,0x10,0x0D,0,0x0F,0x11,0x0E,0,0x10,0x12,0x0F,0,0x11,0x13,0x10,0,
    0x12,0x14,0x11,0,0x13,0x15,0x12,0,0x14,0x16,0x13,0,0x15,0x17,0x14,0,
    0x16,0x18,0x15,0,0x17,0x19,0x16,0,0x18,0x1A,0x17,0,0x19,0x1B,0x18,0,
    0x1A,0x1C,0x19,0,0x1B,0x1D,0x1A,0,0x1C,0x1E,0x1B,0,0x1D,0x1F,0x1C,0,
    0xDF,0xE0,0xDD,0,0xE0,0xE1,0xDE,0,0xE1,0xE2,0xDF,0,0xE2,0xE3,0xE0,0,
    0xE3,0xE4,0xE1,0,0xE4,0xE5,0xE2,0,0xE5,0xE6,0xE3,0,0xE6,0xE7,0xE4,0,
    0xE7,0xE8,0xE5,0,0xE8,0xE9,0xE6,0,0xE9,0xEA,0xE7,0,0xEA,0xEB,0xE8,0,
    0xEB,0xEC,0xE9,0,0xEC,0xED,0xEA,0,0xED,0xEE,0xEB,0,0xEE,0xEF,0xEC,0,
    0xEF,0xF0,0xED,0,0xF0,0xF1,0xEE,0,0xF1,0xF2,0xEF,0,0xF2,0xF3,0xF0,0,
    0xF3,0xF4,0xF1,0,0xF4,0xF5,0xF2,0,0xF5,0xF6,0xF3,0,0xF6,0xF7,0xF4,0,
    0xF7,0xF8,0xF5,0,0xF8,0xF9,0xF6,0,0xF9,0xFA,0xF7,0,0xFA,0xFB,0xF8,0,
    0xFB,0xFC,0xF9,0,0xFC,0xFD,0xFA,0,0xFD,0xFE,0xFB,0,0xFE,0xFF,0xFC,0,
    0xFF,0x00,0xFD,0,0x00,0x01,0xFE,0,0x01,0x02,0xFF,0,0x02,0x03,0x00,0,
    0x03,0x04,0x01,0,0x04,0x05,0x02,0,0x05,0x06,0x03,0,0x06,0x07,0x04,0,
    0x07,0x08,0x05,0,0x08,0x09,0x06,0,0x09,0x0A,0x07,0,0x0A,0x0B,0x08,0,
    0x0B,0x0C,0x09,0,0x0C,0x0D,0x0A,0,0x0D,0x0E,0x0B,0,0x0E,0x0F,0x0C,0,
    0x0F,0x10,0x0D,0,0x10,0x11,0x0E,0,0x11,0x12,0x0F,0,0x12,0x13,0x10,0,
    0x13,0x14,0x11,0,0x14,0x15,0x12,0,0x15,0x16,0x13,0,0x16,0x17,0x14,0,
    0x17,0x18,0x15,0,0x18,0x19,0x16,0,0x19,0x1A,0x17,0,0x1A,0x1B,0x18,0,
    0x1B,0x1C,0x19,0,0x1C,0x1D,0x1A,0,0x1D,0x1E,0x1B,0,0x1E,0x1F,0x1C,0,
    0xE0,0xE0,0xDD,0,0xE1,0xE1,0xDE,0,0xE2,0xE2,0xDF,0,0xE3,0xE3,0xE0,0,
    0xE4,0xE4,0xE1,0,0xE5,0xE5,0xE2,0,0xE6,0xE6,0xE3,0,0xE7,0xE7,0xE4,0,
    0xE8,0xE8,0xE5,0,0xE9,0xE9,0xE6,0,0xEA,0xEA,0xE7,0,0xEB,0xEB,0xE8,0,
    0xEC,0xEC,0xE9,0,0xED,0xED,0xEA,0,0xEE,0xEE,0xEB,0,0xEF,0xEF,0xEC,0,
    0xF0,0xF0,0xED,0,0xF1,0xF1,0xEE,0,0xF2,0xF2,0xEF,0,0xF3,0xF3,0xF0,0,
    0xF4,0xF4,0xF1,0,0xF5,0xF5,0xF2,0,0xF6,0xF6,0xF3,0,0xF7,0xF7,0xF4,0,
    0xF8,0xF8,0xF5,0,0xF9,0xF9,0xF6,0,0xFA,0xFA,0xF7,0,0xFB,0xFB,0xF8,0,
    0xFC,0xFC,0xF9,0,0xFD,0xFD,0xFA,0,0xFE,0xFE,0xFB,0,0xFF,0xFF,0xFC,0,
    0x00,0x00,0xFD,0,0x01,0x01,0xFE,0,0x02,0x02,0xFF,0,0x03,0x03,0x00,0,
    0x04,0x04,0x01,0,0x05,0x05,0x02,0,0x06,0x06,0x03,0,0x07,0x07,0x04,0,
    0x08,0x08,0x05,0,0x09,0x09,0x06,0,0x0A,0x0A,0x07,0,0x0B,0x0B,0x08,0,
    0x0C,0x0C,0x09,0,0x0D,0x0D,0x0A,0,0x0E,0x0E,0x0B,0,0x0F,0x0F,0x0C,0,
    0x10,0x10,0x0D,0,0x11,0x11,0x0E,0,0x12,0x12,0x0F,0,0x13,0x13,0x10,0,
    0x14,0x14,0x11,0,0x15,0x15,0x12,0,0x16,0x16,0x13,0,0x17,0x17,0x14,0,
    0x18,0x18,0x15,0,0x19,0x19,0x16,0,0x1A,0x1A,0x17,0,0x1B,0x1B,0x18,0,
    0x1C,0x1C,0x19,0,0x1D,0x1D,0x1A,0,0x1E,0x1E,0x1B,0,0x1F,0x1F,0x1C,0,
    0xE1,0xE0,0xDD,0,0xE2,0xE1,0xDE,0,0xE3,0xE2,0xDF,0,0xE4,0xE3,0xE0,0,
    0xE5,0xE4,0xE1,0,0xE6,0xE5,0xE2,0,0xE7,0xE6,0xE3,0,0xE8,0xE7,0xE4,0,
    0xE9,0xE8,0xE5,0,0xEA,0xE9,0xE6,0,0xEB,0xEA,0xE7,0,0xEC,0xEB,0xE8,0,
    0xED,0xEC,0xE9,0,0xEE,0xED,0xEA,0,0xEF,0xEE,0xEB,0,0xF0,0xEF,0xEC,0,
    0xF1,0xF0,0xED,0,0xF2,0xF1,0xEE,0,0xF3,0xF2,0xEF,0,0xF4,0xF3,0xF0,0,
    0xF5,0xF4,0xF1,0,0xF6,0xF5,0xF2,0,0xF7,0xF6,0xF3,0,0xF8,0xF7,0xF4,0,
    0xF9,0xF8,0xF5,0,0xFA,0xF9,0xF6,0,0xFB,0xFA,0xF7,0,0xFC,0xFB,0xF8,0,
    0xFD,0xFC,0xF9,0,0xFE,0xFD,0xFA,0,0xFF,0xFE,0xFB,0,0x00,0xFF,0xFC,0,
    0x01,0x00,0xFD,0,0x02,0x01,0xFE,0,0x03,0x02,0xFF,0,0x04,0x03,0x00,0,
    0x05,0x04,0x01,0,0x06,0x05,0x02,0,0x07,0x06,0x03,0,0x08,0x07,0x04,0,
    0x09,0x08,0x05,0,0x0A,0x09,0x06,0,0x0B,0x0A,0x07,0,0x0C,0x0B,0x08,0,
    0x0D,0x0C,0x09,0,0x0E,0x0D,0x0A,0,0x0F,0x0E,0x0B,0,0x10,0x0F,0x0C,0,
    0x11,0x10,0x0D,0,0x12,0x11,0x0E,0,0x13,0x12,0x0F,0,0x14,0x13,0x10,0,
    0x15,0x14,0x11,0,0x16,0x15,0x12,0,0x17,0x16,0x13,0,0x18,0x17,0x14,0,
    0x19,0x18,0x15,0,0x1A,0x19,0x16,0,0x1B,0x1A,0x17,0,0x1C,0x1B,0x18,0,
    0x1D,0x1C,0x19,0,0x1E,0x1D,0x1A,0,0x1F,0x1E,0x1B,0,0x20,0x1F,0x1C,0,
    0xE2,0xE0,0xDD,0,0xE3,0xE1,0xDE,0,0xE4,0xE2,0xDF,0,0xE5,0xE3,0xE0,0,
    0xE6,0xE4,0xE1,0,0xE7,0xE5,0xE2,0,0xE8,0xE6,0xE3,0,0xE9,0xE7,0xE4,0,
    0xEA,0xE8,0xE5,0,0xEB,0xE9,0xE6,0,0xEC,0xEA,0xE7,0,0xED,0xEB,0xE8,0,
    0xEE,0xEC,0xE9,0,0xEF,0xED,0xEA,0,0xF0,0xEE,0xEB,0,0xF1,0xEF,0xEC,0,
    0xF2,0xF0,0xED,0,0xF3,0xF1,0xEE,0,0xF4,0xF2,0xEF,0,0xF5,0xF3,0xF0,0,
    0xF6,0xF4,0xF1,0,0xF7,0xF5,0xF2,0,0xF8,0xF6,0xF3,0,0xF9,0xF7,0xF4,0,
    0xFA,0xF8,0xF5,0,0xFB,0xF9,0xF6,0,0xFC,0xFA,0xF7,0,0xFD,0xFB,0xF8,0,
    0xFE,0xFC,0xF9,0,0xFF,0xFD,0xFA,0,0x00,0xFE,0xFB,0,0x01,0xFF,0xFC,0,
    0x02,0x00,0xFD,0,0x03,0x01,0xFE,0,0x04,0x02,0xFF,0,0x05,0x03,0x00,0,
    0x06,0x04,0x01,0,0x07,0x05,0x02,0,0x08,0x06,0x03,0,0x09,0x07,0x04,0,
    0x0A,0x08,0x05,0,0x0B,0x09,0x06,0,0x0C,0x0A,0x07,0,0x0D,0x0B,0x08,0,
    0x0E,0x0C,0x09,0,0x0F,0x0D,0x0A,0,0x10,0x0E,0x0B,0,0x11,0x0F,0x0C,0,
    0x12,0x10,0x0D,0,0x13,0x11,0x0E,0,0x14,0x12,0x0F,0,0x15,0x13,0x10,0,
    0x16,0x14,0x11,0,0x17,0x15,0x12,0,0x18,0x16,0x13,0,0x19,0x17,0x14,0,
    0x1A,0x18,0x15,0,0x1B,0x19,0x16,0,0x1C,0x1A,0x17,0,0x1D,0x1B,0x18,0,
    0x1E,0x1C,0x19,0,0x1F,0x1D,0x1A,0,0x20,0x1E,0x1B,0,0x21,0x1F,0x1C,0,
    0xE3,0xE0,0xDD,0,0xE4,0xE1,0xDE,0,0xE5,0xE2,0xDF,0,0xE6,0xE3,0xE0,0,
    0xE7,0xE4,0xE1,0,0xE8,0xE5,0xE2,0,0xE9,0xE6,0xE3,0,0xEA,0xE7,0xE4,0,
    0xEB,0xE8,0xE5,0,0xEC,0xE9,0xE6,0,0xED,0xEA,0xE7,0,0xEE,0xEB,0xE8,0,
    0xEF,0xEC,0xE9,0,0xF0,0xED,0xEA,0,0xF1,0xEE,0xEB,0,0xF2,0xEF,0xEC,0,
    0xF3,0xF0,0xED,0,0xF4,0xF1,0xEE,0,0xF5,0xF2,0xEF,0,0xF6,0xF3,0xF0,0,
    0xF7,0xF4,0xF1,0,0xF8,0xF5,0xF2,0,0xF9,0xF6,0xF3,0,0xFA,0xF7,0xF4,0,
    0xFB,0xF8,0xF5,0,0xFC,0xF9,0xF6,0,0xFD,0xFA,0xF7,0,0xFE,0xFB,0xF8,0,
    0xFF,0xFC,0xF9,0,0x00,0xFD,0xFA,0,0x01,0xFE,0xFB,0,0x02,0xFF,0xFC,0,
    0x03,0x00,0xFD,0,0x04,0x01,0xFE,0,0x05,0x02,0xFF,0,0x06,0x03,0x00,0,
    0x07,0x04,0x01,0,0x08,0x05,0x02,0,0x09,0x06,0x03,0,0x0A,0x07,0x04,0,
    0x0B,0x08,0x05,0,0x0C,0x09,0x06,0,0x0D,0x0A,0x07,0,0x0E,0x0B,0x08,0,
    0x0F,0x0C,0x09,0,0x10,0x0D,0x0A,0,0x11,0x0E,0x0B,0,0x12,0x0F,0x0C,0,
    0x13,0x10,0x0D,0,0x14,0x11,0x0E,0,0x15,0x12,0x0F,0,0x16,0x13,0x10,0,
    0x17,0x14,0x11,0,0x18,0x15,0x12,0,0x19,0x16,0x13,0,0x1A,0x17,0x14,0,
    0x1B,0x18,0x15,0,0x1C,0x19,0x16,0,0x1D,0x1A,0x17,0,0x1E,0x1B,0x18,0,
    0x1F,0x1C,0x19,0,0x20,0x1D,0x1A,0,0x21,0x1E,0x1B,0,0x22,0x1F,0x1C,0,
    0xE4,0xE0,0xDD,0,0xE5,0xE1,0xDE,0,0xE6,0xE2,0xDF,0,0xE7,0xE3,0xE0,0,
    0xE8,0xE4,0xE1,0,0xE9,0xE5,0xE2,0,0xEA,0xE6,0xE3,0,0xEB,0xE7,0xE4,0,
    0xEC,0xE8,0xE5,0,0xED,0xE9,0xE6,0,0xEE,0xEA,0xE7,0,0xEF,0xEB,0xE8,0,
    0xF0,0xEC,0xE9,0,0xF1,0xED,0xEA,0,0xF2,0xEE,0xEB,0,0xF3,0xEF,0xEC,0,
    0xF4,0xF0,0xED,0,0xF5,0xF1,0xEE,0,0xF6,0xF2,0xEF,0,0xF7,0xF3,0xF0,0,
    0xF8,0xF4,0xF1,0,0xF9,0xF5,0xF2,0,0xFA,0xF6,0xF3,0,0xFB,0xF7,0xF4,0,
    0xFC,0xF8,0xF5,0,0xFD,0xF9,0xF6,0,0xFE,0xFA,0xF7,0,0xFF,0xFB,0xF8,0,
    0x00,0xFC,0xF9,0,0x01,0xFD,0xFA,0,0x02,0xFE,0xFB,0,0x03,0xFF,0xFC,0,
    0x04,0x00,0xFD,0,0x05,0x01,0xFE,0,0x06,0x02,0xFF,0,0x07,0x03,0x00,0,
    0x08,0x04,0x01,0,0x09,0x05,0x02,0,0x0A,0x06,0x03,0,0x0B,0x07,0x04,0,
    0x0C,0x08,0x05,0,0x0D,0x09,0x06,0,0x0E,0x0A,0x07,0,0x0F,0x0B,0x08,0,
    0x10,0x0C,0x09,0,0x11,0x0D,0x0A,0,0x12,0x0E,0x0B,0,0x13,0x0F,0x0C,0,
    0x14,0x10,0x0D,0,0x15,0x11,0x0E,0,0x16,0x12,0x0F,0,0x17,0x13,0x10,0,
    0x18,0x14,0x11,0,0x19,0x15,0x12,0,0x1A,0x16,0x13,0,0x1B,0x17,0x14,0,
    0x1C,0x18,0x15,0,0x1D,0x19,0x16,0,0x1E,0x1A,0x17,0,0x1F,0x1B,0x18,0,
    0x20,0x1C,0x19,0,0x21,0x1D,0x1A,0,0x22,0x1E,0x1B,0,0x23,0x1F,0x1C,0,
    0xE5,0xE0,0xDD,0,0xE6,0xE1,0xDE,0,0xE7,0xE2,0xDF,0,0xE8,0xE3,0xE0,0,
    0xE9,0xE4,0xE1,0,0xEA,0xE5,0xE2,0,0xEB,0xE6,0xE3,0,0xEC,0xE7,0xE4,0,
    0xED,0xE8,0xE5,0,0xEE,0xE9,0xE6,0,0xEF,0xEA,0xE7,0,0xF0,0xEB,0xE8,0,
    0xF1,0xEC,0xE9,0,0xF2,0xED,0xEA,0,0xF3,0xEE,0xEB,0,0xF4,0xEF,0xEC,0,
    0xF5,0xF0,0xED,0,0xF6,0xF1,0xEE,0,0xF7,0xF2,0xEF,0,0xF8,0xF3,0xF0,0,
    0xF9,0xF4,0xF1,0,0xFA,0xF5,0xF2,0,0xFB,0xF6,0xF3,0,0xFC,0xF7,0xF4,0,
    0xFD,0xF8,0xF5,0,0xFE,0xF9,0xF6,0,0xFF,0xFA,0xF7,0,0x00,0xFB,0xF8,0,
    0x01,0xFC,0xF9,0,0x02,0xFD,0xFA,0,0x03,0xFE,0xFB,0,0x04,0xFF,0xFC,0,
    0x05,0x00,0xFD,0,0x06,0x01,0xFE,0,0x07,0x02,0xFF,0,0x08,0x03,0x00,0,
    0x09,0x04,0x01,0,0x0A,0x05,0x02,0,0x0B,0x06,0x03,0,0x0C,0x07,0x04,0,
    0x0D,0x08,0x05,0,0x0E,0x09,0x06,0,0x0F,0x0A,0x07,0,0x10,0x0B,0x08,0,
    0x11,0x0C,0x09,0,0x12,0x0D,0x0A,0,0x13,0x0E,0x0B,0,0x14,0x0F,0x0C,0,
    0x15,0x10,0x0D,0,0x16,0x11,0x0E,0,0x17,0x12,0x0F,0,0x18,0x13,0x10,0,
    0x19,0x14,0x11,0,0x1A,0x15,0x12,0,0x1B,0x16,0x13,0,0x1C,0x17,0x14,0,
    0x1D,0x18,0x15,0,0x1E,0x19,0x16,0,0x1F,0x1A,0x17,0,0x20,0x1B,0x18,0,
    0x21,0x1C,0x19,0,0x22,0x1D,0x1A,0,0x23,0x1E,0x1B,0,0x24,0x1F,0x1C,0,
    0xE6,0xE0,0xDD,0,0xE7,0xE1,0xDE,0,0xE8,0xE2,0xDF,0,0xE9,0xE3,0xE0,0,
    0xEA,0xE4,0xE1,0,0xEB,0xE5,0xE2,0,0xEC,0xE6,0xE3,0,0xED,0xE7,0xE4,0,
    0xEE,0xE8,0xE5,0,0xEF,0xE9,0xE6,0,0xF0,0xEA,0xE7,0,0xF1,0xEB,0xE8,0,
    0xF2,0xEC,0xE9,0,0xF3,0xED,0xEA,0,0xF4,0xEE,0xEB,0,0xF5,0xEF,0xEC,0,
    0xF6,0xF0,0xED,0,0xF7,0xF1,0xEE,0,0xF8,0xF2,0xEF,0,0xF9,0xF3,0xF0,0,
    0xFA,0xF4,0xF1,0,0xFB,0xF5,0xF2,0,0xFC,0xF6,0xF3,0,0xFD,0xF7,0xF4,0,
    0xFE,0xF8,0xF5,0,0xFF,0xF9,0xF6,0,0x00,0xFA,0xF7,0,0x01,0xFB,0xF8,0,
    0x02,0xFC,0xF9,0,0x03,0xFD,0xFA,0,0x04,0xFE,0xFB,0,0x05,0xFF,0xFC,0,
    0x06,0x00,0xFD,0,0x07,0x01,0xFE,0,0x08,0x02,0xFF,0,0x09,0x03,0x00,0,
    0x0A,0x04,0x01,0,0x0B,0x05,0x02,0,0x0C,0x06,0x03,0,0x0D,0x07,0x04,0,
    0x0E,0x08,0x05,0,0x0F,0x09,0x06,0,0x10,0x0A,0x07,0,0x11,0x0B,0x08,0,
    0x12,0x0C,0x09,0,0x13,0x0D,0x0A,0,0x14,0x0E,0x0B,0,0x15,0x0F,0x0C,0,
    0x16,0x10,0x0D,0,0x17,0x11,0x0E,0,0x18,0x12,0x0F,0,0x19,0x13,0x10,0,
    0x1A,0x14,0x11,0,0x1B,0x15,0x12,0,0x1C,0x16,0x13,0,0x1D,0x17,0x14,0,
    0x1E,0x18,0x15,0,0x1F,0x19,0x16,0,0x20,0x1A,0x17,0,0x21,0x1B,0x18,0,
    0x22,0x1C,0x19,0,0x23,0x1D,0x1A,0,0x24,0x1E,0x1B,0,0x25,0x1F,0x1C,0,
    0xE7,0xE0,0xDD,0,0xE8,0xE1,0xDE,0,0xE9,0xE2,0xDF,0,0xEA,0xE3,0xE0,0,
    0xEB,0xE4,0xE1,0,0xEC,0xE5,0xE2,0,0xED,0xE6,0xE3,0,0xEE,0xE7,0xE4,0,
    0xEF,0xE8,0xE5,0,0xF0,0xE9,0xE6,0,0xF1,0xEA,0xE7,0,0xF2,0xEB,0xE8,0,
    0xF3,0xEC,0xE9,0,0xF4,0xED,0xEA,0,0xF5,0xEE,0xEB,0,0xF6,0xEF,0xEC,0,
    0xF7,0xF0,0xED,0,0xF8,0xF1,0xEE,0,0xF9,0xF2,0xEF,0,0xFA,0xF3,0xF0,0,
    0xFB,0xF4,0xF1,0,0xFC,0xF5,0xF2,0,0xFD,0xF6,0xF3,0,0xFE,0xF7,0xF4,0,
    0xFF,0xF8,0xF5,0,0x00,0xF9,0xF6,0,0x01,0xFA,0xF7,0,0x02,0xFB,0xF8,0,
    0x03,0xFC,0xF9,0,0x04,0xFD,0xFA,0,0x05,0xFE,0xFB,0,0x06,0xFF,0xFC,0,
    0x07,0x00,0xFD,0,0x08,0x01,0xFE,0,0x09,0x02,0xFF,0,0x0A,0x03,0x00,0,
    0x0B,0x04,0x01,0,0x0C,0x05,0x02,0,0x0D,0x06,0x03,0,0x0E,0x07,0x04,0,
    0x0F,0x08,0x05,0,0x10,0x09,0x06,0,0x11,0x0A,0x07,0,0x12,0x0B,0x08,0,
    0x13,0x0C,0x09,0,0x14,0x0D,0x0A,0,0x15,0x0E,0x0B,0,0x16,0x0F,0x0C,0,
    0x17,0x10,0x0D,0,0x18,0x11,0x0E,0,0x19,0x12,0x0F,0,0x1A,0x13,0x10,0,
    0x1B,0x14,0x11,0,0x1C,0x15,0x12,0,0x1D,0x16,0x13,0,0x1E,0x17,0x14,0,
    0x1F,0x18,0x15,0,0x20,0x19,0x16,0,0x21,0x1A,0x17,0,0x22,0x1B,0x18,0,
    0x23,0x1C,0x19,0,0x24,0x1D,0x1A,0,0x25,0x1E,0x1B,0,0x26,0x1F,0x1C,0,
    0xD8,0xE0,0xDE,0,0xD9,0xE1,0xDF,0,0xDA,0xE2,0xE0,0,0xDB,0xE3,0xE1,0,
    0xDC,0xE4,0xE2,0,0xDD,0xE5,0xE3,0,0xDE,0xE6,0xE4,0,0xDF,0xE7,0xE5,0,
    0xE0,0xE8,0xE6,0,0xE1,0xE9,0xE7,0,0xE2,0xEA,0xE8,0,0xE3,0xEB,0xE9,0,
    0xE4,0xEC,0xEA,0,0xE5,0xED,0xEB,0,0xE6,0xEE,0xEC,0,0xE7,0xEF,0xED,0,
    0xE8,0xF0,0xEE,0,0xE9,0xF1,0xEF,0,0xEA,0xF2,0xF0,0,0xEB,0xF3,0xF1,0,
    0xEC,0xF4,0xF2,0,0xED,0xF5,0xF3,0,0xEE,0xF6,0xF4,0,0xEF,0xF7,0xF5,0,
    0xF0,0xF8,0xF6,0,0xF1,0xF9,0xF7,0,0xF2,0xFA,0xF8,0,0xF3,0xFB,0xF9,0,
    0xF4,0xFC,0xFA,0,0xF5,0xFD,0xFB,0,0xF6,0xFE,0xFC,0,0xF7,0xFF,0xFD,0,
    0xF8,0x00,0xFE,0,0xF9,0x01,0xFF,0,0xFA,0x02,0x00,0,0xFB,0x03,0x01,0,
    0xFC,0x04,0x02,0,0xFD,0x05,0x03,0,0xFE,0x06,0x04,0,0xFF,0x07,0x05,0,
    0x00,0x08,0x06,0,0x01,0x09,0x07,0,0x02,0x0A,0x08,0,0x03,0x0B,0x09,0,
    0x04,0x0C,0x0A,0,0x05,0x0D,0x0B,0,0x06,0x0E,0x0C,0,0x07,0x0F,0x0D,0,
    0x08,0x10,0x0E,0,0x09,0x11,0x0F,0,0x0A,0x12,0x10,0,0x0B,0x13,0x11,0,
    0x0C,0x14,0x12,0,0x0D,0x15,0x13,0,0x0E,0x16,0x14,0,0x0F,0x17,0x15,0,
    0x10,0x18,0x16,0,0x11,0x19,0x17,0,0x12,0x1A,0x18,0,0x13,0x1B,0x19,0,
    0x14,0x1C,0x1A,0,0x15,0x1D,0x1B,0,0x16,0x1E,0x1C,0,0x17,0x1F,0x1D,0,
    0xD9,0xE0,0xDE,0,0xDA,0xE1,0xDF,0,0xDB,0xE2,0xE0,0,0xDC,0xE3,0xE1,0,
    0xDD,0xE4,0xE2,0,0xDE,0xE5,0xE3,0,0xDF,0xE6,0xE4,0,0xE0,0xE7,0xE5,0,
    0xE1,0xE8,0xE6,0,0xE2,0xE9,0xE7,0,0xE3,0xEA,0xE8,0,0xE4,0xEB,0xE9,0,
    0xE5,0xEC,0xEA,0,0xE6,0xED,0xEB,0,0xE7,0xEE,0xEC,0,0xE8,0xEF,0xED,0,
    0xE9,0xF0,0xEE,0,0xEA,0xF1,0xEF,0,0xEB,0xF2,0xF0,0,0xEC,0xF3,0xF1,0,
    0xED,0xF4,0xF2,0,0xEE,0xF5,0xF3,0,0xEF,0xF6,0xF4,0,0xF0,0xF7,0xF5,0,
    0xF1,0xF8,0xF6,0,0xF2,0xF9,0xF7,0,0xF3,0xFA,0xF8,0,0xF4,0xFB,0xF9,0,
    0xF5,0xFC,0xFA,0,0xF6,0xFD,0xFB,0,0xF7,0xFE,0xFC,0,0xF8,0xFF,0xFD,0,
    0xF9,0x00,0xFE,0,0xFA,0x01,0xFF,0,0xFB,0x02,0x00,0,0xFC,0x03,0x01,0,
    0xFD,0x04,0x02,0,0xFE,0x05,0x03,0,0xFF,0x06,0x04,0,0x00,0x07,0x05,0,
    0x01,0x08,0x06,0,0x02,0x09,0x07,0,0x03,0x0A,0x08,0,0x04,0x0B,0x09,0,
    0x05,0x0C,0x0A,0,0x06,0x0D,0x0B,0,0x07,0x0E,0x0C,0,0x08,0x0F,0x0D,0,
    0x09,0x10,0x0E,0,0x0A,0x11,0x0F,0,0x0B,0x12,0x10,0,0x0C,0x13,0x11,0,
    0x0D,0x14,0x12,0,0x0E,0x15,0x13,0,0x0F,0x16,0x14,0,0x10,0x17,0x15,0,
    0x11,0x18,0x16,0,0x12,0x19,0x17,0,0x13,0x1A,0x18,0,0x14,0x1B,0x19,0,
    0x15,0x1C,0x1A,0,0x16,0x1D,0x1B,0,0x17,0x1E,0x1C,0,0x18,0x1F,0x1D,0,
    0xDA,0xE0,0xDE,0,0xDB,0xE1,0xDF,0,0xDC,0xE2,0xE0,0,0xDD,0xE3,0xE1,0,
    0xDE,0xE4,0xE2,0,0xDF,0xE5,0xE3,0,0xE0,0xE6,0xE4,0,0xE1,0xE7,0xE5,0,
    0xE2,0xE8,0xE6,0,0xE3,0xE9,0xE7,0,0xE4,0xEA,0xE8,0,0xE5,0xEB,0xE9,0,
    0xE6,0xEC,0xEA,0,0xE7,0xED,0xEB,0,0xE8,0xEE,0xEC,0,0xE9,0xEF,0xED,0,
    0xEA,0xF0,0xEE,0,0xEB,0xF1,0xEF,0,0xEC,0xF2,0xF0,0,0xED,0xF3,0xF1,0,
    0xEE,0xF4,0xF2,0,0xEF,0xF5,0xF3,0,0xF0,0xF6,0xF4,0,0xF1,0xF7,0xF5,0,
    0xF2,0xF8,0xF6,0,0xF3,0xF9,0xF7,0,0xF4,0xFA,0xF8,0,0xF5,0xFB,0xF9,0,
    0xF6,0xFC,0xFA,0,0xF7,0xFD,0xFB,0,0xF8,0xFE,0xFC,0,0xF9,0xFF,0xFD,0,
    0xFA,0x00,0xFE,0,0xFB,0x01,0xFF,0,0xFC,0x02,0x00,0,0xFD,0x03,0x01,0,
    0xFE,0x04,0x02,0,0xFF,0x05,0x03,0,0x00,0x06,0x04,0,0x01,0x07,0x05,0,
    0x02,0x08,0x06,0,0x03,0x09,0x07,0,0x04,0x0A,0x08,0,0x05,0x0B,0x09,0,
    0x06,0x0C,0x0A,0,0x07,0x0D,0x0B,0,0x08,0x0E,0x0C,0,0x09,0x0F,0x0D,0,
    0x0A,0x10,0x0E,0,0x0B,0x11,0x0F,0,0x0C,0x12,0x10,0,0x0D,0x13,0x11,0,
    0x0E,0x14,0x12,0,0x0F,0x15,0x13,0,0x10,0x16,0x14,0,0x11,0x17,0x15,0,
    0x12,0x18,0x16,0,0x13,0x19,0x17,0,0x14,0x1A,0x18,0,0x15,0x1B,0x19,0,
    0x16,0x1C,0x1A,0,0x17,0x1D,0x1B,0,0x18,0x1E,0x1C,0,0x19,0x1F,0x1D,0,
    0xDB,0xE0,0xDE,0,0xDC,0xE1,0xDF,0,0xDD,0xE2,0xE0,0,0xDE,0xE3,0xE1,0,
    0xDF,0xE4,0xE2,0,0xE0,0xE5,0xE3,0,0xE1,0xE6,0xE4,0,0xE2,0xE7,0xE5,0,
    0xE3,0xE8,0xE6,0,0xE4,0xE9,0xE7,0,0xE5,0xEA,0xE8,0,0xE6,0xEB,0xE9,0,
    0xE7,0xEC,0xEA,0,0xE8,0xED,0xEB,0,0xE9,0xEE,0xEC,0,0xEA,0xEF,0xED,0,
    0xEB,0xF0,0xEE,0,0xEC,0xF1,0xEF,0,0xED,0xF2,0xF0,0,0xEE,0xF3,0xF1,0,
    0xEF,0xF4,0xF2,0,0xF0,0xF5,0xF3,0,0xF1,0xF6,0xF4,0,0xF2,0xF7,0xF5,0,
    0xF3,0xF8,0xF6,0,0xF4,0xF9,0xF7,0,0xF5,0xFA,0xF8,0,0xF6,0xFB,0xF9,0,
    0xF7,0xFC,0xFA,0,0xF8,0xFD,0xFB,0,0xF9,0xFE,0xFC,0,0xFA,0xFF,0xFD,0,
    0xFB,0x00,0xFE,0,0xFC,0x01,0xFF,0,0xFD,0x02,0x00,0,0xFE,0x03,0x01,0,
    0xFF,0x04,0x02,0,0x00,0x05,0x03,0,0x01,0x06,0x04,0,0x02,0x07,0x05,0,
    0x03,0x08,0x06,0,0x04,0x09,0x07,0,0x05,0x0A,0x08,0,0x06,0x0B,0x09,0,
    0x07,0x0C,0x0A,0,0x08,0x0D,0x0B,0,0x09,0x0E,0x0C,0,0x0A,0x0F,0x0D,0,
    0x0B,0x10,0x0E,0,0x0C,0x11,0x0F,0,0x0D,0x12,0x10,0,0x0E,0x13,0x11,0,
    0x0F,0x14,0x12,0,0x10,0x15,0x13,0,0x11,0x16,0x14,0,0x12,0x17,0x15,0,
    0x13,0x18,0x16,0,0x14,0x19,0x17,0,0x15,0x1A,0x18,0,0x16,0x1B,0x19,0,
    0x17,0x1C,0x1A,0,0x18,0x1D,0x1B,0,0x19,0x1E,0x1C,0,0x1A,0x1F,0x1D,0,
    0xDC,0xE0,0xDE,0,0xDD,0xE1,0xDF,0,0xDE,0xE2,0xE0,0,0xDF,0xE3,0xE1,0,
    0xE0,0xE4,0xE2,0,0xE1,0xE5,0xE3,0,0xE2,0xE6,0xE4,0,0xE3,0xE7,0xE5,0,
    0xE4,0xE8,0xE6,0,0xE5,0xE9,0xE7,0,0xE6,0xEA,0xE8,0,0xE7,0xEB,0xE9,0,
    0xE8,0xEC,0xEA,0,0xE9,0xED,0xEB,0,0xEA,0xEE,0xEC,0,0xEB,0xEF,0xED,0,
    0xEC,0xF0,0xEE,0,0xED,0xF1,0xEF,0,0xEE,0xF2,0xF0,0,0xEF,0xF3,0xF1,0,
    0xF0,0xF4,0xF2,0,0xF1,0xF5,0xF3,0,0xF2,0xF6,0xF4,0,0xF3,0xF7,0xF5,0,
    0xF4,0xF8,0xF6,0,0xF5,0xF9,0xF7,0,0xF6,0xFA,0xF8,0,0xF7,0xFB,0xF9,0,
    0xF8,0xFC,0xFA,0,0xF9,0xFD,0xFB,0,0xFA,0xFE,0xFC,0,0xFB,0xFF,0xFD,0,
    0xFC,0x00,0xFE,0,0xFD,0x01,0xFF,0,0xFE,0x02,0x00,0,0xFF,0x03,0x01,0,
    0x00,0x04,0x02,0,0x01,0x05,0x03,0,0x02,0x06,0x04,0,0x03,0x07,0x05,0,
    0x04,0x08,0x06,0,0x05,0x09,0x07,0,0x06,0x0A,0x08,0,0x07,0x0B,0x09,0,
    0x08,0x0C,0x0A,0,0x09,0x0D,0x0B,0,0x0A,0x0E,0x0C,0,0x0B,0x0F,0x0D,0,
    0x0C,0x10,0x0E,0,0x0D,0x11,0x0F,0,0x0E,0x12,0x10,0,0x0F,0x13,0x11,0,
    0x10,0x14,0x12,0,0x11,0x15,0x13,0,0x12,0x16,0x14,0,0x13,0x17,0x15,0,
    0x14,0x18,0x16,0,0x15,0x19,0x17,0,0x16,0x1A,0x18,0,0x17,0x1B,0x19,0,
    0x18,0x1C,0x1A,0,0x19,0x1D,0x1B,0,0x1A,0x1E,0x1C,0,0x1B,0x1F,0x1D,0,
    0xDD,0xE0,0xDE,0,0xDE,0xE1,0xDF,0,0xDF,0xE2,0xE0,0,0xE0,0xE3,0xE1,0,
    0xE1,0xE4,0xE2,0,0xE2,0xE5,0xE3,0,0xE3,0xE6,0xE4,0,0xE4,0xE7,0xE5,0,
    0xE5,0xE8,0xE6,0,0xE6,0xE9,0xE7,0,0xE7,0xEA,0xE8,0,0xE8,0xEB,0xE9,0,
    0xE9,0xEC,0xEA,0,0xEA,0xED,0xEB,0,0xEB,0xEE,0xEC,0,0xEC,0xEF,0xED,0,
    0xED,0xF0,0xEE,0,0xEE,0xF1,0xEF,0,0xEF,0xF2,0xF0,0,0xF0,0xF3,0xF1,0,
    0xF1,0xF4,0xF2,0,0xF2,0xF5,0xF3,0,0xF3,0xF6,0xF4,0,0xF4,0xF7,0xF5,0,
    0xF5,0xF8,0xF6,0,0xF6,0xF9,0xF7,0,0xF7,0xFA,0xF8,0,0xF8,0xFB,0xF9,0,
    0xF9,0xFC,0xFA,0,0xFA,0xFD,0xFB,0,0xFB,0xFE,0xFC,0,0xFC,0xFF,0xFD,0,
    0xFD,0x00,0xFE,0,0xFE,0x01,0xFF,0,0xFF,0x02,0x00,0,0x00,0x03,0x01,0,
    0x01,0x04,0x02,0,0x02,0x05,0x03,0,0x03,0x06,0x04,0,0x04,0x07,0x05,0,
    0x05,0x08,0x06,0,0x06,0x09,0x07,0,0x07,0x0A,0x08,0,0x08,0x0B,0x09,0,
    0x09,0x0C,0x0A,0,0x0A,0x0D,0x0B,0,0x0B,0x0E,0x0C,0,0x0C,0x0F,0x0D,0,
    0x0D,0x10,0x0E,0,0x0E,0x11,0x0F,0,0x0F,0x12,0x10,0,0x10,0x13,0x11,0,
    0x11,0x14,0x12,0,0x12,0x15,0x13,0,0x13,0x16,0x14,0,0x14,0x17,0x15,0,
    0x15,0x18,0x16,0,0x16,0x19,0x17,0,0x17,0x1A,0x18,0,0x18,0x1B,0x19,0,
    0x19,0x1C,0x1A,0,0x1A,0x1D,0x1B,0,0x1B,0x1E,0x1C,0,0x1C,0x1F,0x1D,0,
    0xDE,0xE0,0xDE,0,0xDF,0xE1,0xDF,0,0xE0,0xE2,0xE0,0,0xE1,0xE3,0xE1,0,
    0xE2,0xE4,0xE2,0,0xE3,0xE5,0xE3,0,0xE4,0xE6,0xE4,0,0xE5,0xE7,0xE5,0,
    0xE6,0xE8,0xE6,0,0xE7,0xE9,0xE7,0,0xE8,0xEA,0xE8,0,0xE9,0xEB,0xE9,0,
    0xEA,0xEC,0xEA,0,0xEB,0xED,0xEB,0,0xEC,0xEE,0xEC,0,0xED,0xEF,0xED,0,
    0xEE,0xF0,0xEE,0,0xEF,0xF1,0xEF,0,0xF0,0xF2,0xF0,0,0xF1,0xF3,0xF1,0,
    0xF2,0xF4,0xF2,0,0xF3,0xF5,0xF3,0,0xF4,0xF6,0xF4,0,0xF5,0xF7,0xF5,0,
    0xF6,0xF8,0xF6,0,0xF7,0xF9,0xF7,0,0xF8,0xFA,0xF8,0,0xF9,0xFB,0xF9,0,
    0xFA,0xFC,0xFA,0,0xFB,0xFD,0xFB,0,0xFC,0xFE,0xFC,0,0xFD,0xFF,0xFD,0,
    0xFE,0x00,0xFE,0,0xFF,0x01,0xFF,0,0x00,0x02,0x00,0,0x01,0x03,0x01,0,
    0x02,0x04,0x02,0,0x03,0x05,0x03,0,0x04,0x06,0x04,0,0x05,0x07,0x05,0,
    0x06,0x08,0x06,0,0x07,0x09,0x07,0,0x08,0x0A,0x08,0,0x09,0x0B,0x09,0,
    0x0A,0x0C,0x0A,0,0x0B,0x0D,0x0B,0,0x0C,0x0E,0x0C,0,0x0D,0x0F,0x0D,0,
    0x0E,0x10,0x0E,0,0x0F,0x11,0x0F,0,0x10,0x12,0x10,0,0x11,0x13,0x11,0,
    0x12,0x14,0x12,0,0x13,0x15,0x13,0,0x14,0x16,0x14,0,0x15,0x17,0x15,0,
    0x16,0x18,0x16,0,0x17,0x19,0x17,0,0x18,0x1A,0x18,0,0x19,0x1B,0x19,0,
    0x1A,0x1C,0x1A,0,0x1B,0x1D,0x1B,0,0x1C,0x1E,0x1C,0,0x1D,0x1F,0x1D,0,
    0xDF,0xE0,0xDE,0,0xE0,0xE1,0xDF,0,0xE1,0xE2,0xE0,0,0xE2,0xE3,0xE1,0,
    0xE3,0xE4,0xE2,0,0xE4,0xE5,0xE3,0,0xE5,0xE6,0xE4,0,0xE6,0xE7,0xE5,0,
    0xE7,0xE8,0xE6,0,0xE8,0xE9,0xE7,0,0xE9,0xEA,0xE8,0,0xEA,0xEB,0xE9,0,
    0xEB,0xEC,0xEA,0,0xEC,0xED,0xEB,0,0xED,0xEE,0xEC,0,0xEE,0xEF,0xED,0,
    0xEF,0xF0,0xEE,0,0xF0,0xF1,0xEF,0,0xF1,0xF2,0xF0,0,0xF2,0xF3,0xF1,0,
    0xF3,0xF4,0xF2,0,0xF4,0xF5,0xF3,0,0xF5,0xF6,0xF4,0,0xF6,0xF7,0xF5,0,
    0xF7,0xF8,0xF6,0,0xF8,0xF9,0xF7,0,0xF9,0xFA,0xF8,0,0xFA,0xFB,0xF9,0,
    0xFB,0xFC,0xFA,0,0xFC,0xFD,0xFB,0,0xFD,0xFE,0xFC,0,0xFE,0xFF,0xFD,0,
    0xFF,0x00,0xFE,0,0x00,0x01,0xFF,0,0x01,0x02,0x00,0,0x02,0x03,0x01,0,
    0x03,0x04,0x02,0,0x04,0x05,0x03,0,0x05,0x06,0x04,0,0x06,0x07,0x05,0,
    0x07,0x08,0x06,0,0x08,0x09,0x07,0,0x09,0x0A,0x08,0,0x0A,0x0B,0x09,0,
    0x0B,0x0C,0x0A,0,0x0C,0x0D,0x0B,0,0x0D,0x0E,0x0C,0,0x0E,0x0F,0x0D,0,
    0x0F,0x10,0x0E,0,0x10,0x11,0x0F,0,0x11,0x12,0x10,0,0x12,0x13,0x11,0,
    0x13,0x14,0x12,0,0x14,0x15,0x13,0,0x15,0x16,0x14,0,0x16,0x17,0x15,0,
    0x17,0x18,0x16,0,0x18,0x19,0x17,0,0x19,0x1A,0x18,0,0x1A,0x1B,0x19,0,
    0x1B,0x1C,0x1A,0,0x1C,0x1D,0x1B,0,0x1D,0x1E,0x1C,0,0x1E,0x1F,0x1D,0,
    0xE0,0xE0,0xDE,0,0xE1,0xE1,0xDF,0,0xE2,0xE2,0xE0,0,0xE3,0xE3,0xE1,0,
    0xE4,0xE4,0xE2,0,0xE5,0xE5,0xE3,0,0xE6,0xE6,0xE4,0,0xE7,0xE7,0xE5,0,
    0xE8,0xE8,0xE6,0,0xE9,0xE9,0xE7,0,0xEA,0xEA,0xE8,0,0xEB,0xEB,0xE9,0,
    0xEC,0xEC,0xEA,0,0xED,0xED,0xEB,0,0xEE,0xEE,0xEC,0,0xEF,0xEF,0xED,0,
    0xF0,0xF0,0xEE,0,0xF1,0xF1,0xEF,0,0xF2,0xF2,0xF0,0,0xF3,0xF3,0xF1,0,
    0xF4,0xF4,0xF2,0,0xF5,0xF5,0xF3,0,0xF6,0xF6,0xF4,0,0xF7,0xF7,0xF5,0,
    0xF8,0xF8,0xF6,0,0xF9,0xF9,0xF7,0,0xFA,0xFA,0xF8,0,0xFB,0xFB,0xF9,0,
    0xFC,0xFC,0xFA,0,0xFD,0xFD,0xFB,0,0xFE,0xFE,0xFC,0,0xFF,0xFF,0xFD,0,
    0x00,0x00,0xFE,0,0x01,0x01,0xFF,0,0x02,0x02,0x00,0,0x03,0x03,0x01,0,
    0x04,0x04,0x02,0,0x05,0x05,0x03,0,0x06,0x06,0x04,0,0x07,0x07,0x05,0,
    0x08,0x08,0x06,0,0x09,0x09,0x07,0,0x0A,0x0A,0x08,0,0x0B,0x0B,0x09,0,
    0x0C,0x0C,0x0A,0,0x0D,0x0D,0x0B,0,0x0E,0x0E,0x0C,0,0x0F,0x0F,0x0D,0,
    0x10,0x10,0x0E,0,0x11,0x11,0x0F,0,0x12,0x12,0x10,0,0x13,0x13,0x11,0,
    0x14,0x14,0x12,0,0x15,0x15,0x13,0,0x16,0x16,0x14,0,0x17,0x17,0x15,0,
    0x18,0x18,0x16,0,0x19,0x19,0x17,0,0x1A,0x1A,0x18,0,0x1B,0x1B,0x19,0,
    0x1C,0x1C,0x1A,0,0x1D,0x1D,0x1B,0,0x1E,0x1E,0x1C,0,0x1F,0x1F,0x1D,0,
    0xE1,0xE0,0xDE,0,0xE2,0xE1,0xDF,0,0xE3,0xE2,0xE0,0,0xE4,0xE3,0xE1,0,
    0xE5,0xE4,0xE2,0,0xE6,0xE5,0xE3,0,0xE7,0xE6,0xE4,0,0xE8,0xE7,0xE5,0,
    0xE9,0xE8,0xE6,0,0xEA,0xE9,0xE7,0,0xEB,0xEA,0xE8,0,0xEC,0xEB,0xE9,0,
    0xED,0xEC,0xEA,0,0xEE,0xED,0xEB,0,0xEF,0xEE,0xEC,0,0xF0,0xEF,0xED,0,
    0xF1,0xF0,0xEE,0,0xF2,0xF1,0xEF,0,0xF3,0xF2,0xF0,0,0xF4,0xF3,0xF1,0,
    0xF5,0xF4,0xF2,0,0xF6,0xF5,0xF3,0,0xF7,0xF6,0xF4,0,0xF8,0xF7,0xF5,0,
    0xF9,0xF8,0xF6,0,0xFA,0xF9,0xF7,0,0xFB,0xFA,0xF8,0,0xFC,0xFB,0xF9,0,
    0xFD,0xFC,0xFA,0,0xFE,0xFD,0xFB,0,0xFF,0xFE,0xFC,0,0x00,0xFF,0xFD,0,
    0x01,0x00,0xFE,0,0x02,0x01,0xFF,0,0x03,0x02,0x00,0,0x04,0x03,0x01,0,
    0x05,0x04,0x02,0,0x06,0x05,0x03,0,0x07,0x06,0x04,0,0x08,0x07,0x05,0,
    0x09,0x08,0x06,0,0x0A,0x09,0x07,0,0x0B,0x0A,0x08,0,0x0C,0x0B,0x09,0,
    0x0D,0x0C,0x0A,0,0x0E,0x0D,0x0B,0,0x0F,0x0E,0x0C,0,0x10,0x0F,0x0D,0,
    0x11,0x10,0x0E,0,0x12,0x11,0x0F,0,0x13,0x12,0x10,0,0x14,0x13,0x11,0,
    0x15,0x14,0x12,0,0x16,0x15,0x13,0,0x17,0x16,0x14,0,0x18,0x17,0x15,0,
    0x19,0x18,0x16,0,0x1A,0x19,0x17,0,0x1B,0x1A,0x18,0,0x1C,0x1B,0x19,0,
    0x1D,0x1C,0x1A,0,0x1E,0x1D,0x1B,0,0x1F,0x1E,0x1C,0,0x20,0x1F,0x1D,0,
    0xE2,0xE0,0xDE,0,0xE3,0xE1,0xDF,0,0xE4,0xE2,0xE0,0,0xE5,0xE3,0xE1,0,
    0xE6,0xE4,0xE2,0,0xE7,0xE5,0xE3,0,0xE8,0xE6,0xE4,0,0xE9,0xE7,0xE5,0,
    0xEA,0xE8,0xE6,0,0xEB,0xE9,0xE7,0,0xEC,0xEA,0xE8,0,0xED,0xEB,0xE9,0,
    0xEE,0xEC,0xEA,0,0xEF,0xED,0xEB,0,0xF0,0xEE,0xEC,0,0xF1,0xEF,0xED,0,
    0xF2,0xF0,0xEE,0,0xF3,0xF1,0xEF,0,0xF4,0xF2,0xF0,0,0xF5,0xF3,0xF1,0,
    0xF6,0xF4,0xF2,0,0xF7,0xF5,0xF3,0,0xF8,0xF6,0xF4,0,0xF9,0xF7,0xF5,0,
    0xFA,0xF8,0xF6,0,0xFB,0xF9,0xF7,0,0xFC,0xFA,0xF8,0,0xFD,0xFB,0xF9,0,
    0xFE,0xFC,0xFA,0,0xFF,0xFD,0xFB,0,0x00,0xFE,0xFC,0,0x01,0xFF,0xFD,0,
    0x02,0x00,0xFE,0,0x03,0x01,0xFF,0,0x04,0x02,0x00,0,0x05,0x03,0x01,0,
    0x06,0x04,0x02,0,0x07,0x05,0x03,0,0x08,0x06,0x04,0,0x09,0x07,0x05,0,
    0x0A,0x08,0x06,0,0x0B,0x09,0x07,0,0x0C,0x0A,0x08,0,0x0D,0x0B,0x09,0,
    0x0E,0x0C,0x0A,0,0x0F,0x0D,0x0B,0,0x10,0x0E,0x0C,0,0x11,0x0F,0x0D,0,
    0x12,0x10,0x0E,0,0x13,0x11,0x0F,0,0x14,0x12,0x10,0,0x15,0x13,0x11,0,
    0x16,0x14,0x12,0,0x17,0x15,0x13,0,0x18,0x16,0x14,0,0x19,0x17,0x15,0,
    0x1A,0x18,0x16,0,0x1B,0x19,0x17,0,0x1C,0x1A,0x18,0,0x1D,0x1B,0x19,0,
    0x1E,0x1C,0x1A,0,0x1F,0x1D,0x1B,0,0x20,0x1E,0x1C,0,0x21,0x1F,0x1D,0,
    0xE3,0xE0,0xDE,0,0xE4,0xE1,0xDF,0,0xE5,0xE2,0xE0,0,0xE6,0xE3,0xE1,0,
    0xE7,0xE4,0xE2,0,0xE8,0xE5,0xE3,0,0xE9,0xE6,0xE4,0,0xEA,0xE7,0xE5,0,
    0xEB,0xE8,0xE6,0,0xEC,0xE9,0xE7,0,0xED,0xEA,0xE8,0,0xEE,0xEB,0xE9,0,
    0xEF,0xEC,0xEA,0,0xF0,0xED,0xEB,0,0xF1,0xEE,0xEC,0,0xF2,0xEF,0xED,0,
    0xF3,0xF0,0xEE,0,0xF4,0xF1,0xEF,0,0xF5,0xF2,0xF0,0,0xF6,0xF3,0xF1,0,
    0xF7,0xF4,0xF2,0,0xF8,0xF5,0xF3,0,0xF9,0xF6,0xF4,0,0xFA,0xF7,0xF5,0,
    0xFB,0xF8,0xF6,0,0xFC,0xF9,0xF7,0,0xFD,0xFA,0xF8,0,0xFE,0xFB,0xF9,0,
    0xFF,0xFC,0xFA,0,0x00,0xFD,0xFB,0,0x01,0xFE,0xFC,0,0x02,0xFF,0xFD,0,
    0x03,0x00,0xFE,0,0x04,0x01,0xFF,0,0x05,0x02,0x00,0,0x06,0x03,0x01,0,
    0x07,0x04,0x02,0,0x08,0x05,0x03,0,0x09,0x06,0x04,0,0x0A,0x07,0x05,0,
    0x0B,0x08,0x06,0,0x0C,0x09,0x07,0,0x0D,0x0A,0x08,0,0x0E,0x0B,0x09,0,
    0x0F,0x0C,0x0A,0,0x10,0x0D,0x0B,0,0x11,0x0E,0x0C,0,0x12,0x0F,0x0D,0,
    0x13,0x10,0x0E,0,0x14,0x11,0x0F,0,0x15,0x12,0x10,0,0x16,0x13,0x11,0,
    0x17,0x14,0x12,0,0x18,0x15,0x13,0,0x19,0x16,0x14,0,0x1A,0x17,0x15,0,
    0x1B,0x18,0x16,0,0x1C,0x19,0x17,0,0x1D,0x1A,0x18,0,0x1E,0x1B,0x19,0,
    0x1F,0x1C,0x1A,0,0x20,0x1D,0x1B,0,0x21,0x1E,0x1C,0,0x22,0x1F,0x1D,0,
    0xE4,0xE0,0xDE,0,0xE5,0xE1,0xDF,0,0xE6,0xE2,0xE0,0,0xE7,0xE3,0xE1,0,
    0xE8,0xE4,0xE2,0,0xE9,0xE5,0xE3,0,0xEA,0xE6,0xE4,0,0xEB,0xE7,0xE5,0,
    0xEC,0xE8,0xE6,0,0xED,0xE9,0xE7,0,0xEE,0xEA,0xE8,0,0xEF,0xEB,0xE9,0,
    0xF0,0xEC,0xEA,0,0xF1,0xED,0xEB,0,0xF2,0xEE,0xEC,0,0xF3,0xEF,0xED,0,
    0xF4,0xF0,0xEE,0,0xF5,0xF1,0xEF,0,0xF6,0xF2,0xF0,0,0xF7,0xF3,0xF1,0,
    0xF8,0xF4,0xF2,0,0xF9,0xF5,0xF3,0,0xFA,0xF6,0xF4,0,0xFB,0xF7,0xF5,0,
    0xFC,0xF8,0xF6,0,0xFD,0xF9,0xF7,0,0xFE,0xFA,0xF8,0,0xFF,0xFB,0xF9,0,
    0x00,0xFC,0xFA,0,0x01,0xFD,0xFB,0,0x02,0xFE,0xFC,0,0x03,0xFF,0xFD,0,
    0x04,0x00,0xFE,0,0x05,0x01,0xFF,0,0x06,0x02,0x00,0,0x07,0x03,0x01,0,
    0x08,0x04,0x02,0,0x09,0x05,0x03,0,0x0A,0x06,0x04,0,0x0B,0x07,0x05,0,
    0x0C,0x08,0x06,0,0x0D,0x09,0x07,0,0x0E,0x0A,0x08,0,0x0F,0x0B,0x09,0,
    0x10,0x0C,0x0A,0,0x11,0x0D,0x0B,0,0x12,0x0E,0x0C,0,0x13,0x0F,0x0D,0,
    0x14,0x10,0x0E,0,0x15,0x11,0x0F,0,0x16,0x12,0x10,0,0x17,0x13,0x11,0,
    0x18,0x14,0x12,0,0x19,0x15,0x13,0,0x1A,0x16,0x14,0,0x1B,0x17,0x15,0,
    0x1C,0x18,0x16,0,0x1D,0x19,0x17,0,0x1E,0x1A,0x18,0,0x1F,0x1B,0x19,0,
    0x20,0x1C,0x1A,0,0x21,0x1D,0x1B,0,0x22,0x1E,0x1C,0,0x23,0x1F,0x1D,0,
    0xE5,0xE0,0xDE,0,0xE6,0xE1,0xDF,0,0xE7,0xE2,0xE0,0,0xE8,0xE3,0xE1,0,
    0xE9,0xE4,0xE2,0,0xEA,0xE5,0xE3,0,0xEB,0xE6,0xE4,0,0xEC,0xE7,0xE5,0,
    0xED,0xE8,0xE6,0,0xEE,0xE9,0xE7,0,0xEF,0xEA,0xE8,0,0xF0,0xEB,0xE9,0,
    0xF1,0xEC,0xEA,0,0xF2,0xED,0xEB,0,0xF3,0xEE,0xEC,0,0xF4,0xEF,0xED,0,
    0xF5,0xF0,0xEE,0,0xF6,0xF1,0xEF,0,0xF7,0xF2,0xF0,0,0xF8,0xF3,0xF1,0,
    0xF9,0xF4,0xF2,0,0xFA,0xF5,0xF3,0,0xFB,0xF6,0xF4,0,0xFC,0xF7,0xF5,0,
    0xFD,0xF8,0xF6,0,0xFE,0xF9,0xF7,0,0xFF,0xFA,0xF8,0,0x00,0xFB,0xF9,0,
    0x01,0xFC,0xFA,0,0x02,0xFD,0xFB,0,0x03,0xFE,0xFC,0,0x04,0xFF,0xFD,0,
    0x05,0x00,0xFE,0,0x06,0x01,0xFF,0,0x07,0x02,0x00,0,0x08,0x03,0x01,0,
    0x09,0x04,0x02,0,0x0A,0x05,0x03,0,0x0B,0x06,0x04,0,0x0C,0x07,0x05,0,
    0x0D,0x08,0x06,0,0x0E,0x09,0x07,0,0x0F,0x0A,0x08,0,0x10,0x0B,0x09,0,
    0x11,0x0C,0x0A,0,0x12,0x0D,0x0B,0,0x13,0x0E,0x0C,0,0x14,0x0F,0x0D,0,
    0x15,0x10,0x0E,0,0x16,0x11,0x0F,0,0x17,0x12,0x10,0,0x18,0x13,0x11,0,
    0x19,0x14,0x12,0,0x1A,0x15,0x13,0,0x1B,0x16,0x14,0,0x1C,0x17,0x15,0,
    0x1D,0x18,0x16,0,0x1E,0x19,0x17,0,0x1F,0x1A,0x18,0,0x20,0x1B,0x19,0,
    0x21,0x1C,0x1A,0,0x22,0x1D,0x1B,0,0x23,0x1E,0x1C,0,0x24,0x1F,0x1D,0,
    0xE6,0xE0,0xDE,0,0xE7,0xE1,0xDF,0,0xE8,0xE2,0xE0,0,0xE9,0xE3,0xE1,0,
    0xEA,0xE4,0xE2,0,0xEB,0xE5,0xE3,0,0xEC,0xE6,0xE4,0,0xED,0xE7,0xE5,0,
    0xEE,0xE8,0xE6,0,0xEF,0xE9,0xE7,0,0xF0,0xEA,0xE8,0,0xF1,0xEB,0xE9,0,
    0xF2,0xEC,0xEA,0,0xF3,0xED,0xEB,0,0xF4,0xEE,0xEC,0,0xF5,0xEF,0xED,0,
    0xF6,0xF0,0xEE,0,0xF7,0xF1,0xEF,0,0xF8,0xF2,0xF0,0,0xF9,0xF3,0xF1,0,
    0xFA,0xF4,0xF2,0,0xFB,0xF5,0xF3,0,0xFC,0xF6,0xF4,0,0xFD,0xF7,0xF5,0,
    0xFE,0xF8,0xF6,0,0xFF,0xF9,0xF7,0,0x00,0xFA,0xF8,0,0x01,0xFB,0xF9,0,
    0x02,0xFC,0xFA,0,0x03,0xFD,0xFB,0,0x04,0xFE,0xFC,0,0x05,0xFF,0xFD,0,
    0x06,0x00,0xFE,0,0x07,0x01,0xFF,0,0x08,0x02,0x00,0,0x09,0x03,0x01,0,
    0x0A,0x04,0x02,0,0x0B,0x05,0x03,0,0x0C,0x06,0x04,0,0x0D,0x07,0x05,0,
    0x0E,0x08,0x06,0,0x0F,0x09,0x07,0,0x10,0x0A,0x08,0,0x11,0x0B,0x09,0,
    0x12,0x0C,0x0A,0,0x13,0x0D,0x0B,0,0x14,0x0E,0x0C,0,0x15,0x0F,0x0D,0,
    0x16,0x10,0x0E,0,0x17,0x11,0x0F,0,0x18,0x12,0x10,0,0x19,0x13,0x11,0,
    0x1A,0x14,0x12,0,0x1B,0x15,0x13,0,0x1C,0x16,0x14,0,0x1D,0x17,0x15,0,
    0x1E,0x18,0x16,0,0x1F,0x19,0x17,0,0x20,0x1A,0x18,0,0x21,0x1B,0x19,0,
    0x22,0x1C,0x1A,0,0x23,0x1D,0x1B,0,0x24,0x1E,0x1C,0,0x25,0x1F,0x1D,0,
    0xE7,0xE0,0xDE,0,0xE8,0xE1,0xDF,0,0xE9,0xE2,0xE0,0,0xEA,0xE3,0xE1,0,
    0xEB,0xE4,0xE2,0,0xEC,0xE5,0xE3,0,0xED,0xE6,0xE4,0,0xEE,0xE7,0xE5,0,
    0xEF,0xE8,0xE6,0,0xF0,0xE9,0xE7,0,0xF1,0xEA,0xE8,0,0xF2,0xEB,0xE9,0,
    0xF3,0xEC,0xEA,0,0xF4,0xED,0xEB,0,0xF5,0xEE,0xEC,0,0xF6,0xEF,0xED,0,
    0xF7,0xF0,0xEE,0,0xF8,0xF1,0xEF,0,0xF9,0xF2,0xF0,0,0xFA,0xF3,0xF1,0,
    0xFB,0xF4,0xF2,0,0xFC,0xF5,0xF3,0,0xFD,0xF6,0xF4,0,0xFE,0xF7,0xF5,0,
    0xFF,0xF8,0xF6,0,0x00,0xF9,0xF7,0,0x01,0xFA,0xF8,0,0x02,0xFB,0xF9,0,
    0x03,0xFC,0xFA,0,0x04,0xFD,0xFB,0,0x05,0xFE,0xFC,0,0x06,0xFF,0xFD,0,
    0x07,0x00,0xFE,0,0x08,0x01,0xFF,0,0x09,0x02,0x00,0,0x0A,0x03,0x01,0,
    0x0B,0x04,0x02,0,0x0C,0x05,0x03,0,0x0D,0x06,0x04,0,0x0E,0x07,0x05,0,
    0x0F,0x08,0x06,0,0x10,0x09,0x07,0,0x11,0x0A,0x08,0,0x12,0x0B,0x09,0,
    0x13,0x0C,0x0A,0,0x14,0x0D,0x0B,0,0x15,0x0E,0x0C,0,0x16,0x0F,0x0D,0,
    0x17,0x10,0x0E,0,0x18,0x11,0x0F,0,0x19,0x12,0x10,0,0x1A,0x13,0x11,0,
    0x1B,0x14,0x12,0,0x1C,0x15,0x13,0,0x1D,0x16,0x14,0,0x1E,0x17,0x15,0,
    0x1F,0x18,0x16,0,0x20,0x19,0x17,0,0x21,0x1A,0x18,0,0x22,0x1B,0x19,0,
    0x23,0x1C,0x1A,0,0x24,0x1D,0x1B,0,0x25,0x1E,0x1C,0,0x26,0x1F,0x1D,0,
    0xD8,0xE0,0xDF,0,0xD9,0xE1,0xE0,0,0xDA,0xE2,0xE1,0,0xDB,0xE3,0xE2,0,
    0xDC,0xE4,0xE3,0,0xDD,0xE5,0xE4,0,0xDE,0xE6,0xE5,0,0xDF,0xE7,0xE6,0,
    0xE0,0xE8,0xE7,0,0xE1,0xE9,0xE8,0,0xE2,0xEA,0xE9,0,0xE3,0xEB,0xEA,0,
    0xE4,0xEC,0xEB,0,0xE5,0xED,0xEC,0,0xE6,0xEE,0xED,0,0xE7,0xEF,0xEE,0,
    0xE8,0xF0,0xEF,0,0xE9,0xF1,0xF0,0,0xEA,0xF2,0xF1,0,0xEB,0xF3,0xF2,0,
    0xEC,0xF4,0xF3,0,0xED,0xF5,0xF4,0,0xEE,0xF6,0xF5,0,0xEF,0xF7,0xF6,0,
    0xF0,0xF8,0xF7,0,0xF1,0xF9,0xF8,0,0xF2,0xFA,0xF9,0,0xF3,0xFB,0xFA,0,
    0xF4,0xFC,0xFB,0,0xF5,0xFD,0xFC,0,0xF6,0xFE,0xFD,0,0xF7,0xFF,0xFE,0,
    0xF8,0x00,0xFF,0,0xF9,0x01,0x00,0,0xFA,0x02,0x01,0,0xFB,0x03,0x02,0,
    0xFC,0x04,0x03,0,0xFD,0x05,0x04,0,0xFE,0x06,0x05,0,0xFF,0x07,0x06,0,
    0x00,0x08,0x07,0,0x01,0x09,0x08,0,0x02,0x0A,0x09,0,0x03,0x0B,0x0A,0,
    0x04,0x0C,0x0B,0,0x05,0x0D,0x0C,0,0x06,0x0E,0x0D,0,0x07,0x0F,0x0E,0,
    0x08,0x10,0x0F,0,0x09,0x11,0x10,0,0x0A,0x12,0x11,0,0x0B,0x13,0x12,0,
    0x0C,0x14,0x13,0,0x0D,0x15,0x14,0,0x0E,0x16,0x15,0,0x0F,0x17,0x16,0,
    0x10,0x18,0x17,0,0x11,0x19,0x18,0,0x12,0x1A,0x19,0,0x13,0x1B,0x1A,0,
    0x14,0x1C,0x1B,0,0x15,0x1D,0x1C,0,0x16,0x1E,0x1D,0,0x17,0x1F,0x1E,0,
    0xD9,0xE0,0xDF,0,0xDA,0xE1,0xE0,0,0xDB,0xE2,0xE1,0,0xDC,0xE3,0xE2,0,
    0xDD,0xE4,0xE3,0,0xDE,0xE5,0xE4,0,0xDF,0xE6,0xE5,0,0xE0,0xE7,0xE6,0,
    0xE1,0xE8,0xE7,0,0xE2,0xE9,0xE8,0,0xE3,0xEA,0xE9,0,0xE4,0xEB,0xEA,0,
    0xE5,0xEC,0xEB,0,0xE6,0xED,0xEC,0,0xE7,0xEE,0xED,0,0xE8,0xEF,0xEE,0,
    0xE9,0xF0,0xEF,0,0xEA,0xF1,0xF0,0,0xEB,0xF2,0xF1,0,0xEC,0xF3,0xF2,0,
    0xED,0xF4,0xF3,0,0xEE,0xF5,0xF4,0,0xEF,0xF6,0xF5,0,0xF0,0xF7,0xF6,0,
    0xF1,0xF8,0xF7,0,0xF2,0xF9,0xF8,0,0xF3,0xFA,0xF9,0,0xF4,0xFB,0xFA,0,
    0xF5,0xFC,0xFB,0,0xF6,0xFD,0xFC,0,0xF7,0xFE,0xFD,0,0xF8,0xFF,0xFE,0,
    0xF9,0x00,0xFF,0,0xFA,0x01,0x00,0,0xFB,0x02,0x01,0,0xFC,0x03,0x02,0,
    0xFD,0x04,0x03,0,0xFE,0x05,0x04,0,0xFF,0x06,0x05,0,0x00,0x07,0x06,0,
    0x01,0x08,0x07,0,0x02,0x09,0x08,0,0x03,0x0A,0x09,0,0x04,0x0B,0x0A,0,
    0x05,0x0C,0x0B,0,0x06,0x0D,0x0C,0,0x07,0x0E,0x0D,0,0x08,0x0F,0x0E,0,
    0x09,0x10,0x0F,0,0x0A,0x11,0x10,0,0x0B,0x12,0x11,0,0x0C,0x13,0x12,0,
    0x0D,0x14,0x13,0,0x0E,0x15,0x14,0,0x0F,0x16,0x15,0,0x10,0x17,0x16,0,
    0x11,0x18,0x17,0,0x12,0x19,0x18,0,0x13,0x1A,0x19,0,0x14,0x1B,0x1A,0,
    0x15,0x1C,0x1B,0,0x16,0x1D,0x1C,0,0x17,0x1E,0x1D,0,0x18,0x1F,0x1E,0,
    0xDA,0xE0,0xDF,0,0xDB,0xE1,0xE0,0,0xDC,0xE2,0xE1,0,0xDD,0xE3,0xE2,0,
    0xDE,0xE4,0xE3,0,0xDF,0xE5,0xE4,0,0xE0,0xE6,0xE5,0,0xE1,0xE7,0xE6,0,
    0xE2,0xE8,0xE7,0,0xE3,0xE9,0xE8,0,0xE4,0xEA,0xE9,0,0xE5,0xEB,0xEA,0,
    0xE6,0xEC,0xEB,0,0xE7,0xED,0xEC,0,0xE8,0xEE,0xED,0,0xE9,0xEF,0xEE,0,
    0xEA,0xF0,0xEF,0,0xEB,0xF1,0xF0,0,0xEC,0xF2,0xF1,0,0xED,0xF3,0xF2,0,
    0xEE,0xF4,0xF3,0,0xEF,0xF5,0xF4,0,0xF0,0xF6,0xF5,0,0xF1,0xF7,0xF6,0,
    0xF2,0xF8,0xF7,0,0xF3,0xF9,0xF8,0,0xF4,0xFA,0xF9,0,0xF5,0xFB,0xFA,0,
    0xF6,0xFC,0xFB,0,0xF7,0xFD,0xFC,0,0xF8,0xFE,0xFD,0,0xF9,0xFF,0xFE,0,
    0xFA,0x00,0xFF,0,0xFB,0x01,0x00,0,0xFC,0x02,0x01,0,0xFD,0x03,0x02,0,
    0xFE,0x04,0x03,0,0xFF,0x05,0x04,0,0x00,0x06,0x05,0,0x01,0x07,0x06,0,
    0x02,0x08,0x07,0,0x03,0x09,0x08,0,0x04,0x0A,0x09,0,0x05,0x0B,0x0A,0,
    0x06,0x0C,0x0B,0,0x07,0x0D,0x0C,0,0x08,0x0E,0x0D,0,0x09,0x0F,0x0E,0,
    0x0A,0x10,0x0F,0,0x0B,0x11,0x10,0,0x0C,0x12,0x11,0,0x0D,0x13,0x12,0,
    0x0E,0x14,0x13,0,0x0F,0x15,0x14,0,0x10,0x16,0x15,0,0x11,0x17,0x16,0,
    0x12,0x18,0x17,0,0x13,0x19,0x18,0,0x14,0x1A,0x19,0,0x15,0x1B,0x1A,0,
    0x16,0x1C,0x1B,0,0x17,0x1D,0x1C,0,0x18,0x1E,0x1D,0,0x19,0x1F,0x1E,0,
    0xDB,0xE0,0xDF,0,0xDC,0xE1,0xE0,0,0xDD,0xE2,0xE1,0,0xDE,0xE3,0xE2,0,
    0xDF,0xE4,0xE3,0,0xE0,0xE5,0xE4,0,0xE1,0xE6,0xE5,0,0xE2,0xE7,0xE6,0,
    0xE3,0xE8,0xE7,0,0xE4,0xE9,0xE8,0,0xE5,0xEA,0xE9,0,0xE6,0xEB,0xEA,0,
    0xE7,0xEC,0xEB,0,0xE8,0xED,0xEC,0,0xE9,0xEE,0xED,0,0xEA,0xEF,0xEE,0,
    0xEB,0xF0,0xEF,0,0xEC,0xF1,0xF0,0,0xED,0xF2,0xF1,0,0xEE,0xF3,0xF2,0,
    0xEF,0xF4,0xF3,0,0xF0,0xF5,0xF4,0,0xF1,0xF6,0xF5,0,0xF2,0xF7,0xF6,0,
    0xF3,0xF8,0xF7,0,0xF4,0xF9,0xF8,0,0xF5,0xFA,0xF9,0,0xF6,0xFB,0xFA,0,
    0xF7,0xFC,0xFB,0,0xF8,0xFD,0xFC,0,0xF9,0xFE,0xFD,0,0xFA,0xFF,0xFE,0,
    0xFB,0x00,0xFF,0,0xFC,0x01,0x00,0,0xFD,0x02,0x01,0,0xFE,0x03,0x02,0,
    0xFF,0x04,0x03,0,0x00,0x05,0x04,0,0x01,0x06,0x05,0,0x02,0x07,0x06,0,
    0x03,0x08,0x07,0,0x04,0x09,0x08,0,0x05,0x0A,0x09,0,0x06,0x0B,0x0A,0,
    0x07,0x0C,0x0B,0,0x08,0x0D,0x0C,0,0x09,0x0E,0x0D,0,0x0A,0x0F,0x0E,0,
    0x0B,0x10,0x0F,0,0x0C,0x11,0x10,0,0x0D,0x12,0x11,0,0x0E,0x13,0x12,0,
    0x0F,0x14,0x13,0,0x10,0x15,0x14,0,0x11,0x16,0x15,0,0x12,0x17,0x16,0,
    0x13,0x18,0x17,0,0x14,0x19,0x18,0,0x15,0x1A,0x19,0,0x16,0x1B,0x1A,0,
    0x17,0x1C,0x1B,0,0x18,0x1D,0x1C,0,0x19,0x1E,0x1D,0,0x1A,0x1F,0x1E,0,
    0xDC,0xE0,0xDF,0,0xDD,0xE1,0xE0,0,0xDE,0xE2,0xE1,0,0xDF,0xE3,0xE2,0,
    0xE0,0xE4,0xE3,0,0xE1,0xE5,0xE4,0,0xE2,0xE6,0xE5,0,0xE3,0xE7,0xE6,0,
    0xE4,0xE8,0xE7,0,0xE5,0xE9,0xE8,0,0xE6,0xEA,0xE9,0,0xE7,0xEB,0xEA,0,
    0xE8,0xEC,0xEB,0,0xE9,0xED,0xEC,0,0xEA,0xEE,0xED,0,0xEB,0xEF,0xEE,0,
    0xEC,0xF0,0xEF,0,0xED,0xF1,0xF0,0,0xEE,0xF2,0xF1,0,0xEF,0xF3,0xF2,0,
    0xF0,0xF4,0xF3,0,0xF1,0xF5,0xF4,0,0xF2,0xF6,0xF5,0,0xF3,0xF7,0xF6,0,
    0xF4,0xF8,0xF7,0,0xF5,0xF9,0xF8,0,0xF6,0xFA,0xF9,0,0xF7,0xFB,0xFA,0,
    0xF8,0xFC,0xFB,0,0xF9,0xFD,0xFC,0,0xFA,0xFE,0xFD,0,0xFB,0xFF,0xFE,0,
    0xFC,0x00,0xFF,0,0xFD,0x01,0x00,0,0xFE,0x02,0x01,0,0xFF,0x03,0x02,0,
    0x00,0x04,0x03,0,0x01,0x05,0x04,0,0x02,0x06,0x05,0,0x03,0x07,0x06,0,
    0x04,0x08,0x07,0,0x05,0x09,0x08,0,0x06,0x0A,0x09,0,0x07,0x0B,0x0A,0,
    0x08,0x0C,0x0B,0,0x09,0x0D,0x0C,0,0x0A,0x0E,0x0D,0,0x0B,0x0F,0x0E,0,
    0x0C,0x10,0x0F,0,0x0D,0x11,0x10,0,0x0E,0x12,0x11,0,0x0F,0x13,0x12,0,
    0x10,0x14,0x13,0,0x11,0x15,0x14,0,0x12,0x16,0x15,0,0x13,0x17,0x16,0,
    0x14,0x18,0x17,0,0x15,0x19,0x18,0,0x16,0x1A,0x19,0,0x17,0x1B,0x1A,0,
    0x18,0x1C,0x1B,0,0x19,0x1D,0x1C,0,0x1A,0x1E,0x1D,0,0x1B,0x1F,0x1E,0,
    0xDD,0xE0,0xDF,0,0xDE,0xE1,0xE0,0,0xDF,0xE2,0xE1,0,0xE0,0xE3,0xE2,0,
    0xE1,0xE4,0xE3,0,0xE2,0xE5,0xE4,0,0xE3,0xE6,0xE5,0,0xE4,0xE7,0xE6,0,
    0xE5,0xE8,0xE7,0,0xE6,0xE9,0xE8,0,0xE7,0xEA,0xE9,0,0xE8,0xEB,0xEA,0,
    0xE9,0xEC,0xEB,0,0xEA,0xED,0xEC,0,0xEB,0xEE,0xED,0,0xEC,0xEF,0xEE,0,
    0xED,0xF0,0xEF,0,0xEE,0xF1,0xF0,0,0xEF,0xF2,0xF1,0,0xF0,0xF3,0xF2,0,
    0xF1,0xF4,0xF3,0,0xF2,0xF5,0xF4,0,0xF3,0xF6,0xF5,0,0xF4,0xF7,0xF6,0,
    0xF5,0xF8,0xF7,0,0xF6,0xF9,0xF8,0,0xF7,0xFA,0xF9,0,0xF8,0xFB,0xFA,0,
    0xF9,0xFC,0xFB,0,0xFA,0xFD,0xFC,0,0xFB,0xFE,0xFD,0,0xFC,0xFF,0xFE,0,
    0xFD,0x00,0xFF,0,0xFE,0x01,0x00,0,0xFF,0x02,0x01,0,0x00,0x03,0x02,0,
    0x01,0x04,0x03,0,0x02,0x05,0x04,0,0x03,0x06,0x05,0,0x04,0x07,0x06,0,
    0x05,0x08,0x07,0,0x06,0x09,0x08,0,0x07,0x0A,0x09,0,0x08,0x0B,0x0A,0,
    0x09,0x0C,0x0B,0,0x0A,0x0D,0x0C,0,0x0B,0x0E,0x0D,0,0x0C,0x0F,0x0E,0,
    0x0D,0x10,0x0F,0,0x0E,0x11,0x10,0,0x0F,0x12,0x11,0,0x10,0x13,0x12,0,
    0x11,0x14,0x13,0,0x12,0x15,0x14,0,0x13,0x16,0x15,0,0x14,0x17,0x16,0,
    0x15,0x18,0x17,0,0x16,0x19,0x18,0,0x17,0x1A,0x19,0,0x18,0x1B,0x1A,0,
    0x19,0x1C,0x1B,0,0x1A,0x1D,0x1C,0,0x1B,0x1E,0x1D,0,0x1C,0x1F,0x1E,0,
    0xDE,0xE0,0xDF,0,0xDF,0xE1,0xE0,0,0xE0,0xE2,0xE1,0,0xE1,0xE3,0xE2,0,
    0xE2,0xE4,0xE3,0,0xE3,0xE5,0xE4,0,0xE4,0xE6,0xE5,0,0xE5,0xE7,0xE6,0,
    0xE6,0xE8,0xE7,0,0xE7,0xE9,0xE8,0,0xE8,0xEA,0xE9,0,0xE9,0xEB,0xEA,0,
    0xEA,0xEC,0xEB,0,0xEB,0xED,0xEC,0,0xEC,0xEE,0xED,0,0xED,0xEF,0xEE,0,
    0xEE,0xF0,0xEF,0,0xEF,0xF1,0xF0,0,0xF0,0xF2,0xF1,0,0xF1,0xF3,0xF2,0,
    0xF2,0xF4,0xF3,0,0xF3,0xF5,0xF4,0,0xF4,0xF6,0xF5,0,0xF5,0xF7,0xF6,0,
    0xF6,0xF8,0xF7,0,0xF7,0xF9,0xF8,0,0xF8,0xFA,0xF9,0,0xF9,0xFB,0xFA,0,
    0xFA,0xFC,0xFB,0,0xFB,0xFD,0xFC,0,0xFC,0xFE,0xFD,0,0xFD,0xFF,0xFE,0,
    0xFE,0x00,0xFF,0,0xFF,0x01,0x00,0,0x00,0x02,0x01,0,0x01,0x03,0x02,0,
    0x02,0x04,0x03,0,0x03,0x05,0x04,0,0x04,0x06,0x05,0,0x05,0x07,0x06,0,
    0x06,0x08,0x07,0,0x07,0x09,0x08,0,0x08,0x0A,0x09,0,0x09,0x0B,0x0A,0,
    0x0A,0x0C,0x0B,0,0x0B,0x0D,0x0C,0,0x0C,0x0E,0x0D,0,0x0D,0x0F,0x0E,0,
    0x0E,0x10,0x0F,0,0x0F,0x11,0x10,0,0x10,0x12,0x11,0,0x11,0x13,0x12,0,
    0x12,0x14,0x13,0,0x13,0x15,0x14,0,0x14,0x16,0x15,0,0x15,0x17,0x16,0,
    0x16,0x18,0x17,0,0x17,0x19,0x18,0,0x18,0x1A,0x19,0,0x19,0x1B,0x1A,0,
    0x1A,0x1C,0x1B,0,0x1B,0x1D,0x1C,0,0x1C,0x1E,0x1D,0,0x1D,0x1F,0x1E,0,
    0xDF,0xE0,0xDF,0,0xE0,0xE1,0xE0,0,0xE1,0xE2,0xE1,0,0xE2,0xE3,0xE2,0,
    0xE3,0xE4,0xE3,0,0xE4,0xE5,0xE4,0,0xE5,0xE6,0xE5,0,0xE6,0xE7,0xE6,0,
    0xE7,0xE8,0xE7,0,0xE8,0xE9,0xE8,0,0xE9,0xEA,0xE9,0,0xEA,0xEB,0xEA,0,
    0xEB,0xEC,0xEB,0,0xEC,0xED,0xEC,0,0xED,0xEE,0xED,0,0xEE,0xEF,0xEE,0,
    0xEF,0xF0,0xEF,0,0xF0,0xF1,0xF0,0,0xF1,0xF2,0xF1,0,0xF2,0xF3,0xF2,0,
    0xF3,0xF4,0xF3,0,0xF4,0xF5,0xF4,0,0xF5,0xF6,0xF5,0,0xF6,0xF7,0xF6,0,
    0xF7,0xF8,0xF7,0,0xF8,0xF9,0xF8,0,0xF9,0xFA,0xF9,0,0xFA,0xFB,0xFA,0,
    0xFB,0xFC,0xFB,0,0xFC,0xFD,0xFC,0,0xFD,0xFE,0xFD,0,0xFE,0xFF,0xFE,0,
    0xFF,0x00,0xFF,0,0x00,0x01,0x00,0,0x01,0x02,0x01,0,0x02,0x03,0x02,0,
    0x03,0x04,0x03,0,0x04,0x05,0x04,0,0x05,0x06,0x05,0,0x06,0x07,0x06,0,
    0x07,0x08,0x07,0,0x08,0x09,0x08,0,0x09,0x0A,0x09,0,0x0A,0x0B,0x0A,0,
    0x0B,0x0C,0x0B,0,0x0C,0x0D,0x0C,0,0x0D,0x0E,0x0D,0,0x0E,0x0F,0x0E,0,
    0x0F,0x10,0x0F,0,0x10,0x11,0x10,0,0x11,0x12,0x11,0,0x12,0x13,0x12,0,
    0x13,0x14,0x13,0,0x14,0x15,0x14,0,0x15,0x16,0x15,0,0x16,0x17,0x16,0,
    0x17,0x18,0x17,0,0x18,0x19,0x18,0,0x19,0x1A,0x19,0,0x1A,0x1B,0x1A,0,
    0x1B,0x1C,0x1B,0,0x1C,0x1D,0x1C,0,0x1D,0x1E,0x1D,0,0x1E,0x1F,0x1E,0,
    0xE0,0xE0,0xDF,0,0xE1,0xE1,0xE0,0,0xE2,0xE2,0xE1,0,0xE3,0xE3,0xE2,0,
    0xE4,0xE4,0xE3,0,0xE5,0xE5,0xE4,0,0xE6,0xE6,0xE5,0,0xE7,0xE7,0xE6,0,
    0xE8,0xE8,0xE7,0,0xE9,0xE9,0xE8,0,0xEA,0xEA,0xE9,0,0xEB,0xEB,0xEA,0,
    0xEC,0xEC,0xEB,0,0xED,0xED,0xEC,0,0xEE,0xEE,0xED,0,0xEF,0xEF,0xEE,0,
    0xF0,0xF0,0xEF,0,0xF1,0xF1,0xF0,0,0xF2,0xF2,0xF1,0,0xF3,0xF3,0xF2,0,
    0xF4,0xF4,0xF3,0,0xF5,0xF5,0xF4,0,0xF6,0xF6,0xF5,0,0xF7,0xF7,0xF6,0,
    0xF8,0xF8,0xF7,0,0xF9,0xF9,0xF8,0,0xFA,0xFA,0xF9,0,0xFB,0xFB,0xFA,0,
    0xFC,0xFC,0xFB,0,0xFD,0xFD,0xFC,0,0xFE,0xFE,0xFD,0,0xFF,0xFF,0xFE,0,
    0x00,0x00,0xFF,0,0x01,0x01,0x00,0,0x02,0x02,0x01,0,0x03,0x03,0x02,0,
    0x04,0x04,0x03,0,0x05,0x05,0x04,0,0x06,0x06,0x05,0,0x07,0x07,0x06,0,
    0x08,0x08,0x07,0,0x09,0x09,0x08,0,0x0A,0x0A,0x09,0,0x0B,0x0B,0x0A,0,
    0x0C,0x0C,0x0B,0,0x0D,0x0D,0x0C,0,0x0E,0x0E,0x0D,0,0x0F,0x0F,0x0E,0,
    0x10,0x10,0x0F,0,0x11,0x11,0x10,0,0x12,0x12,0x11,0,0x13,0x13,0x12,0,
    0x14,0x14,0x13,0,0x15,0x15,0x14,0,0x16,0x16,0x15,0,0x17,0x17,0x16,0,
    0x18,0x18,0x17,0,0x19,0x19,0x18,0,0x1A,0x1A,0x19,0,0x1B,0x1B,0x1A,0,
    0x1C,0x1C,0x1B,0,0x1D,0x1D,0x1C,0,0x1E,0x1E,0x1D,0,0x1F,0x1F,0x1E,0,
    0xE1,0xE0,0xDF,0,0xE2,0xE1,0xE0,0,0xE3,0xE2,0xE1,0,0xE4,0xE3,0xE2,0,
    0xE5,0xE4,0xE3,0,0xE6,0xE5,0xE4,0,0xE7,0xE6,0xE5,0,0xE8,0xE7,0xE6,0,
    0xE9,0xE8,0xE7,0,0xEA,0xE9,0xE8,0,0xEB,0xEA,0xE9,0,0xEC,0xEB,0xEA,0,
    0xED,0xEC,0xEB,0,0xEE,0xED,0xEC,0,0xEF,0xEE,0xED,0,0xF0,0xEF,0xEE,0,
    0xF1,0xF0,0xEF,0,0xF2,0xF1,0xF0,0,0xF3,0xF2,0xF1,0,0xF4,0xF3,0xF2,0,
    0xF5,0xF4,0xF3,0,0xF6,0xF5,0xF4,0,0xF7,0xF6,0xF5,0,0xF8,0xF7,0xF6,0,
    0xF9,0xF8,0xF7,0,0xFA,0xF9,0xF8,0,0xFB,0xFA,0xF9,0,0xFC,0xFB,0xFA,0,
    0xFD,0xFC,0xFB,0,0xFE,0xFD,0xFC,0,0xFF,0xFE,0xFD,0,0x00,0xFF,0xFE,0,
    0x01,0x00,0xFF,0,0x02,0x01,0x00,0,0x03,0x02,0x01,0,0x04,0x03,0x02,0,
    0x05,0x04,0x03,0,0x06,0x05,0x04,0,0x07,0x06,0x05,0,0x08,0x07,0x06,0,
    0x09,0x08,0x07,0,0x0A,0x09,0x08,0,0x0B,0x0A,0x09,0,0x0C,0x0B,0x0A,0,
    0x0D,0x0C,0x0B,0,0x0E,0x0D,0x0C,0,0x0F,0x0E,0x0D,0,0x10,0x0F,0x0E,0,
    0x11,0x10,0x0F,0,0x12,0x11,0x10,0,0x13,0x12,0x11,0,0x14,0x13,0x12,0,
    0x15,0x14,0x13,0,0x16,0x15,0x14,0,0x17,0x16,0x15,0,0x18,0x17,0x16,0,
    0x19,0x18,0x17,0,0x1A,0x19,0x18,0,0x1B,0x1A,0x19,0,0x1C,0x1B,0x1A,0,
    0x1D,0x1C,0x1B,0,0x1E,0x1D,0x1C,0,0x1F,0x1E,0x1D,0,0x20,0x1F,0x1E,0,
    0xE2,0xE0,0xDF,0,0xE3,0xE1,0xE0,0,0xE4,0xE2,0xE1,0,0xE5,0xE3,0xE2,0,
    0xE6,0xE4,0xE3,0,0xE7,0xE5,0xE4,0,0xE8,0xE6,0xE5,0,0xE9,0xE7,0xE6,0,
    0xEA,0xE8,0xE7,0,0xEB,0xE9,0xE8,0,0xEC,0xEA,0xE9,0,0xED,0xEB,0xEA,0,
    0xEE,0xEC,0xEB,0,0xEF,0xED,0xEC,0,0xF0,0xEE,0xED,0,0xF1,0xEF,0xEE,0,
    0xF2,0xF0,0xEF,0,0xF3,0xF1,0xF0,0,0xF4,0xF2,0xF1,0,0xF5,0xF3,0xF2,0,
    0xF6,0xF4,0xF3,0,0xF7,0xF5,0xF4,0,0xF8,0xF6,0xF5,0,0xF9,0xF7,0xF6,0,
    0xFA,0xF8,0xF7,0,0xFB,0xF9,0xF8,0,0xFC,0xFA,0xF9,0,0xFD,0xFB,0xFA,0,
    0xFE,0xFC,0xFB,0,0xFF,0xFD,0xFC,0,0x00,0xFE,0xFD,0,0x01,0xFF,0xFE,0,
    0x02,0x00,0xFF,0,0x03,0x01,0x00,0,0x04,0x02,0x01,0,0x05,0x03,0x02,0,
    0x06,0x04,0x03,0,0x07,0x05,0x04,0,0x08,0x06,0x05,0,0x09,0x07,0x06,0,
    0x0A,0x08,0x07,0,0x0B,0x09,0x08,0,0x0C,0x0A,0x09,0,0x0D,0x0B,0x0A,0,
    0x0E,0x0C,0x0B,0,0x0F,0x0D,0x0C,0,0x10,0x0E,0x0D,0,0x11,0x0F,0x0E,0,
    0x12,0x10,0x0F,0,0x13,0x11,0x10,0,0x14,0x12,0x11,0,0x15,0x13,0x12,0,
    0x16,0x14,0x13,0,0x17,0x15,0x14,0,0x18,0x16,0x15,0,0x19,0x17,0x16,0,
    0x1A,0x18,0x17,0,0x1B,0x19,0x18,0,0x1C,0x1A,0x19,0,0x1D,0x1B,0x1A,0,
    0x1E,0x1C,0x1B,0,0x1F,0x1D,0x1C,0,0x20,0x1E,0x1D,0,0x21,0x1F,0x1E,0,
    0xE3,0xE0,0xDF,0,0xE4,0xE1,0xE0,0,0xE5,0xE2,0xE1,0,0xE6,0xE3,0xE2,0,
    0xE7,0xE4,0xE3,0,0xE8,0xE5,0xE4,0,0xE9,0xE6,0xE5,0,0xEA,0xE7,0xE6,0,
    0xEB,0xE8,0xE7,0,0xEC,0xE9,0xE8,0,0xED,0xEA,0xE9,0,0xEE,0xEB,0xEA,0,
    0xEF,0xEC,0xEB,0,0xF0,0xED,0xEC,0,0xF1,0xEE,0xED,0,0xF2,0xEF,0xEE,0,
    0xF3,0xF0,0xEF,0,0xF4,0xF1,0xF0,0,0xF5,0xF2,0xF1,0,0xF6,0xF3,0xF2,0,
    0xF7,0xF4,0xF3,0,0xF8,0xF5,0xF4,0,0xF9,0xF6,0xF5,0,0xFA,0xF7,0xF6,0,
    0xFB,0xF8,0xF7,0,0xFC,0xF9,0xF8,0,0xFD,0xFA,0xF9,0,0xFE,0xFB,0xFA,0,
    0xFF,0xFC,0xFB,0,0x00,0xFD,0xFC,0,0x01,0xFE,0xFD,0,0x02,0xFF,0xFE,0,
    0x03,0x00,0xFF,0,0x04,0x01,0x00,0,0x05,0x02,0x01,0,0x06,0x03,0x02,0,
    0x07,0x04,0x03,0,0x08,0x05,0x04,0,0x09,0x06,0x05,0,0x0A,0x07,0x06,0,
    0x0B,0x08,0x07,0,0x0C,0x09,0x08,0,0x0D,0x0A,0x09,0,0x0E,0x0B,0x0A,0,
    0x0F,0x0C,0x0B,0,0x10,0x0D,0x0C,0,0x11,0x0E,0x0D,0,0x12,0x0F,0x0E,0,
    0x13,0x10,0x0F,0,0x14,0x11,0x10,0,0x15,0x12,0x11,0,0x16,0x13,0x12,0,
    0x17,0x14,0x13,0,0x18,0x15,0x14,0,0x19,0x16,0x15,0,0x1A,0x17,0x16,0,
    0x1B,0x18,0x17,0,0x1C,0x19,0x18,0,0x1D,0x1A,0x19,0,0x1E,0x1B,0x1A,0,
    0x1F,0x1C,0x1B,0,0x20,0x1D,0x1C,0,0x21,0x1E,0x1D,0,0x22,0x1F,0x1E,0,
    0xE4,0xE0,0xDF,0,0xE5,0xE1,0xE0,0,0xE6,0xE2,0xE1,0,0xE7,0xE3,0xE2,0,
    0xE8,0xE4,0xE3,0,0xE9,0xE5,0xE4,0,0xEA,0xE6,0xE5,0,0xEB,0xE7,0xE6,0,
    0xEC,0xE8,0xE7,0,0xED,0xE9,0xE8,0,0xEE,0xEA,0xE9,0,0xEF,0xEB,0xEA,0,
    0xF0,0xEC,0xEB,0,0xF1,0xED,0xEC,0,0xF2,0xEE,0xED,0,0xF3,0xEF,0xEE,0,
    0xF4,0xF0,0xEF,0,0xF5,0xF1,0xF0,0,0xF6,0xF2,0xF1,0,0xF7,0xF3,0xF2,0,
    0xF8,0xF4,0xF3,0,0xF9,0xF5,0xF4,0,0xFA,0xF6,0xF5,0,0xFB,0xF7,0xF6,0,
    0xFC,0xF8,0xF7,0,0xFD,0xF9,0xF8,0,0xFE,0xFA,0xF9,0,0xFF,0xFB,0xFA,0,
    0x00,0xFC,0xFB,0,0x01,0xFD,0xFC,0,0x02,0xFE,0xFD,0,0x03,0xFF,0xFE,0,
    0x04,0x00,0xFF,0,0x05,0x01,0x00,0,0x06,0x02,0x01,0,0x07,0x03,0x02,0,
    0x08,0x04,0x03,0,0x09,0x05,0x04,0,0x0A,0x06,0x05,0,0x0B,0x07,0x06,0,
    0x0C,0x08,0x07,0,0x0D,0x09,0x08,0,0x0E,0x0A,0x09,0,0x0F,0x0B,0x0A,0,
    0x10,0x0C,0x0B,0,0x11,0x0D,0x0C,0,0x12,0x0E,0x0D,0,0x13,0x0F,0x0E,0,
    0x14,0x10,0x0F,0,0x15,0x11,0x10,0,0x16,0x12,0x11,0,0x17,0x13,0x12,0,
    0x18,0x14,0x13,0,0x19,0x15,0x14,0,0x1A,0x16,0x15,0,0x1B,0x17,0x16,0,
    0x1C,0x18,0x17,0,0x1D,0x19,0x18,0,0x1E,0x1A,0x19,0,0x1F,0x1B,0x1A,0,
    0x20,0x1C,0x1B,0,0x21,0x1D,0x1C,0,0x22,0x1E,0x1D,0,0x23,0x1F,0x1E,0,
    0xE5,0xE0,0xDF,0,0xE6,0xE1,0xE0,0,0xE7,0xE2,0xE1,0,0xE8,0xE3,0xE2,0,
    0xE9,0xE4,0xE3,0,0xEA,0xE5,0xE4,0,0xEB,0xE6,0xE5,0,0xEC,0xE7,0xE6,0,
    0xED,0xE8,0xE7,0,0xEE,0xE9,0xE8,0,0xEF,0xEA,0xE9,0,0xF0,0xEB,0xEA,0,
    0xF1,0xEC,0xEB,0,0xF2,0xED,0xEC,0,0xF3,0xEE,0xED,0,0xF4,0xEF,0xEE,0,
    0xF5,0xF0,0xEF,0,0xF6,0xF1,0xF0,0,0xF7,0xF2,0xF1,0,0xF8,0xF3,0xF2,0,
    0xF9,0xF4,0xF3,0,0xFA,0xF5,0xF4,0,0xFB,0xF6,0xF5,0,0xFC,0xF7,0xF6,0,
    0xFD,0xF8,0xF7,0,0xFE,0xF9,0xF8,0,0xFF,0xFA,0xF9,0,0x00,0xFB,0xFA,0,
    0x01,0xFC,0xFB,0,0x02,0xFD,0xFC,0,0x03,0xFE,0xFD,0,0x04,0xFF,0xFE,0,
    0x05,0x00,0xFF,0,0x06,0x01,0x00,0,0x07,0x02,0x01,0,0x08,0x03,0x02,0,
    0x09,0x04,0x03,0,0x0A,0x05,0x04,0,0x0B,0x06,0x05,0,0x0C,0x07,0x06,0,
    0x0D,0x08,0x07,0,0x0E,0x09,0x08,0,0x0F,0x0A,0x09,0,0x10,0x0B,0x0A,0,
    0x11,0x0C,0x0B,0,0x12,0x0D,0x0C,0,0x13,0x0E,0x0D,0,0x14,0x0F,0x0E,0,
    0x15,0x10,0x0F,0,0x16,0x11,0x10,0,0x17,0x12,0x11,0,0x18,0x13,0x12,0,
    0x19,0x14,0x13,0,0x1A,0x15,0x14,0,0x1B,0x16,0x15,0,0x1C,0x17,0x16,0,
    0x1D,0x18,0x17,0,0x1E,0x19,0x18,0,0x1F,0x1A,0x19,0,0x20,0x1B,0x1A,0,
    0x21,0x1C,0x1B,0,0x22,0x1D,0x1C,0,0x23,0x1E,0x1D,0,0x24,0x1F,0x1E,0,
    0xE6,0xE0,0xDF,0,0xE7,0xE1,0xE0,0,0xE8,0xE2,0xE1,0,0xE9,0xE3,0xE2,0,
    0xEA,0xE4,0xE3,0,0xEB,0xE5,0xE4,0,0xEC,0xE6,0xE5,0,0xED,0xE7,0xE6,0,
    0xEE,0xE8,0xE7,0,0xEF,0xE9,0xE8,0,0xF0,0xEA,0xE9,0,0xF1,0xEB,0xEA,0,
    0xF2,0xEC,0xEB,0,0xF3,0xED,0xEC,0,0xF4,0xEE,0xED,0,0xF5,0xEF,0xEE,0,
    0xF6,0xF0,0xEF,0,0xF7,0xF1,0xF0,0,0xF8,0xF2,0xF1,0,0xF9,0xF3,0xF2,0,
    0xFA,0xF4,0xF3,0,0xFB,0xF5,0xF4,0,0xFC,0xF6,0xF5,0,0xFD,0xF7,0xF6,0,
    0xFE,0xF8,0xF7,0,0xFF,0xF9,0xF8,0,0x00,0xFA,0xF9,0,0x01,0xFB,0xFA,0,
    0x02,0xFC,0xFB,0,0x03,0xFD,0xFC,0,0x04,0xFE,0xFD,0,0x05,0xFF,0xFE,0,
    0x06,0x00,0xFF,0,0x07,0x01,0x00,0,0x08,0x02,0x01,0,0x09,0x03,0x02,0,
    0x0A,0x04,0x03,0,0x0B,0x05,0x04,0,0x0C,0x06,0x05,0,0x0D,0x07,0x06,0,
    0x0E,0x08,0x07,0,0x0F,0x09,0x08,0,0x10,0x0A,0x09,0,0x11,0x0B,0x0A,0,
    0x12,0x0C,0x0B,0,0x13,0x0D,0x0C,0,0x14,0x0E,0x0D,0,0x15,0x0F,0x0E,0,
    0x16,0x10,0x0F,0,0x17,0x11,0x10,0,0x18,0x12,0x11,0,0x19,0x13,0x12,0,
    0x1A,0x14,0x13,0,0x1B,0x15,0x14,0,0x1C,0x16,0x15,0,0x1D,0x17,0x16,0,
    0x1E,0x18,0x17,0,0x1F,0x19,0x18,0,0x20,0x1A,0x19,0,0x21,0x1B,0x1A,0,
    0x22,0x1C,0x1B,0,0x23,0x1D,0x1C,0,0x24,0x1E,0x1D,0,0x25,0x1F,0x1E,0,
    0xE7,0xE0,0xDF,0,0xE8,0xE1,0xE0,0,0xE9,0xE2,0xE1,0,0xEA,0xE3,0xE2,0,
    0xEB,0xE4,0xE3,0,0xEC,0xE5,0xE4,0,0xED,0xE6,0xE5,0,0xEE,0xE7,0xE6,0,
    0xEF,0xE8,0xE7,0,0xF0,0xE9,0xE8,0,0xF1,0xEA,0xE9,0,0xF2,0xEB,0xEA,0,
    0xF3,0xEC,0xEB,0,0xF4,0xED,0xEC,0,0xF5,0xEE,0xED,0,0xF6,0xEF,0xEE,0,
    0xF7,0xF0,0xEF,0,0xF8,0xF1,0xF0,0,0xF9,0xF2,0xF1,0,0xFA,0xF3,0xF2,0,
    0xFB,0xF4,0xF3,0,0xFC,0xF5,0xF4,0,0xFD,0xF6,0xF5,0,0xFE,0xF7,0xF6,0,
    0xFF,0xF8,0xF7,0,0x00,0xF9,0xF8,0,0x01,0xFA,0xF9,0,0x02,0xFB,0xFA,0,
    0x03,0xFC,0xFB,0,0x04,0xFD,0xFC,0,0x05,0xFE,0xFD,0,0x06,0xFF,0xFE,0,
    0x07,0x00,0xFF,0,0x08,0x01,0x00,0,0x09,0x02,0x01,0,0x0A,0x03,0x02,0,
    0x0B,0x04,0x03,0,0x0C,0x05,0x04,0,0x0D,0x06,0x05,0,0x0E,0x07,0x06,0,
    0x0F,0x08,0x07,0,0x10,0x09,0x08,0,0x11,0x0A,0x09,0,0x12,0x0B,0x0A,0,
    0x13,0x0C,0x0B,0,0x14,0x0D,0x0C,0,0x15,0x0E,0x0D,0,0x16,0x0F,0x0E,0,
    0x17,0x10,0x0F,0,0x18,0x11,0x10,0,0x19,0x12,0x11,0,0x1A,0x13,0x12,0,
    0x1B,0x14,0x13,0,0x1C,0x15,0x14,0,0x1D,0x16,0x15,0,0x1E,0x17,0x16,0,
    0x1F,0x18,0x17,0,0x20,0x19,0x18,0,0x21,0x1A,0x19,0,0x22,0x1B,0x1A,0,
    0x23,0x1C,0x1B,0,0x24,0x1D,0x1C,0,0x25,0x1E,0x1D,0,0x26,0x1F,0x1E,0,
    0xD8,0xE0,0xE0,0,0xD9,0xE1,0xE1,0,0xDA,0xE2,0xE2,0,0xDB,0xE3,0xE3,0,
    0xDC,0xE4,0xE4,0,0xDD,0xE5,0xE5,0,0xDE,0xE6,0xE6,0,0xDF,0xE7,0xE7,0,
    0xE0,0xE8,0xE8,0,0xE1,0xE9,0xE9,0,0xE2,0xEA,0xEA,0,0xE3,0xEB,0xEB,0,
    0xE4,0xEC,0xEC,0,0xE5,0xED,0xED,0,0xE6,0xEE,0xEE,0,0xE7,0xEF,0xEF,0,
    0xE8,0xF0,0xF0,0,0xE9,0xF1,0xF1,0,0xEA,0xF2,0xF2,0,0xEB,0xF3,0xF3,0,
    0xEC,0xF4,0xF4,0,0xED,0xF5,0xF5,0,0xEE,0xF6,0xF6,0,0xEF,0xF7,0xF7,0,
    0xF0,0xF8,0xF8,0,0xF1,0xF9,0xF9,0,0xF2,0xFA,0xFA,0,0xF3,0xFB,0xFB,0,
    0xF4,0xFC,0xFC,0,0xF5,0xFD,0xFD,0,0xF6,0xFE,0xFE,0,0xF7,0xFF,0xFF,0,
    0xF8,0x00,0x00,0,0xF9,0x01,0x01,0,0xFA,0x02,0x02,0,0xFB,0x03,0x03,0,
    0xFC,0x04,0x04,0,0xFD,0x05,0x05,0,0xFE,0x06,0x06,0,0xFF,0x07,0x07,0,
    0x00,0x08,0x08,0,0x01,0x09,0x09,0,0x02,0x0A,0x0A,0,0x03,0x0B,0x0B,0,
    0x04,0x0C,0x0C,0,0x05,0x0D,0x0D,0,0x06,0x0E,0x0E,0,0x07,0x0F,0x0F,0,
    0x08,0x10,0x10,0,0x09,0x11,0x11,0,0x0A,0x12,0x12,0,0x0B,0x13,0x13,0,
    0x0C,0x14,0x14,0,0x0D,0x15,0x15,0,0x0E,0x16,0x16,0,0x0F,0x17,0x17,0,
    0x10,0x18,0x18,0,0x11,0x19,0x19,0,0x12,0x1A,0x1A,0,0x13,0x1B,0x1B,0,
    0x14,0x1C,0x1C,0,0x15,0x1D,0x1D,0,0x16,0x1E,0x1E,0,0x17,0x1F,0x1F,0,
    0xD9,0xE0,0xE0,0,0xDA,0xE1,0xE1,0,0xDB,0xE2,0xE2,0,0xDC,0xE3,0xE3,0,
    0xDD,0xE4,0xE4,0,0xDE,0xE5,0xE5,0,0xDF,0xE6,0xE6,0,0xE0,0xE7,0xE7,0,
    0xE1,0xE8,0xE8,0,0xE2,0xE9,0xE9,0,0xE3,0xEA,0xEA,0,0xE4,0xEB,0xEB,0,
    0xE5,0xEC,0xEC,0,0xE6,0xED,0xED,0,0xE7,0xEE,0xEE,0,0xE8,0xEF,0xEF,0,
    0xE9,0xF0,0xF0,0,0xEA,0xF1,0xF1,0,0xEB,0xF2,0xF2,0,0xEC,0xF3,0xF3,0,
    0xED,0xF4,0xF4,0,0xEE,0xF5,0xF5,0,0xEF,0xF6,0xF6,0,0xF0,0xF7,0xF7,0,
    0xF1,0xF8,0xF8,0,0xF2,0xF9,0xF9,0,0xF3,0xFA,0xFA,0,0xF4,0xFB,0xFB,0,
    0xF5,0xFC,0xFC,0,0xF6,0xFD,0xFD,0,0xF7,0xFE,0xFE,0,0xF8,0xFF,0xFF,0,
    0xF9,0x00,0x00,0,0xFA,0x01,0x01,0,0xFB,0x02,0x02,0,0xFC,0x03,0x03,0,
    0xFD,0x04,0x04,0,0xFE,0x05,0x05,0,0xFF,0x06,0x06,0,0x00,0x07,0x07,0,
    0x01,0x08,0x08,0,0x02,0x09,0x09,0,0x03,0x0A,0x0A,0,0x04,0x0B,0x0B,0,
    0x05,0x0C,0x0C,0,0x06,0x0D,0x0D,0,0x07,0x0E,0x0E,0,0x08,0x0F,0x0F,0,
    0x09,0x10,0x10,0,0x0A,0x11,0x11,0,0x0B,0x12,0x12,0,0x0C,0x13,0x13,0,
    0x0D,0x14,0x14,0,0x0E,0x15,0x15,0,0x0F,0x16,0x16,0,0x10,0x17,0x17,0,
    0x11,0x18,0x18,0,0x12,0x19,0x19,0,0x13,0x1A,0x1A,0,0x14,0x1B,0x1B,0,
    0x15,0x1C,0x1C,0,0x16,0x1D,0x1D,0,0x17,0x1E,0x1E,0,0x18,0x1F,0x1F,0,
    0xDA,0xE0,0xE0,0,0xDB,0xE1,0xE1,0,0xDC,0xE2,0xE2,0,0xDD,0xE3,0xE3,0,
    0xDE,0xE4,0xE4,0,0xDF,0xE5,0xE5,0,0xE0,0xE6,0xE6,0,0xE1,0xE7,0xE7,0,
    0xE2,0xE8,0xE8,0,0xE3,0xE9,0xE9,0,0xE4,0xEA,0xEA,0,0xE5,0xEB,0xEB,0,
    0xE6,0xEC,0xEC,0,0xE7,0xED,0xED,0,0xE8,0xEE,0xEE,0,0xE9,0xEF,0xEF,0,
    0xEA,0xF0,0xF0,0,0xEB,0xF1,0xF1,0,0xEC,0xF2,0xF2,0,0xED,0xF3,0xF3,0,
    0xEE,0xF4,0xF4,0,0xEF,0xF5,0xF5,0,0xF0,0xF6,0xF6,0,0xF1,0xF7,0xF7,0,
    0xF2,0xF8,0xF8,0,0xF3,0xF9,0xF9,0,0xF4,0xFA,0xFA,0,0xF5,0xFB,0xFB,0,
    0xF6,0xFC,0xFC,0,0xF7,0xFD,0xFD,0,0xF8,0xFE,0xFE,0,0xF9,0xFF,0xFF,0,
    0xFA,0x00,0x00,0,0xFB,0x01,0x01,0,0xFC,0x02,0x02,0,0xFD,0x03,0x03,0,
    0xFE,0x04,0x04,0,0xFF,0x05,0x05,0,0x00,0x06,0x06,0,0x01,0x07,0x07,0,
    0x02,0x08,0x08,0,0x03,0x09,0x09,0,0x04,0x0A,0x0A,0,0x05,0x0B,0x0B,0,
    0x06,0x0C,0x0C,0,0x07,0x0D,0x0D,0,0x08,0x0E,0x0E,0,0x09,0x0F,0x0F,0,
    0x0A,0x10,0x10,0,0x0B,0x11,0x11,0,0x0C,0x12,0x12,0,0x0D,0x13,0x13,0,
    0x0E,0x14,0x14,0,0x0F,0x15,0x15,0,0x10,0x16,0x16,0,0x11,0x17,0x17,0,
    0x12,0x18,0x18,0,0x13,0x19,0x19,0,0x14,0x1A,0x1A,0,0x15,0x1B,0x1B,0,
    0x16,0x1C,0x1C,0,0x17,0x1D,0x1D,0,0x18,0x1E,0x1E,0,0x19,0x1F,0x1F,0,
    0xDB,0xE0,0xE0,0,0xDC,0xE1,0xE1,0,0xDD,0xE2,0xE2,0,0xDE,0xE3,0xE3,0,
    0xDF,0xE4,0xE4,0,0xE0,0xE5,0xE5,0,0xE1,0xE6,0xE6,0,0xE2,0xE7,0xE7,0,
    0xE3,0xE8,0xE8,0,0xE4,0xE9,0xE9,0,0xE5,0xEA,0xEA,0,0xE6,0xEB,0xEB,0,
    0xE7,0xEC,0xEC,0,0xE8,0xED,0xED,0,0xE9,0xEE,0xEE,0,0xEA,0xEF,0xEF,0,
    0xEB,0xF0,0xF0,0,0xEC,0xF1,0xF1,0,0xED,0xF2,0xF2,0,0xEE,0xF3,0xF3,0,
    0xEF,0xF4,0xF4,0,0xF0,0xF5,0xF5,0,0xF1,0xF6,0xF6,0,0xF2,0xF7,0xF7,0,
    0xF3,0xF8,0xF8,0,0xF4,0xF9,0xF9,0,0xF5,0xFA,0xFA,0,0xF6,0xFB,0xFB,0,
    0xF7,0xFC,0xFC,0,0xF8,0xFD,0xFD,0,0xF9,0xFE,0xFE,0,0xFA,0xFF,0xFF,0,
    0xFB,0x00,0x00,0,0xFC,0x01,0x01,0,0xFD,0x02,0x02,0,0xFE,0x03,0x03,0,
    0xFF,0x04,0x04,0,0x00,0x05,0x05,0,0x01,0x06,0x06,0,0x02,0x07,0x07,0,
    0x03,0x08,0x08,0,0x04,0x09,0x09,0,0x05,0x0A,0x0A,0,0x06,0x0B,0x0B,0,
    0x07,0x0C,0x0C,0,0x08,0x0D,0x0D,0,0x09,0x0E,0x0E,0,0x0A,0x0F,0x0F,0,
    0x0B,0x10,0x10,0,0x0C,0x11,0x11,0,0x0D,0x12,0x12,0,0x0E,0x13,0x13,0,
    0x0F,0x14,0x14,0,0x10,0x15,0x15,0,0x11,0x16,0x16,0,0x12,0x17,0x17,0,
    0x13,0x18,0x18,0,0x14,0x19,0x19,0,0x15,0x1A,0x1A,0,0x16,0x1B,0x1B,0,
    0x17,0x1C,0x1C,0,0x18,0x1D,0x1D,0,0x19,0x1E,0x1E,0,0x1A,0x1F,0x1F,0,
    0xDC,0xE0,0xE0,0,0xDD,0xE1,0xE1,0,0xDE,0xE2,0xE2,0,0xDF,0xE3,0xE3,0,
    0xE0,0xE4,0xE4,0,0xE1,0xE5,0xE5,0,0xE2,0xE6,0xE6,0,0xE3,0xE7,0xE7,0,
    0xE4,0xE8,0xE8,0,0xE5,0xE9,0xE9,0,0xE6,0xEA,0xEA,0,0xE7,0xEB,0xEB,0,
    0xE8,0xEC,0xEC,0,0xE9,0xED,0xED,0,0xEA,0xEE,0xEE,0,0xEB,0xEF,0xEF,0,
    0xEC,0xF0,0xF0,0,0xED,0xF1,0xF1,0,0xEE,0xF2,0xF2,0,0xEF,0xF3,0xF3,0,
    0xF0,0xF4,0xF4,0,0xF1,0xF5,0xF5,0,0xF2,0xF6,0xF6,0,0xF3,0xF7,0xF7,0,
    0xF4,0xF8,0xF8,0,0xF5,0xF9,0xF9,0,0xF6,0xFA,0xFA,0,0xF7,0xFB,0xFB,0,
    0xF8,0xFC,0xFC,0,0xF9,0xFD,0xFD,0,0xFA,0xFE,0xFE,0,0xFB,0xFF,0xFF,0,
    0xFC,0x00,0x00,0,0xFD,0x01,0x01,0,0xFE,0x02,0x02,0,0xFF,0x03,0x03,0,
    0x00,0x04,0x04,0,0x01,0x05,0x05,0,0x02,0x06,0x06,0,0x03,0x07,0x07,0,
    0x04,0x08,0x08,0,0x05,0x09,0x09,0,0x06,0x0A,0x0A,0,0x07,0x0B,0x0B,0,
    0x08,0x0C,0x0C,0,0x09,0x0D,0x0D,0,0x0A,0x0E,0x0E,0,0x0B,0x0F,0x0F,0,
    0x0C,0x10,0x10,0,0x0D,0x11,0x11,0,0x0E,0x12,0x12,0,0x0F,0x13,0x13,0,
    0x10,0x14,0x14,0,0x11,0x15,0x15,0,0x12,0x16,0x16,0,0x13,0x17,0x17,0,
    0x14,0x18,0x18,0,0x15,0x19,0x19,0,0x16,0x1A,0x1A,0,0x17,0x1B,0x1B,0,
    0x18,0x1C,0x1C,0,0x19,0x1D,0x1D,0,0x1A,0x1E,0x1E,0,0x1B,0x1F,0x1F,0,
    0xDD,0xE0,0xE0,0,0xDE,0xE1,0xE1,0,0xDF,0xE2,0xE2,0,0xE0,0xE3,0xE3,0,
    0xE1,0xE4,0xE4,0,0xE2,0xE5,0xE5,0,0xE3,0xE6,0xE6,0,0xE4,0xE7,0xE7,0,
    0xE5,0xE8,0xE8,0,0xE6,0xE9,0xE9,0,0xE7,0xEA,0xEA,0,0xE8,0xEB,0xEB,0,
    0xE9,0xEC,0xEC,0,0xEA,0xED,0xED,0,0xEB,0xEE,0xEE,0,0xEC,0xEF,0xEF,0,
    0xED,0xF0,0xF0,0,0xEE,0xF1,0xF1,0,0xEF,0xF2,0xF2,0,0xF0,0xF3,0xF3,0,
    0xF1,0xF4,0xF4,0,0xF2,0xF5,0xF5,0,0xF3,0xF6,0xF6,0,0xF4,0xF7,0xF7,0,
    0xF5,0xF8,0xF8,0,0xF6,0xF9,0xF9,0,0xF7,0xFA,0xFA,0,0xF8,0xFB,0xFB,0,
    0xF9,0xFC,0xFC,0,0xFA,0xFD,0xFD,0,0xFB,0xFE,0xFE,0,0xFC,0xFF,0xFF,0,
    0xFD,0x00,0x00,0,0xFE,0x01,0x01,0,0xFF,0x02,0x02,0,0x00,0x03,0x03,0,
    0x01,0x04,0x04,0,0x02,0x05,0x05,0,0x03,0x06,0x06,0,0x04,0x07,0x07,0,
    0x05,0x08,0x08,0,0x06,0x09,0x09,0,0x07,0x0A,0x0A,0,0x08,0x0B,0x0B,0,
    0x09,0x0C,0x0C,0,0x0A,0x0D,0x0D,0,0x0B,0x0E,0x0E,0,0x0C,0x0F,0x0F,0,
    0x0D,0x10,0x10,0,0x0E,0x11,0x11,0,0x0F,0x12,0x12,0,0x10,0x13,0x13,0,
    0x11,0x14,0x14,0,0x12,0x15,0x15,0,0x13,0x16,0x16,0,0x14,0x17,0x17,0,
    0x15,0x18,0x18,0,0x16,0x19,0x19,0,0x17,0x1A,0x1A,0,0x18,0x1B,0x1B,0,
    0x19,0x1C,0x1C,0,0x1A,0x1D,0x1D,0,0x1B,0x1E,0x1E,0,0x1C,0x1F,0x1F,0,
    0xDE,0xE0,0xE0,0,0xDF,0xE1,0xE1,0,0xE0,0xE2,0xE2,0,0xE1,0xE3,0xE3,0,
    0xE2,0xE4,0xE4,0,0xE3,0xE5,0xE5,0,0xE4,0xE6,0xE6,0,0xE5,0xE7,0xE7,0,
    0xE6,0xE8,0xE8,0,0xE7,0xE9,0xE9,0,0xE8,0xEA,0xEA,0,0xE9,0xEB,0xEB,0,
    0xEA,0xEC,0xEC,0,0xEB,0xED,0xED,0,0xEC,0xEE,0xEE,0,0xED,0xEF,0xEF,0,
    0xEE,0xF0,0xF0,0,0xEF,0xF1,0xF1,0,0xF0,0xF2,0xF2,0,0xF1,0xF3,0xF3,0,
    0xF2,0xF4,0xF4,0,0xF3,0xF5,0xF5,0,0xF4,0xF6,0xF6,0,0xF5,0xF7,0xF7,0,
    0xF6,0xF8,0xF8,0,0xF7,0xF9,0xF9,0,0xF8,0xFA,0xFA,0,0xF9,0xFB,0xFB,0,
    0xFA,0xFC,0xFC,0,0xFB,0xFD,0xFD,0,0xFC,0xFE,0xFE,0,0xFD,0xFF,0xFF,0,
    0xFE,0x00,0x00,0,0xFF,0x01,0x01,0,0x00,0x02,0x02,0,0x01,0x03,0x03,0,
    0x02,0x04,0x04,0,0x03,0x05,0x05,0,0x04,0x06,0x06,0,0x05,0x07,0x07,0,
    0x06,0x08,0x08,0,0x07,0x09,0x09,0,0x08,0x0A,0x0A,0,0x09,0x0B,0x0B,0,
    0x0A,0x0C,0x0C,0,0x0B,0x0D,0x0D,0,0x0C,0x0E,0x0E,0,0x0D,0x0F,0x0F,0,
    0x0E,0x10,0x10,0,0x0F,0x11,0x11,0,0x10,0x12,0x12,0,0x11,0x13,0x13,0,
    0x12,0x14,0x14,0,0x13,0x15,0x15,0,0x14,0x16,0x16,0,0x15,0x17,0x17,0,
    0x16,0x18,0x18,0,0x17,0x19,0x19,0,0x18,0x1A,0x1A,0,0x19,0x1B,0x1B,0,
    0x1A,0x1C,0x1C,0,0x1B,0x1D,0x1D,0,0x1C,0x1E,0x1E,0,0x1D,0x1F,0x1F,0,
    0xDF,0xE0,0xE0,0,0xE0,0xE1,0xE1,0,0xE1,0xE2,0xE2,0,0xE2,0xE3,0xE3,0,
    0xE3,0xE4,0xE4,0,0xE4,0xE5,0xE5,0,0xE5,0xE6,0xE6,0,0xE6,0xE7,0xE7,0,
    0xE7,0xE8,0xE8,0,0xE8,0xE9,0xE9,0,0xE9,0xEA,0xEA,0,0xEA,0xEB,0xEB,0,
    0xEB,0xEC,0xEC,0,0xEC,0xED,0xED,0,0xED,0xEE,0xEE,0,0xEE,0xEF,0xEF,0,
    0xEF,0xF0,0xF0,0,0xF0,0xF1,0xF1,0,0xF1,0xF2,0xF2,0,0xF2,0xF3,0xF3,0,
    0xF3,0xF4,0xF4,0,0xF4,0xF5,0xF5,0,0xF5,0xF6,0xF6,0,0xF6,0xF7,0xF7,0,
    0xF7,0xF8,0xF8,0,0xF8,0xF9,0xF9,0,0xF9,0xFA,0xFA,0,0xFA,0xFB,0xFB,0,
    0xFB,0xFC,0xFC,0,0xFC,0xFD,0xFD,0,0xFD,0xFE,0xFE,0,0xFE,0xFF,0xFF,0,
    0xFF,0x00,0x00,0,0x00,0x01,0x01,0,0x01,0x02,0x02,0,0x02,0x03,0x03,0,
    0x03,0x04,0x04,0,0x04,0x05,0x05,0,0x05,0x06,0x06,0,0x06,0x07,0x07,0,
    0x07,0x08,0x08,0,0x08,0x09,0x09,0,0x09,0x0A,0x0A,0,0x0A,0x0B,0x0B,0,
    0x0B,0x0C,0x0C,0,0x0C,0x0D,0x0D,0,0x0D,0x0E,0x0E,0,0x0E,0x0F,0x0F,0,
    0x0F,0x10,0x10,0,0x10,0x11,0x11,0,0x11,0x12,0x12,0,0x12,0x13,0x13,0,
    0x13,0x14,0x14,0,0x14,0x15,0x15,0,0x15,0x16,0x16,0,0x16,0x17,0x17,0,
    0x17,0x18,0x18,0,0x18,0x19,0x19,0,0x19,0x1A,0x1A,0,0x1A,0x1B,0x1B,0,
    0x1B,0x1C,0x1C,0,0x1C,0x1D,0x1D,0,0x1D,0x1E,0x1E,0,0x1E,0x1F,0x1F,0,
    0xE0,0xE0,0xE0,0,0xE1,0xE1,0xE1,0,0xE2,0xE2,0xE2,0,0xE3,0xE3,0xE3,0,
    0xE4,0xE4,0xE4,0,0xE5,0xE5,0xE5,0,0xE6,0xE6,0xE6,0,0xE7,0xE7,0xE7,0,
    0xE8,0xE8,0xE8,0,0xE9,0xE9,0xE9,0,0xEA,0xEA,0xEA,0,0xEB,0xEB,0xEB,0,
    0xEC,0xEC,0xEC,0,0xED,0xED,0xED,0,0xEE,0xEE,0xEE,0,0xEF,0xEF,0xEF,0,
    0xF0,0xF0,0xF0,0,0xF1,0xF1,0xF1,0,0xF2,0xF2,0xF2,0,0xF3,0xF3,0xF3,0,
    0xF4,0xF4,0xF4,0,0xF5,0xF5,0xF5,0,0xF6,0xF6,0xF6,0,0xF7,0xF7,0xF7,0,
    0xF8,0xF8,0xF8,0,0xF9,0xF9,0xF9,0,0xFA,0xFA,0xFA,0,0xFB,0xFB,0xFB,0,
    0xFC,0xFC,0xFC,0,0xFD,0xFD,0xFD,0,0xFE,0xFE,0xFE,0,0xFF,0xFF,0xFF,0,
    0x00,0x00,0x00,0,0x01,0x01,0x01,0,0x02,0x02,0x02,0,0x03,0x03,0x03,0,
    0x04,0x04,0x04,0,0x05,0x05,0x05,0,0x06,0x06,0x06,0,0x07,0x07,0x07,0,
    0x08,0x08,0x08,0,0x09,0x09,0x09,0,0x0A,0x0A,0x0A,0,0x0B,0x0B,0x0B,0,
    0x0C,0x0C,0x0C,0,0x0D,0x0D,0x0D,0,0x0E,0x0E,0x0E,0,0x0F,0x0F,0x0F,0,
    0x10,0x10,0x10,0,0x11,0x11,0x11,0,0x12,0x12,0x12,0,0x13,0x13,0x13,0,
    0x14,0x14,0x14,0,0x15,0x15,0x15,0,0x16,0x16,0x16,0,0x17,0x17,0x17,0,
    0x18,0x18,0x18,0,0x19,0x19,0x19,0,0x1A,0x1A,0x1A,0,0x1B,0x1B,0x1B,0,
    0x1C,0x1C,0x1C,0,0x1D,0x1D,0x1D,0,0x1E,0x1E,0x1E,0,0x1F,0x1F,0x1F,0,
    0xE1,0xE0,0xE0,0,0xE2,0xE1,0xE1,0,0xE3,0xE2,0xE2,0,0xE4,0xE3,0xE3,0,
    0xE5,0xE4,0xE4,0,0xE6,0xE5,0xE5,0,0xE7,0xE6,0xE6,0,0xE8,0xE7,0xE7,0,
    0xE9,0xE8,0xE8,0,0xEA,0xE9,0xE9,0,0xEB,0xEA,0xEA,0,0xEC,0xEB,0xEB,0,
    0xED,0xEC,0xEC,0,0xEE,0xED,0xED,0,0xEF,0xEE,0xEE,0,0xF0,0xEF,0xEF,0,
    0xF1,0xF0,0xF0,0,0xF2,0xF1,0xF1,0,0xF3,0xF2,0xF2,0,0xF4,0xF3,0xF3,0,
    0xF5,0xF4,0xF4,0,0xF6,0xF5,0xF5,0,0xF7,0xF6,0xF6,0,0xF8,0xF7,0xF7,0,
    0xF9,0xF8,0xF8,0,0xFA,0xF9,0xF9,0,0xFB,0xFA,0xFA,0,0xFC,0xFB,0xFB,0,
    0xFD,0xFC,0xFC,0,0xFE,0xFD,0xFD,0,0xFF,0xFE,0xFE,0,0x00,0xFF,0xFF,0,
    0x01,0x00,0x00,0,0x02,0x01,0x01,0,0x03,0x02,0x02,0,0x04,0x03,0x03,0,
    0x05,0x04,0x04,0,0x06,0x05,0x05,0,0x07,0x06,0x06,0,0x08,0x07,0x07,0,
    0x09,0x08,0x08,0,0x0A,0x09,0x09,0,0x0B,0x0A,0x0A,0,0x0C,0x0B,0x0B,0,
    0x0D,0x0C,0x0C,0,0x0E,0x0D,0x0D,0,0x0F,0x0E,0x0E,0,0x10,0x0F,0x0F,0,
    0x11,0x10,0x10,0,0x12,0x11,0x11,0,0x13,0x12,0x12,0,0x14,0x13,0x13,0,
    0x15,0x14,0x14,0,0x16,0x15,0x15,0,0x17,0x16,0x16,0,0x18,0x17,0x17,0,
    0x19,0x18,0x18,0,0x1A,0x19,0x19,0,0x1B,0x1A,0x1A,0,0x1C,0x1B,0x1B,0,
    0x1D,0x1C,0x1C,0,0x1E,0x1D,0x1D,0,0x1F,0x1E,0x1E,0,0x20,0x1F,0x1F,0,
    0xE2,0xE0,0xE0,0,0xE3,0xE1,0xE1,0,0xE4,0xE2,0xE2,0,0xE5,0xE3,0xE3,0,
    0xE6,0xE4,0xE4,0,0xE7,0xE5,0xE5,0,0xE8,0xE6,0xE6,0,0xE9,0xE7,0xE7,0,
    0xEA,0xE8,0xE8,0,0xEB,0xE9,0xE9,0,0xEC,0xEA,0xEA,0,0xED,0xEB,0xEB,0,
    0xEE,0xEC,0xEC,0,0xEF,0xED,0xED,0,0xF0,0xEE,0xEE,0,0xF1,0xEF,0xEF,0,
    0xF2,0xF0,0xF0,0,0xF3,0xF1,0xF1,0,0xF4,0xF2,0xF2,0,0xF5,0xF3,0xF3,0,
    0xF6,0xF4,0xF4,0,0xF7,0xF5,0xF5,0,0xF8,0xF6,0xF6,0,0xF9,0xF7,0xF7,0,
    0xFA,0xF8,0xF8,0,0xFB,0xF9,0xF9,0,0xFC,0xFA,0xFA,0,0xFD,0xFB,0xFB,0,
    0xFE,0xFC,0xFC,0,0xFF,0xFD,0xFD,0,0x00,0xFE,0xFE,0,0x01,0xFF,0xFF,0,
    0x02,0x00,0x00,0,0x03,0x01,0x01,0,0x04,0x02,0x02,0,0x05,0x03,0x03,0,
    0x06,0x04,0x04,0,0x07,0x05,0x05,0,0x08,0x06,0x06,0,0x09,0x07,0x07,0,
    0x0A,0x08,0x08,0,0x0B,0x09,0x09,0,0x0C,0x0A,0x0A,0,0x0D,0x0B,0x0B,0,
    0x0E,0x0C,0x0C,0,0x0F,0x0D,0x0D,0,0x10,0x0E,0x0E,0,0x11,0x0F,0x0F,0,
    0x12,0x10,0x10,0,0x13,0x11,0x11,0,0x14,0x12,0x12,0,0x15,0x13,0x13,0,
    0x16,0x14,0x14,0,0x17,0x15,0x15,0,0x18,0x16,0x16,0,0x19,0x17,0x17,0,
    0x1A,0x18,0x18,0,0x1B,0x19,0x19,0,0x1C,0x1A,0x1A,0,0x1D,0x1B,0x1B,0,
    0x1E,0x1C,0x1C,0,0x1F,0x1D,0x1D,0,0x20,0x1E,0x1E,0,0x21,0x1F,0x1F,0,
    0xE3,0xE0,0xE0,0,0xE4,0xE1,0xE1,0,0xE5,0xE2,0xE2,0,0xE6,0xE3,0xE3,0,
    0xE7,0xE4,0xE4,0,0xE8,0xE5,0xE5,0,0xE9,0xE6,0xE6,0,0xEA,0xE7,0xE7,0,
    0xEB,0xE8,0xE8,0,0xEC,0xE9,0xE9,0,0xED,0xEA,0xEA,0,0xEE,0xEB,0xEB,0,
    0xEF,0xEC,0xEC,0,0xF0,0xED,0xED,0,0xF1,0xEE,0xEE,0,0xF2,0xEF,0xEF,0,
    0xF3,0xF0,0xF0,0,0xF4,0xF1,0xF1,0,0xF5,0xF2,0xF2,0,0xF6,0xF3,0xF3,0,
    0xF7,0xF4,0xF4,0,0xF8,0xF5,0xF5,0,0xF9,0xF6,0xF6,0,0xFA,0xF7,0xF7,0,
    0xFB,0xF8,0xF8,0,0xFC,0xF9,0xF9,0,0xFD,0xFA,0xFA,0,0xFE,0xFB,0xFB,0,
    0xFF,0xFC,0xFC,0,0x00,0xFD,0xFD,0,0x01,0xFE,0xFE,0,0x02,0xFF,0xFF,0,
    0x03,0x00,0x00,0,0x04,0x01,0x01,0,0x05,0x02,0x02,0,0x06,0x03,0x03,0,
    0x07,0x04,0x04,0,0x08,0x05,0x05,0,0x09,0x06,0x06,0,0x0A,0x07,0x07,0,
    0x0B,0x08,0x08,0,0x0C,0x09,0x09,0,0x0D,0x0A,0x0A,0,0x0E,0x0B,0x0B,0,
    0x0F,0x0C,0x0C,0,0x10,0x0D,0x0D,0,0x11,0x0E,0x0E,0,0x12,0x0F,0x0F,0,
    0x13,0x10,0x10,0,0x14,0x11,0x11,0,0x15,0x12,0x12,0,0x16,0x13,0x13,0,
    0x17,0x14,0x14,0,0x18,0x15,0x15,0,0x19,0x16,0x16,0,0x1A,0x17,0x17,0,
    0x1B,0x18,0x18,0,0x1C,0x19,0x19,0,0x1D,0x1A,0x1A,0,0x1E,0x1B,0x1B,0,
    0x1F,0x1C,0x1C,0,0x20,0x1D,0x1D,0,0x21,0x1E,0x1E,0,0x22,0x1F,0x1F,0,
    0xE4,0xE0,0xE0,0,0xE5,0xE1,0xE1,0,0xE6,0xE2,0xE2,0,0xE7,0xE3,0xE3,0,
    0xE8,0xE4,0xE4,0,0xE9,0xE5,0xE5,0,0xEA,0xE6,0xE6,0,0xEB,0xE7,0xE7,0,
    0xEC,0xE8,0xE8,0,0xED,0xE9,0xE9,0,0xEE,0xEA,0xEA,0,0xEF,0xEB,0xEB,0,
    0xF0,0xEC,0xEC,0,0xF1,0xED,0xED,0,0xF2,0xEE,0xEE,0,0xF3,0xEF,0xEF,0,
    0xF4,0xF0,0xF0,0,0xF5,0xF1,0xF1,0,0xF6,0xF2,0xF2,0,0xF7,0xF3,0xF3,0,
    0xF8,0xF4,0xF4,0,0xF9,0xF5,0xF5,0,0xFA,0xF6,0xF6,0,0xFB,0xF7,0xF7,0,
    0xFC,0xF8,0xF8,0,0xFD,0xF9,0xF9,0,0xFE,0xFA,0xFA,0,0xFF,0xFB,0xFB,0,
    0x00,0xFC,0xFC,0,0x01,0xFD,0xFD,0,0x02,0xFE,0xFE,0,0x03,0xFF,0xFF,0,
    0x04,0x00,0x00,0,0x05,0x01,0x01,0,0x06,0x02,0x02,0,0x07,0x03,0x03,0,
    0x08,0x04,0x04,0,0x09,0x05,0x05,0,0x0A,0x06,0x06,0,0x0B,0x07,0x07,0,
    0x0C,0x08,0x08,0,0x0D,0x09,0x09,0,0x0E,0x0A,0x0A,0,0x0F,0x0B,0x0B,0,
    0x10,0x0C,0x0C,0,0x11,0x0D,0x0D,0,0x12,0x0E,0x0E,0,0x13,0x0F,0x0F,0,
    0x14,0x10,0x10,0,0x15,0x11,0x11,0,0x16,0x12,0x12,0,0x17,0x13,0x13,0,
    0x18,0x14,0x14,0,0x19,0x15,0x15,0,0x1A,0x16,0x16,0,0x1B,0x17,0x17,0,
    0x1C,0x18,0x18,0,0x1D,0x19,0x19,0,0x1E,0x1A,0x1A,0,0x1F,0x1B,0x1B,0,
    0x20,0x1C,0x1C,0,0x21,0x1D,0x1D,0,0x22,0x1E,0x1E,0,0x23,0x1F,0x1F,0,
    0xE5,0xE0,0xE0,0,0xE6,0xE1,0xE1,0,0xE7,0xE2,0xE2,0,0xE8,0xE3,0xE3,0,
    0xE9,0xE4,0xE4,0,0xEA,0xE5,0xE5,0,0xEB,0xE6,0xE6,0,0xEC,0xE7,0xE7,0,
    0xED,0xE8,0xE8,0,0xEE,0xE9,0xE9,0,0xEF,0xEA,0xEA,0,0xF0,0xEB,0xEB,0,
    0xF1,0xEC,0xEC,0,0xF2,0xED,0xED,0,0xF3,0xEE,0xEE,0,0xF4,0xEF,0xEF,0,
    0xF5,0xF0,0xF0,0,0xF6,0xF1,0xF1,0,0xF7,0xF2,0xF2,0,0xF8,0xF3,0xF3,0,
    0xF9,0xF4,0xF4,0,0xFA,0xF5,0xF5,0,0xFB,0xF6,0xF6,0,0xFC,0xF7,0xF7,0,
    0xFD,0xF8,0xF8,0,0xFE,0xF9,0xF9,0,0xFF,0xFA,0xFA,0,0x00,0xFB,0xFB,0,
    0x01,0xFC,0xFC,0,0x02,0xFD,0xFD,0,0x03,0xFE,0xFE,0,0x04,0xFF,0xFF,0,
    0x05,0x00,0x00,0,0x06,0x01,0x01,0,0x07,0x02,0x02,0,0x08,0x03,0x03,0,
    0x09,0x04,0x04,0,0x0A,0x05,0x05,0,0x0B,0x06,0x06,0,0x0C,0x07,0x07,0,
    0x0D,0x08,0x08,0,0x0E,0x09,0x09,0,0x0F,0x0A,0x0A,0,0x10,0x0B,0x0B,0,
    0x11,0x0C,0x0C,0,0x12,0x0D,0x0D,0,0x13,0x0E,0x0E,0,0x14,0x0F,0x0F,0,
    0x15,0x10,0x10,0,0x16,0x11,0x11,0,0x17,0x12,0x12,0,0x18,0x13,0x13,0,
    0x19,0x14,0x14,0,0x1A,0x15,0x15,0,0x1B,0x16,0x16,0,0x1C,0x17,0x17,0,
    0x1D,0x18,0x18,0,0x1E,0x19,0x19,0,0x1F,0x1A,0x1A,0,0x20,0x1B,0x1B,0,
    0x21,0x1C,0x1C,0,0x22,0x1D,0x1D,0,0x23,0x1E,0x1E,0,0x24,0x1F,0x1F,0,
    0xE6,0xE0,0xE0,0,0xE7,0xE1,0xE1,0,0xE8,0xE2,0xE2,0,0xE9,0xE3,0xE3,0,
    0xEA,0xE4,0xE4,0,0xEB,0xE5,0xE5,0,0xEC,0xE6,0xE6,0,0xED,0xE7,0xE7,0,
    0xEE,0xE8,0xE8,0,0xEF,0xE9,0xE9,0,0xF0,0xEA,0xEA,0,0xF1,0xEB,0xEB,0,
    0xF2,0xEC,0xEC,0,0xF3,0xED,0xED,0,0xF4,0xEE,0xEE,0,0xF5,0xEF,0xEF,0,
    0xF6,0xF0,0xF0,0,0xF7,0xF1,0xF1,0,0xF8,0xF2,0xF2,0,0xF9,0xF3,0xF3,0,
    0xFA,0xF4,0xF4,0,0xFB,0xF5,0xF5,0,0xFC,0xF6,0xF6,0,0xFD,0xF7,0xF7,0,
    0xFE,0xF8,0xF8,0,0xFF,0xF9,0xF9,0,0x00,0xFA,0xFA,0,0x01,0xFB,0xFB,0,
    0x02,0xFC,0xFC,0,0x03,0xFD,0xFD,0,0x04,0xFE,0xFE,0,0x05,0xFF,0xFF,0,
    0x06,0x00,0x00,0,0x07,0x01,0x01,0,0x08,0x02,0x02,0,0x09,0x03,0x03,0,
    0x0A,0x04,0x04,0,0x0B,0x05,0x05,0,0x0C,0x06,0x06,0,0x0D,0x07,0x07,0,
    0x0E,0x08,0x08,0,0x0F,0x09,0x09,0,0x10,0x0A,0x0A,0,0x11,0x0B,0x0B,0,
    0x12,0x0C,0x0C,0,0x13,0x0D,0x0D,0,0x14,0x0E,0x0E,0,0x15,0x0F,0x0F,0,
    0x16,0x10,0x10,0,0x17,0x11,0x11,0,0x18,0x12,0x12,0,0x19,0x13,0x13,0,
    0x1A,0x14,0x14,0,0x1B,0x15,0x15,0,0x1C,0x16,0x16,0,0x1D,0x17,0x17,0,
    0x1E,0x18,0x18,0,0x1F,0x19,0x19,0,0x20,0x1A,0x1A,0,0x21,0x1B,0x1B,0,
    0x22,0x1C,0x1C,0,0x23,0x1D,0x1D,0,0x24,0x1E,0x1E,0,0x25,0x1F,0x1F,0,
    0xE7,0xE0,0xE0,0,0xE8,0xE1,0xE1,0,0xE9,0xE2,0xE2,0,0xEA,0xE3,0xE3,0,
    0xEB,0xE4,0xE4,0,0xEC,0xE5,0xE5,0,0xED,0xE6,0xE6,0,0xEE,0xE7,0xE7,0,
    0xEF,0xE8,0xE8,0,0xF0,0xE9,0xE9,0,0xF1,0xEA,0xEA,0,0xF2,0xEB,0xEB,0,
    0xF3,0xEC,0xEC,0,0xF4,0xED,0xED,0,0xF5,0xEE,0xEE,0,0xF6,0xEF,0xEF,0,
    0xF7,0xF0,0xF0,0,0xF8,0xF1,0xF1,0,0xF9,0xF2,0xF2,0,0xFA,0xF3,0xF3,0,
    0xFB,0xF4,0xF4,0,0xFC,0xF5,0xF5,0,0xFD,0xF6,0xF6,0,0xFE,0xF7,0xF7,0,
    0xFF,0xF8,0xF8,0,0x00,0xF9,0xF9,0,0x01,0xFA,0xFA,0,0x02,0xFB,0xFB,0,
    0x03,0xFC,0xFC,0,0x04,0xFD,0xFD,0,0x05,0xFE,0xFE,0,0x06,0xFF,0xFF,0,
    0x07,0x00,0x00,0,0x08,0x01,0x01,0,0x09,0x02,0x02,0,0x0A,0x03,0x03,0,
    0x0B,0x04,0x04,0,0x0C,0x05,0x05,0,0x0D,0x06,0x06,0,0x0E,0x07,0x07,0,
    0x0F,0x08,0x08,0,0x10,0x09,0x09,0,0x11,0x0A,0x0A,0,0x12,0x0B,0x0B,0,
    0x13,0x0C,0x0C,0,0x14,0x0D,0x0D,0,0x15,0x0E,0x0E,0,0x16,0x0F,0x0F,0,
    0x17,0x10,0x10,0,0x18,0x11,0x11,0,0x19,0x12,0x12,0,0x1A,0x13,0x13,0,
    0x1B,0x14,0x14,0,0x1C,0x15,0x15,0,0x1D,0x16,0x16,0,0x1E,0x17,0x17,0,
    0x1F,0x18,0x18,0,0x20,0x19,0x19,0,0x21,0x1A,0x1A,0,0x22,0x1B,0x1B,0,
    0x23,0x1C,0x1C,0,0x24,0x1D,0x1D,0,0x25,0x1E,0x1E,0,0x26,0x1F,0x1F,0,
    0xD8,0xE0,0xE1,0,0xD9,0xE1,0xE2,0,0xDA,0xE2,0xE3,0,0xDB,0xE3,0xE4,0,
    0xDC,0xE4,0xE5,0,0xDD,0xE5,0xE6,0,0xDE,0xE6,0xE7,0,0xDF,0xE7,0xE8,0,
    0xE0,0xE8,0xE9,0,0xE1,0xE9,0xEA,0,0xE2,0xEA,0xEB,0,0xE3,0xEB,0xEC,0,
    0xE4,0xEC,0xED,0,0xE5,0xED,0xEE,0,0xE6,0xEE,0xEF,0,0xE7,0xEF,0xF0,0,
    0xE8,0xF0,0xF1,0,0xE9,0xF1,0xF2,0,0xEA,0xF2,0xF3,0,0xEB,0xF3,0xF4,0,
    0xEC,0xF4,0xF5,0,0xED,0xF5,0xF6,0,0xEE,0xF6,0xF7,0,0xEF,0xF7,0xF8,0,
    0xF0,0xF8,0xF9,0,0xF1,0xF9,0xFA,0,0xF2,0xFA,0xFB,0,0xF3,0xFB,0xFC,0,
    0xF4,0xFC,0xFD,0,0xF5,0xFD,0xFE,0,0xF6,0xFE,0xFF,0,0xF7,0xFF,0x00,0,
    0xF8,0x00,0x01,0,0xF9,0x01,0x02,0,0xFA,0x02,0x03,0,0xFB,0x03,0x04,0,
    0xFC,0x04,0x05,0,0xFD,0x05,0x06,0,0xFE,0x06,0x07,0,0xFF,0x07,0x08,0,
    0x00,0x08,0x09,0,0x01,0x09,0x0A,0,0x02,0x0A,0x0B,0,0x03,0x0B,0x0C,0,
    0x04,0x0C,0x0D,0,0x05,0x0D,0x0E,0,0x06,0x0E,0x0F,0,0x07,0x0F,0x10,0,
    0x08,0x10,0x11,0,0x09,0x11,0x12,0,0x0A,0x12,0x13,0,0x0B,0x13,0x14,0,
    0x0C,0x14,0x15,0,0x0D,0x15,0x16,0,0x0E,0x16,0x17,0,0x0F,0x17,0x18,0,
    0x10,0x18,0x19,0,0x11,0x19,0x1A,0,0x12,0x1A,0x1B,0,0x13,0x1B,0x1C,0,
    0x14,0x1C,0x1D,0,0x15,0x1D,0x1E,0,0x16,0x1E,0x1F,0,0x17,0x1F,0x20,0,
    0xD9,0xE0,0xE1,0,0xDA,0xE1,0xE2,0,0xDB,0xE2,0xE3,0,0xDC,0xE3,0xE4,0,
    0xDD,0xE4,0xE5,0,0xDE,0xE5,0xE6,0,0xDF,0xE6,0xE7,0,0xE0,0xE7,0xE8,0,
    0xE1,0xE8,0xE9,0,0xE2,0xE9,0xEA,0,0xE3,0xEA,0xEB,0,0xE4,0xEB,0xEC,0,
    0xE5,0xEC,0xED,0,0xE6,0xED,0xEE,0,0xE7,0xEE,0xEF,0,0xE8,0xEF,0xF0,0,
    0xE9,0xF0,0xF1,0,0xEA,0xF1,0xF2,0,0xEB,0xF2,0xF3,0,0xEC,0xF3,0xF4,0,
    0xED,0xF4,0xF5,0,0xEE,0xF5,0xF6,0,0xEF,0xF6,0xF7,0,0xF0,0xF7,0xF8,0,
    0xF1,0xF8,0xF9,0,0xF2,0xF9,0xFA,0,0xF3,0xFA,0xFB,0,0xF4,0xFB,0xFC,0,
    0xF5,0xFC,0xFD,0,0xF6,0xFD,0xFE,0,0xF7,0xFE,0xFF,0,0xF8,0xFF,0x00,0,
    0xF9,0x00,0x01,0,0xFA,0x01,0x02,0,0xFB,0x02,0x03,0,0xFC,0x03,0x04,0,
    0xFD,0x04,0x05,0,0xFE,0x05,0x06,0,0xFF,0x06,0x07,0,0x00,0x07,0x08,0,
    0x01,0x08,0x09,0,0x02,0x09,0x0A,0,0x03,0x0A,0x0B,0,0x04,0x0B,0x0C,0,
    0x05,0x0C,0x0D,0,0x06,0x0D,0x0E,0,0x07,0x0E,0x0F,0,0x08,0x0F,0x10,0,
    0x09,0x10,0x11,0,0x0A,0x11,0x12,0,0x0B,0x12,0x13,0,0x0C,0x13,0x14,0,
    0x0D,0x14,0x15,0,0x0E,0x15,0x16,0,0x0F,0x16,0x17,0,0x10,0x17,0x18,0,
    0x11,0x18,0x19,0,0x12,0x19,0x1A,0,0x13,0x1A,0x1B,0,0x14,0x1B,0x1C,0,
    0x15,0x1C,0x1D,0,0x16,0x1D,0x1E,0,0x17,0x1E,0x1F,0,0x18,0x1F,0x20,0,
    0xDA,0xE0,0xE1,0,0xDB,0xE1,0xE2,0,0xDC,0xE2,0xE3,0,0xDD,0xE3,0xE4,0,
    0xDE,0xE4,0xE5,0,0xDF,0xE5,0xE6,0,0xE0,0xE6,0xE7,0,0xE1,0xE7,0xE8,0,
    0xE2,0xE8,0xE9,0,0xE3,0xE9,0xEA,0,0xE4,0xEA,0xEB,0,0xE5,0xEB,0xEC,0,
    0xE6,0xEC,0xED,0,0xE7,0xED,0xEE,0,0xE8,0xEE,0xEF,0,0xE9,0xEF,0xF0,0,
    0xEA,0xF0,0xF1,0,0xEB,0xF1,0xF2,0,0xEC,0xF2,0xF3,0,0xED,0xF3,0xF4,0,
    0xEE,0xF4,0xF5,0,0xEF,0xF5,0xF6,0,0xF0,0xF6,0xF7,0,0xF1,0xF7,0xF8,0,
    0xF2,0xF8,0xF9,0,0xF3,0xF9,0xFA,0,0xF4,0xFA,0xFB,0,0xF5,0xFB,0xFC,0,
    0xF6,0xFC,0xFD,0,0xF7,0xFD,0xFE,0,0xF8,0xFE,0xFF,0,0xF9,0xFF,0x00,0,
    0xFA,0x00,0x01,0,0xFB,0x01,0x02,0,0xFC,0x02,0x03,0,0xFD,0x03,0x04,0,
    0xFE,0x04,0x05,0,0xFF,0x05,0x06,0,0x00,0x06,0x07,0,0x01,0x07,0x08,0,
    0x02,0x08,0x09,0,0x03,0x09,0x0A,0,0x04,0x0A,0x0B,0,0x05,0x0B,0x0C,0,
    0x06,0x0C,0x0D,0,0x07,0x0D,0x0E,0,0x08,0x0E,0x0F,0,0x09,0x0F,0x10,0,
    0x0A,0x10,0x11,0,0x0B,0x11,0x12,0,0x0C,0x12,0x13,0,0x0D,0x13,0x14,0,
    0x0E,0x14,0x15,0,0x0F,0x15,0x16,0,0x10,0x16,0x17,0,0x11,0x17,0x18,0,
    0x12,0x18,0x19,0,0x13,0x19,0x1A,0,0x14,0x1A,0x1B,0,0x15,0x1B,0x1C,0,
    0x16,0x1C,0x1D,0,0x17,0x1D,0x1E,0,0x18,0x1E,0x1F,0,0x19,0x1F,0x20,0,
    0xDB,0xE0,0xE1,0,0xDC,0xE1,0xE2,0,0xDD,0xE2,0xE3,0,0xDE,0xE3,0xE4,0,
    0xDF,0xE4,0xE5,0,0xE0,0xE5,0xE6,0,0xE1,0xE6,0xE7,0,0xE2,0xE7,0xE8,0,
    0xE3,0xE8,0xE9,0,0xE4,0xE9,0xEA,0,0xE5,0xEA,0xEB,0,0xE6,0xEB,0xEC,0,
    0xE7,0xEC,0xED,0,0xE8,0xED,0xEE,0,0xE9,0xEE,0xEF,0,0xEA,0xEF,0xF0,0,
    0xEB,0xF0,0xF1,0,0xEC,0xF1,0xF2,0,0xED,0xF2,0xF3,0,0xEE,0xF3,0xF4,0,
    0xEF,0xF4,0xF5,0,0xF0,0xF5,0xF6,0,0xF1,0xF6,0xF7,0,0xF2,0xF7,0xF8,0,
    0xF3,0xF8,0xF9,0,0xF4,0xF9,0xFA,0,0xF5,0xFA,0xFB,0,0xF6,0xFB,0xFC,0,
    0xF7,0xFC,0xFD,0,0xF8,0xFD,0xFE,0,0xF9,0xFE,0xFF,0,0xFA,0xFF,0x00,0,
    0xFB,0x00,0x01,0,0xFC,0x01,0x02,0,0xFD,0x02,0x03,0,0xFE,0x03,0x04,0,
    0xFF,0x04,0x05,0,0x00,0x05,0x06,0,0x01,0x06,0x07,0,0x02,0x07,0x08,0,
    0x03,0x08,0x09,0,0x04,0x09,0x0A,0,0x05,0x0A,0x0B,0,0x06,0x0B,0x0C,0,
    0x07,0x0C,0x0D,0,0x08,0x0D,0x0E,0,0x09,0x0E,0x0F,0,0x0A,0x0F,0x10,0,
    0x0B,0x10,0x11,0,0x0C,0x11,0x12,0,0x0D,0x12,0x13,0,0x0E,0x13,0x14,0,
    0x0F,0x14,0x15,0,0x10,0x15,0x16,0,0x11,0x16,0x17,0,0x12,0x17,0x18,0,
    0x13,0x18,0x19,0,0x14,0x19,0x1A,0,0x15,0x1A,0x1B,0,0x16,0x1B,0x1C,0,
    0x17,0x1C,0x1D,0,0x18,0x1D,0x1E,0,0x19,0x1E,0x1F,0,0x1A,0x1F,0x20,0,
    0xDC,0xE0,0xE1,0,0xDD,0xE1,0xE2,0,0xDE,0xE2,0xE3,0,0xDF,0xE3,0xE4,0,
    0xE0,0xE4,0xE5,0,0xE1,0xE5,0xE6,0,0xE2,0xE6,0xE7,0,0xE3,0xE7,0xE8,0,
    0xE4,0xE8,0xE9,0,0xE5,0xE9,0xEA,0,0xE6,0xEA,0xEB,0,0xE7,0xEB,0xEC,0,
    0xE8,0xEC,0xED,0,0xE9,0xED,0xEE,0,0xEA,0xEE,0xEF,0,0xEB,0xEF,0xF0,0,
    0xEC,0xF0,0xF1,0,0xED,0xF1,0xF2,0,0xEE,0xF2,0xF3,0,0xEF,0xF3,0xF4,0,
    0xF0,0xF4,0xF5,0,0xF1,0xF5,0xF6,0,0xF2,0xF6,0xF7,0,0xF3,0xF7,0xF8,0,
    0xF4,0xF8,0xF9,0,0xF5,0xF9,0xFA,0,0xF6,0xFA,0xFB,0,0xF7,0xFB,0xFC,0,
    0xF8,0xFC,0xFD,0,0xF9,0xFD,0xFE,0,0xFA,0xFE,0xFF,0,0xFB,0xFF,0x00,0,
    0xFC,0x00,0x01,0,0xFD,0x01,0x02,0,0xFE,0x02,0x03,0,0xFF,0x03,0x04,0,
    0x00,0x04,0x05,0,0x01,0x05,0x06,0,0x02,0x06,0x07,0,0x03,0x07,0x08,0,
    0x04,0x08,0x09,0,0x05,0x09,0x0A,0,0x06,0x0A,0x0B,0,0x07,0x0B,0x0C,0,
    0x08,0x0C,0x0D,0,0x09,0x0D,0x0E,0,0x0A,0x0E,0x0F,0,0x0B,0x0F,0x10,0,
    0x0C,0x10,0x11,0,0x0D,0x11,0x12,0,0x0E,0x12,0x13,0,0x0F,0x13,0x14,0,
    0x10,0x14,0x15,0,0x11,0x15,0x16,0,0x12,0x16,0x17,0,0x13,0x17,0x18,0,
    0x14,0x18,0x19,0,0x15,0x19,0x1A,0,0x16,0x1A,0x1B,0,0x17,0x1B,0x1C,0,
    0x18,0x1C,0x1D,0,0x19,0x1D,0x1E,0,0x1A,0x1E,0x1F,0,0x1B,0x1F,0x20,0,
    0xDD,0xE0,0xE1,0,0xDE,0xE1,0xE2,0,0xDF,0xE2,0xE3,0,0xE0,0xE3,0xE4,0,
    0xE1,0xE4,0xE5,0,0xE2,0xE5,0xE6,0,0xE3,0xE6,0xE7,0,0xE4,0xE7,0xE8,0,
    0xE5,0xE8,0xE9,0,0xE6,0xE9,0xEA,0,0xE7,0xEA,0xEB,0,0xE8,0xEB,0xEC,0,
    0xE9,0xEC,0xED,0,0xEA,0xED,0xEE,0,0xEB,0xEE,0xEF,0,0xEC,0xEF,0xF0,0,
    0xED,0xF0,0xF1,0,0xEE,0xF1,0xF2,0,0xEF,0xF2,0xF3,0,0xF0,0xF3,0xF4,0,
    0xF1,0xF4,0xF5,0,0xF2,0xF5,0xF6,0,0xF3,0xF6,0xF7,0,0xF4,0xF7,0xF8,0,
    0xF5,0xF8,0xF9,0,0xF6,0xF9,0xFA,0,0xF7,0xFA,0xFB,0,0xF8,0xFB,0xFC,0,
    0xF9,0xFC,0xFD,0,0xFA,0xFD,0xFE,0,0xFB,0xFE,0xFF,0,0xFC,0xFF,0x00,0,
    0xFD,0x00,0x01,0,0xFE,0x01,0x02,0,0xFF,0x02,0x03,0,0x00,0x03,0x04,0,
    0x01,0x04,0x05,0,0x02,0x05,0x06,0,0x03,0x06,0x07,0,0x04,0x07,0x08,0,
    0x05,0x08,0x09,0,0x06,0x09,0x0A,0,0x07,0x0A,0x0B,0,0x08,0x0B,0x0C,0,
    0x09,0x0C,0x0D,0,0x0A,0x0D,0x0E,0,0x0B,0x0E,0x0F,0,0x0C,0x0F,0x10,0,
    0x0D,0x10,0x11,0,0x0E,0x11,0x12,0,0x0F,0x12,0x13,0,0x10,0x13,0x14,0,
    0x11,0x14,0x15,0,0x12,0x15,0x16,0,0x13,0x16,0x17,0,0x14,0x17,0x18,0,
    0x15,0x18,0x19,0,0x16,0x19,0x1A,0,0x17,0x1A,0x1B,0,0x18,0x1B,0x1C,0,
    0x19,0x1C,0x1D,0,0x1A,0x1D,0x1E,0,0x1B,0x1E,0x1F,0,0x1C,0x1F,0x20,0,
    0xDE,0xE0,0xE1,0,0xDF,0xE1,0xE2,0,0xE0,0xE2,0xE3,0,0xE1,0xE3,0xE4,0,
    0xE2,0xE4,0xE5,0,0xE3,0xE5,0xE6,0,0xE4,0xE6,0xE7,0,0xE5,0xE7,0xE8,0,
    0xE6,0xE8,0xE9,0,0xE7,0xE9,0xEA,0,0xE8,0xEA,0xEB,0,0xE9,0xEB,0xEC,0,
    0xEA,0xEC,0xED,0,0xEB,0xED,0xEE,0,0xEC,0xEE,0xEF,0,0xED,0xEF,0xF0,0,
    0xEE,0xF0,0xF1,0,0xEF,0xF1,0xF2,0,0xF0,0xF2,0xF3,0,0xF1,0xF3,0xF4,0,
    0xF2,0xF4,0xF5,0,0xF3,0xF5,0xF6,0,0xF4,0xF6,0xF7,0,0xF5,0xF7,0xF8,0,
    0xF6,0xF8,0xF9,0,0xF7,0xF9,0xFA,0,0xF8,0xFA,0xFB,0,0xF9,0xFB,0xFC,0,
    0xFA,0xFC,0xFD,0,0xFB,0xFD,0xFE,0,0xFC,0xFE,0xFF,0,0xFD,0xFF,0x00,0,
    0xFE,0x00,0x01,0,0xFF,0x01,0x02,0,0x00,0x02,0x03,0,0x01,0x03,0x04,0,
    0x02,0x04,0x05,0,0x03,0x05,0x06,0,0x04,0x06,0x07,0,0x05,0x07,0x08,0,
    0x06,0x08,0x09,0,0x07,0x09,0x0A,0,0x08,0x0A,0x0B,0,0x09,0x0B,0x0C,0,
    0x0A,0x0C,0x0D,0,0x0B,0x0D,0x0E,0,0x0C,0x0E,0x0F,0,0x0D,0x0F,0x10,0,
    0x0E,0x10,0x11,0,0x0F,0x11,0x12,0,0x10,0x12,0x13,0,0x11,0x13,0x14,0,
    0x12,0x14,0x15,0,0x13,0x15,0x16,0,0x14,0x16,0x17,0,0x15,0x17,0x18,0,
    0x16,0x18,0x19,0,0x17,0x19,0x1A,0,0x18,0x1A,0x1B,0,0x19,0x1B,0x1C,0,
    0x1A,0x1C,0x1D,0,0x1B,0x1D,0x1E,0,0x1C,0x1E,0x1F,0,0x1D,0x1F,0x20,0,
    0xDF,0xE0,0xE1,0,0xE0,0xE1,0xE2,0,0xE1,0xE2,0xE3,0,0xE2,0xE3,0xE4,0,
    0xE3,0xE4,0xE5,0,0xE4,0xE5,0xE6,0,0xE5,0xE6,0xE7,0,0xE6,0xE7,0xE8,0,
    0xE7,0xE8,0xE9,0,0xE8,0xE9,0xEA,0,0xE9,0xEA,0xEB,0,0xEA,0xEB,0xEC,0,
    0xEB,0xEC,0xED,0,0xEC,0xED,0xEE,0,0xED,0xEE,0xEF,0,0xEE,0xEF,0xF0,0,
    0xEF,0xF0,0xF1,0,0xF0,0xF1,0xF2,0,0xF1,0xF2,0xF3,0,0xF2,0xF3,0xF4,0,
    0xF3,0xF4,0xF5,0,0xF4,0xF5,0xF6,0,0xF5,0xF6,0xF7,0,0xF6,0xF7,0xF8,0,
    0xF7,0xF8,0xF9,0,0xF8,0xF9,0xFA,0,0xF9,0xFA,0xFB,0,0xFA,0xFB,0xFC,0,
    0xFB,0xFC,0xFD,0,0xFC,0xFD,0xFE,0,0xFD,0xFE,0xFF,0,0xFE,0xFF,0x00,0,
    0xFF,0x00,0x01,0,0x00,0x01,0x02,0,0x01,0x02,0x03,0,0x02,0x03,0x04,0,
    0x03,0x04,0x05,0,0x04,0x05,0x06,0,0x05,0x06,0x07,0,0x06,0x07,0x08,0,
    0x07,0x08,0x09,0,0x08,0x09,0x0A,0,0x09,0x0A,0x0B,0,0x0A,0x0B,0x0C,0,
    0x0B,0x0C,0x0D,0,0x0C,0x0D,0x0E,0,0x0D,0x0E,0x0F,0,0x0E,0x0F,0x10,0,
    0x0F,0x10,0x11,0,0x10,0x11,0x12,0,0x11,0x12,0x13,0,0x12,0x13,0x14,0,
    0x13,0x14,0x15,0,0x14,0x15,0x16,0,0x15,0x16,0x17,0,0x16,0x17,0x18,0,
    0x17,0x18,0x19,0,0x18,0x19,0x1A,0,0x19,0x1A,0x1B,0,0x1A,0x1B,0x1C,0,
    0x1B,0x1C,0x1D,0,0x1C,0x1D,0x1E,0,0x1D,0x1E,0x1F,0,0x1E,0x1F,0x20,0,
    0xE0,0xE0,0xE1,0,0xE1,0xE1,0xE2,0,0xE2,0xE2,0xE3,0,0xE3,0xE3,0xE4,0,
    0xE4,0xE4,0xE5,0,0xE5,0xE5,0xE6,0,0xE6,0xE6,0xE7,0,0xE7,0xE7,0xE8,0,
    0xE8,0xE8,0xE9,0,0xE9,0xE9,0xEA,0,0xEA,0xEA,0xEB,0,0xEB,0xEB,0xEC,0,
    0xEC,0xEC,0xED,0,0xED,0xED,0xEE,0,0xEE,0xEE,0xEF,0,0xEF,0xEF,0xF0,0,
    0xF0,0xF0,0xF1,0,0xF1,0xF1,0xF2,0,0xF2,0xF2,0xF3,0,0xF3,0xF3,0xF4,0,
    0xF4,0xF4,0xF5,0,0xF5,0xF5,0xF6,0,0xF6,0xF6,0xF7,0,0xF7,0xF7,0xF8,0,
    0xF8,0xF8,0xF9,0,0xF9,0xF9,0xFA,0,0xFA,0xFA,0xFB,0,0xFB,0xFB,0xFC,0,
    0xFC,0xFC,0xFD,0,0xFD,0xFD,0xFE,0,0xFE,0xFE,0xFF,0,0xFF,0xFF,0x00,0,
    0x00,0x00,0x01,0,0x01,0x01,0x02,0,0x02,0x02,0x03,0,0x03,0x03,0x04,0,
    0x04,0x04,0x05,0,0x05,0x05,0x06,0,0x06,0x06,0x07,0,0x07,0x07,0x08,0,
    0x08,0x08,0x09,0,0x09,0x09,0x0A,0,0x0A,0x0A,0x0B,0,0x0B,0x0B,0x0C,0,
    0x0C,0x0C,0x0D,0,0x0D,0x0D,0x0E,0,0x0E,0x0E,0x0F,0,0x0F,0x0F,0x10,0,
    0x10,0x10,0x11,0,0x11,0x11,0x12,0,0x12,0x12,0x13,0,0x13,0x13,0x14,0,
    0x14,0x14,0x15,0,0x15,0x15,0x16,0,0x16,0x16,0x17,0,0x17,0x17,0x18,0,
    0x18,0x18,0x19,0,0x19,0x19,0x1A,0,0x1A,0x1A,0x1B,0,0x1B,0x1B,0x1C,0,
    0x1C,0x1C,0x1D,0,0x1D,0x1D,0x1E,0,0x1E,0x1E,0x1F,0,0x1F,0x1F,0x20,0,
    0xE1,0xE0,0xE1,0,0xE2,0xE1,0xE2,0,0xE3,0xE2,0xE3,0,0xE4,0xE3,0xE4,0,
    0xE5,0xE4,0xE5,0,0xE6,0xE5,0xE6,0,0xE7,0xE6,0xE7,0,0xE8,0xE7,0xE8,0,
    0xE9,0xE8,0xE9,0,0xEA,0xE9,0xEA,0,0xEB,0xEA,0xEB,0,0xEC,0xEB,0xEC,0,
    0xED,0xEC,0xED,0,0xEE,0xED,0xEE,0,0xEF,0xEE,0xEF,0,0xF0,0xEF,0xF0,0,
    0xF1,0xF0,0xF1,0,0xF2,0xF1,0xF2,0,0xF3,0xF2,0xF3,0,0xF4,0xF3,0xF4,0,
    0xF5,0xF4,0xF5,0,0xF6,0xF5,0xF6,0,0xF7,0xF6,0xF7,0,0xF8,0xF7,0xF8,0,
    0xF9,0xF8,0xF9,0,0xFA,0xF9,0xFA,0,0xFB,0xFA,0xFB,0,0xFC,0xFB,0xFC,0,
    0xFD,0xFC,0xFD,0,0xFE,0xFD,0xFE,0,0xFF,0xFE,0xFF,0,0x00,0xFF,0x00,0,
    0x01,0x00,0x01,0,0x02,0x01,0x02,0,0x03,0x02,0x03,0,0x04,0x03,0x04,0,
    0x05,0x04,0x05,0,0x06,0x05,0x06,0,0x07,0x06,0x07,0,0x08,0x07,0x08,0,
    0x09,0x08,0x09,0,0x0A,0x09,0x0A,0,0x0B,0x0A,0x0B,0,0x0C,0x0B,0x0C,0,
    0x0D,0x0C,0x0D,0,0x0E,0x0D,0x0E,0,0x0F,0x0E,0x0F,0,0x10,0x0F,0x10,0,
    0x11,0x10,0x11,0,0x12,0x11,0x12,0,0x13,0x12,0x13,0,0x14,0x13,0x14,0,
    0x15,0x14,0x15,0,0x16,0x15,0x16,0,0x17,0x16,0x17,0,0x18,0x17,0x18,0,
    0x19,0x18,0x19,0,0x1A,0x19,0x1A,0,0x1B,0x1A,0x1B,0,0x1C,0x1B,0x1C,0,
    0x1D,0x1C,0x1D,0,0x1E,0x1D,0x1E,0,0x1F,0x1E,0x1F,0,0x20,0x1F,0x20,0,
    0xE2,0xE0,0xE1,0,0xE3,0xE1,0xE2,0,0xE4,0xE2,0xE3,0,0xE5,0xE3,0xE4,0,
    0xE6,0xE4,0xE5,0,0xE7,0xE5,0xE6,0,0xE8,0xE6,0xE7,0,0xE9,0xE7,0xE8,0,
    0xEA,0xE8,0xE9,0,0xEB,0xE9,0xEA,0,0xEC,0xEA,0xEB,0,0xED,0xEB,0xEC,0,
    0xEE,0xEC,0xED,0,0xEF,0xED,0xEE,0,0xF0,0xEE,0xEF,0,0xF1,0xEF,0xF0,0,
    0xF2,0xF0,0xF1,0,0xF3,0xF1,0xF2,0,0xF4,0xF2,0xF3,0,0xF5,0xF3,0xF4,0,
    0xF6,0xF4,0xF5,0,0xF7,0xF5,0xF6,0,0xF8,0xF6,0xF7,0,0xF9,0xF7,0xF8,0,
    0xFA,0xF8,0xF9,0,0xFB,0xF9,0xFA,0,0xFC,0xFA,0xFB,0,0xFD,0xFB,0xFC,0,
    0xFE,0xFC,0xFD,0,0xFF,0xFD,0xFE,0,0x00,0xFE,0xFF,0,0x01,0xFF,0x00,0,
    0x02,0x00,0x01,0,0x03,0x01,0x02,0,0x04,0x02,0x03,0,0x05,0x03,0x04,0,
    0x06,0x04,0x05,0,0x07,0x05,0x06,0,0x08,0x06,0x07,0,0x09,0x07,0x08,0,
    0x0A,0x08,0x09,0,0x0B,0x09,0x0A,0,0x0C,0x0A,0x0B,0,0x0D,0x0B,0x0C,0,
    0x0E,0x0C,0x0D,0,0x0F,0x0D,0x0E,0,0x10,0x0E,0x0F,0,0x11,0x0F,0x10,0,
    0x12,0x10,0x11,0,0x13,0x11,0x12,0,0x14,0x12,0x13,0,0x15,0x13,0x14,0,
    0x16,0x14,0x15,0,0x17,0x15,0x16,0,0x18,0x16,0x17,0,0x19,0x17,0x18,0,
    0x1A,0x18,0x19,0,0x1B,0x19,0x1A,0,0x1C,0x1A,0x1B,0,0x1D,0x1B,0x1C,0,
    0x1E,0x1C,0x1D,0,0x1F,0x1D,0x1E,0,0x20,0x1E,0x1F,0,0x21,0x1F,0x20,0,
    0xE3,0xE0,0xE1,0,0xE4,0xE1,0xE2,0,0xE5,0xE2,0xE3,0,0xE6,0xE3,0xE4,0,
    0xE7,0xE4,0xE5,0,0xE8,0xE5,0xE6,0,0xE9,0xE6,0xE7,0,0xEA,0xE7,0xE8,0,
    0xEB,0xE8,0xE9,0,0xEC,0xE9,0xEA,0,0xED,0xEA,0xEB,0,0xEE,0xEB,0xEC,0,
    0xEF,0xEC,0xED,0,0xF0,0xED,0xEE,0,0xF1,0xEE,0xEF,0,0xF2,0xEF,0xF0,0,
    0xF3,0xF0,0xF1,0,0xF4,0xF1,0xF2,0,0xF5,0xF2,0xF3,0,0xF6,0xF3,0xF4,0,
    0xF7,0xF4,0xF5,0,0xF8,0xF5,0xF6,0,0xF9,0xF6,0xF7,0,0xFA,0xF7,0xF8,0,
    0xFB,0xF8,0xF9,0,0xFC,0xF9,0xFA,0,0xFD,0xFA,0xFB,0,0xFE,0xFB,0xFC,0,
    0xFF,0xFC,0xFD,0,0x00,0xFD,0xFE,0,0x01,0xFE,0xFF,0,0x02,0xFF,0x00,0,
    0x03,0x00,0x01,0,0x04,0x01,0x02,0,0x05,0x02,0x03,0,0x06,0x03,0x04,0,
    0x07,0x04,0x05,0,0x08,0x05,0x06,0,0x09,0x06,0x07,0,0x0A,0x07,0x08,0,
    0x0B,0x08,0x09,0,0x0C,0x09,0x0A,0,0x0D,0x0A,0x0B,0,0x0E,0x0B,0x0C,0,
    0x0F,0x0C,0x0D,0,0x10,0x0D,0x0E,0,0x11,0x0E,0x0F,0,0x12,0x0F,0x10,0,
    0x13,0x10,0x11,0,0x14,0x11,0x12,0,0x15,0x12,0x13,0,0x16,0x13,0x14,0,
    0x17,0x14,0x15,0,0x18,0x15,0x16,0,0x19,0x16,0x17,0,0x1A,0x17,0x18,0,
    0x1B,0x18,0x19,0,0x1C,0x19,0x1A,0,0x1D,0x1A,0x1B,0,0x1E,0x1B,0x1C,0,
    0x1F,0x1C,0x1D,0,0x20,0x1D,0x1E,0,0x21,0x1E,0x1F,0,0x22,0x1F,0x20,0,
    0xE4,0xE0,0xE1,0,0xE5,0xE1,0xE2,0,0xE6,0xE2,0xE3,0,0xE7,0xE3,0xE4,0,
    0xE8,0xE4,0xE5,0,0xE9,0xE5,0xE6,0,0xEA,0xE6,0xE7,0,0xEB,0xE7,0xE8,0,
    0xEC,0xE8,0xE9,0,0xED,0xE9,0xEA,0,0xEE,0xEA,0xEB,0,0xEF,0xEB,0xEC,0,
    0xF0,0xEC,0xED,0,0xF1,0xED,0xEE,0,0xF2,0xEE,0xEF,0,0xF3,0xEF,0xF0,0,
    0xF4,0xF0,0xF1,0,0xF5,0xF1,0xF2,0,0xF6,0xF2,0xF3,0,0xF7,0xF3,0xF4,0,
    0xF8,0xF4,0xF5,0,0xF9,0xF5,0xF6,0,0xFA,0xF6,0xF7,0,0xFB,0xF7,0xF8,0,
    0xFC,0xF8,0xF9,0,0xFD,0xF9,0xFA,0,0xFE,0xFA,0xFB,0,0xFF,0xFB,0xFC,0,
    0x00,0xFC,0xFD,0,0x01,0xFD,0xFE,0,0x02,0xFE,0xFF,0,0x03,0xFF,0x00,0,
    0x04,0x00,0x01,0,0x05,0x01,0x02,0,0x06,0x02,0x03,0,0x07,0x03,0x04,0,
    0x08,0x04,0x05,0,0x09,0x05,0x06,0,0x0A,0x06,0x07,0,0x0B,0x07,0x08,0,
    0x0C,0x08,0x09,0,0x0D,0x09,0x0A,0,0x0E,0x0A,0x0B,0,0x0F,0x0B,0x0C,0,
    0x10,0x0C,0x0D,0,0x11,0x0D,0x0E,0,0x12,0x0E,0x0F,0,0x13,0x0F,0x10,0,
    0x14,0x10,0x11,0,0x15,0x11,0x12,0,0x16,0x12,0x13,0,0x17,0x13,0x14,0,
    0x18,0x14,0x15,0,0x19,0x15,0x16,0,0x1A,0x16,0x17,0,0x1B,0x17,0x18,0,
    0x1C,0x18,0x19,0,0x1D,0x19,0x1A,0,0x1E,0x1A,0x1B,0,0x1F,0x1B,0x1C,0,
    0x20,0x1C,0x1D,0,0x21,0x1D,0x1E,0,0x22,0x1E,0x1F,0,0x23,0x1F,0x20,0,
    0xE5,0xE0,0xE1,0,0xE6,0xE1,0xE2,0,0xE7,0xE2,0xE3,0,0xE8,0xE3,0xE4,0,
    0xE9,0xE4,0xE5,0,0xEA,0xE5,0xE6,0,0xEB,0xE6,0xE7,0,0xEC,0xE7,0xE8,0,
    0xED,0xE8,0xE9,0,0xEE,0xE9,0xEA,0,0xEF,0xEA,0xEB,0,0xF0,0xEB,0xEC,0,
    0xF1,0xEC,0xED,0,0xF2,0xED,0xEE,0,0xF3,0xEE,0xEF,0,0xF4,0xEF,0xF0,0,
    0xF5,0xF0,0xF1,0,0xF6,0xF1,0xF2,0,0xF7,0xF2,0xF3,0,0xF8,0xF3,0xF4,0,
    0xF9,0xF4,0xF5,0,0xFA,0xF5,0xF6,0,0xFB,0xF6,0xF7,0,0xFC,0xF7,0xF8,0,
    0xFD,0xF8,0xF9,0,0xFE,0xF9,0xFA,0,0xFF,0xFA,0xFB,0,0x00,0xFB,0xFC,0,
    0x01,0xFC,0xFD,0,0x02,0xFD,0xFE,0,0x03,0xFE,0xFF,0,0x04,0xFF,0x00,0,
    0x05,0x00,0x01,0,0x06,0x01,0x02,0,0x07,0x02,0x03,0,0x08,0x03,0x04,0,
    0x09,0x04,0x05,0,0x0A,0x05,0x06,0,0x0B,0x06,0x07,0,0x0C,0x07,0x08,0,
    0x0D,0x08,0x09,0,0x0E,0x09,0x0A,0,0x0F,0x0A,0x0B,0,0x10,0x0B,0x0C,0,
    0x11,0x0C,0x0D,0,0x12,0x0D,0x0E,0,0x13,0x0E,0x0F,0,0x14,0x0F,0x10,0,
    0x15,0x10,0x11,0,0x16,0x11,0x12,0,0x17,0x12,0x13,0,0x18,0x13,0x14,0,
    0x19,0x14,0x15,0,0x1A,0x15,0x16,0,0x1B,0x16,0x17,0,0x1C,0x17,0x18,0,
    0x1D,0x18,0x19,0,0x1E,0x19,0x1A,0,0x1F,0x1A,0x1B,0,0x20,0x1B,0x1C,0,
    0x21,0x1C,0x1D,0,0x22,0x1D,0x1E,0,0x23,0x1E,0x1F,0,0x24,0x1F,0x20,0,
    0xE6,0xE0,0xE1,0,0xE7,0xE1,0xE2,0,0xE8,0xE2,0xE3,0,0xE9,0xE3,0xE4,0,
    0xEA,0xE4,0xE5,0,0xEB,0xE5,0xE6,0,0xEC,0xE6,0xE7,0,0xED,0xE7,0xE8,0,
    0xEE,0xE8,0xE9,0,0xEF,0xE9,0xEA,0,0xF0,0xEA,0xEB,0,0xF1,0xEB,0xEC,0,
    0xF2,0xEC,0xED,0,0xF3,0xED,0xEE,0,0xF4,0xEE,0xEF,0,0xF5,0xEF,0xF0,0,
    0xF6,0xF0,0xF1,0,0xF7,0xF1,0xF2,0,0xF8,0xF2,0xF3,0,0xF9,0xF3,0xF4,0,
    0xFA,0xF4,0xF5,0,0xFB,0xF5,0xF6,0,0xFC,0xF6,0xF7,0,0xFD,0xF7,0xF8,0,
    0xFE,0xF8,0xF9,0,0xFF,0xF9,0xFA,0,0x00,0xFA,0xFB,0,0x01,0xFB,0xFC,0,
    0x02,0xFC,0xFD,0,0x03,0xFD,0xFE,0,0x04,0xFE,0xFF,0,0x05,0xFF,0x00,0,
    0x06,0x00,0x01,0,0x07,0x01,0x02,0,0x08,0x02,0x03,0,0x09,0x03,0x04,0,
    0x0A,0x04,0x05,0,0x0B,0x05,0x06,0,0x0C,0x06,0x07,0,0x0D,0x07,0x08,0,
    0x0E,0x08,0x09,0,0x0F,0x09,0x0A,0,0x10,0x0A,0x0B,0,0x11,0x0B,0x0C,0,
    0x12,0x0C,0x0D,0,0x13,0x0D,0x0E,0,0x14,0x0E,0x0F,0,0x15,0x0F,0x10,0,
    0x16,0x10,0x11,0,0x17,0x11,0x12,0,0x18,0x12,0x13,0,0x19,0x13,0x14,0,
    0x1A,0x14,0x15,0,0x1B,0x15,0x16,0,0x1C,0x16,0x17,0,0x1D,0x17,0x18,0,
    0x1E,0x18,0x19,0,0x1F,0x19,0x1A,0,0x20,0x1A,0x1B,0,0x21,0x1B,0x1C,0,
    0x22,0x1C,0x1D,0,0x23,0x1D,0x1E,0,0x24,0x1E,0x1F,0,0x25,0x1F,0x20,0,
    0xE7,0xE0,0xE1,0,0xE8,0xE1,0xE2,0,0xE9,0xE2,0xE3,0,0xEA,0xE3,0xE4,0,
    0xEB,0xE4,0xE5,0,0xEC,0xE5,0xE6,0,0xED,0xE6,0xE7,0,0xEE,0xE7,0xE8,0,
    0xEF,0xE8,0xE9,0,0xF0,0xE9,0xEA,0,0xF1,0xEA,0xEB,0,0xF2,0xEB,0xEC,0,
    0xF3,0xEC,0xED,0,0xF4,0xED,0xEE,0,0xF5,0xEE,0xEF,0,0xF6,0xEF,0xF0,0,
    0xF7,0xF0,0xF1,0,0xF8,0xF1,0xF2,0,0xF9,0xF2,0xF3,0,0xFA,0xF3,0xF4,0,
    0xFB,0xF4,0xF5,0,0xFC,0xF5,0xF6,0,0xFD,0xF6,0xF7,0,0xFE,0xF7,0xF8,0,
    0xFF,0xF8,0xF9,0,0x00,0xF9,0xFA,0,0x01,0xFA,0xFB,0,0x02,0xFB,0xFC,0,
    0x03,0xFC,0xFD,0,0x04,0xFD,0xFE,0,0x05,0xFE,0xFF,0,0x06,0xFF,0x00,0,
    0x07,0x00,0x01,0,0x08,0x01,0x02,0,0x09,0x02,0x03,0,0x0A,0x03,0x04,0,
    0x0B,0x04,0x05,0,0x0C,0x05,0x06,0,0x0D,0x06,0x07,0,0x0E,0x07,0x08,0,
    0x0F,0x08,0x09,0,0x10,0x09,0x0A,0,0x11,0x0A,0x0B,0,0x12,0x0B,0x0C,0,
    0x13,0x0C,0x0D,0,0x14,0x0D,0x0E,0,0x15,0x0E,0x0F,0,0x16,0x0F,0x10,0,
    0x17,0x10,0x11,0,0x18,0x11,0x12,0,0x19,0x12,0x13,0,0x1A,0x13,0x14,0,
    0x1B,0x14,0x15,0,0x1C,0x15,0x16,0,0x1D,0x16,0x17,0,0x1E,0x17,0x18,0,
    0x1F,0x18,0x19,0,0x20,0x19,0x1A,0,0x21,0x1A,0x1B,0,0x22,0x1B,0x1C,0,
    0x23,0x1C,0x1D,0,0x24,0x1D,0x1E,0,0x25,0x1E,0x1F,0,0x26,0x1F,0x20,0,
    0xD8,0xE0,0xE2,0,0xD9,0xE1,0xE3,0,0xDA,0xE2,0xE4,0,0xDB,0xE3,0xE5,0,
    0xDC,0xE4,0xE6,0,0xDD,0xE5,0xE7,0,0xDE,0xE6,0xE8,0,0xDF,0xE7,0xE9,0,
    0xE0,0xE8,0xEA,0,0xE1,0xE9,0xEB,0,0xE2,0xEA,0xEC,0,0xE3,0xEB,0xED,0,
    0xE4,0xEC,0xEE,0,0xE5,0xED,0xEF,0,0xE6,0xEE,0xF0,0,0xE7,0xEF,0xF1,0,
    0xE8,0xF0,0xF2,0,0xE9,0xF1,0xF3,0,0xEA,0xF2,0xF4,0,0xEB,0xF3,0xF5,0,
    0xEC,0xF4,0xF6,0,0xED,0xF5,0xF7,0,0xEE,0xF6,0xF8,0,0xEF,0xF7,0xF9,0,
    0xF0,0xF8,0xFA,0,0xF1,0xF9,0xFB,0,0xF2,0xFA,0xFC,0,0xF3,0xFB,0xFD,0,
    0xF4,0xFC,0xFE,0,0xF5,0xFD,0xFF,0,0xF6,0xFE,0x00,0,0xF7,0xFF,0x01,0,
    0xF8,0x00,0x02,0,0xF9,0x01,0x03,0,0xFA,0x02,0x04,0,0xFB,0x03,0x05,0,
    0xFC,0x04,0x06,0,0xFD,0x05,0x07,0,0xFE,0x06,0x08,0,0xFF,0x07,0x09,0,
    0x00,0x08,0x0A,0,0x01,0x09,0x0B,0,0x02,0x0A,0x0C,0,0x03,0x0B,0x0D,0,
    0x04,0x0C,0x0E,0,0x05,0x0D,0x0F,0,0x06,0x0E,0x10,0,0x07,0x0F,0x11,0,
    0x08,0x10,0x12,0,0x09,0x11,0x13,0,0x0A,0x12,0x14,0,0x0B,0x13,0x15,0,
    0x0C,0x14,0x16,0,0x0D,0x15,0x17,0,0x0E,0x16,0x18,0,0x0F,0x17,0x19,0,
    0x10,0x18,0x1A,0,0x11,0x19,0x1B,0,0x12,0x1A,0x1C,0,0x13,0x1B,0x1D,0,
    0x14,0x1C,0x1E,0,0x15,0x1D,0x1F,0,0x16,0x1E,0x20,0,0x17,0x1F,0x21,0,
    0xD9,0xE0,0xE2,0,0xDA,0xE1,0xE3,0,0xDB,0xE2,0xE4,0,0xDC,0xE3,0xE5,0,
    0xDD,0xE4,0xE6,0,0xDE,0xE5,0xE7,0,0xDF,0xE6,0xE8,0,0xE0,0xE7,0xE9,0,
    0xE1,0xE8,0xEA,0,0xE2,0xE9,0xEB,0,0xE3,0xEA,0xEC,0,0xE4,0xEB,0xED,0,
    0xE5,0xEC,0xEE,0,0xE6,0xED,0xEF,0,0xE7,0xEE,0xF0,0,0xE8,0xEF,0xF1,0,
    0xE9,0xF0,0xF2,0,0xEA,0xF1,0xF3,0,0xEB,0xF2,0xF4,0,0xEC,0xF3,0xF5,0,
    0xED,0xF4,0xF6,0,0xEE,0xF5,0xF7,0,0xEF,0xF6,0xF8,0,0xF0,0xF7,0xF9,0,
    0xF1,0xF8,0xFA,0,0xF2,0xF9,0xFB,0,0xF3,0xFA,0xFC,0,0xF4,0xFB,0xFD,0,
    0xF5,0xFC,0xFE,0,0xF6,0xFD,0xFF,0,0xF7,0xFE,0x00,0,0xF8,0xFF,0x01,0,
    0xF9,0x00,0x02,0,0xFA,0x01,0x03,0,0xFB,0x02,0x04,0,0xFC,0x03,0x05,0,
    0xFD,0x04,0x06,0,0xFE,0x05,0x07,0,0xFF,0x06,0x08,0,0x00,0x07,0x09,0,
    0x01,0x08,0x0A,0,0x02,0x09,0x0B,0,0x03,0x0A,0x0C,0,0x04,0x0B,0x0D,0,
    0x05,0x0C,0x0E,0,0x06,0x0D,0x0F,0,0x07,0x0E,0x10,0,0x08,0x0F,0x11,0,
    0x09,0x10,0x12,0,0x0A,0x11,0x13,0,0x0B,0x12,0x14,0,0x0C,0x13,0x15,0,
    0x0D,0x14,0x16,0,0x0E,0x15,0x17,0,0x0F,0x16,0x18,0,0x10,0x17,0x19,0,
    0x11,0x18,0x1A,0,0x12,0x19,0x1B,0,0x13,0x1A,0x1C,0,0x14,0x1B,0x1D,0,
    0x15,0x1C,0x1E,0,0x16,0x1D,0x1F,0,0x17,0x1E,0x20,0,0x18,0x1F,0x21,0,
    0xDA,0xE0,0xE2,0,0xDB,0xE1,0xE3,0,0xDC,0xE2,0xE4,0,0xDD,0xE3,0xE5,0,
    0xDE,0xE4,0xE6,0,0xDF,0xE5,0xE7,0,0xE0,0xE6,0xE8,0,0xE1,0xE7,0xE9,0,
    0xE2,0xE8,0xEA,0,0xE3,0xE9,0xEB,0,0xE4,0xEA,0xEC,0,0xE5,0xEB,0xED,0,
    0xE6,0xEC,0xEE,0,0xE7,0xED,0xEF,0,0xE8,0xEE,0xF0,0,0xE9,0xEF,0xF1,0,
    0xEA,0xF0,0xF2,0,0xEB,0xF1,0xF3,0,0xEC,0xF2,0xF4,0,0xED,0xF3,0xF5,0,
    0xEE,0xF4,0xF6,0,0xEF,0xF5,0xF7,0,0xF0,0xF6,0xF8,0,0xF1,0xF7,0xF9,0,
    0xF2,0xF8,0xFA,0,0xF3,0xF9,0xFB,0,0xF4,0xFA,0xFC,0,0xF5,0xFB,0xFD,0,
    0xF6,0xFC,0xFE,0,0xF7,0xFD,0xFF,0,0xF8,0xFE,0x00,0,0xF9,0xFF,0x01,0,
    0xFA,0x00,0x02,0,0xFB,0x01,0x03,0,0xFC,0x02,0x04,0,0xFD,0x03,0x05,0,
    0xFE,0x04,0x06,0,0xFF,0x05,0x07,0,0x00,0x06,0x08,0,0x01,0x07,0x09,0,
    0x02,0x08,0x0A,0,0x03,0x09,0x0B,0,0x04,0x0A,0x0C,0,0x05,0x0B,0x0D,0,
    0x06,0x0C,0x0E,0,0x07,0x0D,0x0F,0,0x08,0x0E,0x10,0,0x09,0x0F,0x11,0,
    0x0A,0x10,0x12,0,0x0B,0x11,0x13,0,0x0C,0x12,0x14,0,0x0D,0x13,0x15,0,
    0x0E,0x14,0x16,0,0x0F,0x15,0x17,0,0x10,0x16,0x18,0,0x11,0x17,0x19,0,
    0x12,0x18,0x1A,0,0x13,0x19,0x1B,0,0x14,0x1A,0x1C,0,0x15,0x1B,0x1D,0,
    0x16,0x1C,0x1E,0,0x17,0x1D,0x1F,0,0x18,0x1E,0x20,0,0x19,0x1F,0x21,0,
    0xDB,0xE0,0xE2,0,0xDC,0xE1,0xE3,0,0xDD,0xE2,0xE4,0,0xDE,0xE3,0xE5,0,
    0xDF,0xE4,0xE6,0,0xE0,0xE5,0xE7,0,0xE1,0xE6,0xE8,0,0xE2,0xE7,0xE9,0,
    0xE3,0xE8,0xEA,0,0xE4,0xE9,0xEB,0,0xE5,0xEA,0xEC,0,0xE6,0xEB,0xED,0,
    0xE7,0xEC,0xEE,0,0xE8,0xED,0xEF,0,0xE9,0xEE,0xF0,0,0xEA,0xEF,0xF1,0,
    0xEB,0xF0,0xF2,0,0xEC,0xF1,0xF3,0,0xED,0xF2,0xF4,0,0xEE,0xF3,0xF5,0,
    0xEF,0xF4,0xF6,0,0xF0,0xF5,0xF7,0,0xF1,0xF6,0xF8,0,0xF2,0xF7,0xF9,0,
    0xF3,0xF8,0xFA,0,0xF4,0xF9,0xFB,0,0xF5,0xFA,0xFC,0,0xF6,0xFB,0xFD,0,
    0xF7,0xFC,0xFE,0,0xF8,0xFD,0xFF,0,0xF9,0xFE,0x00,0,0xFA,0xFF,0x01,0,
    0xFB,0x00,0x02,0,0xFC,0x01,0x03,0,0xFD,0x02,0x04,0,0xFE,0x03,0x05,0,
    0xFF,0x04,0x06,0,0x00,0x05,0x07,0,0x01,0x06,0x08,0,0x02,0x07,0x09,0,
    0x03,0x08,0x0A,0,0x04,0x09,0x0B,0,0x05,0x0A,0x0C,0,0x06,0x0B,0x0D,0,
    0x07,0x0C,0x0E,0,0x08,0x0D,0x0F,0,0x09,0x0E,0x10,0,0x0A,0x0F,0x11,0,
    0x0B,0x10,0x12,0,0x0C,0x11,0x13,0,0x0D,0x12,0x14,0,0x0E,0x13,0x15,0,
    0x0F,0x14,0x16,0,0x10,0x15,0x17,0,0x11,0x16,0x18,0,0x12,0x17,0x19,0,
    0x13,0x18,0x1A,0,0x14,0x19,0x1B,0,0x15,0x1A,0x1C,0,0x16,0x1B,0x1D,0,
    0x17,0x1C,0x1E,0,0x18,0x1D,0x1F,0,0x19,0x1E,0x20,0,0x1A,0x1F,0x21,0,
    0xDC,0xE0,0xE2,0,0xDD,0xE1,0xE3,0,0xDE,0xE2,0xE4,0,0xDF,0xE3,0xE5,0,
    0xE0,0xE4,0xE6,0,0xE1,0xE5,0xE7,0,0xE2,0xE6,0xE8,0,0xE3,0xE7,0xE9,0,
    0xE4,0xE8,0xEA,0,0xE5,0xE9,0xEB,0,0xE6,0xEA,0xEC,0,0xE7,0xEB,0xED,0,
    0xE8,0xEC,0xEE,0,0xE9,0xED,0xEF,0,0xEA,0xEE,0xF0,0,0xEB,0xEF,0xF1,0,
    0xEC,0xF0,0xF2,0,0xED,0xF1,0xF3,0,0xEE,0xF2,0xF4,0,0xEF,0xF3,0xF5,0,
    0xF0,0xF4,0xF6,0,0xF1,0xF5,0xF7,0,0xF2,0xF6,0xF8,0,0xF3,0xF7,0xF9,0,
    0xF4,0xF8,0xFA,0,0xF5,0xF9,0xFB,0,0xF6,0xFA,0xFC,0,0xF7,0xFB,0xFD,0,
    0xF8,0xFC,0xFE,0,0xF9,0xFD,0xFF,0,0xFA,0xFE,0x00,0,0xFB,0xFF,0x01,0,
    0xFC,0x00,0x02,0,0xFD,0x01,0x03,0,0xFE,0x02,0x04,0,0xFF,0x03,0x05,0,
    0x00,0x04,0x06,0,0x01,0x05,0x07,0,0x02,0x06,0x08,0,0x03,0x07,0x09,0,
    0x04,0x08,0x0A,0,0x05,0x09,0x0B,0,0x06,0x0A,0x0C,0,0x07,0x0B,0x0D,0,
    0x08,0x0C,0x0E,0,0x09,0x0D,0x0F,0,0x0A,0x0E,0x10,0,0x0B,0x0F,0x11,0,
    0x0C,0x10,0x12,0,0x0D,0x11,0x13,0,0x0E,0x12,0x14,0,0x0F,0x13,0x15,0,
    0x10,0x14,0x16,0,0x11,0x15,0x17,0,0x12,0x16,0x18,0,0x13,0x17,0x19,0,
    0x14,0x18,0x1A,0,0x15,0x19,0x1B,0,0x16,0x1A,0x1C,0,0x17,0x1B,0x1D,0,
    0x18,0x1C,0x1E,0,0x19,0x1D,0x1F,0,0x1A,0x1E,0x20,0,0x1B,0x1F,0x21,0,
    0xDD,0xE0,0xE2,0,0xDE,0xE1,0xE3,0,0xDF,0xE2,0xE4,0,0xE0,0xE3,0xE5,0,
    0xE1,0xE4,0xE6,0,0xE2,0xE5,0xE7,0,0xE3,0xE6,0xE8,0,0xE4,0xE7,0xE9,0,
    0xE5,0xE8,0xEA,0,0xE6,0xE9,0xEB,0,0xE7,0xEA,0xEC,0,0xE8,0xEB,0xED,0,
    0xE9,0xEC,0xEE,0,0xEA,0xED,0xEF,0,0xEB,0xEE,0xF0,0,0xEC,0xEF,0xF1,0,
    0xED,0xF0,0xF2,0,0xEE,0xF1,0xF3,0,0xEF,0xF2,0xF4,0,0xF0,0xF3,0xF5,0,
    0xF1,0xF4,0xF6,0,0xF2,0xF5,0xF7,0,0xF3,0xF6,0xF8,0,0xF4,0xF7,0xF9,0,
    0xF5,0xF8,0xFA,0,0xF6,0xF9,0xFB,0,0xF7,0xFA,0xFC,0,0xF8,0xFB,0xFD,0,
    0xF9,0xFC,0xFE,0,0xFA,0xFD,0xFF,0,0xFB,0xFE,0x00,0,0xFC,0xFF,0x01,0,
    0xFD,0x00,0x02,0,0xFE,0x01,0x03,0,0xFF,0x02,0x04,0,0x00,0x03,0x05,0,
    0x01,0x04,0x06,0,0x02,0x05,0x07,0,0x03,0x06,0x08,0,0x04,0x07,0x09,0,
    0x05,0x08,0x0A,0,0x06,0x09,0x0B,0,0x07,0x0A,0x0C,0,0x08,0x0B,0x0D,0,
    0x09,0x0C,0x0E,0,0x0A,0x0D,0x0F,0,0x0B,0x0E,0x10,0,0x0C,0x0F,0x11,0,
    0x0D,0x10,0x12,0,0x0E,0x11,0x13,0,0x0F,0x12,0x14,0,0x10,0x13,0x15,0,
    0x11,0x14,0x16,0,0x12,0x15,0x17,0,0x13,0x16,0x18,0,0x14,0x17,0x19,0,
    0x15,0x18,0x1A,0,0x16,0x19,0x1B,0,0x17,0x1A,0x1C,0,0x18,0x1B,0x1D,0,
    0x19,0x1C,0x1E,0,0x1A,0x1D,0x1F,0,0x1B,0x1E,0x20,0,0x1C,0x1F,0x21,0,
    0xDE,0xE0,0xE2,0,0xDF,0xE1,0xE3,0,0xE0,0xE2,0xE4,0,0xE1,0xE3,0xE5,0,
    0xE2,0xE4,0xE6,0,0xE3,0xE5,0xE7,0,0xE4,0xE6,0xE8,0,0xE5,0xE7,0xE9,0,
    0xE6,0xE8,0xEA,0,0xE7,0xE9,0xEB,0,0xE8,0xEA,0xEC,0,0xE9,0xEB,0xED,0,
    0xEA,0xEC,0xEE,0,0xEB,0xED,0xEF,0,0xEC,0xEE,0xF0,0,0xED,0xEF,0xF1,0,
    0xEE,0xF0,0xF2,0,0xEF,0xF1,0xF3,0,0xF0,0xF2,0xF4,0,0xF1,0xF3,0xF5,0,
    0xF2,0xF4,0xF6,0,0xF3,0xF5,0xF7,0,0xF4,0xF6,0xF8,0,0xF5,0xF7,0xF9,0,
    0xF6,0xF8,0xFA,0,0xF7,0xF9,0xFB,0,0xF8,0xFA,0xFC,0,0xF9,0xFB,0xFD,0,
    0xFA,0xFC,0xFE,0,0xFB,0xFD,0xFF,0,0xFC,0xFE,0x00,0,0xFD,0xFF,0x01,0,
    0xFE,0x00,0x02,0,0xFF,0x01,0x03,0,0x00,0x02,0x04,0,0x01,0x03,0x05,0,
    0x02,0x04,0x06,0,0x03,0x05,0x07,0,0x04,0x06,0x08,0,0x05,0x07,0x09,0,
    0x06,0x08,0x0A,0,0x07,0x09,0x0B,0,0x08,0x0A,0x0C,0,0x09,0x0B,0x0D,0,
    0x0A,0x0C,0x0E,0,0x0B,0x0D,0x0F,0,0x0C,0x0E,0x10,0,0x0D,0x0F,0x11,0,
    0x0E,0x10,0x12,0,0x0F,0x11,0x13,0,0x10,0x12,0x14,0,0x11,0x13,0x15,0,
    0x12,0x14,0x16,0,0x13,0x15,0x17,0,0x14,0x16,0x18,0,0x15,0x17,0x19,0,
    0x16,0x18,0x1A,0,0x17,0x19,0x1B,0,0x18,0x1A,0x1C,0,0x19,0x1B,0x1D,0,
    0x1A,0x1C,0x1E,0,0x1B,0x1D,0x1F,0,0x1C,0x1E,0x20,0,0x1D,0x1F,0x21,0,
    0xDF,0xE0,0xE2,0,0xE0,0xE1,0xE3,0,0xE1,0xE2,0xE4,0,0xE2,0xE3,0xE5,0,
    0xE3,0xE4,0xE6,0,0xE4,0xE5,0xE7,0,0xE5,0xE6,0xE8,0,0xE6,0xE7,0xE9,0,
    0xE7,0xE8,0xEA,0,0xE8,0xE9,0xEB,0,0xE9,0xEA,0xEC,0,0xEA,0xEB,0xED,0,
    0xEB,0xEC,0xEE,0,0xEC,0xED,0xEF,0,0xED,0xEE,0xF0,0,0xEE,0xEF,0xF1,0,
    0xEF,0xF0,0xF2,0,0xF0,0xF1,0xF3,0,0xF1,0xF2,0xF4,0,0xF2,0xF3,0xF5,0,
    0xF3,0xF4,0xF6,0,0xF4,0xF5,0xF7,0,0xF5,0xF6,0xF8,0,0xF6,0xF7,0xF9,0,
    0xF7,0xF8,0xFA,0,0xF8,0xF9,0xFB,0,0xF9,0xFA,0xFC,0,0xFA,0xFB,0xFD,0,
    0xFB,0xFC,0xFE,0,0xFC,0xFD,0xFF,0,0xFD,0xFE,0x00,0,0xFE,0xFF,0x01,0,
    0xFF,0x00,0x02,0,0x00,0x01,0x03,0,0x01,0x02,0x04,0,0x02,0x03,0x05,0,
    0x03,0x04,0x06,0,0x04,0x05,0x07,0,0x05,0x06,0x08,0,0x06,0x07,0x09,0,
    0x07,0x08,0x0A,0,0x08,0x09,0x0B,0,0x09,0x0A,0x0C,0,0x0A,0x0B,0x0D,0,
    0x0B,0x0C,0x0E,0,0x0C,0x0D,0x0F,0,0x0D,0x0E,0x10,0,0x0E,0x0F,0x11,0,
    0x0F,0x10,0x12,0,0x10,0x11,0x13,0,0x11,0x12,0x14,0,0x12,0x13,0x15,0,
    0x13,0x14,0x16,0,0x14,0x15,0x17,0,0x15,0x16,0x18,0,0x16,0x17,0x19,0,
    0x17,0x18,0x1A,0,0x18,0x19,0x1B,0,0x19,0x1A,0x1C,0,0x1A,0x1B,0x1D,0,
    0x1B,0x1C,0x1E,0,0x1C,0x1D,0x1F,0,0x1D,0x1E,0x20,0,0x1E,0x1F,0x21,0,
    0xE0,0xE0,0xE2,0,0xE1,0xE1,0xE3,0,0xE2,0xE2,0xE4,0,0xE3,0xE3,0xE5,0,
    0xE4,0xE4,0xE6,0,0xE5,0xE5,0xE7,0,0xE6,0xE6,0xE8,0,0xE7,0xE7,0xE9,0,
    0xE8,0xE8,0xEA,0,0xE9,0xE9,0xEB,0,0xEA,0xEA,0xEC,0,0xEB,0xEB,0xED,0,
    0xEC,0xEC,0xEE,0,0xED,0xED,0xEF,0,0xEE,0xEE,0xF0,0,0xEF,0xEF,0xF1,0,
    0xF0,0xF0,0xF2,0,0xF1,0xF1,0xF3,0,0xF2,0xF2,0xF4,0,0xF3,0xF3,0xF5,0,
    0xF4,0xF4,0xF6,0,0xF5,0xF5,0xF7,0,0xF6,0xF6,0xF8,0,0xF7,0xF7,0xF9,0,
    0xF8,0xF8,0xFA,0,0xF9,0xF9,0xFB,0,0xFA,0xFA,0xFC,0,0xFB,0xFB,0xFD,0,
    0xFC,0xFC,0xFE,0,0xFD,0xFD,0xFF,0,0xFE,0xFE,0x00,0,0xFF,0xFF,0x01,0,
    0x00,0x00,0x02,0,0x01,0x01,0x03,0,0x02,0x02,0x04,0,0x03,0x03,0x05,0,
    0x04,0x04,0x06,0,0x05,0x05,0x07,0,0x06,0x06,0x08,0,0x07,0x07,0x09,0,
    0x08,0x08,0x0A,0,0x09,0x09,0x0B,0,0x0A,0x0A,0x0C,0,0x0B,0x0B,0x0D,0,
    0x0C,0x0C,0x0E,0,0x0D,0x0D,0x0F,0,0x0E,0x0E,0x10,0,0x0F,0x0F,0x11,0,
    0x10,0x10,0x12,0,0x11,0x11,0x13,0,0x12,0x12,0x14,0,0x13,0x13,0x15,0,
    0x14,0x14,0x16,0,0x15,0x15,0x17,0,0x16,0x16,0x18,0,0x17,0x17,0x19,0,
    0x18,0x18,0x1A,0,0x19,0x19,0x1B,0,0x1A,0x1A,0x1C,0,0x1B,0x1B,0x1D,0,
    0x1C,0x1C,0x1E,0,0x1D,0x1D,0x1F,0,0x1E,0x1E,0x20,0,0x1F,0x1F,0x21,0,
    0xE1,0xE0,0xE2,0,0xE2,0xE1,0xE3,0,0xE3,0xE2,0xE4,0,0xE4,0xE3,0xE5,0,
    0xE5,0xE4,0xE6,0,0xE6,0xE5,0xE7,0,0xE7,0xE6,0xE8,0,0xE8,0xE7,0xE9,0,
    0xE9,0xE8,0xEA,0,0xEA,0xE9,0xEB,0,0xEB,0xEA,0xEC,0,0xEC,0xEB,0xED,0,
    0xED,0xEC,0xEE,0,0xEE,0xED,0xEF,0,0xEF,0xEE,0xF0,0,0xF0,0xEF,0xF1,0,
    0xF1,0xF0,0xF2,0,0xF2,0xF1,0xF3,0,0xF3,0xF2,0xF4,0,0xF4,0xF3,0xF5,0,
    0xF5,0xF4,0xF6,0,0xF6,0xF5,0xF7,0,0xF7,0xF6,0xF8,0,0xF8,0xF7,0xF9,0,
    0xF9,0xF8,0xFA,0,0xFA,0xF9,0xFB,0,0xFB,0xFA,0xFC,0,0xFC,0xFB,0xFD,0,
    0xFD,0xFC,0xFE,0,0xFE,0xFD,0xFF,0,0xFF,0xFE,0x00,0,0x00,0xFF,0x01,0,
    0x01,0x00,0x02,0,0x02,0x01,0x03,0,0x03,0x02,0x04,0,0x04,0x03,0x05,0,
    0x05,0x04,0x06,0,0x06,0x05,0x07,0,0x07,0x06,0x08,0,0x08,0x07,0x09,0,
    0x09,0x08,0x0A,0,0x0A,0x09,0x0B,0,0x0B,0x0A,0x0C,0,0x0C,0x0B,0x0D,0,
    0x0D,0x0C,0x0E,0,0x0E,0x0D,0x0F,0,0x0F,0x0E,0x10,0,0x10,0x0F,0x11,0,
    0x11,0x10,0x12,0,0x12,0x11,0x13,0,0x13,0x12,0x14,0,0x14,0x13,0x15,0,
    0x15,0x14,0x16,0,0x16,0x15,0x17,0,0x17,0x16,0x18,0,0x18,0x17,0x19,0,
    0x19,0x18,0x1A,0,0x1A,0x19,0x1B,0,0x1B,0x1A,0x1C,0,0x1C,0x1B,0x1D,0,
    0x1D,0x1C,0x1E,0,0x1E,0x1D,0x1F,0,0x1F,0x1E,0x20,0,0x20,0x1F,0x21,0,
    0xE2,0xE0,0xE2,0,0xE3,0xE1,0xE3,0,0xE4,0xE2,0xE4,0,0xE5,0xE3,0xE5,0,
    0xE6,0xE4,0xE6,0,0xE7,0xE5,0xE7,0,0xE8,0xE6,0xE8,0,0xE9,0xE7,0xE9,0,
    0xEA,0xE8,0xEA,0,0xEB,0xE9,0xEB,0,0xEC,0xEA,0xEC,0,0xED,0xEB,0xED,0,
    0xEE,0xEC,0xEE,0,0xEF,0xED,0xEF,0,0xF0,0xEE,0xF0,0,0xF1,0xEF,0xF1,0,
    0xF2,0xF0,0xF2,0,0xF3,0xF1,0xF3,0,0xF4,0xF2,0xF4,0,0xF5,0xF3,0xF5,0,
    0xF6,0xF4,0xF6,0,0xF7,0xF5,0xF7,0,0xF8,0xF6,0xF8,0,0xF9,0xF7,0xF9,0,
    0xFA,0xF8,0xFA,0,0xFB,0xF9,0xFB,0,0xFC,0xFA,0xFC,0,0xFD,0xFB,0xFD,0,
    0xFE,0xFC,0xFE,0,0xFF,0xFD,0xFF,0,0x00,0xFE,0x00,0,0x01,0xFF,0x01,0,
    0x02,0x00,0x02,0,0x03,0x01,0x03,0,0x04,0x02,0x04,0,0x05,0x03,0x05,0,
    0x06,0x04,0x06,0,0x07,0x05,0x07,0,0x08,0x06,0x08,0,0x09,0x07,0x09,0,
    0x0A,0x08,0x0A,0,0x0B,0x09,0x0B,0,0x0C,0x0A,0x0C,0,0x0D,0x0B,0x0D,0,
    0x0E,0x0C,0x0E,0,0x0F,0x0D,0x0F,0,0x10,0x0E,0x10,0,0x11,0x0F,0x11,0,
    0x12,0x10,0x12,0,0x13,0x11,0x13,0,0x14,0x12,0x14,0,0x15,0x13,0x15,0,
    0x16,0x14,0x16,0,0x17,0x15,0x17,0,0x18,0x16,0x18,0,0x19,0x17,0x19,0,
    0x1A,0x18,0x1A,0,0x1B,0x19,0x1B,0,0x1C,0x1A,0x1C,0,0x1D,0x1B,0x1D,0,
    0x1E,0x1C,0x1E,0,0x1F,0x1D,0x1F,0,0x20,0x1E,0x20,0,0x21,0x1F,0x21,0,
    0xE3,0xE0,0xE2,0,0xE4,0xE1,0xE3,0,0xE5,0xE2,0xE4,0,0xE6,0xE3,0xE5,0,
    0xE7,0xE4,0xE6,0,0xE8,0xE5,0xE7,0,0xE9,0xE6,0xE8,0,0xEA,0xE7,0xE9,0,
    0xEB,0xE8,0xEA,0,0xEC,0xE9,0xEB,0,0xED,0xEA,0xEC,0,0xEE,0xEB,0xED,0,
    0xEF,0xEC,0xEE,0,0xF0,0xED,0xEF,0,0xF1,0xEE,0xF0,0,0xF2,0xEF,0xF1,0,
    0xF3,0xF0,0xF2,0,0xF4,0xF1,0xF3,0,0xF5,0xF2,0xF4,0,0xF6,0xF3,0xF5,0,
    0xF7,0xF4,0xF6,0,0xF8,0xF5,0xF7,0,0xF9,0xF6,0xF8,0,0xFA,0xF7,0xF9,0,
    0xFB,0xF8,0xFA,0,0xFC,0xF9,0xFB,0,0xFD,0xFA,0xFC,0,0xFE,0xFB,0xFD,0,
    0xFF,0xFC,0xFE,0,0x00,0xFD,0xFF,0,0x01,0xFE,0x00,0,0x02,0xFF,0x01,0,
    0x03,0x00,0x02,0,0x04,0x01,0x03,0,0x05,0x02,0x04,0,0x06,0x03,0x05,0,
    0x07,0x04,0x06,0,0x08,0x05,0x07,0,0x09,0x06,0x08,0,0x0A,0x07,0x09,0,
    0x0B,0x08,0x0A,0,0x0C,0x09,0x0B,0,0x0D,0x0A,0x0C,0,0x0E,0x0B,0x0D,0,
    0x0F,0x0C,0x0E,0,0x10,0x0D,0x0F,0,0x11,0x0E,0x10,0,0x12,0x0F,0x11,0,
    0x13,0x10,0x12,0,0x14,0x11,0x13,0,0x15,0x12,0x14,0,0x16,0x13,0x15,0,
    0x17,0x14,0x16,0,0x18,0x15,0x17,0,0x19,0x16,0x18,0,0x1A,0x17,0x19,0,
    0x1B,0x18,0x1A,0,0x1C,0x19,0x1B,0,0x1D,0x1A,0x1C,0,0x1E,0x1B,0x1D,0,
    0x1F,0x1C,0x1E,0,0x20,0x1D,0x1F,0,0x21,0x1E,0x20,0,0x22,0x1F,0x21,0,
    0xE4,0xE0,0xE2,0,0xE5,0xE1,0xE3,0,0xE6,0xE2,0xE4,0,0xE7,0xE3,0xE5,0,
    0xE8,0xE4,0xE6,0,0xE9,0xE5,0xE7,0,0xEA,0xE6,0xE8,0,0xEB,0xE7,0xE9,0,
    0xEC,0xE8,0xEA,0,0xED,0xE9,0xEB,0,0xEE,0xEA,0xEC,0,0xEF,0xEB,0xED,0,
    0xF0,0xEC,0xEE,0,0xF1,0xED,0xEF,0,0xF2,0xEE,0xF0,0,0xF3,0xEF,0xF1,0,
    0xF4,0xF0,0xF2,0,0xF5,0xF1,0xF3,0,0xF6,0xF2,0xF4,0,0xF7,0xF3,0xF5,0,
    0xF8,0xF4,0xF6,0,0xF9,0xF5,0xF7,0,0xFA,0xF6,0xF8,0,0xFB,0xF7,0xF9,0,
    0xFC,0xF8,0xFA,0,0xFD,0xF9,0xFB,0,0xFE,0xFA,0xFC,0,0xFF,0xFB,0xFD,0,
    0x00,0xFC,0xFE,0,0x01,0xFD,0xFF,0,0x02,0xFE,0x00,0,0x03,0xFF,0x01,0,
    0x04,0x00,0x02,0,0x05,0x01,0x03,0,0x06,0x02,0x04,0,0x07,0x03,0x05,0,
    0x08,0x04,0x06,0,0x09,0x05,0x07,0,0x0A,0x06,0x08,0,0x0B,0x07,0x09,0,
    0x0C,0x08,0x0A,0,0x0D,0x09,0x0B,0,0x0E,0x0A,0x0C,0,0x0F,0x0B,0x0D,0,
    0x10,0x0C,0x0E,0,0x11,0x0D,0x0F,0,0x12,0x0E,0x10,0,0x13,0x0F,0x11,0,
    0x14,0x10,0x12,0,0x15,0x11,0x13,0,0x16,0x12,0x14,0,0x17,0x13,0x15,0,
    0x18,0x14,0x16,0,0x19,0x15,0x17,0,0x1A,0x16,0x18,0,0x1B,0x17,0x19,0,
    0x1C,0x18,0x1A,0,0x1D,0x19,0x1B,0,0x1E,0x1A,0x1C,0,0x1F,0x1B,0x1D,0,
    0x20,0x1C,0x1E,0,0x21,0x1D,0x1F,0,0x22,0x1E,0x20,0,0x23,0x1F,0x21,0,
    0xE5,0xE0,0xE2,0,0xE6,0xE1,0xE3,0,0xE7,0xE2,0xE4,0,0xE8,0xE3,0xE5,0,
    0xE9,0xE4,0xE6,0,0xEA,0xE5,0xE7,0,0xEB,0xE6,0xE8,0,0xEC,0xE7,0xE9,0,
    0xED,0xE8,0xEA,0,0xEE,0xE9,0xEB,0,0xEF,0xEA,0xEC,0,0xF0,0xEB,0xED,0,
    0xF1,0xEC,0xEE,0,0xF2,0xED,0xEF,0,0xF3,0xEE,0xF0,0,0xF4,0xEF,0xF1,0,
    0xF5,0xF0,0xF2,0,0xF6,0xF1,0xF3,0,0xF7,0xF2,0xF4,0,0xF8,0xF3,0xF5,0,
    0xF9,0xF4,0xF6,0,0xFA,0xF5,0xF7,0,0xFB,0xF6,0xF8,0,0xFC,0xF7,0xF9,0,
    0xFD,0xF8,0xFA,0,0xFE,0xF9,0xFB,0,0xFF,0xFA,0xFC,0,0x00,0xFB,0xFD,0,
    0x01,0xFC,0xFE,0,0x02,0xFD,0xFF,0,0x03,0xFE,0x00,0,0x04,0xFF,0x01,0,
    0x05,0x00,0x02,0,0x06,0x01,0x03,0,0x07,0x02,0x04,0,0x08,0x03,0x05,0,
    0x09,0x04,0x06,0,0x0A,0x05,0x07,0,0x0B,0x06,0x08,0,0x0C,0x07,0x09,0,
    0x0D,0x08,0x0A,0,0x0E,0x09,0x0B,0,0x0F,0x0A,0x0C,0,0x10,0x0B,0x0D,0,
    0x11,0x0C,0x0E,0,0x12,0x0D,0x0F,0,0x13,0x0E,0x10,0,0x14,0x0F,0x11,0,
    0x15,0x10,0x12,0,0x16,0x11,0x13,0,0x17,0x12,0x14,0,0x18,0x13,0x15,0,
    0x19,0x14,0x16,0,0x1A,0x15,0x17,0,0x1B,0x16,0x18,0,0x1C,0x17,0x19,0,
    0x1D,0x18,0x1A,0,0x1E,0x19,0x1B,0,0x1F,0x1A,0x1C,0,0x20,0x1B,0x1D,0,
    0x21,0x1C,0x1E,0,0x22,0x1D,0x1F,0,0x23,0x1E,0x20,0,0x24,0x1F,0x21,0,
    0xE6,0xE0,0xE2,0,0xE7,0xE1,0xE3,0,0xE8,0xE2,0xE4,0,0xE9,0xE3,0xE5,0,
    0xEA,0xE4,0xE6,0,0xEB,0xE5,0xE7,0,0xEC,0xE6,0xE8,0,0xED,0xE7,0xE9,0,
    0xEE,0xE8,0xEA,0,0xEF,0xE9,0xEB,0,0xF0,0xEA,0xEC,0,0xF1,0xEB,0xED,0,
    0xF2,0xEC,0xEE,0,0xF3,0xED,0xEF,0,0xF4,0xEE,0xF0,0,0xF5,0xEF,0xF1,0,
    0xF6,0xF0,0xF2,0,0xF7,0xF1,0xF3,0,0xF8,0xF2,0xF4,0,0xF9,0xF3,0xF5,0,
    0xFA,0xF4,0xF6,0,0xFB,0xF5,0xF7,0,0xFC,0xF6,0xF8,0,0xFD,0xF7,0xF9,0,
    0xFE,0xF8,0xFA,0,0xFF,0xF9,0xFB,0,0x00,0xFA,0xFC,0,0x01,0xFB,0xFD,0,
    0x02,0xFC,0xFE,0,0x03,0xFD,0xFF,0,0x04,0xFE,0x00,0,0x05,0xFF,0x01,0,
    0x06,0x00,0x02,0,0x07,0x01,0x03,0,0x08,0x02,0x04,0,0x09,0x03,0x05,0,
    0x0A,0x04,0x06,0,0x0B,0x05,0x07,0,0x0C,0x06,0x08,0,0x0D,0x07,0x09,0,
    0x0E,0x08,0x0A,0,0x0F,0x09,0x0B,0,0x10,0x0A,0x0C,0,0x11,0x0B,0x0D,0,
    0x12,0x0C,0x0E,0,0x13,0x0D,0x0F,0,0x14,0x0E,0x10,0,0x15,0x0F,0x11,0,
    0x16,0x10,0x12,0,0x17,0x11,0x13,0,0x18,0x12,0x14,0,0x19,0x13,0x15,0,
    0x1A,0x14,0x16,0,0x1B,0x15,0x17,0,0x1C,0x16,0x18,0,0x1D,0x17,0x19,0,
    0x1E,0x18,0x1A,0,0x1F,0x19,0x1B,0,0x20,0x1A,0x1C,0,0x21,0x1B,0x1D,0,
    0x22,0x1C,0x1E,0,0x23,0x1D,0x1F,0,0x24,0x1E,0x20,0,0x25,0x1F,0x21,0,
    0xE7,0xE0,0xE2,0,0xE8,0xE1,0xE3,0,0xE9,0xE2,0xE4,0,0xEA,0xE3,0xE5,0,
    0xEB,0xE4,0xE6,0,0xEC,0xE5,0xE7,0,0xED,0xE6,0xE8,0,0xEE,0xE7,0xE9,0,
    0xEF,0xE8,0xEA,0,0xF0,0xE9,0xEB,0,0xF1,0xEA,0xEC,0,0xF2,0xEB,0xED,0,
    0xF3,0xEC,0xEE,0,0xF4,0xED,0xEF,0,0xF5,0xEE,0xF0,0,0xF6,0xEF,0xF1,0,
    0xF7,0xF0,0xF2,0,0xF8,0xF1,0xF3,0,0xF9,0xF2,0xF4,0,0xFA,0xF3,0xF5,0,
    0xFB,0xF4,0xF6,0,0xFC,0xF5,0xF7,0,0xFD,0xF6,0xF8,0,0xFE,0xF7,0xF9,0,
    0xFF,0xF8,0xFA,0,0x00,0xF9,0xFB,0,0x01,0xFA,0xFC,0,0x02,0xFB,0xFD,0,
    0x03,0xFC,0xFE,0,0x04,0xFD,0xFF,0,0x05,0xFE,0x00,0,0x06,0xFF,0x01,0,
    0x07,0x00,0x02,0,0x08,0x01,0x03,0,0x09,0x02,0x04,0,0x0A,0x03,0x05,0,
    0x0B,0x04,0x06,0,0x0C,0x05,0x07,0,0x0D,0x06,0x08,0,0x0E,0x07,0x09,0,
    0x0F,0x08,0x0A,0,0x10,0x09,0x0B,0,0x11,0x0A,0x0C,0,0x12,0x0B,0x0D,0,
    0x13,0x0C,0x0E,0,0x14,0x0D,0x0F,0,0x15,0x0E,0x10,0,0x16,0x0F,0x11,0,
    0x17,0x10,0x12,0,0x18,0x11,0x13,0,0x19,0x12,0x14,0,0x1A,0x13,0x15,0,
    0x1B,0x14,0x16,0,0x1C,0x15,0x17,0,0x1D,0x16,0x18,0,0x1E,0x17,0x19,0,
    0x1F,0x18,0x1A,0,0x20,0x19,0x1B,0,0x21,0x1A,0x1C,0,0x22,0x1B,0x1D,0,
    0x23,0x1C,0x1E,0,0x24,0x1D,0x1F,0,0x25,0x1E,0x20,0,0x26,0x1F,0x21,0,
    0xD8,0xE0,0xE3,0,0xD9,0xE1,0xE4,0,0xDA,0xE2,0xE5,0,0xDB,0xE3,0xE6,0,
    0xDC,0xE4,0xE7,0,0xDD,0xE5,0xE8,0,0xDE,0xE6,0xE9,0,0xDF,0xE7,0xEA,0,
    0xE0,0xE8,0xEB,0,0xE1,0xE9,0xEC,0,0xE2,0xEA,0xED,0,0xE3,0xEB,0xEE,0,
    0xE4,0xEC,0xEF,0,0xE5,0xED,0xF0,0,0xE6,0xEE,0xF1,0,0xE7,0xEF,0xF2,0,
    0xE8,0xF0,0xF3,0,0xE9,0xF1,0xF4,0,0xEA,0xF2,0xF5,0,0xEB,0xF3,0xF6,0,
    0xEC,0xF4,0xF7,0,0xED,0xF5,0xF8,0,0xEE,0xF6,0xF9,0,0xEF,0xF7,0xFA,0,
    0xF0,0xF8,0xFB,0,0xF1,0xF9,0xFC,0,0xF2,0xFA,0xFD,0,0xF3,0xFB,0xFE,0,
    0xF4,0xFC,0xFF,0,0xF5,0xFD,0x00,0,0xF6,0xFE,0x01,0,0xF7,0xFF,0x02,0,
    0xF8,0x00,0x03,0,0xF9,0x01,0x04,0,0xFA,0x02,0x05,0,0xFB,0x03,0x06,0,
    0xFC,0x04,0x07,0,0xFD,0x05,0x08,0,0xFE,0x06,0x09,0,0xFF,0x07,0x0A,0,
    0x00,0x08,0x0B,0,0x01,0x09,0x0C,0,0x02,0x0A,0x0D,0,0x03,0x0B,0x0E,0,
    0x04,0x0C,0x0F,0,0x05,0x0D,0x10,0,0x06,0x0E,0x11,0,0x07,0x0F,0x12,0,
    0x08,0x10,0x13,0,0x09,0x11,0x14,0,0x0A,0x12,0x15,0,0x0B,0x13,0x16,0,
    0x0C,0x14,0x17,0,0x0D,0x15,0x18,0,0x0E,0x16,0x19,0,0x0F,0x17,0x1A,0,
    0x10,0x18,0x1B,0,0x11,0x19,0x1C,0,0x12,0x1A,0x1D,0,0x13,0x1B,0x1E,0,
    0x14,0x1C,0x1F,0,0x15,0x1D,0x20,0,0x16,0x1E,0x21,0,0x17,0x1F,0x22,0,
    0xD9,0xE0,0xE3,0,0xDA,0xE1,0xE4,0,0xDB,0xE2,0xE5,0,0xDC,0xE3,0xE6,0,
    0xDD,0xE4,0xE7,0,0xDE,0xE5,0xE8,0,0xDF,0xE6,0xE9,0,0xE0,0xE7,0xEA,0,
    0xE1,0xE8,0xEB,0,0xE2,0xE9,0xEC,0,0xE3,0xEA,0xED,0,0xE4,0xEB,0xEE,0,
    0xE5,0xEC,0xEF,0,0xE6,0xED,0xF0,0,0xE7,0xEE,0xF1,0,0xE8,0xEF,0xF2,0,
    0xE9,0xF0,0xF3,0,0xEA,0xF1,0xF4,0,0xEB,0xF2,0xF5,0,0xEC,0xF3,0xF6,0,
    0xED,0xF4,0xF7,0,0xEE,0xF5,0xF8,0,0xEF,0xF6,0xF9,0,0xF0,0xF7,0xFA,0,
    0xF1,0xF8,0xFB,0,0xF2,0xF9,0xFC,0,0xF3,0xFA,0xFD,0,0xF4,0xFB,0xFE,0,
    0xF5,0xFC,0xFF,0,0xF6,0xFD,0x00,0,0xF7,0xFE,0x01,0,0xF8,0xFF,0x02,0,
    0xF9,0x00,0x03,0,0xFA,0x01,0x04,0,0xFB,0x02,0x05,0,0xFC,0x03,0x06,0,
    0xFD,0x04,0x07,0,0xFE,0x05,0x08,0,0xFF,0x06,0x09,0,0x00,0x07,0x0A,0,
    0x01,0x08,0x0B,0,0x02,0x09,0x0C,0,0x03,0x0A,0x0D,0,0x04,0x0B,0x0E,0,
    0x05,0x0C,0x0F,0,0x06,0x0D,0x10,0,0x07,0x0E,0x11,0,0x08,0x0F,0x12,0,
    0x09,0x10,0x13,0,0x0A,0x11,0x14,0,0x0B,0x12,0x15,0,0x0C,0x13,0x16,0,
    0x0D,0x14,0x17,0,0x0E,0x15,0x18,0,0x0F,0x16,0x19,0,0x10,0x17,0x1A,0,
    0x11,0x18,0x1B,0,0x12,0x19,0x1C,0,0x13,0x1A,0x1D,0,0x14,0x1B,0x1E,0,
    0x15,0x1C,0x1F,0,0x16,0x1D,0x20,0,0x17,0x1E,0x21,0,0x18,0x1F,0x22,0,
    0xDA,0xE0,0xE3,0,0xDB,0xE1,0xE4,0,0xDC,0xE2,0xE5,0,0xDD,0xE3,0xE6,0,
    0xDE,0xE4,0xE7,0,0xDF,0xE5,0xE8,0,0xE0,0xE6,0xE9,0,0xE1,0xE7,0xEA,0,
    0xE2,0xE8,0xEB,0,0xE3,0xE9,0xEC,0,0xE4,0xEA,0xED,0,0xE5,0xEB,0xEE,0,
    0xE6,0xEC,0xEF,0,0xE7,0xED,0xF0,0,0xE8,0xEE,0xF1,0,0xE9,0xEF,0xF2,0,
    0xEA,0xF0,0xF3,0,0xEB,0xF1,0xF4,0,0xEC,0xF2,0xF5,0,0xED,0xF3,0xF6,0,
    0xEE,0xF4,0xF7,0,0xEF,0xF5,0xF8,0,0xF0,0xF6,0xF9,0,0xF1,0xF7,0xFA,0,
    0xF2,0xF8,0xFB,0,0xF3,0xF9,0xFC,0,0xF4,0xFA,0xFD,0,0xF5,0xFB,0xFE,0,
    0xF6,0xFC,0xFF,0,0xF7,0xFD,0x00,0,0xF8,0xFE,0x01,0,0xF9,0xFF,0x02,0,
    0xFA,0x00,0x03,0,0xFB,0x01,0x04,0,0xFC,0x02,0x05,0,0xFD,0x03,0x06,0,
    0xFE,0x04,0x07,0,0xFF,0x05,0x08,0,0x00,0x06,0x09,0,0x01,0x07,0x0A,0,
    0x02,0x08,0x0B,0,0x03,0x09,0x0C,0,0x04,0x0A,0x0D,0,0x05,0x0B,0x0E,0,
    0x06,0x0C,0x0F,0,0x07,0x0D,0x10,0,0x08,0x0E,0x11,0,0x09,0x0F,0x12,0,
    0x0A,0x10,0x13,0,0x0B,0x11,0x14,0,0x0C,0x12,0x15,0,0x0D,0x13,0x16,0,
    0x0E,0x14,0x17,0,0x0F,0x15,0x18,0,0x10,0x16,0x19,0,0x11,0x17,0x1A,0,
    0x12,0x18,0x1B,0,0x13,0x19,0x1C,0,0x14,0x1A,0x1D,0,0x15,0x1B,0x1E,0,
    0x16,0x1C,0x1F,0,0x17,0x1D,0x20,0,0x18,0x1E,0x21,0,0x19,0x1F,0x22,0,
    0xDB,0xE0,0xE3,0,0xDC,0xE1,0xE4,0,0xDD,0xE2,0xE5,0,0xDE,0xE3,0xE6,0,
    0xDF,0xE4,0xE7,0,0xE0,0xE5,0xE8,0,0xE1,0xE6,0xE9,0,0xE2,0xE7,0xEA,0,
    0xE3,0xE8,0xEB,0,0xE4,0xE9,0xEC,0,0xE5,0xEA,0xED,0,0xE6,0xEB,0xEE,0,
    0xE7,0xEC,0xEF,0,0xE8,0xED,0xF0,0,0xE9,0xEE,0xF1,0,0xEA,0xEF,0xF2,0,
    0xEB,0xF0,0xF3,0,0xEC,0xF1,0xF4,0,0xED,0xF2,0xF5,0,0xEE,0xF3,0xF6,0,
    0xEF,0xF4,0xF7,0,0xF0,0xF5,0xF8,0,0xF1,0xF6,0xF9,0,0xF2,0xF7,0xFA,0,
    0xF3,0xF8,0xFB,0,0xF4,0xF9,0xFC,0,0xF5,0xFA,0xFD,0,0xF6,0xFB,0xFE,0,
    0xF7,0xFC,0xFF,0,0xF8,0xFD,0x00,0,0xF9,0xFE,0x01,0,0xFA,0xFF,0x02,0,
    0xFB,0x00,0x03,0,0xFC,0x01,0x04,0,0xFD,0x02,0x05,0,0xFE,0x03,0x06,0,
    0xFF,0x04,0x07,0,0x00,0x05,0x08,0,0x01,0x06,0x09,0,0x02,0x07,0x0A,0,
    0x03,0x08,0x0B,0,0x04,0x09,0x0C,0,0x05,0x0A,0x0D,0,0x06,0x0B,0x0E,0,
    0x07,0x0C,0x0F,0,0x08,0x0D,0x10,0,0x09,0x0E,0x11,0,0x0A,0x0F,0x12,0,
    0x0B,0x10,0x13,0,0x0C,0x11,0x14,0,0x0D,0x12,0x15,0,0x0E,0x13,0x16,0,
    0x0F,0x14,0x17,0,0x10,0x15,0x18,0,0x11,0x16,0x19,0,0x12,0x17,0x1A,0,
    0x13,0x18,0x1B,0,0x14,0x19,0x1C,0,0x15,0x1A,0x1D,0,0x16,0x1B,0x1E,0,
    0x17,0x1C,0x1F,0,0x18,0x1D,0x20,0,0x19,0x1E,0x21,0,0x1A,0x1F,0x22,0,
    0xDC,0xE0,0xE3,0,0xDD,0xE1,0xE4,0,0xDE,0xE2,0xE5,0,0xDF,0xE3,0xE6,0,
    0xE0,0xE4,0xE7,0,0xE1,0xE5,0xE8,0,0xE2,0xE6,0xE9,0,0xE3,0xE7,0xEA,0,
    0xE4,0xE8,0xEB,0,0xE5,0xE9,0xEC,0,0xE6,0xEA,0xED,0,0xE7,0xEB,0xEE,0,
    0xE8,0xEC,0xEF,0,0xE9,0xED,0xF0,0,0xEA,0xEE,0xF1,0,0xEB,0xEF,0xF2,0,
    0xEC,0xF0,0xF3,0,0xED,0xF1,0xF4,0,0xEE,0xF2,0xF5,0,0xEF,0xF3,0xF6,0,
    0xF0,0xF4,0xF7,0,0xF1,0xF5,0xF8,0,0xF2,0xF6,0xF9,0,0xF3,0xF7,0xFA,0,
    0xF4,0xF8,0xFB,0,0xF5,0xF9,0xFC,0,0xF6,0xFA,0xFD,0,0xF7,0xFB,0xFE,0,
    0xF8,0xFC,0xFF,0,0xF9,0xFD,0x00,0,0xFA,0xFE,0x01,0,0xFB,0xFF,0x02,0,
    0xFC,0x00,0x03,0,0xFD,0x01,0x04,0,0xFE,0x02,0x05,0,0xFF,0x03,0x06,0,
    0x00,0x04,0x07,0,0x01,0x05,0x08,0,0x02,0x06,0x09,0,0x03,0x07,0x0A,0,
    0x04,0x08,0x0B,0,0x05,0x09,0x0C,0,0x06,0x0A,0x0D,0,0x07,0x0B,0x0E,0,
    0x08,0x0C,0x0F,0,0x09,0x0D,0x10,0,0x0A,0x0E,0x11,0,0x0B,0x0F,0x12,0,
    0x0C,0x10,0x13,0,0x0D,0x11,0x14,0,0x0E,0x12,0x15,0,0x0F,0x13,0x16,0,
    0x10,0x14,0x17,0,0x11,0x15,0x18,0,0x12,0x16,0x19,0,0x13,0x17,0x1A,0,
    0x14,0x18,0x1B,0,0x15,0x19,0x1C,0,0x16,0x1A,0x1D,0,0x17,0x1B,0x1E,0,
    0x18,0x1C,0x1F,0,0x19,0x1D,0x20,0,0x1A,0x1E,0x21,0,0x1B,0x1F,0x22,0,
    0xDD,0xE0,0xE3,0,0xDE,0xE1,0xE4,0,0xDF,0xE2,0xE5,0,0xE0,0xE3,0xE6,0,
    0xE1,0xE4,0xE7,0,0xE2,0xE5,0xE8,0,0xE3,0xE6,0xE9,0,0xE4,0xE7,0xEA,0,
    0xE5,0xE8,0xEB,0,0xE6,0xE9,0xEC,0,0xE7,0xEA,0xED,0,0xE8,0xEB,0xEE,0,
    0xE9,0xEC,0xEF,0,0xEA,0xED,0xF0,0,0xEB,0xEE,0xF1,0,0xEC,0xEF,0xF2,0,
    0xED,0xF0,0xF3,0,0xEE,0xF1,0xF4,0,0xEF,0xF2,0xF5,0,0xF0,0xF3,0xF6,0,
    0xF1,0xF4,0xF7,0,0xF2,0xF5,0xF8,0,0xF3,0xF6,0xF9,0,0xF4,0xF7,0xFA,0,
    0xF5,0xF8,0xFB,0,0xF6,0xF9,0xFC,0,0xF7,0xFA,0xFD,0,0xF8,0xFB,0xFE,0,
    0xF9,0xFC,0xFF,0,0xFA,0xFD,0x00,0,0xFB,0xFE,0x01,0,0xFC,0xFF,0x02,0,
    0xFD,0x00,0x03,0,0xFE,0x01,0x04,0,0xFF,0x02,0x05,0,0x00,0x03,0x06,0,
    0x01,0x04,0x07,0,0x02,0x05,0x08,0,0x03,0x06,0x09,0,0x04,0x07,0x0A,0,
    0x05,0x08,0x0B,0,0x06,0x09,0x0C,0,0x07,0x0A,0x0D,0,0x08,0x0B,0x0E,0,
    0x09,0x0C,0x0F,0,0x0A,0x0D,0x10,0,0x0B,0x0E,0x11,0,0x0C,0x0F,0x12,0,
    0x0D,0x10,0x13,0,0x0E,0x11,0x14,0,0x0F,0x12,0x15,0,0x10,0x13,0x16,0,
    0x11,0x14,0x17,0,0x12,0x15,0x18,0,0x13,0x16,0x19,0,0x14,0x17,0x1A,0,
    0x15,0x18,0x1B,0,0x16,0x19,0x1C,0,0x17,0x1A,0x1D,0,0x18,0x1B,0x1E,0,
    0x19,0x1C,0x1F,0,0x1A,0x1D,0x20,0,0x1B,0x1E,0x21,0,0x1C,0x1F,0x22,0,
    0xDE,0xE0,0xE3,0,0xDF,0xE1,0xE4,0,0xE0,0xE2,0xE5,0,0xE1,0xE3,0xE6,0,
    0xE2,0xE4,0xE7,0,0xE3,0xE5,0xE8,0,0xE4,0xE6,0xE9,0,0xE5,0xE7,0xEA,0,
    0xE6,0xE8,0xEB,0,0xE7,0xE9,0xEC,0,0xE8,0xEA,0xED,0,0xE9,0xEB,0xEE,0,
    0xEA,0xEC,0xEF,0,0xEB,0xED,0xF0,0,0xEC,0xEE,0xF1,0,0xED,0xEF,0xF2,0,
    0xEE,0xF0,0xF3,0,0xEF,0xF1,0xF4,0,0xF0,0xF2,0xF5,0,0xF1,0xF3,0xF6,0,
    0xF2,0xF4,0xF7,0,0xF3,0xF5,0xF8,0,0xF4,0xF6,0xF9,0,0xF5,0xF7,0xFA,0,
    0xF6,0xF8,0xFB,0,0xF7,0xF9,0xFC,0,0xF8,0xFA,0xFD,0,0xF9,0xFB,0xFE,0,
    0xFA,0xFC,0xFF,0,0xFB,0xFD,0x00,0,0xFC,0xFE,0x01,0,0xFD,0xFF,0x02,0,
    0xFE,0x00,0x03,0,0xFF,0x01,0x04,0,0x00,0x02,0x05,0,0x01,0x03,0x06,0,
    0x02,0x04,0x07,0,0x03,0x05,0x08,0,0x04,0x06,0x09,0,0x05,0x07,0x0A,0,
    0x06,0x08,0x0B,0,0x07,0x09,0x0C,0,0x08,0x0A,0x0D,0,0x09,0x0B,0x0E,0,
    0x0A,0x0C,0x0F,0,0x0B,0x0D,0x10,0,0x0C,0x0E,0x11,0,0x0D,0x0F,0x12,0,
    0x0E,0x10,0x13,0,0x0F,0x11,0x14,0,0x10,0x12,0x15,0,0x11,0x13,0x16,0,
    0x12,0x14,0x17,0,0x13,0x15,0x18,0,0x14,0x16,0x19,0,0x15,0x17,0x1A,0,
    0x16,0x18,0x1B,0,0x17,0x19,0x1C,0,0x18,0x1A,0x1D,0,0x19,0x1B,0x1E,0,
    0x1A,0x1C,0x1F,0,0x1B,0x1D,0x20,0,0x1C,0x1E,0x21,0,0x1D,0x1F,0x22,0,
    0xDF,0xE0,0xE3,0,0xE0,0xE1,0xE4,0,0xE1,0xE2,0xE5,0,0xE2,0xE3,0xE6,0,
    0xE3,0xE4,0xE7,0,0xE4,0xE5,0xE8,0,0xE5,0xE6,0xE9,0,0xE6,0xE7,0xEA,0,
    0xE7,0xE8,0xEB,0,0xE8,0xE9,0xEC,0,0xE9,0xEA,0xED,0,0xEA,0xEB,0xEE,0,
    0xEB,0xEC,0xEF,0,0xEC,0xED,0xF0,0,0xED,0xEE,0xF1,0,0xEE,0xEF,0xF2,0,
    0xEF,0xF0,0xF3,0,0xF0,0xF1,0xF4,0,0xF1,0xF2,0xF5,0,0xF2,0xF3,0xF6,0,
    0xF3,0xF4,0xF7,0,0xF4,0xF5,0xF8,0,0xF5,0xF6,0xF9,0,0xF6,0xF7,0xFA,0,
    0xF7,0xF8,0xFB,0,0xF8,0xF9,0xFC,0,0xF9,0xFA,0xFD,0,0xFA,0xFB,0xFE,0,
    0xFB,0xFC,0xFF,0,0xFC,0xFD,0x00,0,0xFD,0xFE,0x01,0,0xFE,0xFF,0x02,0,
    0xFF,0x00,0x03,0,0x00,0x01,0x04,0,0x01,0x02,0x05,0,0x02,0x03,0x06,0,
    0x03,0x04,0x07,0,0x04,0x05,0x08,0,0x05,0x06,0x09,0,0x06,0x07,0x0A,0,
    0x07,0x08,0x0B,0,0x08,0x09,0x0C,0,0x09,0x0A,0x0D,0,0x0A,0x0B,0x0E,0,
    0x0B,0x0C,0x0F,0,0x0C,0x0D,0x10,0,0x0D,0x0E,0x11,0,0x0E,0x0F,0x12,0,
    0x0F,0x10,0x13,0,0x10,0x11,0x14,0,0x11,0x12,0x15,0,0x12,0x13,0x16,0,
    0x13,0x14,0x17,0,0x14,0x15,0x18,0,0x15,0x16,0x19,0,0x16,0x17,0x1A,0,
    0x17,0x18,0x1B,0,0x18,0x19,0x1C,0,0x19,0x1A,0x1D,0,0x1A,0x1B,0x1E,0,
    0x1B,0x1C,0x1F,0,0x1C,0x1D,0x20,0,0x1D,0x1E,0x21,0,0x1E,0x1F,0x22,0,
    0xE0,0xE0,0xE3,0,0xE1,0xE1,0xE4,0,0xE2,0xE2,0xE5,0,0xE3,0xE3,0xE6,0,
    0xE4,0xE4,0xE7,0,0xE5,0xE5,0xE8,0,0xE6,0xE6,0xE9,0,0xE7,0xE7,0xEA,0,
    0xE8,0xE8,0xEB,0,0xE9,0xE9,0xEC,0,0xEA,0xEA,0xED,0,0xEB,0xEB,0xEE,0,
    0xEC,0xEC,0xEF,0,0xED,0xED,0xF0,0,0xEE,0xEE,0xF1,0,0xEF,0xEF,0xF2,0,
    0xF0,0xF0,0xF3,0,0xF1,0xF1,0xF4,0,0xF2,0xF2,0xF5,0,0xF3,0xF3,0xF6,0,
    0xF4,0xF4,0xF7,0,0xF5,0xF5,0xF8,0,0xF6,0xF6,0xF9,0,0xF7,0xF7,0xFA,0,
    0xF8,0xF8,0xFB,0,0xF9,0xF9,0xFC,0,0xFA,0xFA,0xFD,0,0xFB,0xFB,0xFE,0,
    0xFC,0xFC,0xFF,0,0xFD,0xFD,0x00,0,0xFE,0xFE,0x01,0,0xFF,0xFF,0x02,0,
    0x00,0x00,0x03,0,0x01,0x01,0x04,0,0x02,0x02,0x05,0,0x03,0x03,0x06,0,
    0x04,0x04,0x07,0,0x05,0x05,0x08,0,0x06,0x06,0x09,0,0x07,0x07,0x0A,0,
    0x08,0x08,0x0B,0,0x09,0x09,0x0C,0,0x0A,0x0A,0x0D,0,0x0B,0x0B,0x0E,0,
    0x0C,0x0C,0x0F,0,0x0D,0x0D,0x10,0,0x0E,0x0E,0x11,0,0x0F,0x0F,0x12,0,
    0x10,0x10,0x13,0,0x11,0x11,0x14,0,0x12,0x12,0x15,0,0x13,0x13,0x16,0,
    0x14,0x14,0x17,0,0x15,0x15,0x18,0,0x16,0x16,0x19,0,0x17,0x17,0x1A,0,
    0x18,0x18,0x1B,0,0x19,0x19,0x1C,0,0x1A,0x1A,0x1D,0,0x1B,0x1B,0x1E,0,
    0x1C,0x1C,0x1F,0,0x1D,0x1D,0x20,0,0x1E,0x1E,0x21,0,0x1F,0x1F,0x22,0,
    0xE1,0xE0,0xE3,0,0xE2,0xE1,0xE4,0,0xE3,0xE2,0xE5,0,0xE4,0xE3,0xE6,0,
    0xE5,0xE4,0xE7,0,0xE6,0xE5,0xE8,0,0xE7,0xE6,0xE9,0,0xE8,0xE7,0xEA,0,
    0xE9,0xE8,0xEB,0,0xEA,0xE9,0xEC,0,0xEB,0xEA,0xED,0,0xEC,0xEB,0xEE,0,
    0xED,0xEC,0xEF,0,0xEE,0xED,0xF0,0,0xEF,0xEE,0xF1,0,0xF0,0xEF,0xF2,0,
    0xF1,0xF0,0xF3,0,0xF2,0xF1,0xF4,0,0xF3,0xF2,0xF5,0,0xF4,0xF3,0xF6,0,
    0xF5,0xF4,0xF7,0,0xF6,0xF5,0xF8,0,0xF7,0xF6,0xF9,0,0xF8,0xF7,0xFA,0,
    0xF9,0xF8,0xFB,0,0xFA,0xF9,0xFC,0,0xFB,0xFA,0xFD,0,0xFC,0xFB,0xFE,0,
    0xFD,0xFC,0xFF,0,0xFE,0xFD,0x00,0,0xFF,0xFE,0x01,0,0x00,0xFF,0x02,0,
    0x01,0x00,0x03,0,0x02,0x01,0x04,0,0x03,0x02,0x05,0,0x04,0x03,0x06,0,
    0x05,0x04,0x07,0,0x06,0x05,0x08,0,0x07,0x06,0x09,0,0x08,0x07,0x0A,0,
    0x09,0x08,0x0B,0,0x0A,0x09,0x0C,0,0x0B,0x0A,0x0D,0,0x0C,0x0B,0x0E,0,
    0x0D,0x0C,0x0F,0,0x0E,0x0D,0x10,0,0x0F,0x0E,0x11,0,0x10,0x0F,0x12,0,
    0x11,0x10,0x13,0,0x12,0x11,0x14,0,0x13,0x12,0x15,0,0x14,0x13,0x16,0,
    0x15,0x14,0x17,0,0x16,0x15,0x18,0,0x17,0x16,0x19,0,0x18,0x17,0x1A,0,
    0x19,0x18,0x1B,0,0x1A,0x19,0x1C,0,0x1B,0x1A,0x1D,0,0x1C,0x1B,0x1E,0,
    0x1D,0x1C,0x1F,0,0x1E,0x1D,0x20,0,0x1F,0x1E,0x21,0,0x20,0x1F,0x22,0,
    0xE2,0xE0,0xE3,0,0xE3,0xE1,0xE4,0,0xE4,0xE2,0xE5,0,0xE5,0xE3,0xE6,0,
    0xE6,0xE4,0xE7,0,0xE7,0xE5,0xE8,0,0xE8,0xE6,0xE9,0,0xE9,0xE7,0xEA,0,
    0xEA,0xE8,0xEB,0,0xEB,0xE9,0xEC,0,0xEC,0xEA,0xED,0,0xED,0xEB,0xEE,0,
    0xEE,0xEC,0xEF,0,0xEF,0xED,0xF0,0,0xF0,0xEE,0xF1,0,0xF1,0xEF,0xF2,0,
    0xF2,0xF0,0xF3,0,0xF3,0xF1,0xF4,0,0xF4,0xF2,0xF5,0,0xF5,0xF3,0xF6,0,
    0xF6,0xF4,0xF7,0,0xF7,0xF5,0xF8,0,0xF8,0xF6,0xF9,0,0xF9,0xF7,0xFA,0,
    0xFA,0xF8,0xFB,0,0xFB,0xF9,0xFC,0,0xFC,0xFA,0xFD,0,0xFD,0xFB,0xFE,0,
    0xFE,0xFC,0xFF,0,0xFF,0xFD,0x00,0,0x00,0xFE,0x01,0,0x01,0xFF,0x02,0,
    0x02,0x00,0x03,0,0x03,0x01,0x04,0,0x04,0x02,0x05,0,0x05,0x03,0x06,0,
    0x06,0x04,0x07,0,0x07,0x05,0x08,0,0x08,0x06,0x09,0,0x09,0x07,0x0A,0,
    0x0A,0x08,0x0B,0,0x0B,0x09,0x0C,0,0x0C,0x0A,0x0D,0,0x0D,0x0B,0x0E,0,
    0x0E,0x0C,0x0F,0,0x0F,0x0D,0x10,0,0x10,0x0E,0x11,0,0x11,0x0F,0x12,0,
    0x12,0x10,0x13,0,0x13,0x11,0x14,0,0x14,0x12,0x15,0,0x15,0x13,0x16,0,
    0x16,0x14,0x17,0,0x17,0x15,0x18,0,0x18,0x16,0x19,0,0x19,0x17,0x1A,0,
    0x1A,0x18,0x1B,0,0x1B,0x19,0x1C,0,0x1C,0x1A,0x1D,0,0x1D,0x1B,0x1E,0,
    0x1E,0x1C,0x1F,0,0x1F,0x1D,0x20,0,0x20,0x1E,0x21,0,0x21,0x1F,0x22,0,
    0xE3,0xE0,0xE3,0,0xE4,0xE1,0xE4,0,0xE5,0xE2,0xE5,0,0xE6,0xE3,0xE6,0,
    0xE7,0xE4,0xE7,0,0xE8,0xE5,0xE8,0,0xE9,0xE6,0xE9,0,0xEA,0xE7,0xEA,0,
    0xEB,0xE8,0xEB,0,0xEC,0xE9,0xEC,0,0xED,0xEA,0xED,0,0xEE,0xEB,0xEE,0,
    0xEF,0xEC,0xEF,0,0xF0,0xED,0xF0,0,0xF1,0xEE,0xF1,0,0xF2,0xEF,0xF2,0,
    0xF3,0xF0,0xF3,0,0xF4,0xF1,0xF4,0,0xF5,0xF2,0xF5,0,0xF6,0xF3,0xF6,0,
    0xF7,0xF4,0xF7,0,0xF8,0xF5,0xF8,0,0xF9,0xF6,0xF9,0,0xFA,0xF7,0xFA,0,
    0xFB,0xF8,0xFB,0,0xFC,0xF9,0xFC,0,0xFD,0xFA,0xFD,0,0xFE,0xFB,0xFE,0,
    0xFF,0xFC,0xFF,0,0x00,0xFD,0x00,0,0x01,0xFE,0x01,0,0x02,0xFF,0x02,0,
    0x03,0x00,0x03,0,0x04,0x01,0x04,0,0x05,0x02,0x05,0,0x06,0x03,0x06,0,
    0x07,0x04,0x07,0,0x08,0x05,0x08,0,0x09,0x06,0x09,0,0x0A,0x07,0x0A,0,
    0x0B,0x08,0x0B,0,0x0C,0x09,0x0C,0,0x0D,0x0A,0x0D,0,0x0E,0x0B,0x0E,0,
    0x0F,0x0C,0x0F,0,0x10,0x0D,0x10,0,0x11,0x0E,0x11,0,0x12,0x0F,0x12,0,
    0x13,0x10,0x13,0,0x14,0x11,0x14,0,0x15,0x12,0x15,0,0x16,0x13,0x16,0,
    0x17,0x14,0x17,0,0x18,0x15,0x18,0,0x19,0x16,0x19,0,0x1A,0x17,0x1A,0,
    0x1B,0x18,0x1B,0,0x1C,0x19,0x1C,0,0x1D,0x1A,0x1D,0,0x1E,0x1B,0x1E,0,
    0x1F,0x1C,0x1F,0,0x20,0x1D,0x20,0,0x21,0x1E,0x21,0,0x22,0x1F,0x22,0,
    0xE4,0xE0,0xE3,0,0xE5,0xE1,0xE4,0,0xE6,0xE2,0xE5,0,0xE7,0xE3,0xE6,0,
    0xE8,0xE4,0xE7,0,0xE9,0xE5,0xE8,0,0xEA,0xE6,0xE9,0,0xEB,0xE7,0xEA,0,
    0xEC,0xE8,0xEB,0,0xED,0xE9,0xEC,0,0xEE,0xEA,0xED,0,0xEF,0xEB,0xEE,0,
    0xF0,0xEC,0xEF,0,0xF1,0xED,0xF0,0,0xF2,0xEE,0xF1,0,0xF3,0xEF,0xF2,0,
    0xF4,0xF0,0xF3,0,0xF5,0xF1,0xF4,0,0xF6,0xF2,0xF5,0,0xF7,0xF3,0xF6,0,
    0xF8,0xF4,0xF7,0,0xF9,0xF5,0xF8,0,0xFA,0xF6,0xF9,0,0xFB,0xF7,0xFA,0,
    0xFC,0xF8,0xFB,0,0xFD,0xF9,0xFC,0,0xFE,0xFA,0xFD,0,0xFF,0xFB,0xFE,0,
    0x00,0xFC,0xFF,0,0x01,0xFD,0x00,0,0x02,0xFE,0x01,0,0x03,0xFF,0x02,0,
    0x04,0x00,0x03,0,0x05,0x01,0x04,0,0x06,0x02,0x05,0,0x07,0x03,0x06,0,
    0x08,0x04,0x07,0,0x09,0x05,0x08,0,0x0A,0x06,0x09,0,0x0B,0x07,0x0A,0,
    0x0C,0x08,0x0B,0,0x0D,0x09,0x0C,0,0x0E,0x0A,0x0D,0,0x0F,0x0B,0x0E,0,
    0x10,0x0C,0x0F,0,0x11,0x0D,0x10,0,0x12,0x0E,0x11,0,0x13,0x0F,0x12,0,
    0x14,0x10,0x13,0,0x15,0x11,0x14,0,0x16,0x12,0x15,0,0x17,0x13,0x16,0,
    0x18,0x14,0x17,0,0x19,0x15,0x18,0,0x1A,0x16,0x19,0,0x1B,0x17,0x1A,0,
    0x1C,0x18,0x1B,0,0x1D,0x19,0x1C,0,0x1E,0x1A,0x1D,0,0x1F,0x1B,0x1E,0,
    0x20,0x1C,0x1F,0,0x21,0x1D,0x20,0,0x22,0x1E,0x21,0,0x23,0x1F,0x22,0,
    0xE5,0xE0,0xE3,0,0xE6,0xE1,0xE4,0,0xE7,0xE2,0xE5,0,0xE8,0xE3,0xE6,0,
    0xE9,0xE4,0xE7,0,0xEA,0xE5,0xE8,0,0xEB,0xE6,0xE9,0,0xEC,0xE7,0xEA,0,
    0xED,0xE8,0xEB,0,0xEE,0xE9,0xEC,0,0xEF,0xEA,0xED,0,0xF0,0xEB,0xEE,0,
    0xF1,0xEC,0xEF,0,0xF2,0xED,0xF0,0,0xF3,0xEE,0xF1,0,0xF4,0xEF,0xF2,0,
    0xF5,0xF0,0xF3,0,0xF6,0xF1,0xF4,0,0xF7,0xF2,0xF5,0,0xF8,0xF3,0xF6,0,
    0xF9,0xF4,0xF7,0,0xFA,0xF5,0xF8,0,0xFB,0xF6,0xF9,0,0xFC,0xF7,0xFA,0,
    0xFD,0xF8,0xFB,0,0xFE,0xF9,0xFC,0,0xFF,0xFA,0xFD,0,0x00,0xFB,0xFE,0,
    0x01,0xFC,0xFF,0,0x02,0xFD,0x00,0,0x03,0xFE,0x01,0,0x04,0xFF,0x02,0,
    0x05,0x00,0x03,0,0x06,0x01,0x04,0,0x07,0x02,0x05,0,0x08,0x03,0x06,0,
    0x09,0x04,0x07,0,0x0A,0x05,0x08,0,0x0B,0x06,0x09,0,0x0C,0x07,0x0A,0,
    0x0D,0x08,0x0B,0,0x0E,0x09,0x0C,0,0x0F,0x0A,0x0D,0,0x10,0x0B,0x0E,0,
    0x11,0x0C,0x0F,0,0x12,0x0D,0x10,0,0x13,0x0E,0x11,0,0x14,0x0F,0x12,0,
    0x15,0x10,0x13,0,0x16,0x11,0x14,0,0x17,0x12,0x15,0,0x18,0x13,0x16,0,
    0x19,0x14,0x17,0,0x1A,0x15,0x18,0,0x1B,0x16,0x19,0,0x1C,0x17,0x1A,0,
    0x1D,0x18,0x1B,0,0x1E,0x19,0x1C,0,0x1F,0x1A,0x1D,0,0x20,0x1B,0x1E,0,
    0x21,0x1C,0x1F,0,0x22,0x1D,0x20,0,0x23,0x1E,0x21,0,0x24,0x1F,0x22,0,
    0xE6,0xE0,0xE3,0,0xE7,0xE1,0xE4,0,0xE8,0xE2,0xE5,0,0xE9,0xE3,0xE6,0,
    0xEA,0xE4,0xE7,0,0xEB,0xE5,0xE8,0,0xEC,0xE6,0xE9,0,0xED,0xE7,0xEA,0,
    0xEE,0xE8,0xEB,0,0xEF,0xE9,0xEC,0,0xF0,0xEA,0xED,0,0xF1,0xEB,0xEE,0,
    0xF2,0xEC,0xEF,0,0xF3,0xED,0xF0,0,0xF4,0xEE,0xF1,0,0xF5,0xEF,0xF2,0,
    0xF6,0xF0,0xF3,0,0xF7,0xF1,0xF4,0,0xF8,0xF2,0xF5,0,0xF9,0xF3,0xF6,0,
    0xFA,0xF4,0xF7,0,0xFB,0xF5,0xF8,0,0xFC,0xF6,0xF9,0,0xFD,0xF7,0xFA,0,
    0xFE,0xF8,0xFB,0,0xFF,0xF9,0xFC,0,0x00,0xFA,0xFD,0,0x01,0xFB,0xFE,0,
    0x02,0xFC,0xFF,0,0x03,0xFD,0x00,0,0x04,0xFE,0x01,0,0x05,0xFF,0x02,0,
    0x06,0x00,0x03,0,0x07,0x01,0x04,0,0x08,0x02,0x05,0,0x09,0x03,0x06,0,
    0x0A,0x04,0x07,0,0x0B,0x05,0x08,0,0x0C,0x06,0x09,0,0x0D,0x07,0x0A,0,
    0x0E,0x08,0x0B,0,0x0F,0x09,0x0C,0,0x10,0x0A,0x0D,0,0x11,0x0B,0x0E,0,
    0x12,0x0C,0x0F,0,0x13,0x0D,0x10,0,0x14,0x0E,0x11,0,0x15,0x0F,0x12,0,
    0x16,0x10,0x13,0,0x17,0x11,0x14,0,0x18,0x12,0x15,0,0x19,0x13,0x16,0,
    0x1A,0x14,0x17,0,0x1B,0x15,0x18,0,0x1C,0x16,0x19,0,0x1D,0x17,0x1A,0,
    0x1E,0x18,0x1B,0,0x1F,0x19,0x1C,0,0x20,0x1A,0x1D,0,0x21,0x1B,0x1E,0,
    0x22,0x1C,0x1F,0,0x23,0x1D,0x20,0,0x24,0x1E,0x21,0,0x25,0x1F,0x22,0,
    0xE7,0xE0,0xE3,0,0xE8,0xE1,0xE4,0,0xE9,0xE2,0xE5,0,0xEA,0xE3,0xE6,0,
    0xEB,0xE4,0xE7,0,0xEC,0xE5,0xE8,0,0xED,0xE6,0xE9,0,0xEE,0xE7,0xEA,0,
    0xEF,0xE8,0xEB,0,0xF0,0xE9,0xEC,0,0xF1,0xEA,0xED,0,0xF2,0xEB,0xEE,0,
    0xF3,0xEC,0xEF,0,0xF4,0xED,0xF0,0,0xF5,0xEE,0xF1,0,0xF6,0xEF,0xF2,0,
    0xF7,0xF0,0xF3,0,0xF8,0xF1,0xF4,0,0xF9,0xF2,0xF5,0,0xFA,0xF3,0xF6,0,
    0xFB,0xF4,0xF7,0,0xFC,0xF5,0xF8,0,0xFD,0xF6,0xF9,0,0xFE,0xF7,0xFA,0,
    0xFF,0xF8,0xFB,0,0x00,0xF9,0xFC,0,0x01,0xFA,0xFD,0,0x02,0xFB,0xFE,0,
    0x03,0xFC,0xFF,0,0x04,0xFD,0x00,0,0x05,0xFE,0x01,0,0x06,0xFF,0x02,0,
    0x07,0x00,0x03,0,0x08,0x01,0x04,0,0x09,0x02,0x05,0,0x0A,0x03,0x06,0,
    0x0B,0x04,0x07,0,0x0C,0x05,0x08,0,0x0D,0x06,0x09,0,0x0E,0x07,0x0A,0,
    0x0F,0x08,0x0B,0,0x10,0x09,0x0C,0,0x11,0x0A,0x0D,0,0x12,0x0B,0x0E,0,
    0x13,0x0C,0x0F,0,0x14,0x0D,0x10,0,0x15,0x0E,0x11,0,0x16,0x0F,0x12,0,
    0x17,0x10,0x13,0,0x18,0x11,0x14,0,0x19,0x12,0x15,0,0x1A,0x13,0x16,0,
    0x1B,0x14,0x17,0,0x1C,0x15,0x18,0,0x1D,0x16,0x19,0,0x1E,0x17,0x1A,0,
    0x1F,0x18,0x1B,0,0x20,0x19,0x1C,0,0x21,0x1A,0x1D,0,0x22,0x1B,0x1E,0,
    0x23,0x1C,0x1F,0,0x24,0x1D,0x20,0,0x25,0x1E,0x21,0,0x26,0x1F,0x22,0,
    0xD8,0xE0,0xE4,0,0xD9,0xE1,0xE5,0,0xDA,0xE2,0xE6,0,0xDB,0xE3,0xE7,0,
    0xDC,0xE4,0xE8,0,0xDD,0xE5,0xE9,0,0xDE,0xE6,0xEA,0,0xDF,0xE7,0xEB,0,
    0xE0,0xE8,0xEC,0,0xE1,0xE9,0xED,0,0xE2,0xEA,0xEE,0,0xE3,0xEB,0xEF,0,
    0xE4,0xEC,0xF0,0,0xE5,0xED,0xF1,0,0xE6,0xEE,0xF2,0,0xE7,0xEF,0xF3,0,
    0xE8,0xF0,0xF4,0,0xE9,0xF1,0xF5,0,0xEA,0xF2,0xF6,0,0xEB,0xF3,0xF7,0,
    0xEC,0xF4,0xF8,0,0xED,0xF5,0xF9,0,0xEE,0xF6,0xFA,0,0xEF,0xF7,0xFB,0,
    0xF0,0xF8,0xFC,0,0xF1,0xF9,0xFD,0,0xF2,0xFA,0xFE,0,0xF3,0xFB,0xFF,0,
    0xF4,0xFC,0x00,0,0xF5,0xFD,0x01,0,0xF6,0xFE,0x02,0,0xF7,0xFF,0x03,0,
    0xF8,0x00,0x04,0,0xF9,0x01,0x05,0,0xFA,0x02,0x06,0,0xFB,0x03,0x07,0,
    0xFC,0x04,0x08,0,0xFD,0x05,0x09,0,0xFE,0x06,0x0A,0,0xFF,0x07,0x0B,0,
    0x00,0x08,0x0C,0,0x01,0x09,0x0D,0,0x02,0x0A,0x0E,0,0x03,0x0B,0x0F,0,
    0x04,0x0C,0x10,0,0x05,0x0D,0x11,0,0x06,0x0E,0x12,0,0x07,0x0F,0x13,0,
    0x08,0x10,0x14,0,0x09,0x11,0x15,0,0x0A,0x12,0x16,0,0x0B,0x13,0x17,0,
    0x0C,0x14,0x18,0,0x0D,0x15,0x19,0,0x0E,0x16,0x1A,0,0x0F,0x17,0x1B,0,
    0x10,0x18,0x1C,0,0x11,0x19,0x1D,0,0x12,0x1A,0x1E,0,0x13,0x1B,0x1F,0,
    0x14,0x1C,0x20,0,0x15,0x1D,0x21,0,0x16,0x1E,0x22,0,0x17,0x1F,0x23,0,
    0xD9,0xE0,0xE4,0,0xDA,0xE1,0xE5,0,0xDB,0xE2,0xE6,0,0xDC,0xE3,0xE7,0,
    0xDD,0xE4,0xE8,0,0xDE,0xE5,0xE9,0,0xDF,0xE6,0xEA,0,0xE0,0xE7,0xEB,0,
    0xE1,0xE8,0xEC,0,0xE2,0xE9,0xED,0,0xE3,0xEA,0xEE,0,0xE4,0xEB,0xEF,0,
    0xE5,0xEC,0xF0,0,0xE6,0xED,0xF1,0,0xE7,0xEE,0xF2,0,0xE8,0xEF,0xF3,0,
    0xE9,0xF0,0xF4,0,0xEA,0xF1,0xF5,0,0xEB,0xF2,0xF6,0,0xEC,0xF3,0xF7,0,
    0xED,0xF4,0xF8,0,0xEE,0xF5,0xF9,0,0xEF,0xF6,0xFA,0,0xF0,0xF7,0xFB,0,
    0xF1,0xF8,0xFC,0,0xF2,0xF9,0xFD,0,0xF3,0xFA,0xFE,0,0xF4,0xFB,0xFF,0,
    0xF5,0xFC,0x00,0,0xF6,0xFD,0x01,0,0xF7,0xFE,0x02,0,0xF8,0xFF,0x03,0,
    0xF9,0x00,0x04,0,0xFA,0x01,0x05,0,0xFB,0x02,0x06,0,0xFC,0x03,0x07,0,
    0xFD,0x04,0x08,0,0xFE,0x05,0x09,0,0xFF,0x06,0x0A,0,0x00,0x07,0x0B,0,
    0x01,0x08,0x0C,0,0x02,0x09,0x0D,0,0x03,0x0A,0x0E,0,0x04,0x0B,0x0F,0,
    0x05,0x0C,0x10,0,0x06,0x0D,0x11,0,0x07,0x0E,0x12,0,0x08,0x0F,0x13,0,
    0x09,0x10,0x14,0,0x0A,0x11,0x15,0,0x0B,0x12,0x16,0,0x0C,0x13,0x17,0,
    0x0D,0x14,0x18,0,0x0E,0x15,0x19,0,0x0F,0x16,0x1A,0,0x10,0x17,0x1B,0,
    0x11,0x18,0x1C,0,0x12,0x19,0x1D,0,0x13,0x1A,0x1E,0,0x14,0x1B,0x1F,0,
    0x15,0x1C,0x20,0,0x16,0x1D,0x21,0,0x17,0x1E,0x22,0,0x18,0x1F,0x23,0,
    0xDA,0xE0,0xE4,0,0xDB,0xE1,0xE5,0,0xDC,0xE2,0xE6,0,0xDD,0xE3,0xE7,0,
    0xDE,0xE4,0xE8,0,0xDF,0xE5,0xE9,0,0xE0,0xE6,0xEA,0,0xE1,0xE7,0xEB,0,
    0xE2,0xE8,0xEC,0,0xE3,0xE9,0xED,0,0xE4,0xEA,0xEE,0,0xE5,0xEB,0xEF,0,
    0xE6,0xEC,0xF0,0,0xE7,0xED,0xF1,0,0xE8,0xEE,0xF2,0,0xE9,0xEF,0xF3,0,
    0xEA,0xF0,0xF4,0,0xEB,0xF1,0xF5,0,0xEC,0xF2,0xF6,0,0xED,0xF3,0xF7,0,
    0xEE,0xF4,0xF8,0,0xEF,0xF5,0xF9,0,0xF0,0xF6,0xFA,0,0xF1,0xF7,0xFB,0,
    0xF2,0xF8,0xFC,0,0xF3,0xF9,0xFD,0,0xF4,0xFA,0xFE,0,0xF5,0xFB,0xFF,0,
    0xF6,0xFC,0x00,0,0xF7,0xFD,0x01,0,0xF8,0xFE,0x02,0,0xF9,0xFF,0x03,0,
    0xFA,0x00,0x04,0,0xFB,0x01,0x05,0,0xFC,0x02,0x06,0,0xFD,0x03,0x07,0,
    0xFE,0x04,0x08,0,0xFF,0x05,0x09,0,0x00,0x06,0x0A,0,0x01,0x07,0x0B,0,
    0x02,0x08,0x0C,0,0x03,0x09,0x0D,0,0x04,0x0A,0x0E,0,0x05,0x0B,0x0F,0,
    0x06,0x0C,0x10,0,0x07,0x0D,0x11,0,0x08,0x0E,0x12,0,0x09,0x0F,0x13,0,
    0x0A,0x10,0x14,0,0x0B,0x11,0x15,0,0x0C,0x12,0x16,0,0x0D,0x13,0x17,0,
    0x0E,0x14,0x18,0,0x0F,0x15,0x19,0,0x10,0x16,0x1A,0,0x11,0x17,0x1B,0,
    0x12,0x18,0x1C,0,0x13,0x19,0x1D,0,0x14,0x1A,0x1E,0,0x15,0x1B,0x1F,0,
    0x16,0x1C,0x20,0,0x17,0x1D,0x21,0,0x18,0x1E,0x22,0,0x19,0x1F,0x23,0,
    0xDB,0xE0,0xE4,0,0xDC,0xE1,0xE5,0,0xDD,0xE2,0xE6,0,0xDE,0xE3,0xE7,0,
    0xDF,0xE4,0xE8,0,0xE0,0xE5,0xE9,0,0xE1,0xE6,0xEA,0,0xE2,0xE7,0xEB,0,
    0xE3,0xE8,0xEC,0,0xE4,0xE9,0xED,0,0xE5,0xEA,0xEE,0,0xE6,0xEB,0xEF,0,
    0xE7,0xEC,0xF0,0,0xE8,0xED,0xF1,0,0xE9,0xEE,0xF2,0,0xEA,0xEF,0xF3,0,
    0xEB,0xF0,0xF4,0,0xEC,0xF1,0xF5,0,0xED,0xF2,0xF6,0,0xEE,0xF3,0xF7,0,
    0xEF,0xF4,0xF8,0,0xF0,0xF5,0xF9,0,0xF1,0xF6,0xFA,0,0xF2,0xF7,0xFB,0,
    0xF3,0xF8,0xFC,0,0xF4,0xF9,0xFD,0,0xF5,0xFA,0xFE,0,0xF6,0xFB,0xFF,0,
    0xF7,0xFC,0x00,0,0xF8,0xFD,0x01,0,0xF9,0xFE,0x02,0,0xFA,0xFF,0x03,0,
    0xFB,0x00,0x04,0,0xFC,0x01,0x05,0,0xFD,0x02,0x06,0,0xFE,0x03,0x07,0,
    0xFF,0x04,0x08,0,0x00,0x05,0x09,0,0x01,0x06,0x0A,0,0x02,0x07,0x0B,0,
    0x03,0x08,0x0C,0,0x04,0x09,0x0D,0,0x05,0x0A,0x0E,0,0x06,0x0B,0x0F,0,
    0x07,0x0C,0x10,0,0x08,0x0D,0x11,0,0x09,0x0E,0x12,0,0x0A,0x0F,0x13,0,
    0x0B,0x10,0x14,0,0x0C,0x11,0x15,0,0x0D,0x12,0x16,0,0x0E,0x13,0x17,0,
    0x0F,0x14,0x18,0,0x10,0x15,0x19,0,0x11,0x16,0x1A,0,0x12,0x17,0x1B,0,
    0x13,0x18,0x1C,0,0x14,0x19,0x1D,0,0x15,0x1A,0x1E,0,0x16,0x1B,0x1F,0,
    0x17,0x1C,0x20,0,0x18,0x1D,0x21,0,0x19,0x1E,0x22,0,0x1A,0x1F,0x23,0,
    0xDC,0xE0,0xE4,0,0xDD,0xE1,0xE5,0,0xDE,0xE2,0xE6,0,0xDF,0xE3,0xE7,0,
    0xE0,0xE4,0xE8,0,0xE1,0xE5,0xE9,0,0xE2,0xE6,0xEA,0,0xE3,0xE7,0xEB,0,
    0xE4,0xE8,0xEC,0,0xE5,0xE9,0xED,0,0xE6,0xEA,0xEE,0,0xE7,0xEB,0xEF,0,
    0xE8,0xEC,0xF0,0,0xE9,0xED,0xF1,0,0xEA,0xEE,0xF2,0,0xEB,0xEF,0xF3,0,
    0xEC,0xF0,0xF4,0,0xED,0xF1,0xF5,0,0xEE,0xF2,0xF6,0,0xEF,0xF3,0xF7,0,
    0xF0,0xF4,0xF8,0,0xF1,0xF5,0xF9,0,0xF2,0xF6,0xFA,0,0xF3,0xF7,0xFB,0,
    0xF4,0xF8,0xFC,0,0xF5,0xF9,0xFD,0,0xF6,0xFA,0xFE,0,0xF7,0xFB,0xFF,0,
    0xF8,0xFC,0x00,0,0xF9,0xFD,0x01,0,0xFA,0xFE,0x02,0,0xFB,0xFF,0x03,0,
    0xFC,0x00,0x04,0,0xFD,0x01,0x05,0,0xFE,0x02,0x06,0,0xFF,0x03,0x07,0,
    0x00,0x04,0x08,0,0x01,0x05,0x09,0,0x02,0x06,0x0A,0,0x03,0x07,0x0B,0,
    0x04,0x08,0x0C,0,0x05,0x09,0x0D,0,0x06,0x0A,0x0E,0,0x07,0x0B,0x0F,0,
    0x08,0x0C,0x10,0,0x09,0x0D,0x11,0,0x0A,0x0E,0x12,0,0x0B,0x0F,0x13,0,
    0x0C,0x10,0x14,0,0x0D,0x11,0x15,0,0x0E,0x12,0x16,0,0x0F,0x13,0x17,0,
    0x10,0x14,0x18,0,0x11,0x15,0x19,0,0x12,0x16,0x1A,0,0x13,0x17,0x1B,0,
    0x14,0x18,0x1C,0,0x15,0x19,0x1D,0,0x16,0x1A,0x1E,0,0x17,0x1B,0x1F,0,
    0x18,0x1C,0x20,0,0x19,0x1D,0x21,0,0x1A,0x1E,0x22,0,0x1B,0x1F,0x23,0,
    0xDD,0xE0,0xE4,0,0xDE,0xE1,0xE5,0,0xDF,0xE2,0xE6,0,0xE0,0xE3,0xE7,0,
    0xE1,0xE4,0xE8,0,0xE2,0xE5,0xE9,0,0xE3,0xE6,0xEA,0,0xE4,0xE7,0xEB,0,
    0xE5,0xE8,0xEC,0,0xE6,0xE9,0xED,0,0xE7,0xEA,0xEE,0,0xE8,0xEB,0xEF,0,
    0xE9,0xEC,0xF0,0,0xEA,0xED,0xF1,0,0xEB,0xEE,0xF2,0,0xEC,0xEF,0xF3,0,
    0xED,0xF0,0xF4,0,0xEE,0xF1,0xF5,0,0xEF,0xF2,0xF6,0,0xF0,0xF3,0xF7,0,
    0xF1,0xF4,0xF8,0,0xF2,0xF5,0xF9,0,0xF3,0xF6,0xFA,0,0xF4,0xF7,0xFB,0,
    0xF5,0xF8,0xFC,0,0xF6,0xF9,0xFD,0,0xF7,0xFA,0xFE,0,0xF8,0xFB,0xFF,0,
    0xF9,0xFC,0x00,0,0xFA,0xFD,0x01,0,0xFB,0xFE,0x02,0,0xFC,0xFF,0x03,0,
    0xFD,0x00,0x04,0,0xFE,0x01,0x05,0,0xFF,0x02,0x06,0,0x00,0x03,0x07,0,
    0x01,0x04,0x08,0,0x02,0x05,0x09,0,0x03,0x06,0x0A,0,0x04,0x07,0x0B,0,
    0x05,0x08,0x0C,0,0x06,0x09,0x0D,0,0x07,0x0A,0x0E,0,0x08,0x0B,0x0F,0,
    0x09,0x0C,0x10,0,0x0A,0x0D,0x11,0,0x0B,0x0E,0x12,0,0x0C,0x0F,0x13,0,
    0x0D,0x10,0x14,0,0x0E,0x11,0x15,0,0x0F,0x12,0x16,0,0x10,0x13,0x17,0,
    0x11,0x14,0x18,0,0x12,0x15,0x19,0,0x13,0x16,0x1A,0,0x14,0x17,0x1B,0,
    0x15,0x18,0x1C,0,0x16,0x19,0x1D,0,0x17,0x1A,0x1E,0,0x18,0x1B,0x1F,0,
    0x19,0x1C,0x20,0,0x1A,0x1D,0x21,0,0x1B,0x1E,0x22,0,0x1C,0x1F,0x23,0,
    0xDE,0xE0,0xE4,0,0xDF,0xE1,0xE5,0,0xE0,0xE2,0xE6,0,0xE1,0xE3,0xE7,0,
    0xE2,0xE4,0xE8,0,0xE3,0xE5,0xE9,0,0xE4,0xE6,0xEA,0,0xE5,0xE7,0xEB,0,
    0xE6,0xE8,0xEC,0,0xE7,0xE9,0xED,0,0xE8,0xEA,0xEE,0,0xE9,0xEB,0xEF,0,
    0xEA,0xEC,0xF0,0,0xEB,0xED,0xF1,0,0xEC,0xEE,0xF2,0,0xED,0xEF,0xF3,0,
    0xEE,0xF0,0xF4,0,0xEF,0xF1,0xF5,0,0xF0,0xF2,0xF6,0,0xF1,0xF3,0xF7,0,
    0xF2,0xF4,0xF8,0,0xF3,0xF5,0xF9,0,0xF4,0xF6,0xFA,0,0xF5,0xF7,0xFB,0,
    0xF6,0xF8,0xFC,0,0xF7,0xF9,0xFD,0,0xF8,0xFA,0xFE,0,0xF9,0xFB,0xFF,0,
    0xFA,0xFC,0x00,0,0xFB,0xFD,0x01,0,0xFC,0xFE,0x02,0,0xFD,0xFF,0x03,0,
    0xFE,0x00,0x04,0,0xFF,0x01,0x05,0,0x00,0x02,0x06,0,0x01,0x03,0x07,0,
    0x02,0x04,0x08,0,0x03,0x05,0x09,0,0x04,0x06,0x0A,0,0x05,0x07,0x0B,0,
    0x06,0x08,0x0C,0,0x07,0x09,0x0D,0,0x08,0x0A,0x0E,0,0x09,0x0B,0x0F,0,
    0x0A,0x0C,0x10,0,0x0B,0x0D,0x11,0,0x0C,0x0E,0x12,0,0x0D,0x0F,0x13,0,
    0x0E,0x10,0x14,0,0x0F,0x11,0x15,0,0x10,0x12,0x16,0,0x11,0x13,0x17,0,
    0x12,0x14,0x18,0,0x13,0x15,0x19,0,0x14,0x16,0x1A,0,0x15,0x17,0x1B,0,
    0x16,0x18,0x1C,0,0x17,0x19,0x1D,0,0x18,0x1A,0x1E,0,0x19,0x1B,0x1F,0,
    0x1A,0x1C,0x20,0,0x1B,0x1D,0x21,0,0x1C,0x1E,0x22,0,0x1D,0x1F,0x23,0,
    0xDF,0xE0,0xE4,0,0xE0,0xE1,0xE5,0,0xE1,0xE2,0xE6,0,0xE2,0xE3,0xE7,0,
    0xE3,0xE4,0xE8,0,0xE4,0xE5,0xE9,0,0xE5,0xE6,0xEA,0,0xE6,0xE7,0xEB,0,
    0xE7,0xE8,0xEC,0,0xE8,0xE9,0xED,0,0xE9,0xEA,0xEE,0,0xEA,0xEB,0xEF,0,
    0xEB,0xEC,0xF0,0,0xEC,0xED,0xF1,0,0xED,0xEE,0xF2,0,0xEE,0xEF,0xF3,0,
    0xEF,0xF0,0xF4,0,0xF0,0xF1,0xF5,0,0xF1,0xF2,0xF6,0,0xF2,0xF3,0xF7,0,
    0xF3,0xF4,0xF8,0,0xF4,0xF5,0xF9,0,0xF5,0xF6,0xFA,0,0xF6,0xF7,0xFB,0,
    0xF7,0xF8,0xFC,0,0xF8,0xF9,0xFD,0,0xF9,0xFA,0xFE,0,0xFA,0xFB,0xFF,0,
    0xFB,0xFC,0x00,0,0xFC,0xFD,0x01,0,0xFD,0xFE,0x02,0,0xFE,0xFF,0x03,0,
    0xFF,0x00,0x04,0,0x00,0x01,0x05,0,0x01,0x02,0x06,0,0x02,0x03,0x07,0,
    0x03,0x04,0x08,0,0x04,0x05,0x09,0,0x05,0x06,0x0A,0,0x06,0x07,0x0B,0,
    0x07,0x08,0x0C,0,0x08,0x09,0x0D,0,0x09,0x0A,0x0E,0,0x0A,0x0B,0x0F,0,
    0x0B,0x0C,0x10,0,0x0C,0x0D,0x11,0,0x0D,0x0E,0x12,0,0x0E,0x0F,0x13,0,
    0x0F,0x10,0x14,0,0x10,0x11,0x15,0,0x11,0x12,0x16,0,0x12,0x13,0x17,0,
    0x13,0x14,0x18,0,0x14,0x15,0x19,0,0x15,0x16,0x1A,0,0x16,0x17,0x1B,0,
    0x17,0x18,0x1C,0,0x18,0x19,0x1D,0,0x19,0x1A,0x1E,0,0x1A,0x1B,0x1F,0,
    0x1B,0x1C,0x20,0,0x1C,0x1D,0x21,0,0x1D,0x1E,0x22,0,0x1E,0x1F,0x23,0,
    0xE0,0xE0,0xE4,0,0xE1,0xE1,0xE5,0,0xE2,0xE2,0xE6,0,0xE3,0xE3,0xE7,0,
    0xE4,0xE4,0xE8,0,0xE5,0xE5,0xE9,0,0xE6,0xE6,0xEA,0,0xE7,0xE7,0xEB,0,
    0xE8,0xE8,0xEC,0,0xE9,0xE9,0xED,0,0xEA,0xEA,0xEE,0,0xEB,0xEB,0xEF,0,
    0xEC,0xEC,0xF0,0,0xED,0xED,0xF1,0,0xEE,0xEE,0xF2,0,0xEF,0xEF,0xF3,0,
    0xF0,0xF0,0xF4,0,0xF1,0xF1,0xF5,0,0xF2,0xF2,0xF6,0,0xF3,0xF3,0xF7,0,
    0xF4,0xF4,0xF8,0,0xF5,0xF5,0xF9,0,0xF6,0xF6,0xFA,0,0xF7,0xF7,0xFB,0,
    0xF8,0xF8,0xFC,0,0xF9,0xF9,0xFD,0,0xFA,0xFA,0xFE,0,0xFB,0xFB,0xFF,0,
    0xFC,0xFC,0x00,0,0xFD,0xFD,0x01,0,0xFE,0xFE,0x02,0,0xFF,0xFF,0x03,0,
    0x00,0x00,0x04,0,0x01,0x01,0x05,0,0x02,0x02,0x06,0,0x03,0x03,0x07,0,
    0x04,0x04,0x08,0,0x05,0x05,0x09,0,0x06,0x06,0x0A,0,0x07,0x07,0x0B,0,
    0x08,0x08,0x0C,0,0x09,0x09,0x0D,0,0x0A,0x0A,0x0E,0,0x0B,0x0B,0x0F,0,
    0x0C,0x0C,0x10,0,0x0D,0x0D,0x11,0,0x0E,0x0E,0x12,0,0x0F,0x0F,0x13,0,
    0x10,0x10,0x14,0,0x11,0x11,0x15,0,0x12,0x12,0x16,0,0x13,0x13,0x17,0,
    0x14,0x14,0x18,0,0x15,0x15,0x19,0,0x16,0x16,0x1A,0,0x17,0x17,0x1B,0,
    0x18,0x18,0x1C,0,0x19,0x19,0x1D,0,0x1A,0x1A,0x1E,0,0x1B,0x1B,0x1F,0,
    0x1C,0x1C,0x20,0,0x1D,0x1D,0x21,0,0x1E,0x1E,0x22,0,0x1F,0x1F,0x23,0,
    0xE1,0xE0,0xE4,0,0xE2,0xE1,0xE5,0,0xE3,0xE2,0xE6,0,0xE4,0xE3,0xE7,0,
    0xE5,0xE4,0xE8,0,0xE6,0xE5,0xE9,0,0xE7,0xE6,0xEA,0,0xE8,0xE7,0xEB,0,
    0xE9,0xE8,0xEC,0,0xEA,0xE9,0xED,0,0xEB,0xEA,0xEE,0,0xEC,0xEB,0xEF,0,
    0xED,0xEC,0xF0,0,0xEE,0xED,0xF1,0,0xEF,0xEE,0xF2,0,0xF0,0xEF,0xF3,0,
    0xF1,0xF0,0xF4,0,0xF2,0xF1,0xF5,0,0xF3,0xF2,0xF6,0,0xF4,0xF3,0xF7,0,
    0xF5,0xF4,0xF8,0,0xF6,0xF5,0xF9,0,0xF7,0xF6,0xFA,0,0xF8,0xF7,0xFB,0,
    0xF9,0xF8,0xFC,0,0xFA,0xF9,0xFD,0,0xFB,0xFA,0xFE,0,0xFC,0xFB,0xFF,0,
    0xFD,0xFC,0x00,0,0xFE,0xFD,0x01,0,0xFF,0xFE,0x02,0,0x00,0xFF,0x03,0,
    0x01,0x00,0x04,0,0x02,0x01,0x05,0,0x03,0x02,0x06,0,0x04,0x03,0x07,0,
    0x05,0x04,0x08,0,0x06,0x05,0x09,0,0x07,0x06,0x0A,0,0x08,0x07,0x0B,0,
    0x09,0x08,0x0C,0,0x0A,0x09,0x0D,0,0x0B,0x0A,0x0E,0,0x0C,0x0B,0x0F,0,
    0x0D,0x0C,0x10,0,0x0E,0x0D,0x11,0,0x0F,0x0E,0x12,0,0x10,0x0F,0x13,0,
    0x11,0x10,0x14,0,0x12,0x11,0x15,0,0x13,0x12,0x16,0,0x14,0x13,0x17,0,
    0x15,0x14,0x18,0,0x16,0x15,0x19,0,0x17,0x16,0x1A,0,0x18,0x17,0x1B,0,
    0x19,0x18,0x1C,0,0x1A,0x19,0x1D,0,0x1B,0x1A,0x1E,0,0x1C,0x1B,0x1F,0,
    0x1D,0x1C,0x20,0,0x1E,0x1D,0x21,0,0x1F,0x1E,0x22,0,0x20,0x1F,0x23,0,
    0xE2,0xE0,0xE4,0,0xE3,0xE1,0xE5,0,0xE4,0xE2,0xE6,0,0xE5,0xE3,0xE7,0,
    0xE6,0xE4,0xE8,0,0xE7,0xE5,0xE9,0,0xE8,0xE6,0xEA,0,0xE9,0xE7,0xEB,0,
    0xEA,0xE8,0xEC,0,0xEB,0xE9,0xED,0,0xEC,0xEA,0xEE,0,0xED,0xEB,0xEF,0,
    0xEE,0xEC,0xF0,0,0xEF,0xED,0xF1,0,0xF0,0xEE,0xF2,0,0xF1,0xEF,0xF3,0,
    0xF2,0xF0,0xF4,0,0xF3,0xF1,0xF5,0,0xF4,0xF2,0xF6,0,0xF5,0xF3,0xF7,0,
    0xF6,0xF4,0xF8,0,0xF7,0xF5,0xF9,0,0xF8,0xF6,0xFA,0,0xF9,0xF7,0xFB,0,
    0xFA,0xF8,0xFC,0,0xFB,0xF9,0xFD,0,0xFC,0xFA,0xFE,0,0xFD,0xFB,0xFF,0,
    0xFE,0xFC,0x00,0,0xFF,0xFD,0x01,0,0x00,0xFE,0x02,0,0x01,0xFF,0x03,0,
    0x02,0x00,0x04,0,0x03,0x01,0x05,0,0x04,0x02,0x06,0,0x05,0x03,0x07,0,
    0x06,0x04,0x08,0,0x07,0x05,0x09,0,0x08,0x06,0x0A,0,0x09,0x07,0x0B,0,
    0x0A,0x08,0x0C,0,0x0B,0x09,0x0D,0,0x0C,0x0A,0x0E,0,0x0D,0x0B,0x0F,0,
    0x0E,0x0C,0x10,0,0x0F,0x0D,0x11,0,0x10,0x0E,0x12,0,0x11,0x0F,0x13,0,
    0x12,0x10,0x14,0,0x13,0x11,0x15,0,0x14,0x12,0x16,0,0x15,0x13,0x17,0,
    0x16,0x14,0x18,0,0x17,0x15,0x19,0,0x18,0x16,0x1A,0,0x19,0x17,0x1B,0,
    0x1A,0x18,0x1C,0,0x1B,0x19,0x1D,0,0x1C,0x1A,0x1E,0,0x1D,0x1B,0x1F,0,
    0x1E,0x1C,0x20,0,0x1F,0x1D,0x21,0,0x20,0x1E,0x22,0,0x21,0x1F,0x23,0,
    0xE3,0xE0,0xE4,0,0xE4,0xE1,0xE5,0,0xE5,0xE2,0xE6,0,0xE6,0xE3,0xE7,0,
    0xE7,0xE4,0xE8,0,0xE8,0xE5,0xE9,0,0xE9,0xE6,0xEA,0,0xEA,0xE7,0xEB,0,
    0xEB,0xE8,0xEC,0,0xEC,0xE9,0xED,0,0xED,0xEA,0xEE,0,0xEE,0xEB,0xEF,0,
    0xEF,0xEC,0xF0,0,0xF0,0xED,0xF1,0,0xF1,0xEE,0xF2,0,0xF2,0xEF,0xF3,0,
    0xF3,0xF0,0xF4,0,0xF4,0xF1,0xF5,0,0xF5,0xF2,0xF6,0,0xF6,0xF3,0xF7,0,
    0xF7,0xF4,0xF8,0,0xF8,0xF5,0xF9,0,0xF9,0xF6,0xFA,0,0xFA,0xF7,0xFB,0,
    0xFB,0xF8,0xFC,0,0xFC,0xF9,0xFD,0,0xFD,0xFA,0xFE,0,0xFE,0xFB,0xFF,0,
    0xFF,0xFC,0x00,0,0x00,0xFD,0x01,0,0x01,0xFE,0x02,0,0x02,0xFF,0x03,0,
    0x03,0x00,0x04,0,0x04,0x01,0x05,0,0x05,0x02,0x06,0,0x06,0x03,0x07,0,
    0x07,0x04,0x08,0,0x08,0x05,0x09,0,0x09,0x06,0x0A,0,0x0A,0x07,0x0B,0,
    0x0B,0x08,0x0C,0,0x0C,0x09,0x0D,0,0x0D,0x0A,0x0E,0,0x0E,0x0B,0x0F,0,
    0x0F,0x0C,0x10,0,0x10,0x0D,0x11,0,0x11,0x0E,0x12,0,0x12,0x0F,0x13,0,
    0x13,0x10,0x14,0,0x14,0x11,0x15,0,0x15,0x12,0x16,0,0x16,0x13,0x17,0,
    0x17,0x14,0x18,0,0x18,0x15,0x19,0,0x19,0x16,0x1A,0,0x1A,0x17,0x1B,0,
    0x1B,0x18,0x1C,0,0x1C,0x19,0x1D,0,0x1D,0x1A,0x1E,0,0x1E,0x1B,0x1F,0,
    0x1F,0x1C,0x20,0,0x20,0x1D,0x21,0,0x21,0x1E,0x22,0,0x22,0x1F,0x23,0,
    0xE4,0xE0,0xE4,0,0xE5,0xE1,0xE5,0,0xE6,0xE2,0xE6,0,0xE7,0xE3,0xE7,0,
    0xE8,0xE4,0xE8,0,0xE9,0xE5,0xE9,0,0xEA,0xE6,0xEA,0,0xEB,0xE7,0xEB,0,
    0xEC,0xE8,0xEC,0,0xED,0xE9,0xED,0,0xEE,0xEA,0xEE,0,0xEF,0xEB,0xEF,0,
    0xF0,0xEC,0xF0,0,0xF1,0xED,0xF1,0,0xF2,0xEE,0xF2,0,0xF3,0xEF,0xF3,0,
    0xF4,0xF0,0xF4,0,0xF5,0xF1,0xF5,0,0xF6,0xF2,0xF6,0,0xF7,0xF3,0xF7,0,
    0xF8,0xF4,0xF8,0,0xF9,0xF5,0xF9,0,0xFA,0xF6,0xFA,0,0xFB,0xF7,0xFB,0,
    0xFC,0xF8,0xFC,0,0xFD,0xF9,0xFD,0,0xFE,0xFA,0xFE,0,0xFF,0xFB,0xFF,0,
    0x00,0xFC,0x00,0,0x01,0xFD,0x01,0,0x02,0xFE,0x02,0,0x03,0xFF,0x03,0,
    0x04,0x00,0x04,0,0x05,0x01,0x05,0,0x06,0x02,0x06,0,0x07,0x03,0x07,0,
    0x08,0x04,0x08,0,0x09,0x05,0x09,0,0x0A,0x06,0x0A,0,0x0B,0x07,0x0B,0,
    0x0C,0x08,0x0C,0,0x0D,0x09,0x0D,0,0x0E,0x0A,0x0E,0,0x0F,0x0B,0x0F,0,
    0x10,0x0C,0x10,0,0x11,0x0D,0x11,0,0x12,0x0E,0x12,0,0x13,0x0F,0x13,0,
    0x14,0x10,0x14,0,0x15,0x11,0x15,0,0x16,0x12,0x16,0,0x17,0x13,0x17,0,
    0x18,0x14,0x18,0,0x19,0x15,0x19,0,0x1A,0x16,0x1A,0,0x1B,0x17,0x1B,0,
    0x1C,0x18,0x1C,0,0x1D,0x19,0x1D,0,0x1E,0x1A,0x1E,0,0x1F,0x1B,0x1F,0,
    0x20,0x1C,0x20,0,0x21,0x1D,0x21,0,0x22,0x1E,0x22,0,0x23,0x1F,0x23,0,
    0xE5,0xE0,0xE4,0,0xE6,0xE1,0xE5,0,0xE7,0xE2,0xE6,0,0xE8,0xE3,0xE7,0,
    0xE9,0xE4,0xE8,0,0xEA,0xE5,0xE9,0,0xEB,0xE6,0xEA,0,0xEC,0xE7,0xEB,0,
    0xED,0xE8,0xEC,0,0xEE,0xE9,0xED,0,0xEF,0xEA,0xEE,0,0xF0,0xEB,0xEF,0,
    0xF1,0xEC,0xF0,0,0xF2,0xED,0xF1,0,0xF3,0xEE,0xF2,0,0xF4,0xEF,0xF3,0,
    0xF5,0xF0,0xF4,0,0xF6,0xF1,0xF5,0,0xF7,0xF2,0xF6,0,0xF8,0xF3,0xF7,0,
    0xF9,0xF4,0xF8,0,0xFA,0xF5,0xF9,0,0xFB,0xF6,0xFA,0,0xFC,0xF7,0xFB,0,
    0xFD,0xF8,0xFC,0,0xFE,0xF9,0xFD,0,0xFF,0xFA,0xFE,0,0x00,0xFB,0xFF,0,
    0x01,0xFC,0x00,0,0x02,0xFD,0x01,0,0x03,0xFE,0x02,0,0x04,0xFF,0x03,0,
    0x05,0x00,0x04,0,0x06,0x01,0x05,0,0x07,0x02,0x06,0,0x08,0x03,0x07,0,
    0x09,0x04,0x08,0,0x0A,0x05,0x09,0,0x0B,0x06,0x0A,0,0x0C,0x07,0x0B,0,
    0x0D,0x08,0x0C,0,0x0E,0x09,0x0D,0,0x0F,0x0A,0x0E,0,0x10,0x0B,0x0F,0,
    0x11,0x0C,0x10,0,0x12,0x0D,0x11,0,0x13,0x0E,0x12,0,0x14,0x0F,0x13,0,
    0x15,0x10,0x14,0,0x16,0x11,0x15,0,0x17,0x12,0x16,0,0x18,0x13,0x17,0,
    0x19,0x14,0x18,0,0x1A,0x15,0x19,0,0x1B,0x16,0x1A,0,0x1C,0x17,0x1B,0,
    0x1D,0x18,0x1C,0,0x1E,0x19,0x1D,0,0x1F,0x1A,0x1E,0,0x20,0x1B,0x1F,0,
    0x21,0x1C,0x20,0,0x22,0x1D,0x21,0,0x23,0x1E,0x22,0,0x24,0x1F,0x23,0,
    0xE6,0xE0,0xE4,0,0xE7,0xE1,0xE5,0,0xE8,0xE2,0xE6,0,0xE9,0xE3,0xE7,0,
    0xEA,0xE4,0xE8,0,0xEB,0xE5,0xE9,0,0xEC,0xE6,0xEA,0,0xED,0xE7,0xEB,0,
    0xEE,0xE8,0xEC,0,0xEF,0xE9,0xED,0,0xF0,0xEA,0xEE,0,0xF1,0xEB,0xEF,0,
    0xF2,0xEC,0xF0,0,0xF3,0xED,0xF1,0,0xF4,0xEE,0xF2,0,0xF5,0xEF,0xF3,0,
    0xF6,0xF0,0xF4,0,0xF7,0xF1,0xF5,0,0xF8,0xF2,0xF6,0,0xF9,0xF3,0xF7,0,
    0xFA,0xF4,0xF8,0,0xFB,0xF5,0xF9,0,0xFC,0xF6,0xFA,0,0xFD,0xF7,0xFB,0,
    0xFE,0xF8,0xFC,0,0xFF,0xF9,0xFD,0,0x00,0xFA,0xFE,0,0x01,0xFB,0xFF,0,
    0x02,0xFC,0x00,0,0x03,0xFD,0x01,0,0x04,0xFE,0x02,0,0x05,0xFF,0x03,0,
    0x06,0x00,0x04,0,0x07,0x01,0x05,0,0x08,0x02,0x06,0,0x09,0x03,0x07,0,
    0x0A,0x04,0x08,0,0x0B,0x05,0x09,0,0x0C,0x06,0x0A,0,0x0D,0x07,0x0B,0,
    0x0E,0x08,0x0C,0,0x0F,0x09,0x0D,0,0x10,0x0A,0x0E,0,0x11,0x0B,0x0F,0,
    0x12,0x0C,0x10,0,0x13,0x0D,0x11,0,0x14,0x0E,0x12,0,0x15,0x0F,0x13,0,
    0x16,0x10,0x14,0,0x17,0x11,0x15,0,0x18,0x12,0x16,0,0x19,0x13,0x17,0,
    0x1A,0x14,0x18,0,0x1B,0x15,0x19,0,0x1C,0x16,0x1A,0,0x1D,0x17,0x1B,0,
    0x1E,0x18,0x1C,0,0x1F,0x19,0x1D,0,0x20,0x1A,0x1E,0,0x21,0x1B,0x1F,0,
    0x22,0x1C,0x20,0,0x23,0x1D,0x21,0,0x24,0x1E,0x22,0,0x25,0x1F,0x23,0,
    0xE7,0xE0,0xE4,0,0xE8,0xE1,0xE5,0,0xE9,0xE2,0xE6,0,0xEA,0xE3,0xE7,0,
    0xEB,0xE4,0xE8,0,0xEC,0xE5,0xE9,0,0xED,0xE6,0xEA,0,0xEE,0xE7,0xEB,0,
    0xEF,0xE8,0xEC,0,0xF0,0xE9,0xED,0,0xF1,0xEA,0xEE,0,0xF2,0xEB,0xEF,0,
    0xF3,0xEC,0xF0,0,0xF4,0xED,0xF1,0,0xF5,0xEE,0xF2,0,0xF6,0xEF,0xF3,0,
    0xF7,0xF0,0xF4,0,0xF8,0xF1,0xF5,0,0xF9,0xF2,0xF6,0,0xFA,0xF3,0xF7,0,
    0xFB,0xF4,0xF8,0,0xFC,0xF5,0xF9,0,0xFD,0xF6,0xFA,0,0xFE,0xF7,0xFB,0,
    0xFF,0xF8,0xFC,0,0x00,0xF9,0xFD,0,0x01,0xFA,0xFE,0,0x02,0xFB,0xFF,0,
    0x03,0xFC,0x00,0,0x04,0xFD,0x01,0,0x05,0xFE,0x02,0,0x06,0xFF,0x03,0,
    0x07,0x00,0x04,0,0x08,0x01,0x05,0,0x09,0x02,0x06,0,0x0A,0x03,0x07,0,
    0x0B,0x04,0x08,0,0x0C,0x05,0x09,0,0x0D,0x06,0x0A,0,0x0E,0x07,0x0B,0,
    0x0F,0x08,0x0C,0,0x10,0x09,0x0D,0,0x11,0x0A,0x0E,0,0x12,0x0B,0x0F,0,
    0x13,0x0C,0x10,0,0x14,0x0D,0x11,0,0x15,0x0E,0x12,0,0x16,0x0F,0x13,0,
    0x17,0x10,0x14,0,0x18,0x11,0x15,0,0x19,0x12,0x16,0,0x1A,0x13,0x17,0,
    0x1B,0x14,0x18,0,0x1C,0x15,0x19,0,0x1D,0x16,0x1A,0,0x1E,0x17,0x1B,0,
    0x1F,0x18,0x1C,0,0x20,0x19,0x1D,0,0x21,0x1A,0x1E,0,0x22,0x1B,0x1F,0,
    0x23,0x1C,0x20,0,0x24,0x1D,0x21,0,0x25,0x1E,0x22,0,0x26,0x1F,0x23,0,
    0xD8,0xE0,0xE5,0,0xD9,0xE1,0xE6,0,0xDA,0xE2,0xE7,0,0xDB,0xE3,0xE8,0,
    0xDC,0xE4,0xE9,0,0xDD,0xE5,0xEA,0,0xDE,0xE6,0xEB,0,0xDF,0xE7,0xEC,0,
    0xE0,0xE8,0xED,0,0xE1,0xE9,0xEE,0,0xE2,0xEA,0xEF,0,0xE3,0xEB,0xF0,0,
    0xE4,0xEC,0xF1,0,0xE5,0xED,0xF2,0,0xE6,0xEE,0xF3,0,0xE7,0xEF,0xF4,0,
    0xE8,0xF0,0xF5,0,0xE9,0xF1,0xF6,0,0xEA,0xF2,0xF7,0,0xEB,0xF3,0xF8,0,
    0xEC,0xF4,0xF9,0,0xED,0xF5,0xFA,0,0xEE,0xF6,0xFB,0,0xEF,0xF7,0xFC,0,
    0xF0,0xF8,0xFD,0,0xF1,0xF9,0xFE,0,0xF2,0xFA,0xFF,0,0xF3,0xFB,0x00,0,
    0xF4,0xFC,0x01,0,0xF5,0xFD,0x02,0,0xF6,0xFE,0x03,0,0xF7,0xFF,0x04,0,
    0xF8,0x00,0x05,0,0xF9,0x01,0x06,0,0xFA,0x02,0x07,0,0xFB,0x03,0x08,0,
    0xFC,0x04,0x09,0,0xFD,0x05,0x0A,0,0xFE,0x06,0x0B,0,0xFF,0x07,0x0C,0,
    0x00,0x08,0x0D,0,0x01,0x09,0x0E,0,0x02,0x0A,0x0F,0,0x03,0x0B,0x10,0,
    0x04,0x0C,0x11,0,0x05,0x0D,0x12,0,0x06,0x0E,0x13,0,0x07,0x0F,0x14,0,
    0x08,0x10,0x15,0,0x09,0x11,0x16,0,0x0A,0x12,0x17,0,0x0B,0x13,0x18,0,
    0x0C,0x14,0x19,0,0x0D,0x15,0x1A,0,0x0E,0x16,0x1B,0,0x0F,0x17,0x1C,0,
    0x10,0x18,0x1D,0,0x11,0x19,0x1E,0,0x12,0x1A,0x1F,0,0x13,0x1B,0x20,0,
    0x14,0x1C,0x21,0,0x15,0x1D,0x22,0,0x16,0x1E,0x23,0,0x17,0x1F,0x24,0,
    0xD9,0xE0,0xE5,0,0xDA,0xE1,0xE6,0,0xDB,0xE2,0xE7,0,0xDC,0xE3,0xE8,0,
    0xDD,0xE4,0xE9,0,0xDE,0xE5,0xEA,0,0xDF,0xE6,0xEB,0,0xE0,0xE7,0xEC,0,
    0xE1,0xE8,0xED,0,0xE2,0xE9,0xEE,0,0xE3,0xEA,0xEF,0,0xE4,0xEB,0xF0,0,
    0xE5,0xEC,0xF1,0,0xE6,0xED,0xF2,0,0xE7,0xEE,0xF3,0,0xE8,0xEF,0xF4,0,
    0xE9,0xF0,0xF5,0,0xEA,0xF1,0xF6,0,0xEB,0xF2,0xF7,0,0xEC,0xF3,0xF8,0,
    0xED,0xF4,0xF9,0,0xEE,0xF5,0xFA,0,0xEF,0xF6,0xFB,0,0xF0,0xF7,0xFC,0,
    0xF1,0xF8,0xFD,0,0xF2,0xF9,0xFE,0,0xF3,0xFA,0xFF,0,0xF4,0xFB,0x00,0,
    0xF5,0xFC,0x01,0,0xF6,0xFD,0x02,0,0xF7,0xFE,0x03,0,0xF8,0xFF,0x04,0,
    0xF9,0x00,0x05,0,0xFA,0x01,0x06,0,0xFB,0x02,0x07,0,0xFC,0x03,0x08,0,
    0xFD,0x04,0x09,0,0xFE,0x05,0x0A,0,0xFF,0x06,0x0B,0,0x00,0x07,0x0C,0,
    0x01,0x08,0x0D,0,0x02,0x09,0x0E,0,0x03,0x0A,0x0F,0,0x04,0x0B,0x10,0,
    0x05,0x0C,0x11,0,0x06,0x0D,0x12,0,0x07,0x0E,0x13,0,0x08,0x0F,0x14,0,
    0x09,0x10,0x15,0,0x0A,0x11,0x16,0,0x0B,0x12,0x17,0,0x0C,0x13,0x18,0,
    0x0D,0x14,0x19,0,0x0E,0x15,0x1A,0,0x0F,0x16,0x1B,0,0x10,0x17,0x1C,0,
    0x11,0x18,0x1D,0,0x12,0x19,0x1E,0,0x13,0x1A,0x1F,0,0x14,0x1B,0x20,0,
    0x15,0x1C,0x21,0,0x16,0x1D,0x22,0,0x17,0x1E,0x23,0,0x18,0x1F,0x24,0,
    0xDA,0xE0,0xE5,0,0xDB,0xE1,0xE6,0,0xDC,0xE2,0xE7,0,0xDD,0xE3,0xE8,0,
    0xDE,0xE4,0xE9,0,0xDF,0xE5,0xEA,0,0xE0,0xE6,0xEB,0,0xE1,0xE7,0xEC,0,
    0xE2,0xE8,0xED,0,0xE3,0xE9,0xEE,0,0xE4,0xEA,0xEF,0,0xE5,0xEB,0xF0,0,
    0xE6,0xEC,0xF1,0,0xE7,0xED,0xF2,0,0xE8,0xEE,0xF3,0,0xE9,0xEF,0xF4,0,
    0xEA,0xF0,0xF5,0,0xEB,0xF1,0xF6,0,0xEC,0xF2,0xF7,0,0xED,0xF3,0xF8,0,
    0xEE,0xF4,0xF9,0,0xEF,0xF5,0xFA,0,0xF0,0xF6,0xFB,0,0xF1,0xF7,0xFC,0,
    0xF2,0xF8,0xFD,0,0xF3,0xF9,0xFE,0,0xF4,0xFA,0xFF,0,0xF5,0xFB,0x00,0,
    0xF6,0xFC,0x01,0,0xF7,0xFD,0x02,0,0xF8,0xFE,0x03,0,0xF9,0xFF,0x04,0,
    0xFA,0x00,0x05,0,0xFB,0x01,0x06,0,0xFC,0x02,0x07,0,0xFD,0x03,0x08,0,
    0xFE,0x04,0x09,0,0xFF,0x05,0x0A,0,0x00,0x06,0x0B,0,0x01,0x07,0x0C,0,
    0x02,0x08,0x0D,0,0x03,0x09,0x0E,0,0x04,0x0A,0x0F,0,0x05,0x0B,0x10,0,
    0x06,0x0C,0x11,0,0x07,0x0D,0x12,0,0x08,0x0E,0x13,0,0x09,0x0F,0x14,0,
    0x0A,0x10,0x15,0,0x0B,0x11,0x16,0,0x0C,0x12,0x17,0,0x0D,0x13,0x18,0,
    0x0E,0x14,0x19,0,0x0F,0x15,0x1A,0,0x10,0x16,0x1B,0,0x11,0x17,0x1C,0,
    0x12,0x18,0x1D,0,0x13,0x19,0x1E,0,0x14,0x1A,0x1F,0,0x15,0x1B,0x20,0,
    0x16,0x1C,0x21,0,0x17,0x1D,0x22,0,0x18,0x1E,0x23,0,0x19,0x1F,0x24,0,
    0xDB,0xE0,0xE5,0,0xDC,0xE1,0xE6,0,0xDD,0xE2,0xE7,0,0xDE,0xE3,0xE8,0,
    0xDF,0xE4,0xE9,0,0xE0,0xE5,0xEA,0,0xE1,0xE6,0xEB,0,0xE2,0xE7,0xEC,0,
    0xE3,0xE8,0xED,0,0xE4,0xE9,0xEE,0,0xE5,0xEA,0xEF,0,0xE6,0xEB,0xF0,0,
    0xE7,0xEC,0xF1,0,0xE8,0xED,0xF2,0,0xE9,0xEE,0xF3,0,0xEA,0xEF,0xF4,0,
    0xEB,0xF0,0xF5,0,0xEC,0xF1,0xF6,0,0xED,0xF2,0xF7,0,0xEE,0xF3,0xF8,0,
    0xEF,0xF4,0xF9,0,0xF0,0xF5,0xFA,0,0xF1,0xF6,0xFB,0,0xF2,0xF7,0xFC,0,
    0xF3,0xF8,0xFD,0,0xF4,0xF9,0xFE,0,0xF5,0xFA,0xFF,0,0xF6,0xFB,0x00,0,
    0xF7,0xFC,0x01,0,0xF8,0xFD,0x02,0,0xF9,0xFE,0x03,0,0xFA,0xFF,0x04,0,
    0xFB,0x00,0x05,0,0xFC,0x01,0x06,0,0xFD,0x02,0x07,0,0xFE,0x03,0x08,0,
    0xFF,0x04,0x09,0,0x00,0x05,0x0A,0,0x01,0x06,0x0B,0,0x02,0x07,0x0C,0,
    0x03,0x08,0x0D,0,0x04,0x09,0x0E,0,0x05,0x0A,0x0F,0,0x06,0x0B,0x10,0,
    0x07,0x0C,0x11,0,0x08,0x0D,0x12,0,0x09,0x0E,0x13,0,0x0A,0x0F,0x14,0,
    0x0B,0x10,0x15,0,0x0C,0x11,0x16,0,0x0D,0x12,0x17,0,0x0E,0x13,0x18,0,
    0x0F,0x14,0x19,0,0x10,0x15,0x1A,0,0x11,0x16,0x1B,0,0x12,0x17,0x1C,0,
    0x13,0x18,0x1D,0,0x14,0x19,0x1E,0,0x15,0x1A,0x1F,0,0x16,0x1B,0x20,0,
    0x17,0x1C,0x21,0,0x18,0x1D,0x22,0,0x19,0x1E,0x23,0,0x1A,0x1F,0x24,0,
    0xDC,0xE0,0xE5,0,0xDD,0xE1,0xE6,0,0xDE,0xE2,0xE7,0,0xDF,0xE3,0xE8,0,
    0xE0,0xE4,0xE9,0,0xE1,0xE5,0xEA,0,0xE2,0xE6,0xEB,0,0xE3,0xE7,0xEC,0,
    0xE4,0xE8,0xED,0,0xE5,0xE9,0xEE,0,0xE6,0xEA,0xEF,0,0xE7,0xEB,0xF0,0,
    0xE8,0xEC,0xF1,0,0xE9,0xED,0xF2,0,0xEA,0xEE,0xF3,0,0xEB,0xEF,0xF4,0,
    0xEC,0xF0,0xF5,0,0xED,0xF1,0xF6,0,0xEE,0xF2,0xF7,0,0xEF,0xF3,0xF8,0,
    0xF0,0xF4,0xF9,0,0xF1,0xF5,0xFA,0,0xF2,0xF6,0xFB,0,0xF3,0xF7,0xFC,0,
    0xF4,0xF8,0xFD,0,0xF5,0xF9,0xFE,0,0xF6,0xFA,0xFF,0,0xF7,0xFB,0x00,0,
    0xF8,0xFC,0x01,0,0xF9,0xFD,0x02,0,0xFA,0xFE,0x03,0,0xFB,0xFF,0x04,0,
    0xFC,0x00,0x05,0,0xFD,0x01,0x06,0,0xFE,0x02,0x07,0,0xFF,0x03,0x08,0,
    0x00,0x04,0x09,0,0x01,0x05,0x0A,0,0x02,0x06,0x0B,0,0x03,0x07,0x0C,0,
    0x04,0x08,0x0D,0,0x05,0x09,0x0E,0,0x06,0x0A,0x0F,0,0x07,0x0B,0x10,0,
    0x08,0x0C,0x11,0,0x09,0x0D,0x12,0,0x0A,0x0E,0x13,0,0x0B,0x0F,0x14,0,
    0x0C,0x10,0x15,0,0x0D,0x11,0x16,0,0x0E,0x12,0x17,0,0x0F,0x13,0x18,0,
    0x10,0x14,0x19,0,0x11,0x15,0x1A,0,0x12,0x16,0x1B,0,0x13,0x17,0x1C,0,
    0x14,0x18,0x1D,0,0x15,0x19,0x1E,0,0x16,0x1A,0x1F,0,0x17,0x1B,0x20,0,
    0x18,0x1C,0x21,0,0x19,0x1D,0x22,0,0x1A,0x1E,0x23,0,0x1B,0x1F,0x24,0,
    0xDD,0xE0,0xE5,0,0xDE,0xE1,0xE6,0,0xDF,0xE2,0xE7,0,0xE0,0xE3,0xE8,0,
    0xE1,0xE4,0xE9,0,0xE2,0xE5,0xEA,0,0xE3,0xE6,0xEB,0,0xE4,0xE7,0xEC,0,
    0xE5,0xE8,0xED,0,0xE6,0xE9,0xEE,0,0xE7,0xEA,0xEF,0,0xE8,0xEB,0xF0,0,
    0xE9,0xEC,0xF1,0,0xEA,0xED,0xF2,0,0xEB,0xEE,0xF3,0,0xEC,0xEF,0xF4,0,
    0xED,0xF0,0xF5,0,0xEE,0xF1,0xF6,0,0xEF,0xF2,0xF7,0,0xF0,0xF3,0xF8,0,
    0xF1,0xF4,0xF9,0,0xF2,0xF5,0xFA,0,0xF3,0xF6,0xFB,0,0xF4,0xF7,0xFC,0,
    0xF5,0xF8,0xFD,0,0xF6,0xF9,0xFE,0,0xF7,0xFA,0xFF,0,0xF8,0xFB,0x00,0,
    0xF9,0xFC,0x01,0,0xFA,0xFD,0x02,0,0xFB,0xFE,0x03,0,0xFC,0xFF,0x04,0,
    0xFD,0x00,0x05,0,0xFE,0x01,0x06,0,0xFF,0x02,0x07,0,0x00,0x03,0x08,0,
    0x01,0x04,0x09,0,0x02,0x05,0x0A,0,0x03,0x06,0x0B,0,0x04,0x07,0x0C,0,
    0x05,0x08,0x0D,0,0x06,0x09,0x0E,0,0x07,0x0A,0x0F,0,0x08,0x0B,0x10,0,
    0x09,0x0C,0x11,0,0x0A,0x0D,0x12,0,0x0B,0x0E,0x13,0,0x0C,0x0F,0x14,0,
    0x0D,0x10,0x15,0,0x0E,0x11,0x16,0,0x0F,0x12,0x17,0,0x10,0x13,0x18,0,
    0x11,0x14,0x19,0,0x12,0x15,0x1A,0,0x13,0x16,0x1B,0,0x14,0x17,0x1C,0,
    0x15,0x18,0x1D,0,0x16,0x19,0x1E,0,0x17,0x1A,0x1F,0,0x18,0x1B,0x20,0,
    0x19,0x1C,0x21,0,0x1A,0x1D,0x22,0,0x1B,0x1E,0x23,0,0x1C,0x1F,0x24,0,
    0xDE,0xE0,0xE5,0,0xDF,0xE1,0xE6,0,0xE0,0xE2,0xE7,0,0xE1,0xE3,0xE8,0,
    0xE2,0xE4,0xE9,0,0xE3,0xE5,0xEA,0,0xE4,0xE6,0xEB,0,0xE5,0xE7,0xEC,0,
    0xE6,0xE8,0xED,0,0xE7,0xE9,0xEE,0,0xE8,0xEA,0xEF,0,0xE9,0xEB,0xF0,0,
    0xEA,0xEC,0xF1,0,0xEB,0xED,0xF2,0,0xEC,0xEE,0xF3,0,0xED,0xEF,0xF4,0,
    0xEE,0xF0,0xF5,0,0xEF,0xF1,0xF6,0,0xF0,0xF2,0xF7,0,0xF1,0xF3,0xF8,0,
    0xF2,0xF4,0xF9,0,0xF3,0xF5,0xFA,0,0xF4,0xF6,0xFB,0,0xF5,0xF7,0xFC,0,
    0xF6,0xF8,0xFD,0,0xF7,0xF9,0xFE,0,0xF8,0xFA,0xFF,0,0xF9,0xFB,0x00,0,
    0xFA,0xFC,0x01,0,0xFB,0xFD,0x02,0,0xFC,0xFE,0x03,0,0xFD,0xFF,0x04,0,
    0xFE,0x00,0x05,0,0xFF,0x01,0x06,0,0x00,0x02,0x07,0,0x01,0x03,0x08,0,
    0x02,0x04,0x09,0,0x03,0x05,0x0A,0,0x04,0x06,0x0B,0,0x05,0x07,0x0C,0,
    0x06,0x08,0x0D,0,0x07,0x09,0x0E,0,0x08,0x0A,0x0F,0,0x09,0x0B,0x10,0,
    0x0A,0x0C,0x11,0,0x0B,0x0D,0x12,0,0x0C,0x0E,0x13,0,0x0D,0x0F,0x14,0,
    0x0E,0x10,0x15,0,0x0F,0x11,0x16,0,0x10,0x12,0x17,0,0x11,0x13,0x18,0,
    0x12,0x14,0x19,0,0x13,0x15,0x1A,0,0x14,0x16,0x1B,0,0x15,0x17,0x1C,0,
    0x16,0x18,0x1D,0,0x17,0x19,0x1E,0,0x18,0x1A,0x1F,0,0x19,0x1B,0x20,0,
    0x1A,0x1C,0x21,0,0x1B,0x1D,0x22,0,0x1C,0x1E,0x23,0,0x1D,0x1F,0x24,0,
    0xDF,0xE0,0xE5,0,0xE0,0xE1,0xE6,0,0xE1,0xE2,0xE7,0,0xE2,0xE3,0xE8,0,
    0xE3,0xE4,0xE9,0,0xE4,0xE5,0xEA,0,0xE5,0xE6,0xEB,0,0xE6,0xE7,0xEC,0,
    0xE7,0xE8,0xED,0,0xE8,0xE9,0xEE,0,0xE9,0xEA,0xEF,0,0xEA,0xEB,0xF0,0,
    0xEB,0xEC,0xF1,0,0xEC,0xED,0xF2,0,0xED,0xEE,0xF3,0,0xEE,0xEF,0xF4,0,
    0xEF,0xF0,0xF5,0,0xF0,0xF1,0xF6,0,0xF1,0xF2,0xF7,0,0xF2,0xF3,0xF8,0,
    0xF3,0xF4,0xF9,0,0xF4,0xF5,0xFA,0,0xF5,0xF6,0xFB,0,0xF6,0xF7,0xFC,0,
    0xF7,0xF8,0xFD,0,0xF8,0xF9,0xFE,0,0xF9,0xFA,0xFF,0,0xFA,0xFB,0x00,0,
    0xFB,0xFC,0x01,0,0xFC,0xFD,0x02,0,0xFD,0xFE,0x03,0,0xFE,0xFF,0x04,0,
    0xFF,0x00,0x05,0,0x00,0x01,0x06,0,0x01,0x02,0x07,0,0x02,0x03,0x08,0,
    0x03,0x04,0x09,0,0x04,0x05,0x0A,0,0x05,0x06,0x0B,0,0x06,0x07,0x0C,0,
    0x07,0x08,0x0D,0,0x08,0x09,0x0E,0,0x09,0x0A,0x0F,0,0x0A,0x0B,0x10,0,
    0x0B,0x0C,0x11,0,0x0C,0x0D,0x12,0,0x0D,0x0E,0x13,0,0x0E,0x0F,0x14,0,
    0x0F,0x10,0x15,0,0x10,0x11,0x16,0,0x11,0x12,0x17,0,0x12,0x13,0x18,0,
    0x13,0x14,0x19,0,0x14,0x15,0x1A,0,0x15,0x16,0x1B,0,0x16,0x17,0x1C,0,
    0x17,0x18,0x1D,0,0x18,0x19,0x1E,0,0x19,0x1A,0x1F,0,0x1A,0x1B,0x20,0,
    0x1B,0x1C,0x21,0,0x1C,0x1D,0x22,0,0x1D,0x1E,0x23,0,0x1E,0x1F,0x24,0,
    0xE0,0xE0,0xE5,0,0xE1,0xE1,0xE6,0,0xE2,0xE2,0xE7,0,0xE3,0xE3,0xE8,0,
    0xE4,0xE4,0xE9,0,0xE5,0xE5,0xEA,0,0xE6,0xE6,0xEB,0,0xE7,0xE7,0xEC,0,
    0xE8,0xE8,0xED,0,0xE9,0xE9,0xEE,0,0xEA,0xEA,0xEF,0,0xEB,0xEB,0xF0,0,
    0xEC,0xEC,0xF1,0,0xED,0xED,0xF2,0,0xEE,0xEE,0xF3,0,0xEF,0xEF,0xF4,0,
    0xF0,0xF0,0xF5,0,0xF1,0xF1,0xF6,0,0xF2,0xF2,0xF7,0,0xF3,0xF3,0xF8,0,
    0xF4,0xF4,0xF9,0,0xF5,0xF5,0xFA,0,0xF6,0xF6,0xFB,0,0xF7,0xF7,0xFC,0,
    0xF8,0xF8,0xFD,0,0xF9,0xF9,0xFE,0,0xFA,0xFA,0xFF,0,0xFB,0xFB,0x00,0,
    0xFC,0xFC,0x01,0,0xFD,0xFD,0x02,0,0xFE,0xFE,0x03,0,0xFF,0xFF,0x04,0,
    0x00,0x00,0x05,0,0x01,0x01,0x06,0,0x02,0x02,0x07,0,0x03,0x03,0x08,0,
    0x04,0x04,0x09,0,0x05,0x05,0x0A,0,0x06,0x06,0x0B,0,0x07,0x07,0x0C,0,
    0x08,0x08,0x0D,0,0x09,0x09,0x0E,0,0x0A,0x0A,0x0F,0,0x0B,0x0B,0x10,0,
    0x0C,0x0C,0x11,0,0x0D,0x0D,0x12,0,0x0E,0x0E,0x13,0,0x0F,0x0F,0x14,0,
    0x10,0x10,0x15,0,0x11,0x11,0x16,0,0x12,0x12,0x17,0,0x13,0x13,0x18,0,
    0x14,0x14,0x19,0,0x15,0x15,0x1A,0,0x16,0x16,0x1B,0,0x17,0x17,0x1C,0,
    0x18,0x18,0x1D,0,0x19,0x19,0x1E,0,0x1A,0x1A,0x1F,0,0x1B,0x1B,0x20,0,
    0x1C,0x1C,0x21,0,0x1D,0x1D,0x22,0,0x1E,0x1E,0x23,0,0x1F,0x1F,0x24,0,
    0xE1,0xE0,0xE5,0,0xE2,0xE1,0xE6,0,0xE3,0xE2,0xE7,0,0xE4,0xE3,0xE8,0,
    0xE5,0xE4,0xE9,0,0xE6,0xE5,0xEA,0,0xE7,0xE6,0xEB,0,0xE8,0xE7,0xEC,0,
    0xE9,0xE8,0xED,0,0xEA,0xE9,0xEE,0,0xEB,0xEA,0xEF,0,0xEC,0xEB,0xF0,0,
    0xED,0xEC,0xF1,0,0xEE,0xED,0xF2,0,0xEF,0xEE,0xF3,0,0xF0,0xEF,0xF4,0,
    0xF1,0xF0,0xF5,0,0xF2,0xF1,0xF6,0,0xF3,0xF2,0xF7,0,0xF4,0xF3,0xF8,0,
    0xF5,0xF4,0xF9,0,0xF6,0xF5,0xFA,0,0xF7,0xF6,0xFB,0,0xF8,0xF7,0xFC,0,
    0xF9,0xF8,0xFD,0,0xFA,0xF9,0xFE,0,0xFB,0xFA,0xFF,0,0xFC,0xFB,0x00,0,
    0xFD,0xFC,0x01,0,0xFE,0xFD,0x02,0,0xFF,0xFE,0x03,0,0x00,0xFF,0x04,0,
    0x01,0x00,0x05,0,0x02,0x01,0x06,0,0x03,0x02,0x07,0,0x04,0x03,0x08,0,
    0x05,0x04,0x09,0,0x06,0x05,0x0A,0,0x07,0x06,0x0B,0,0x08,0x07,0x0C,0,
    0x09,0x08,0x0D,0,0x0A,0x09,0x0E,0,0x0B,0x0A,0x0F,0,0x0C,0x0B,0x10,0,
    0x0D,0x0C,0x11,0,0x0E,0x0D,0x12,0,0x0F,0x0E,0x13,0,0x10,0x0F,0x14,0,
    0x11,0x10,0x15,0,0x12,0x11,0x16,0,0x13,0x12,0x17,0,0x14,0x13,0x18,0,
    0x15,0x14,0x19,0,0x16,0x15,0x1A,0,0x17,0x16,0x1B,0,0x18,0x17,0x1C,0,
    0x19,0x18,0x1D,0,0x1A,0x19,0x1E,0,0x1B,0x1A,0x1F,0,0x1C,0x1B,0x20,0,
    0x1D,0x1C,0x21,0,0x1E,0x1D,0x22,0,0x1F,0x1E,0x23,0,0x20,0x1F,0x24,0,
    0xE2,0xE0,0xE5,0,0xE3,0xE1,0xE6,0,0xE4,0xE2,0xE7,0,0xE5,0xE3,0xE8,0,
    0xE6,0xE4,0xE9,0,0xE7,0xE5,0xEA,0,0xE8,0xE6,0xEB,0,0xE9,0xE7,0xEC,0,
    0xEA,0xE8,0xED,0,0xEB,0xE9,0xEE,0,0xEC,0xEA,0xEF,0,0xED,0xEB,0xF0,0,
    0xEE,0xEC,0xF1,0,0xEF,0xED,0xF2,0,0xF0,0xEE,0xF3,0,0xF1,0xEF,0xF4,0,
    0xF2,0xF0,0xF5,0,0xF3,0xF1,0xF6,0,0xF4,0xF2,0xF7,0,0xF5,0xF3,0xF8,0,
    0xF6,0xF4,0xF9,0,0xF7,0xF5,0xFA,0,0xF8,0xF6,0xFB,0,0xF9,0xF7,0xFC,0,
    0xFA,0xF8,0xFD,0,0xFB,0xF9,0xFE,0,0xFC,0xFA,0xFF,0,0xFD,0xFB,0x00,0,
    0xFE,0xFC,0x01,0,0xFF,0xFD,0x02,0,0x00,0xFE,0x03,0,0x01,0xFF,0x04,0,
    0x02,0x00,0x05,0,0x03,0x01,0x06,0,0x04,0x02,0x07,0,0x05,0x03,0x08,0,
    0x06,0x04,0x09,0,0x07,0x05,0x0A,0,0x08,0x06,0x0B,0,0x09,0x07,0x0C,0,
    0x0A,0x08,0x0D,0,0x0B,0x09,0x0E,0,0x0C,0x0A,0x0F,0,0x0D,0x0B,0x10,0,
    0x0E,0x0C,0x11,0,0x0F,0x0D,0x12,0,0x10,0x0E,0x13,0,0x11,0x0F,0x14,0,
    0x12,0x10,0x15,0,0x13,0x11,0x16,0,0x14,0x12,0x17,0,0x15,0x13,0x18,0,
    0x16,0x14,0x19,0,0x17,0x15,0x1A,0,0x18,0x16,0x1B,0,0x19,0x17,0x1C,0,
    0x1A,0x18,0x1D,0,0x1B,0x19,0x1E,0,0x1C,0x1A,0x1F,0,0x1D,0x1B,0x20,0,
    0x1E,0x1C,0x21,0,0x1F,0x1D,0x22,0,0x20,0x1E,0x23,0,0x21,0x1F,0x24,0,
    0xE3,0xE0,0xE5,0,0xE4,0xE1,0xE6,0,0xE5,0xE2,0xE7,0,0xE6,0xE3,0xE8,0,
    0xE7,0xE4,0xE9,0,0xE8,0xE5,0xEA,0,0xE9,0xE6,0xEB,0,0xEA,0xE7,0xEC,0,
    0xEB,0xE8,0xED,0,0xEC,0xE9,0xEE,0,0xED,0xEA,0xEF,0,0xEE,0xEB,0xF0,0,
    0xEF,0xEC,0xF1,0,0xF0,0xED,0xF2,0,0xF1,0xEE,0xF3,0,0xF2,0xEF,0xF4,0,
    0xF3,0xF0,0xF5,0,0xF4,0xF1,0xF6,0,0xF5,0xF2,0xF7,0,0xF6,0xF3,0xF8,0,
    0xF7,0xF4,0xF9,0,0xF8,0xF5,0xFA,0,0xF9,0xF6,0xFB,0,0xFA,0xF7,0xFC,0,
    0xFB,0xF8,0xFD,0,0xFC,0xF9,0xFE,0,0xFD,0xFA,0xFF,0,0xFE,0xFB,0x00,0,
    0xFF,0xFC,0x01,0,0x00,0xFD,0x02,0,0x01,0xFE,0x03,0,0x02,0xFF,0x04,0,
    0x03,0x00,0x05,0,0x04,0x01,0x06,0,0x05,0x02,0x07,0,0x06,0x03,0x08,0,
    0x07,0x04,0x09,0,0x08,0x05,0x0A,0,0x09,0x06,0x0B,0,0x0A,0x07,0x0C,0,
    0x0B,0x08,0x0D,0,0x0C,0x09,0x0E,0,0x0D,0x0A,0x0F,0,0x0E,0x0B,0x10,0,
    0x0F,0x0C,0x11,0,0x10,0x0D,0x12,0,0x11,0x0E,0x13,0,0x12,0x0F,0x14,0,
    0x13,0x10,0x15,0,0x14,0x11,0x16,0,0x15,0x12,0x17,0,0x16,0x13,0x18,0,
    0x17,0x14,0x19,0,0x18,0x15,0x1A,0,0x19,0x16,0x1B,0,0x1A,0x17,0x1C,0,
    0x1B,0x18,0x1D,0,0x1C,0x19,0x1E,0,0x1D,0x1A,0x1F,0,0x1E,0x1B,0x20,0,
    0x1F,0x1C,0x21,0,0x20,0x1D,0x22,0,0x21,0x1E,0x23,0,0x22,0x1F,0x24,0,
    0xE4,0xE0,0xE5,0,0xE5,0xE1,0xE6,0,0xE6,0xE2,0xE7,0,0xE7,0xE3,0xE8,0,
    0xE8,0xE4,0xE9,0,0xE9,0xE5,0xEA,0,0xEA,0xE6,0xEB,0,0xEB,0xE7,0xEC,0,
    0xEC,0xE8,0xED,0,0xED,0xE9,0xEE,0,0xEE,0xEA,0xEF,0,0xEF,0xEB,0xF0,0,
    0xF0,0xEC,0xF1,0,0xF1,0xED,0xF2,0,0xF2,0xEE,0xF3,0,0xF3,0xEF,0xF4,0,
    0xF4,0xF0,0xF5,0,0xF5,0xF1,0xF6,0,0xF6,0xF2,0xF7,0,0xF7,0xF3,0xF8,0,
    0xF8,0xF4,0xF9,0,0xF9,0xF5,0xFA,0,0xFA,0xF6,0xFB,0,0xFB,0xF7,0xFC,0,
    0xFC,0xF8,0xFD,0,0xFD,0xF9,0xFE,0,0xFE,0xFA,0xFF,0,0xFF,0xFB,0x00,0,
    0x00,0xFC,0x01,0,0x01,0xFD,0x02,0,0x02,0xFE,0x03,0,0x03,0xFF,0x04,0,
    0x04,0x00,0x05,0,0x05,0x01,0x06,0,0x06,0x02,0x07,0,0x07,0x03,0x08,0,
    0x08,0x04,0x09,0,0x09,0x05,0x0A,0,0x0A,0x06,0x0B,0,0x0B,0x07,0x0C,0,
    0x0C,0x08,0x0D,0,0x0D,0x09,0x0E,0,0x0E,0x0A,0x0F,0,0x0F,0x0B,0x10,0,
    0x10,0x0C,0x11,0,0x11,0x0D,0x12,0,0x12,0x0E,0x13,0,0x13,0x0F,0x14,0,
    0x14,0x10,0x15,0,0x15,0x11,0x16,0,0x16,0x12,0x17,0,0x17,0x13,0x18,0,
    0x18,0x14,0x19,0,0x19,0x15,0x1A,0,0x1A,0x16,0x1B,0,0x1B,0x17,0x1C,0,
    0x1C,0x18,0x1D,0,0x1D,0x19,0x1E,0,0x1E,0x1A,0x1F,0,0x1F,0x1B,0x20,0,
    0x20,0x1C,0x21,0,0x21,0x1D,0x22,0,0x22,0x1E,0x23,0,0x23,0x1F,0x24,0,
    0xE5,0xE0,0xE5,0,0xE6,0xE1,0xE6,0,0xE7,0xE2,0xE7,0,0xE8,0xE3,0xE8,0,
    0xE9,0xE4,0xE9,0,0xEA,0xE5,0xEA,0,0xEB,0xE6,0xEB,0,0xEC,0xE7,0xEC,0,
    0xED,0xE8,0xED,0,0xEE,0xE9,0xEE,0,0xEF,0xEA,0xEF,0,0xF0,0xEB,0xF0,0,
    0xF1,0xEC,0xF1,0,0xF2,0xED,0xF2,0,0xF3,0xEE,0xF3,0,0xF4,0xEF,0xF4,0,
    0xF5,0xF0,0xF5,0,0xF6,0xF1,0xF6,0,0xF7,0xF2,0xF7,0,0xF8,0xF3,0xF8,0,
    0xF9,0xF4,0xF9,0,0xFA,0xF5,0xFA,0,0xFB,0xF6,0xFB,0,0xFC,0xF7,0xFC,0,
    0xFD,0xF8,0xFD,0,0xFE,0xF9,0xFE,0,0xFF,0xFA,0xFF,0,0x00,0xFB,0x00,0,
    0x01,0xFC,0x01,0,0x02,0xFD,0x02,0,0x03,0xFE,0x03,0,0x04,0xFF,0x04,0,
    0x05,0x00,0x05,0,0x06,0x01,0x06,0,0x07,0x02,0x07,0,0x08,0x03,0x08,0,
    0x09,0x04,0x09,0,0x0A,0x05,0x0A,0,0x0B,0x06,0x0B,0,0x0C,0x07,0x0C,0,
    0x0D,0x08,0x0D,0,0x0E,0x09,0x0E,0,0x0F,0x0A,0x0F,0,0x10,0x0B,0x10,0,
    0x11,0x0C,0x11,0,0x12,0x0D,0x12,0,0x13,0x0E,0x13,0,0x14,0x0F,0x14,0,
    0x15,0x10,0x15,0,0x16,0x11,0x16,0,0x17,0x12,0x17,0,0x18,0x13,0x18,0,
    0x19,0x14,0x19,0,0x1A,0x15,0x1A,0,0x1B,0x16,0x1B,0,0x1C,0x17,0x1C,0,
    0x1D,0x18,0x1D,0,0x1E,0x19,0x1E,0,0x1F,0x1A,0x1F,0,0x20,0x1B,0x20,0,
    0x21,0x1C,0x21,0,0x22,0x1D,0x22,0,0x23,0x1E,0x23,0,0x24,0x1F,0x24,0,
    0xE6,0xE0,0xE5,0,0xE7,0xE1,0xE6,0,0xE8,0xE2,0xE7,0,0xE9,0xE3,0xE8,0,
    0xEA,0xE4,0xE9,0,0xEB,0xE5,0xEA,0,0xEC,0xE6,0xEB,0,0xED,0xE7,0xEC,0,
    0xEE,0xE8,0xED,0,0xEF,0xE9,0xEE,0,0xF0,0xEA,0xEF,0,0xF1,0xEB,0xF0,0,
    0xF2,0xEC,0xF1,0,0xF3,0xED,0xF2,0,0xF4,0xEE,0xF3,0,0xF5,0xEF,0xF4,0,
    0xF6,0xF0,0xF5,0,0xF7,0xF1,0xF6,0,0xF8,0xF2,0xF7,0,0xF9,0xF3,0xF8,0,
    0xFA,0xF4,0xF9,0,0xFB,0xF5,0xFA,0,0xFC,0xF6,0xFB,0,0xFD,0xF7,0xFC,0,
    0xFE,0xF8,0xFD,0,0xFF,0xF9,0xFE,0,0x00,0xFA,0xFF,0,0x01,0xFB,0x00,0,
    0x02,0xFC,0x01,0,0x03,0xFD,0x02,0,0x04,0xFE,0x03,0,0x05,0xFF,0x04,0,
    0x06,0x00,0x05,0,0x07,0x01,0x06,0,0x08,0x02,0x07,0,0x09,0x03,0x08,0,
    0x0A,0x04,0x09,0,0x0B,0x05,0x0A,0,0x0C,0x06,0x0B,0,0x0D,0x07,0x0C,0,
    0x0E,0x08,0x0D,0,0x0F,0x09,0x0E,0,0x10,0x0A,0x0F,0,0x11,0x0B,0x10,0,
    0x12,0x0C,0x11,0,0x13,0x0D,0x12,0,0x14,0x0E,0x13,0,0x15,0x0F,0x14,0,
    0x16,0x10,0x15,0,0x17,0x11,0x16,0,0x18,0x12,0x17,0,0x19,0x13,0x18,0,
    0x1A,0x14,0x19,0,0x1B,0x15,0x1A,0,0x1C,0x16,0x1B,0,0x1D,0x17,0x1C,0,
    0x1E,0x18,0x1D,0,0x1F,0x19,0x1E,0,0x20,0x1A,0x1F,0,0x21,0x1B,0x20,0,
    0x22,0x1C,0x21,0,0x23,0x1D,0x22,0,0x24,0x1E,0x23,0,0x25,0x1F,0x24,0,
    0xE7,0xE0,0xE5,0,0xE8,0xE1,0xE6,0,0xE9,0xE2,0xE7,0,0xEA,0xE3,0xE8,0,
    0xEB,0xE4,0xE9,0,0xEC,0xE5,0xEA,0,0xED,0xE6,0xEB,0,0xEE,0xE7,0xEC,0,
    0xEF,0xE8,0xED,0,0xF0,0xE9,0xEE,0,0xF1,0xEA,0xEF,0,0xF2,0xEB,0xF0,0,
    0xF3,0xEC,0xF1,0,0xF4,0xED,0xF2,0,0xF5,0xEE,0xF3,0,0xF6,0xEF,0xF4,0,
    0xF7,0xF0,0xF5,0,0xF8,0xF1,0xF6,0,0xF9,0xF2,0xF7,0,0xFA,0xF3,0xF8,0,
    0xFB,0xF4,0xF9,0,0xFC,0xF5,0xFA,0,0xFD,0xF6,0xFB,0,0xFE,0xF7,0xFC,0,
    0xFF,0xF8,0xFD,0,0x00,0xF9,0xFE,0,0x01,0xFA,0xFF,0,0x02,0xFB,0x00,0,
    0x03,0xFC,0x01,0,0x04,0xFD,0x02,0,0x05,0xFE,0x03,0,0x06,0xFF,0x04,0,
    0x07,0x00,0x05,0,0x08,0x01,0x06,0,0x09,0x02,0x07,0,0x0A,0x03,0x08,0,
    0x0B,0x04,0x09,0,0x0C,0x05,0x0A,0,0x0D,0x06,0x0B,0,0x0E,0x07,0x0C,0,
    0x0F,0x08,0x0D,0,0x10,0x09,0x0E,0,0x11,0x0A,0x0F,0,0x12,0x0B,0x10,0,
    0x13,0x0C,0x11,0,0x14,0x0D,0x12,0,0x15,0x0E,0x13,0,0x16,0x0F,0x14,0,
    0x17,0x10,0x15,0,0x18,0x11,0x16,0,0x19,0x12,0x17,0,0x1A,0x13,0x18,0,
    0x1B,0x14,0x19,0,0x1C,0x15,0x1A,0,0x1D,0x16,0x1B,0,0x1E,0x17,0x1C,0,
    0x1F,0x18,0x1D,0,0x20,0x19,0x1E,0,0x21,0x1A,0x1F,0,0x22,0x1B,0x20,0,
    0x23,0x1C,0x21,0,0x24,0x1D,0x22,0,0x25,0x1E,0x23,0,0x26,0x1F,0x24,0,
    0xD8,0xE0,0xE6,0,0xD9,0xE1,0xE7,0,0xDA,0xE2,0xE8,0,0xDB,0xE3,0xE9,0,
    0xDC,0xE4,0xEA,0,0xDD,0xE5,0xEB,0,0xDE,0xE6,0xEC,0,0xDF,0xE7,0xED,0,
    0xE0,0xE8,0xEE,0,0xE1,0xE9,0xEF,0,0xE2,0xEA,0xF0,0,0xE3,0xEB,0xF1,0,
    0xE4,0xEC,0xF2,0,0xE5,0xED,0xF3,0,0xE6,0xEE,0xF4,0,0xE7,0xEF,0xF5,0,
    0xE8,0xF0,0xF6,0,0xE9,0xF1,0xF7,0,0xEA,0xF2,0xF8,0,0xEB,0xF3,0xF9,0,
    0xEC,0xF4,0xFA,0,0xED,0xF5,0xFB,0,0xEE,0xF6,0xFC,0,0xEF,0xF7,0xFD,0,
    0xF0,0xF8,0xFE,0,0xF1,0xF9,0xFF,0,0xF2,0xFA,0x00,0,0xF3,0xFB,0x01,0,
    0xF4,0xFC,0x02,0,0xF5,0xFD,0x03,0,0xF6,0xFE,0x04,0,0xF7,0xFF,0x05,0,
    0xF8,0x00,0x06,0,0xF9,0x01,0x07,0,0xFA,0x02,0x08,0,0xFB,0x03,0x09,0,
    0xFC,0x04,0x0A,0,0xFD,0x05,0x0B,0,0xFE,0x06,0x0C,0,0xFF,0x07,0x0D,0,
    0x00,0x08,0x0E,0,0x01,0x09,0x0F,0,0x02,0x0A,0x10,0,0x03,0x0B,0x11,0,
    0x04,0x0C,0x12,0,0x05,0x0D,0x13,0,0x06,0x0E,0x14,0,0x07,0x0F,0x15,0,
    0x08,0x10,0x16,0,0x09,0x11,0x17,0,0x0A,0x12,0x18,0,0x0B,0x13,0x19,0,
    0x0C,0x14,0x1A,0,0x0D,0x15,0x1B,0,0x0E,0x16,0x1C,0,0x0F,0x17,0x1D,0,
    0x10,0x18,0x1E,0,0x11,0x19,0x1F,0,0x12,0x1A,0x20,0,0x13,0x1B,0x21,0,
    0x14,0x1C,0x22,0,0x15,0x1D,0x23,0,0x16,0x1E,0x24,0,0x17,0x1F,0x25,0,
    0xD9,0xE0,0xE6,0,0xDA,0xE1,0xE7,0,0xDB,0xE2,0xE8,0,0xDC,0xE3,0xE9,0,
    0xDD,0xE4,0xEA,0,0xDE,0xE5,0xEB,0,0xDF,0xE6,0xEC,0,0xE0,0xE7,0xED,0,
    0xE1,0xE8,0xEE,0,0xE2,0xE9,0xEF,0,0xE3,0xEA,0xF0,0,0xE4,0xEB,0xF1,0,
    0xE5,0xEC,0xF2,0,0xE6,0xED,0xF3,0,0xE7,0xEE,0xF4,0,0xE8,0xEF,0xF5,0,
    0xE9,0xF0,0xF6,0,0xEA,0xF1,0xF7,0,0xEB,0xF2,0xF8,0,0xEC,0xF3,0xF9,0,
    0xED,0xF4,0xFA,0,0xEE,0xF5,0xFB,0,0xEF,0xF6,0xFC,0,0xF0,0xF7,0xFD,0,
    0xF1,0xF8,0xFE,0,0xF2,0xF9,0xFF,0,0xF3,0xFA,0x00,0,0xF4,0xFB,0x01,0,
    0xF5,0xFC,0x02,0,0xF6,0xFD,0x03,0,0xF7,0xFE,0x04,0,0xF8,0xFF,0x05,0,
    0xF9,0x00,0x06,0,0xFA,0x01,0x07,0,0xFB,0x02,0x08,0,0xFC,0x03,0x09,0,
    0xFD,0x04,0x0A,0,0xFE,0x05,0x0B,0,0xFF,0x06,0x0C,0,0x00,0x07,0x0D,0,
    0x01,0x08,0x0E,0,0x02,0x09,0x0F,0,0x03,0x0A,0x10,0,0x04,0x0B,0x11,0,
    0x05,0x0C,0x12,0,0x06,0x0D,0x13,0,0x07,0x0E,0x14,0,0x08,0x0F,0x15,0,
    0x09,0x10,0x16,0,0x0A,0x11,0x17,0,0x0B,0x12,0x18,0,0x0C,0x13,0x19,0,
    0x0D,0x14,0x1A,0,0x0E,0x15,0x1B,0,0x0F,0x16,0x1C,0,0x10,0x17,0x1D,0,
    0x11,0x18,0x1E,0,0x12,0x19,0x1F,0,0x13,0x1A,0x20,0,0x14,0x1B,0x21,0,
    0x15,0x1C,0x22,0,0x16,0x1D,0x23,0,0x17,0x1E,0x24,0,0x18,0x1F,0x25,0,
    0xDA,0xE0,0xE6,0,0xDB,0xE1,0xE7,0,0xDC,0xE2,0xE8,0,0xDD,0xE3,0xE9,0,
    0xDE,0xE4,0xEA,0,0xDF,0xE5,0xEB,0,0xE0,0xE6,0xEC,0,0xE1,0xE7,0xED,0,
    0xE2,0xE8,0xEE,0,0xE3,0xE9,0xEF,0,0xE4,0xEA,0xF0,0,0xE5,0xEB,0xF1,0,
    0xE6,0xEC,0xF2,0,0xE7,0xED,0xF3,0,0xE8,0xEE,0xF4,0,0xE9,0xEF,0xF5,0,
    0xEA,0xF0,0xF6,0,0xEB,0xF1,0xF7,0,0xEC,0xF2,0xF8,0,0xED,0xF3,0xF9,0,
    0xEE,0xF4,0xFA,0,0xEF,0xF5,0xFB,0,0xF0,0xF6,0xFC,0,0xF1,0xF7,0xFD,0,
    0xF2,0xF8,0xFE,0,0xF3,0xF9,0xFF,0,0xF4,0xFA,0x00,0,0xF5,0xFB,0x01,0,
    0xF6,0xFC,0x02,0,0xF7,0xFD,0x03,0,0xF8,0xFE,0x04,0,0xF9,0xFF,0x05,0,
    0xFA,0x00,0x06,0,0xFB,0x01,0x07,0,0xFC,0x02,0x08,0,0xFD,0x03,0x09,0,
    0xFE,0x04,0x0A,0,0xFF,0x05,0x0B,0,0x00,0x06,0x0C,0,0x01,0x07,0x0D,0,
    0x02,0x08,0x0E,0,0x03,0x09,0x0F,0,0x04,0x0A,0x10,0,0x05,0x0B,0x11,0,
    0x06,0x0C,0x12,0,0x07,0x0D,0x13,0,0x08,0x0E,0x14,0,0x09,0x0F,0x15,0,
    0x0A,0x10,0x16,0,0x0B,0x11,0x17,0,0x0C,0x12,0x18,0,0x0D,0x13,0x19,0,
    0x0E,0x14,0x1A,0,0x0F,0x15,0x1B,0,0x10,0x16,0x1C,0,0x11,0x17,0x1D,0,
    0x12,0x18,0x1E,0,0x13,0x19,0x1F,0,0x14,0x1A,0x20,0,0x15,0x1B,0x21,0,
    0x16,0x1C,0x22,0,0x17,0x1D,0x23,0,0x18,0x1E,0x24,0,0x19,0x1F,0x25,0,
    0xDB,0xE0,0xE6,0,0xDC,0xE1,0xE7,0,0xDD,0xE2,0xE8,0,0xDE,0xE3,0xE9,0,
    0xDF,0xE4,0xEA,0,0xE0,0xE5,0xEB,0,0xE1,0xE6,0xEC,0,0xE2,0xE7,0xED,0,
    0xE3,0xE8,0xEE,0,0xE4,0xE9,0xEF,0,0xE5,0xEA,0xF0,0,0xE6,0xEB,0xF1,0,
    0xE7,0xEC,0xF2,0,0xE8,0xED,0xF3,0,0xE9,0xEE,0xF4,0,0xEA,0xEF,0xF5,0,
    0xEB,0xF0,0xF6,0,0xEC,0xF1,0xF7,0,0xED,0xF2,0xF8,0,0xEE,0xF3,0xF9,0,
    0xEF,0xF4,0xFA,0,0xF0,0xF5,0xFB,0,0xF1,0xF6,0xFC,0,0xF2,0xF7,0xFD,0,
    0xF3,0xF8,0xFE,0,0xF4,0xF9,0xFF,0,0xF5,0xFA,0x00,0,0xF6,0xFB,0x01,0,
    0xF7,0xFC,0x02,0,0xF8,0xFD,0x03,0,0xF9,0xFE,0x04,0,0xFA,0xFF,0x05,0,
    0xFB,0x00,0x06,0,0xFC,0x01,0x07,0,0xFD,0x02,0x08,0,0xFE,0x03,0x09,0,
    0xFF,0x04,0x0A,0,0x00,0x05,0x0B,0,0x01,0x06,0x0C,0,0x02,0x07,0x0D,0,
    0x03,0x08,0x0E,0,0x04,0x09,0x0F,0,0x05,0x0A,0x10,0,0x06,0x0B,0x11,0,
    0x07,0x0C,0x12,0,0x08,0x0D,0x13,0,0x09,0x0E,0x14,0,0x0A,0x0F,0x15,0,
    0x0B,0x10,0x16,0,0x0C,0x11,0x17,0,0x0D,0x12,0x18,0,0x0E,0x13,0x19,0,
    0x0F,0x14,0x1A,0,0x10,0x15,0x1B,0,0x11,0x16,0x1C,0,0x12,0x17,0x1D,0,
    0x13,0x18,0x1E,0,0x14,0x19,0x1F,0,0x15,0x1A,0x20,0,0x16,0x1B,0x21,0,
    0x17,0x1C,0x22,0,0x18,0x1D,0x23,0,0x19,0x1E,0x24,0,0x1A,0x1F,0x25,0,
    0xDC,0xE0,0xE6,0,0xDD,0xE1,0xE7,0,0xDE,0xE2,0xE8,0,0xDF,0xE3,0xE9,0,
    0xE0,0xE4,0xEA,0,0xE1,0xE5,0xEB,0,0xE2,0xE6,0xEC,0,0xE3,0xE7,0xED,0,
    0xE4,0xE8,0xEE,0,0xE5,0xE9,0xEF,0,0xE6,0xEA,0xF0,0,0xE7,0xEB,0xF1,0,
    0xE8,0xEC,0xF2,0,0xE9,0xED,0xF3,0,0xEA,0xEE,0xF4,0,0xEB,0xEF,0xF5,0,
    0xEC,0xF0,0xF6,0,0xED,0xF1,0xF7,0,0xEE,0xF2,0xF8,0,0xEF,0xF3,0xF9,0,
    0xF0,0xF4,0xFA,0,0xF1,0xF5,0xFB,0,0xF2,0xF6,0xFC,0,0xF3,0xF7,0xFD,0,
    0xF4,0xF8,0xFE,0,0xF5,0xF9,0xFF,0,0xF6,0xFA,0x00,0,0xF7,0xFB,0x01,0,
    0xF8,0xFC,0x02,0,0xF9,0xFD,0x03,0,0xFA,0xFE,0x04,0,0xFB,0xFF,0x05,0,
    0xFC,0x00,0x06,0,0xFD,0x01,0x07,0,0xFE,0x02,0x08,0,0xFF,0x03,0x09,0,
    0x00,0x04,0x0A,0,0x01,0x05,0x0B,0,0x02,0x06,0x0C,0,0x03,0x07,0x0D,0,
    0x04,0x08,0x0E,0,0x05,0x09,0x0F,0,0x06,0x0A,0x10,0,0x07,0x0B,0x11,0,
    0x08,0x0C,0x12,0,0x09,0x0D,0x13,0,0x0A,0x0E,0x14,0,0x0B,0x0F,0x15,0,
    0x0C,0x10,0x16,0,0x0D,0x11,0x17,0,0x0E,0x12,0x18,0,0x0F,0x13,0x19,0,
    0x10,0x14,0x1A,0,0x11,0x15,0x1B,0,0x12,0x16,0x1C,0,0x13,0x17,0x1D,0,
    0x14,0x18,0x1E,0,0x15,0x19,0x1F,0,0x16,0x1A,0x20,0,0x17,0x1B,0x21,0,
    0x18,0x1C,0x22,0,0x19,0x1D,0x23,0,0x1A,0x1E,0x24,0,0x1B,0x1F,0x25,0,
    0xDD,0xE0,0xE6,0,0xDE,0xE1,0xE7,0,0xDF,0xE2,0xE8,0,0xE0,0xE3,0xE9,0,
    0xE1,0xE4,0xEA,0,0xE2,0xE5,0xEB,0,0xE3,0xE6,0xEC,0,0xE4,0xE7,0xED,0,
    0xE5,0xE8,0xEE,0,0xE6,0xE9,0xEF,0,0xE7,0xEA,0xF0,0,0xE8,0xEB,0xF1,0,
    0xE9,0xEC,0xF2,0,0xEA,0xED,0xF3,0,0xEB,0xEE,0xF4,0,0xEC,0xEF,0xF5,0,
    0xED,0xF0,0xF6,0,0xEE,0xF1,0xF7,0,0xEF,0xF2,0xF8,0,0xF0,0xF3,0xF9,0,
    0xF1,0xF4,0xFA,0,0xF2,0xF5,0xFB,0,0xF3,0xF6,0xFC,0,0xF4,0xF7,0xFD,0,
    0xF5,0xF8,0xFE,0,0xF6,0xF9,0xFF,0,0xF7,0xFA,0x00,0,0xF8,0xFB,0x01,0,
    0xF9,0xFC,0x02,0,0xFA,0xFD,0x03,0,0xFB,0xFE,0x04,0,0xFC,0xFF,0x05,0,
    0xFD,0x00,0x06,0,0xFE,0x01,0x07,0,0xFF,0x02,0x08,0,0x00,0x03,0x09,0,
    0x01,0x04,0x0A,0,0x02,0x05,0x0B,0,0x03,0x06,0x0C,0,0x04,0x07,0x0D,0,
    0x05,0x08,0x0E,0,0x06,0x09,0x0F,0,0x07,0x0A,0x10,0,0x08,0x0B,0x11,0,
    0x09,0x0C,0x12,0,0x0A,0x0D,0x13,0,0x0B,0x0E,0x14,0,0x0C,0x0F,0x15,0,
    0x0D,0x10,0x16,0,0x0E,0x11,0x17,0,0x0F,0x12,0x18,0,0x10,0x13,0x19,0,
    0x11,0x14,0x1A,0,0x12,0x15,0x1B,0,0x13,0x16,0x1C,0,0x14,0x17,0x1D,0,
    0x15,0x18,0x1E,0,0x16,0x19,0x1F,0,0x17,0x1A,0x20,0,0x18,0x1B,0x21,0,
    0x19,0x1C,0x22,0,0x1A,0x1D,0x23,0,0x1B,0x1E,0x24,0,0x1C,0x1F,0x25,0,
    0xDE,0xE0,0xE6,0,0xDF,0xE1,0xE7,0,0xE0,0xE2,0xE8,0,0xE1,0xE3,0xE9,0,
    0xE2,0xE4,0xEA,0,0xE3,0xE5,0xEB,0,0xE4,0xE6,0xEC,0,0xE5,0xE7,0xED,0,
    0xE6,0xE8,0xEE,0,0xE7,0xE9,0xEF,0,0xE8,0xEA,0xF0,0,0xE9,0xEB,0xF1,0,
    0xEA,0xEC,0xF2,0,0xEB,0xED,0xF3,0,0xEC,0xEE,0xF4,0,0xED,0xEF,0xF5,0,
    0xEE,0xF0,0xF6,0,0xEF,0xF1,0xF7,0,0xF0,0xF2,0xF8,0,0xF1,0xF3,0xF9,0,
    0xF2,0xF4,0xFA,0,0xF3,0xF5,0xFB,0,0xF4,0xF6,0xFC,0,0xF5,0xF7,0xFD,0,
    0xF6,0xF8,0xFE,0,0xF7,0xF9,0xFF,0,0xF8,0xFA,0x00,0,0xF9,0xFB,0x01,0,
    0xFA,0xFC,0x02,0,0xFB,0xFD,0x03,0,0xFC,0xFE,0x04,0,0xFD,0xFF,0x05,0,
    0xFE,0x00,0x06,0,0xFF,0x01,0x07,0,0x00,0x02,0x08,0,0x01,0x03,0x09,0,
    0x02,0x04,0x0A,0,0x03,0x05,0x0B,0,0x04,0x06,0x0C,0,0x05,0x07,0x0D,0,
    0x06,0x08,0x0E,0,0x07,0x09,0x0F,0,0x08,0x0A,0x10,0,0x09,0x0B,0x11,0,
    0x0A,0x0C,0x12,0,0x0B,0x0D,0x13,0,0x0C,0x0E,0x14,0,0x0D,0x0F,0x15,0,
    0x0E,0x10,0x16,0,0x0F,0x11,0x17,0,0x10,0x12,0x18,0,0x11,0x13,0x19,0,
    0x12,0x14,0x1A,0,0x13,0x15,0x1B,0,0x14,0x16,0x1C,0,0x15,0x17,0x1D,0,
    0x16,0x18,0x1E,0,0x17,0x19,0x1F,0,0x18,0x1A,0x20,0,0x19,0x1B,0x21,0,
    0x1A,0x1C,0x22,0,0x1B,0x1D,0x23,0,0x1C,0x1E,0x24,0,0x1D,0x1F,0x25,0,
    0xDF,0xE0,0xE6,0,0xE0,0xE1,0xE7,0,0xE1,0xE2,0xE8,0,0xE2,0xE3,0xE9,0,
    0xE3,0xE4,0xEA,0,0xE4,0xE5,0xEB,0,0xE5,0xE6,0xEC,0,0xE6,0xE7,0xED,0,
    0xE7,0xE8,0xEE,0,0xE8,0xE9,0xEF,0,0xE9,0xEA,0xF0,0,0xEA,0xEB,0xF1,0,
    0xEB,0xEC,0xF2,0,0xEC,0xED,0xF3,0,0xED,0xEE,0xF4,0,0xEE,0xEF,0xF5,0,
    0xEF,0xF0,0xF6,0,0xF0,0xF1,0xF7,0,0xF1,0xF2,0xF8,0,0xF2,0xF3,0xF9,0,
    0xF3,0xF4,0xFA,0,0xF4,0xF5,0xFB,0,0xF5,0xF6,0xFC,0,0xF6,0xF7,0xFD,0,
    0xF7,0xF8,0xFE,0,0xF8,0xF9,0xFF,0,0xF9,0xFA,0x00,0,0xFA,0xFB,0x01,0,
    0xFB,0xFC,0x02,0,0xFC,0xFD,0x03,0,0xFD,0xFE,0x04,0,0xFE,0xFF,0x05,0,
    0xFF,0x00,0x06,0,0x00,0x01,0x07,0,0x01,0x02,0x08,0,0x02,0x03,0x09,0,
    0x03,0x04,0x0A,0,0x04,0x05,0x0B,0,0x05,0x06,0x0C,0,0x06,0x07,0x0D,0,
    0x07,0x08,0x0E,0,0x08,0x09,0x0F,0,0x09,0x0A,0x10,0,0x0A,0x0B,0x11,0,
    0x0B,0x0C,0x12,0,0x0C,0x0D,0x13,0,0x0D,0x0E,0x14,0,0x0E,0x0F,0x15,0,
    0x0F,0x10,0x16,0,0x10,0x11,0x17,0,0x11,0x12,0x18,0,0x12,0x13,0x19,0,
    0x13,0x14,0x1A,0,0x14,0x15,0x1B,0,0x15,0x16,0x1C,0,0x16,0x17,0x1D,0,
    0x17,0x18,0x1E,0,0x18,0x19,0x1F,0,0x19,0x1A,0x20,0,0x1A,0x1B,0x21,0,
    0x1B,0x1C,0x22,0,0x1C,0x1D,0x23,0,0x1D,0x1E,0x24,0,0x1E,0x1F,0x25,0,
    0xE0,0xE0,0xE6,0,0xE1,0xE1,0xE7,0,0xE2,0xE2,0xE8,0,0xE3,0xE3,0xE9,0,
    0xE4,0xE4,0xEA,0,0xE5,0xE5,0xEB,0,0xE6,0xE6,0xEC,0,0xE7,0xE7,0xED,0,
    0xE8,0xE8,0xEE,0,0xE9,0xE9,0xEF,0,0xEA,0xEA,0xF0,0,0xEB,0xEB,0xF1,0,
    0xEC,0xEC,0xF2,0,0xED,0xED,0xF3,0,0xEE,0xEE,0xF4,0,0xEF,0xEF,0xF5,0,
    0xF0,0xF0,0xF6,0,0xF1,0xF1,0xF7,0,0xF2,0xF2,0xF8,0,0xF3,0xF3,0xF9,0,
    0xF4,0xF4,0xFA,0,0xF5,0xF5,0xFB,0,0xF6,0xF6,0xFC,0,0xF7,0xF7,0xFD,0,
    0xF8,0xF8,0xFE,0,0xF9,0xF9,0xFF,0,0xFA,0xFA,0x00,0,0xFB,0xFB,0x01,0,
    0xFC,0xFC,0x02,0,0xFD,0xFD,0x03,0,0xFE,0xFE,0x04,0,0xFF,0xFF,0x05,0,
    0x00,0x00,0x06,0,0x01,0x01,0x07,0,0x02,0x02,0x08,0,0x03,0x03,0x09,0,
    0x04,0x04,0x0A,0,0x05,0x05,0x0B,0,0x06,0x06,0x0C,0,0x07,0x07,0x0D,0,
    0x08,0x08,0x0E,0,0x09,0x09,0x0F,0,0x0A,0x0A,0x10,0,0x0B,0x0B,0x11,0,
    0x0C,0x0C,0x12,0,0x0D,0x0D,0x13,0,0x0E,0x0E,0x14,0,0x0F,0x0F,0x15,0,
    0x10,0x10,0x16,0,0x11,0x11,0x17,0,0x12,0x12,0x18,0,0x13,0x13,0x19,0,
    0x14,0x14,0x1A,0,0x15,0x15,0x1B,0,0x16,0x16,0x1C,0,0x17,0x17,0x1D,0,
    0x18,0x18,0x1E,0,0x19,0x19,0x1F,0,0x1A,0x1A,0x20,0,0x1B,0x1B,0x21,0,
    0x1C,0x1C,0x22,0,0x1D,0x1D,0x23,0,0x1E,0x1E,0x24,0,0x1F,0x1F,0x25,0,
    0xE1,0xE0,0xE6,0,0xE2,0xE1,0xE7,0,0xE3,0xE2,0xE8,0,0xE4,0xE3,0xE9,0,
    0xE5,0xE4,0xEA,0,0xE6,0xE5,0xEB,0,0xE7,0xE6,0xEC,0,0xE8,0xE7,0xED,0,
    0xE9,0xE8,0xEE,0,0xEA,0xE9,0xEF,0,0xEB,0xEA,0xF0,0,0xEC,0xEB,0xF1,0,
    0xED,0xEC,0xF2,0,0xEE,0xED,0xF3,0,0xEF,0xEE,0xF4,0,0xF0,0xEF,0xF5,0,
    0xF1,0xF0,0xF6,0,0xF2,0xF1,0xF7,0,0xF3,0xF2,0xF8,0,0xF4,0xF3,0xF9,0,
    0xF5,0xF4,0xFA,0,0xF6,0xF5,0xFB,0,0xF7,0xF6,0xFC,0,0xF8,0xF7,0xFD,0,
    0xF9,0xF8,0xFE,0,0xFA,0xF9,0xFF,0,0xFB,0xFA,0x00,0,0xFC,0xFB,0x01,0,
    0xFD,0xFC,0x02,0,0xFE,0xFD,0x03,0,0xFF,0xFE,0x04,0,0x00,0xFF,0x05,0,
    0x01,0x00,0x06,0,0x02,0x01,0x07,0,0x03,0x02,0x08,0,0x04,0x03,0x09,0,
    0x05,0x04,0x0A,0,0x06,0x05,0x0B,0,0x07,0x06,0x0C,0,0x08,0x07,0x0D,0,
    0x09,0x08,0x0E,0,0x0A,0x09,0x0F,0,0x0B,0x0A,0x10,0,0x0C,0x0B,0x11,0,
    0x0D,0x0C,0x12,0,0x0E,0x0D,0x13,0,0x0F,0x0E,0x14,0,0x10,0x0F,0x15,0,
    0x11,0x10,0x16,0,0x12,0x11,0x17,0,0x13,0x12,0x18,0,0x14,0x13,0x19,0,
    0x15,0x14,0x1A,0,0x16,0x15,0x1B,0,0x17,0x16,0x1C,0,0x18,0x17,0x1D,0,
    0x19,0x18,0x1E,0,0x1A,0x19,0x1F,0,0x1B,0x1A,0x20,0,0x1C,0x1B,0x21,0,
    0x1D,0x1C,0x22,0,0x1E,0x1D,0x23,0,0x1F,0x1E,0x24,0,0x20,0x1F,0x25,0,
    0xE2,0xE0,0xE6,0,0xE3,0xE1,0xE7,0,0xE4,0xE2,0xE8,0,0xE5,0xE3,0xE9,0,
    0xE6,0xE4,0xEA,0,0xE7,0xE5,0xEB,0,0xE8,0xE6,0xEC,0,0xE9,0xE7,0xED,0,
    0xEA,0xE8,0xEE,0,0xEB,0xE9,0xEF,0,0xEC,0xEA,0xF0,0,0xED,0xEB,0xF1,0,
    0xEE,0xEC,0xF2,0,0xEF,0xED,0xF3,0,0xF0,0xEE,0xF4,0,0xF1,0xEF,0xF5,0,
    0xF2,0xF0,0xF6,0,0xF3,0xF1,0xF7,0,0xF4,0xF2,0xF8,0,0xF5,0xF3,0xF9,0,
    0xF6,0xF4,0xFA,0,0xF7,0xF5,0xFB,0,0xF8,0xF6,0xFC,0,0xF9,0xF7,0xFD,0,
    0xFA,0xF8,0xFE,0,0xFB,0xF9,0xFF,0,0xFC,0xFA,0x00,0,0xFD,0xFB,0x01,0,
    0xFE,0xFC,0x02,0,0xFF,0xFD,0x03,0,0x00,0xFE,0x04,0,0x01,0xFF,0x05,0,
    0x02,0x00,0x06,0,0x03,0x01,0x07,0,0x04,0x02,0x08,0,0x05,0x03,0x09,0,
    0x06,0x04,0x0A,0,0x07,0x05,0x0B,0,0x08,0x06,0x0C,0,0x09,0x07,0x0D,0,
    0x0A,0x08,0x0E,0,0x0B,0x09,0x0F,0,0x0C,0x0A,0x10,0,0x0D,0x0B,0x11,0,
    0x0E,0x0C,0x12,0,0x0F,0x0D,0x13,0,0x10,0x0E,0x14,0,0x11,0x0F,0x15,0,
    0x12,0x10,0x16,0,0x13,0x11,0x17,0,0x14,0x12,0x18,0,0x15,0x13,0x19,0,
    0x16,0x14,0x1A,0,0x17,0x15,0x1B,0,0x18,0x16,0x1C,0,0x19,0x17,0x1D,0,
    0x1A,0x18,0x1E,0,0x1B,0x19,0x1F,0,0x1C,0x1A,0x20,0,0x1D,0x1B,0x21,0,
    0x1E,0x1C,0x22,0,0x1F,0x1D,0x23,0,0x20,0x1E,0x24,0,0x21,0x1F,0x25,0,
    0xE3,0xE0,0xE6,0,0xE4,0xE1,0xE7,0,0xE5,0xE2,0xE8,0,0xE6,0xE3,0xE9,0,
    0xE7,0xE4,0xEA,0,0xE8,0xE5,0xEB,0,0xE9,0xE6,0xEC,0,0xEA,0xE7,0xED,0,
    0xEB,0xE8,0xEE,0,0xEC,0xE9,0xEF,0,0xED,0xEA,0xF0,0,0xEE,0xEB,0xF1,0,
    0xEF,0xEC,0xF2,0,0xF0,0xED,0xF3,0,0xF1,0xEE,0xF4,0,0xF2,0xEF,0xF5,0,
    0xF3,0xF0,0xF6,0,0xF4,0xF1,0xF7,0,0xF5,0xF2,0xF8,0,0xF6,0xF3,0xF9,0,
    0xF7,0xF4,0xFA,0,0xF8,0xF5,0xFB,0,0xF9,0xF6,0xFC,0,0xFA,0xF7,0xFD,0,
    0xFB,0xF8,0xFE,0,0xFC,0xF9,0xFF,0,0xFD,0xFA,0x00,0,0xFE,0xFB,0x01,0,
    0xFF,0xFC,0x02,0,0x00,0xFD,0x03,0,0x01,0xFE,0x04,0,0x02,0xFF,0x05,0,
    0x03,0x00,0x06,0,0x04,0x01,0x07,0,0x05,0x02,0x08,0,0x06,0x03,0x09,0,
    0x07,0x04,0x0A,0,0x08,0x05,0x0B,0,0x09,0x06,0x0C,0,0x0A,0x07,0x0D,0,
    0x0B,0x08,0x0E,0,0x0C,0x09,0x0F,0,0x0D,0x0A,0x10,0,0x0E,0x0B,0x11,0,
    0x0F,0x0C,0x12,0,0x10,0x0D,0x13,0,0x11,0x0E,0x14,0,0x12,0x0F,0x15,0,
    0x13,0x10,0x16,0,0x14,0x11,0x17,0,0x15,0x12,0x18,0,0x16,0x13,0x19,0,
    0x17,0x14,0x1A,0,0x18,0x15,0x1B,0,0x19,0x16,0x1C,0,0x1A,0x17,0x1D,0,
    0x1B,0x18,0x1E,0,0x1C,0x19,0x1F,0,0x1D,0x1A,0x20,0,0x1E,0x1B,0x21,0,
    0x1F,0x1C,0x22,0,0x20,0x1D,0x23,0,0x21,0x1E,0x24,0,0x22,0x1F,0x25,0,
    0xE4,0xE0,0xE6,0,0xE5,0xE1,0xE7,0,0xE6,0xE2,0xE8,0,0xE7,0xE3,0xE9,0,
    0xE8,0xE4,0xEA,0,0xE9,0xE5,0xEB,0,0xEA,0xE6,0xEC,0,0xEB,0xE7,0xED,0,
    0xEC,0xE8,0xEE,0,0xED,0xE9,0xEF,0,0xEE,0xEA,0xF0,0,0xEF,0xEB,0xF1,0,
    0xF0,0xEC,0xF2,0,0xF1,0xED,0xF3,0,0xF2,0xEE,0xF4,0,0xF3,0xEF,0xF5,0,
    0xF4,0xF0,0xF6,0,0xF5,0xF1,0xF7,0,0xF6,0xF2,0xF8,0,0xF7,0xF3,0xF9,0,
    0xF8,0xF4,0xFA,0,0xF9,0xF5,0xFB,0,0xFA,0xF6,0xFC,0,0xFB,0xF7,0xFD,0,
    0xFC,0xF8,0xFE,0,0xFD,0xF9,0xFF,0,0xFE,0xFA,0x00,0,0xFF,0xFB,0x01,0,
    0x00,0xFC,0x02,0,0x01,0xFD,0x03,0,0x02,0xFE,0x04,0,0x03,0xFF,0x05,0,
    0x04,0x00,0x06,0,0x05,0x01,0x07,0,0x06,0x02,0x08,0,0x07,0x03,0x09,0,
    0x08,0x04,0x0A,0,0x09,0x05,0x0B,0,0x0A,0x06,0x0C,0,0x0B,0x07,0x0D,0,
    0x0C,0x08,0x0E,0,0x0D,0x09,0x0F,0,0x0E,0x0A,0x10,0,0x0F,0x0B,0x11,0,
    0x10,0x0C,0x12,0,0x11,0x0D,0x13,0,0x12,0x0E,0x14,0,0x13,0x0F,0x15,0,
    0x14,0x10,0x16,0,0x15,0x11,0x17,0,0x16,0x12,0x18,0,0x17,0x13,0x19,0,
    0x18,0x14,0x1A,0,0x19,0x15,0x1B,0,0x1A,0x16,0x1C,0,0x1B,0x17,0x1D,0,
    0x1C,0x18,0x1E,0,0x1D,0x19,0x1F,0,0x1E,0x1A,0x20,0,0x1F,0x1B,0x21,0,
    0x20,0x1C,0x22,0,0x21,0x1D,0x23,0,0x22,0x1E,0x24,0,0x23,0x1F,0x25,0,
    0xE5,0xE0,0xE6,0,0xE6,0xE1,0xE7,0,0xE7,0xE2,0xE8,0,0xE8,0xE3,0xE9,0,
    0xE9,0xE4,0xEA,0,0xEA,0xE5,0xEB,0,0xEB,0xE6,0xEC,0,0xEC,0xE7,0xED,0,
    0xED,0xE8,0xEE,0,0xEE,0xE9,0xEF,0,0xEF,0xEA,0xF0,0,0xF0,0xEB,0xF1,0,
    0xF1,0xEC,0xF2,0,0xF2,0xED,0xF3,0,0xF3,0xEE,0xF4,0,0xF4,0xEF,0xF5,0,
    0xF5,0xF0,0xF6,0,0xF6,0xF1,0xF7,0,0xF7,0xF2,0xF8,0,0xF8,0xF3,0xF9,0,
    0xF9,0xF4,0xFA,0,0xFA,0xF5,0xFB,0,0xFB,0xF6,0xFC,0,0xFC,0xF7,0xFD,0,
    0xFD,0xF8,0xFE,0,0xFE,0xF9,0xFF,0,0xFF,0xFA,0x00,0,0x00,0xFB,0x01,0,
    0x01,0xFC,0x02,0,0x02,0xFD,0x03,0,0x03,0xFE,0x04,0,0x04,0xFF,0x05,0,
    0x05,0x00,0x06,0,0x06,0x01,0x07,0,0x07,0x02,0x08,0,0x08,0x03,0x09,0,
    0x09,0x04,0x0A,0,0x0A,0x05,0x0B,0,0x0B,0x06,0x0C,0,0x0C,0x07,0x0D,0,
    0x0D,0x08,0x0E,0,0x0E,0x09,0x0F,0,0x0F,0x0A,0x10,0,0x10,0x0B,0x11,0,
    0x11,0x0C,0x12,0,0x12,0x0D,0x13,0,0x13,0x0E,0x14,0,0x14,0x0F,0x15,0,
    0x15,0x10,0x16,0,0x16,0x11,0x17,0,0x17,0x12,0x18,0,0x18,0x13,0x19,0,
    0x19,0x14,0x1A,0,0x1A,0x15,0x1B,0,0x1B,0x16,0x1C,0,0x1C,0x17,0x1D,0,
    0x1D,0x18,0x1E,0,0x1E,0x19,0x1F,0,0x1F,0x1A,0x20,0,0x20,0x1B,0x21,0,
    0x21,0x1C,0x22,0,0x22,0x1D,0x23,0,0x23,0x1E,0x24,0,0x24,0x1F,0x25,0,
    0xE6,0xE0,0xE6,0,0xE7,0xE1,0xE7,0,0xE8,0xE2,0xE8,0,0xE9,0xE3,0xE9,0,
    0xEA,0xE4,0xEA,0,0xEB,0xE5,0xEB,0,0xEC,0xE6,0xEC,0,0xED,0xE7,0xED,0,
    0xEE,0xE8,0xEE,0,0xEF,0xE9,0xEF,0,0xF0,0xEA,0xF0,0,0xF1,0xEB,0xF1,0,
    0xF2,0xEC,0xF2,0,0xF3,0xED,0xF3,0,0xF4,0xEE,0xF4,0,0xF5,0xEF,0xF5,0,
    0xF6,0xF0,0xF6,0,0xF7,0xF1,0xF7,0,0xF8,0xF2,0xF8,0,0xF9,0xF3,0xF9,0,
    0xFA,0xF4,0xFA,0,0xFB,0xF5,0xFB,0,0xFC,0xF6,0xFC,0,0xFD,0xF7,0xFD,0,
    0xFE,0xF8,0xFE,0,0xFF,0xF9,0xFF,0,0x00,0xFA,0x00,0,0x01,0xFB,0x01,0,
    0x02,0xFC,0x02,0,0x03,0xFD,0x03,0,0x04,0xFE,0x04,0,0x05,0xFF,0x05,0,
    0x06,0x00,0x06,0,0x07,0x01,0x07,0,0x08,0x02,0x08,0,0x09,0x03,0x09,0,
    0x0A,0x04,0x0A,0,0x0B,0x05,0x0B,0,0x0C,0x06,0x0C,0,0x0D,0x07,0x0D,0,
    0x0E,0x08,0x0E,0,0x0F,0x09,0x0F,0,0x10,0x0A,0x10,0,0x11,0x0B,0x11,0,
    0x12,0x0C,0x12,0,0x13,0x0D,0x13,0,0x14,0x0E,0x14,0,0x15,0x0F,0x15,0,
    0x16,0x10,0x16,0,0x17,0x11,0x17,0,0x18,0x12,0x18,0,0x19,0x13,0x19,0,
    0x1A,0x14,0x1A,0,0x1B,0x15,0x1B,0,0x1C,0x16,0x1C,0,0x1D,0x17,0x1D,0,
    0x1E,0x18,0x1E,0,0x1F,0x19,0x1F,0,0x20,0x1A,0x20,0,0x21,0x1B,0x21,0,
    0x22,0x1C,0x22,0,0x23,0x1D,0x23,0,0x24,0x1E,0x24,0,0x25,0x1F,0x25,0,
    0xE7,0xE0,0xE6,0,0xE8,0xE1,0xE7,0,0xE9,0xE2,0xE8,0,0xEA,0xE3,0xE9,0,
    0xEB,0xE4,0xEA,0,0xEC,0xE5,0xEB,0,0xED,0xE6,0xEC,0,0xEE,0xE7,0xED,0,
    0xEF,0xE8,0xEE,0,0xF0,0xE9,0xEF,0,0xF1,0xEA,0xF0,0,0xF2,0xEB,0xF1,0,
    0xF3,0xEC,0xF2,0,0xF4,0xED,0xF3,0,0xF5,0xEE,0xF4,0,0xF6,0xEF,0xF5,0,
    0xF7,0xF0,0xF6,0,0xF8,0xF1,0xF7,0,0xF9,0xF2,0xF8,0,0xFA,0xF3,0xF9,0,
    0xFB,0xF4,0xFA,0,0xFC,0xF5,0xFB,0,0xFD,0xF6,0xFC,0,0xFE,0xF7,0xFD,0,
    0xFF,0xF8,0xFE,0,0x00,0xF9,0xFF,0,0x01,0xFA,0x00,0,0x02,0xFB,0x01,0,
    0x03,0xFC,0x02,0,0x04,0xFD,0x03,0,0x05,0xFE,0x04,0,0x06,0xFF,0x05,0,
    0x07,0x00,0x06,0,0x08,0x01,0x07,0,0x09,0x02,0x08,0,0x0A,0x03,0x09,0,
    0x0B,0x04,0x0A,0,0x0C,0x05,0x0B,0,0x0D,0x06,0x0C,0,0x0E,0x07,0x0D,0,
    0x0F,0x08,0x0E,0,0x10,0x09,0x0F,0,0x11,0x0A,0x10,0,0x12,0x0B,0x11,0,
    0x13,0x0C,0x12,0,0x14,0x0D,0x13,0,0x15,0x0E,0x14,0,0x16,0x0F,0x15,0,
    0x17,0x10,0x16,0,0x18,0x11,0x17,0,0x19,0x12,0x18,0,0x1A,0x13,0x19,0,
    0x1B,0x14,0x1A,0,0x1C,0x15,0x1B,0,0x1D,0x16,0x1C,0,0x1E,0x17,0x1D,0,
    0x1F,0x18,0x1E,0,0x20,0x19,0x1F,0,0x21,0x1A,0x20,0,0x22,0x1B,0x21,0,
    0x23,0x1C,0x22,0,0x24,0x1D,0x23,0,0x25,0x1E,0x24,0,0x26,0x1F,0x25,0,
    0xD8,0xE0,0xE7,0,0xD9,0xE1,0xE8,0,0xDA,0xE2,0xE9,0,0xDB,0xE3,0xEA,0,
    0xDC,0xE4,0xEB,0,0xDD,0xE5,0xEC,0,0xDE,0xE6,0xED,0,0xDF,0xE7,0xEE,0,
    0xE0,0xE8,0xEF,0,0xE1,0xE9,0xF0,0,0xE2,0xEA,0xF1,0,0xE3,0xEB,0xF2,0,
    0xE4,0xEC,0xF3,0,0xE5,0xED,0xF4,0,0xE6,0xEE,0xF5,0,0xE7,0xEF,0xF6,0,
    0xE8,0xF0,0xF7,0,0xE9,0xF1,0xF8,0,0xEA,0xF2,0xF9,0,0xEB,0xF3,0xFA,0,
    0xEC,0xF4,0xFB,0,0xED,0xF5,0xFC,0,0xEE,0xF6,0xFD,0,0xEF,0xF7,0xFE,0,
    0xF0,0xF8,0xFF,0,0xF1,0xF9,0x00,0,0xF2,0xFA,0x01,0,0xF3,0xFB,0x02,0,
    0xF4,0xFC,0x03,0,0xF5,0xFD,0x04,0,0xF6,0xFE,0x05,0,0xF7,0xFF,0x06,0,
    0xF8,0x00,0x07,0,0xF9,0x01,0x08,0,0xFA,0x02,0x09,0,0xFB,0x03,0x0A,0,
    0xFC,0x04,0x0B,0,0xFD,0x05,0x0C,0,0xFE,0x06,0x0D,0,0xFF,0x07,0x0E,0,
    0x00,0x08,0x0F,0,0x01,0x09,0x10,0,0x02,0x0A,0x11,0,0x03,0x0B,0x12,0,
    0x04,0x0C,0x13,0,0x05,0x0D,0x14,0,0x06,0x0E,0x15,0,0x07,0x0F,0x16,0,
    0x08,0x10,0x17,0,0x09,0x11,0x18,0,0x0A,0x12,0x19,0,0x0B,0x13,0x1A,0,
    0x0C,0x14,0x1B,0,0x0D,0x15,0x1C,0,0x0E,0x16,0x1D,0,0x0F,0x17,0x1E,0,
    0x10,0x18,0x1F,0,0x11,0x19,0x20,0,0x12,0x1A,0x21,0,0x13,0x1B,0x22,0,
    0x14,0x1C,0x23,0,0x15,0x1D,0x24,0,0x16,0x1E,0x25,0,0x17,0x1F,0x26,0,
    0xD9,0xE0,0xE7,0,0xDA,0xE1,0xE8,0,0xDB,0xE2,0xE9,0,0xDC,0xE3,0xEA,0,
    0xDD,0xE4,0xEB,0,0xDE,0xE5,0xEC,0,0xDF,0xE6,0xED,0,0xE0,0xE7,0xEE,0,
    0xE1,0xE8,0xEF,0,0xE2,0xE9,0xF0,0,0xE3,0xEA,0xF1,0,0xE4,0xEB,0xF2,0,
    0xE5,0xEC,0xF3,0,0xE6,0xED,0xF4,0,0xE7,0xEE,0xF5,0,0xE8,0xEF,0xF6,0,
    0xE9,0xF0,0xF7,0,0xEA,0xF1,0xF8,0,0xEB,0xF2,0xF9,0,0xEC,0xF3,0xFA,0,
    0xED,0xF4,0xFB,0,0xEE,0xF5,0xFC,0,0xEF,0xF6,0xFD,0,0xF0,0xF7,0xFE,0,
    0xF1,0xF8,0xFF,0,0xF2,0xF9,0x00,0,0xF3,0xFA,0x01,0,0xF4,0xFB,0x02,0,
    0xF5,0xFC,0x03,0,0xF6,0xFD,0x04,0,0xF7,0xFE,0x05,0,0xF8,0xFF,0x06,0,
    0xF9,0x00,0x07,0,0xFA,0x01,0x08,0,0xFB,0x02,0x09,0,0xFC,0x03,0x0A,0,
    0xFD,0x04,0x0B,0,0xFE,0x05,0x0C,0,0xFF,0x06,0x0D,0,0x00,0x07,0x0E,0,
    0x01,0x08,0x0F,0,0x02,0x09,0x10,0,0x03,0x0A,0x11,0,0x04,0x0B,0x12,0,
    0x05,0x0C,0x13,0,0x06,0x0D,0x14,0,0x07,0x0E,0x15,0,0x08,0x0F,0x16,0,
    0x09,0x10,0x17,0,0x0A,0x11,0x18,0,0x0B,0x12,0x19,0,0x0C,0x13,0x1A,0,
    0x0D,0x14,0x1B,0,0x0E,0x15,0x1C,0,0x0F,0x16,0x1D,0,0x10,0x17,0x1E,0,
    0x11,0x18,0x1F,0,0x12,0x19,0x20,0,0x13,0x1A,0x21,0,0x14,0x1B,0x22,0,
    0x15,0x1C,0x23,0,0x16,0x1D,0x24,0,0x17,0x1E,0x25,0,0x18,0x1F,0x26,0,
    0xDA,0xE0,0xE7,0,0xDB,0xE1,0xE8,0,0xDC,0xE2,0xE9,0,0xDD,0xE3,0xEA,0,
    0xDE,0xE4,0xEB,0,0xDF,0xE5,0xEC,0,0xE0,0xE6,0xED,0,0xE1,0xE7,0xEE,0,
    0xE2,0xE8,0xEF,0,0xE3,0xE9,0xF0,0,0xE4,0xEA,0xF1,0,0xE5,0xEB,0xF2,0,
    0xE6,0xEC,0xF3,0,0xE7,0xED,0xF4,0,0xE8,0xEE,0xF5,0,0xE9,0xEF,0xF6,0,
    0xEA,0xF0,0xF7,0,0xEB,0xF1,0xF8,0,0xEC,0xF2,0xF9,0,0xED,0xF3,0xFA,0,
    0xEE,0xF4,0xFB,0,0xEF,0xF5,0xFC,0,0xF0,0xF6,0xFD,0,0xF1,0xF7,0xFE,0,
    0xF2,0xF8,0xFF,0,0xF3,0xF9,0x00,0,0xF4,0xFA,0x01,0,0xF5,0xFB,0x02,0,
    0xF6,0xFC,0x03,0,0xF7,0xFD,0x04,0,0xF8,0xFE,0x05,0,0xF9,0xFF,0x06,0,
    0xFA,0x00,0x07,0,0xFB,0x01,0x08,0,0xFC,0x02,0x09,0,0xFD,0x03,0x0A,0,
    0xFE,0x04,0x0B,0,0xFF,0x05,0x0C,0,0x00,0x06,0x0D,0,0x01,0x07,0x0E,0,
    0x02,0x08,0x0F,0,0x03,0x09,0x10,0,0x04,0x0A,0x11,0,0x05,0x0B,0x12,0,
    0x06,0x0C,0x13,0,0x07,0x0D,0x14,0,0x08,0x0E,0x15,0,0x09,0x0F,0x16,0,
    0x0A,0x10,0x17,0,0x0B,0x11,0x18,0,0x0C,0x12,0x19,0,0x0D,0x13,0x1A,0,
    0x0E,0x14,0x1B,0,0x0F,0x15,0x1C,0,0x10,0x16,0x1D,0,0x11,0x17,0x1E,0,
    0x12,0x18,0x1F,0,0x13,0x19,0x20,0,0x14,0x1A,0x21,0,0x15,0x1B,0x22,0,
    0x16,0x1C,0x23,0,0x17,0x1D,0x24,0,0x18,0x1E,0x25,0,0x19,0x1F,0x26,0,
    0xDB,0xE0,0xE7,0,0xDC,0xE1,0xE8,0,0xDD,0xE2,0xE9,0,0xDE,0xE3,0xEA,0,
    0xDF,0xE4,0xEB,0,0xE0,0xE5,0xEC,0,0xE1,0xE6,0xED,0,0xE2,0xE7,0xEE,0,
    0xE3,0xE8,0xEF,0,0xE4,0xE9,0xF0,0,0xE5,0xEA,0xF1,0,0xE6,0xEB,0xF2,0,
    0xE7,0xEC,0xF3,0,0xE8,0xED,0xF4,0,0xE9,0xEE,0xF5,0,0xEA,0xEF,0xF6,0,
    0xEB,0xF0,0xF7,0,0xEC,0xF1,0xF8,0,0xED,0xF2,0xF9,0,0xEE,0xF3,0xFA,0,
    0xEF,0xF4,0xFB,0,0xF0,0xF5,0xFC,0,0xF1,0xF6,0xFD,0,0xF2,0xF7,0xFE,0,
    0xF3,0xF8,0xFF,0,0xF4,0xF9,0x00,0,0xF5,0xFA,0x01,0,0xF6,0xFB,0x02,0,
    0xF7,0xFC,0x03,0,0xF8,0xFD,0x04,0,0xF9,0xFE,0x05,0,0xFA,0xFF,0x06,0,
    0xFB,0x00,0x07,0,0xFC,0x01,0x08,0,0xFD,0x02,0x09,0,0xFE,0x03,0x0A,0,
    0xFF,0x04,0x0B,0,0x00,0x05,0x0C,0,0x01,0x06,0x0D,0,0x02,0x07,0x0E,0,
    0x03,0x08,0x0F,0,0x04,0x09,0x10,0,0x05,0x0A,0x11,0,0x06,0x0B,0x12,0,
    0x07,0x0C,0x13,0,0x08,0x0D,0x14,0,0x09,0x0E,0x15,0,0x0A,0x0F,0x16,0,
    0x0B,0x10,0x17,0,0x0C,0x11,0x18,0,0x0D,0x12,0x19,0,0x0E,0x13,0x1A,0,
    0x0F,0x14,0x1B,0,0x10,0x15,0x1C,0,0x11,0x16,0x1D,0,0x12,0x17,0x1E,0,
    0x13,0x18,0x1F,0,0x14,0x19,0x20,0,0x15,0x1A,0x21,0,0x16,0x1B,0x22,0,
    0x17,0x1C,0x23,0,0x18,0x1D,0x24,0,0x19,0x1E,0x25,0,0x1A,0x1F,0x26,0,
    0xDC,0xE0,0xE7,0,0xDD,0xE1,0xE8,0,0xDE,0xE2,0xE9,0,0xDF,0xE3,0xEA,0,
    0xE0,0xE4,0xEB,0,0xE1,0xE5,0xEC,0,0xE2,0xE6,0xED,0,0xE3,0xE7,0xEE,0,
    0xE4,0xE8,0xEF,0,0xE5,0xE9,0xF0,0,0xE6,0xEA,0xF1,0,0xE7,0xEB,0xF2,0,
    0xE8,0xEC,0xF3,0,0xE9,0xED,0xF4,0,0xEA,0xEE,0xF5,0,0xEB,0xEF,0xF6,0,
    0xEC,0xF0,0xF7,0,0xED,0xF1,0xF8,0,0xEE,0xF2,0xF9,0,0xEF,0xF3,0xFA,0,
    0xF0,0xF4,0xFB,0,0xF1,0xF5,0xFC,0,0xF2,0xF6,0xFD,0,0xF3,0xF7,0xFE,0,
    0xF4,0xF8,0xFF,0,0xF5,0xF9,0x00,0,0xF6,0xFA,0x01,0,0xF7,0xFB,0x02,0,
    0xF8,0xFC,0x03,0,0xF9,0xFD,0x04,0,0xFA,0xFE,0x05,0,0xFB,0xFF,0x06,0,
    0xFC,0x00,0x07,0,0xFD,0x01,0x08,0,0xFE,0x02,0x09,0,0xFF,0x03,0x0A,0,
    0x00,0x04,0x0B,0,0x01,0x05,0x0C,0,0x02,0x06,0x0D,0,0x03,0x07,0x0E,0,
    0x04,0x08,0x0F,0,0x05,0x09,0x10,0,0x06,0x0A,0x11,0,0x07,0x0B,0x12,0,
    0x08,0x0C,0x13,0,0x09,0x0D,0x14,0,0x0A,0x0E,0x15,0,0x0B,0x0F,0x16,0,
    0x0C,0x10,0x17,0,0x0D,0x11,0x18,0,0x0E,0x12,0x19,0,0x0F,0x13,0x1A,0,
    0x10,0x14,0x1B,0,0x11,0x15,0x1C,0,0x12,0x16,0x1D,0,0x13,0x17,0x1E,0,
    0x14,0x18,0x1F,0,0x15,0x19,0x20,0,0x16,0x1A,0x21,0,0x17,0x1B,0x22,0,
    0x18,0x1C,0x23,0,0x19,0x1D,0x24,0,0x1A,0x1E,0x25,0,0x1B,0x1F,0x26,0,
    0xDD,0xE0,0xE7,0,0xDE,0xE1,0xE8,0,0xDF,0xE2,0xE9,0,0xE0,0xE3,0xEA,0,
    0xE1,0xE4,0xEB,0,0xE2,0xE5,0xEC,0,0xE3,0xE6,0xED,0,0xE4,0xE7,0xEE,0,
    0xE5,0xE8,0xEF,0,0xE6,0xE9,0xF0,0,0xE7,0xEA,0xF1,0,0xE8,0xEB,0xF2,0,
    0xE9,0xEC,0xF3,0,0xEA,0xED,0xF4,0,0xEB,0xEE,0xF5,0,0xEC,0xEF,0xF6,0,
    0xED,0xF0,0xF7,0,0xEE,0xF1,0xF8,0,0xEF,0xF2,0xF9,0,0xF0,0xF3,0xFA,0,
    0xF1,0xF4,0xFB,0,0xF2,0xF5,0xFC,0,0xF3,0xF6,0xFD,0,0xF4,0xF7,0xFE,0,
    0xF5,0xF8,0xFF,0,0xF6,0xF9,0x00,0,0xF7,0xFA,0x01,0,0xF8,0xFB,0x02,0,
    0xF9,0xFC,0x03,0,0xFA,0xFD,0x04,0,0xFB,0xFE,0x05,0,0xFC,0xFF,0x06,0,
    0xFD,0x00,0x07,0,0xFE,0x01,0x08,0,0xFF,0x02,0x09,0,0x00,0x03,0x0A,0,
    0x01,0x04,0x0B,0,0x02,0x05,0x0C,0,0x03,0x06,0x0D,0,0x04,0x07,0x0E,0,
    0x05,0x08,0x0F,0,0x06,0x09,0x10,0,0x07,0x0A,0x11,0,0x08,0x0B,0x12,0,
    0x09,0x0C,0x13,0,0x0A,0x0D,0x14,0,0x0B,0x0E,0x15,0,0x0C,0x0F,0x16,0,
    0x0D,0x10,0x17,0,0x0E,0x11,0x18,0,0x0F,0x12,0x19,0,0x10,0x13,0x1A,0,
    0x11,0x14,0x1B,0,0x12,0x15,0x1C,0,0x13,0x16,0x1D,0,0x14,0x17,0x1E,0,
    0x15,0x18,0x1F,0,0x16,0x19,0x20,0,0x17,0x1A,0x21,0,0x18,0x1B,0x22,0,
    0x19,0x1C,0x23,0,0x1A,0x1D,0x24,0,0x1B,0x1E,0x25,0,0x1C,0x1F,0x26,0,
    0xDE,0xE0,0xE7,0,0xDF,0xE1,0xE8,0,0xE0,0xE2,0xE9,0,0xE1,0xE3,0xEA,0,
    0xE2,0xE4,0xEB,0,0xE3,0xE5,0xEC,0,0xE4,0xE6,0xED,0,0xE5,0xE7,0xEE,0,
    0xE6,0xE8,0xEF,0,0xE7,0xE9,0xF0,0,0xE8,0xEA,0xF1,0,0xE9,0xEB,0xF2,0,
    0xEA,0xEC,0xF3,0,0xEB,0xED,0xF4,0,0xEC,0xEE,0xF5,0,0xED,0xEF,0xF6,0,
    0xEE,0xF0,0xF7,0,0xEF,0xF1,0xF8,0,0xF0,0xF2,0xF9,0,0xF1,0xF3,0xFA,0,
    0xF2,0xF4,0xFB,0,0xF3,0xF5,0xFC,0,0xF4,0xF6,0xFD,0,0xF5,0xF7,0xFE,0,
    0xF6,0xF8,0xFF,0,0xF7,0xF9,0x00,0,0xF8,0xFA,0x01,0,0xF9,0xFB,0x02,0,
    0xFA,0xFC,0x03,0,0xFB,0xFD,0x04,0,0xFC,0xFE,0x05,0,0xFD,0xFF,0x06,0,
    0xFE,0x00,0x07,0,0xFF,0x01,0x08,0,0x00,0x02,0x09,0,0x01,0x03,0x0A,0,
    0x02,0x04,0x0B,0,0x03,0x05,0x0C,0,0x04,0x06,0x0D,0,0x05,0x07,0x0E,0,
    0x06,0x08,0x0F,0,0x07,0x09,0x10,0,0x08,0x0A,0x11,0,0x09,0x0B,0x12,0,
    0x0A,0x0C,0x13,0,0x0B,0x0D,0x14,0,0x0C,0x0E,0x15,0,0x0D,0x0F,0x16,0,
    0x0E,0x10,0x17,0,0x0F,0x11,0x18,0,0x10,0x12,0x19,0,0x11,0x13,0x1A,0,
    0x12,0x14,0x1B,0,0x13,0x15,0x1C,0,0x14,0x16,0x1D,0,0x15,0x17,0x1E,0,
    0x16,0x18,0x1F,0,0x17,0x19,0x20,0,0x18,0x1A,0x21,0,0x19,0x1B,0x22,0,
    0x1A,0x1C,0x23,0,0x1B,0x1D,0x24,0,0x1C,0x1E,0x25,0,0x1D,0x1F,0x26,0,
    0xDF,0xE0,0xE7,0,0xE0,0xE1,0xE8,0,0xE1,0xE2,0xE9,0,0xE2,0xE3,0xEA,0,
    0xE3,0xE4,0xEB,0,0xE4,0xE5,0xEC,0,0xE5,0xE6,0xED,0,0xE6,0xE7,0xEE,0,
    0xE7,0xE8,0xEF,0,0xE8,0xE9,0xF0,0,0xE9,0xEA,0xF1,0,0xEA,0xEB,0xF2,0,
    0xEB,0xEC,0xF3,0,0xEC,0xED,0xF4,0,0xED,0xEE,0xF5,0,0xEE,0xEF,0xF6,0,
    0xEF,0xF0,0xF7,0,0xF0,0xF1,0xF8,0,0xF1,0xF2,0xF9,0,0xF2,0xF3,0xFA,0,
    0xF3,0xF4,0xFB,0,0xF4,0xF5,0xFC,0,0xF5,0xF6,0xFD,0,0xF6,0xF7,0xFE,0,
    0xF7,0xF8,0xFF,0,0xF8,0xF9,0x00,0,0xF9,0xFA,0x01,0,0xFA,0xFB,0x02,0,
    0xFB,0xFC,0x03,0,0xFC,0xFD,0x04,0,0xFD,0xFE,0x05,0,0xFE,0xFF,0x06,0,
    0xFF,0x00,0x07,0,0x00,0x01,0x08,0,0x01,0x02,0x09,0,0x02,0x03,0x0A,0,
    0x03,0x04,0x0B,0,0x04,0x05,0x0C,0,0x05,0x06,0x0D,0,0x06,0x07,0x0E,0,
    0x07,0x08,0x0F,0,0x08,0x09,0x10,0,0x09,0x0A,0x11,0,0x0A,0x0B,0x12,0,
    0x0B,0x0C,0x13,0,0x0C,0x0D,0x14,0,0x0D,0x0E,0x15,0,0x0E,0x0F,0x16,0,
    0x0F,0x10,0x17,0,0x10,0x11,0x18,0,0x11,0x12,0x19,0,0x12,0x13,0x1A,0,
    0x13,0x14,0x1B,0,0x14,0x15,0x1C,0,0x15,0x16,0x1D,0,0x16,0x17,0x1E,0,
    0x17,0x18,0x1F,0,0x18,0x19,0x20,0,0x19,0x1A,0x21,0,0x1A,0x1B,0x22,0,
    0x1B,0x1C,0x23,0,0x1C,0x1D,0x24,0,0x1D,0x1E,0x25,0,0x1E,0x1F,0x26,0,
    0xE0,0xE0,0xE7,0,0xE1,0xE1,0xE8,0,0xE2,0xE2,0xE9,0,0xE3,0xE3,0xEA,0,
    0xE4,0xE4,0xEB,0,0xE5,0xE5,0xEC,0,0xE6,0xE6,0xED,0,0xE7,0xE7,0xEE,0,
    0xE8,0xE8,0xEF,0,0xE9,0xE9,0xF0,0,0xEA,0xEA,0xF1,0,0xEB,0xEB,0xF2,0,
    0xEC,0xEC,0xF3,0,0xED,0xED,0xF4,0,0xEE,0xEE,0xF5,0,0xEF,0xEF,0xF6,0,
    0xF0,0xF0,0xF7,0,0xF1,0xF1,0xF8,0,0xF2,0xF2,0xF9,0,0xF3,0xF3,0xFA,0,
    0xF4,0xF4,0xFB,0,0xF5,0xF5,0xFC,0,0xF6,0xF6,0xFD,0,0xF7,0xF7,0xFE,0,
    0xF8,0xF8,0xFF,0,0xF9,0xF9,0x00,0,0xFA,0xFA,0x01,0,0xFB,0xFB,0x02,0,
    0xFC,0xFC,0x03,0,0xFD,0xFD,0x04,0,0xFE,0xFE,0x05,0,0xFF,0xFF,0x06,0,
    0x00,0x00,0x07,0,0x01,0x01,0x08,0,0x02,0x02,0x09,0,0x03,0x03,0x0A,0,
    0x04,0x04,0x0B,0,0x05,0x05,0x0C,0,0x06,0x06,0x0D,0,0x07,0x07,0x0E,0,
    0x08,0x08,0x0F,0,0x09,0x09,0x10,0,0x0A,0x0A,0x11,0,0x0B,0x0B,0x12,0,
    0x0C,0x0C,0x13,0,0x0D,0x0D,0x14,0,0x0E,0x0E,0x15,0,0x0F,0x0F,0x16,0,
    0x10,0x10,0x17,0,0x11,0x11,0x18,0,0x12,0x12,0x19,0,0x13,0x13,0x1A,0,
    0x14,0x14,0x1B,0,0x15,0x15,0x1C,0,0x16,0x16,0x1D,0,0x17,0x17,0x1E,0,
    0x18,0x18,0x1F,0,0x19,0x19,0x20,0,0x1A,0x1A,0x21,0,0x1B,0x1B,0x22,0,
    0x1C,0x1C,0x23,0,0x1D,0x1D,0x24,0,0x1E,0x1E,0x25,0,0x1F,0x1F,0x26,0,
    0xE1,0xE0,0xE7,0,0xE2,0xE1,0xE8,0,0xE3,0xE2,0xE9,0,0xE4,0xE3,0xEA,0,
    0xE5,0xE4,0xEB,0,0xE6,0xE5,0xEC,0,0xE7,0xE6,0xED,0,0xE8,0xE7,0xEE,0,
    0xE9,0xE8,0xEF,0,0xEA,0xE9,0xF0,0,0xEB,0xEA,0xF1,0,0xEC,0xEB,0xF2,0,
    0xED,0xEC,0xF3,0,0xEE,0xED,0xF4,0,0xEF,0xEE,0xF5,0,0xF0,0xEF,0xF6,0,
    0xF1,0xF0,0xF7,0,0xF2,0xF1,0xF8,0,0xF3,0xF2,0xF9,0,0xF4,0xF3,0xFA,0,
    0xF5,0xF4,0xFB,0,0xF6,0xF5,0xFC,0,0xF7,0xF6,0xFD,0,0xF8,0xF7,0xFE,0,
    0xF9,0xF8,0xFF,0,0xFA,0xF9,0x00,0,0xFB,0xFA,0x01,0,0xFC,0xFB,0x02,0,
    0xFD,0xFC,0x03,0,0xFE,0xFD,0x04,0,0xFF,0xFE,0x05,0,0x00,0xFF,0x06,0,
    0x01,0x00,0x07,0,0x02,0x01,0x08,0,0x03,0x02,0x09,0,0x04,0x03,0x0A,0,
    0x05,0x04,0x0B,0,0x06,0x05,0x0C,0,0x07,0x06,0x0D,0,0x08,0x07,0x0E,0,
    0x09,0x08,0x0F,0,0x0A,0x09,0x10,0,0x0B,0x0A,0x11,0,0x0C,0x0B,0x12,0,
    0x0D,0x0C,0x13,0,0x0E,0x0D,0x14,0,0x0F,0x0E,0x15,0,0x10,0x0F,0x16,0,
    0x11,0x10,0x17,0,0x12,0x11,0x18,0,0x13,0x12,0x19,0,0x14,0x13,0x1A,0,
    0x15,0x14,0x1B,0,0x16,0x15,0x1C,0,0x17,0x16,0x1D,0,0x18,0x17,0x1E,0,
    0x19,0x18,0x1F,0,0x1A,0x19,0x20,0,0x1B,0x1A,0x21,0,0x1C,0x1B,0x22,0,
    0x1D,0x1C,0x23,0,0x1E,0x1D,0x24,0,0x1F,0x1E,0x25,0,0x20,0x1F,0x26,0,
    0xE2,0xE0,0xE7,0,0xE3,0xE1,0xE8,0,0xE4,0xE2,0xE9,0,0xE5,0xE3,0xEA,0,
    0xE6,0xE4,0xEB,0,0xE7,0xE5,0xEC,0,0xE8,0xE6,0xED,0,0xE9,0xE7,0xEE,0,
    0xEA,0xE8,0xEF,0,0xEB,0xE9,0xF0,0,0xEC,0xEA,0xF1,0,0xED,0xEB,0xF2,0,
    0xEE,0xEC,0xF3,0,0xEF,0xED,0xF4,0,0xF0,0xEE,0xF5,0,0xF1,0xEF,0xF6,0,
    0xF2,0xF0,0xF7,0,0xF3,0xF1,0xF8,0,0xF4,0xF2,0xF9,0,0xF5,0xF3,0xFA,0,
    0xF6,0xF4,0xFB,0,0xF7,0xF5,0xFC,0,0xF8,0xF6,0xFD,0,0xF9,0xF7,0xFE,0,
    0xFA,0xF8,0xFF,0,0xFB,0xF9,0x00,0,0xFC,0xFA,0x01,0,0xFD,0xFB,0x02,0,
    0xFE,0xFC,0x03,0,0xFF,0xFD,0x04,0,0x00,0xFE,0x05,0,0x01,0xFF,0x06,0,
    0x02,0x00,0x07,0,0x03,0x01,0x08,0,0x04,0x02,0x09,0,0x05,0x03,0x0A,0,
    0x06,0x04,0x0B,0,0x07,0x05,0x0C,0,0x08,0x06,0x0D,0,0x09,0x07,0x0E,0,
    0x0A,0x08,0x0F,0,0x0B,0x09,0x10,0,0x0C,0x0A,0x11,0,0x0D,0x0B,0x12,0,
    0x0E,0x0C,0x13,0,0x0F,0x0D,0x14,0,0x10,0x0E,0x15,0,0x11,0x0F,0x16,0,
    0x12,0x10,0x17,0,0x13,0x11,0x18,0,0x14,0x12,0x19,0,0x15,0x13,0x1A,0,
    0x16,0x14,0x1B,0,0x17,0x15,0x1C,0,0x18,0x16,0x1D,0,0x19,0x17,0x1E,0,
    0x1A,0x18,0x1F,0,0x1B,0x19,0x20,0,0x1C,0x1A,0x21,0,0x1D,0x1B,0x22,0,
    0x1E,0x1C,0x23,0,0x1F,0x1D,0x24,0,0x20,0x1E,0x25,0,0x21,0x1F,0x26,0,
    0xE3,0xE0,0xE7,0,0xE4,0xE1,0xE8,0,0xE5,0xE2,0xE9,0,0xE6,0xE3,0xEA,0,
    0xE7,0xE4,0xEB,0,0xE8,0xE5,0xEC,0,0xE9,0xE6,0xED,0,0xEA,0xE7,0xEE,0,
    0xEB,0xE8,0xEF,0,0xEC,0xE9,0xF0,0,0xED,0xEA,0xF1,0,0xEE,0xEB,0xF2,0,
    0xEF,0xEC,0xF3,0,0xF0,0xED,0xF4,0,0xF1,0xEE,0xF5,0,0xF2,0xEF,0xF6,0,
    0xF3,0xF0,0xF7,0,0xF4,0xF1,0xF8,0,0xF5,0xF2,0xF9,0,0xF6,0xF3,0xFA,0,
    0xF7,0xF4,0xFB,0,0xF8,0xF5,0xFC,0,0xF9,0xF6,0xFD,0,0xFA,0xF7,0xFE,0,
    0xFB,0xF8,0xFF,0,0xFC,0xF9,0x00,0,0xFD,0xFA,0x01,0,0xFE,0xFB,0x02,0,
    0xFF,0xFC,0x03,0,0x00,0xFD,0x04,0,0x01,0xFE,0x05,0,0x02,0xFF,0x06,0,
    0x03,0x00,0x07,0,0x04,0x01,0x08,0,0x05,0x02,0x09,0,0x06,0x03,0x0A,0,
    0x07,0x04,0x0B,0,0x08,0x05,0x0C,0,0x09,0x06,0x0D,0,0x0A,0x07,0x0E,0,
    0x0B,0x08,0x0F,0,0x0C,0x09,0x10,0,0x0D,0x0A,0x11,0,0x0E,0x0B,0x12,0,
    0x0F,0x0C,0x13,0,0x10,0x0D,0x14,0,0x11,0x0E,0x15,0,0x12,0x0F,0x16,0,
    0x13,0x10,0x17,0,0x14,0x11,0x18,0,0x15,0x12,0x19,0,0x16,0x13,0x1A,0,
    0x17,0x14,0x1B,0,0x18,0x15,0x1C,0,0x19,0x16,0x1D,0,0x1A,0x17,0x1E,0,
    0x1B,0x18,0x1F,0,0x1C,0x19,0x20,0,0x1D,0x1A,0x21,0,0x1E,0x1B,0x22,0,
    0x1F,0x1C,0x23,0,0x20,0x1D,0x24,0,0x21,0x1E,0x25,0,0x22,0x1F,0x26,0,
    0xE4,0xE0,0xE7,0,0xE5,0xE1,0xE8,0,0xE6,0xE2,0xE9,0,0xE7,0xE3,0xEA,0,
    0xE8,0xE4,0xEB,0,0xE9,0xE5,0xEC,0,0xEA,0xE6,0xED,0,0xEB,0xE7,0xEE,0,
    0xEC,0xE8,0xEF,0,0xED,0xE9,0xF0,0,0xEE,0xEA,0xF1,0,0xEF,0xEB,0xF2,0,
    0xF0,0xEC,0xF3,0,0xF1,0xED,0xF4,0,0xF2,0xEE,0xF5,0,0xF3,0xEF,0xF6,0,
    0xF4,0xF0,0xF7,0,0xF5,0xF1,0xF8,0,0xF6,0xF2,0xF9,0,0xF7,0xF3,0xFA,0,
    0xF8,0xF4,0xFB,0,0xF9,0xF5,0xFC,0,0xFA,0xF6,0xFD,0,0xFB,0xF7,0xFE,0,
    0xFC,0xF8,0xFF,0,0xFD,0xF9,0x00,0,0xFE,0xFA,0x01,0,0xFF,0xFB,0x02,0,
    0x00,0xFC,0x03,0,0x01,0xFD,0x04,0,0x02,0xFE,0x05,0,0x03,0xFF,0x06,0,
    0x04,0x00,0x07,0,0x05,0x01,0x08,0,0x06,0x02,0x09,0,0x07,0x03,0x0A,0,
    0x08,0x04,0x0B,0,0x09,0x05,0x0C,0,0x0A,0x06,0x0D,0,0x0B,0x07,0x0E,0,
    0x0C,0x08,0x0F,0,0x0D,0x09,0x10,0,0x0E,0x0A,0x11,0,0x0F,0x0B,0x12,0,
    0x10,0x0C,0x13,0,0x11,0x0D,0x14,0,0x12,0x0E,0x15,0,0x13,0x0F,0x16,0,
    0x14,0x10,0x17,0,0x15,0x11,0x18,0,0x16,0x12,0x19,0,0x17,0x13,0x1A,0,
    0x18,0x14,0x1B,0,0x19,0x15,0x1C,0,0x1A,0x16,0x1D,0,0x1B,0x17,0x1E,0,
    0x1C,0x18,0x1F,0,0x1D,0x19,0x20,0,0x1E,0x1A,0x21,0,0x1F,0x1B,0x22,0,
    0x20,0x1C,0x23,0,0x21,0x1D,0x24,0,0x22,0x1E,0x25,0,0x23,0x1F,0x26,0,
    0xE5,0xE0,0xE7,0,0xE6,0xE1,0xE8,0,0xE7,0xE2,0xE9,0,0xE8,0xE3,0xEA,0,
    0xE9,0xE4,0xEB,0,0xEA,0xE5,0xEC,0,0xEB,0xE6,0xED,0,0xEC,0xE7,0xEE,0,
    0xED,0xE8,0xEF,0,0xEE,0xE9,0xF0,0,0xEF,0xEA,0xF1,0,0xF0,0xEB,0xF2,0,
    0xF1,0xEC,0xF3,0,0xF2,0xED,0xF4,0,0xF3,0xEE,0xF5,0,0xF4,0xEF,0xF6,0,
    0xF5,0xF0,0xF7,0,0xF6,0xF1,0xF8,0,0xF7,0xF2,0xF9,0,0xF8,0xF3,0xFA,0,
    0xF9,0xF4,0xFB,0,0xFA,0xF5,0xFC,0,0xFB,0xF6,0xFD,0,0xFC,0xF7,0xFE,0,
    0xFD,0xF8,0xFF,0,0xFE,0xF9,0x00,0,0xFF,0xFA,0x01,0,0x00,0xFB,0x02,0,
    0x01,0xFC,0x03,0,0x02,0xFD,0x04,0,0x03,0xFE,0x05,0,0x04,0xFF,0x06,0,
    0x05,0x00,0x07,0,0x06,0x01,0x08,0,0x07,0x02,0x09,0,0x08,0x03,0x0A,0,
    0x09,0x04,0x0B,0,0x0A,0x05,0x0C,0,0x0B,0x06,0x0D,0,0x0C,0x07,0x0E,0,
    0x0D,0x08,0x0F,0,0x0E,0x09,0x10,0,0x0F,0x0A,0x11,0,0x10,0x0B,0x12,0,
    0x11,0x0C,0x13,0,0x12,0x0D,0x14,0,0x13,0x0E,0x15,0,0x14,0x0F,0x16,0,
    0x15,0x10,0x17,0,0x16,0x11,0x18,0,0x17,0x12,0x19,0,0x18,0x13,0x1A,0,
    0x19,0x14,0x1B,0,0x1A,0x15,0x1C,0,0x1B,0x16,0x1D,0,0x1C,0x17,0x1E,0,
    0x1D,0x18,0x1F,0,0x1E,0x19,0x20,0,0x1F,0x1A,0x21,0,0x20,0x1B,0x22,0,
    0x21,0x1C,0x23,0,0x22,0x1D,0x24,0,0x23,0x1E,0x25,0,0x24,0x1F,0x26,0,
    0xE6,0xE0,0xE7,0,0xE7,0xE1,0xE8,0,0xE8,0xE2,0xE9,0,0xE9,0xE3,0xEA,0,
    0xEA,0xE4,0xEB,0,0xEB,0xE5,0xEC,0,0xEC,0xE6,0xED,0,0xED,0xE7,0xEE,0,
    0xEE,0xE8,0xEF,0,0xEF,0xE9,0xF0,0,0xF0,0xEA,0xF1,0,0xF1,0xEB,0xF2,0,
    0xF2,0xEC,0xF3,0,0xF3,0xED,0xF4,0,0xF4,0xEE,0xF5,0,0xF5,0xEF,0xF6,0,
    0xF6,0xF0,0xF7,0,0xF7,0xF1,0xF8,0,0xF8,0xF2,0xF9,0,0xF9,0xF3,0xFA,0,
    0xFA,0xF4,0xFB,0,0xFB,0xF5,0xFC,0,0xFC,0xF6,0xFD,0,0xFD,0xF7,0xFE,0,
    0xFE,0xF8,0xFF,0,0xFF,0xF9,0x00,0,0x00,0xFA,0x01,0,0x01,0xFB,0x02,0,
    0x02,0xFC,0x03,0,0x03,0xFD,0x04,0,0x04,0xFE,0x05,0,0x05,0xFF,0x06,0,
    0x06,0x00,0x07,0,0x07,0x01,0x08,0,0x08,0x02,0x09,0,0x09,0x03,0x0A,0,
    0x0A,0x04,0x0B,0,0x0B,0x05,0x0C,0,0x0C,0x06,0x0D,0,0x0D,0x07,0x0E,0,
    0x0E,0x08,0x0F,0,0x0F,0x09,0x10,0,0x10,0x0A,0x11,0,0x11,0x0B,0x12,0,
    0x12,0x0C,0x13,0,0x13,0x0D,0x14,0,0x14,0x0E,0x15,0,0x15,0x0F,0x16,0,
    0x16,0x10,0x17,0,0x17,0x11,0x18,0,0x18,0x12,0x19,0,0x19,0x13,0x1A,0,
    0x1A,0x14,0x1B,0,0x1B,0x15,0x1C,0,0x1C,0x16,0x1D,0,0x1D,0x17,0x1E,0,
    0x1E,0x18,0x1F,0,0x1F,0x19,0x20,0,0x20,0x1A,0x21,0,0x21,0x1B,0x22,0,
    0x22,0x1C,0x23,0,0x23,0x1D,0x24,0,0x24,0x1E,0x25,0,0x25,0x1F,0x26,0,
    0xE7,0xE0,0xE7,0,0xE8,0xE1,0xE8,0,0xE9,0xE2,0xE9,0,0xEA,0xE3,0xEA,0,
    0xEB,0xE4,0xEB,0,0xEC,0xE5,0xEC,0,0xED,0xE6,0xED,0,0xEE,0xE7,0xEE,0,
    0xEF,0xE8,0xEF,0,0xF0,0xE9,0xF0,0,0xF1,0xEA,0xF1,0,0xF2,0xEB,0xF2,0,
    0xF3,0xEC,0xF3,0,0xF4,0xED,0xF4,0,0xF5,0xEE,0xF5,0,0xF6,0xEF,0xF6,0,
    0xF7,0xF0,0xF7,0,0xF8,0xF1,0xF8,0,0xF9,0xF2,0xF9,0,0xFA,0xF3,0xFA,0,
    0xFB,0xF4,0xFB,0,0xFC,0xF5,0xFC,0,0xFD,0xF6,0xFD,0,0xFE,0xF7,0xFE,0,
    0xFF,0xF8,0xFF,0,0x00,0xF9,0x00,0,0x01,0xFA,0x01,0,0x02,0xFB,0x02,0,
    0x03,0xFC,0x03,0,0x04,0xFD,0x04,0,0x05,0xFE,0x05,0,0x06,0xFF,0x06,0,
    0x07,0x00,0x07,0,0x08,0x01,0x08,0,0x09,0x02,0x09,0,0x0A,0x03,0x0A,0,
    0x0B,0x04,0x0B,0,0x0C,0x05,0x0C,0,0x0D,0x06,0x0D,0,0x0E,0x07,0x0E,0,
    0x0F,0x08,0x0F,0,0x10,0x09,0x10,0,0x11,0x0A,0x11,0,0x12,0x0B,0x12,0,
    0x13,0x0C,0x13,0,0x14,0x0D,0x14,0,0x15,0x0E,0x15,0,0x16,0x0F,0x16,0,
    0x17,0x10,0x17,0,0x18,0x11,0x18,0,0x19,0x12,0x19,0,0x1A,0x13,0x1A,0,
    0x1B,0x14,0x1B,0,0x1C,0x15,0x1C,0,0x1D,0x16,0x1D,0,0x1E,0x17,0x1E,0,
    0x1F,0x18,0x1F,0,0x20,0x19,0x20,0,0x21,0x1A,0x21,0,0x22,0x1B,0x22,0,
    0x23,0x1C,0x23,0,0x24,0x1D,0x24,0,0x25,0x1E,0x25,0,0x26,0x1F,0x26,0,
};
#endif

/* --- Basics --- */

// This implementation assumes that:
//  - converting a uint32_t to a size_t will never overflow.
//  - converting a size_t to a uint64_t will never overflow.
#if defined(__WORDSIZE)
#if (__WORDSIZE != 32) && (__WORDSIZE != 64)
#error "qoir.h requires a word size of either 32 or 64 bits"
#endif
#endif

// Normally, the qoir_private_peek_etc and qoir_private_poke_etc
// implementations are both (1) correct regardless of CPU endianness and (2)
// very fast (e.g. an inlined qoir_private_peek_u32le call, in an optimized
// clang or gcc build, is a single MOV instruction on x86_64).
//
// However, the endian-agnostic implementations are slow on Microsoft's C
// compiler (MSC). Alternative memcpy-based implementations restore speed, but
// they are only correct on little-endian CPU architectures. Defining
// QOIR_USE_MEMCPY_LE_PEEK_POKE opts in to these implementations.
//
// https://godbolt.org/z/q4MfjzTPh

#define get32(a) le32toh(*(uint32_t*)(a))
#define get64(a) le64toh(*(uint64_t*)(a))

static inline uint32_t
_peek_u32le(const uint8_t* p)
{
#if defined(QOIR_USE_MEMCPY_LE_PEEK_POKE)
    uint32_t x;
    memcpy(&x, p, 4);
    return x;
#else
    return ((uint32_t)(p[0]) << 0) | ((uint32_t)(p[1]) << 8) |
           ((uint32_t)(p[2]) << 16) | ((uint32_t)(p[3]) << 24);
#endif
}

static inline uint64_t
_peek_u64le(const uint8_t* p)
{
#if defined(QOIR_USE_MEMCPY_LE_PEEK_POKE)
    uint64_t x;
    memcpy(&x, p, 8);
    return x;
#else
    return ((uint64_t)(p[0]) << 0) | ((uint64_t)(p[1]) << 8) |
           ((uint64_t)(p[2]) << 16) | ((uint64_t)(p[3]) << 24) |
           ((uint64_t)(p[4]) << 32) | ((uint64_t)(p[5]) << 40) |
           ((uint64_t)(p[6]) << 48) | ((uint64_t)(p[7]) << 56);
#endif
}


#if 0
static inline bool              //
qoir_private_u64_overflow_add(  //
    uint64_t* x,                //
    uint64_t y) {
  uint64_t old = *x;
  *x += y;
  return *x < old;
}
#endif


static inline uint32_t
_tile_dimension(int interior, uint32_t pixel_dimension)
{
    return interior ? QOIR_TILE_SIZE : (((pixel_dimension - 1) & QOIR_TILE_MASK) + 1);
}

// QOIR_TILE_LZ4_COMPRESSION_WORST_CASE equals
// qoir_lz4_block_encode_worst_case_dst_len(4 * QTS * QTS).
#define QOIR_TILE_LZ4_COMPRESSION_WORST_CASE \
  ((4 * QOIR_TILE_SIZE * QOIR_TILE_SIZE) +   \
   ((4 * QOIR_TILE_SIZE * QOIR_TILE_SIZE) / 255) + 16)

/* --- SWAR (SIMD Within a Register) --- */

// These treat a u32 as a u8x4 vector.

// QOIR_SWAR_PADDB returns (a + b).
#define QOIR_SWAR_PADDB(a, b) \
  (((a & 0x7F7F7F7F) + (b & 0x7F7F7F7F)) ^ ((a ^ b) & 0x80808080))

// QOIR_SWAR_PSUBB returns (a - b).
#define QOIR_SWAR_PSUBB(a, b) \
  ((a | 0x80808080) - (b & 0x7F7F7F7F)) ^ ((a ^ ~b) & 0x80808080)

/* --- Status Messages --- */

#if 0
const char qoir_lz4_status_message__error_dst_is_too_short[] = "#qoir/lz4: dst is too short";
const char qoir_lz4_status_message__error_invalid_data[] = "#qoir/lz4: invalid data";
const char qoir_lz4_status_message__error_src_is_too_long[] = "#qoir/lz4: src is too long";

const char qoir_status_message__error_invalid_argument[] = "#qoir: invalid argument";
const char qoir_status_message__error_invalid_data[] = "#qoir: invalid data";
const char qoir_status_message__error_out_of_memory[] = "#qoir: out of memory";
const char qoir_status_message__error_unsupported_metadata_size[] = "#qoir: unsupported metadata size";
const char qoir_status_message__error_unsupported_pixbuf_dimensions[] = "#qoir: unsupported pixbuf dimensions";
const char qoir_status_message__error_unsupported_pixfmt[] = "#qoir: unsupported pixfmt";
const char qoir_status_message__error_unsupported_tile_format[] = "#qoir: unsupported tile format";
#endif

/* --- Pixel Swizzlers --- */

// The DST and SRC in qoir_private_swizzle__DST__SRC means:
//  - bgr,  rgb  = 3 bytes per pixel, opaque.
//  - bgra, rgba = 4 bytes per pixel, some sort of alpha.
//  - bgrn, rgbn = 4 bytes per pixel, nonpremultiplied alpha.
//  - bgrp, rgbp = 4 bytes per pixel, premultiplied alpha.
//  - bgrx, rgbx = 4 bytes per pixel, opaque (every 4th byte is ignored).
//
// As for bgr versus rgb, letter order matches in-memory byte order,
// independent of endianness.
//
// Unlike PNG, it's always 8 bits per channel, in memory. No more, no less.
// However, when converting between nonpremultiplied and premultiplied alpha,
// the computation happens in 16-bit space (and hence the multiplication by one
// or two factors of 0x101).
//
// Working in the higher bit depth can produce slightly different (and arguably
// slightly more accurate) results. For example, given 8-bit blue and alpha of
// 0x80 and 0x81:
//
//  - ((0x80   * 0x81  ) / 0xFF  )      = 0x40        = 0x40
//  - ((0x8080 * 0x8181) / 0xFFFF) >> 8 = 0x4101 >> 8 = 0x41

typedef void (*swizzle_func_t)(uint8_t *dst, size_t dst_stride, const uint8_t *src, size_t src_stride, size_t width, size_t height);


static void
_swizzle__copy_4(uint8_t *dst_ptr, size_t dst_stride, const uint8_t *src_ptr, size_t src_stride, size_t width, size_t height)
{
    size_t n = 4 * width;
    for(;height > 0;--height){
        memcpy(dst_ptr, src_ptr, n);
        dst_ptr += dst_stride;
        src_ptr += src_stride;
    }
}


static void
_swizzle__bgr__bgrn(uint8_t *dst, size_t dst_stride, const uint8_t *src, size_t src_stride, size_t width, size_t height)
{
  for (; height > 0; height--) {
    for (size_t n = width; n > 0; n--) {
      uint8_t s0 = *src++;
      uint8_t s1 = *src++;
      uint8_t s2 = *src++;
      uint8_t s3 = *src++;
      uint32_t alpha_101_101 = ((uint32_t)s3) * (0x101 * 0x101);
      *dst++ = (uint8_t)(((s0 * alpha_101_101) / 0xFFFF) >> 8);
      *dst++ = (uint8_t)(((s1 * alpha_101_101) / 0xFFFF) >> 8);
      *dst++ = (uint8_t)(((s2 * alpha_101_101) / 0xFFFF) >> 8);
    }
    dst += dst_stride - (3 * width);
    src += src_stride - (4 * width);
  }
}


static void
_swizzle__bgr__bgrp(uint8_t *dst, size_t dst_stride, const uint8_t *src, size_t src_stride, size_t width, size_t height)
{
  for (; height > 0; height--) {
    for (size_t n = width; n > 0; n--) {
      uint8_t s0 = *src++;
      uint8_t s1 = *src++;
      uint8_t s2 = *src++;
      src++;
      *dst++ = s0;
      *dst++ = s1;
      *dst++ = s2;
    }
    dst += dst_stride - (3 * width);
    src += src_stride - (4 * width);
  }
}


static void
_swizzle__bgr__rgbn(uint8_t *dst, size_t dst_stride, const uint8_t *src, size_t src_stride, size_t width, size_t height)
{
  for (; height > 0; height--) {
    for (size_t n = width; n > 0; n--) {
      uint8_t s0 = *src++;
      uint8_t s1 = *src++;
      uint8_t s2 = *src++;
      uint8_t s3 = *src++;
      uint32_t alpha_101_101 = ((uint32_t)s3) * (0x101 * 0x101);
      *dst++ = (uint8_t)(((s2 * alpha_101_101) / 0xFFFF) >> 8);
      *dst++ = (uint8_t)(((s1 * alpha_101_101) / 0xFFFF) >> 8);
      *dst++ = (uint8_t)(((s0 * alpha_101_101) / 0xFFFF) >> 8);
    }
    dst += dst_stride - (3 * width);
    src += src_stride - (4 * width);
  }
}


static void
_swizzle__bgr__rgbp(uint8_t *dst, size_t dst_stride, const uint8_t * src, size_t src_stride, size_t width, size_t height)
{
  for (; height > 0; height--) {
    for (size_t n = width; n > 0; n--) {
      uint8_t s0 = *src++;
      uint8_t s1 = *src++;
      uint8_t s2 = *src++;
      src++;
      *dst++ = s2;
      *dst++ = s1;
      *dst++ = s0;
    }
    dst += dst_stride - (3 * width);
    src += src_stride - (4 * width);
  }
}


#if 0
static void
_swizzle__bgra__bgr(uint8_t *dst_ptr, size_t dst_stride, const uint8_t *src_ptr, size_t src_stride, size_t width, size_t height)
{
  for (; height_in_pixels > 0; height_in_pixels--) {
    for (size_t n = width_in_pixels; n > 0; n--) {
      uint8_t s0 = *src_ptr++;
      uint8_t s1 = *src_ptr++;
      uint8_t s2 = *src_ptr++;
      *dst_ptr++ = s0;
      *dst_ptr++ = s1;
      *dst_ptr++ = s2;
      *dst_ptr++ = 0xFF;
    }
    dst_ptr += dst_stride_in_bytes - (4 * width_in_pixels);
    src_ptr += src_stride_in_bytes - (3 * width_in_pixels);
  }
}
#endif


static void
_swizzle__bgra__bgrx(uint8_t *dst, size_t dst_stride, const uint8_t *src, size_t src_stride, size_t width, size_t height)
{
  for (; height > 0; height--) {
    for (size_t n = width; n > 0; n--) {
      uint8_t s0 = *src++;
      uint8_t s1 = *src++;
      uint8_t s2 = *src++;
      src++;
      *dst++ = s0;
      *dst++ = s1;
      *dst++ = s2;
      *dst++ = 0xFF;
    }
    dst += dst_stride - (4 * width);
    src += src_stride - (4 * width);
  }
}


#if 0
static void                                //
qoir_private_swizzle__bgra__rgb(           //
    uint8_t* QOIR_RESTRICT dst_ptr,        //
    size_t dst_stride_in_bytes,            //
    const uint8_t* QOIR_RESTRICT src_ptr,  //
    size_t src_stride_in_bytes,            //
    size_t width_in_pixels,                //
    size_t height_in_pixels) {
  for (; height_in_pixels > 0; height_in_pixels--) {
    for (size_t n = width_in_pixels; n > 0; n--) {
      uint8_t s0 = *src_ptr++;
      uint8_t s1 = *src_ptr++;
      uint8_t s2 = *src_ptr++;
      *dst_ptr++ = s2;
      *dst_ptr++ = s1;
      *dst_ptr++ = s0;
      *dst_ptr++ = 0xFF;
    }
    dst_ptr += dst_stride_in_bytes - (4 * width_in_pixels);
    src_ptr += src_stride_in_bytes - (3 * width_in_pixels);
  }
}
#endif


static void
_swizzle__bgra__rgba(uint8_t *dst, size_t dst_stride, const uint8_t *src, size_t src_stride, size_t width, size_t height)
{
  for (; height > 0; height--) {
    for (size_t n = width; n > 0; n--) {
      uint8_t s0 = *src++;
      uint8_t s1 = *src++;
      uint8_t s2 = *src++;
      uint8_t s3 = *src++;
      *dst++ = s2;
      *dst++ = s1;
      *dst++ = s0;
      *dst++ = s3;
    }
    dst += dst_stride - (4 * width);
    src += src_stride - (4 * width);
  }
}


static void
_swizzle__bgra__rgbx(uint8_t *dst, size_t dst_stride, const uint8_t *src, size_t src_stride, size_t width, size_t height)
{
  for (; height > 0; height--) {
    for (size_t n = width; n > 0; n--) {
      uint8_t s0 = *src++;
      uint8_t s1 = *src++;
      uint8_t s2 = *src++;
      src++;
      *dst++ = s2;
      *dst++ = s1;
      *dst++ = s0;
      *dst++ = 0xFF;
    }
    dst += dst_stride - (4 * width);
    src += src_stride - (4 * width);
  }
}


static void
_swizzle__bgrn__bgrp(uint8_t *dst, size_t dst_stride, const uint8_t *src, size_t src_stride, size_t width, size_t height)
{
  for (; height > 0; height--) {
    for (size_t n = width; n > 0; n--) {
      uint8_t s0 = *src++;
      uint8_t s1 = *src++;
      uint8_t s2 = *src++;
      uint8_t s3 = *src++;
      if (s3 == 0xFF) {
        *dst++ = s0;
        *dst++ = s1;
        *dst++ = s2;
        *dst++ = s3;
      } else if (s3 == 0x00) {
        *dst++ = 0x00;
        *dst++ = 0x00;
        *dst++ = 0x00;
        *dst++ = 0x00;
      } else {
        const uint32_t ffff_101 = 0xFFFF * 0x101;
        uint32_t alpha_101 = (uint32_t)s3 * 0x101;
        *dst++ = (uint8_t)(((s0 * ffff_101) / alpha_101) >> 8);
        *dst++ = (uint8_t)(((s1 * ffff_101) / alpha_101) >> 8);
        *dst++ = (uint8_t)(((s2 * ffff_101) / alpha_101) >> 8);
        *dst++ = s3;
      }
    }
    dst += dst_stride - (4 * width);
    src += src_stride - (4 * width);
  }
}


static void
_swizzle__bgrn__rgbp(uint8_t *dst, size_t dst_stride, const uint8_t *src, size_t src_stride, size_t width, size_t height)
{
  for (; height > 0; height--) {
    for (size_t n = width; n > 0; n--) {
      uint8_t s0 = *src++;
      uint8_t s1 = *src++;
      uint8_t s2 = *src++;
      uint8_t s3 = *src++;
      if (s3 == 0xFF) {
        *dst++ = s0;
        *dst++ = s1;
        *dst++ = s2;
        *dst++ = s3;
      } else if (s3 == 0x00) {
        *dst++ = 0x00;
        *dst++ = 0x00;
        *dst++ = 0x00;
        *dst++ = 0x00;
      } else {
        const uint32_t ffff_101 = 0xFFFF * 0x101;
        uint32_t alpha_101 = (uint32_t)s3 * 0x101;
        *dst++ = (uint8_t)(((s2 * ffff_101) / alpha_101) >> 8);
        *dst++ = (uint8_t)(((s1 * ffff_101) / alpha_101) >> 8);
        *dst++ = (uint8_t)(((s0 * ffff_101) / alpha_101) >> 8);
        *dst++ = s3;
      }
    }
    dst += dst_stride - (4 * width);
    src += src_stride - (4 * width);
  }
}


static void
_swizzle__bgrp__bgrn(uint8_t *dst, size_t dst_stride, const uint8_t *src, size_t src_stride, size_t width, size_t height)
{
  for (; height > 0; height--) {
    for (size_t n = width; n > 0; n--) {
      uint8_t s0 = *src++;
      uint8_t s1 = *src++;
      uint8_t s2 = *src++;
      uint8_t s3 = *src++;
      uint32_t alpha_101_101 = ((uint32_t)s3) * (0x101 * 0x101);
      *dst++ = (uint8_t)(((s0 * alpha_101_101) / 0xFFFF) >> 8);
      *dst++ = (uint8_t)(((s1 * alpha_101_101) / 0xFFFF) >> 8);
      *dst++ = (uint8_t)(((s2 * alpha_101_101) / 0xFFFF) >> 8);
      *dst++ = s3;
    }
    dst += dst_stride - (4 * width);
    src += src_stride - (4 * width);
  }
}


static void
_swizzle__bgrp__rgbn(uint8_t *dst, size_t dst_stride, const uint8_t *src, size_t src_stride, size_t width, size_t height)
{
  for (; height > 0; height--) {
    for (size_t n = width; n > 0; n--) {
      uint8_t s0 = *src++;
      uint8_t s1 = *src++;
      uint8_t s2 = *src++;
      uint8_t s3 = *src++;
      uint32_t alpha_101_101 = ((uint32_t)s3) * (0x101 * 0x101);
      *dst++ = (uint8_t)(((s2 * alpha_101_101) / 0xFFFF) >> 8);
      *dst++ = (uint8_t)(((s1 * alpha_101_101) / 0xFFFF) >> 8);
      *dst++ = (uint8_t)(((s0 * alpha_101_101) / 0xFFFF) >> 8);
      *dst++ = s3;
    }
    dst += dst_stride - (4 * width);
    src += src_stride - (4 * width);
  }
}

// -------- LZ4 Decode

static ssize_t
qoir_lz4_block_decode(uint8_t *dst_ptr, size_t dst_len, const uint8_t *src_ptr, size_t src_len)
{
    //size_result_t result = {0};
    if (src_len > QOIR_LZ4_BLOCK_DECODE_MAX_INCL_SRC_LEN) {
        //result.status_message = qoir_lz4_status_message__error_src_is_too_long;
        return -1;
    }

#if defined(QOIR_CONFIG__USE_OFFICIAL_LZ4_LIBRARY)
    int n = LZ4_decompress_safe((const char*)src_ptr, (char*)dst_ptr, (int)src_len,
                                ((dst_len > INT_MAX) ? INT_MAX : (int)dst_len));
    if (n < 0) {
        //result.status_message = qoir_lz4_status_message__error_invalid_data;
    }
    return n;

#else
    uint8_t* const original_dst_ptr = dst_ptr;

    // See https://github.com/lz4/lz4/blob/dev/doc/lz4_Block_format.md for file
    // format details, such as the LZ4 token's bit patterns.
    while (src_len > 0) {
        uint32_t token = *src_ptr++;
        src_len--;

        uint32_t literal_len = token >> 4;
        if (literal_len > 0) {
            if (literal_len == 15) {
                while (1) {
                    if (src_len == 0) {
                        goto fail_invalid_data;
                    }
                    uint32_t s = *src_ptr++;
                    src_len--;
                    literal_len += s;
                    if (s != 255) {
                        break;
                    }
                }
            }

            if (literal_len > src_len){
                goto fail_invalid_data;
            }else if (literal_len > dst_len){
                //result.status_message = qoir_lz4_status_message__error_dst_is_too_short;
                return -1;
            }
            memcpy(dst_ptr, src_ptr, literal_len);
            dst_ptr += literal_len;
            dst_len -= literal_len;
            src_ptr += literal_len;
            src_len -= literal_len;
            if (src_len == 0) {
                //result.value = ((size_t)(dst_ptr - original_dst_ptr));
                //return result;
                return ((size_t)(dst_ptr - original_dst_ptr));
            }
        }

        if (src_len < 2) {
            goto fail_invalid_data;
        }
        uint32_t copy_off = ((uint32_t)src_ptr[0]) | (((uint32_t)src_ptr[1]) << 8);
        src_ptr += 2;
        src_len -= 2;
        if ((copy_off == 0) || (copy_off > ((size_t)(dst_ptr - original_dst_ptr)))){
            goto fail_invalid_data;
        }

        uint32_t copy_len = (token & 15) + 4;
        if (copy_len == 19) {
            while (1) {
                if (src_len == 0) {
                    goto fail_invalid_data;
                }
                uint32_t s = *src_ptr++;
                src_len--;
                copy_len += s;
                if (s != 255) {
                    break;
                }
            }
        }

        if (dst_len < copy_len) {
            //result.status_message = qoir_lz4_status_message__error_dst_is_too_short;
            return -1;
        }
        dst_len -= copy_len;
        for (const uint8_t* from = dst_ptr - copy_off; copy_len > 0; copy_len--) {
            *dst_ptr++ = *from++;
        }
    }

fail_invalid_data:
    //result.status_message = qoir_lz4_status_message__error_invalid_data;
    return -1;
#endif  // QOIR_CONFIG__USE_OFFICIAL_LZ4_LIBRARY
}


// -------- QOIR Decode

#if 0
static qoir_decode_pixel_configuration_result  //
qoir_decode_pixel_configuration(                          //
    const uint8_t* src_ptr,                               //
    size_t src_len) {
  qoir_decode_pixel_configuration_result result = {0};

  if ((src_len < 20) ||
      (qoir_private_peek_u32le(src_ptr) != 0x52494F51)) {  // "QOIR"le.
    result.status_message = qoir_status_message__error_invalid_data;
    return result;
  }
  uint64_t qoir_chunk_payload_len = qoir_private_peek_u64le(src_ptr + 4);
  if ((qoir_chunk_payload_len < 8) ||
      (qoir_chunk_payload_len > 0x7FFFFFFFFFFFFFFFull)) {
    result.status_message = qoir_status_message__error_invalid_data;
    return result;
  }

  uint32_t header0 = qoir_private_peek_u32le(src_ptr + 12);
  uint32_t width_in_pixels = 0xFFFFFF & header0;
  qoir_pixel_format src_pixfmt = 0x0F & (header0 >> 24);
  switch (src_pixfmt) {
    case QOIR_PIXEL_FORMAT__BGRX:
    case QOIR_PIXEL_FORMAT__BGRA_NONPREMUL:
    case QOIR_PIXEL_FORMAT__BGRA_PREMUL:
      break;
    default:
      result.status_message = qoir_status_message__error_invalid_data;
      return result;
  }
  uint32_t header1 = qoir_private_peek_u32le(src_ptr + 16);
  uint32_t height_in_pixels = 0xFFFFFF & header1;

  result.dst_pixcfg.pixfmt = src_pixfmt;
  result.dst_pixcfg.width_in_pixels = width_in_pixels;
  result.dst_pixcfg.height_in_pixels = height_in_pixels;
  return result;
}
#endif

static swizzle_func_t
_choose_decode_swizzle_func(pixel_format_t dst_pixfmt, pixel_format_t src_pixfmt)
{
    switch (dst_pixfmt) {
    case PIXEL_FORMAT__BGRX:
        switch (src_pixfmt) {
        case PIXEL_FORMAT__BGRX: return _swizzle__copy_4;
        case PIXEL_FORMAT__BGRA_NONPREMUL: return _swizzle__bgrp__bgrn;
        case PIXEL_FORMAT__BGRA_PREMUL: return _swizzle__copy_4;
        default: return NULL;
        }
        break;
    case PIXEL_FORMAT__BGRA_NONPREMUL:
        switch (src_pixfmt){
        case PIXEL_FORMAT__BGRX: return _swizzle__bgra__bgrx;
        case PIXEL_FORMAT__BGRA_NONPREMUL: return _swizzle__copy_4;
        case PIXEL_FORMAT__BGRA_PREMUL: return _swizzle__bgrn__bgrp;
        default: return NULL;
        }
        break;
    case PIXEL_FORMAT__BGRA_PREMUL:
        switch (src_pixfmt){
        case PIXEL_FORMAT__BGRX: return _swizzle__bgra__bgrx;
        case PIXEL_FORMAT__BGRA_NONPREMUL: return _swizzle__bgrp__bgrn;
        case PIXEL_FORMAT__BGRA_PREMUL: return _swizzle__copy_4;
        default: return NULL;
        }
        break;
    case PIXEL_FORMAT__BGR:
        switch (src_pixfmt){
        case PIXEL_FORMAT__BGRX: return _swizzle__bgr__bgrp;
        case PIXEL_FORMAT__BGRA_NONPREMUL: return _swizzle__bgr__bgrn;
        case PIXEL_FORMAT__BGRA_PREMUL: return _swizzle__bgr__bgrp;
        default: return NULL;
        }
        break;
    case PIXEL_FORMAT__RGBX:
        switch (src_pixfmt){
        case PIXEL_FORMAT__BGRX: return _swizzle__bgra__rgba;
        case PIXEL_FORMAT__BGRA_NONPREMUL: return _swizzle__bgrp__rgbn;
        case PIXEL_FORMAT__BGRA_PREMUL: return _swizzle__bgra__rgba;
        default: return NULL;
        }
        break;
    case PIXEL_FORMAT__RGBA_NONPREMUL:
        switch (src_pixfmt){
        case PIXEL_FORMAT__BGRX: return _swizzle__bgra__rgbx;
        case PIXEL_FORMAT__BGRA_NONPREMUL: return _swizzle__bgra__rgba;
        case PIXEL_FORMAT__BGRA_PREMUL: return _swizzle__bgrn__rgbp;
        default: return NULL;
        }
        break;
    case PIXEL_FORMAT__RGBA_PREMUL:
        switch (src_pixfmt){
        case PIXEL_FORMAT__BGRX: return _swizzle__bgra__rgbx;
        case PIXEL_FORMAT__BGRA_NONPREMUL: return _swizzle__bgrp__rgbn;
        case PIXEL_FORMAT__BGRA_PREMUL: return _swizzle__bgra__rgba;
        default: return NULL;
        }
        break;
    case PIXEL_FORMAT__RGB:
        switch (src_pixfmt){
        case PIXEL_FORMAT__BGRX: return _swizzle__bgr__rgbp;
        case PIXEL_FORMAT__BGRA_NONPREMUL: return _swizzle__bgr__rgbn;
        case PIXEL_FORMAT__BGRA_PREMUL: return _swizzle__bgr__rgbp;
        default: return NULL;
        }
        break;
    default: return NULL;
    }
}

// Callers should pass (QOIR_LITERALS_PRE_PADDING + (4 * tw * th)) for dst_len,
// so that the decode loop can always refer (by a simple offset of -4) to the
// pixel left of the current pixel.
//
// Callers should pass (total_ops_length + 8) for src_len so the decode loop
// can always peek for 8 bytes, even at the end of the stream. Reference: §
static ssize_t
_decode_tile_ops(uint8_t *dst_ptr, size_t dst_len, const uint8_t *src_ptr, size_t src_len)
{
    //size_result_t result = {0};

    if ((dst_len < QOIR_LITERALS_PRE_PADDING) || (src_len < 8)) {
        //result.status_message = qoir_status_message__error_invalid_argument;
        return -1;
    }

    // color_cache is conceptually "uint8_t color_cache[64][4]" but is flattened
    // for performance. It is always indexed by a uint8_t value that is a
    // multiple of four. The four uint8_t elements are in B, G, R, A order.
    uint8_t color_cache[256];
    for (int i = 0; i < 256; i += 4) {
        color_cache[i + 0] = 0x00;
        color_cache[i + 1] = 0x00;
        color_cache[i + 2] = 0x00;
        color_cache[i + 3] = 0xFF;
    }
    uint8_t next_color_index = 0;

  uint8_t* dp = dst_ptr + QOIR_LITERALS_PRE_PADDING;
  uint8_t* dq = dst_ptr + dst_len;
  const uint8_t* sp = src_ptr;
  const uint8_t* sq = src_ptr + src_len - 8;
  while (dp < dq) {
    if (sp >= sq) {
      //result.status_message = qoir_status_message__error_invalid_data;
      //return result;
      return -1;
    }

    uint8_t pixel[4];
    memcpy(pixel, dp - 4, 4);

    uint64_t s64 = _peek_u64le(sp);
    if ((s64 & 0xFF) == 0xF7) {  // QOIR_OP_BGR8
      pixel[0] += (uint8_t)(s64 >> 0x08);
      pixel[1] += (uint8_t)(s64 >> 0x10);
      pixel[2] += (uint8_t)(s64 >> 0x18);
      sp += 4;
      memcpy(color_cache + next_color_index, pixel, 4);
      next_color_index += 4;
      memcpy(dp, pixel, 4);
      dp += 4;

    } else if ((s64 & 0x03) == 0) {  // QOIR_OP_INDEX
      sp += 1;
      memcpy(pixel, color_cache + (uint8_t)s64, 4);
      memcpy(dp, pixel, 4);
      dp += 4;

    } else if ((s64 & 0x03) == 1) {  // QOIR_OP_BGR2
      uint32_t delta8x4 = (uint32_t)(((s64 >> 0x02) & 0x000003) |  //
                                     ((s64 << 0x04) & 0x000300) |  //
                                     ((s64 << 0x0A) & 0x030000));
      delta8x4 = QOIR_SWAR_PSUBB(delta8x4, 0x020202);
      uint32_t pixel8x4;
      memcpy(&pixel8x4, pixel, 4);
#ifdef QOIR_USE_SIMD_SSE2
      pixel8x4 = (uint32_t)_mm_cvtsi128_si32(_mm_add_epi8(
          _mm_cvtsi32_si128((int)pixel8x4), _mm_cvtsi32_si128((int)delta8x4)));
#else
      pixel8x4 = QOIR_SWAR_PADDB(pixel8x4, delta8x4);
#endif
      memcpy(pixel, &pixel8x4, 4);

      sp += 1;
      memcpy(color_cache + next_color_index, pixel, 4);
      next_color_index += 4;
      memcpy(dp, pixel, 4);
      dp += 4;

    } else if ((s64 & 0x03) == 2) {  // QOIR_OP_LUMA
#ifndef QOIR_CONFIG__DISABLE_LARGE_LOOK_UP_TABLES
      uint32_t delta8x4;
      memcpy(&delta8x4, qoir_private_table_luma - 2 + (uint16_t)s64, 4);
      uint32_t pixel8x4;
      memcpy(&pixel8x4, pixel, 4);
# if defined(QOIR_USE_SIMD_SSE2)
      pixel8x4 = (uint32_t)_mm_cvtsi128_si32(_mm_add_epi8(
          _mm_cvtsi32_si128((int)pixel8x4), _mm_cvtsi32_si128((int)delta8x4)));
# else
      pixel8x4 = QOIR_SWAR_PADDB(pixel8x4, delta8x4);
# endif
      memcpy(pixel, &pixel8x4, 4);
#else
      uint8_t delta_g = ((uint8_t)s64 >> 0x02) - 32;
      pixel[0] += delta_g - 8 + ((s64 >> 0x08) & 0x0F);
      pixel[1] += delta_g;
      pixel[2] += delta_g - 8 + ((s64 >> 0x0C) & 0x0F);
#endif
      sp += 2;
      memcpy(color_cache + next_color_index, pixel, 4);
      next_color_index += 4;
      memcpy(dp, pixel, 4);
      dp += 4;

    } else if ((s64 & 0x07) == 3) {  // QOIR_OP_BGR7
      uint32_t delta8x4 = (uint32_t)((((s64 >> 0x03) - 0x000040) & 0x00007F) |
                                     (((s64 >> 0x02) - 0x004000) & 0x007F00) |
                                     (((s64 >> 0x01) - 0x400000) & 0x7F0000));
      delta8x4 |= (delta8x4 & 0x404040) << 1;
      uint32_t pixel8x4;
      memcpy(&pixel8x4, pixel, 4);
#if defined(QOIR_USE_SIMD_SSE2)
      pixel8x4 = (uint32_t)_mm_cvtsi128_si32(_mm_add_epi8(
          _mm_cvtsi32_si128((int)pixel8x4), _mm_cvtsi32_si128((int)delta8x4)));
#else
      pixel8x4 = QOIR_SWAR_PADDB(pixel8x4, delta8x4);
#endif
      memcpy(pixel, &pixel8x4, 4);
      sp += 3;
      memcpy(color_cache + next_color_index, pixel, 4);
      next_color_index += 4;
      memcpy(dp, pixel, 4);
      dp += 4;

    } else if ((s64 & 0xFF) < 0xD7) {  // QOIR_OP_RUNS
      size_t run_length = (s64 & 0xFF) >> 0x03;
      if (((size_t)(dq - dp)) < (4 * (run_length + 1))) {
          //result.status_message = qoir_status_message__error_invalid_data;
          return -1;
      }
      do {
        memcpy(dp, pixel, 4);
        dp += 4;
      } while (run_length--);
      sp += 1;

    } else if ((s64 & 0xFF) == 0xD7) {  // QOIR_OP_RUNL
      size_t run_length = (s64 >> 0x08) & 0xFF;
      if (((size_t)(dq - dp)) < (4 * (run_length + 1))) {
          //result.status_message = qoir_status_message__error_invalid_data;
          return -1;
      }
      do {
        memcpy(dp, pixel, 4);
        dp += 4;
      } while (run_length--);
      sp += 2;

    } else if ((s64 & 0xFF) == 0xDF) {  // QOIR_OP_BGRA2
      pixel[0] += ((s64 >> 0x08) & 0x03) - 2;
      pixel[1] += ((s64 >> 0x0A) & 0x03) - 2;
      pixel[2] += ((s64 >> 0x0C) & 0x03) - 2;
      pixel[3] += ((s64 >> 0x0E) & 0x03) - 2;
      sp += 2;
      memcpy(color_cache + next_color_index, pixel, 4);
      next_color_index += 4;
      memcpy(dp, pixel, 4);
      dp += 4;

    } else if ((s64 & 0xFF) == 0xE7) {  // QOIR_OP_BGRA4
      pixel[0] += ((s64 >> 0x08) & 0x0F) - 8;
      pixel[1] += ((s64 >> 0x0C) & 0x0F) - 8;
      pixel[2] += ((s64 >> 0x10) & 0x0F) - 8;
      pixel[3] += ((s64 >> 0x14) & 0x0F) - 8;
      sp += 3;
      memcpy(color_cache + next_color_index, pixel, 4);
      next_color_index += 4;
      memcpy(dp, pixel, 4);
      dp += 4;

    } else if ((s64 & 0xFF) == 0xEF) {  // QOIR_OP_BGRA8
      pixel[0] += (uint8_t)(s64 >> 0x08);
      pixel[1] += (uint8_t)(s64 >> 0x10);
      pixel[2] += (uint8_t)(s64 >> 0x18);
      pixel[3] += (uint8_t)(s64 >> 0x20);
      sp += 5;
      memcpy(color_cache + next_color_index, pixel, 4);
      next_color_index += 4;
      memcpy(dp, pixel, 4);
      dp += 4;

    } else {  // QOIR_OP_A8
      pixel[3] += (uint8_t)(s64 >> 0x08);
      sp += 2;
      memcpy(color_cache + next_color_index, pixel, 4);
      next_color_index += 4;
      memcpy(dp, pixel, 4);
      dp += 4;
    }
  }

    if (sp != sq) {
        //result.status_message = qoir_status_message__error_invalid_data;
        return -1;
    }
    //result.value = (size_t)(dp - dst_ptr);
    //return result;
    return (size_t)(dp - dst_ptr);
}


static int
_decode_qpix_payload(decode_buffer_t *decbuf,
                     pixel_buffer_t *dst,
                     rectangle_t _dst_clip_rectangle,
                     pixel_format_t src_format,
                     uint32_t src_width,
                     uint32_t src_height,
                     const uint8_t* src,
                     size_t src_len,
                     rectangle_t src_clip_rectangle,
                     int32_t _offset_x,
                     int32_t _offset_y,
                     uint32_t lossiness)
{
  do {
    rectangle_t dst_clip_rect = make_rectangle(0, 0, dst->width, dst->height);
    //dst_clip_rect = rectangle__intersect(dst_clip_rect, dst_clip_rectangle);
    rectangle_t dst_clip_rect_in_src_space;
    dst_clip_rect_in_src_space.x0 = dst_clip_rect.x0;
    dst_clip_rect_in_src_space.y0 = dst_clip_rect.y0;
    dst_clip_rect_in_src_space.x1 = dst_clip_rect.x1;
    dst_clip_rect_in_src_space.y1 = dst_clip_rect.y1;

    size_t height_in_tiles = calculate_number_of_tiles_1d(src_height);
    size_t width_in_tiles = calculate_number_of_tiles_1d(src_width);
    if ((height_in_tiles == 0) || (width_in_tiles == 0))
        goto done;

    size_t ty1 = (height_in_tiles - 1) << QOIR_TILE_SHIFT;
    size_t tx1 = (width_in_tiles - 1) << QOIR_TILE_SHIFT;

    swizzle_func_t swizzle_func = _choose_decode_swizzle_func(dst->format, src_format);
    if(!swizzle_func){
        //return qoir_status_message__error_unsupported_pixfmt;
        return -1;
    }

    int num_dst_channels = _pixel_format_get_bpc(dst->format);

    uint8_t* literals_pre_padding = decbuf->private_impl.literals;
    for (int i = 0; i < QOIR_LITERALS_PRE_PADDING; i += 4) {
      literals_pre_padding[i + 0] = 0x00;
      literals_pre_padding[i + 1] = 0x00;
      literals_pre_padding[i + 2] = 0x00;
      literals_pre_padding[i + 3] = 0xFF;
    }

    // ty, tx, tw and th are the tile's top-left offset, width and height, all
    // measured in pixels.
    for (size_t ty = 0; ty <= ty1; ty += QOIR_TILE_SIZE) {
      for (size_t tx = 0; tx <= tx1; tx += QOIR_TILE_SIZE) {
        size_t tw = _tile_dimension(tx < tx1, src_width);
        size_t th = _tile_dimension(ty < ty1, src_height);
        rectangle_t src_clip_rect =
            make_rectangle((int32_t)(tx + 0), (int32_t)(ty + 0),
                           (int32_t)(tx + tw), (int32_t)(ty + th));
        src_clip_rect = rectangle__intersect(src_clip_rect, src_clip_rectangle);
        src_clip_rect = rectangle__intersect(src_clip_rect, dst_clip_rect_in_src_space);

        if (src_len < 4){
            //return qoir_status_message__error_invalid_data;
            return -1;
        }

        uint32_t prefix = _peek_u32le(src);
        src += 4;
        src_len -= 4;
        size_t tile_len = prefix & 0xFFFFFF;
        if ((src_len < (tile_len + 8)) || (((4 * QOIR_TS2) < tile_len) && ((prefix >> 31) != 0))) {
            //return qoir_status_message__error_invalid_data;
            return -1;
        }

        if (rectangle__is_empty(src_clip_rect)) {
          src += tile_len;
          src_len -= tile_len;
          continue;
        }

        const uint8_t* literals = NULL;
        switch (prefix >> 24) {
          case 0: {  // Literals tile format.
            if (tile_len != (4 * tw * th)) {
                //return qoir_status_message__error_invalid_data;
                return -1;
            }
            literals = src;
            break;
          }
          case 1: {  // Ops tile format.
            ssize_t r = _decode_tile_ops(
                decbuf->private_impl.literals,              //
                QOIR_LITERALS_PRE_PADDING + (4 * tw * th),  //
                src, tile_len + 8);                         // See § for +8.
            if (r < 0) {
                //return r.status_message;
                return -1;
            } else if (r != (QOIR_LITERALS_PRE_PADDING + (4 * tw * th))) {
                //return qoir_status_message__error_invalid_data;
                return -1;
            }
            literals =
                decbuf->private_impl.literals + QOIR_LITERALS_PRE_PADDING;
            break;
          }
          case 2: {  // LZ4-Literals tile format.
            ssize_t r = qoir_lz4_block_decode(
                decbuf->private_impl.literals + QOIR_LITERALS_PRE_PADDING,
                sizeof(decbuf->private_impl.literals) -
                    QOIR_LITERALS_PRE_PADDING,
                src, tile_len);
            if (r < 0) {
                //return qoir_status_message__error_invalid_data;
                return -1;
            } else if (r != (4 * tw * th)) {
                //return qoir_status_message__error_invalid_data;
                return -1;
            }
            literals =
                decbuf->private_impl.literals + QOIR_LITERALS_PRE_PADDING;
            break;
          }
          case 3: {  // LZ4-Ops tile format.
            ssize_t r0 = qoir_lz4_block_decode(
                decbuf->private_impl.ops, sizeof(decbuf->private_impl.ops),
                src, tile_len);
            if (r0 < 0) {
                //return qoir_status_message__error_invalid_data;
                return -1;
            }
            ssize_t r1 = _decode_tile_ops(
                decbuf->private_impl.literals,
                QOIR_LITERALS_PRE_PADDING + (4 * tw * th),
                decbuf->private_impl.ops, r0 + 8);          // See § for +8.
            if(r1 < 0) {
                //return r1.status_message;
                return -1;
            } else if (r1 != (QOIR_LITERALS_PRE_PADDING + (4 * tw * th))){
                //return qoir_status_message__error_invalid_data;
                return -1;
            }
            literals =
                decbuf->private_impl.literals + QOIR_LITERALS_PRE_PADDING;
            break;
          }
          default:
            //return qoir_status_message__error_unsupported_tile_format;
            return -1;
        }

        src += tile_len;
        src_len -= tile_len;

        if (lossiness) {
          uint8_t* p = decbuf->private_impl.ops;
          const uint8_t* q = literals;
          const uint8_t* unlossify =
              qoir_private_table_unlossify[lossiness - 1];
          for (uint32_t i = 4 * tw * th; i > 0; i--) {
            *p++ = unlossify[*q++];
          }
          literals = decbuf->private_impl.ops;
        }

        uint8_t* dp =
            dst->data +
            ((src_clip_rect.y0) * dst->stride) +
            ((src_clip_rect.x0) * num_dst_channels);
        const uint8_t* sp = literals +
                            ((src_clip_rect.y0 & QOIR_TILE_MASK) * 4 * tw) +
                            ((src_clip_rect.x0 & QOIR_TILE_MASK) * 4);
        (*swizzle_func)(dp, dst->stride, sp, 4 * tw,
                        rectangle__width(src_clip_rect),
                        rectangle__height(src_clip_rect));
      }
    }
    }while(0);

done:
    if (src_len != 8) {
        //return qoir_status_message__error_invalid_data;
        return -1;
    }
    return 0;
}


#if 0
static decode_result_t
_make_decode_result_error(const char* status_message)
{
    decode_result_t result = {0};
    result.status_message = status_message;
    return result;
}
#endif


typedef struct{
    int width;
    int height;
    pixel_format_t format;
    int lossiness;
    int bpc;                            /* bytes per color */
#ifdef ENABLE_METADATA
    // Optional metadata chunks.
    const uint8_t *metadata_cicp_ptr;
    size_t metadata_cicp_len;
    const uint8_t *metadata_iccp_ptr;
    size_t metadata_iccp_len;
    const uint8_t *metadata_exif_ptr;
    size_t metadata_exif_len;
    const uint8_t *metadata_xmp_ptr;
    size_t metadata_xmp_len;
#endif
}context_t;


static int
_handle_qoir(const uint8_t *src, size_t len, context_t *c)
{
    uint32_t d;
    if(len < 8)
        return -1;

    d = _peek_u32le(src);
    c->width = 0xFFFFFF & d;
    c->format = 0x0F & (d >> 24);
    d = _peek_u32le(src + 4);
    c->height = 0xFFFFFF & d;
    c->lossiness = 0x07 & (d >> 24);
    c->bpc = _pixel_format_get_bpc(c->format);
    return 0;

}


static int
_handle_qpix(const uint8_t *src, size_t len, context_t *c, pixel_buffer_t *dst)
{
    uint64_t pixbuf_len = c->width * c->height * c->bpc;
    decode_buffer_t *decbuf;
    if(pixbuf_len <= 0){
        /* An empty image is OK. */
        return 0;
    }
    if(pixbuf_len > SIZE_MAX){
        //return _make_decode_result_error(qoir_status_message__error_unsupported_pixbuf_dimensions);
        return -1;
    }

    //decbuf = options ? options->decbuf : NULL;
    decbuf = (decode_buffer_t*)malloc(sizeof(decode_buffer_t));
    if(!decbuf){
        //return _make_decode_result_error(qoir_status_message__error_out_of_memory);
        return -1;
    }
    rectangle_t dst_clip_rectangle = make_rectangle(0, 0, 0xFFFFFF, 0xFFFFFF);
    rectangle_t src_clip_rectangle = make_rectangle(0, 0, 0xFFFFFF, 0xFFFFFF);
    int ret = _decode_qpix_payload(decbuf, dst, dst_clip_rectangle,
                                   c->format,
                                   c->width, c->height, src,
                                   len + 8,  // See § for +8.
                                   src_clip_rectangle, 0, 0, c->lossiness);
    free(decbuf);

    return ret;
}



static const char **
_qoir_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-qoir",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_qoir_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;

    return h;
}


static void
_qoir_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


#if 0
static cairo_surface_t *
_surface_from_pixbuf(pixel_buffer_t *pb)
{
    uint8_t *src = pb->data;
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, pb->pixcfg.width, pb->pixcfg.height);
    int stride = cairo_image_surface_get_stride(surface) / sizeof(uint32_t);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int y;
    for(y=0;y<pb->pixcfg.height;++y){
#if 1
        memcpy(dst, src, pb->pixcfg.width * 4);
        src += pb->pixcfg.width * 4;
        dst += stride;
#else
        int x;
        for(x=0;x<pb->pixcfg.width;++x){
            int a = src[3];
            int r = a * src[0] / 255;
            int g = a * src[1] / 255;
            int b = a * src[2] / 255;
            *dst++ = r << 16 | g << 8 | b << 0 | a << 24;
            src += 4;
        }
#endif
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}
#endif


static int
_qoir_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    const uint8_t *src = (uint8_t*)data;
    const uint8_t *end = src + len;
    //decode_result_t result = {0};
    context_t c;
    pixel_buffer_t dst;
    int seen_header = 0;
    int seen_body = 0;

    for(;;){
        uint32_t chunk_type;
        uint64_t chunk_size;
        if(src + 12 > end)
            goto fail_invalid_data;

        chunk_type = _peek_u32le(src + 0);
        chunk_size = _peek_u64le(src + 4);
        src += 12;
        if(src + chunk_size > end || chunk_size > 0x7FFFFFFFFFFFFFFFull)
            goto fail_invalid_data;

        switch(chunk_type){
        case ID_QOIR:
            if(seen_header)
                goto fail_invalid_data;
            if(_handle_qoir(src, chunk_size, &c))
                goto fail_invalid_data;
            switch(c.format){
            case PIXEL_FORMAT__BGRX:
            case PIXEL_FORMAT__BGRA_NONPREMUL:
            case PIXEL_FORMAT__BGRA_PREMUL:
                break;
            default:
                goto fail_invalid_data;
            }
            h->info->width = c.width;
            h->info->height = c.height;
            seen_header = 1;
            break;
        case ID_QEND:
            if ((chunk_size != 0) || (src != end))
                goto fail_invalid_data;
             goto out;
        case ID_QPIX:
            if(!seen_header || seen_body)
                goto fail_invalid_data;
            if(_pixel_format_has_alpha(c.format)){
                h->surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, c.width, c.height);
                dst.format = PIXEL_FORMAT_CAIRO_ARGB32;
            }else{
                h->surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, c.width, c.height);
                dst.format = PIXEL_FORMAT_CAIRO_RGB24;
            }
            dst.width = c.width;
            dst.height = c.height;
            dst.data = cairo_image_surface_get_data(h->surface);
            dst.stride = cairo_image_surface_get_stride(h->surface);;
            if(_handle_qpix(src, chunk_size, &c, &dst))
                goto fail_invalid_data;
            seen_body = 1;
            break;
#ifdef ENABLE_METADATA
        case ID_CICP:
            if (result.metadata_cicp_ptr)
                goto fail_invalid_data;
            result.metadata_cicp_ptr = src;
            result.metadata_cicp_len = chunk_size;
            break;
        case ID_ICCP:
            if (result.metadata_iccp_ptr)
                goto fail_invalid_data;
            result.metadata_iccp_ptr = src;
            result.metadata_iccp_len = chunk_size;
            break;
        case ID_EXIF:
            if (result.metadata_exif_ptr)
                goto fail_invalid_data;
            result.metadata_exif_ptr = src;
            result.metadata_exif_len = chunk_size;
            break;
        case ID_XMP:
            if (result.metadata_xmp_ptr)
                goto fail_invalid_data;
            result.metadata_xmp_ptr = src;
            result.metadata_xmp_len = chunk_size;
            break;
#endif
        }
        src += chunk_size;
    }

out:
    if (!seen_body)
        goto fail_invalid_data;
    return 0;

fail_invalid_data:
    //return _make_decode_result_error(qoir_status_message__error_invalid_data);
    return -1;
}


static cairo_surface_t *
_qoir_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _qoir_get_mime_types,
        .new = _qoir_new,
        .free = _qoir_free,
        .create_from_data = _qoir_create_from_data,
        .get_image_surface = _qoir_get_image_surface
    };
    return &p;
}
